/* -*- tab-width:4 -*- */
/*
Part of the Epsilon Programmer's Editor by Lugaru Software, Ltd.
www.lugaru.com
Copyright 1999, 2017 Lugaru Software, Ltd.  All Rights Reserved

nsshelp is a helper program which Epsilon must use on systems
lacking the shared NSS libraries it normally requires.  These
libraries let Epsilon look up host names and get information on
users and groups.

With a command-line parameter (a host name) it prints the
corresponding IP address and exits with code 0.  If the host isn't
found or isn't valid, it exits with code 1.

Without a command-line parameter, it begins looping as a filter,
waiting for requests from Epsilon on stdin and responding on stdout.
Each request is one line, and specified by the first character on
the line.  Each response is also one line, except for the list of all
users.

A u request is followed by a user name; nsshelp responds with info on
that user in the format of /etc/passwd, delimiting with : characters.
It responds with a blank line if the user is unknown.  A U request is
similar, but Epsilon sends a user number, not a name.

A g request is followed by a group name; nsshelp responds with info on
that group in the format of /etc/group, delimiting with : characters.
(It omits the group's members for simplicity.)  A G request is similar,
but Epsilon sends a group number, not a name.

An a request asks for a list of all users.  Nsshelp responds with a
series of lines in /etc/passwd format, followed by a blank line.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <pwd.h>
#include <sys/types.h>
#include <grp.h>
#include <netdb.h>

#define _XOPEN_SOURCE 700
#include <unistd.h>						// For gethostname()

void dump_user(p)
struct passwd *p;
{
	if (p)
		printf("%s:%s:%d:%d:%s:%s:%s",
			   p->pw_name, p->pw_passwd, p->pw_uid, p->pw_gid,
			   p->pw_gecos, p->pw_dir, p->pw_shell);
	printf("\n");
}

void dump_group(p)
struct group *p;
{
	if (p)
		printf("%s:%s:%d", p->gr_name, p->gr_passwd, p->gr_gid);
	printf("\n");
}

void dump_all_users()
{
	struct passwd *p;

	setpwent();
	while ((p = getpwent()))
		dump_user(p);
	printf("\n");
}

int main(argc, argv)
int argc;
char *argv[];
{
	char buf[1000];

	setvbuf(stdin, NULL, _IOLBF, 0);
	setvbuf(stdout, NULL, _IOLBF, 0);

	if (argc > 1) {		/* Report on one host name. */
		struct hostent *he;

		if (!strcmp(argv[1], "-current-host")) {
			if (!gethostname(buf, sizeof(buf)) && (he = gethostbyname(buf)))
				printf("%s\n", he->h_name);
			else
				printf("?\n");
			exit(0);		/* Print the IP address in hex. */
		}
		he = gethostbyname(argv[1]);
		if (!he)
			exit(1);	/* An unknown host. */
		printf("%x\n", (unsigned int) *(uint32_t *)he->h_addr);
		exit(0);		/* Print the IP address in hex. */
	}

	while (fgets(buf, sizeof(buf), stdin)) {
		if (strchr(buf, '\n'))		 /* Strip the final \n if any. */
			*strchr(buf, '\n') = 0;
		switch (buf[0]) {
			case 'u':
				dump_user(getpwnam(buf + 1));
				break;
			case 'U':
				dump_user(getpwuid(atoi(buf + 1)));
				break;
			case 'g':
				dump_group(getgrnam(buf + 1));
				break;
			case 'G':
				dump_group(getgrgid(atoi(buf + 1)));
				break;
			case 'a':
				dump_all_users();
				break;
			default:	/* Unknown request. */
				exit(1);
		}
	}
	exit(0);
}
