/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1985, 2020 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// This file has basic commands for moving around and inserting
// characters, and some handy general-purpose subroutines.

#include "eel.h"

/* insert the following extended ascii character into the buffer */

command quoted_insert() on reg_tab[CTRL('Q')]
{
	mention("Quote: ");
	do {
		getkey();
		if (IS_WIN_KEY(key)) {
			iter = 0;
			say("");
			return;
		}
	} while (!IS_ASCII_KEY(key & ~KEY_ALT));
	if (key == CTRL('Q') && ungot_key != -1) {	// auto-quoted
		key = ungot_key;	// Strip extra auto-quoting.
		ungot_key = -1;
	}
	say("");
	if (user_abort == 2)
		check_abort();
	user_abort = 0;
	if (!IS_ASCII_KEY(full_key))
		full_key = generic_key(full_key);
	if (IS_ALT_KEY(full_key) && KEY_PLAIN(full_key) < 128)
		full_key |= 128;			// turn on meta bit
	delete_if_highlighted();
	while (iter-- > 0)
		insert(full_key);
}

/* toggle all self-inserting characters with self-overwrite */
command overwrite_mode() on reg_tab[KEYINSERT]
{
	over_mode = (has_arg? (iter != 0) : !over_mode);
	make_mode();
	iter = 1;
}

do_reindent_previous_line()
{
	int i = get_mode_variable("%s-reindent-previous-line");

	if (got_bad_number)
		i = default_reindent_previous_line;
	if (!i)
		return;
	save_spot point;
	nl_reverse();
	to_indentation();
	save_var indenter_action = IACT_REINDENT_PREV;
	if (curchar() != '\n')
		(*indenter)();
}

/* Delete trailing whitespace, then break line here and indent it. */
break_and_indent_line()
{
	delete(point, point - parse_string(-1, "[ \t]*"));
	insert('\n');
	do_reindent_previous_line();
	save_var indenter_action = IACT_AUTO_INDENT;
	point += parse_string(1, "[ \t]*");
	(*indenter)();
}

/* insert or overwrite according to mode */
command normal_character() on reg_tab['\n']		/* more bindings, see below */
{
	int i;

	check_abort();
	delete_if_highlighted();
	if (!call_maybe_break_line(3))
		return;
	if (IS_ALT_KEY(key) && KEY_PLAIN(key) < 128)
		key |= 128;			// turn on meta bit
	if (!IS_ASCII_KEY(full_key))
		if (IS_ASCII_KEY(key))
			full_key = key;
		else if (IS_ASCII_KEY(KEY_PLAIN(key)))
			full_key = KEY_PLAIN(key);
	if (full_key == '\n') {
		if (over_mode) {
			if (!nl_forward())
				insert('\n');
		} else if (auto_indent && indenter)
			break_and_indent_line();
		else
			insert('\n');
	} else if (over_mode && point < size() && curchar() != '\n') {
		if (curchar() == '\t')
			i = display_width('\t', current_column());
		else
			i = 0;
		replace(point++, full_key);
		while (--i > 0)		/* insert i-1 spaces after pt */
			insert(' '), point--;
	} else
		insert(full_key);
}

when_loading()		/* put it on the other keys, too */
{
	set_range(reg_tab, ' ', (short) normal_character, '~' - ' ' + 1);
	set_range(reg_tab, 128, (short) normal_character, MAX_CHAR - 128 + 1);
}

// If there's a highlighted region, maybe delete it so we can replace it
// with new chars.  Return 1 if we did.
delete_if_highlighted()
{
	if (typing_deletes_highlight && is_highlight_on()
		  && (typing_deletes_highlight != 2 || _doing_input)) {
		delete_region();
		return 1;
	}
	return 0;
}

/* Go to the beginning of the current line */
command beginning_of_line() on reg_tab[CTRL('A')], reg_tab[NUMALT(KEYLEFT)],
							   reg_tab[NUMALT(NUMDIGIT(4))]
{
	iter=0;				/* throw away any argument */
	do_shift_selects();
	cursor_to_column = -1;
	to_begin_line();
}

/* Go to the end of the current line */
command end_of_line() on reg_tab[CTRL('E')], reg_tab[NUMALT(KEYRIGHT)],
						 reg_tab[NUMALT(NUMDIGIT(6))]
{
	iter=0;			/* throw away any argument */
	do_shift_selects();
	cursor_to_column = -1;
	to_end_line();
}

give_begin_line()	/* return the position of this line's start */
{
	save_var point;
	to_begin_line();
	return point;
}

give_end_line()		/* return the position of this line's end */
{
	save_var point;
	to_end_line();
	return point;
}

command down_line() on reg_tab[CTRL('N')], reg_tab[KEYDOWN]
{
	int true;

	down_or_up_lines(1);
	if (in_macro() && window_extra_lines())		// Don't scroll unnecessarily.
		return;
	true = point;
	while (scroll_at_end > 0 && true > window_end) {
		window_scroll(scroll_at_end);
		if (window_end + 1 >= size() - narrow_end || in_macro())
			break;		/* make sure we stop at end */
		point = window_end + 1;	/* compute window_end explicitely */
		window_end = next_screen_line(scroll_at_end) - 1;
		check_abort();
	}
	point = true;
}

command up_line() on reg_tab[CTRL('P')], reg_tab[KEYUP]
{
	int max = iter + 5;

	down_or_up_lines(0);
	while (scroll_at_end > 0 && point < window_start) {
		window_scroll(-scroll_at_end);
		check_abort();
		if (max-- < 0)
			break;
	}
}

virtual_column()
{
	return (cursor_to_column >= 0) ? cursor_to_column : current_column();
}

virtual_mark_column()
{
	return (mark_to_column >= 0) ? mark_to_column : get_column(mark);
}

down_or_up_lines(down_lines)
{
	int hit_end = 0, wantcol;

	do_shift_selects();
	if (iter < 0)
		iter = -iter, down_lines = !down_lines;
	if (prev_cmd != SAMECOL)
		_preserve_column = virtual_space ? virtual_column()
					: current_column();
	while (iter-- > 0)
		if (!down_lines)
			nl_reverse();
		else if (!nl_forward()) {
			hit_end = 1;
			break;
		}
	if (is_highlight_on())
		save_var goal_column = -1;
	wantcol = (goal_column >= 0) ? goal_column : _preserve_column;
	if (virtual_space)
		to_virtual_column(wantcol);
	else if (!hit_end || goal_column >= 0)	/* stay at end if we tried */
		move_to_column(wantcol);	/* to move past it */
	this_cmd = SAMECOL;
	iter = 0;
}

to_virtual_column(col)
{
	move_to_column(col);
	if (current_column() != col) {
		cursor_to_column = col;
		call_on_modify = 1;
	} else
		cursor_to_column = -1;
}

/* Set point to the previous character in the buffer */
command backward_character() on reg_tab[CTRL('B')], reg_tab[KEYLEFT]
{
	do_shift_selects();
	if (virtual_space < 2) {
		point -= iter;
		iter = 0;
	} else
		to_virtual_column(virtual_column() - 1);
}


/* Set the point to the next character in the buffer */
command forward_character() on reg_tab[CTRL('F')], reg_tab[KEYRIGHT]
{
	do_shift_selects();
	if (virtual_space < 2) {
		point += iter;
		iter = 0;
	} else
		to_virtual_column(virtual_column() + 1);
}

// Return nonzero if the shift key's down.  Handles keyboard macro
// case too; records shift state if recording one, and reads shift state
// if not.

was_key_shifted()
{
	int shifted;

	shifted = (full_key & KEY_SHIFT) != 0;
	if (shifted && _len_def_mac > 0 && !recording_suspended
		&& key == def_text[_len_def_mac - 1])	// And record in macro.
		def_text[_len_def_mac - 1] |= EXTEND_SEL_KEY;
	return shifted;
}

// Called by movement commands that want to use Windows-style arrow key
// selection rules.  Starts or continues defining a region, or turns
// off highlighting, as appropriate.

do_shift_selects()
{
	int arrow = (NUMPLAIN(key) >= KEYEND && NUMPLAIN(key) <= KEYPGUP)
				|| (NUMPLAIN(key) >= NUMDIGIT(0) && NUMPLAIN(key) <= NUMDIGIT(9));
	if (!shift_selects || IS_ASCII_KEY(key) || !arrow) {	// only affects cursor keys or numpad keys
		if (shift_selecting == 2)		// except at prompts
			highlight_off();
		return;
	}
	if (was_key_shifted()) {
		if (!is_highlight_on())		// if shift key is down, begin/extend
			do_set_mark(point);		// selection
		set_region_type(REGNORM);
		highlight_on();
		shift_selecting = 1;
	} else if (arrow && shift_selecting)	// if shift was just removed, remove highlight
		highlight_off();
}

command open_line() on reg_tab[CTRL('O')]
{
	insert('\n');
	point--;
}

command goto_beginning() on reg_tab[ALT('<')], reg_tab[NUMCTRL(KEYHOME)]
{
	do_shift_selects();
	point = 0;
}

command goto_end() on reg_tab[ALT('>')], reg_tab[NUMCTRL(KEYEND)]
{
	do_shift_selects();
	point = size();
}

go_line(line)
{
	point = 0;
	if (line > 0)
		move_by_lines(line - 1);
}

buf_go_line(int buf, int line)
{
	save_var bufnum = buf;
	go_line(line);
}

// Prompt for a line number and go to that line.  Also recognizes:

// 40%     Go to a line 40% of the way into the buffer.
// :20     Go to column 20 on this line.
// 10:20   Go to column 20 on line 10.
// 40%:20  Go to column 20 on a line 40% of the way into the buffer.

// With a numeric argument, go to the indicated line number without prompting.

command goto_line() on cx_tab['g']
{
	char res[FNAMELEN], *p;
	int num = iter, percent = 0;

	got_bad_number = iter = 0;
	if (has_arg) {
		go_line(num);
		return;
	}
	get_string(res, "Go to line: ");
	if (tolower(*res) == 'p') {
		num = exptoi(res + 1);
		if (got_bad_number)
			error("Bad position %s.", res + 1);
		point = num;
		return;
	}
	if (res[0] && *(p = res + strlen(res) - 1) == '%') {
		percent = 1;
		*p = 0;
	}
	num = exptoi(res);
	if (!got_bad_number) {
		if (percent) {
			point = muldiv(size(), num, 100);	// handles intermediates > 32 bits
			if (num)
				nl_forward();
		} else
			go_line(num);
	}
	p = strchr(res, ':');		// a number after : or . is a column
	if (!p)
		p = strchr(res, '.');
	if (p) {
		num = exptoi(p + 1);
		if (!got_bad_number)
			to_column_and_center(num);
	}
}

int isalnum(int c)
{
	return isalpha(c) || isdigit(c);
}

int isident(int c)
{
	return isalpha(c) || isdigit(c) || c == '_';
}

int get_key_response(char *pr, char *valid, int def, char *helpcmd)
{
	do {
		noteput("%s", pr);
		getkey();
		check_abort();
		if (IS_ENTER_KEY(key) && def > 0)
			key = def;
		if (IS_HELP_KEY(key) || key == '?')
			if (helpcmd)
				help_on_command(helpcmd);
	} while (!IS_ASCII_KEY(key) || !strchr(valid, toupper(key)));
	say("");
	return toupper(key);
}

full_getkey(char *prompt, int code)
{			/* get a key using ^Q, ^[, and ^^ prefixes */
	char msg[80];
	int allcode = 0, refreshed = 0;

	strcpy(msg, prompt);
	for (;;) {
		switch (code) {
			case ALTIFY_KEY:  strcat(msg, "<Alt> "); break;
			case CTRLIFY_KEY: strcat(msg, "<Ctrl> "); break;
		}
		allcode |= code, code = 0;
		for (;;) {
			mention("%s", msg);
			getkey();
			if (key != MOUSE_MOVE && !IS_WIN_KEY(key))
				break;
			if (!refreshed)
				refresh();
			refreshed = 1;
		}
		if (allcode & ALTIFY_KEY)
			key = generic_key(full_key = make_alt(full_key));
		if (allcode & CTRLIFY_KEY)
			key = generic_key(full_key = make_ctrl(full_key));
		if (root_keys[key] == (short) alt_prefix)
			code = ALTIFY_KEY;
		else if (root_keys[key] == (short) ctrl_prefix)
			code = CTRLIFY_KEY;
		else
			return;
	}
}

ask_yn(char *title, char *question, char *yes_button,
	   char *no_button, int def_button)
{
	if (is_gui && want_gui_prompts) {
		show_text(-2, 1, "%s", question);
		return button_dialog(title, question, yes_button, no_button, "",
							 def_button) == 1;
	} else {
		if (windows_state() & WINSTATE_MINIMIZED)
			windows_restore();
		for (;;) {
			noteput("%s%s", question,
					def_button == 1 ? "[Y] " : def_button == 2 ? "[N] " : "");
			getkey();
			if (user_abort)
				return 0;
			switch (toupper(key)) {
				case 'Y': say(""); return 1;
				case 'N': say(""); return 0;
			}
			if (IS_ENTER_KEY(key) && def_button) {
				say("");
				return def_button == 1;
			}
			refresh();
		}
	}
}

// Ask a 3-way question.  The first letter of each answer is its
// shortcut key; put & before another letter to use it instead.  Title
// and question are for GUI, prompt is for others.  Return 1, 2, 3, or 0
// if aborted.
ask_3way(char *title, char *question, char *prompt,
		 char *a1, char *a2, char *a3, int def)
{
	char msg[FNAMELEN], keys[10];
	int res;

	keys[0] = strchr(a1, '&') ? strchr(a1, '&')[1] : a1[0];
	keys[1] = strchr(a2, '&') ? strchr(a2, '&')[1] : a2[0];
	keys[2] = strchr(a3, '&') ? strchr(a3, '&')[1] : a3[0];
	keys[3] = 0;
	if (is_gui && want_gui_prompts) {
		show_text(-2, 1, "%s", prompt);
		res = button_dialog(title, question, a1, a2, a3, def);
		if (user_abort)
			res = 0;
		user_abort = 0;
	} else {
		*msg = 0;
		if (def >= 1 && def <= 3)
			sprintf(msg, "[%c] ", keys[def - 1]);
		for (;;) {
			noteput("%s%s", prompt, msg);
			getkey();
			if (user_abort) {
				user_abort = 0;
				return 0;
			}
			if (strchr(keys, toupper(key))) {
				say("");
				return strchr(keys, toupper(key)) + 1 - keys;
			}
			if (IS_ENTER_KEY(key) && def) {
				say("");
				return def;
			}
			refresh();
		}
	}
	return res;
}

charname_lister(char *s, char *(*finder)(), int flags, int mbuf)
{
	int i, len = strlen(s);
	char buf[FNAMELEN];

	UNREFERENCED_PARAMETER(flags);
	UNREFERENCED_PARAMETER(finder);
	if (strchr(s, '>'))
		len = strchr(s, '>') - s + 1;
	buf_zap(mbuf);
	save_var case_fold = 1;
	if (!*s || *s == '<') {
		note("Listing matches...");
		for (i = 0; i < MAX_CHAR; i++) {
			*buf = 0;
			show_char(buf, i, 3);
			if (buf[0] == '<' && buf[1] && buf[1] != '#'
				  && !strnfcmp(s, buf, len))
				buf_printf(mbuf, "%s  U+%04x (%c)\n", buf, i, i);
		}
		note("");
	}
	point = 0;
}

char *charname_matcher(s, flags)
char *s;
{
	if (!*s)
		return 0;
	if (flags & STARTMATCH)
		charname_lister(s, charname_matcher, flags, zap(_MATCH_BUF));
	return general_matcher(s, flags);
}

command insert_ascii() on reg_tab[ALT('#')] /* insert character by number */
{
	int res;
	char buf[FNAMELEN];

	if (run_by_mouse)
		has_arg = 0;
	save_var completion_lister = charname_lister;
	save_var menu_width = MAX(menu_width, 70);
	comp_read(buf, "Insert character with code: ",
			  charname_matcher, COMP_FOLD | SPACE_VALID, "");
	res = exptoi(buf);
	if (got_bad_number || res < 0 || res > MAX_CHAR)
		error("Invalid code %s.", buf);
	insert(res);
}

/* Prompt for a key, filling in name, and returning a pointer for
 * binding.  Flags: ask_if_ambiguous=1: ask for clarification if the resulting
 * key is a variation of a generic key (e.g. Ctrl-H in general, or the
 * Backspace key in particular); otherwise assume the more specific key
 * is meant (e.g. Backspace).  2=warn_if_std: ask for confirmation
 * before rebinding if the key isn't often rebound. */

short *ask_key(char *pr, char *keyname, int flags)
{
	int index, k, fk, ask_if_ambiguous = flags & 1, warn_if_std = flags & 2;
	short *tab, *alttab;
	char msg[FNAMELEN], *p;

	tab = mode_keys;
	alttab = root_keys;
	*keyname = 0;			// put readable form of key here
	key_binding[0] = 1;		// also save sequence in this global
	for (;;) {
		sprintf(msg, "%s%s", pr, keyname);
		full_getkey(msg, 0);
		if (root_keys[key] == (short) quoted_insert) {
			mention("%s<Quote> ", msg);
			getkey();
			user_abort = 0;
		}
		check_abort();
		p = keyname + strlen(keyname);
		k = full_key;
		index = (tab[k] == 0) ? alttab[k] : tab[k];
		if (index <= 0) {
			k = key;
			index = (tab[k] == 0) ? alttab[k] : tab[k];
		}
		show_char(keyname, ask_if_ambiguous ? k : full_key, 1);
		key_binding[key_binding[0]] = k; // add to constructed key
		key_binding[0]++;
		if (index <= 0 || name_type(index) != NT_TABLE)
			break;
		alttab = index_table((tab[k] && !alttab[k]) ?
							 tab[k] : alttab[k]);
		tab = index_table(index);
		if (!alttab)
			alttab = tab;
		strcat(keyname, " ");
		warn_if_std = 0;
	}
	say("");
	if (warn_if_std && (key >= ' ' && key <= 65535 || key == '\r' || key == '\n')) {
		sprintf(msg, "Are you sure you want to rebind the standard key %k? ", key);
		if (!ask_yn("Confirm rebinding a standard key", msg, "Yes", "No", 0))
			quick_abort();
	}
	if (ask_if_ambiguous && full_key != key) {
		k = key;
		fk = full_key;
		sprintf(msg, "Bind only specific key %k (not general key %k)? ",
				fk, k);
		if (ask_yn("Specific key or general?", msg, "Yes", "No", 2)) {
			key = k = fk;
			*p = 0;
			show_char(keyname, k, 1);
		}
	} else if (full_key != key && k == full_key)
		key = k;
	return ((tab[k] == 0) ? alttab : tab) + k;
}

char *temp_buf()	/* create a unique temporary buffer, return name */
{
	return bufnum_to_name(tmp_buf());
}

tmp_buf()	/* create a unique temporary buffer, return number */
{
	char bname[80];
	int uniq = 1;

	do {
		sprintf(bname, "-temp%d", uniq++);
	} while (exist(bname));
	return create(bname);
}

int expand_string_template(char *dest, char *template, char *keys, char *vals[])
{
	int err = 0;
	char *p;

	for (; *template; )
		if (*template != '%')
			*dest++ = *template++;
		else if (!*++template)
			err |= 2;
		else if (tolower(*template) == 'x') {
			get_executable_directory(dest);
			if (*template++ == 'X')
				convert_to_8_3_filename(dest);
			dest += strlen(dest);
		} else if (!(p = strchr(keys, *template++)))
			err |= 1;
		else {
			strcpy(dest, vals[p - keys]);
			dest += strlen(dest);
		}
	*dest = 0;
	return err;
}

int strtoi(char *s, int base)
{	/* convert a string of digits to a number */
	char *p, *digits = "0123456789ABCDEF";
	int number, neg = 0;

	got_bad_number = 0;
	while (*s == '-')
		s++, neg = !neg;
	number = 0;
	if (!*s || !(p = index(digits, toupper(*s))) || p - digits >= base)
		got_bad_number = 1;
	while (*s && (p = index(digits, toupper(*s++))) && p - digits < base)
		number = number * base + (p - digits);
	return neg ? -number : number;
}

/* Return value of a number in the string s. Interpret prefixes 0x for
 * hex, 0o for octal, and 0b for binary, plus character codes like '\n' or
 * 'G' or <THAI CHARACTER RU>. If no base, assume base 10. */
int numtoi(char *s)
{
	return numtoi_with_base(s, 10);
}

/* Return value of a number in the string s. Interpret prefixes 0x for
 * hex, 0o for octal, and 0b for binary, plus character codes like '\n' or
 * 'G' or <THAI CHARACTER RU>. If no base is given in the string,
 * assume the specified base. */
int numtoi_with_base(char *s, int base)
{
	int neg = 0, i;
	char *from = "nbtrf", *to = "\n\b\t\r\f", *p;

	got_bad_number = 0;
	while (isspace(*s))
		s++;
	if (*s == '\'') {		/* allow 'x' as a number (give ascii code) */
		if (s[1] != '\\' && s[2] == '\'')
			return s[1];
		if (s[1] == '\\' && s[2] && s[3] == '\'') {
			p = strchr(from, s[2]);
			if (p)
				return to[p - from];
			return s[2];
		}
		got_bad_number = 1;
		return s[1];
	}
	if (*s == '<') {
		i = key_value(s);
		if (i >= 0)
			return i;
		got_bad_number = 1;
		return 0;
	}
	while (*s == '-')
		s++, neg = !neg;
	if (s[0] == '0' && tolower(s[1]) == 'x')
		i = strtoi(s + 2, 16);
	else if (s[0] == '0' && tolower(s[1]) == 'o')
		i = strtoi(s + 2, 8);
	else if (s[0] == '0' && tolower(s[1]) == 'b')
		i = strtoi(s + 2, 2);
	else
		i = strtoi(s, base);
	return neg ? -i : i;
}

// Return value of expression s.  Only compile it if we have to, for
// speed.  If expression starts with <, it's a key name numtoi
// recognizes specially, so don't compile it.  Otherwise, compile it if
// it has an operator, or doesn't look like a number.  Exception: numtoi
// handles an initial -.
int exptoi(char *s)
{
	save_var use_default = 0;	// Don't run compile with use_default 1.
	while (isspace(*s))
		s++;
	if (!strchr("<", *s) && (strpbrk(s, "+/*^|&<>%=")
							  || *s && strpbrk(s + 1, "-") || isalpha(*s))
		  && (has_feature & FEAT_EEL_COMPILE))
		return evaluate_numeric_expression(s);
	return numtoi(s);
}
