/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*	Copyright (C) 2003, 2020 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// A context-sensitive help command.

#include "eel.h"
#include "colcode.h"
#include "lowlevel.h"
#include "chelp.h"

// An override function for a mode lets us use a different context for
// various parts of the same file (i.e. for scripting embedded in html)
// or for different types of files that share the same major mode.

// In C mode, which context should we use?
context_help_override_c(char *mode)
{
	if (c_extra_keywords == EEL_KEYWORDS)	// .h files have this bit too,
		strcpy(mode, "eel");		// but don't assume they're for EEL.
	else if (c_extra_keywords & JAVA_KEYWORDS)
		strcpy(mode, "java");
	else if (c_extra_keywords & JS_KEYWORDS)
		strcpy(mode, "jscript");
	else if (c_extra_keywords & OBJC_KEYWORDS)
		strcpy(mode, "objc");
	else
		strcpy(mode, "c");
}

// Put the name of the mode to use for context-sensitive help in mode.
get_context_name(char *context)
{
	char func[FNAMELEN], mode[FNAMELEN];
	int i;

	clean_mode(mode);
	sprintf(func, "context_help_override_%s", mode);
	if (i = find_index(func))		// The mode specified a different context
		(*(int (*)())i)(context);	// (for embedding etc.).
	else
		strcpy(context, mode);
}

// Load the current word into item, using pat as the word pattern.
context_help_get_item(char *item, char *pat)
{
	save_var point--;
	re_search(1, pat);
	re_search(-1, pat);
	grab(matchstart, matchend, item);
}

context_help_perldoc()
{
	char item[FNAMELEN], cmd[FNAMELEN];
	int orig = bufnum;

	context_help_get_item(item, "<word|:>+");
	if (!strcmp(item, "use")) {
		save_var point;
		point += parse_string(1, "<word|:>*<space>*");
		context_help_get_item(item, "<word|:>+");
	}
	sprintf(cmd, "-f %s", item);
	make_man_page(cmd, 1);
	if (!strncmp(bufname, "man--f ", 7) && lines_between(0, size()) < 3) {
		int b = bufnum;
		to_buffer_num(orig);
		buf_delete(b);
		make_man_page(item, 1);
	}
}

context_help_man()
{
	char item[FNAMELEN];

	context_help_get_item(item, word_pattern);
	make_man_page(item, 0);
}

// If this info file has an entry for item, go there and return 1.  If
// there's no such info file or it lacks such an entry, return 0.
int check_context_help_info(char *item, char *file)
{
	char node[FNAMELEN];

	sprintf(node, "(%s)", file);
	return *file && info_go_node(node, 3) && do_info_index(item, 0) > 0;
}

// Look up word in this comma-separated list of info files.
do_context_help_info(char *file)
{
	char f[FNAMELEN], item[FNAMELEN], *s, *t;

	context_help_get_item(item, word_pattern);
	for (s = file; t = strchr(s, ','); s = t + 1) {
		sprintf(f, "%.*s", t - s, s);
		if (check_context_help_info(item, f))
			return 1;
	}
	if (!check_context_help_info(item, s)) {
		say("%s not found in index of Info file %s.", item, file);
		sprintf(f, "(%s)", file);
		if (!info_go_node(f, 3))
			info_go_node("(dir)", 0);
	}
}

// Run this program on the current word.
do_context_help_run(char *prog)
{
	char item[FNAMELEN], cmdline[FNAMELEN], buf[FNAMELEN];
	char winprog[FNAMELEN], fixprog[FNAMELEN], *s;
	int sync = 1;

	context_help_get_item(item, word_pattern);
	strcpy(fixprog, prog);
	if (*fixprog == '"' && (s = strchr(fixprog + 1, '"'))) {
		strncpy(winprog, fixprog + 1, s - fixprog - 1);
		convert_to_8_3_filename(winprog);
		strcat(winprog, s + 1);
		strcpy(fixprog, winprog);
	}
	s = fixprog + strlen(fixprog) - 1;
	if (strlen(fixprog) > 1 && *s == '&') {
		*s = 0;
		sync = 0;
	}
	if (s = strchr(fixprog, '*')) {
		strncpy(cmdline, fixprog, s - fixprog);
		quote_for_shell(cmdline + strlen(cmdline), item);
		strcat(cmdline, s + 1);
	} else {
		sprintf(cmdline, "%s ", fixprog);
		quote_for_shell(cmdline + strlen(cmdline), item);
	}
	sprintf(buf, "context-help-%s", item);
	if (opsys == OS_UNIX || !sync) {
		pipe_text(NULL, buf, cmdline, NULL, (sync ? PIPE_SYNCH : 0) |
				  PIPE_CLEAR_BUF | PIPE_NOREFRESH, 0);
		if (buffer_size(buf)) {
			to_buffer(buf);
			point = 0;
		}
	} else {
		strcpy(winprog, cmdline);
		if (index(winprog, ' '))
			*index(winprog, ' ') = 0;
		if (winexec(winprog, cmdline, SW_SHOWNORMAL, 0) == -1)
			say("Could not start %s.", winprog);
	}
}

// Copy src to dest, replacing the first instance of c with repl and
// return 1. If none, copy src to dest and return 0.
do_context_help_replace_in_string(char *dest, char *src, char c, char *repl)
{
	char *s;
	if (s = strchr(src, c)) {
		strncpy(dest, src, s - src);
		strcat(dest, repl);
		strcat(dest, s + 1);
		return 1;
	} 
	strcpy(dest, src);
	return 0;
}

// Display this url in a web browser.
do_context_help_url(char *url)
{
	char item[FNAMELEN], full_url[FNAMELEN];

	context_help_get_item(item, word_pattern);
	if (!do_context_help_replace_in_string(full_url, url, '*', item))
		strcat(full_url, item);
	show_url(full_url); 
}

do_context_help_site_search(char *site, char *site_alt)
{
	char replacement[FNAMELEN];
	do_context_help_replace_in_string(replacement, site_alt, '@', site);
	do_context_help_url(replacement);
}

do_context_help_winhelp(char *file)
{
	char item[FNAMELEN];

	context_help_get_item(item, word_pattern);
	if (!strncmp(file, "epsilon", 7) || !strncmp(file, "edoc", 4))
		win_help_string("", item);
	else
		win_help_string(file, item);
}

// Interpret this rule for context help.
use_context_help_rule(char *rule, char *var)
{
	switch (*rule) {
		case '>':	do_context_help_info(rule + 1); break;
		case '!':	do_context_help_run(rule + 1); break;
		case '=':	do_context_help_url(rule + 1); break;
		case '&':	do_context_help_site_search(rule + 1, context_help_site_search); break;
		case '@':	do_context_help_site_search(rule + 1, context_help_site_search_lucky); break;
		case '$':	if (!(has_feature & FEAT_WINHELP))
						error("In variable %s, Winhelp rule unavailable.", var);
					do_context_help_winhelp(rule + 1); break;
		case '+':	try_calling(rule + 1); break;
		default:	if (*rule)
						error("Unknown rule code %c in variable %s.",
							  *rule, var);
					else
						error("Set the %s variable to locate help sources.",
							  var);
	}
}

command context_help() on reg_tab[NUMSHIFT(FKEY(1))]
{
	char var[FNAMELEN], osvar[FNAMELEN], context[FNAMELEN], *s;
	int i;

	tiled_only();
	get_context_name(context);
	sprintf(var, "context_help_rule_%s", context);
	sprintf(osvar, "%s%s", var, is_unix ? "_unix" : "_windows");
	if ((i = find_index(osvar)) && get_str_var(i))
		strcpy(var, osvar);	// Select it even if it's empty.
	else if ((!(i = find_index(var)))
		  && *context_help_default_rule)
		strcpy(var, "context_help_default_rule");
	s = get_str_var(find_index(var));
	if (!s)
		error("No help is configured for %s context.", context);
	use_context_help_rule(s, var);
}
