/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1993, 2020 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

/* Integrated access to the clipboard. */

#include "eel.h"
#include "lowlevel.h"
#include "kill.h"

// Change \n to \r\n when copying to clipboard, \r\n to \n when retrieving?
clipboard_convert_newlines()
{
	return (GET_LINE_TRANSLATE(translation_type) != FILETYPE_BINARY
			&& opsys != OS_UNIX) ? CLIP_CONVERT_NEWLINES : 0;
}

char *clipboard_buf_name()
{
	if (opsys == OS_UNIX && clipboard_format == 1)
		return "-primary-selection-";
	return "-clipboard-";
}

command copy_to_clipboard()	// copy region to clipboard
{
	int err = 0, tmp = zap(clipboard_buf_name());
	save_spot point, mark;
	fix_region();
	if (region_type() == REGRECT)
		extract_rectangle(tmp, 0);
	else
		buf_xfer(tmp, point, mark);
	if (!buffer_to_clipboard(tmp, clipboard_convert_newlines(), clipboard_format))
		err = errno;
	zap(clipboard_buf_name());	// Load it next time.
	if (err)
		file_error(err, "", "Couldn't copy to clipboard.");
	else {
		say("Selection copied to clipboard.");
		highlight_off();
	}
}

get_clipboard_fix_lines(int buf)
{
	if (GET_LINE_TRANSLATE(translation_type) == FILETYPE_BINARY)
		return;
	switch (clipboard_convert_mac_lines) {
		case 0:
			return;
		case 1:
			break;
		default:
			if (is_unix_flavor != IS_UNIX_MACOS)
				return;
	}
	save_var bufnum = buf, point = 0;	// If \r's but no \n's, convert.
	if (!nl_forward() && re_search(-1, "\r"))
		simple_re_replace(0, "\r", "\n");
}

command insert_clipboard()
{
	if (buffer_flags(0) & 1)	// Turn off converting if orig buffer's
		save_var clipboard_convert_unicode = 0;	// already Unicode.
	int buf = zap(clipboard_buf_name());
	if (clipboard_to_buffer(buf,
							clipboard_convert_newlines(), clipboard_format)) {
		get_clipboard_fix_lines(buf);
		mark = point;
		set_region_type(REGNORM);
		grab_buffer(buf);
		build_first = 1;		/* redisplay hint */
	} else
		file_error(errno, "", "Couldn't retrieve clipboard.");
}

int should_use_clipboard()
{
	switch (clipboard_access) {
		case 0:	return 0;
		case 2:	return 1;		// Always, even in a macro.
	}
	return !in_macro() && !key_from_macro;
}

save_to_clipboard()	// try to save current kill buffer on clipboard
{
	int strip = clipboard_convert_newlines();

	clipboard_translation_type = GET_LINE_TRANSLATE(translation_type) | FORCED_TRANS;
	if (should_use_clipboard() && clipboard_available()) {
		buffer_to_clipboard(name_to_bufnum(_cur_kill_buf), strip,
							clipboard_format);
		clipboard_to_buffer(zap(clipboard_buf_name()), strip,
							clipboard_format);
	}
}

get_clipboard()		// push clipboard text on kill ring
{
	int new, old;

	if (!should_use_clipboard() || !clipboard_available())
		return;
	if (buffer_flags(0) & 1)	// Turn off converting if orig buffer's
		save_var clipboard_convert_unicode = 0;	// already Unicode.
	new = zap("-new-clipboard-");
	int trans = GET_LINE_TRANSLATE(translation_type);
	if (clipboard_translation_type & FORCED_TRANS)
		trans = GET_LINE_TRANSLATE(clipboard_translation_type);
	int conv = (trans != FILETYPE_BINARY && opsys != OS_UNIX) ? CLIP_CONVERT_NEWLINES : 0;
	if (!clipboard_to_buffer(new, conv, clipboard_format))
		return;
	get_clipboard_fix_lines(new);
	old = create(clipboard_buf_name());
	if (!buffers_identical(old, new)) {
		save_var bufnum = new;
		check_num_kill_buffers();	// copy clipboard to kill ring
		push_kill();
		_kill_width[_num_kill] = REGION;
		xfer(_cur_kill_buf, 0, size());
		buf_zap(old);		// copy new text to old clipboard buf
		buf_xfer(old, 0, size());
	}
	buf_zap(new);
}

// Return the clipboard format code for HTML fragments.
int html_clipboard_format()
{
	return call_dll("USER32.DLL", "RegisterClipboardFormatW",
					"p", "S", "HTML Format");
}

// Epsilon's RGB encoding isn't the same as HTML's; convert.
// (In v14 it is, so this could go away.)
int color_to_html(int color)
{
	int r = GETRED(COLOR_STRIP_ATTR(color));
	int g = GETGREEN(COLOR_STRIP_ATTR(color));
	int b = GETBLUE(COLOR_STRIP_ATTR(color));
	return (r << 16) | (g << 8) | b;
}

// Encode the 8-bit characters of buffer buf into the specified Unicode
// encoding, preserving colors.
int color_perform_unicode_conversion(int buf, char *enc)
{
	int pos, sz, color;
	save_var bufnum = buf;
	save_var point = 0;			// Convert any non-ascii to UTF8.
	while (re_search(1, "<^ascii>")) {
		pos = point;
		sz = size();	// Do it one character at a time to preserve colors.
		color = get_character_color(point - 1);
		if (perform_unicode_conversion(bufnum, point - 1, point, 0, enc))
			error("Failed to convert to %s at %d in %s.", enc, pos, bufname);
		point = pos + size() - sz;
		set_character_color(pos - 1, point, color);
		check_abort();
	}
}

// Look for each use of open following close.  If it has the same value
// (in match #1) as the previous one, delete this open and close.
html_clean_repeated(char *open, char *close)
{
	char last[FNAMELEN], this[FNAMELEN];

	*last = 0;
	save_var point = 0;
	while (re_search(1, open)) {
		grab(find_group(1, 1), find_group(1, 0), this);
		int end = matchend;
		if (!strcmp(last, this) && matches_at(matchstart, -1, close))
			delete(matchend, end);
		strcpy(last, this);
	}
}

// Put an HTML representation of the region's text on the clipboard,
// preserving its syntax coloring.  For Windows, also put on a version
// using the special clipboard format defined for HTML fragments.

command copy_formatting_as_html()
{
	int buf = zap("#html-clipboard#");
	int color, colstart, colend, tab = tab_size;
	save_var point, mark;
	fix_region();
	buf_xfer_colors(buf, point, mark);		// Make a copy of the region.
	save_var bufnum = buf;
	tab_size = tab;
	tab_convert(0, size(), 0);				// Untabify it.
	color_perform_unicode_conversion(bufnum, "utf8");	// To UTF8.
	point = 0;						// Convert special chars to HTML.
	simple_re_replace(1, "%&", "&amp;");
	simple_re_replace(1, "%<", "&lt;");
	simple_re_replace(1, "%>", "&gt;");
	spot before = alloc_spot(), after = alloc_spot(1);
	on_exit free_spot(before), free_spot(after);
	while (point < size()) {	// Insert HTML for the colors.
		color = get_character_color(point, &colstart, &colend);
		if (color == -1)
			color = color_class text;
		*before = *after = colend;
		bprintf("<span style=\"background: #%06x\"><span style=\"color: #%06x\">",
				color_to_html(get_background_color(color)),
				color_to_html(get_foreground_color(color)));
		if (get_foreground_color(color) & EFONT_ITALIC)
			bprintf("<i>");
		if (get_foreground_color(color) & EFONT_BOLD)
			bprintf("<b>");
		if (get_foreground_color(color) & EFONT_UNDERLINED)
			bprintf("<u>");
		point = *before;
		if (get_foreground_color(color) & EFONT_UNDERLINED)
			bprintf("</u>");
		if (get_foreground_color(color) & EFONT_BOLD)
			bprintf("</b>");
		if (get_foreground_color(color) & EFONT_ITALIC)
			bprintf("</i>");
		bprintf("</span></span>");
		point = *after;
		check_abort();
	}				// Now simplify the HTML coloring just inserted.
	html_clean_repeated("%<span style=\"background: #(<word>+)\"%>", "%</span%>");
	html_clean_repeated("%<span style=\"color: #(<word>+)\"%>", "%</span%>");
	html_mode();
	point = size();
	bprintf("</pre>");
	point = 0;
	bprintf("<pre>");
	point = 0;
	mark = size();
	copy_to_clipboard();
	restore_vars();
	highlight_off();
	say("Selection copied to clipboard.");
	if (!is_mswindows)
		return;
	int frag = tmp_buf();	// For Win32 only, make an HTML Fragment version.
	save_var bufnum = frag;
	grab_buffer(buf);
	simple_re_replace(1, "\n", "\r\n");		// Data will be raw binary.
	translation_type = FILETYPE_BINARY;
	point = size() - 6;		// Before </pre>.
	bprintf("<!--EndFragment-->");
	point = size();
	bprintf("</body></html>");
	point = 0;
	bprintf("Version:0.9\r\nStartHTML:00000097\r\nEndHTML:%08d\r\n",
			size() + 135);		// Create char offsets.
	bprintf("StartFragment:00000135\r\nEndFragment:%08d\r\n",
			size() + 46);
	bprintf("<html><body>");
	point += 5;			// After <pre>.
	bprintf("<!--StartFragment -->");
	if (!buffer_to_clipboard(bufnum, CLIP_ADD_FORMAT, html_clipboard_format()))
		say("Error %d copying HTML fragment to clipboard.", errno);
	restore_vars();
	buf_delete(frag);
}

// Put the current buffer's file name on the clipboard.
command copy_file_name() on cc_tab[ALT('n')]
{
	char buf[FNAMELEN], *s = filename;

	if (is_dired_buf())
		dired_getname(s = buf, 1); 
	else if (!*s)
		s = bufname;
	copy_line_to_clipboard(s, 0);
} 

command copy_include_file_name() on cc_tab[ALT('i')]
{
	char *s, line[FNAMELEN], mode[FNAMELEN], name[FNAMELEN];
	char orig_dir[FNAMELEN], dir[FNAMELEN];
	int i;

	if (!*filename)
		error("This buffer has no file name.");
	clean_mode(mode);
	sprintf(name, "copy-include-file-name-%s", mode);
	i = find_index(name);
	switch (name_type(i)) {		// i has name table index of subr/var
		case NT_SUBR: case NT_COMMAND:
		case NT_AUTOSUBR: case NT_AUTOLOAD:
			(*(void (*)())i)();
			return;
		case NT_VAR: case NT_BUFVAR:
		case NT_WINVAR: case NT_BUILTVAR:
			strcpy(dir, filename);
			*get_tail(dir, 1) = 0;
			getcd(orig_dir);
			on_exit chdir(orig_dir);
			chdir(dir);
			s = get_str_var(i);
			if (s && *s) {		// Got a template back.
				build_filename(line, s, filename);
				copy_line_to_clipboard(line, 1);
				return;
			}
			break;
	}
	sprintf(line, "import %r", get_tail(filename, 0));	// Default.
	copy_line_to_clipboard(line, 1);
}

// Copy some text to the clipboard.  Flags: 1=add a newline after it,
// 2=don't display msg.
copy_line_to_clipboard(char *line, int flags)
{
	int tmp = tmp_buf();
	buf_printf(tmp, "%s%s", line, (flags & 1) ? "\n" : "");
	buffer_to_clipboard(tmp, CLIP_CONVERT_NEWLINES, 0);
	buf_delete(tmp);
	if (!(flags & 2))
		say("Copied to clipboard: %s", line);
}
