/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1985, 2020 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// This file interprets Epsilon's command line switches.

#include "eel.h"
#include "menu.h"
#include "session.h"

early_restoring()
{
	char fname[FNAMELEN];
	int i;

	after_loading(); /* call early, so we can display errors, etc. */
	putenv("EPSRUNS=Y");	/* for catching nested Epsilons */
	if (opsys == OS_WINDOWS)
		putenv("PKNOFASTCHAR=Y"); //make PKZIP, UNZIP output capturable
	char *s = getenv("PERLDB_OPTS"), *t = malloc((s ? strlen(s) : 0) + 30);
	sprintf(t, "PERLDB_OPTS=ReadLine=0 %s", s ? s : "");
	putenv(t);	// Make Perl debugger run in process buffer.
	root_keys = reg_tab;
		/* now set buffer-specific variables */
	word_pattern.default = default_word;
	_display_class.default = _std_disp_class;
	_echo_display_class = _echo_disp_class;
	i = *get_color_scheme_variable();
	if (name_type(i) == NT_COLSCHEME)
		selected_color_scheme = i;
	else
		get_standard_color_scheme();
	use_default = 1;		/* default is fundamental mode */
	mode_default_settings();
	use_default = 0;
	apply_defaults();	/* apply defaults to init buffer & window */
	do_set_show_graphic(get_show_graphic());

	force_save_as = FSA_NEWFILE;	// make startup buffer prompt
	strcpy(fname, "startup.txt");	// for filename
	absolute(fname);
	filename = fname;
	strcpy(_docname, "edoc");
	if (load_from_state)
		show_text(-2, 0, "Loaded from state file %s", state_file);
	else if (*state_file)
		show_text(-2, 0, "Loaded from state file %s (quickup)", state_file);
	else
		show_text(-2, 0, "Loaded from bytecode files");
	for (i = 0; original_argv(i); i++)
		show_text(-2, 0, "Argv[%d]=%s", i, original_argv(i));
	get_customization_directory(fname);
	show_text(-2, 0, "Customization directory is %s", fname);
	char *path = verenv("EPSPATH");
	show_text(-2, 0, "EPSPATH is %s", path ? path : "undefined");
}

int quickup_would_work()
{
	char def[FNAMELEN];
	if (!*state_file)
		return 0;
	strcpy(def, state_file);
	set_default_state_file_name(def);
	absolute(def);
	return get_file_read_only(state_file) != 1 && !strcmp(state_file, def);
}

offer_quickup()
{
	char cmdline[16384];
	int i;

	get_string(cmdline, "Run quickup now, then restart this instance [Y]: ");
	if (tolower(*cmdline) == 'n')
		return;
	if (shell_with_args(argv[0], "-quickup") != 42)
		error("Auto-quickup failed, please update manually.");
	*cmdline = 0;
	for (i = 1; original_argv(i); i++) {
		quote_for_shell(cmdline + strlen(cmdline), original_argv(i));
		if (original_argv(i + 1))
			strcat(cmdline, " ");
	}
	int err = shell_with_args_flags(original_argv(0), cmdline,
									SHELL_NO_SYNCH | SHELL_KEEP_ENV);
	if (err)
		file_error(err, original_argv(0),
				   "Couldn't restart Epsilon, please restart manually");
	else
		leave(0);
}

// To deal with change in version string format, process a version
// string into a standard format: strip colon and make part after b
// two-digit.
standardize_version_string(char *dest, char *src)
{
	char *s;
	strcpy(dest, src);
	while (s = strpbrk(dest, ": "))	// Strip all these chars.
		strcpy(s, s + 1);
	s = dest + strlen(dest) - 1;	// Last char.
	if (strchr("abcd.", s[-1])) {	// Second-to-last char.
		s[2] = 0;		// One digit, make two: Convert 13b4 to 13b04
		s[1] = s[0];
		s[0] = '0';
	}
	for (; *dest; dest++)
		*dest = tolower(*dest);
}

restore_check_version()
{
	char clean_exe_version[30], clean_eel_version[30];

	// To deal with change in version string format, process
	// versioned_file_string into a standard format: strip colon and
	// make part after b two-digit.
	standardize_version_string(clean_exe_version, versioned_file_string);
	standardize_version_string(clean_eel_version, eel_version);
	if (!load_from_state || strpbrk(clean_exe_version, "ad"))
		return;
	if (strcmp(clean_eel_version, clean_exe_version) >= 0)
		if (!strchr(clean_eel_version, 'b') || strchr(clean_exe_version, 'b'))
			return;
	// Each user has a private state file, so we can't update them
	// all automatically.
	save_var bufnum = zap(HELPBUF);
	bprintf("Your version %s state file requires updating.\n"
			"You need to start Epsilon with its -quickup flag.\n", 
			eel_version);
	bprintf("Until you do this, Epsilon will not work correctly.\n");
	view_help(" Update Needed! ", 0);
	if (quickup_would_work())
		offer_quickup();
}

// Flavor: 1 means do pre-einit -d flags, 2 means do all other -d flags
// plus -p and -l flags, 3 means do -r flags.
parse_command_line_flags(int flavor)
{
	int i, j;
	char *s, *flag, *arg, buf[1024];

	for (i = 1; i < argc; i++) {
		flag = argv[i];
		if (*flag != '-')
			continue;
		if (!strcmp(flag, "--"))
			return;
		if (!strcmp(flag, "-dir")) {
			i++;
			continue;
		}
		arg = (flag[1] && flag[2] || i + 1 >= argc) ? (flag + 2) : argv[++i];
		if (!strchr((flavor == 3 ? "r" : flavor == 2 ? "lpd"  : "d"), flag[1]))
			continue;
		switch (flag[1]) {
			case 'p':
				if (!flag[2])
					preserve_session_once = preserve_session ? 1 : 2;
				else if (isdigit(flag[2]))
					preserve_session_once = numtoi(flag + 2) + 1;
				else
					cmd_line_session_file = flag + 2;
				break;
			case 'l':
				load_commands(arg); break;
			case 'd':
			case 'r': 
				strcpy(buf, arg);
				s = index(buf, '=');
				if (!s)		// accept either VAR=VAL or VAR!VAL
					s = index(buf, '!');
				if (s)
					*s = 0;
				if (j = find_index(buf)) {
					if (flag[1] == 'd') {
						if (!s)
							bad_flag(flag);
						if ((flavor == 1) != !strcmp(buf, "einit-file-name"))
							continue;	// Do this var only if flavor is 1.
						now_parsing_cmds = 1;	// Don't record this.
						safe_set_var_val(j, s + 1, 1);
						now_parsing_cmds = 0;
					} else {
						if (s)
							(*(int (*)())j)(s + 1);
						else
							(*(int (*)())j)();
					}
					break;
				}
				if (flag[1] == 'd') {
					bprintf("No such variable %s.\n", buf);
					init_help_abort();
				} else
					say("No command %s, %s ignored.", buf, flag);
				break;
		}
	}
}

// Find Windows file names of the form z:file on the command line and
// make them absolute, so we don't lose the per-drive current directory
// info while switching directories later.
fix_drive_only_file_names(int i)
{
	char fname[FNAMELEN];
	int all_files = 0;
	for (; i < argc; i++) {
		if (argv[i][0] == '+') {
		} else if (argv[i][0] == '-' && !all_files) {
			if (argv[i][1] == '-')
				all_files = 1;
			else if (!strcmp(argv[i], "-dir")
					 || strchr("ldr", argv[i][1]) && !argv[i][2]
					 || strchr("f", argv[i][1]) && !argv[i][3])
				i++;
		} else if (isalpha(argv[i][0]) && argv[i][1] == ':' && argv[i][2]
				   && !is_path_separator(argv[i][2])) {
			strcpy(fname, argv[i]);
			convert_cygwin_filename(fname, 0);
			absolute(fname);
			argv[i] = strsave(fname);
		}
	}
}

when_restoring()
{
	int tflag = 0, i, splitmode = HORIZONTAL, have_files = 0;
	char *flag, *arg, init_dir[FNAMELEN], *dir = init_dir;

	getcd(init_dir);
	early_restoring();
	early_init();
	set_toolbar();

	for (i = 1; i < argc; i++) {
		if (argv[i][0] != '-') {
			have_files = 1;
			continue;
		}
		if (!strcmp(argv[i], "--"))
			break;
		if (!stricmp(argv[i], "-noinit")) {
			noinit_flag = 1;
			continue;
		} else if (!strcmp(argv[i], "-dir")) {
			if (++i < argc)
				dir = argv[i];
			continue;
		}
		flag = argv[i];
		switch (flag[1]) {
		case 'f':
			arg = (flag[2] && flag[3] || i + 1 >= argc) ? (flag + 3) : argv[++i];
			switch (flag[2]) {
			case 's':
				set_swapname(arg);
				break;
			case 'd':
				strcpy(_docname, arg);
				break;
			default:
				bad_flag(flag);
			}
			break;
		case 't':
			tflag = 1;
			break;
		case 'd':
		case 'r':
		case 'l': if (!flag[2] && i + 1 < argc)
					  i++;
				  break;
		case 'p': break;
		case 'h': case '?': usage(); break;
		case 'v':
			switch (flag[2]) {
				case 'v':
					splitmode = VERTICAL;
					break;
				default:
					bad_flag(flag);
			}
			break;
		case '-':
			if (!strcmp(flag + 2, "help"))
				usage();
			else
				bad_flag(flag);
			break;
		default:
			bad_flag(flag);
		}			/* i is 1st file argument */
	}
	if (i < argc)
		have_files = 1;
	term_init();
	try_calling("key-setup");
	middle_init();
	init_menu();
	term_cmd_line();
	delete_buffer("-customizations");	// Don't record any settings before now.
	parse_command_line_flags(1);
	fix_drive_only_file_names(1);
	if (!noinit_flag && load_from_state)
		init_from_file();
	parse_command_line_flags(2);
	if (tflag) {
		preserve_session_once = 1;
		tutorial();
	} else {
		if (load_from_state && (!have_files || session_always_restore)) {
			if (session_restore_files == 2 && have_files)	// Have cmdline files,
				save_var session_restore_files = 0;	// skip session files.
			load_session();
		}
		if (have_files)
			load_cmd_line_files(1, splitmode, dir);
		delete_dull_buffer("startup");
	}
	switching_to_buffer();
	restore_check_version();
	show_text(0, 1, "Epsilon Programmer's Editor %s, EEL version %s",
			  version, eel_version);
	try_calling("start-up");
	parse_command_line_flags(3);
	if (!noinit_flag && load_from_state)
		load_version_patch_file();
}

/* Start at argv[i].  If dirname, interpret filenames relative to it. */
load_cmd_line_files(int i, int splitmode, char *dirname)
{
	int wins = 0, tot_size, line, column, pos, j, all_files = 0;
	char fname[FNAMELEN], fixed[FNAMELEN], orig_dir[FNAMELEN], *s, first = 1;

	if (dirname) {
		getcd(orig_dir);
		on_exit chdir(orig_dir);
		chdir(dirname);
	}
	one_window();			// Don't use windows specified by session
	tot_size = window_height;
	if (is_gui && i + 1 < argc && argv[i][0] != '+') {	// multiple args
		for (j = i, *fixed = 0; j < argc; j++) {
			if (strlen(fixed) + strlen(argv[j]) >= FNAMELEN - 1) {
				*fixed = 0;
				break;		// See if perhaps a stray association is
			}				// passing us a file name w/o the quotes.
			if (*fixed)
				strcat(fixed, " ");
			strcat(fixed, argv[j]);
		}
		if (check_file(fixed) == 1 && !check_file(argv[i])) {
			argv[i] = strsave(fixed);
			argc = i + 1;
		}
	}
	line = column = pos = -1;
	for (; i < argc; i++) {
		if (argv[i][0] == '+') {
			if (argv[i][1] != 'p') {
				line = strtoi(argv[i] + 1, 10);
				s = strchr(argv[i], ':');
				if (s)
					column = strtoi(s + 1, 10);
			} else
				pos = strtoi(argv[i] + 2, 10);
		} else if (argv[i][0] == '-' && !all_files) {
			if (argv[i][1] == '-')
				all_files = 1;
			else if (!strcmp(argv[i], "-dir")
					 || strchr("ldr", argv[i][1]) && !argv[i][2]
					 || strchr("f", argv[i][1]) && !argv[i][3])
				i++;
		} else {
			strcpy(fname, argv[i]);
			convert_cygwin_filename(fname, 0);
			absolute(fname);
			if ((directory_flags & DIR_CWD_FROM_FILE) && first) {
				char dir[FNAMELEN];
				strcpy(dir, fname);
				*get_tail(dir, 1) = 0;
				chdir(dir);
				first = 0;
			}
			s = expand_wildcards ? do_file_match(fname, STARTMATCH | EXACTONLY)
				: fname;
			if (!s)		/* new file */
				s = fname;
			do {
				if (wins < max_initial_windows && wins++ > 0) {
					window_split(splitmode);
					window_height = 1000;
				}
				if (!(is_pattern(s) & 2) && !is_directory(s)
					&& !is_remote_file(s))	// Has quotable wildcards.
					find_it(s, get_default_translation_type(s));
				else
					do_find(s, get_default_translation_type(s));
				if (line > 0)
					go_line(line);
				if (pos >= 0)
					point = pos;
				if (column >= 0)
					move_to_column(column);
			} while (expand_wildcards
				&& (s = do_file_match(fname, EXACTONLY)));
			line = column = pos = -1;
		}
	}
	for (; window_number > 0; window_number--)
		window_height = tot_size / number_of_user_windows();
}

delete_dull_buffer(buf)		/* delete if no file & not in window */
char *buf;
{
	int orig, keep;

	orig = bufnum;
	bufname = buf;
	keep = (*filename && force_save_as != FSA_NEWFILE)
		   || buf_in_window(bufnum) > 0;
	bufnum = orig;
	if (!keep)
		delete_buffer(buf);
}

// See if there's an EEL patch file for this version available, and load it.
load_version_patch_file()
{
	char file[FNAMELEN], ver[FNAMELEN], *s;
	strcpy(ver, eel_version);
	while (s = strchr(ver, '.'))
		strcpy(s, s + 1);
	sprintf(file, "v%spatch.e", ver);
	show_text(-2, 0, "Looking for patch file %s on EPSPATH", file);
	if (!load_eel_from_path(file, 1))
		show_text(-2, 0, "Loaded %s", (s = lookpath(file)) ? s : file);
}

init_from_file()
{
	char *p, file[FNAMELEN];
	int res;
	jmp_buf this_level;

	save_var top_level = &this_level;	// Don't let errors abort,
	if (setjmp(top_level)) {		// so session file can still load.
		show_text(0, -1, "Error loading %f, see %s buffer.",
				  einit_file_name, MSG_BUFFER);
		return;
	}
	save_var _abort_load_on_error = 2;
	if ((load_customizations & 1) && (p = lookpath(einit_file_name))) {
		strcpy(file, p);
		show_text(-2, 0, "Loaded init file %s", file);
		save_var bufnum = zap("-customizations");
		do_file_read(file, FILETYPE_AUTO);
		ini_mode();
		res = parse_cmds();
		tab_size = tab_size.default;
		restore_vars();
		if (res)
			show_text(0, -1, "Error loading %f, see %s buffer.", file, MSG_BUFFER);
	}
}

usage_new_page()
{
	if (!is_gui) {
		if (is_unix != IS_UNIX_XWIN)
			bprintf("\nPress any key for more flags.");
		file_write(TERMINAL, 1);
		leave_blank = 1;
		if (is_unix == IS_UNIX_TERM)
			term_position(0, screen_lines - 1);
		if (is_unix != IS_UNIX_XWIN)
			getkey();

		buf_zap(bufnum);
	}
}

usage()
{
	char name[FNAMELEN], *p;

	bufnum = zap("usage");
	bprintf("\n");
	strcpy(name, get_tail(argv[0], 0));
	*get_extension(name) = 0;
	if (opsys != OS_UNIX)
		for (p = name; *p; p++)
			*p = tolower(*p);
	bprintf("Usage: %s [flags...] [files...]\n", name);
	bprintf("A file name may have +linenum or +linenum:column before it.\n\n");

	bprintf("Flags:\n");
	bprintf("-sFILE   Load state from FILE.sta (default: epsilon.sta).\n");
	bprintf("-bFILE   Build new system from FILE.b, not .sta file "
			"(default: epsilon.b).\n");
	bprintf("-quickup Load quickup.b after loading state."
			"    -nologo Skip initial msg.\n");
	bprintf("-noinit  Skip loading einit.ecm.\n\n");

	bprintf("-lFILE   Load FILE.b after loading files. ");
	bprintf("-rCMD    Run command or macro CMD.\n");
	bprintf("-teach   Run the tutorial.                ");
	bprintf("-dVAR!VAL Set variable to value.\n");
	bprintf("-add     Send files to prior instance.    ");
	bprintf("-noserver Don't accept sent files.\n");
	bprintf("-wait    Send files, then wait.           ");
	bprintf("-server SVR Set the server name for -add.\n");
	bprintf("-dir DIR  Interp. relative paths on cmd line.\n");
	bprintf("-p[VAL]  Set/toggle session-preserving.   ");
	bprintf("-pFILE    Load session from FILE.\n");
	bprintf("-wNUM    Set various directory control flags.\n");

	usage_new_page();
	bprintf("\n");
	bprintf("-fsDIRS  Use DIRS for temp space.         ");
	if (!(has_feature & FEAT_WINHELP))
		bprintf("-fdFILE  Documentation in file FILE.");
	bprintf("\n");

	bprintf("-vmono   Use monochrome color scheme.     ");
	bprintf("-vlNUM   Screen has NUM lines.\n");
	bprintf("-vcolor  Use colors.                      ");
	bprintf("-vcNUM   Screen has NUM columns.\n");
	bprintf("-vv      Put cmd line files side by side.\n\n");
	if (opsys == OS_UNIX) {
		bprintf("-vt      Don't use X11.                   ");
		bprintf("-bw PIX Window border width.\n");
		bprintf("-display DISP  Use this X11 display.      ");
		bprintf("-fn FONT Use this X11 font.\n");
		bprintf("-geometry GEOM  X11 window geometry spec. ");
		bprintf("-name NAME Look up X11 resouces under NAME.\n");
		bprintf("-title TITLE  Set X11 window title.       ");
		bprintf("-xrm RES Specify an X11 resource name/val.\n");
	}
	bprintf("-kaNUM   Compatibility flags.\n");
	if (opsys == OS_WINDOWS && !is_gui) {
		bprintf("-kcNUM   Mouse graphic cursor/block/off.  ");
		bprintf("-kmNUM   Mouse relative/absolute/off.\n");
	}
	bprintf("-ksNUM   Speed (>100) vs. abort response.\n");
	if (opsys == OS_WINDOWS && !is_gui)
		usage_new_page();

	bprintf("\n");
	bprintf("-mNUM    Use NUM bytes of main memory.    ");
	bprintf("-m0      Use minimum memory.\n");
	bprintf("-mNUM    Use NUM kbytes (if NUM < 1000)   ");
	bprintf("-m       Use as needed.\n\n");
	bprintf("  NUMk means kilobytes, NUMm means megabytes.\n");
	bprintf("  Without suffix, NUM is in kilobytes if < 1000, "
			"otherwise bytes.\n");
	restore_vars();
	init_abort();
}

bad_flag(s)
char *s;
{
	bprintf("\n");
	bprintf("Unrecognized flag %s on command line.\n", s);
	init_help_abort();
}

init_help_abort()
{
	bprintf("For help, type \"epsilon -help\".\n\n");
	init_abort();
}

init_abort()	/* show current buffer, then abort */
{
	if (has_feature & FEAT_GUI_DIALOGS) {
		int buf = bufnum, win;
		window_bufnum = zap("Error");
		save_var _doing_input = DI_VIEW, paging_retains_view = 1;
		win = add_popup(0, 0, screen_cols, screen_lines, 0, buf);
		point = 0;
		go_to_link = 0;
		display_dialog_box("UsageBox", "Usage", win, 0, 0, "OK", "", "");
		view_loop(win);
		leave();
	} else {
		if (is_unix == IS_UNIX_TERM)
			bprintf("\nPress any key to continue.\n");
		file_write(TERMINAL, 1);
		leave_blank = 1;
		if (is_unix == IS_UNIX_TERM) {
			term_position(0, screen_lines - 1);
			getkey();
		}
		leave();
	}
}

early_init()
{
}

middle_init()
{
}
