/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1995, 2020 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// This file has Epsilon's code coloring functions.
// There are routines that handle code coloring in general, and
// routines specific to code coloring in C-mode buffers.

#include "eel.h"
#include "c.h"
#include "colcode.h"

// Helpful for debugging. Outputs info on a tagged region in the current
// buffer within a given range to outbuf.
dump_tagged_region(char *tag, char *outbuf, int from, int to)
{
//	return;
	int val, p1, p2;
	buffer_printf(outbuf, "Tag %s ranges in buffer %s:\n", tag, bufname);
	while (from < to) {
		val = get_tagged_region(tag, from, &p1, &p2);
		buffer_printf(outbuf, "%d from %d to %d: ", val, p1, p2);
		if (p2 - p1 > 60) {
			dump_tagged_region_helper(outbuf, p1, p1 + 29);
			buffer_printf(outbuf, "...");
			dump_tagged_region_helper(outbuf, p2 - 29, p2);
		} else
			dump_tagged_region_helper(outbuf, p1, p2);
		buffer_printf(outbuf, "\n");
		if (from >= p2 || p2 >= size() - 1)
			return;
		from = p2;
	}
}

dump_tagged_region_helper(char *outbuf, int start, int end)
{
	int c;
	while (start < end) {
		c = character(start++);
		if (c == '\n')
			buffer_printf(outbuf, "\\n");
		else
			buffer_printf(outbuf, "%c", c);
	}
}

init_color_code()
{
	coloring_flags |= COLOR_DO_COLORING;
}

#define MAX_BUFFER_WHEN_DISPLAYING	5
buffer int (*buffer_when_displaying[MAX_BUFFER_WHEN_DISPLAYING])();

// Just call the buffer-specific functions for this buffer.
default_when_displaying(int from, int to)
{
	UNREFERENCED_PARAMETER(from);
	for (to = 0; to < MAX_BUFFER_WHEN_DISPLAYING; to++)
		if (buffer_when_displaying[to])
			(*buffer_when_displaying[to])();
}

add_buffer_when_displaying(int buf, int (*func)())
{
	if (!add_to_buffer_func_list(buffer_when_displaying,
								 MAX_BUFFER_WHEN_DISPLAYING, buf, func))
		error("Too many when-displaying functions for this buffer.");
	if (!when_displaying)
		when_displaying = default_when_displaying;
}

delete_buffer_when_displaying(int buf, int (*func)())
{
	delete_from_buffer_func_list(buffer_when_displaying,
								 MAX_BUFFER_WHEN_DISPLAYING, buf, func);
}

// On each refresh, recolor_partial_code() will be called with the
// buffer range that was modified since the last redisplay. It handles
// changes in syntax highlighting rules, and sees that the visible
// buffer region is syntax highlighted.
recolor_partial_code(from, to)
{
	init_color_code();
	if (!(coloring_flags & COLOR_IN_PROGRESS)) {
		coloring_flags |= COLOR_IN_PROGRESS; // ensure errors don't restart us
		if (!(coloring_flags & COLOR_MINIMAL) != !minimal_coloring) {
			drop_all_colored_regions();		// If setting's changed.
			coloring_flags &= ~COLOR_MINIMAL;
			if (minimal_coloring)
				coloring_flags |= COLOR_MINIMAL;
		}
		save_var point, matchstart, matchend, abort_searching = 0;
		if (!(coloring_flags & COLOR_RETAIN_NARROWING))
			save_var narrow_start = 0, narrow_end = 0;
		shrink_colored_area();		// Buffer modifications shrink list of regions.
		if (size() < color_whole_buffer) // Now expand list of regions.
			color_for_area(0, size());
		else {
			int end = MIN(window_end + 1, size());
			if (display_column >= 0)
				hscroll_color_for_area(window_start, end);
			else
				color_for_area(window_start, end);
		}
		coloring_flags &= ~COLOR_IN_PROGRESS;	// no errors aborted us,
	}					// let us run next time
	default_when_displaying(from, to);
}

// We're displaying the buffer region start...end, so syntax-highlight
// it if needed. But since we're horizontally scrolling, and lines might
// extend far to the right, we'll avoid wasting time coloring those
// parts, by coloring line by line.
void hscroll_color_for_area(int start, int end)
{
	int i, from, to, next, max = text_height();
	int last_column = text_width() + display_column;
	save_var point;
	for (i = 0; i < max; i++) {		// For each screen line of this window
		start = window_line_to_position(i);
		if (start >= end)	// If its start is past the region to color, done.
			break;
		// To avoid an expensive move_to_column(), we first check if
		// this line is already entirely colored.
		if (!get_tagged_region("needs-color", start, &from, &to)) {
			// The start of this line is already colored and it's part
			// of the already-colored region from...to.
			next = window_line_to_position(i + 1);	// Start of next screen line.
			if (next == -1 && i == max - 1)	// Last screen line, so we can't get
				next = end;		// start of the next one -- substitute.
			if (next != -1 && next <= to)
				continue;		// Entire line is already colored.
		}
		point = start;		// Find buffer pos of the last column we need to color.
		move_to_column(last_column);
		if (curchar() == '\n') {
			point++;
			hscroll_color_extra(start);
		}
		color_for_area(start, point);
	}
}

// We're at the start of a line. Move forward over additional lines as
// long as we stay within a certain number of characters of start.
int hscroll_color_extra(int start)
{
	save_var narrow_end = size() - (start + 2000);
	int p = point;
	while (nl_forward())
		p = point;
	point = p;		// Start of last full line in region.
}

// Return -1 if this particular call to wait_for_key() doesn't need to
// call us again, or the amount of time in hundredths of a second until
// we wish to be called again.

when_idle_coloring(cnt)	// Called when idle, waiting for a key.
{			// Cnt increases while waiting for same key.
	int last = 0, from, to;

	if (idle_coloring_delay < 0 || !(coloring_flags & COLOR_DO_COLORING)
		|| (coloring_flags & COLOR_IN_PROGRESS) || !size())
		return -1;
	if (cnt < 1)
		time_begin(&idle_color_time, idle_coloring_delay);
	if (!time_done(&idle_color_time))
		return time_remaining(&idle_color_time);
	save_var narrow_start = 0, narrow_end = 0;
	if (!get_tagged_region("needs-color", 0, &from, &to)) {
		if (to >= size())	// We're done if the good region extends to eof.
			return -1;		// Otherwise, get the following (bad) region.
		get_tagged_region("needs-color", to, &from, &to);
	}
	last = from;
	save_var point, matchstart, matchend, abort_searching = 0;
	coloring_flags |= COLOR_IN_PROGRESS; // ensure errors don't restart us
	color_for_area(last, last + idle_coloring_size);
	coloring_flags &= ~COLOR_IN_PROGRESS;	// no errors aborted us, do more
	return 0;
}

/*
First invalidate the modified part of the buffer.  This typically
happens once on each refresh.

First, get the region of the buffer that has been modified since last
time we were called.

Then expand that region to include entire lines.

Then (unless this buffer uses line-based coloring), expand that region
to include entire colored regions.  For instance, if the modification
was inside a block comment, expand the region to the whole comment.
(We omit this step during line-based coloring because each region might
be very large.)

*/

shrink_colored_area()
{
	int from, to, adjfrom, adjto, by_lines;

	if (!(coloring_flags & COLOR_DO_COLORING))
		return;		// No coloring yet, so nothing to do.
	by_lines = (recolor_from_here == recolor_by_lines);
	if (modified_buffer_region(&from, &to, "needs-color")) {
		point = from - 1;
		if (by_lines)
			adjfrom = give_begin_line();
		else
			get_character_color(give_begin_line(), &adjfrom, (int *) 0);
		point = to;
		if (by_lines)
			adjto = give_end_line() + 1;
		else
			get_character_color(give_end_line(), (int *) 0, &adjto);
		if (coloring_flags & COLOR_INVALIDATE_FORWARD)
			adjto = size();
		if (coloring_flags & COLOR_INVALIDATE_BACKWARD)
			adjfrom = 0;
		set_tagged_region("needs-color", adjfrom, adjto, -1);
		if (coloring_flags & COLOR_INVALIDATE_RESETS)
			set_character_color(adjfrom, adjto, -1);
	}
	reset_modified_buffer_region("needs-color");
}

// Make sure this range in the current buffer has the right coloring.
// If the buffer has been modified since its last coloring, mark the
// modified area uncolored first.
recolor_buffer_range(start, end)
{
	if (want_code_coloring && recolor_range) {
		save_var point;
		shrink_colored_area();
		if (!(coloring_flags & COLOR_RETAIN_NARROWING))
			save_var narrow_start = 0, narrow_end = 0;
		color_for_area(start, end);
	}
}

int get_character_syntax_color(int pos)
{
	save_var narrow_start = 0, narrow_end = 0, matchstart, matchend, point;
	recolor_buffer_range(pos, pos + 1);
	return get_character_color(pos);
}

// Return nonzero if some characters in this buffer don't have
// up-to-date coloring applied already.
int chars_need_coloring()
{
	int from, to;

	if (!want_code_coloring || !recolor_range)
		return 0;
	save_var point;
	shrink_colored_area();
	return get_tagged_region("needs-color", 0, &from, &to) || to != size();
}

// To speed up scrolling, go back a few lines, so we color multiple
// lines at a time.  But limit this if lines are very long.
color_go_back_lines()
{
	save_var narrow_start = point - 10000;
	nl_reverse();
	nl_reverse();
	nl_reverse();
}

/*
Now deal with the buffer range we have to color (typically the
current window).  If our starting point isn't correctly colored
already, then begin coloring at a position somewhere before,
determined by the mode's recolor_from_here() function.
Then, for each uncolored region between from and to, color it,
until the good region extends between these points.
*/

color_for_area(from, to)	// we need to color region in this range
{
	save_var point, abort_searching = 0;
	init_color_code();
	must_color_through = to;
	color_for_narrowed_area(from, to);
}

// Like above, but don't disturb narrowing.
color_for_narrowed_area(from, to)
{
	int i, j = 0, give_up = 0, last_j;

	save_var point, abort_searching = 0;
	init_color_code();
	to = MIN(to, size() - narrow_end);
	for (;;) {
		last_j = j;
		if (get_tagged_region("needs-color", from, &i, &j)) {
			if (i < from) {
				point = from;		// From isn't colored, so start before it.
				color_go_back_lines();	// Color a few extra lines before
										// to speed up scrolling.
				if (point > i) // Next, find starting place known to be on 
					i = recolor_from_here(i); // the edge of a colored region.
			}				// (Or use known safe place if it's closer.)
		} else {
			if (j >= to) {		// From is colored, see if good region extends
				return;			// far enough.  If not, get following bad reg.
			} else if (!get_tagged_region("needs-color", j, &i, &j)) {
				int a, b, res;

				say("Coloring internal error 1");
				buffer_printf("debug-color", "Coloring internal error 1:\n");
				buffer_printf("debug-color", "bad color 0 from %d to %d in %s, size %d\n",
							  i, j, bufname, size());
				res = get_tagged_region("needs-color", from, &a, &b);
				buffer_printf("debug-color", "color %d at %d (from %d to %d) in %s, size %d\n",
							  res, from, a, b, bufname, size());
				res = get_tagged_region("needs-color", j - 1, &a, &b);
				buffer_printf("debug-color", "color %d at %d (from %d to %d) in %s, size %d\n",
							  res, j - 1, a, b, bufname, size());
				return;
			}
		}
		j = MIN(j, to);		// Color only some of it.
		if (i >= j)
			return;
		j = call_recolor_range(i, j);
		set_tagged_region("needs-color", i, j, 0);
		if (j <= last_j && give_up++ > 100) {
			say("Coloring internal error 2");
			when_displaying = 0;
			return;
		}
	}
}

int call_recolor_range(int from, int to)
{
	if (want_debug_coloring)		// Clean slate to see just what it colors.
		set_tagged_region("debug-color", 0, size(), -1);
	int ret = recolor_range(from, to);
	if (want_debug_coloring)
		check_debug_coloring(from, to, ret);
	return ret;
}

// Record that the area between from & to is now correctly colored.
add_colored_area(from, to)
{
	set_tagged_region("needs-color", from, to, 0);
}

drop_all_colored_regions()
{
	set_tagged_region("needs-color", 0, size(), -1);
}

// A recolor_from_here function good for line-based buffers.
recolor_by_lines(safe)
{
	safe = safe;		// eliminate compiler warning
	return give_begin_line();
}

// A recolor_from_here function for always coloring from start of buffer.
recolor_from_top(safe)
{
	return safe;
}

// functions for coloring C/C++/EEL code

color_c_from_here(safe)	// Move backward to the nearest line guaranteed
{			// to start outside any colored region, return pos.
	int pos, limit; // We know safe is ok value to return.

	limit = point - safe;
	if (color_look_back && color_look_back < limit)
		limit = color_look_back;
	to_begin_line();
	pos = point;
	while ((in_c_comment(limit) == IN_OLD_COMMENT)
				       && search(-1, "/*")) {
		to_begin_line(); // move back out of an old-style comment
		pos = point;
	}
	point = pos;
	while (character(point - 2) == '\\') {	// don't stay on contin line
		point--;
		to_begin_line();
	}
	return point;
}

 // Recolor just this section. Last colored region may go past to.
 // Return where we stopped coloring.
int color_c_range(int from, int to)
{
	if (c_extra_keywords & JS_KEYWORDS) {	// Recolor all of a JS template literal
		int start;
		if (get_tagged_region("color-c-as-unit", from, &start) != -1)
			from = start;
	}
	return do_color_c_range(from, to);
}

int do_color_c_range(int from, int to)
{
	int t = -1, talk, s, talk_at = 0;
	char pat[200];
	TIMER talk_now;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend;
	c_init_color(from, to);
	point = from;
	if (talk = (to - from > 30000))	// Show status during long delays.
		time_begin(&talk_now, 50);		// Wait a bit before chatter.
	save_var case_fold = 0;
	strcpy(pat, "/<*>|//|^[ \t]*#|[\"']");
	if (!minimal_coloring)
		strcat(pat, "|[A-Za-z_][A-Za-z0-9_]*"
		       "|-?%.?[0-9]([A-Za-z0-9._']|[Ee]-)*");
	if (c_extra_keywords & JS_KEYWORDS)
		strcat(pat, "|`");	// JavaScript: `${template} ${literals}` syntax
	while (point < to) {
		if (!re_search(1, pat)) {
			t = size();
			break;
		}
		t = matchstart;
		switch (character(point - 1)) {		// check last char
		case '/':			// found // one-line comment
			while (nl_forward() && character(point - 2) == '\\')
				;		// Continued // comment.
			set_character_color(t, point, color_class c_comment);
			break;
		case '*':			// found /* starting comment
			search(1, "*/");
			set_character_color(t, point, color_class c_comment);
			if (get_character_color(point, (int *) 0, &s)
				  == color_class c_comment && s > to) {
				c_init_color(point, s);
				to = s;	// Comment extended further before; rescan it.
			}
			break;
		case '#':			// found preproc line
			if (matches_at(t, 1, c_color_preproc_if_pattern)) {
				if (locate_by_ifdef(1))
					nl_forward();
				else
					point = size();
				set_character_color(t, point, color_class c_comment);
			} else {
				c_preproc_color();	// move past correct part
				set_character_color(t, point, color_class c_preprocessor);
			}
			break;
		case '"':		// found a string literal
			point = t;
			if ((c_extra_keywords & OBJC_KEYWORDS) && character(t - 1) == '@')
				t--;
			if ((c_extra_keywords & CPP_KEYWORDS) && character(t - 1) == 'R'
				  && move_past_cpp_raw_literal(0, 0))
				t--;	// Color R too.
			else if ((c_extra_keywords & CSHARP_KEYWORDS)
				  && character(t - 1) == '@')
				t--, re_search(1, "\"([^\"]|\"\")*\"![^\"]");
			else
				re_search(1, "\"([^\"\\\n]|\\(.|\n))*[\"\n]");
			set_character_color(t, point, color_class c_string);
			if (get_character_color(point, (int *) 0, &s) == 
			    color_class c_string && s > to)  // fix up after
				c_init_color(point, to = s); // quoted "'s
			break;
		case '\'':		// found a char const
			point = t;
			re_search(1, "\'([^\'\\\n]|\\(.|\n))*[\'\n]");
			set_character_color(t, point, color_class c_charconst);
			break;
		case '`':		// found JavaScript template literal
			color_js_template_literal(t);
			break;
		default:		// found identifier, kywd, or number
			if ((c_extra_keywords & (OBJC_KEYWORDS | JAVA_KEYWORDS))
				  && character(t - 1) == '@')
				t--;
			set_character_color(t, point, c_keyword_color(t));
			if ((c_extra_keywords & IDL_KEYWORDS)	// Special rule for
				&& (point - t == 4)			// uuid's in IDL files.
				&& parse_string(-1, "uuid")
				&& parse_string(1, "%([0-9A-Fa-f]"))
				maybe_color_uuid();
			break;
		}
		if (talk && point > talk_at + 30000 && time_done(&talk_now)) {
			note("Coloring C program: %d%% complete...",
				 muldiv(point - from, 100, to - from));
			talk_at = point;
		}
	}
	c_init_color(to, t);
	if (talk_at)
		note("");
	return point;
}

// Parse a C++ raw string literal like R"delim(text)delim". For forward
// (reverse is 0), point+offset is just before the opening ". Move to
// just after the final " and return 1. If it wasn't a valid raw string
// literal, just return 0 without moving. If unterminated, move to EOF
// and return 2.

// If reverse is nonzero, we try instead to parse it backwards, and
// return 0 if it wasn't a raw string literal after all. We expect
// point+offset to be just after the final ", and move to just before
// the starting R on success.
int move_past_cpp_raw_literal(int offset, int reverse)
{
	char delim[FNAMELEN], pat[FNAMELEN];

	save_var case_fold = 0;
	if (reverse) {
		int orig = point;
		if (!matches_at(point + offset, -1, "%)([^\n()\\]*)\""))
			return 0;
		grab(find_group(1, 1), find_group(1, 0), delim);
		sprintf(pat, "R\"%q%%(", delim);
		if (re_search(-1, pat))
			return 1;
		point = orig;
		return 0;
	}		
	if (character(point + offset - 1) != 'R'
		  || !matches_at(point + offset, 1, "\"([^\n()\\]*)%("))
		return 0;
	grab(find_group(1, 1), find_group(1, 0), delim);
	sprintf(pat, "%%)%q\"", delim);
	return re_search(1, pat) ? 1 : 2;
}

c_init_color(from, to)
{
	if (from < to)
		set_character_color(from, to, minimal_coloring
				    ? color_class c_identifier
				    : color_class c_punctuation);
}

// Return color for "identifier" from here to point (something with
// alpha or digits).  May adjust point.
c_keyword_color(from)
{
	char buf[500];

	if (point - from > sizeof(buf) / sizeof(char) - 10)
		return color_class c_identifier;
	buf[0] = '|';		// get identifier, between | chars
	grab(from, point, buf + 1);
	if (index("0123456789-.", buf[1])) {
		if ((c_extra_keywords & CSHARP_KEYWORDS)) {
			point = from;		// Rules for C# numbers are slightly different.
			point += parse_string(1, "-?%.?[0-9]([A-Za-z0-9_]|[Ee]-|.[0-9])*");
			grab(from, point, buf + 1);
		}
		return c_number_color(buf + 1);
	}
	strcpy(buf + point - from + 1, "|");
	if (is_c_keyword(buf) && (!(c_extra_keywords & CSHARP_KEYWORDS) ||
							  character(from - 1) != '@'))
		return color_class c_keyword;
	if (color_class c_function != color_class c_identifier
		  && paren_follows())
		return color_class c_function;
	return color_class c_identifier;
}

is_c_keyword(p)		// is text in p (must be surrounded by |'s) a keyword?
char *p;
{
	if (strstr("|if|else|for|do|while|return|break|continue|switch|"
		   "case|default|goto|char|short|int|static|volatile|const|"
		   "float|double|long|void|enum|", p))
		return 1;
	if (p[1] == '_' &&
	    strstr("|_asm|_based|_cdecl|_emit|_except|_export|_far|_fastcall|"
		   "_fortran|_huge|_interrupt|_loadds|_near|_pascal|_saveregs|"
		   "_segment|_segname|_stdcall|_syscall|_try|__int64|", p))
		return 1;
	if (!(c_extra_keywords & (EEL_KEYWORDS | CPP_KEYWORDS | JAVA_KEYWORDS |
							  JS_KEYWORDS | IDL_KEYWORDS | CSHARP_KEYWORDS | OBJC_KEYWORDS))
		  && strstr("|restrict|_Alignas|_Alignof|_Atomic|_Bool|_Complex|"
					"_Generic|_Imaginary|_Noreturn|_Static_assert|_Thread_local|"
					"_Decimal128|_Decimal32|_Decimal64|", p))
		return 1;		// C language only
	if ((c_extra_keywords & EEL_KEYWORDS) &&
	    strstr("|keytable|buffer|window|command|on|color_class|"
		   "spot|save_spot|save_var|on_exit|user|zeroed|byte|int32|", p))
		return 1;
	if ((c_extra_keywords & CPP_KEYWORDS) &&
	    strstr("|asm|catch|class|delete|friend|inline|new|operator|"
			   "private|protected|public|template|try|this|virtual|"
			   "throw|_self|and|and_eq|bitand|bitor|bool|compl|const_cast|"
			   "dynamic_cast|explicit|false|mutable|namespace|not|not_eq|"
			   "or|or_eq|reinterpret_cast|static_cast|true|typeid|typename|"
			   "using|wchar_t|xor|xor_eq|", p))
		return 1;
	if ((c_extra_keywords & CPP_KEYWORDS) &&
		  strstr("|alignas|alignof|char8_t|char16_t|char32_t|concept|consteval|"
				 "constexpr|constinit|co_await|co_return|co_yield|decltype|"
				 "export|noexcept|nullptr|requires|static_assert|thread_local|", p))
		return 1;		// Later C++ keywords
	if ((c_extra_keywords & CPP_KEYWORDS) &&
		  strstr("|override|final|import|module|", p))
		return 1;		// C++ "identifiers with special meaning"
	if ((c_extra_keywords & (JAVA_KEYWORDS | JS_KEYWORDS)) &&
	    strstr("|abstract|assert|boolean|byte|catch|class|extends|false|final|"
			   "finally|function|implements|import|in|instanceof|"
			   "interface|native|new|null|package|private|protected|"
			   "public|strictfp|super|synchronized|this|throw|throws|transient|"
			   "true|try|var|with|", p))
		return 1;
	if (!(c_extra_keywords & (JAVA_KEYWORDS | JS_KEYWORDS)) &&
	    strstr("|auto|extern|register|unsigned|signed|sizeof|"
			   "struct|union|typedef|", p))
		return 1;		// C keywords not in Java.
	if ((c_extra_keywords & IDL_KEYWORDS) &&
		strstr("|appobject|auto_handle|bindable|boolean|broadcast|byte|"
			   "callback|coclass|code|comm_status|context_handle|"
			   "control|decode|defaultbind|dispinterface|displaybind|"
			   "dual|enable_allocate|encode|error_status_t|"
			   "explicit_handle|fault_status|handle|handle_t|hidden|"
			   "hyper|idempotent|ignore|immediatebind|import|in|"
			   "include|interface|library|licensed|local|maybe|module|", p))
		return 1;
	if ((c_extra_keywords & IDL_KEYWORDS) &&
		strstr("|ms_union|nocode|nonextensible|notify|object|odl|"
			   "oleautomation|optional|out|pipe|propget|propput|"
			   "propputref|public|ptr|readonly|ref|requestedit|"
			   "restricted|retval|small|source|string|unique|"
			   "usesgetlasterror|v1_enum|vararg|", p))
		return 1;
	if ((c_extra_keywords & CSHARP_KEYWORDS) &&
		  strstr("|abstract|as|base|bool|byte|catch|checked|class|decimal|delegate|"
				 "event|explicit|false|finally|fixed|foreach|implicit|in|"
				 "interface|internal|is|lock|namespace|new|null|object|"
				 "operator|out|override|params|private|protected|public|"
				 "readonly|ref|sbyte|sealed|sizeof|stackalloc|string|this|"
				 "throw|true|try|typeof|uint|ulong|unchecked|unsafe|ushort|"
				 "using|virtual|", p))
		return 1;
	if ((c_extra_keywords & OBJC_KEYWORDS) &&
		  strstr("|self|super|BOOL|id|Class|nil|YES|NO|NS_DURING|NS_HANDLER|"
				 "NS_ENDHANDLER|in|out|inout|bycopy|byref|oneway|@interface|"
				 "@implementation|@end|@class|@private|@protected|@public|"
				 "@encode|@protocol|@selector|@try|@catch|@finally|@throw|"
				 "@synchronized|Nil|IBAction|IBOutlet|", p))
			return 1;
	return 0;
}

// Is there whitespace followed by a "(" at point in buffer?
// If so, we assume it's a function name. We also handle other things
// that can appear before the "(".
paren_follows()
{
	char pat[FNAMELEN];
	save_var point;
	sprintf(pat, "[ \t\n]*((%s)</word>)?[ \t\n]*", c_prototype_macro_names);
	point += parse_string(1, pat);
	for (;; point++) {
		switch (curchar()) {
			case ' ': case '\t': case '\n':
				continue;
			case '(':		return 1;
			case '<':		if (find_template_end()) {
								point--;
								break;
							}
			default:		return 0;
		}
	}
}

c_number_color(s)	// return color for number in s
char *s;		// redefine to distinguish different bases
{
	s = s;		// silence eel's warning
	return color_class c_number;
}

maybe_color_uuid()
{
	save_var case_fold = 1;		// Does it look like a quoteless uuid?
	if (!parse_string(1, "%([0-9A-F][0-9A-F][0-9A-F][0-9A-F]"
					  "[0-9A-F][0-9A-F][0-9A-F][0-9A-F]-")
		|| !parse_string(1, "%([-0-9A-F]+%)"))
		return;			// No, so keep searching.
	set_character_color(matchstart, matchstart + 1,	// Do special colors.
						color_class c_punctuation);
	set_character_color(matchstart + 1, matchend - 1,
						color_class c_number);
	set_character_color(matchend - 1, matchend,
						color_class c_punctuation);
	point = matchend;			// And move past this.
}


c_preproc_color()	// color from t to end of preprocessor text
{
	int orig = point;

	re_search(1, "[ \t]*");
	if (parse_string(1, "include|import|tryinclude")) {
		while (nl_forward() && character(point - 2) == '\\')
			;			// color contin lines
		save_var narrow_end = size() - point;
		point = orig;
		while (re_search(1, "/[*/]"))	// stop before comment
			if (!(in_c_comment(c_look_back) & IN_STRING)) {
				point -= 2;
				break;
			}
	} else if (parse_string(1, "(define|undef)[ \t]+[A-Za-z0-9_]+")) {
		point = matchend;
		point += parse_string(1, "%([A-Za-z0-9_, \t]*%)");
	} else
		re_search(1, "[A-Za-z_0-9]*");
}

// Color a JavaScript template literal like `${template} ${literals}`.
// These can be nested, since any expression can be inside one.
void color_js_template_literal(int p)
{
	coloring_flags |= COLOR_INVALIDATE_FORWARD | COLOR_INVALIDATE_RESETS;
	int start = p;
	while (re_search(1, "[`\\$]")) {
		if (character(matchstart) == '`')
			break;
		else if (character(matchstart) == '\\')
			point++;
		else if (curchar() == '{') {
			set_character_color(p, point - 1, color_class c_string);
			p = point - 1;
			int braces = 0;
			while (re_search(1, "[`\\{}\"']")) {
				if (character(matchstart) == '}') {
					if (--braces <= 0)
						break;
				} else if (character(matchstart) == '{')
					braces++;
				else if (character(matchstart) == '\\')
					point++;
				else if (character(matchstart) == '`')
					color_js_template_literal(matchstart);
				else if (character(matchstart) == '"')
					re_search(1, "([^\"\\\n]|\\(.|\n))*[\"\n]");
				else if (character(matchstart) == '\'')
					re_search(1, "([^\'\\\n]|\\(.|\n))*[\'\n]");
			}
			p = point = color_narrowed_c_range(p, point);
		}
	}
	set_character_color(p, point, color_class c_string);
	set_tagged_region("color-c-as-unit", start, point, 1);
}

int color_narrowed_c_range(int start, int end)
{
	save_var narrow_start = start, narrow_end = size() - end;
	return do_color_c_range(start, end);
}

drop_coloring(b)
{
	save_var bufnum = b;
	set_character_color(0, size(), -1);
	if (coloring_flags & COLOR_DO_COLORING) {
		drop_all_colored_regions();
		full_redraw = 1;
	}
	coloring_flags &= ~(COLOR_DO_COLORING | COLOR_IN_PROGRESS);
}

command change_code_coloring()
{
	want_code_coloring = has_arg ? (iter != 0) : !want_code_coloring;
	iter = 0;
	say("Code coloring %sabled for this buffer.",
	    want_code_coloring ? "en" : "dis");
	when_setting_want_code_coloring();
}

// always call the following after setting want_code_coloring variable
// set-variable calls this automatically
when_setting_want_code_coloring()
{
	when_displaying = default_when_displaying;
	if (want_code_coloring)
		want_code_coloring = 1;
	if (!want_code_coloring)
		drop_coloring(bufnum);
	if (want_code_coloring && recolor_range && recolor_from_here)
		when_displaying = recolor_partial_code;
	coloring_flags &= ~(COLOR_DO_COLORING | COLOR_IN_PROGRESS);
}

#undef set_character_color
// Record that we're set some color for this region.
debug_set_character_color(int from, int to, int color)
{
	set_character_color(from, to, color);
	if (want_debug_coloring) {
		set_tagged_region("debug-color", from, to, 1);
		if (want_debug_coloring & 16) {
			char bname[FNAMELEN];
			sprintf(bname, "debug-color-%s", bufname);
			create(bname);
			buffer_to_end(bname, -1);
			buffer_printf(bname, "Painted %d...%d in ", from, to);
			char *name = name_color_class(color);
			if (name)
				buffer_printf(bname, "%s\n", name);
			else
				buffer_printf(bname, "color class %d\n", color);
		}
	}
}

// A coloring function was asked to color from...to, and reported that
// it colored from...ret. Check if it really did. If not, report the
// bug.
check_debug_coloring(int from, int to, int got)
{
	char bname[FNAMELEN];
	sprintf(bname, "debug-color-%s", bufname);
	int start, end, buf = -1, ok = 0, p, i;
	if (get_tagged_region("debug-color", from, &start, &end) == 1
		  && end >= got)
		ok = 1;
	if (!ok || (want_debug_coloring & 8)) {
		buf = create(bname);
		buffer_to_end(bname, -1);
	}
	if (ok) {		// No errors detected.
		if (want_debug_coloring & 8)
			buf_printf(buf, "Colored %d...%d, went to %d, OK\n", from, to, got);
		return;
	}
	save_var narrow_start = 0, narrow_end = 0;
	say("Possible coloring bug detected, see report in buffer %s", bname);
	buf_printf(buf, "Colored %d...%d, went to %d, some regions not colored:\n", from, to, got);
	p = from;
	do {
		i = get_tagged_region("debug-color", p, &start, &end);
		if (i == -1) {	// Missed
			buf_printf(buf, "  %d...%d was not colored\n", start, end);
			if (want_debug_coloring & 2)
				set_character_color(start, end, color_class debug_text);
		} else if (want_debug_coloring & 4)
			buf_printf(buf, "  %d...%d was colored\n", start, end);
		p = end;
	} while (p < to);
}
