/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1985, 2020 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// This file defines the set-color command, which changes Epsilon's
// screen colors.

#include "eel.h"
#include "proc.h"

#define FG(c)		((c) & 15)	/* extract foreground & background */
#define BG(c)		(((c) >> 4) & 7)
#define COLOR_COL	57
#define COLOR_LINE	2
#define SCRPLACE(c)	(COLOR_COL + FG(c)), (COLOR_LINE + BG(c))

// Is the name table entry m a color scheme?
int is_color_scheme(int m)
{
	return name_type(m) == NT_COLSCHEME
			&& strcmp(name_name(m), "color-defaults");

}
// Fill buffer b with a list of all color schemes.
make_color_scheme_list(int b)
{
	int i;

	buf_zap(b);
	for (i = 1; i <= final_index(); i++)
		if (is_color_scheme(i))
			buf_printf(b, "%s\n", name_name(i));
}

// Get color class name on current line, return its index.  If expanded
// category name, return -2.  If collapsed category name, return -3.
// Otherwise return -1.
int this_color_class_to_index()
{
	int i, classes = number_of_color_classes();

	char name[FNAMELEN];
	save_var point = give_begin_line();
	point += parse_string(1, " *");
	if (curchar() == '-')
		return -2;
	else if (curchar() == '+')
		return -3;
	grab(point, give_end_line(), name);
	for (i = 0; i < classes; i++)
		if (!strcmp(name, name_color_class(i)))
			return i;
	return -1;
}

int count_color_schemes()
{
	int i, j = 0;

	for (i = 1; i <= final_index(); i++)
		if (is_color_scheme(i))
			j++;
	return j;
}

record_color_settings(int **orig_fg, int **orig_bg, int **orig_schemes)
{
	int i, j, k, classes = number_of_color_classes();
	int schemes = count_color_schemes();

	*orig_schemes = (int *) malloc(schemes * sizeof(int));
	*orig_fg = (int *) malloc(classes * schemes * sizeof(int));
	*orig_bg = (int *) malloc(classes * schemes * sizeof(int));
	for (i = 1, k = 0; i <= final_index(); i++)
		if (is_color_scheme(i)) {
			selected_color_scheme = i;
			for (j = 0; j < classes; j++) {
				(*orig_fg)[k * classes + j] = get_foreground_color(j, 1);
				(*orig_bg)[k * classes + j] = get_background_color(j, 1);
			}
			(*orig_schemes)[k++] = i;
		}
}

restore_color_settings(int *orig_fg, int *orig_bg, int *orig_schemes)
{
	int i, j, k, classes = number_of_color_classes();
	int schemes = ptrlen(orig_schemes) / sizeof(int);

	for (i = 0; i < schemes; i++) {
		selected_color_scheme = orig_schemes[i];
		for (j = 0; j < classes; j++) {
			k = i * classes + j;
			set_color_pair(j, orig_fg[k], orig_bg[k]);
		}
	}
}

preserve_color_settings(int *orig_fg, int *orig_bg, int *orig_schemes,
						int orig_scheme)
{
	int i, j, k, n, classes = number_of_color_classes();
	int schemes = ptrlen(orig_schemes) / sizeof(int), fg_color, bg_color;

	*get_color_scheme_variable() = selected_color_scheme;
	if (want_record_customizations()) {
		k = prep_customizations();
		if (selected_color_scheme != orig_scheme)
			list_one(k, color_scheme_variable_index(), 1);
		for (i = 0; i < schemes; i++) {
			selected_color_scheme = orig_schemes[i];
			for (j = 0; j < classes; j++) {
				fg_color = get_foreground_color(j, 1);
				bg_color = get_background_color(j, 1);
				n = i * classes+ j;
				if (fg_color != orig_fg[n] || bg_color != orig_bg[n])
					list_one_color_setting(k, j, 1);
			}
		}
		selected_color_scheme = *get_color_scheme_variable();
		say("");
	} else
		show_binding("To remember these colors for "
					 "future sessions, use %s.", "list-customizations");
}

// Prompt for which font attribute to set, and return a code like
// EFONT_BOLD, or -1 if canceled. For X Windows.
int xwin_ask_attribute(int class, int fg_color, int bg_color)
{
	int buf, win;
	buf = tmp_buf();
	on_exit buf_delete(buf);
	save_var bufnum = buf;
	win = add_popup(4, 1, 70, 5, ALL_BORD(BSINGLE), buf);
	on_exit remove_window(win);
	save_var window_handle = win;
	window_title(win, BTOP, TITLECENTER, " Set Font Attribute ");
	set_menu_look(win);
	for (;;) {
		buf_zap(bufnum);
		bprintf("Click a box, or press B to toggle bold, I to toggle italic,\n");
		bprintf("or U to toggle underlined. Press D when done.\n");
		bprintf("[%c] Bold       [%c] Italic     [%c] Underlined           [Done]",
				(fg_color & EFONT_BOLD) ? 'X' : ' ',
				(fg_color & EFONT_ITALIC) ? 'X' : ' ',
				(fg_color & EFONT_UNDERLINED) ? 'X' : ' ');
		point = 0;
		refresh();
		getkey();
		key = toupper(key);
		if (IS_MOUSE_DOWN(key) && mouse_y == 4)
			switch ((mouse_x - 4) / 15 ) {
				case 0:	fg_color ^= EFONT_BOLD; break;
				case 1:	fg_color ^= EFONT_ITALIC; break;
				case 2:	fg_color ^= EFONT_UNDERLINED; break;
				case 3: return fg_color;
			}
		switch (key) {
			case 'B':	fg_color ^= EFONT_BOLD; break;
			case 'I':	fg_color ^= EFONT_ITALIC; break;
			case 'U':	fg_color ^= EFONT_UNDERLINED; break;
			case 'D':
			case 'Q':
			case '\n':
			case '\r':	return fg_color;
		}
		set_color_pair(class, fg_color, bg_color);
		screen_messed();
	}
}

#define XWIN_COLOR_BAR_HEIGHT	15

// Switch to win and make its scroll bar reflect the color (0...255).
// Also ensure it's filled with lines.
void xwin_update_color(int win, int this_color, int full_color)
{
	int i;
	save_var window_handle = win;
	if (!size())
		for (i = 0; i < 255; i++)
			bprintf("\n");
	point = this_color;
	text_color = i = define_color_class(NULL, 0, full_color);
	mouse_cursor = left_pointer;
}

color_class _color_setting_helper;

int xwin_custom_color(int old, char *title)
{
	int buf[4], win[4], i, color = old, drag_win = -1;
	for (i = 0; i < 4; i++)
		buf[i] = tmp_buf();
	on_exit for (i = 0; i < 4; i++) buf_delete(buf[i]);
	save_var bufnum = buf[0];
	bprintf("\nDrag the scroll bar for each color, press R, G, or B to type a value,\n");
	bprintf("or H to enter in hex. Enter to accept, C to cancel.\n\n\n");
	for (i = 0; i < XWIN_COLOR_BAR_HEIGHT; i++) {
//		if (i > 0 && i < XWIN_COLOR_BAR_HEIGHT)
			bprintf("                                                  %p          %p ",
					color_class _color_setting_helper, -1);
		bprintf("\n");
	}
	bprintf("\n");
// Maybe first show a selection of std colors based on the list in eel.h.

	point = 0;
	win[0] = add_popup(4, 1, 78, XWIN_COLOR_BAR_HEIGHT + 10,
					   ALL_BORD(BSINGLE), buf[0]);
	on_exit remove_window(win[0]);
	for (i = 1; i < 4; i++) {
		win[i] = add_popup(i * 14 - 4, 7, 10, XWIN_COLOR_BAR_HEIGHT,
						   BORD(BRIGHT, BSINGLE), buf[1]);
	}
	on_exit for (i = 0; i < 4; i++) remove_window(win[i]);
	int topline = 8 * y_pixels_per_char();
	int bottomline = (XWIN_COLOR_BAR_HEIGHT - 2) * y_pixels_per_char();

	save_var window_handle = win[1];
	window_title(win[0], BTOP, TITLECENTER, title);
	set_menu_look(win[0]);
	save_var normal_gui_cursor = GUI_CURSOR_SHAPE(1, 1, 1);
	for (;;) {
		xwin_update_color(win[1], GETRED(color), MAKE_RGB(255, 0, 0));
		xwin_update_color(win[2], GETGREEN(color), MAKE_RGB(0, 255, 0));
		xwin_update_color(win[3], GETBLUE(color), MAKE_RGB(0, 0, 255));
		window_handle = win[0];
		mouse_cursor = left_pointer;
		display_scroll_bar = 0;
		point = size();
		to_begin_line();
		delete(point, size());
		bprintf("      Red %3d       Green %3d     Blue %3d     This Color 0x%06X",
				GETRED(color),  GETGREEN(color), GETBLUE(color), color);
		point = 0;
		window_handle = (drag_win == win[1]) ? win[2] : win[1];
		set_color_pair(color_class _color_setting_helper, 0, color);
		refresh();
		getkey();
		if (user_abort)
			return old;
		key = toupper(key);
		save_var mouse_mask |= MASK_MOVE | MASK_UP;
		switch (key) {
			case MOUSE_LEFT_DN:
			case MOUSE_DBL_LEFT:
				drag_win = get_window_area();	// Note start of drag
				if (in_text_area(mouse_x, mouse_y) || mouse_x != w_right + 1)
					drag_win = -1;	// But only if it's on a scroll bar.
				break;
			case MOUSE_LEFT_UP:
				drag_win = -1;
				break;
			case MOUSE_MOVE:
				if (!mouse_pressed())
					break;
				i = (mouse_pixel_y - topline) * 256 / bottomline;
				i = MAX(i, 0);
				i = MIN(i, 255);
				if (drag_win == win[1])
					color = MAKE_RGB(i, GETGREEN(color), GETBLUE(color));
				else if (drag_win == win[2])
					color = MAKE_RGB(GETRED(color), i, GETBLUE(color));
				else if (drag_win == win[3])
					color = MAKE_RGB(GETRED(color), GETGREEN(color), i);
				break;
			case 'R':
				color = MAKE_RGB(get_number("Red 0-255: "), GETGREEN(color), GETBLUE(color));
				break;
			case 'G':
				color = MAKE_RGB(GETRED(color), get_number("Green 0-255: "), GETBLUE(color));
				break;
			case 'B':
				color = MAKE_RGB(GETRED(color), GETGREEN(color), get_number("Blue 0-255: "));
				break;
			case 'H':
				char resp[FNAMELEN], curhex[100];
				sprintf(curhex, "%06x", color);
				get_strdef(resp, "Hex RGB code for the color", curhex);
				if (*resp)
					color = numtoi_with_base(resp, 16);
				break;
			case ESC:
			case 'C':	return old;
			case 'S':
			case 'D':
			case 'Q':
			case '\n':
			case '\r':	return color;
		}
	}
}

// Insert a list of all color codes in all the defined color schemes.
// Sort and remove dups. Then add a standard palette at the start.
xwin_list_color_palette()
{
	int i, j, max = final_index(), r, g, b;
	save_var selected_color_scheme;
	for (i = 1; i <= max; i++)
		if (name_type(i) == NT_COLSCHEME) {
			selected_color_scheme = i;
			for (j = 0; j < number_of_color_classes(); j++)
				bprintf("%06x\n%06x\n",
						COLOR_STRIP_ATTR(get_foreground_color(j)),
						COLOR_STRIP_ATTR(get_background_color(j)));
		}
	do_sort_region(0, size(), 0, 0);
	point = 0;
	do_uniq(1, 1, 0);
	point = 0;	// Now add a 5-6-4 palette at the start.
	char *reds   = "\x00\x40\x80\xBF\xFF";
	char *greens = "\x00\x33\x66\x99\xCC\xFF";
	char *blues  = "\x00\x55\xAA\xFF";
	for (r = 0; r < 5; r++)
		for (g = 0; g < 6; g++)
			for (b = 0; b < 4; b++)
				bprintf("%02x%02x%02x\n", reds[r], greens[g], blues[b]);
}

int xwin_ask_color(int old, char *title)
{
	int color = old, buf, win, colors[160], i, j, k, max, x, y;
	char line[100];
	buf = tmp_buf();
	on_exit buf_delete(buf);
	save_var bufnum = buf;
	xwin_list_color_palette();
	point = 0;
	for (max = 0; max < sizeof(colors) / sizeof(*colors); max++) {
		if (!grab_line(bufnum, line))
			break;
		colors[max] = strtoi(line, 16);
	}
	delete(0, size());
	bprintf("\n  Click on a color, or press C or click Custom Color to "
			"choose another.\n\n");
	for (k = i = 0; i < 16 && k < max; i++) {
		bprintf("   ");
		for (j = 0; j < 16 && k < max; j++)
			bprintf("%p   %p ", define_color_class(NULL, 0, colors[k++]), 0);
		bprintf("\n\n");
	}
	bprintf("\n  Original color was: 0x%06x %p   %p          [Custom Color]",
			old, define_color_class(NULL, 0, old), 0);
	point = 0;
	win = add_popup(4, 0, 78, 30, ALL_BORD(BSINGLE), buf);
	window_title(win, BTOP, TITLECENTER, title);
	on_exit remove_window(win);
	mouse_cursor = left_pointer;
	display_scroll_bar = 0;
	save_var mouse_mask |= MASK_MOVE;
	save_var mouse_panning = 1;	// Otherwise we'd only get moves when button's down.
	for (;;) {
		refresh();
		getkey();
		if (user_abort)
			return old;
		key = toupper(key);
		x = (mouse_x - 8) / 4;
		y = (mouse_y - 4) / 2;
		i = y * 16 + x;
		switch (key) {
			case MOUSE_MOVE:
				if (i >= 0 && i < max)
					note("Color code: 0x%06x", colors[i]);
				break;
			case MOUSE_LEFT_DN:
			case MOUSE_DBL_LEFT:
				if (i < 0)
					break;
				if (i < max)
					return colors[i];
			case 'C':			// Fall through
				remove_window(win);
				return xwin_custom_color(old, title);
			case ESC:
				return old;
			case 'S':
			case 'D':
			case 'Q':
			case '\n':
			case '\r':
				return color;
		}
	}
	return old;
}


// Is the current key bound to a command that will search in our list?
is_color_search_key()
{
	char *s;
	return key == CTRL('S') || key == CTRL('R') || (s = name_name(root_keys[key]))
			&& string_matches_pattern(s, "(reverse-)?(incremental|regex"
									  "|string)-search");
}

// The current key is Ctrl-S or similar, so run it to search.  But
// first expand everything.
run_color_search_key(int hier)
{
	save_var point;
	if (hier)
		while (point = 0, re_search(1, " *<+>"))
			set_color_expand(1);
	restore_vars();
	run_topkey();
	ungot_key = -1;
	in_echo_area = 0;
}

// Replace tab with \n for this category if expand == 1, or vice versa
// if expand == 0, or toggle if expand == 2.  Return new class number.
int set_color_expand(int expand)
{
	int level, start, end;

	to_begin_line();
	if (expand == 2)	// Toggle expand/collapse.
		expand = parse_string(1, " *<+>") > 0;
	if (!expand)		// Collapse: find prev - line.
		if (parse_string(1, "<+>"))	// Can't collapse any more.
			return this_color_class_to_index();
		else if (!parse_string(1, " *<->"))	// Search for the right - line.
			re_search(-1, parse_string(1, "    ") ? "^  <->" :  "^<->");
	start = point;
	level = parse_string(1, " *");
	point += level;
	if (curchar() != (expand ? '+' : '-'))
		return this_color_class_to_index();
	replace(point, expand ? '-' : '+');
	if (expand) {		// Do this one line.
		save_var narrow_start = point, narrow_end = size() - give_end_line();
		end = give_end_line();
		if (!level)
			simple_re_replace(1, "\t  ![^ \n]", "\n  ");
		else
			simple_re_replace(1, "\t", "\n");
	} else { // Collapse: Find next at this level, then replace before there.
		save_var narrow_start = ++point;
		if (re_search(1, level ? "^ *[-+]" : "^[-+]"))
			to_begin_line();
		end = point - 1;
		save_var narrow_end = size() - point + 1;
		point = 0;
		simple_re_replace(1, "  -", "  +");
		simple_re_replace(1, "\n", "\t");
	}
	restore_vars();
	build_window();
	if (end > window_end)	// Make sure contents of expansion is visible.
		window_start = start;
	if (expand)				// Start on 1st item in category.
		nl_forward();
	return this_color_class_to_index();
}

// Write category for color class i into cat.  Categories may use | to
// indicate a nested subcategory.  Eventually each color class
// definition will include its category and description, but for now,
// use heuristics to assign a category.  We assume any new color classes
// will be for language modes.

set_color_give_category(int i, char *cat)
{
	char *s = name_color_class(i), *t;
	if (word_in_list(s, "|text|highlight|echo-area|mode-line|horiz-border"
					 "|vert-border|", 0))
		strcpy(cat, "General");
	else if (starts_with_in_list(s, "|show-matching-|", 0))
		strcpy(cat, "Special-Purpose Modes|Delimiters");
	else if (word_in_list(s, "|viewed-gui-text|dialog-gui-text"
							"|menu-highlight|", 0))
		strcpy(cat, "Dialogs");
	else if (word_in_list(s, "|menu|popup-title|popup-border|viewed-text"
						  "|viewed-border|viewed-link|selected-link|", 0))
		strcpy(cat, "Non-GUI Dialogs");
	else if (word_in_list(s, "|menu-bar|menu-bar-highlight|submenu"
						  "|submenu-highlight|", 0))
		strcpy(cat, "Non-GUI Menu Bar");
	else if (starts_with_in_list(s, "|debug-|overlap-|screen-|", 0)
			 || word_in_list(s, "|after-exiting|default|ident-completion"
							 "|pull-highlight|", 0))
		strcpy(cat, "Rarely Modified");
	else if (starts_with_in_list(s, "|html-|", 0))	// More descriptive names.
		strcpy(cat, "Language Modes|HTML/XML Modes");
	else if (starts_with_in_list(s, "|vbasic-|", 0))
		strcpy(cat, "Language Modes|Visual Basic Mode");
	else if (starts_with_in_list(s, "|tex-|", 0))
		strcpy(cat, "Language Modes|TeX/LaTeX Modes");
	else if (strchr(s, '-')) {
		sprintf(cat, "%s Modes|%s", starts_with_in_list(s,
					"|process-|diff-|grep-|dired-|man-|hex-|info-|spell-|", 0)
				? "Special-Purpose": "Language", s);
		t = strchr(cat, '|') + 1;
		*t = toupper(*t);
		if (starts_with_in_list(t, "|gams-|css-|", 1))
			for (; isalpha(*t); t++)
				*t = toupper(*t);
		t = strchr(t, '-');
		if (t)
			strcpy(t, " Mode");
	} else
		strcpy(cat, "Other");
}

// Create class list in the current buffer.
make_color_class_list(int expanded)
{
	int i, classes = number_of_color_classes();

	// First, set up a good top-level order.
	stuff("-General\n-Language Modes\n-Special-Purpose Modes\n-Dialogs\n");
	stuff("-Non-GUI Dialogs\n-Non-GUI Menu Bar\n");
	add_region(point_spot, point_spot, color_class menu_highlight,
			   REGLINE);
	tab_size = 100;		// Hide rest of tree.
	save_var narrow_start, narrow_end;
	for (i = 0; i < classes; i++) {
		char cat[FNAMELEN], pat[FNAMELEN], *s, firstcat[FNAMELEN];
		if (*name_color_class(i) == '-')
			continue;
		set_color_give_category(i, cat);
		point = 0;
		strcpy(firstcat, cat);
		if (s = strchr(firstcat, '|'))
			*s = 0;
		sprintf(pat, "^-%q\n", firstcat);
		if (!re_search(1, pat))
			bprintf("-%s\n", firstcat);
		narrow_start = point;
		if (re_search(1, "^-"))
			to_begin_line();
		narrow_end = size() - point;
		if (s = strchr(cat, '|')) {
			sprintf(pat, "^  -%q\n", s + 1);
			point = 0;
			if (!re_search(1, pat))
				bprintf("  -%s\n", s + 1);
			else
				re_search(1, "^!([^ \n]|  [^ \n])");
			bprintf("  ");
		}
		narrow_start = narrow_end = 0;
		bprintf("  %s\n", name_color_class(i));
	}
	if (!expanded) {	// Now collapse all categories.
		delete(size() - 1, size());
		simple_re_replace(1, "^  -", "  +");
		simple_re_replace(1, "^-", "+");
		simple_re_replace(1, "\n ", "\t ");
	}
}

// If we find color class name, expand it and return 1.  Call this
// repeatedly to expand each level and reposition.
int set_color_search_expand(char *name)
{
	char pat[FNAMELEN];
	if (!name || !*name)
		return 0;
	sprintf(pat, "(^|\t)[-+ ]*%q($|\t)", name);
	point = 0;
	if (!re_search(1, pat))
		return 0;
	set_color_expand(1);
	return 1;
}	

// Insert key help, using current line from class window.
set_color_explain_keys(char *class_line)
{
	if (class_line[0] == ' ' && strchr("-+", class_line[2]))
		bprintf("\nType N or P to set other color classes, + to expand "
				"or - to collapse\na category, T to try out these colors "
				"temporarily, S to set these\n"
				"colors permanently, or C to cancel.  ");
	else if (class_line[0] == '+')
		bprintf("\nType N or P to set other color classes, + to expand "
				"a category, T to\ntry out these colors temporarily, "
				"S to set these colors permanently,\n"
				"or C to cancel.  ");
	else
		bprintf("\nType N or P to set other color classes, - to collapse "
				"a category, T to\ntry out these colors temporarily, "
				"S to set these colors permanently,\n"
				"or C to cancel.  ");
}

command set_color()		/* query user to set colors */
{
	int class, color, i, j, *orig, classes, docbuf, wantwin, cmd;
	int swin, ccwin, cwin = 0, infowin, newcolor, newscheme, orig_scheme;
	int *orig_fg, *orig_bg, *orig_schemes;
	int fg_color, bg_color;
	char buf[100], class_line[100], *s;
	int sbuf, ccbuf, cbuf, infobuf, orig_win = window_handle, win, page;
	char *new, *old;
	jmp_buf this_level;

	int orig_class = get_character_syntax_color(point);
	save_var abort_searching = 0, shift_selects = 0;
	save_var top_level = &this_level;
	classes = number_of_color_classes();

	sbuf = tmp_buf();	/* list of color schemes */
	bufnum = sbuf;	/* now highlight current line in buffers */
	add_region(point_spot, point_spot, color_class menu_highlight,
			   REGLINE);
	make_color_scheme_list(sbuf);

	orig_scheme = selected_color_scheme;
	record_color_settings(&orig_fg, &orig_bg, &orig_schemes);
	selected_color_scheme = orig_scheme;
	orig = (int *) malloc(classes * sizeof(int));
	for (i = 0; i < classes; i++)
		orig[i] = alter_color(i, -1);

	class = iter = page = 0;

	bufnum = docbuf = tmp_buf();	// read help text for classes into buf
	on_exit buf_delete(docbuf);
	if (new = lookpath("colclass.txt"))
		file_read(new, FILETYPE_AUTO);
	ccbuf = tmp_buf();	/* list of color classes */
	cbuf = tmp_buf();	/* window with colors */ 
	infobuf = tmp_buf();	/* help buffer with options */ 

	bufnum = ccbuf;
	make_color_class_list(0);

winagain:
	swin = add_popup(4, 1, 20, 10, ALL_BORD(BSINGLE), sbuf);
	window_title(swin, BTOP, TITLECENTER, " Color Schemes ");
	set_menu_look(swin);

	ccwin = add_popup(28, 1, 25, 10, ALL_BORD(BSINGLE), ccbuf);
	display_column = 0;		/* prevent continuation lines */
	window_title(ccwin, BTOP, TITLECENTER, " Color Classes ");
	set_menu_look(ccwin);
	point = 0;
	set_color_expand(1);	// Expand first category.
	if (orig_class >= 0 && set_color_search_expand(name_color_class(orig_class))
		  && set_color_search_expand(name_color_class(orig_class))
		  && set_color_search_expand(name_color_class(orig_class)))
		class = this_color_class_to_index();

	if (!(has_feature & FEAT_ANYCOLOR)) {		// Matrix of color choices.
		cwin = add_popup(56, 1, 18, 10, ALL_BORD(BSINGLE), cbuf);
		window_title(cwin, BTOP, TITLECENTER, " Colors ");
		invisible_window = 1;
		insert('\n');
		if (selectable_colors > 128)
			window_title(cwin, BBOTTOM, TITLERIGHT(1), " More ");
	}

	infowin = add_popup(2, 14, 75, 9, _view_border, infobuf);
	display_column = 0;		/* prevent continuation lines */
	mouse_cursor = left_pointer;
	text_color = is_gui ? color_class text : color_class menu;
	display_scroll_bar = 0;
	newscheme = 2;
	if (has_feature & FEAT_COLOR_DIALOG) {
		display_dialog_box("SetColorBox", "Set colors",
						   swin, ccwin, infowin, "OK", "Cancel", "Help");
		save_var permit_window_keys = PERMIT_SCROLL_KEY;
		save_var return_raw_buttons = 1;
	}
	wantwin = ccwin;

	if (setjmp(top_level))	/* come here if subcommand's aborted */
		;
	for (;;) {
		if (class >= 0) {
			color = alter_color(class, -1);
			fg_color = get_foreground_color(class, 0);
			bg_color = get_background_color(class, 0);
			if ((color >= 128) != page && !newscheme)
				newscheme = 1;
			page = (color >= 128);
		}
		if (newscheme) {
			if (!(has_feature & FEAT_ANYCOLOR)) {	// Sample window has 2 pages.
				window_handle = cwin;
				point = page;
			}
			window_handle = swin;
			if (newscheme == 2) {
				make_color_scheme_list(bufnum);
				delete(size() - 1, size());	/* no final \n */
			}
			point = 0;
			col_search(name_name(selected_color_scheme), 0);
		}
		if (!(has_feature & FEAT_ANYCOLOR)) {
			j = page ? 128 : 0;
			for (i = 0; i < 128; i++, j++)
				term_write_attr(SCRPLACE(i), '+',
								(class >= 0) ? j : 0);
		}
		window_handle = ccwin;
		if (class >= 0) {
			point = 0;
			sprintf(buf, "^[-+ ]*%q$", name_color_class(class));
			re_search(1, buf);
		}
		to_begin_line();
		parse_string(1, "[^\n\t]*", class_line);
		to_begin_line();
		buf_zap(infobuf);
		window_handle = infowin;
		if (!(has_feature & FEAT_COLOR_DIALOG))
			bprintf("Type U or D to change the color scheme, M to "
					"make a new color scheme.\n");
		if (class >= 0) {
			if (is_unix == IS_UNIX_XWIN) {
				bprintf("Type F or B to set foreground or background colors, A to set attributes\n");
				bprintf("like bold or italic for the %s color class. Sample text:\n",
						name_color_class(class));
			} else if (!(has_feature & FEAT_COLOR_DIALOG)) {
				bprintf("Or use the arrow keys to change the "
						"%s color class.\n", name_color_class(class));
			} else
				bprintf("Sample text using the %s color class:\n",
						name_color_class(class));
			i = point;
		}
		bufnum = docbuf;
		point = 0;
		s = class_line;
		while (strchr(" -+", *s))
			s++;
		sprintf(buf, "^%q\t+", s);
		if (re_search(1, buf))
			buf_xfer(infobuf, point, give_end_line());
		else if (class >= 0)
			buf_printf(infobuf, "This is sample text.");
		bufnum = infobuf;
		j = point + 1;
		if (!(has_feature & FEAT_COLOR_DIALOG)) {
			set_color_explain_keys(class_line);
			if (class >= 0)
				bprintf("Color was initially 0x%02x, this color is 0x%02x.",
						orig[class], color);
		} else if (string_matches_pattern(class_line, " *<+>"))
			bprintf("\nPress + to expand or - to collapse a category.");
		else if (string_matches_pattern(class_line, " *-"))
			bprintf("\nPress - to collapse this category.");
		if (class >= 0)
			set_character_color(i, j, class);
		point = 0;

		// Disable 2 color-setting buttons and 3 font style checkboxes only.
		disable_dialog_controls = (class < 0) ? 0x1807 : 0;
		dialog_checkboxes
				= (class < 0) ? 0 : (((fg_color & EFONT_BOLD) ? 1 : 0)
									 | ((fg_color & EFONT_ITALIC) ? 2 : 0)
									 | ((fg_color & EFONT_UNDERLINED) ? 4 : 0));

		if (has_feature & FEAT_ANYCOLOR) {
			window_handle = wantwin;
			maybe_refresh();
			getkey();
			wantwin = window_handle;
		} else {
			window_handle = cwin;
			do {
				maybe_refresh();
				if (class >= 0)
					term_write_attr(SCRPLACE(color), 'C', color);
				full_getkey("Your choice: ", 0);
				if (class >= 0)
					term_write_attr(SCRPLACE(color), '+', color);
			} while (!IS_TRUE_KEY(key) && !IS_MOUSE_DOWN(key));
		}
		key = toupper(key);
		newcolor = -100;	// Assume it's an invalid key.
		newscheme = 0;
		cmd = binding_of(key, reg_tab);
		if (cmd == (short) next_page || cmd == (short) previous_page
			  || cmd == (short) goto_beginning || cmd == (short) goto_end) {
			do_command(cmd);
			newcolor = -99;		// Mark as a valid key.
		}
		if (has_feature & FEAT_COLOR_DIALOG) {
			window_handle = ccwin;	// Scrolling might have changed these.
			class = this_color_class_to_index();

			window_handle = swin;
			grab_full_line(bufnum, buf);
			i = find_index(buf);
			if (i > 0 && name_type(i) == NT_COLSCHEME)
				selected_color_scheme = i;
		}
		if (key_is_button == 4 && class >= 0) {
			i = comm_dlg_color(fg_color, "Foreground color");
			if (i != -1)
				set_color_pair(class, i | (fg_color & EFONT_STYLE), bg_color);
		} else if (key_is_button == 5 && class >= 0) {
			i = comm_dlg_color(bg_color, "Background color");
			if (i != -1)
				set_color_pair(class, fg_color, i);
		} else if (key_is_button == 8 && class >= 0) {
			set_color_pair(class, fg_color ^= EFONT_BOLD, bg_color);
		} else if (key_is_button == 9 && class >= 0) {
			set_color_pair(class, fg_color ^= EFONT_ITALIC, bg_color);
		} else if (key_is_button == 10 && class >= 0) {
			set_color_pair(class, fg_color ^= EFONT_UNDERLINED, bg_color);
		} else if (key == 'C' || key == 'Q' || key == ESC || user_abort
				   || key_is_button == 2) {
			restore_color_settings(orig_fg, orig_bg, orig_schemes);
			selected_color_scheme = orig_scheme;
			say("");
			break;
		} else if (key == 'S' || key == ' ' || IS_ENTER_KEY(key)
				   || key_is_button == 1) {
			preserve_color_settings(orig_fg, orig_bg, orig_schemes,
									orig_scheme);
			break;
		} else if (IS_MOUSE_DOWN(key)) {
			win = window_at_coords(mouse_y, mouse_x, mouse_screen);
			if (!is_window(win))
				continue;
			window_handle = win;
			get_text_area(win);
			if (win == cwin) {
				if (bar_on_border()) {
					i = scroll_bar_code();
					if ((i ? (i > 0) : !page) != point)
						newcolor = color ^ 128;
				} else {
					newcolor = (mouse_y - COLOR_LINE) * 16;
					newcolor += (mouse_x - COLOR_COL);
					if (page)
						newcolor += 128;
				}
			} else if (in_text_area(mouse_x, mouse_y))
				point_to_pos(mouse_x, mouse_y, mouse_screen);
			else if (display_scroll_bar && bar_on_border())
				mouse_scroll_bar();
			else
				continue;
			if (win == ccwin) {
				class = this_color_class_to_index();
				if (class < 0)
					class = set_color_expand(2);
			} else if (win == swin)
				newscheme = 1;
			else if (win == infowin && isupper(curchar())
					 && character(point - 1) == ' '
					 && character(point + 1) == ' ')
				ungot_key = curchar();
			if (win != infowin)
				wantwin = win;
		} else if ((has_feature & FEAT_COLOR_DIALOG)
				   && full_key == NUMSHIFT(GREYTAB)) {
			prev_dialog_item();
			wantwin = window_handle;
			continue;
		} else if ((has_feature & FEAT_COLOR_DIALOG) && generic_key(key) == '\t') {
			next_dialog_item();
			wantwin = window_handle;
			continue;
		} else if ((has_feature & FEAT_COLOR_DIALOG) && key == GETFOCUS) {
			win = window_at_coords(mouse_y, mouse_x, mouse_screen);
			if (!is_window(win) || win == infowin)
				continue;
			wantwin = win;
		}
		if (!(has_feature & FEAT_ANYCOLOR))
			switch (get_direction()) {
				case BTOP:		newcolor = color + 256 - 16; break;
				case BBOTTOM:	newcolor = color + 16; break;
				case BLEFT:		newcolor = color + 256 - 1; break;
				case BRIGHT:	newcolor = color + 1; break;
			}
		else
			switch (get_direction()) {
				case BLEFT:
					if (wantwin == ccwin && class == -2) {
						key = '-';
						break;
					}
				case BTOP: 
					key = (wantwin == ccwin) ? 'P' : 'U'; break;
				case BRIGHT:
					if (wantwin == ccwin && class == -3) {
						key = '+';
						break;
					}						
				case BBOTTOM:
					key = (wantwin == ccwin) ? 'N' : 'D'; break;
			}
		if ((has_feature & FEAT_COLOR_DIALOG) && key == 'N' && win == swin)
			key = 'D';
		if ((has_feature & FEAT_COLOR_DIALOG) && key == 'P' && win == swin)
			key = 'U';
		switch (key) {
			case 'N': 
				window_handle = ccwin;
				to_end_line();
				if (!nl_forward())
					point = 0;
				class = this_color_class_to_index();
				break;
			case 'P':
				window_handle = ccwin;
				if (!re_search(-1, "^.+"))
					point = size();
				class = this_color_class_to_index();
				break;
			case 'T':
				show_binding("Continue changing colors "
					"by typing %s", "exit-level");
				remove_window(swin);
				remove_window(ccwin);
				if (!(has_feature & FEAT_ANYCOLOR))
					remove_window(cwin);
				remove_window(infowin);
				abort_searching = 1, shift_selects = 1;
				recursive_edit();
				abort_searching = 0, shift_selects = 0;
				goto winagain;
				break;
			case 'D':
				wantwin = window_handle = swin;
				if (!nl_forward())
					point = 0;
				newscheme = 1;
				break;
			case 'U':
				wantwin = window_handle = swin;
				if (!nl_reverse())
					point = size();
				newscheme = 1;
				break;
			case 'M':
				wantwin = swin;
				maybe_refresh();
				get_strpopup(buf, "New scheme name: ", "", "set-color");
				if (*buf) {
					new_variable(buf, NT_COLSCHEME, 0, 0);
					new = varptr(find_index(buf));
					old = varptr(selected_color_scheme);
					for (i = 0; i < ptrlen(old, 1); i++)
						*new++ = *old++;
					newscheme = 2;
				}
				break;
			default:
				if (is_unix == IS_UNIX_XWIN && class >= 0) {
					if (key == 'F') {
						i = xwin_ask_color(fg_color, " Set Foreground Color ");
						if (i != -1)
							set_color_pair(class, i | (fg_color & EFONT_STYLE), bg_color);
						break;
					} else if (key == 'B') {
						i = xwin_ask_color(bg_color, " Set Background Color ");
						if (i != -1)
							set_color_pair(class, fg_color, i);
						break;
					} else if (key == 'A') {
						fg_color = xwin_ask_attribute(class, fg_color, bg_color);
						break;
					}
					screen_messed();
				}
				if (cmd == (short) help || key == 'h' || key == '?'
					  || key_is_button == 3) {
					if (!(has_feature & FEAT_COLOR_DIALOG))
						say("Press <Space> to continue.");
					help_on_command("set-color");
					if (!(has_feature & FEAT_COLOR_DIALOG))
						newscheme = 1;
				} else if (is_color_search_key()) {
					window_handle = (has_feature & FEAT_COLOR_DIALOG)
									? wantwin : ccwin;
					run_color_search_key(window_handle == ccwin);
					if (window_handle == ccwin)
						class = this_color_class_to_index();
					else if (window_handle == swin)
						newscheme = 1;
				} else if ((key == '+' || key == '-') && (wantwin == ccwin
					|| !(has_feature & FEAT_COLOR_DIALOG))) {
					window_handle = ccwin;
					class = set_color_expand(key == '+');
				} else if (IS_TRUE_KEY(key) && newcolor == -100) {
					maybe_ding(bell_on_bad_key);
					newscheme = 1;
				}
				break;
		}
		if (newcolor > -99 && class >= 0)	/* did it change? */
			alter_color(class, newcolor % selectable_colors);
		else if (newscheme) {
			window_handle = swin;
			grab_full_line(bufnum, buf);
			i = find_index(buf);
			if (i > 0 && name_type(i) == NT_COLSCHEME)
				selected_color_scheme = i;
		}
	}
	remove_window(swin);
	remove_window(ccwin);
	if (!(has_feature & FEAT_ANYCOLOR))
		remove_window(cwin);
	remove_window(infowin);
	buf_delete(ccbuf);
	buf_delete(cbuf);
	buf_delete(infobuf);
	if (is_window(orig_win))
		window_handle = orig_win;
	buf_delete(sbuf);
	full_redraw = 1;
	*get_color_scheme_variable() = selected_color_scheme;
	free(orig);
	free(orig_fg);
	free(orig_bg);
}
