/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1985, 2020 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// This file has dired mode.

#include "eel.h"
#include "dired.h"
#include "colcode.h"
#include "epsnet.h"
#include "grep.h"

when_loading()		/* disable self-inserting chars */
{
	fix_key_table(reg_tab, (short) normal_character, dired_tab, -1);
	set_case_indirect(dired_tab);
}

dired_help() on dired_tab['h'], dired_tab['?']
{
	help_on_command("dired-mode");
}

// If the buffer was manually switched to dired mode, it may lack a
// proper pattern.
char *get_dired_buffer_pattern()
{
	return dired_buffer_pattern ? dired_buffer_pattern : bufname;
}

// Given a prefix like "dired-isdir-", returns an index to
// the subroutine dired-isdir-unix when the buffer has a Unix-style
// directory list, dired-isdir-simple when the buffer has that format
// of directory list, or NULL if the buffer isn't from an FTP or the
// appropriate function doesn't exist.  Func is nonzero if the item
// should be a subroutine, or 0 if it should be a variable.

int get_dired_item(char *prefix, int func)
{
	char item[100];
	int i;

	strcpy(item, prefix);
	switch (dired_format) {
		case DF_UNIX:		strcat(item, "unix"); break;
		case DF_UNIX_FULL:	strcat(item, "unixfull"); break;
		case DF_SIMPLE:		strcat(item, "simple"); break;
		case DF_OLDNT:		strcat(item, "oldnt"); break;
		case DF_VMS:		strcat(item, "vms"); break;
		default:			return 0;
	}
	i = find_index(item);
	switch (name_type(i)) {
		case NT_COMMAND: case NT_SUBR:
		case NT_AUTOLOAD: case NT_AUTOSUBR:
			if (func)
				return i;
			break;
		case NT_VAR: case NT_BUFVAR:
		case NT_WINVAR:
			if (!func)
				return i;
			break;
	}
	return 0;
}

dired_set_hiding() on dired_tab['-']
{
	if (dired_force_show_dotfiles) {
		if (ask_yn("Hide files",
				   "Hide files starting with \".\"? ", "Yes", "No", 1)) {
			dired_force_show_dotfiles = 0;
			dired_refresh();
			say("Press - to show.");
		}
	} else if (ask_yn("Show hidden files",
					  "Show hidden files starting with \".\"? ", "Yes", "No", 1)) {
		dired_force_show_dotfiles = 1;
		dired_refresh();
		say("Press - to hide.");
	}
}

dired_examine() on dired_tab[' '], dired_tab['\n'],
		   dired_tab['e'], dired_tab['\r']
{
	char fname[FNAMELEN];

	to_begin_line();
	if (parse_string(1, OMITLINEPAT)) {
		dired_set_hiding();
		return;
	}
	dired_getname(fname, 1);
	do_dired_examine(fname, dired_isdir());
	iter = 0;
}

dired_examine_deleting() on dired_tab['E']
{
	int buf = bufnum;
	dired_examine();
	if (is_dired_buf() && buf && buf != bufnum)
		buf_delete(buf);
}

dired_parent() on dired_tab['^'], dired_tab['.']
{				/* view parent directory */
	char fname[FNAMELEN];

	to_begin_line();
	dired_try_getdir(fname, 1);
	strcat(fname, "..");
	do_dired_examine(fname, 1);
	iter = 0;
}

// DF_UNIX is for remote and local files; DF_UNIX_FULL is our special
// local format for when dired-layout bit 32 is on (full dates).
int dired_is_unix_format()
{
	return dired_format == DF_UNIX || dired_format == DF_UNIX_FULL;
}

do_dired_examine(fname, isdir)
char *fname;
{
	char file[FNAMELEN], *p;

	save_var case_fold = !dired_is_unix_format();
	save_var file_pattern_wildcards = 0;	// []'s etc in file name are literal.
	strcpy(file, fname);
	absolute(file);
	if (isdir == 3)	// symbolic link to a file
		if (!is_remote_file(file) || !is_ftp_url_a_dir(file))
			isdir = 0;	// if remote, make sure it's not really a dir
		else
			isdir = 1;
	if (isdir) {
		if (is_remote_file(file) && isdir == 1
			&& is_path_separator(file[strlen(file) - 1]) != 1)
			strcat(file, "/");	// If dir, not link, add /.
		if (dired_one(file) || is_remote_file(file))
			return;
		strcpy(file, fname);
		p = get_tail(file, 0);	/* for .., go to right place */
		if (p > file && strcmp(p, "..") == 0) {
			p[-(opsys != OS_UNIX)] = 0;	// Remove / too unless Unix.
			p = get_tail(file, 0);
			strcat(p, "\n");
			col_search(p, goal_column);
			move_to_column(goal_column);
			build_first = 1;
		}
	} else			/* regular file */
		ask_find_it(file);
}

dired_examine_in_window() on dired_tab['1'], dired_tab['2'],
				dired_tab['5'], dired_tab['z']
{		/* do appropriate window cmd, then examine */
	int olditer = iter, oldarg = has_arg;

	has_arg = iter = 0;
	do_command(cx_tab[key]);
	iter = olditer, has_arg = oldarg;
	ungot_key = 'e';
}

dired_examine_in_next_window() on dired_tab['o']
{
	char fname[FNAMELEN];
	int isdir;

	dired_getname(fname, 1);
	isdir = dired_isdir();
	if (number_of_user_windows() == 1)
		window_split(HORIZONTAL);
	next_user_window(1);
	do_dired_examine(fname, isdir);
	iter = 0;
}

dired_refresh() on dired_tab['r']
{
	char msg[FNAMELEN];
	int yes, remote;

	if (modified) {
		sprintf(msg, "Relist %s, discarding markings? ", bufname);
		yes = ask_yn("Discard markings", msg, "Yes", "No", 1);
		check_abort();
		if (!yes)
			return;
	}
	remote = is_remote_file(get_dired_buffer_pattern());
	if (remote && (ftp_job || is_busy_doing_scp(bufnum) != -1))
		error("Still retrieving directory listing in this buffer.");
	if (remote)
		point = 0;
	else
		save_var point, window_start;
	save_var file_pattern_wildcards = dired_buffer_pattern_wildcards;
	create_dired_listing(get_dired_buffer_pattern(), 1);
	restore_vars();
	if (!remote)
		move_to_column(goal_column);
	dired_orig_position = point;
	fix_window_start();
	if (window_start >= size())
		window_start = prev_screen_line(text_height() / 2);
}

dired_create_directory() on dired_tab['+']
{
	save_var _show_help_on = "dired-mode";
	do_create_directory(1);
}

do_create_directory(int in_dired)
{
	char dir[FNAMELEN], cmd[FNAMELEN], url[FNAMELEN];

	if (!prompt_with_buffer_directory)
		save_var prompt_with_buffer_directory = PBD_INSERT;
	get_file_dir(dir, "Create directory: ");
	if (is_remote_file(dir)) {
		if (in_dired)
			dired_getdir(url);
		else
			strcpy(url, dir);
		sprintf(cmd, "MKD %s", get_url_file_part(dir, 0));
		ftp_misc_operation(url, cmd);
		say("Operation queued.");
	} else {
		if (mkdir(dir))
			file_error(errno, dir, "couldn't create");
		if (is_dired_buf())
			dired_refresh();
	}
}

// Copy src to dest, putting \ before shell meta chars.
quote_for_shell(char *dest, char *src)
{
	if (opsys != OS_UNIX) {
		sprintf(dest, "\"%s\"", src);
		return;
	}
	while (*src) {
		if (!isalnum(*src) && !strchr("-~/", *src))
			*dest++ = '\\';
		*dest++ = *src++;
	}
	*dest = 0;
}

do_shell_command(char *file, char *dir)
{
	char cmd[FNAMELEN], cmdline[FNAMELEN], buf[FNAMELEN], pr[FNAMELEN], *s;
	int sync = 1;

	sprintf(pr, "! on %s: ", get_tail(file, 0));
	get_string(cmd, pr);
	if (strlen(cmd) > 1 && cmd[strlen(cmd) - 1] == '&') {
		cmd[strlen(cmd) - 1] = 0;
		sync = 0;
	}
	sprintf(buf, "shell-output-%s-%s", cmd, get_tail(file, 0));
	if (!strchr(cmd, '*'))
		strcat(cmd, " *");
	strcpy(cmdline, cmd);
	strcpy(strchr(cmdline, '*'), " ");
	quote_for_shell(cmdline + strlen(cmdline), file);
	strcat(cmdline, strchr(cmd, '*') + 1);
	while (s = strchr(buf, '"'))	// No " in buffer names.
		*s = '\'';
	if (opsys == OS_UNIX || !sync) {
		pipe_text(NULL, buf, cmdline, dir,
				  (sync ? PIPE_SYNCH : 0) | PIPE_CLEAR_BUF | PIPE_NOREFRESH, 0);
		if (buffer_size(buf)) {
			to_buffer(buf);
			point = 0;
		}
	} else
		do_push(cmdline, 1, 1);
}

dired_shell_command() on dired_tab['!']
{
	char dir[FNAMELEN], file[FNAMELEN];

	save_var _show_help_on = "dired-mode";
	dired_getname(file, 1);
	dired_getdir(dir);
	do_shell_command(file, dir);
}

dired_sort() on dired_tab['s']
{
	char *p, msg[200];
	int rev, sort;

	if (dired_format == DF_VMS)
		error("Sorting VMS directories not supported.");
	rev = isupper(dired_sorts_files);
	switch (sort = toupper(dired_sorts_files)) {
		case 'N': case 'D': case 'S':  case 'E':
			break;
		case 0: case 'U':
			sort = 'U', rev = 0; break;
		default:
			sort = 'N';
	}
	for (;;) {
		strcpy(msg, "Name, Ext, Size, Date, Unsorted,");
		p = index(msg, sort);
		p = index(p, ',');
		msg[strlen(msg) - 1] = 0;
		sayput("Sort by %.*s (now)%s; Reverse; %sGroup dirs"
			   " [NESDURG] ", p - msg, msg, p,
			   dired_groups_dirs ? "un" : "");
		getkey();
		check_abort();
		switch (toupper(key)) {
			case 'N':
			case 'D':
			case 'S':
			case 'E': sort = toupper(key); break;
			case 'U': say("Future listings will be unsorted.");
					  sort = 0; break;
			case 'R': rev = !rev; break;
			case '+': rev = 0; break;
			case '-': rev = 1; break;
			case 'G': dired_groups_dirs = !dired_groups_dirs;
					  break;
			default:  if (IS_ENTER_KEY(key))
						  break;
					  else if (binding_of(key, root_keys) == (short) help
							   || key == 'h' || key == '?')
						  help_on_command("dired-sort");
					  continue;
		}
		break;
	}
	dired_sorts_files = rev ? toupper(sort) : tolower(sort);
	if (sort) {
		sprintf(msg, "%s%s, %s", rev ? "Reverse " : "",
				(sort == 'N') ? "Name" : (sort == 'E') ? "Extension"
			: (sort == 'S') ? "Size" : "Date",
				dired_groups_dirs ? "grouping directories at top"
								  : "not grouping directories");
		noteput("Sorting...");
		sort_for_dired();
		say("Sorted by %s.", msg);
	} else
		say("Future listings will be unsorted.");
	maybe_refresh();
}

dired_show_live_link(char *file, char *msg)
{
	struct file_info fi;

	save_var readonly_warning |= ROWARN_GREP, no_popup_errors = 1;
	readonly_warning &= ~(ROWARN_MSG | ROWARN_BELL);
	save_var bell_on_read_error = 0;
	if (!*msg)
		if (check_file(file, &fi) != CHECK_FILE)
			sprintf(msg, "Can't read %s: ", get_tail(file, 1));
		else if (fi.fsize > dired_live_link_limit)
			sprintf(msg, "%s too big: ", get_tail(file, 1));
	set_read_only(0);
	filename = "";
	zap(bufname);
	fundamental_mode();
	if (!*msg) {
		if (do_file_read(file, FILETYPE_AUTO))
			sprintf(msg, "Can't read %s: ", get_tail(file, 1));
		else {
			unique_filename_identifier(file, unique_file_identifier);
			file_io_converter = 0;
			file_convert_read(FILE_CONVERT_QUIET);
			call_mode(filename);
			check_filevars();
		}
	}
	modified = 0;
	set_window_wrap(wrap_dired_live_link);	// Binary files can be slow to
	make_mode();							// display unless wrapped.
	set_read_only(1);
	buffer_not_saveable = 1;
	force_save_as = FSA_NEWFILE;
	auto_read_changed_file = 1;
	if (!*msg)
		sprintf(msg, "%s: ", get_tail(file, 1));
	set_mode_message(msg);
}

dired_when_idle(int cnt)
{
	char file[FNAMELEN], msg[FNAMELEN];
	int win;
	jmp_buf this_level;

	if (!cnt)			// On our first call when idle, do nothing,
		return 0;		// but ask to be called again.
	if (cnt > 1)		// On our third or later call, do nothing and
		return -1;		// say we don't need to be called again.
	if (!exist(LIVE_LINK_BUF))
		return -1;
	win = is_buffer_in_window(LIVE_LINK_BUF);
	if (win == -1)
		return -1;
	*file = *msg = 0;
	save_var point;
	if (!dired_getname(file, 0) || is_remote_file(file))
		strcpy(msg, "none: ");	// Say why we're not displaying a file.
	else if (is_directory(file))
		strcpy(msg, "dir: ");
	save_var window_handle = win;
	if (!strcmp(filename, file) || modified)
		return -1;	// If file name's unchanged, or user modified buf, stop.
	save_var top_level = &this_level;
	if (!setjmp(top_level))
		dired_show_live_link(file, msg);
	restore_vars();
	refresh();
	return -1;
}

dired_live_link() on dired_tab['l']
{
	iter = 0;
	add_buffer_when_idle(bufnum, dired_when_idle);	// Ensure it's set up.
	if (is_buffer_in_window(LIVE_LINK_BUF) == -1) {
		save_var window_handle;
		if (number_of_user_windows() == 1)
			window_split(has_arg ? VERTICAL : HORIZONTAL);
		else
			to_next_user_window(1);
		to_buffer(LIVE_LINK_BUF);
	}
}

char *dired_skip_line_pat()
{
	if (dired_format == DF_VMS)
		return "^(     Directory of |  total [0-9]+|  .*:(%[.*%])?$|\n"
				"|  Directory .*:%[|  Total of .* (files|blocks)"
				"|                    |  USR%$DISK:"
				"|" OMITLINEPAT ")";
	else
		return "^(     Directory of |  total [0-9]+|\n"
				"|  Directory .*:%[|  Total of .* (files|blocks)"
				"|                    "
				"|" OMITLINEPAT ")";
}

dired_search() on dired_tab['f']
{
	if (!is_highlight_on()) {
		save_var _highlight_control;
		highlight_off();
		save_var restrict_search_type = REGRECT;
		set_spot(&restrict_search.start, 0);
		set_spot(&restrict_search.end, size());
		restrict_search.colstart = goal_column;
		restrict_search.colend = 10000;
		int h = add_region(restrict_search.start, restrict_search.end,
						   color_class pull_highlight, REGRECT);
		on_exit remove_region(h);
		modify_region(h, MRSTARTCOL, restrict_search.colstart);
		modify_region(h, MRENDCOL, restrict_search.colend);
	}
	incremental_search();
}

dired_repl(newchar)
{
	to_begin_line();
	if (parse_string(1, dired_skip_line_pat()))
		nl_forward();
	if (index("DCMULRG ", curchar()) && character(point + 1) == ' ')
		replace(point, newchar);
	save_var iter = 1;		// call us to do each repetition
	dired_down_line();
}

dired_mark_for_delete() on dired_tab['d']
{
	char fname[FNAMELEN];

	dired_getname(fname, 1);
	if (!strcmp(get_tail(fname, 0), ".."))
		error("Edit the parent directory to delete it.");
	dired_repl('D');
}

dired_mark_for_copy() on dired_tab['c']
{
	if (no_remote_copy(get_dired_buffer_pattern()))
		error("Copy operation not implemented for remote files.");
	int isdir = dired_isdir();
	if (isdir == 1 || isdir == 2)
		error("Can't copy directories.");
	dired_repl('C');
}

dired_mark_for_move() on dired_tab['m']
{
	dired_repl('M');
}

dired_mark_for_uppercase() on dired_tab['U']
{
	dired_repl('U');
}

dired_mark_for_lowercase() on dired_tab['L']
{
	dired_repl('L');
}

dired_mark_for_regex_rename() on dired_tab['R']
{
	dired_repl('R');
}

dired_unmark() on dired_tab['u']
{
	dired_repl(' ');
}

dired_do_uppercase(name, to, in_dired)	/* return 1 if error */
char *name, *to;
{
	char newname[FNAMELEN];

	strcpy(newname, name);		// Rename file to uppercase.
	for (to = get_tail(newname, 0); *to; to++)
		*to = toupper(*to);
	return dired_do_move(name, newname, in_dired);
}

dired_do_lowercase(name, to, in_dired)	/* return 1 if error */
char *name, *to;
{
	char newname[FNAMELEN];

	to = to;
	strcpy(newname, name);		// Rename file to lowercase.
	for (to = get_tail(newname, 0); *to; to++)
		*to = tolower(*to);
	return dired_do_move(name, newname, in_dired);
}

// Delete the directory dir and its contents.
dired_delete_hierarchy(char *dir)
{
	char pat[FNAMELEN], tail[FNAMELEN], *s;
	int cnt = -1, last;

	strcpy(pat, dir);
	add_final_slash(pat);
	save_var file_pattern_rules |= FPAT_SKIP_DIR_SYMLINKS
								   | FPAT_SKIP_RECUR_SYMLINKS;
	save_var file_pattern_wildcards = 0;
	strcat(pat, "**/*");
	do {	// Delete everything that matches our pattern.
		s = do_file_match(pat, STARTMATCH | EXACTONLY);
		last = cnt;
		for (cnt = 0; s; s = do_file_match(pat, EXACTONLY), cnt++) {
			check_abort();
			strcpy(tail, get_tail(s, 0));
			drop_final_slash(tail);
			if (!strcmp(tail, ".") || !strcmp(tail, ".."))
				continue;
			if (get_file_read_only(s))
				set_file_read_only(s, 0);
			if (cnt % 10 == 0)
				noteput("Deleting %f ...", s);
			if (is_directory(s))
				rmdir(s);
			else
				delete_file(s);
		}	// Repeat until the number of matches stops decreasing.
	} while (cnt < last || last == -1);
	set_file_read_only(dir, 0);
	return rmdir(dir);
}

// Issue a warning about an operation, and get confirmation.  Return 0
// to go ahead, or 1 if user canceled (possibly setting user_abort too).

int dired_warn_operation(int bit, char *op, char *msg, char *file)
{
	char msg1[FNAMELEN], msg2[FNAMELEN], tmp[FNAMELEN];

	if (dired_confirmation & bit) {
		sprintf(msg1, "%s %s? ", op, file);
		strcpy(tmp, file);
		if (!is_gui || !want_gui_prompts)
			abbreviate_file_name(tmp, screen_cols - 54);
		sprintf(msg2, "%s %s (Yes, yes to All, No)? ", op, tmp);
		switch (ask_3way(msg, msg1, msg2, "Yes", "Yes to &All", "No", 1)) {
			case 0:	user_abort = 1;
					return 1;
			case 3: return 1;
			case 2:	dired_confirmation &= ~bit;
					break;
		}
	}
	return 0;
}

// Offer to delete a hierarchy.  Return nonzero if we can't.
int dired_warn_rmdir(char *dir)
{
	if (!rmdir(dir))
		return 0;
	if (dired_warn_operation(8, "Delete hierarchy",
							 "Confirm deleting hierarchy", dir))
		return 1;
	return dired_delete_hierarchy(dir);
}

// Return 0 if we could, or 1 if error or user cancelled.

int dired_warn_delete(char *file)
{
	if (get_file_read_only(file) != 1)
		return delete_file(file);
	if (dired_warn_operation(1, "Delete read-only file",
							 "Confirm deletion", file))
		return 1;
	set_file_read_only(file, 0);
	return delete_file(file);
}

/* Returns 1 if error.  May set dired_confirmation for all-files
 * behavior, so save it first.  If !in_dired, safe to assume it's a
 * file, not a directory.
 */
dired_do_delete(name, to, in_dired)
char *name, *to;
{
	int isdir = in_dired ? dired_isdir() : 0;

	to = to;
	if (is_remote_file(name)) {
		buffer_printf(REMOTE_FTP_OP_BUFFER,
					  isdir == 1 ? "RMD %s\n" : "DELE %s\n",
					  get_url_file_part(name, 0));
	} else {
		if (in_dired)
			noteput("Deleting %f.", name);
		if (isdir == 1) {
			if (dired_warn_rmdir(name))
				return 1;
		} else if (dired_warn_delete(name)
				   && (isdir <= 1 || !is_mswindows || rmdir(name)))
			return 1;	// If NTFS junction, try rmdir too.
		if (!isdir)
			maybe_delete_spell_ignore_list(name);
	}
	if (in_dired) {
		save_var modified;	// leave modify bit alone
		to_begin_line();	// but delete line
		delete(point - 1, give_end_line());
	}
	return 0;
}

// Warn if file s already exists.  Return 1 to skip, 0 to go ahead.
// Maybe disable future checking.
dired_warn_existing_file(char *s, int bit)
{
	char msg[FNAMELEN], msg2[FNAMELEN];

	if (is_remote_file(s))
		inet_only();
	if (!(dired_confirmation & bit) || check_file(s, NULL) != CHECK_FILE)
		return 0;
	sprintf(msg, "Overwrite %r? ", s);
	sprintf(msg2, "Overwrite %r [Overwrite, overwrite All, Skip]? ", s);
	switch (ask_3way("Confirm overwrite", msg, msg2,
					 "Overwrite", "Overwrite &All", "Skip", 1)) {
		case 1:
			return 0;
		case 2:
			if (key == 'N' && !key_is_button)	// N means Skip, not All.
				return 1;
			dired_confirmation &= ~bit;
			return 0;
		case 0:
			user_abort = 1;
	}
	return 1;
}

// Is copying to or from a filename of this type unsupported?  Right
// now we support local files or scp: for copying.

no_remote_copy(char *fname)
{
	return is_remote_file(fname) && strnfcmp(fname, "scp:", 4);
}

/* Return 1 if error, 2 if we didn't do it but already showed error. */
dired_do_copy(name, to, in_dired)
char *name, *to;
{
	char dest[FNAMELEN];

	make_copy_name(name, to, dest);
	if (no_remote_copy(name) || no_remote_copy(to)) {
		if (!in_dired)
			noteput("Copy operation not implemented for remote files.");
		return 2;
	}
	if (in_dired)
		noteput("Copying %f to %s.", name, dest);

	if (!strcmp(name, dest))	// disallow copying file to itself
		return 2;
	if (!is_remote_file(dest) && dired_warn_existing_file(dest, 2) > 0)
		return 2;
	if (is_remote_file(name) && is_remote_file(to)) {
		char *s = get_url_file_part(name, 0), *t = get_url_file_part(dest, 0);
		if (s - name != t - dest || strncmp(name, dest, s - name)) {
			noteput("Remote=>remote copying with different hosts not supported.");
			return 2;
		}
		buffer_printf(REMOTE_FTP_OP_BUFFER, "CPFR %s\nCPTO %s\n", s, t);
	} else if (is_remote_file(name)) {
		net_receive_file(name, dest);
	} else if (is_remote_file(dest)) {
		net_send_file(name, dest);
	} else {
		if (copyfile(name, dest))
			return 1;
	}
	if (in_dired) {
		save_var modified;	// leave modify bit alone
		dired_unmark();
		nl_reverse();
	}
	return 0;
}

/* Return 1 if error, 2 if we didn't do it but already showed error. */
dired_do_move(name, to, in_dired)
char *name, *to;
{
	char dest[FNAMELEN];

	make_copy_name(name, to, dest);
	if (is_remote_file(name)) {
		buffer_printf(REMOTE_FTP_OP_BUFFER, "RNFR %s\nRNTO %s\n",
				get_url_file_part(name, 0), get_url_file_part(dest, 0));
	} else {
		if (in_dired)
			noteput("Moving %f to %s.", name, dest);
		if (rename_file(name, dest)) {
			if (!strcmp(name, dest) || dired_warn_existing_file(dest, 4) > 0)
				return 2;
			if (copyfile(name, dest))
				return 1;
			if (delete_file(name))
				return 1;
		}
	}
	if (in_dired) {
		save_var modified;	// leave modify bit alone
		to_begin_line();	// but delete line
		delete(point - 1, give_end_line());
	}
	return 0;
}

// Copy destfile to res, but if it's a directory, append basename of src to it.
make_copy_name(char *src, char *destfile, char *res)
{
	strcpy(res, destfile);
	if (url_is_a_directory(res)) {
		destfile = res + strlen(res);
		if (*res && !is_path_separator(destfile[-1]))
			*destfile++ = is_remote_file(destfile) ? '/' : path_sep;
		strcpy(destfile, get_tail(src, 0));
	}
}

// Confirm each change to be made, then make it.
dired_do_regex_rename(name, to, in_dired)	/* return 1 if error */
char *name, *to;
{
	char newname[FNAMELEN];
	int tmp = tmp_buf();

	strcpy(to = newname, name);
	if (!dired_regex_rename_data->whole_name)
		to = get_tail(newname, 0);	// Leave the dir name alone.
	save_var bufnum = tmp;
	stuff(to);
	point = 0;
	save_var abort_searching = ABORT_ERROR;
	string_replace(dired_regex_rename_data->str, dired_regex_rename_data->with,
				   dired_regex_rename_data->flags & ~INCR);
	grab(0, size(), to);
	restore_vars();
	buf_delete(tmp);
	if (user_abort)
		return 0;
	if (dired_regex_rename_data->prompt) {
		sayput("Rename %s to %s? [Y, N, ! for all] ",
			   get_tail(name, 1), get_tail(newname, 1));
		for (;;) {
			getkey();
			if (user_abort)
				return 0;
			if (IS_ENTER_KEY(key))
				key = 'Y';
			switch (toupper(key)) {
				case 'N':	return 1;
				case '!':	dired_regex_rename_data->prompt = 0;
				case 'Y':	break;
				default:	continue;	// loop
			}
			break;		// user said yes
		}
	}
	return dired_do_move(name, newname, in_dired);
}

// Ask for search/replace patterns to transform each file name.
char *dired_prompt_regex_replace()
{
	if (!dired_regex_rename_data)		/* Get space for it. */
		dired_regex_rename_data = (struct dired_regex_rename_data *)
								  malloc(sizeof(struct dired_regex_rename_data));
	dired_regex_rename_data->whole_name = has_arg;
	dired_regex_rename_data->prompt = 1;
	dired_regex_rename_data->flags = get_replace_text(REGEX,
		dired_regex_rename_data->with);
	strncpy(dired_regex_rename_data->str,
			default_search_string(dired_regex_rename_data->flags), FNAMELEN - 1);
	return NULL;
}

// If the buffer of operations is nonempty, start executing it.

send_remote_ftp_op(char *url)
{
	char *p;

	if (!exist(REMOTE_FTP_OP_BUFFER))
		return;
	save_var bufname = REMOTE_FTP_OP_BUFFER;
	delete_final_newlines();
	if (!size())
		return;
	p = malloc(size() + 1);
	grab(0, size(), p);
	restore_vars();
	ftp_misc_operation(url, p);
	free(p);
	say("Operation queued.");
}

dired_execute() on dired_tab['q'], dired_tab['x']
{
	char url[FNAMELEN];
	int del, copy, move, uc, lc, rr;

	save_var dired_confirmation, _show_help_on = "dired-mode";
	if (is_remote_file(get_dired_buffer_pattern()) && exist(REMOTE_FTP_OP_BUFFER))
		zap(REMOTE_FTP_OP_BUFFER);
	del = dired_do_no_param("Delete", "\nD ", "deleted", dired_do_delete, 0);
	copy = dired_do_param("Copy", "\nC ", "copied", dired_do_copy);
	move = dired_do_param("Move", "\nM ", "moved", dired_do_move);
	uc = dired_do_no_param("Uppercase", "\nU ", "uppercased",
						   dired_do_uppercase, 0);
	lc = dired_do_no_param("Lowercase", "\nL ", "lowercased",
						   dired_do_lowercase, 0);
	rr = dired_do_no_param("Regex-rename", "\nR ", "regex-renamed",
							 dired_do_regex_rename, dired_prompt_regex_replace);
	if (del + copy + move + uc + lc + rr == 0)
		error("No files marked");
	modified = 0;
	if (!is_remote_file(get_dired_buffer_pattern()) && (move + copy + uc + lc + rr))
		dired_refresh();	// could be rename or similar, make new listing
	if (is_remote_file(get_dired_buffer_pattern())) {
		dired_getdir(url);
		send_remote_ftp_op(url);
	}
	fix_window_start();
	iter = 0;
}

// For safety, don't permit .. combined with other files when deleting.

dired_strip_parent(char *tag)
{
	char name[FNAMELEN];

	save_var point = 0;
	while (search(1, tag) == 1) {
		dired_getname(name, 3);
		if (!strcmp(get_tail(name, 0), "..")) {
			save_var modified;	// leave modify bit alone
			dired_unmark();
			say("Skipping .. for safety; delete this hierarchy separately.");
		}
	}
}

dired_do_no_param(op, tag, past, func, extra)
char *op, *tag, *past, *(*extra)();
int (*func)();
{
	int count = dired_ask(op, tag, "? ", NULL);

	if (count > 1 && tag[1] == 'D')
		dired_strip_parent(tag);
	if (count > 0)
		dired_do(past, tag, count, func, extra ? (*extra)() : NULL);
	return count;
}

dired_do_param(op, tag, past, func)
char *op, *tag, *past;
int (*func)();
{
	char to[FNAMELEN], recall_tag[FNAMELEN];
	int count;

	sprintf(recall_tag, "dired-%s", op);
	save_var recall_id = recall_tag;
	count = dired_ask(op, tag, " to: ", to);
	if (count > 1 && !is_remote_file(to))	// can only copy/move mult
		switch (check_file(to, NULL)) {		// to a dir
		case 0:		if (mkdir(to)) {
					file_error(errno, to, "make dir");
					quick_abort();
				}
		case CHECK_DIR:	break;
		default:	error("%s is not a directory.", to);
		}
	if (count > 0)
		dired_do(past, tag, count, func, to);
	return count;
}

dired_ask(op, tag, end, file)	// make list of files, confirm/get param if any
char *op, *tag, *end, *file;	// return count or 0 if none were marked
{
	int list = tmp_buf(), count = 0;
	char msg[FNAMELEN], name[FNAMELEN], title[FNAMELEN];

	save_var point = 0;
	save_var abort_searching = ABORT_ERROR;
	while (search(1, tag) == 1) {
		dired_getname(name, 3);
		count++;
		if (count > 1)
			sprintf(msg, "%s these %d files%s", op, count, end);
		else
			sprintf(msg, "%s this file%s", op, end);
		if (count > 1 && !(is_gui && want_gui_prompts))
			note(msg);
		buf_printf(list, "%s\n", name);
	}
	restore_vars();

	save_var inp_get_other_name = name;
	if (!count) {
		buf_delete(list);
		return 0;
	} else if (count == 1 && (!(is_gui && want_gui_prompts) || file)
			   && strlen(name) < screen_cols / 2) {
		sprintf(msg, "%s file %s%s", op, name, end);
	} else {
		if (!(is_gui && want_gui_prompts) || file)
			say(msg);
		sprintf(title, " Files to %s ", op);
		save_var _view_title = (is_gui && want_gui_prompts && !file)
							   ? msg : title;
		buf_delete_final_newlines(list);
		save_var paging_retains_view = 1;
		view_buf(list, 0);
	}
	save_var prompt_with_buffer_directory = PBD_INSERT;
	if (is_gui && want_gui_prompts && key == abort_key)
		file = 0, *name = 'N';		// User clicked Cancel.
	else if (file) {
		get_file_dir(file, msg);
		if (count > 1 && is_remote_file(file)
			&& !is_path_separator(file[strlen(file) - 1]))
			strcat(file, "/");
	} else if (is_gui && want_gui_prompts) {	// Already prompted so don't
		*name = toupper(key) == 'Y' || IS_ENTER_KEY(key) // ask again.
				|| toupper(key) == 'O' ? 'Y' : 'N';
		ungot_key = -1;
	} else
		get_string(name, msg);
	buf_delete(list);
	if (file ? *file : (toupper(*name) == 'Y'))
		return count;
	error("No action taken.");
}

dired_do(op, tag, count, func, to)
char *op, *tag, *to;
int (*func)();
{
	spot final;
	int bad = 0, tried = 0;
	char name[FNAMELEN];

	final = alloc_spot();
	point = 0;
	save_var abort_searching = 0;
	while (search(1, tag)) {
		if (user_abort) {
			if (!bad)
				*final = point;
			break;
		}
		dired_getname(name, 3);
		tried++;
		if ((*func)(name, to, 1))
			if (!bad++)		/* error occurred */
				*final = point;	/* 1st error */
	}
	point = *final;
	free_spot(final);
	move_to_column(goal_column);
	if (bad)
		error("%d of %d file%s not %s", bad, tried,
			tried != 1 ? "s" : "", op);
	say("%d file%s %s", count, count != 1 ? "s" : "", op);
}

dired_del_line()	/* delete current line, leave modify bit alone */
{
	save_var modified;	// leave modify bit alone
	to_begin_line();	// but delete line
	delete(point, give_end_line() + 1);
}

do_shell_execute(char *fname, char *op, char *descr)
{
	char dirname[FNAMELEN], file[FNAMELEN];

	strcpy(file, fname);
	absolute(file);
	strcpy(dirname, file);
	*get_tail(dirname, 1) = 0;
	if (is_remote_file(file) && !*op)
		op = "Open";
	if ( do_run_viewer(file, op, dirname) == 0)
		say("Could not %s for \"%f\"", descr, file);
}

dired_run_viewer() on dired_tab['v']
{
	char fname[FNAMELEN];

	dired_getname(fname, 1);
	do_shell_execute(fname, "", "run viewer"); // Def. action might not be Open.
}

dired_show_properties() on dired_tab['t']
{
	char fname[FNAMELEN];

	dired_getname(fname, 1);
	do_shell_execute(fname,  "Properties", "show properties");
}

// Run the Unix chmod program on the file, passing flags, and return 1.
// If error, show it using a buffer named buf and return 0.
int dired_run_chmod(char *flags, char *fname, char *dir, char *buf)
{
	char cmd[FNAMELEN];

	sprintf(cmd, "chmod %s ", flags);
	quote_for_shell(cmd + strlen(cmd), fname);
	pipe_text(NULL, buf, cmd, dir,
			  PIPE_SYNCH | PIPE_CLEAR_BUF | PIPE_NOREFRESH, 0);
	if (!buffer_size(buf))
		return 1;
	to_buffer(buf);
	discardable_buffer = 1;
	point = 0;
	return 0;
}

// Return 1 if ok, 0 if error.
int dired_do_set_attributes(char *fname, char *dir)
{
	char flags[FNAMELEN], cmd[FNAMELEN], msg[FNAMELEN];
	char abbrev[FNAMELEN];

	save_var _show_help_on = "dired-mode";
	strcpy(abbrev, fname);
	abbreviate_file_name(abbrev, screen_cols - 50);
	if (is_unix || is_remote_file(fname)) {
		sprintf(msg, is_remote_file(fname) ? "Octal mode for chmod command on %s"
										   : "Flags for chmod command on %s",
				abbrev);
		get_str_auto_def(flags, msg);
		if (!*flags)
			return 0;
		if (is_remote_file(fname)) {
			sprintf(cmd, "chmod %s %s", flags, get_url_file_part(fname, 0));
			ftp_misc_operation(dir, cmd);
			say("Operation queued.");
			return 0;	// Suppress refreshing listing.

		} else if (!dired_run_chmod(flags, fname, dir,
									"-dired-set-attributes-result"))
			return 0;
	} else {
		struct file_info fi;
		int attrs;

		if (!check_file(fname, &fi))
			error("Can't get attributes for %f", fname);
		attrs = fi.opsysattr;
		sprintf(cmd, "%s%s%s%s",
				(attrs & DOS_ATTR_RDONLY) ? "R" : "",
				(attrs & DOS_ATTR_SYSTEM) ? "S" : "",
				(attrs & DOS_ATTR_HIDDEN) ? "H" : "",
				(attrs & DOS_ATTR_ARCHIVE) ? "A" : "");
		sprintf(msg, "Set attributes (RO, Sys, Hid, Arch) for %s to: ", abbrev);
		get_strnone(flags, msg, cmd);
		save_var case_fold = 1;
		attrs = (strstr(flags, "R") ?  DOS_ATTR_RDONLY : 0)
				| (strstr(flags, "S") ? DOS_ATTR_SYSTEM : 0)
				| (strstr(flags, "H") ? DOS_ATTR_HIDDEN : 0)
				| (strstr(flags, "A") ? DOS_ATTR_ARCHIVE : 0);
		set_file_opsys_attribute(fname, attrs);
	}
	return 1;
}

dired_set_attributes() on dired_tab['a']
{
	char dir[FNAMELEN], fname[FNAMELEN];

	dired_getname(fname, 1);
	dired_getdir(dir);
	if (dired_do_set_attributes(fname, dir))
		if (modified)
			say("Press r to refresh listing, showing new attributes.");
		else
			dired_refresh();
}

dired_toggle_readonly() on dired_tab['@']
{
	char dir[FNAMELEN], fname[FNAMELEN];

	dired_getname(fname, 1);
	dired_getdir(dir);
	if (is_remote_file(fname))
		error("Can't set attributes for remote file %f", fname);
	int old = get_file_read_only(fname);
	if (old == -1 || !set_file_read_only(fname, !old))
		error("Can't change read-only status of %f", fname);
	if (modified)
		say("Press r to refresh listing, showing new attributes.");
	else
		dired_refresh();
}

dired_goto_dir() on dired_tab['g']
{
	char fname[FNAMELEN];

	dired_getname(fname, 1);
	if (chdir(fname) && (dired_try_getdir(fname, 1), chdir(fname)))
		file_error(errno, fname, "unknown error");
	else {
		getcd(fname);
		say("Current directory: %f", fname);
	}
	iter = 0;
}

dired_print() on dired_tab['P']
{
	char fname[FNAMELEN];
	int had, old = window_bufnum, err = 0, repeated = has_arg, newbuf;

	if (dired_isdir())
		error("Can't print directories.");
	save_var has_arg = 0, iter = 1;
	dired_getname(fname, 1);
	had = look_file(fname);
	if (had)
		to_buffer(bufname);
	else
		err = find_in_other_buf(fname, give_line_translate(fname));
	newbuf = bufnum;
	if (!err)
		err = do_print_region(bufnum, 1, repeated);
	to_buffer_num(old);
	if (!had)
		buf_delete(newbuf);
	if (err)
		quick_abort();
}

dired_unmark_all(char *pat)
{
	save_spot point = 0;
	save_var modified;	// leave modify bit alone
	while (re_search(1, pat))
		replace(give_begin_line(), ' ');
}

dired_mark_files_containing() on dired_tab['G']
{
	int flags = REGEX | UNKNOWN, code;
	char msg[FNAMELEN], *keys = "dmcuULRg", *codes= "DMC ULRG", *p;

	iter = 0;
	sprintf(msg, "Mark files %scontaining text: ", has_arg ? "not " : "");
	flags = get_search_string(msg, flags | FILEGREP);
	flags &= ~(INCR | FAILING | BAD | REVERSE);
	do {
		sayput("Mark for delete/move/copy/unmark/etc. or generic (dmcuULRg): ");
		getkey();
		check_abort();
		refresh();
	} while (!IS_ASCII_KEY(key) || !(p = strchr(keys, key)));
	code = codes[p - keys];
	if (code == 'G')
		dired_unmark_all("\nG ");
	do_dired_mark_files_containing(flags, default_search_string(flags),
								   code, has_arg);
}

// If the current buffer contains the pattern (or doesn't, if invert),
// use the code to mark the current line of the dired buffer orig.
// Return 0=not found, 1=found, 2=aborted, 3=bad pattern.
dired_mark_one_file_containing(int orig, int flags, char *str, int code,
							   int invert)
{
	save_var point = (flags & REVERSE) ? size() : 0;
	int ret = do_color_searching(flags, str);
	save_var bufnum = orig, point;
	if (ret < DSABORT && (ret != 0) != (invert != 0))
		dired_repl(code);
	return ret;
}

do_dired_mark_files_containing(int flags, char *pat, int code, int invert)
{
	int res, cnt = 0, this_err, skipped = 0, tempbuf = 0, aborted = 0;
	int orig = bufnum, endpos = point, setpos = 0;
	char file[FNAMELEN];

	save_var synchronous_file_io = 1;
	save_var readonly_warning;			// Don't show msg or ring bell
	readonly_warning &= ~(ROWARN_MSG | ROWARN_BELL); // unless matches.
	save_var abort_file_matching = ABORT_ERROR;
	save_var bufnum, window_bufnum;
	for (point = 0; point < size() && nl_forward(); ) {
		if (!dired_getname(file, 2) || dired_isdir())
			continue;
		if (!ok_grep_file_match(file, file)) {
			skipped++;
			continue;
		}
		noteput("Searching %r...", file);
		if (!look_file(file)) {
			this_err = std_find_it(file);
			tempbuf = bufnum;
		} else {
			this_err = 0;
			if (grep_on_changed_file & 2)
				save_var auto_read_changed_file = 1;
			if (check_dates(0) == 2)
				say("Searching version of %s in memory, "
					  "but version on disk has changed.\n", file);
		}
		if (this_err == EREADABORT || user_abort) {
			aborted = 1;
			break;
		} else if (this_err) {
			window_bufnum = orig;
			if (tempbuf)
				buf_delete(tempbuf);
			tempbuf = 0;
			continue;
		}

		res = dired_mark_one_file_containing(orig, flags, pat, code, invert);
		window_bufnum = orig;
		if (res >= DSABORT)	/* error occurred during search */
			break;
		if (res) {
			cnt++;
			if (!setpos)
				endpos = point;
			setpos = 1;
			refresh();
		}
		if (tempbuf)
			buf_delete(tempbuf);
		tempbuf = 0;
	}
	window_bufnum = orig;
	point = endpos;
	user_abort = 0;
	restore_vars();
	if (tempbuf)
		buf_delete(tempbuf);
	if (aborted)
		quick_abort();
	else if (res == DSBAD)
		error("Illegal search pattern: %s", pat);
	else if (res != DSABORT && skipped)
		say("Skipped %d file%s due to grep-ignore-file- variables",
			  skipped, (skipped == 1) ? "" : "s");
	else
		say("Marked %d file%s.", cnt, (cnt == 1) ? "" : "s");
}

dired_move_by_marking(int dir)
{
	if (dir < 0)
		nl_reverse();
	to_begin_line();
	char pat[FNAMELEN];
	sprintf(pat, "^(%s|[^%c])", dired_skip_line_pat(), character(give_begin_line()));
	re_search(dir, pat);
	to_begin_line();
	if (dir < 0 || parse_string(1, dired_skip_line_pat()))
		nl_forward();
	move_to_column(goal_column);
}

dired_forward_by_marking() on dired_tab[ALT(']')]
{
	dired_move_by_marking(1);
}

dired_backward_by_marking() on dired_tab[ALT('[')]
{
	dired_move_by_marking(-1);
}

command quick_dired_command() on reg_tab[ALT('o')]
{
	char file[FNAMELEN], rel[FNAMELEN], msg[FNAMELEN], dest[FNAMELEN];
	char dir[FNAMELEN];

	if (is_dired_buf()) {
		dired_getname(file, 3);
		dired_try_getdir(dir, 1);
		save_var inp_get_other_name = file;
	} else {
		strcpy(file, filename);
		strcpy(dir, filename);
		*get_tail(dir, 0) = 0;
	}
	if (!*file)
		error("This buffer has no associated file.");
	strcpy(rel, get_tail(file, 0));
again:
	do {
		sayput("%s: D=Delete, Move, Copy, View, T=Props, .=Dired, "
			   "G=Go dir, ?=Show more: ", rel);
		getkey();
		check_abort();
	} while (!IS_TRUE_KEY(key));
	say("");
	if (is_remote_file(file) && exist(REMOTE_FTP_OP_BUFFER))
		zap(REMOTE_FTP_OP_BUFFER);
	switch (toupper(key)) {
		case '.':
			dired_one(file);
			break;
		case 'D':
			sprintf(msg, "Delete %s? ", rel);
			if (!ask_yn("Delete file", msg, "Delete", "Cancel", 1))
				break;
			save_var dired_confirmation;
			if (dired_do_delete(file, NULL, 0)) {
				check_abort();
				file_error(errno, file, "Could not delete");
			} else if (!is_remote_file(file))
				say("%f deleted.", file);
			break;
		case 'C':
			save_var dired_confirmation;
			if (no_remote_copy(file))
				error("Copy operation not implemented for remote files.");
			if (is_directory(file) == 1)
				error("Can't copy directories.");
			sprintf(msg, "Copy %s to: ", rel);
			get_file_dir(dest, msg);
			if (*dest) {
				switch (dired_do_copy(file, dest, 0)) {
					case 1:
						file_error(errno, dest, "Could not copy");
						break;
					case 0:
						say("%f copied.", file);
						break;
				}
			}
			break;
		case 'M':
			save_var dired_confirmation;
			sprintf(msg, "Move %s to: ", rel);
			get_file_dir(dest, msg);
			if (*dest) {
				if (dired_do_move(file, dest, 0))
					file_error(errno, dest, "Could not move");
				else if (!is_remote_file(file))
					say("%f moved.", file);
			}
			break;
		case 'V':
			do_shell_execute(file, "", "run viewer");
			break;
		case 'T':
			do_shell_execute(file,  "Properties", "show properties");
			break;
		case 'F':
			do_shell_execute(dir, "Open", "run viewer");
			break;
		case 'G':
			if (chdir(dir))
				file_error(errno, dir, "unknown error");
			else {
				getcd(dir);
				say("Current directory: %f", dir);
			}
			break;
		case '+':
			do_create_directory(0);
			break;
		case '!':
			do_shell_command(file, dir);
			break;
		case 'A':
			dired_do_set_attributes(file, dir);
			check_dates(0);
			break;
		case '@':
			if (is_remote_file(file))
				say("Can't on remote files.");
			int old = get_file_read_only(file);
			if (old == -1 || !set_file_read_only(file, !old))
				error("Can't change read-only status of %f", file);
			check_dates(0);
			break;
		default:
			if (key == '?' || key == HELPKEY || key == FKEY(1))
				help_on_command("quick-dired-command");
			else
				say("Invalid key.");
			goto again;
	}
	if (is_remote_file(file))
		send_remote_ftp_op(file);
}

dired_down_line() on dired_tab['n']
{
	down_line();
	to_begin_line();
	while (parse_string(1, dired_skip_line_pat()))
		nl_forward();
	move_to_column(goal_column);
}

dired_up_line() on dired_tab['p']
{
	up_line();
	to_begin_line();
	while (parse_string(1, dired_skip_line_pat())) {
		if (!nl_reverse())
			break;
		to_begin_line();
	}
	move_to_column(goal_column);
}

dired_getdir(fullname)	// grab directory, add / to end, for completion
char *fullname;		// set fullname to "" if error
{
	return dired_try_getdir(fullname, 0);
}

dired_try_getdir(fullname, showerr)	// grab directory, add / or \ to end
char *fullname;		// if we can't find, set fullname "" if !showerr
{			// return nonzero if ok
	int i;
	char *p, path_char;

	*fullname = 0;
	save_var point;
	to_end_line();
	if (!re_search(-1, DIRLINEPAT))	// find prev "Directory of "
		if (showerr)		// bad dired buf
			error("Can't determine directory of files");
		else
			return 0;
	point = matchstart;
	re_search(1, " *");		// skip initial spaces
	parse_string(1, FILECHARS, fullname);
	path_char = is_remote_file(fullname) ? '/' : path_sep;
	i = strlen(fullname);
	if (is_remote_file(fullname))
		while (i > 0 && !is_path_separator(fullname[i - 1]))
			fullname[--i] = 0;
	p = fullname + i;
	if (i > 0 && !is_path_separator(fullname[i - 1])
		  && (strcmp(fullname, "~") || !is_unix))
		p[0] = path_char, p[1] = 0; // add final / or \ if none
	return 1;
}

// Return nonzero if rel (a junction destination name in a dired
// listing) is valid. Windows sometimes makes reparse points that use
// fake names.
int dired_is_valid_link(char *rel)
{
	char name[FNAMELEN], *p;
	dired_try_getdir(name, 1);
	p = name + strlen(name);
	strcpy(p, rel);
	if (is_path_separator(*p) == 1	// If link dest starts with /,
		  || isalpha(p[0]) && p[1] == ':' && is_path_separator(p[2]) == 1)
		strcpy(get_url_file_part(name, 0), p);	// just use that.
	return check_file(name);
}

// Get name of file/dir on current line and return 1.  If none, return 0
// (or, if flags&1, abort with error).  Return destination of a
// symbolic link unless flags&2.

int dired_getname(char *fullname, int flags)
{
	char *p, *q;

	dired_try_getdir(fullname, 1);
	p = fullname + strlen(fullname);
	dired_getindex(p, goal_column);		/* the file name */
	to_begin_line();
	if (parse_string(1, dired_skip_line_pat()))  /* on dir line, make name illegal */
		*p = 0;
	move_to_column(goal_column);
	if (!*p || isspace(*p))
		if (flags & 1)
			error("Can't read file name");
		else
			return 0;
	q = strstr(p, " -> ");			// replace file name with link dest
	if (q)
		if ((flags & 2) || !dired_is_valid_link(q + 4))
			*q = 0;
		else
			strcpy(p, q + 4);
	if (is_path_separator(*p) == 1	// If link dest starts with /,
		  || isalpha(p[0]) && p[1] == ':' && is_path_separator(p[2]) == 1)
		strcpy(get_url_file_part(fullname, 0), p);	// just use that.
	if (dired_format == DF_SIMPLE) {
		q = index(fullname, ' ');		// Remove stuff after file name.
		if (q)
			*q = 0;
	} else if (dired_format == DF_VMS) {
		q = index(fullname, ' ');		// Remove stuff after file name.
		if (q)
			*q = 0;
		q = strstr(fullname, ".DIR;");	// Refer to directories without
		if (q)							// .DIR extension to see contents.
			*q = 0;
		q = strstr(fullname, ";");		// Remove version number.
		if (q && isdigit(q[1]))
			*q = 0;
	}
	p += strlen(p) - 1;		// Unix file name might use "ls -F" format
	if (*p == '/' || *p == '*')
		*p = 0;
	return 1;
}

// Look for continuation lines in VMS directory listing, combine.

dired_vms_fix_lines()
{
	save_var modified;
	simple_re_replace(1, "\n                    ", "");
}

// Look for extra-wide columns in Unix directory listing, repair.

dired_unix_fix_lines()
{
	save_var modified, indent_with_tabs = 0, case_fold = 1;
	right_align_columns("[0-9]+ *" UNIX_DATE);
	right_align_columns(UNIX_DATE);
	save_var point = 0;
	if (re_search(1, ".*" UNIX_DATE))
		goal_column = current_column();
}

// Maybe remove lines for files that start with ".".

dired_strip_dotfiles()
{
	char name[FNAMELEN];
	int filecnt = 0, dircnt = 0, specialcnt = 0;

	if (dired_force_show_dotfiles || dired_show_dotfiles == 15
		  || dired_format && !dired_is_unix_format())
		return;
	if (string_matches_pattern(get_dired_buffer_pattern(),
							   ".*(^|[,;{/\\])%.%.?[^./\\]+$"))
		return;		// Pattern explicitly mentions dotfile.
	save_var modified, point = 0;
	for (; search(1, " ."); to_end_line()) { // Quickly get to possible dotfile.
		if (give_end_line() >= size())	// Don't do partial lines at eof.
			break;
		to_begin_line();
		if (parse_string(1, dired_skip_line_pat()))
			continue;		// Ignore directory separator lines.
		dired_getindex(name, goal_column);
		if (*name == '.') {
			if (strlen(name) > 1 && is_path_separator(name[strlen(name) - 1]))
				  name[strlen(name) - 1] = 0;
			if (!strcmp(name, ".")) {
				if (dired_show_dotfiles & 1)
					continue;
				specialcnt++;
			} else if (!strcmp(name, "..")) {
				if (dired_show_dotfiles & 2)
					continue;
				specialcnt++;
			} else if (!dired_isdir()) {
				if (dired_show_dotfiles & 8)
					continue;
				filecnt++;
			} else {
				if (dired_show_dotfiles & 4)
					continue;
				dircnt++;
			}
			delete(give_begin_line() - 1, give_end_line());
		}
		check_abort();
	}
	point = size();
	if (character(point - 1) == '\n' && (filecnt || dircnt)
		  && !(dired_show_dotfiles & 16))
		bprintf("  %d file%s, %d director%s omitted due to "
				"dired-show-dotfiles\n",
				filecnt, filecnt == 1 ? "" : "s",
				dircnt + specialcnt, dircnt + specialcnt == 1 ? "y" : "ies");
}

// Return 3 if symbolic link to file, 2 if symbolic link to dir, 1 if
// actual dir, else 0.  Unix-specific version.

dired_isdir_unix()
{
	save_var point;
	to_begin_line();
	save_var case_fold = 1;
	if (!parse_string(1, dired_dirsearch_unix))
		return 0;
	switch (character(matchend - 1)) {
		case 'a':
		case 'l':	return 3;
		case 'A':
		case 'L':	return 2;
	}
	return 1;
}

// Return 3 if symbolic link to file, 2 if symbolic link to dir, 1 if
// actual dir, else 0.

dired_isdir()
{
	int i = get_dired_item("dired-isdir-", 1);
	if (i)
		return (*(int (*)())i)();
	i = get_dired_item("dired-dirsearch-", 0);
	if (i) {
		save_var point;
		to_begin_line();
		return parse_string(1, get_str_var(i)) != 0;
	}
	if (matches_at(column_to_pos(ISDIR_START), 1, "  %<SYMLINK%>"))
		return 3;
	else if (matches_at(column_to_pos(ISDIR_START), 1,
						"  %<(JUNCTION|SYMLINKD)%>"))
		return 2;
	return matches_at(column_to_pos(ISDIR_START), 1, DIRSEARCH) != 0;
}

dired_getindex(s,i)	/* put current line in s, starting at column i */
char *s;
{
	save_var point;
	move_to_column(i);
	re_search(1, " *");		/* skip initial spaces */
	if (parse_string(1, FILECHARS) >= ptrlen(s))
		error("File name too long.");
	parse_string(1, FILECHARS, s);
}

is_dired_buf()		/* is the current buffer a dired buffer? */
{
	return mode_keys == dired_tab;
}

color_dired_range(from, to)
{
	int color, namecolor, old, namestart, nameend, isdir;
	char buf[FNAMELEN];

	if (from >= to)
		return to;
	detect_dired_format();
	save_var point, matchstart, matchend;
	point = from;
	save_var narrow_start = give_begin_line();
	point = to;
	nl_forward();
	save_var narrow_end = size() - point;
	point = 0;
	while (point < size() - narrow_end) {
		switch (curchar()) {
			case 'D': color = color_class dired_deleted; break;
			case 'U': case 'L': case 'R': case 'G':
			case 'M': color = color_class dired_moved; break;
			case 'C': color = color_class dired_copied; break;
			default: color = color_class dired_normal; break;
		}
		old = point;
		if (parse_string(1, dired_skip_line_pat()) != 0) {
			color = color_class dired_header;
			namestart = matchend;
			nameend = give_end_line();
			namecolor = color_class dired_dir_name;
		} else {
			dired_getindex(buf, goal_column);
			namestart = matchstart;
			nameend = matchend;
			namecolor = color_class dired_file_name;
			if (dired_format == DF_VMS)
				nameend = point + parse_string(1, "[^\n;]*;?[0-9]*");
			isdir = dired_isdir();
			if (isdir) {
				namecolor = color_class dired_dir_name;
				if (!dired_format)
					namestart = column_to_pos(ISDIR_START);
				if (isdir == 3 && parse_string(1, ".* -> "))
					nameend = matchend;
			}
			if (!is_remote_file(get_dired_buffer_pattern())
				  && character(namestart - 1) == ' ')
				namestart--;
		}
		if (nl_forward())
			set_character_color(point - 1, point, -1);
		set_character_color(old, point - 1, color);
		set_character_color(namestart, nameend, namecolor);
	}
	return point;
}

command dired() on cx_tab['d']
{
	char dir[FNAMELEN];

	tiled_only();
	get_file_dir(dir, "Edit directory: ");
	if (was_quoted)		// []'s etc in file name are literal.
		save_var file_pattern_wildcards = 0;
	return dired_one(dir);
}

detect_dired_format()
{
	int old = dired_format, i;

	goal_column = NAME_START;
	save_var point = 0, case_fold = 1;
	dired_format = DF_VMS;		// Skip VMS lines too for now.
	while (parse_string(1, dired_skip_line_pat()))
		nl_forward();
	dired_format = 0;
	if (point >= size())
		return;
	if (parse_string(1, ".*" UNIX_FULL_DATE)) {
		dired_format = DF_UNIX_FULL;
		goal_column = get_column(matchend);
	} else if (parse_string(1, ".*" UNIX_DATE)) {
		dired_format = DF_UNIX;
		goal_column = get_column(matchend);
	} else if (parse_string(1, ".*" OLDNT_DATE)) {
		dired_format = DF_OLDNT;
		goal_column = 41;
	} else if (parse_string(1, ".*" VMS_DATE)) {
		dired_format = DF_VMS;
		goal_column = 2;
	} else if (!parse_string(1, ". [ 0-9][0-9]-[a-z?][a-z?][a-z?]-[0-9][0-9]")
			  && !parse_string(1, ". [0-9][0-9][0-9][0-9]-[0-9][0-9]-[0-9][0-9]")) {
		dired_format = DF_SIMPLE;
		goal_column = 2;
	} else {
		for (i = 1; i <= 8; i <<= 1)
			if (dired_layout & i)
				goal_column++;
	}
	if (dired_format != old)
		drop_all_colored_regions();
}

char _dired_mode_name[] = "Dired";

dired_mode()
{
	mode_keys = dired_tab;
	major_mode = _dired_mode_name;
	mouse_dbl_selects = 1;
	detect_dired_format();
	recolor_range = color_dired_range;	// set up coloring rules
	recolor_from_here = recolor_by_lines;
	when_setting_want_code_coloring();	// maybe turn on coloring
	discardable_buffer = 1;
	make_mode();
}

/* Dir is absolute pathname of directory or pattern to list; may be
 * modified.  Returns 0 if ok. */
int dired_one(char *dir)
{
	char buf[FNAMELEN], *to;
	int uniq = 1;

	if (is_remote_file(dir))
		standardize_remote_pathname(dir);	// Try to remove . and ..
	if (*dir) {
		if (!is_path_separator(dir[strlen(dir) - 1]))
			if (is_remote_file(dir)) {
				if (remote_file_type(dir) == 8)
					strcat(dir, "/");
			} else if (is_directory(dir))
				add_final_slash(dir);
		strcpy(buf, dir);
	} else if (prompt_with_buffer_directory) {
		get_buffer_directory(buf);	// this file's directory
		dir = buf;
	} else {
		getcd(buf);	/* "" means current directory */
		dir = buf;
	}
	if (!is_remote_file(buf))
		dired_standardize(buf);		/* fix appropriately for this o/s */
	else {
		standardize_remote_pathname(buf);	// Try to remove . and ..
		standardize_remote_pathname(dir);	// Try to remove . and ..
	}
	for (to = buf; *to; to++)
		if (*to < ' ')			// Make it legal.
			*to = '!';
	to = buf + strlen(buf);
	while (exist(buf)) {	/* make it unique */
		bufname = buf;
		if (is_dired_buf())	/* reuse existing dired buffer */
			break;
		sprintf(to, "<%d>", uniq++);
	}
	to_buffer(buf);
	move_to_column(goal_column);
	if (is_remote_file(buf) && is_dired_buf() && lines_between(0, size(), 0) > 1)
		return 0;			// Don't reread unless user explicitly says to.
	return create_dired_listing(dir, 1);
}

int create_dired_listing(char *dir, int talk)
{
	buf_zap(bufnum);
	translation_type = FILETYPE_AUTO;
	dired_mode();
	if (!talk)
		save_var show_status = 0;
	if (talk)
		noteput("Building directory listing...");
	if (make_dired(dir)) {
		if (talk)
			say("No matches.");
		return 1;
	}
	if (talk)
		say("");
	return 0;
}

dired_ftp_activity(int activity, int buf, int from, int to)
{
	to = to;
	save_var bufnum = buf, iter, has_arg;
	switch (activity) {
		case NET_DONE:
			detect_dired_format();
			if (dired_format == DF_VMS)
				dired_vms_fix_lines();
			else if (dired_format == DF_UNIX)
				dired_unix_fix_lines();
			if (!modified && (point == dired_orig_position || dired_orig_position == -1))
				sort_for_dired();
			else
				say("Sorting omitted, press S to sort.");
			dired_strip_dotfiles();
			break;
		case NET_RECV:
			if (dired_orig_position == -1)
				dired_orig_position = point = from;
			save_spot point = from;
			save_var modified;
			while (re_search(1, "^.")) {
				point--;
				stuff("  ");
				point++;
			}
			if (!want_code_coloring)
				detect_dired_format();
			break;
		case NET_LOG_WRITE:
			break;
	}
}

do_remote_dired(char *dir)
{
	char plain[FNAMELEN], *p;

	strcpy(plain, dir);
	p = get_tail(plain, 1);	// If name contains wildcards, strip them.
	if (index(p, '*') || index(p, '?'))
		*p = 0;
	buffer_ftp_activity = dired_ftp_activity;
	bprintf("     Directory of %s\n", plain);
	point = modified = 0;
	dired_orig_position = point;
	if (index(dir, '*') || index(dir, '?'))	// Make FTP listing
		save_var ftp_compatible_dirs |= 1;	// show just the basename.
	return url_operation(dir, FTP_LIST);
}

make_dired(dir)
char *dir;
{
	char tmp[FNAMELEN];

	modified = 0;
	save_var abort_file_matching = ABORT_JUMP;
	strcpy(tmp, dir);		// Our arg might be a dired_buffer_pattern that we're
	if (dired_buffer_pattern)	// about to delete.
		free(dired_buffer_pattern);
	dired_buffer_pattern = strsave(tmp);
	dired_buffer_pattern_wildcards = file_pattern_wildcards;
	dir = tmp;
	if (is_remote_file(dir))
		do_remote_dired(dir);
	else if (do_dired(dir)) {
		modified = 0;
		return 1;
	}
	detect_dired_format();
	dired_strip_dotfiles();
	switching_to_buffer();
	modified = 0;
	add_buffer_when_idle(bufnum, dired_when_idle);
	if (!is_remote_file(dir))
		sort_for_dired();
	else {
		to_end_line();
		type_point = size();
	}
	dired_orig_position = point;
	return 0;
}

sort_for_dired()
{
	int start, done = 0, chat = size() - narrow_end - narrow_start > 100000;

	if (dired_format == DF_VMS)
		return;		// Sorting not supported.
	if (chat)
		noteput("Sorting...");
	save_var modified;
	window_start = point = 0;
	save_var abort_searching = 0;
	while (!done) {
		nl_forward();	/* skip header line */
		to_begin_line();
		start = point;
		if (re_search(1, dired_skip_line_pat()))
			to_begin_line();
		else
			done = 1;
		if (start < point - 1)
			dired_sorter(start, point - (parse_string(-1, "\n\n") ? 1 : 0));
	}
	point = 0;
	dired_down_line();		/* position on 2nd line */
	window_start = 0;
	if (chat)
		say("");
}

dired_sorter(start, end)
{
	int tmp;
	char *pat = DIRSEARCH;
	int i = get_dired_item("dired-dirsearch-", 0);

	if (i)
		pat = get_str_var(i);
	save_var case_fold = 1, sort_case_fold = !dired_is_unix_format();
	save_var narrow_start = start;
	save_var narrow_end = size() - end;
	if (dired_groups_dirs) {
		save_var point = 0;
		tmp = tmp_buf();
		int dirs = tmp_buf();	// Put ., then .., in here, don't sort.
		while (column_color_searching(REGEX, "<.>[/\\]?$", goal_column, -1)
			   && !user_abort)
			move_line_to_buffer(dirs);
		point = 0;
		while (column_color_searching(REGEX, "<.><.>[/\\]?$", goal_column, -1)
			   && !user_abort)
			move_line_to_buffer(dirs);
		point = 0;
		while (re_search(1, pat) && !user_abort)
			move_line_to_buffer(tmp);
		dired_group_sorter(start, point);
		point = 0;
		grab_buffer(dirs);
		buf_delete(dirs);
		grab_buffer(tmp);
		buf_delete(tmp);
		dired_group_sorter(mark, point);
	} else
		dired_group_sorter(start, end);
}

dired_group_sorter(start, end)
{
	int on_col = goal_column, stripwid = 0, i;

	if (!dired_sorts_files || user_abort)
		return;
	save_var narrow_start = start;
	save_var narrow_end = size() - end;
	save_var point;
	save_var abort_searching = ABORT_ERROR;
	switch (toupper(dired_sorts_files)) {
		case 'D': stripwid = dired_sort_by_date(); break;
		case 'E': stripwid = dired_sort_by_exten(); break;
		case 'S': i = get_dired_item("dired-sort-by-size-", 1);
				  if (i)		// Call style-specific function.
					  stripwid = (*(int (*)())i)();
				  else
					  stripwid = dired_sort_by_size_plain();
				  break;
	}
	do_sort_region(0, size(), on_col, isupper(dired_sorts_files));
	if (stripwid)
		strip_columns(goal_column, goal_column + stripwid);
}

dired_sort_by_size_simple()
{
	say("Sorting by size not supported.");
	return 0;
}

// Insert size in each line, return length.
dired_sort_by_size_unix()
{
	char line[FNAMELEN], *p;

	point = 0;
	while (re_search(1, UNIX_DATE)) {
		point = matchstart;
		if (!parse_string(-1, "[0-9]+[MKG]? *", line)) {
			note("Unrecognized format, can't sort.");
			strcpy(line, "0");
		}
		if (p = strchr(line, 'G'))
			strcpy(p, "000000000");
		if (p = strchr(line, 'M'))
			strcpy(p, "000000");
		if (p = strchr(line, 'K'))
			strcpy(p, "000");
		move_to_column(goal_column);
		bprintf("%20s", line);
		to_end_line();
	}
	return 20;
}

int dired_sort_by_size_unixfull()
{
	return dired_sort_by_size_unix();
}

// Some sizes will use commas, but larger sizes won't.  For sorting,
// remove commas from all sizes.
dired_sort_by_size_plain()
{
	char line[FNAMELEN];

	point = 0;
	do {
		if (point == give_end_line())
			break;
		move_to_column(SIZE_COL);
		re_search(1, " *");		/* skip initial spaces */
		*line = 0;
		while (parse_string(1, "[0-9]+", line + strlen(line)))
			while (curchar() == ',')
				point++;
		move_to_column(goal_column);
		bprintf("%15s", line);
	} while (nl_forward());
	return 15;
}

dired_sort_by_date_simple()
{
	say("Sorting by date not supported.");
	return 0;
}

	/* insert numeric sortable date in each line, return length */
dired_sort_by_date_unix()
{
	char line[FNAMELEN * 2], *p, *months, *time;
	int month, year, full;
	struct time_info now;

	time_and_day(&now);
	months = "Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec ";
	point = 0;
	while (re_search(1, UNIX_DATE)) {
		grab(matchstart + 1, point, line);
		full = (strlen(line) == 20);	// Full date format.
		if (line[3] != ' ' || line[6] != ' ') {
			say("Unrecognized date format, can't sort.");
			strcpy(line, " Jan  1  1970 ");	// Fake default date.
		}
		line[3] = line[6] = 0;	// break into pieces
		p = strstr(months, line);
		if (!p) {
			say("Unrecognized month %s, can't sort.", line);
			p = months;
		}
		month = 1 + (p - months) / 4;	// Compute month 1..12.

		if (full) {		// Format of Nov 24 18:22:48 1986.
			time = line + 7;
			year = numtoi(line + 16);
		} else if (line[9] == ':') {	// Either a time (in the last 6 months)...
			time = line + 7;	// Year is this year or late last year.
			year = now.year - (month > now.month + 1);
		} else {			// Or a prior year, so make up a fake time.
			time = "00:00 ";
			year = numtoi(line + 7);
		}
		move_to_column(goal_column);
		bprintf("%5d%02d%2s%*s", year, month, line + 4, full ? 8 : 5, time);
		to_end_line();
	}
	return full ? 18 : 15;
}

dired_sort_by_date_unixfull()
{	// Similar, but it always includes year.
	return dired_sort_by_date_unix() + 4;
}

/* insert numeric sortable date in each line, return length */
dired_sort_by_date_oldnt()
{
	char line[FNAMELEN * 2];
	int hour;

	point = 0;
	while (re_search(1, OLDNT_DATE)) {
		grab(matchstart, point, line);
		hour = numtoi(line + 10);
		if (hour == 12)
			hour = 0;
		if (toupper(line[15]) == 'P')
			hour += 12;
		move_to_column(goal_column);
		bprintf("%.2s%.5s%02d%.2s", line + 6,
				line, hour, line + 13);
		to_end_line();
	}
	return 11;
}

dired_sort_by_date()	/* insert stuff in each line, return length */
{
	char line[FNAMELEN * 3], *p, *months;
	int end, hour, i, month;

	months = "Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec ";
	point = 0;
	i = get_dired_item("dired-sort-by-date-", 1);
	if (i)		// Call style-specific function.
		return (*(int (*)())i)();
	do {
		end = give_end_line();
		if (point == end)
			break;
		grab(point, end, line);
		move_to_column(goal_column);
		hour = numtoi(line + 15);
		if (hour == 12 && strchr("ap", line[23]))
			hour = 0;
		if (line[23] == 'p')
			hour += 12;
		if (dired_layout & 16)
			bprintf("%.4s%.2s%.2s", line + 2, line + 7, line + 10);
		else {
			line[8] = 0;
			p = strstr(months, line + 5);
			if (!p) {
				say("Unrecognized month %s, can't sort.", line + 5);
				p = months;
			}
			month = 1 + (p - months) / 4;	// Compute month 1..12.
			bprintf("%.4s%2d%.2s", line + 9, month, line + 2);
		}
		bprintf("%2d%.5s", hour, line + 18);
		point = end;
	} while (nl_forward());
	return 15;
}

dired_sort_by_exten()	/* insert stuff in each line, return length */
{
	char line[FNAMELEN];

	point = 0;
	do {
		if (point == give_end_line())
			break;
		dired_getindex(line, goal_column);
		move_to_column(goal_column);
		bprintf("%-30.30s", get_extension(line));
	} while (nl_forward());
	return 30;
}

strip_columns(scol, ecol)
{
	point = 0;
	do {
		delete(column_to_pos(scol), column_to_pos(ecol));
	} while (nl_forward());
}
