/* -*- tab-width:4 -*- */
/*
Part of the Epsilon Programmer's Editor by Lugaru Software, Ltd.
www.lugaru.com
Copyright 2003-2011 Lugaru Software, Ltd.  All Rights Reserved

This is a helper program which lets Epsilon perform file operations over
ssh connections.

It prints a welcome message, then waits for commands, which start with
"## ".  The commands:

list diname/pattern: prints a directory listing of the specified directory
or file pattern.

comp pattern: prints the completions for a file pattern.

dele file: deletes a file.

rmd dir: deletes a directory.

mkd dir: creates a directory.

rnfr file1
rnto file2: renames file1 to file2.

cpfr file1
cpto file2: copies file1 to file2.

prdir dir: print absolute path for dir like ~bob, or home dir if empty.

quit: exits.

The comp and list commands output a series of lines of data, each
starting with "d ", followed by a single status line that starts with "s ".
Other commands just output a single status line.

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <glob.h>
#include <sys/types.h>
#include <sys/stat.h>

#define BUFSIZE	1000

// Copy the contents of handle f to stdout, prefixing each line with
// the specified prefix.  Add a \n to the end of the output if needed.
// Then write an OK status line to stdout.

void send_output(FILE *f, char *prefix)
{
	int c, lstart = 1;

	while ((c = getc(f)) != EOF) {
		if (lstart)
			printf("%s", prefix);
		putchar(c);
		lstart = (c == '\n');
	}
	if (!lstart)
		printf("\n");
	fclose(f);
	printf("s OK\n");
}

// List the specified directory to stdout.
void list_dir(char *filepat)
{
	char cmd[BUFSIZE], *s;
	struct stat sb;
	int isdir = 0;

	if (!*filepat)
		isdir = 1;
	else if (!stat(filepat, &sb))
		isdir = S_ISDIR(sb.st_mode);
	if (isdir)
		sprintf(cmd, "ls -aflF %s 2>&1", filepat);
	else if ((s = strrchr(filepat, '/'))) {
		sprintf(cmd, "cd %.*s; ls -adflF %s 2>&1", 
				(int) (s - filepat), filepat, s + 1);
	} else
		sprintf(cmd, "ls -adflF %s 2>&1", filepat);
	send_output(popen(cmd, "r"), "d ");
}

void rename_file(char *from, char *to)
{
	char cmd[BUFSIZE];

	sprintf(cmd, "mv %s %s 2>&1", from, to);
	send_output(popen(cmd, "r"), "");
}

void copy_file(char *from, char *to)
{
	char cmd[BUFSIZE];

	sprintf(cmd, "cp %s %s 2>&1", from, to);
	send_output(popen(cmd, "r"), "");
}

// List possible matches of the specified pattern to stdout.
void list_matches(char *pat)
{
	glob_t g;
	char **p;

	if (!glob(pat, GLOB_NOSORT | GLOB_MARK | GLOB_BRACE, NULL, &g) && g.gl_pathc) {
		for (p = g.gl_pathv; *p; p++)
			printf("d %s\n", *p);
	}
	printf("s OK\n");
	globfree(&g);
}

// Print absolute form of directory name.
void prdir(char *dir)
{
	char cmd[BUFSIZE];

	if (*dir)
		sprintf(cmd, "cd %s ; pwd 2>&1", dir);
	else
		sprintf(cmd, "pwd 2>&1");
	send_output(popen(cmd, "r"), "d ");
}

void bad_cmd(void)
{
	printf("epsilon-xfer-helper-ftclient: Bad command.\n");
	exit(1);
}

void send_error(char *descr, char *file)
{
	printf("%s %s: %s\n", descr, file, strerror(errno));
}

int main(argc, argv)
int argc;
char *argv[];
{
	char buf[BUFSIZE], cmd[BUFSIZE], *args, orig_file[BUFSIZE];

	setvbuf(stdin, NULL, _IOLBF, 0);
	setvbuf(stdout, NULL, _IOLBF, 0);

	printf("Epsilon-xfer-helper version 1\n");
	while (fgets(buf, sizeof(buf), stdin)) {
		if (strchr(buf, '\n'))		 /* Strip the final \n if any. */
			*strchr(buf, '\n') = 0;
		if (strncmp(buf, "## ", 3))
			bad_cmd();
		for (args = buf + 3; *args && *args != ' '; args++)
			;
		strncpy(cmd, buf + 3, args - (buf + 3));
		cmd[args - (buf + 3)] = 0;
		while (*args && *args == ' ')
			args++;
		if (!strcasecmp(cmd, "list")) {
			list_dir(args);
		} else if (!strcasecmp(cmd, "comp")) {
			list_matches(args);
		} else if (!strcasecmp(cmd, "dele")) {
			if (unlink(args))
				send_error("deleting", args);
			printf("s OK\n");
		} else if (!strcasecmp(cmd, "mkd")) {
			if (mkdir(args, 0777))
				send_error("making directory", args);
			printf("s OK\n");
		} else if (!strcasecmp(cmd, "rmd")) {
			if (rmdir(args))
				send_error("removing directory", args);
			printf("s OK\n");
		} else if (!strcasecmp(cmd, "rnfr")) {
			strcpy(orig_file, args);
			printf("s OK\n");
		} else if (!strcasecmp(cmd, "rnto")) {
			rename_file(orig_file, args);
		} else if (!strcasecmp(cmd, "cpfr")) {
			strcpy(orig_file, args);
			printf("s OK\n");
		} else if (!strcasecmp(cmd, "cpto")) {
			copy_file(orig_file, args);
		} else if (!strcasecmp(cmd, "prdir")) {
			prdir(args);
		} else if (!strcasecmp(cmd, "quit")) {
			exit(0);
		} else {	/* Unknown request. */
			bad_cmd();
		}
	}
	exit(0);
}
