/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*	Copyright (C) 1997, 2020 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

#include "eel.h"
#include "epsnet.h"
#include "proc.h"
#include "colcode.h"
#include "dired.h"

inet_only()
{
	if (!(has_feature & FEAT_INTERNET))
		error("Internet features are not available in this version.");
}

error_if_remote(char *file)
{
	if (is_remote_file(file))
		error("Remote files not supported for this command.");
}

// Put a message at the end of the current buffer's mode line.
// Pass NULL to show no message.

set_mode_message(char *s)
{
	if (mode_extra) {
		free(mode_extra);
		mode_extra = 0;
	}
	if (s)
		mode_extra = strsave(s);
	make_mode();
}

set_buf_mode_message(int buf, char *s)
{
	save_var bufnum = buf;
	set_mode_message(s);
}

// Show a message (or erase) for this buffer and the one linked to it.
// Buffers with active ftp sessions are linked to their Ftp Log
// buffers, and vice versa.

set_both_mode_message(char *s)
{
	save_var bufnum;
	if (buf_exist(ftp_job->buf)) {
		bufnum = ftp_job->buf;
		set_mode_message(s);
	}
	if (buf_exist(ftp_job->log_buf)) {
		bufnum = ftp_job->log_buf;
		set_mode_message(s);
	}
}

// Does the status message appear after here in the log?
found_ftp_status(char *pat)
{
	save_var point;
	return re_search(1, pat);
}

show_ftp_percent(char *op, int so_far)
{
	char msg[FNAMELEN];

	ftp_job->so_far = so_far;
	if (ftp_job->file_size)
		sprintf(msg, "[FTP: %s %d%%] ", op,
				muldiv(so_far, 100, ftp_job->file_size));
	else		// We don't know how big the file is.
		sprintf(msg, "[FTP: %s %d Kb] ", op, so_far / 1000);
	set_both_mode_message(msg);
}

// Set point in all windows displaying this buffer.

position_in_all_windows(buf, pos)
{
	int orig = window_number;
	save_var bufnum;
	do {
		if (window_bufnum == buf)
			point = pos;
	} while (++window_number != orig);
}

// Epsilon calls this whenever any FTP-related activity occurs.

ftp_activity(int activity, int buf, int from, int to)
{
	char msg[FNAMELEN];
	int show_log = 0, orig_buf = bufnum, op;

	if (!buf_exist(buf))
		return;
	save_var iter, has_arg;
	save_var bufnum = buf;
	if (!ftp_job)
		return;
	switch (activity) {
		case NET_DONE:
			op = ftp_job->operation;
			if (op == FTP_SEND)
				show_text(0, 1, "%s written.", filename);
			if (translation_type == FILETYPE_AUTO && op == FTP_RECV)
				translation_type = new_buffer_translation_type;
			if (!modified && !point && op == FTP_RECV)
				go_line(ftp_go_to_line);
			free(ftp_job);
			ftp_job = 0;
			set_mode_message(NULL);
			check_filevars();
			if (op == FTP_RECV)
				decompress_after_reading();
			break;
		case NET_LOG_DONE:
			ftp_job = 0;
			set_mode_message(NULL);
			check_filevars();
			break;
		case NET_RECV:
			if (!from) {		// First block.
				position_in_all_windows(buf, 0);
				if (mode_keys == buf_tab)		// Switch out of Fundamental mode.
					call_mode(filename);
			}
			show_ftp_percent("got", from);
			break;
		case NET_SEND:
//			ftp_job->so_far = from;
			ftp_job->file_size = to;
			show_ftp_percent("sent", from);
			break;
		case NET_LOG_WRITE:
			save_var point = from;	// Get file size from msg.
			if (re_search(1, "^150 ") && re_search(1, "[0-9]+ bytes")) {
				char filesize[100];
				grab(matchstart, matchend, filesize);
				if (numtoi(filesize) && !got_bad_number)
					ftp_job->file_size = numtoi(filesize);
			}
			point = from;
			if (found_ftp_status("^220 "))	// Welcome msg.
				set_both_mode_message("[FTP: Sending user name] ");
			if (found_ftp_status("^331 "))	// Host wants password.
				set_both_mode_message("[FTP: Sending password] ");
			if (found_ftp_status("^230 "))	// User logged in.
				set_both_mode_message("[FTP: Working] ");
			if (found_ftp_status("^530 "))	// Bad password.
				discard_password_data();
			if (re_search(1, "^[45][0-9][0-9] ") && !parse_string(-1, "PASV\n... ")) {
				parse_string(1, ".*", msg);		// Some error code.
				say("FTP error: %s", msg);
				// Don't go to log if we're just creating a new file.
				if (ftp_job->operation != FTP_RECV
					|| !parse_string(-1, "^550 .*"))
					show_log = 1;
			}
			break;
	}
	if (buffer_ftp_activity)	// Do some buffer-specific FTP work.
		buffer_ftp_activity(activity, buf, from, to);
	switch (activity) {		// These don't automatically refresh later.
		case NET_DONE:
		case NET_LOG_DONE:
		case NET_SEND:
			refresh();
			break;
	}
	if (show_log) {
		if (num_of_popups() > 0 || orig_buf != ftp_job->buf) {
			say("FTP errors occurred, see buffer %s.",
				buf_exist(ftp_job->log_buf) ?
				bufnum_to_name(ftp_job->log_buf) : "<deleted>");
			if (buf_exist(ftp_job->buf) && ftp_job->operation == FTP_SEND) {
				save_var bufnum = ftp_job->buf;
				modified = 1;
			}
		} else {
			restore_vars();
			save_var iter, has_arg;
			save_var bufnum = buf;
			if (buf_exist(ftp_job->buf)) {
				to_buffer_num(ftp_job->buf);
				if (ftp_job->operation == FTP_SEND)
					modified = 1;
			}
			if (buf_exist(ftp_job->log_buf))
				to_buffer_num(ftp_job->log_buf);
			point = size();
			refresh();
			restore_vars();
			bufnum = window_bufnum;
		}
	}
	restore_vars();
	if (num_of_popups() > 0 && !strcmp(bufname, CONNECT_LIST_BUF)) {
		int line = lines_between(0, point, 0) + 1, col = current_column();
		make_connect_list();
		go_line(line);
		move_to_column(col);
		refresh();
	}
}


// Kill any stuck ftp job in buffer buf, permitting another ftp job to
// run there.  Also kills telnet jobs.
kill_ftp_job(int buf)
{
	save_var bufnum = buf;
	if (ftp_job)
		say("Disconnecting buffer from ftp job in %s.", bufname);
	else if (telnet_id)
		telnet_send(telnet_id, NULL);
	ftp_job = 0;
}

// Is this line in an FTP log part of a response?
// Otherwise, it must be a valid FTP server command like RETR.

is_ftp_log_response_line()
{
	save_var point;
	to_begin_line();
	if (isdigit(curchar()))
		return 1;
	if (re_search(-1, "^[0-9][0-9][0-9]!"))	// Look for prev response line.
		if (curchar() == '-')	// Found a continuation line.
			return 1;
	return 0;
}

color_ftp_log_range(from, to)
{
	int color;

	if (from >= to)
		return to;
	save_var point = from, matchstart, matchend;
	to_begin_line();
	do {
		if (is_ftp_log_response_line())
			color = color_class process_output;
		else
			color = color_class process_input;
		set_character_color(point, give_end_line() + 1, color);
	} while (point < to && nl_forward());
	return to;
}

do_ftp_op(char *dest, char *host, char *port, char *usr, char *pwd,
		  char *fname, int cmd)
{
	char log[FNAMELEN], *p;
	int id, i = 1, portnum;
	struct ftp_job *job;

	portnum = numtoi(port);
	if (got_bad_number)
		portnum = 21;
	save_var bufnum;
	bufname = dest;
	if (ftp_job)
		error("This buffer's FTP job has not yet completed.");
	strcpy(log, "FTP Log");
	p = log + strlen(log);
	for (;;) {
		if (!exist(log))
			break;
		bufname = log;
		if (!ftp_job)
			break;
		sprintf(p, " %d", i++);
	}
	if (ftp_compatible_dirs & 1)
		cmd |= FTP_USE_CWD;
	if (ftp_compatible_dirs & 2)
		cmd |= FTP_PLAIN_LIST;
	id = ftp_op(dest, log, host, portnum, usr, pwd, fname, cmd);
	if (id) {

		bufname = dest;
		job = (struct ftp_job *) malloc(sizeof(struct ftp_job));
		job->id = id;
		job->operation = cmd & FTP_OP_MASK;
		job->buf = bufnum;
		job->log_buf = name_to_bufnum(log);
		job->file_size = 0;
		job->time_started = time_ms();
		ftp_job = job;
		when_net_activity = ftp_activity;

		bufname = log;
		ftp_job = job;
		discardable_buffer = 1;
		when_net_activity = ftp_activity;
		translation_type = FILETYPE_MSDOS;
		recolor_range = color_ftp_log_range;	// set up coloring rules
		recolor_from_here = recolor_by_lines;
		when_setting_want_code_coloring();	// maybe turn on coloring
		set_both_mode_message("[FTP: Connecting] ");
	}
}

int find_remote_file(char *file, int transl)
{
	int op = FTP_RECV;

	inet_only();
	buf_zap(bufnum);
	type_point = 0;
	_read_aborted = 0;
	modified = 0;
	filename = file;
	unique_file_identifier[0] = 0;
	translation_type = transl;
	if (ftp_ascii_transfers && GET_LINE_TRANSLATE(transl) != FILETYPE_BINARY)
		op |= FTP_ASCII;
	url_operation(file, op);
	if (synchronous_file_io)
		wait_for_remote_file(bufnum);
	return 0;
}

int save_remote_file(char *fname, int trans, struct file_info *finfo)
{
	int op = FTP_SEND;

	if (ftp_job)
		error("This buffer's FTP job has not yet completed.");
	inet_only();
	if (finfo)
		finfo->check_type = 0;		// Don't check file dates.
	translation_type = trans;
	modified = 0;
	if (ftp_ascii_transfers && GET_LINE_TRANSLATE(trans) != FILETYPE_BINARY)
		op |= FTP_ASCII;
	url_operation(fname, op);
	if (synchronous_file_io)
		wait_for_remote_file(bufnum);
	return 0;
}

#define HTTP_HEADERS	"HTTP Headers"

http_activity(int activity, int buf, int from, int to)
{
	int headers, pos, orig;
	char err[FNAMELEN];

	UNREFERENCED_PARAMETER(to);
	if (!buf_exist(buf))
		return;
	save_var bufnum = buf;
	if (activity == NET_DONE) {
		set_mode_message(NULL);
		refresh();
		if (when_net_activity == http_activity)
			when_net_activity = 0;
		return;
	}
	if (!from)
		position_in_all_windows(buf, 0);
	bufnum = buf;
	if (mode_extra && !strstr(mode_extra, "Connecting")) {
		if (from < 1000 && mode_keys == buf_tab)
			suffix_none();		// Switch out of Fundamental mode.
		return;
	}
	save_spot point = 0;
	if (re_search(1, "\r?\n\r?\n")) {
		set_mode_message("[HTTP: Retrieving] ");
		pos = point;
		headers = create(HTTP_HEADERS);	// Got end of headers yet?
		set_buf_point(headers, orig = buf_size(headers));
		buffer_printf(HTTP_HEADERS, "Headers for %s\n", filename);
		xfer(HTTP_HEADERS, 0, pos);
		point = 0;
		if (parse_string(1, "HTTP/[.0-9]+ [1345]")) {
			int m = matchend - 1;
			grab(m, MIN(parse_string(1, "[^\n\r]*"), 200), err);
			say("Response: %s, see buffer \"HTTP Headers\" for details.", err);
		}
		save_var modified = 0;	// Ignore following change to buffer.
		delete(0, pos);
		if (mode_keys == buf_tab)
			suffix_none();		// Switch out of Fundamental mode.
		bufname = HTTP_HEADERS;
		discardable_buffer = 1;	// Deleting the headers is ok.
		save_var narrow_start = orig;
		simple_re_replace(1, "\r\n", "\n");
		translation_type = FILETYPE_MSDOS;
	}
}

// Prompt for password for this usr@host, unless we already did.
// Remember previous responses to each usr/host combo, and send it back
// without prompting (but never save this info in a state file).
// Call with res==NULL to discard a particular password and prompt
// again next time it's needed.

int get_password(char *res, char *host, char *usr)
{
	char msg[FNAMELEN];
	struct password_data *p;
	int use_old;

	for (p = password_data; p; p = p->next)
		if (!strcmp(p->host, host) && !strcmp(p->usr, usr)) {
			if (p->invalid)
				break;
			if (res)
				strcpy(res, p->pwd);
			else				// Mark it invalid so we prompt next time.
				p->invalid = 1;
			return 0;
		}
	if (!res)		// Told to discard a pwd, but it's not there.
		return 0;
	if (_doing_input) {
		say("Can't prompt for a password within another prompt.");
		return 1;
	}
	use_old = (p != 0);		// Replace existing pwd if it was refused.
	sprintf(msg, "Password for %s@%s: ", usr, host);
	comp_read(res, msg, (int (*)()) 0, POP_UP_PROMPT | PASSWORD_PROMPT, "");
	if (!use_old)
		p = (struct password_data *) malloc(sizeof(struct password_data));
	strcpy(p->host, host);
	strcpy(p->usr, usr);
	strcpy(p->pwd, res);
	p->invalid = 0;
	if (!use_old) {		// If we're not replacing an invalid pwd,
		p->next = password_data;	// add to the list.
		password_data = p;
	}
	return 0;
}

// Look in this log file for the host and username, then discard
// any saved password data.
discard_password_data()
{
	char hostname[FNAMELEN], username[FNAMELEN];
	save_var point, case_fold = 1;
	if (!re_search(-1, "^000 Connecting to !"))
		return;
	parse_string(1, ".*", hostname);
	if (!re_search(1, "^user "))
		return;
	parse_string(1, ".*", username);
	get_password(NULL, hostname, username);	// Discard password.
}

// Parse and check a URL, filling in its component parts.

int prepare_url_operation(char *file, int op, struct url_parts *parts)
{
	char *p;

	save_var case_fold = 1;
	if (!parse_url(file, parts))
		error("Bad URL %s", file);
	if (!strstr(" ftp scp ", parts->service) && (op & FTP_OP_MASK) != FTP_RECV)
		error("Can't do that with %s: buffers, only ftp: or scp: buffers.",
			  parts->service);
	if (!strstr(" ftp scp http https telnet ssh ", parts->service))
		error("Unknown service %s: in URL.", parts->service);

	if (!strcmp(parts->service, "ftp")) {	// An FTP URL?
		if (!parts->have_usr)		// No user specified.
			strcpy(parts->usr, "anonymous");
		if (!parts->have_password)	// If no password yet,
			if (!strfcmp(parts->usr, "anonymous")	// Use anon pwd?
				|| !strfcmp(parts->usr, "guest"))
				strcpy(parts->pwd, anon_ftp_password);
			else				// Or prompt for one/retrieve old.
				if (get_password(parts->pwd, parts->host, parts->usr))
					return 1;
	} else if (!strcmp(parts->service, "http") || !strcmp(parts->service, "https")) {
		if (parts->have_usr && !parts->have_password)
			get_password(parts->pwd, parts->host, parts->usr);
	} else if (!strcmp(parts->service, "scp") || !strcmp(parts->service, "ssh")) {
	} else if (parts->have_usr)
		error("User name not supported in %s URL's", parts->service);

	while (p = index(parts->fname, '\\'))
		*p = '/';			// Convert all \'s to /'s.
	return 0;
}

do_http(struct url_parts *parts, int port)
{
	char host_pat[FNAMELEN], fname[FNAMELEN], auth[FNAMELEN];

	when_net_activity = http_activity;
	*auth = 0;
	set_mode_message("[HTTP: Connecting] ");
	if (parts->have_usr)
		sprintf(auth, "%s:%s", parts->usr, parts->pwd);
	sprintf(host_pat, "|%s|", parts->host);
	if (http_proxy_port && !strstr(http_proxy_exceptions, host_pat)) {
		sprintf(fname, "http://%s:%d/%s", parts->host, port, parts->fname);
		http_retrieve(fname, http_proxy_server, http_proxy_port,
					  auth, bufname, 0);
	} else {
		sprintf(fname, "/%s", parts->fname);
		http_retrieve(fname, parts->host, port, auth, bufname, 0);
	}
}

// Try to retrieve a file via http and store it on disk.  Sz is the
// expected size, for progress messages.  Returns 0 if OK.
int download_file_to_disk(char *url, char *fname, int sz)
{
	int orig = bufnum, b, got_sz;
	do_find(url, FILETYPE_AUTO);
	b = bufnum;
	if (strcmp(filename, url))
		return 1;
	to_buffer_num(orig);
	bufnum = b;
	while (mode_extra) {
		got_sz = size();
		bufnum = orig;
		if (sz)
			note("Downloading file... %'d of %'d", got_sz, sz);
		delay(1, COND_KEY);
		check_abort();
		bufnum = b;
	}
	bufnum = orig;
	say("Done, writing %f to disk...", fname);
	bufnum = b;
	int err = file_write(fname, FILETYPE_AUTO);
	bufnum = orig;
	drop_buffer(bufnum_to_name(b));
	if (!err)
		say("");
	return err;
}

https_activity(int activity, int buf, int from, int to)
{
	UNREFERENCED_PARAMETER(to);
	if (!buf_exist(buf))
		return;
	save_var bufnum = buf;
	if (activity == NET_DONE) {
		set_mode_message(NULL);
		refresh();
		if (when_activity == https_activity)
			when_activity = 0;
		return;
	}
	if (!from) {
		position_in_all_windows(buf, 0);
		if (mode_keys == buf_tab)
			suffix_none();		// Switch out of Fundamental mode.
	}
}

do_https(struct url_parts *parts, int port)
{
	char cmdline[FNAMELEN], dir[FNAMELEN];

	when_activity = https_activity;
	set_mode_message("[HTTPS: Connecting] ");
	get_executable_directory(dir);
	sprintf(cmdline, ".%cgethttps ", path_sep);
	if (parts->have_usr)
		sprintf(cmdline + strlen(cmdline), "--password=\"%s\" ", parts->pwd);
	sprintf(cmdline + strlen(cmdline), "\"https://");
	if (parts->have_usr)
		sprintf(cmdline + strlen(cmdline), "%s@", parts->usr);
	sprintf(cmdline + strlen(cmdline), "%s", parts->host);
	if (port)
		sprintf(cmdline + strlen(cmdline), ":%d", port);
	sprintf(cmdline + strlen(cmdline), "/%s\"", parts->fname);
	if (!buf_pipe_text(0, bufnum, cmdline, dir, PIPE_CLEAR_BUF))
		file_error(errno, "Running gethttps helper", "Couldn't start gethttps helper");
}

// Feed next cmd to process.
scp_next_cmd(struct scp_job *p)
{
	char *s;

	if (p->flags & SCP_CMD_ERROR) {
		p->cmd = 0;
		if (p->destbuf)
			set_buf_mode_message(p->destbuf, NULL);
	}
	if (p->cmd && *p->cmd) {
		p->busy = 1;
		if (p->flags & SCP_CLIENT_MODE)
			buf_printf(p->buf, "## ");
		if (s = strchr(p->cmd, '\n')) {
			buf_stuff(p->buf, p->cmd, s++ - p->cmd + 1);
		} else {
			buf_stuff(p->buf, p->cmd, strlen(p->cmd));
			buf_stuff(p->buf, "\n", 1);
		}
		p->cmd = s;
	}
}

// Start a process in the buffer for job p running cmdline and using
// curdir.  Also save in case we have to restart later.
scp_start_process(struct scp_job *p, char *cmdline)
{
	if (p->cmdline)
		free(p->cmdline);
	p->cmdline = strsave(cmdline);
	scp_restart_process(p);
}

scp_restart_process(struct scp_job *p)
{
	char dir[FNAMELEN];

	strcpy(dir, p->file);
	if (is_remote_file(dir))
		*dir = 0;
	else
		*get_tail(dir, 1) = 0;
	buffer_printf("-scp-trace", "Run %s in %s\n", p->cmdline, dir);
	if (concur_shell("", p->cmdline, *dir ? dir : NULL,
					 bufnum_to_name(p->buf)))
		error("Couldn't start scp process");
	save_var bufnum = p->buf;
	if (is_unix && strstr(p->cmdline, "sftp"))
		process_echo &= 3;	// Turn off our echoing, since sftp does it too.
}

do_when_idle_scp(int cnt)
{
	struct scp_job *p = scp_job;
	if (cnt == 0)
		for (; p; p = p->next)
			if (p->flags & SCP_RESTART) {
				p->flags &= ~SCP_RESTART;
				scp_restart_process(p);
				say("Connection reestablished.");
			}
	return -1;
}

// We've been notified that an scp buffer's process has exited.  Clean
// up.

scp_handle_done(struct scp_job *p)
{
	char fname[FNAMELEN];

	p->busy = 0;
	if (!buf_exist(p->destbuf))
		return;
	if (p->cmd) {
		if (p->cmd == strchr(p->cmdlist, '\n') + 1) {
			say("Connection to site was lost; reconnecting.");
			buffer_printf("-scp-trace", "Restart connection: |%s| of |%s|\n",
				  p->cmd, p->cmdlist ? p->cmdlist : "<none>");
			p->cmd = p->cmdlist;
			p->flags |= SCP_RESTART;
			return;
		}
	}
	set_buf_mode_message(p->destbuf, NULL);
	if (p->op == FTP_RECV && (p->flags & SCP_TMP_FILE)) {
		save_var bufnum = p->destbuf;		// The cmd running was supposed to
		strcpy(fname, filename);			// retrieve a file.  Now load it.
		do_file_read(p->file, translation_type);
		filename = fname;
		file_info.check_type = 0;	// Prevent checking its date.
		file_info.year = 0;			// Let session save it, though.
		call_mode(filename);
		check_filevars();
	}
	if (p->op == FTP_SEND && (p->flags & SCP_TMP_FILE)
		  && !(p->flags & SCP_CMD_ERROR)) {
		save_var bufnum = p->destbuf;
		say("%f: written.", filename);
	}
	if (p->flags & SCP_TMP_FILE)
		delete_file(p->file);
}

// List is a series of commands separated by \n's.  Return nonzero if
// cmd matches ones.
int scp_echoed_cmdline(char *cmd, char *list)
{
	char *end;
	if (!cmd || !list)
		return 0;
	while (end = strchr(list, '\n')) {
		if (!strncmp(list, cmd, end - list))
			return 1;
		list = end + 1;
	}
	return !strcmp(cmd, list);
}

// The scp program produced some output doing a file xfer, which
// it normally shouldn't.  Show it.
scp_handle_xfer(struct scp_job *p, int from, int to)
{
	char msg[FNAMELEN];

	UNREFERENCED_PARAMETER(p);
	grab(from, MIN(to, from + sizeof(msg) / sizeof(*msg) - 1), msg);
	if (*msg && msg[strlen(msg) - 1] == '\n')
		msg[strlen(msg) - 1] = 0;
	if (strcmp(msg, "Exited"))
		say("%s", msg);
}

// Deal with output from epsilon-xfer-client process.  Return nonzero to
// feed next command.

int scp_client_handle_list(struct scp_job *p)
{
	int i, nextcmd = 0;
	char msg[FNAMELEN];

	while (nl_forward()) {
		i = point;
		point--;
		to_begin_line();
		if (character(point) == 'd' && character(point + 1) == ' ') {
			if (buf_exist(p->destbuf)) {	// Got some data, copy where it belongs.
				if (*p->file)
					buf_printf(p->destbuf, "%s", p->file);
				buf_xfer(p->destbuf, point + 2, i);
			}
		} else if (character(point) == 's' && character(point + 1) == ' ') {
			if (buf_exist(p->destbuf) && (!p->cmd || !*p->cmd))
				set_buf_mode_message(p->destbuf, NULL);
			p->busy = 0;			// Got a status msg.
			nextcmd = 1;			// Send a pending command if any.
		} else if (parse_string(1, "Epsilon-xfer-helper version ")) {
			if (character(matchend) < '1')
				say("Bad epsilon-xfer-helper version: %c", character(matchend));
			p->busy = 0;
			nextcmd = 1;
		} else if (!parse_string(1, "(## |Exited)")) {
			grab(point, MIN(give_end_line(),
							point + sizeof(msg) / sizeof(*msg) - 1), msg);
			if (*msg)		// Any other line is an error msg.
				say("%s", msg);
		}
		point = i;
	}
	return nextcmd;
}

// Deal with output from sftp program.  Return nonzero to feed next
// command.

int scp_sftp_handle_list(struct scp_job *p)
{
	int i, isdata;
	char msg[FNAMELEN];

	save_var case_fold = 1;
	while (nl_forward()) {
		i = point;
		point--;
		to_begin_line();
		isdata = 0;
		if (parse_string(1, "[-dcblsp][-r][-w][-sxt][-r][-w][-sxt]"
						 "[-r][-w][-sxt] .*" UNIX_DATE))
			isdata = 1;
		if ((p->flags & SCP_CHECK_IS_DIR) && parse_string(1,
			".* (not a directory|No such|not found)"))
			isdata = 1;
		if (parse_string(1, ".* No such file or directory")
			  && p->cmd && !strncmp(p->cmd, "ls ", 3)) {
			if (buf_exist(p->destbuf))	// Error part way through, so don't
				set_buf_mode_message(p->destbuf, NULL);	// copy any more.
			p->flags |= SCP_CMD_ERROR;
		}
		if (isdata) { // Got some data, copy where it belongs.
			if (buf_exist(p->destbuf) && !(p->flags & SCP_CMD_ERROR)) {
				if (*p->file && !(p->flags & (SCP_CHECK_IS_DIR
											  | SCP_CHECK_MATCHES)))
					buf_printf(p->destbuf, "%s", p->file);
				buf_xfer(p->destbuf, point, i);
			}
		} else if (parse_string(1, "Remote (working )?directory(:| is| is now) ")) {
			grab(matchend, give_end_line(), msg);
			if (p->lastdir)
				free(p->lastdir);
			p->lastdir = strsave(msg);
			if (!p->homedir)
				p->homedir = strsave(msg);
		} else if ((p->flags & SCP_IS_XFER)	// Ignore this msg.
				   && parse_string(1, "((Fetching|Uploading) .* to |"
								   "(remote|local):.* => (remote|local):|"
								   ".*--:-- ETA)")) {
		} else if (!parse_string(1, "(Exited|Connect(ing|ed) to "
								 "|Connection closed|Listing directory "
								 "|Using username \"|sftp%>"
								 "|(.* |^)pass(word|phrase).*:)")) {
			grab(point, MIN(give_end_line(),
							point + sizeof(msg) / sizeof(*msg) - 1), msg);
			if (*msg && !scp_echoed_cmdline(msg, p->cmdlist))
				say("%s", msg);		// Any other line is an error msg.
			if ((p->flags & SCP_IS_XFER)
				  && parse_string(1, ".*: (Permission denied|No such file"
								  "|Device out of space|Read-only file system"
								  "|Connection timed out|Could not)"))
				p->flags |= SCP_CMD_ERROR;
		}
		point = i;
	}
	if (parse_string(-1, "p?sftp%> *")) {
		if (!p->cmd || !*p->cmd) {
			if (buf_exist(p->destbuf))
				set_buf_mode_message(p->destbuf, NULL);
			if (p->flags & SCP_IS_XFER)
				scp_handle_done(p);
		}
		p->busy = 0;			// Got a prompt for another cmd.
		return 1;			// Send a pending command if any.
	}
	return 0;
}

// The remote program produced some output in response to our request
// for info.  Handle it as appropriate.
scp_handle_list(struct scp_job *p, int from, int to)
{
	int start, old_modified, old, old_busy, nextcmd;

	UNREFERENCED_PARAMETER(to);
	save_var point = from;
	if (buf_exist(p->destbuf)) {
		old = get_buf_point(p->destbuf);
		old_modified = get_buf_modified(p->destbuf);
		start = buf_size(p->destbuf);
		old_busy = p->busy;
		set_buf_point(p->destbuf, start);
	}
	if (p->flags & SCP_CLIENT_MODE)
		nextcmd = scp_client_handle_list(p);
	else
		nextcmd = scp_sftp_handle_list(p);
	if (buf_exist(p->destbuf)) {
		set_buf_point(p->destbuf, old);
		set_buf_modified(p->destbuf, old_modified);
		if (p->op == FTP_LIST) {	// Say we've got more dired results.
			if (!(p->flags & SCP_CMD_ERROR))
				dired_ftp_activity(NET_RECV, p->destbuf, start,
								   buf_size(p->destbuf));
			if (old_busy == 1 && !p->busy && !p->cmd)
				dired_ftp_activity(NET_DONE, p->destbuf, 0, 0);
		} else if ((p->flags & SCP_IS_XFER) && (p->flags & SCP_CMD_ERROR)
				   && p->op == FTP_SEND)
			set_buf_modified(p->destbuf, 1);
	}
	if (nextcmd)	// If there's a pending command, send it now.
		scp_next_cmd(p);
	if (!p->busy && !p->cmd)
		p->destbuf = 0;	// Finished the operation, so don't copy any more.
}

// Handle activity in scp buffer.
scp_handler(int activity, int buf, int from, int to)
{
	struct scp_job *p;

	save_var bufnum = buf, abort_searching = 0;
	for (p = scp_job; p; p = p->next)	// Find the data for this buffer.
		if (bufnum == p->buf)
			break;
	if (!p || !p->destbuf)
		return;
	if (activity == NET_DONE)
		scp_handle_done(p);
	else if (activity == NET_RECV) {
		if (p->op == FTP_LIST || p->op == FTP_MISC || (p->flags & SCP_IS_XFER))
			scp_handle_list(p, from, to);
		else
			scp_handle_xfer(p, from, to);
		if (matches_at(to, -1, "(.* |^)pass(word|phrase).*: ?")) {
			ungot_key = WIN_HELP_REQUEST;	// Make Epsilon run this cmd:
			menu_command = "password-prompt";
		}
	}
	refresh();
}

// Find or allocate an scp_job structure associated with this buffer,
// and make a work buffer for it too.
struct scp_job *find_scp_buffer(char *host, char *uname, int op)
{
	char buf[FNAMELEN];
	struct scp_job *p;
	int i, b, separate_xfers;

	for (p = scp_job; p; p = p->next)
		if (!strfcmp(host, p->host) && !strcmp(uname, p->uname)
			  && !p->busy && buf_exist(p->buf)) {
			separate_xfers = p->flags & (SCP_CLIENT_MODE | SCP_SCP_XFER);
			if (!separate_xfers || (op & 2) == (p->op & 2))
				return p;
		}
	for (i = 1; ; i++) {
		sprintf(buf, "scp-helper-%s-%s-%d", host, uname, i);
		if (!exist(buf))
			break;
	}
	b = create(buf);
	p = (struct scp_job *) malloc(sizeof(struct scp_job));
	p->host = strsave(host);
	p->uname = strsave(uname);
	p->buf = b;
	p->next = scp_job;
	scp_job = p;
	save_var bufnum = b;
	when_activity = scp_handler;
	discardable_buffer = 1;
	buffer_not_saveable = 1;
	return p;
}

// Prepare an scp_job structure associated with this buffer.
struct scp_job *prep_scp_buffer(struct url_parts *parts, int op)
{
	struct scp_job *p;

	if (!(has_feature & FEAT_MULT_CONCUR))
		error("SCP functions require Unix or Windows NT/2K/XP");
	p = find_scp_buffer(parts->host, parts->usr, op);
	p->destbuf = bufnum;
	p->busy = 1;
	p->op = op;
	if (p->cmdlist)
		free(p->cmdlist);
	p->cmdlist = 0;
	p->cmd = 0;
	p->flags = scp_flags(parts);
	set_buf_point(p->buf, buf_size(p->buf));
	set_mode_message("[SCP: Working] ");
	return p;
}

// Add host name to variable var and see if that makes a variable name.
// If so, return its index.  If not, return index of the original
// variable's name, or 0 if none.

int find_variable_by_host(char *ovar, char *host)
{
	char var[FNAMELEN], *s, *h;
	int i;

	strcpy(var, ovar);
	s = var + strlen(var);
	*s++ = '-';
	for (h = host; *h; h++, s++)
		*s = isalnum(*h) ? *h : '-';
	*s = 0;
	if (i = find_index(var))
		return i;
	return find_index(ovar);
}

// Return the contents of the appropriate variable, looking for
// host-name-specific too.

char *scp_variable(struct url_parts *parts, char *varname)
{
	char var[FNAMELEN], *u;

	sprintf(var, "%s%s%s%s-template",
			parts->service, *varname ? "-" : "",
			varname, parts->have_usr ? "" : "-no-user");
	int i = find_variable_by_host(var, parts->host);
	if (!i || !(u = get_str_var(i)))
		error("No variable %s.", var);
	return u;
}

// Return scp flags appropriate for this URL.  Right now we just
// distinguish xfer-client mode from sftp mode.
int scp_flags(struct url_parts *parts)
{
	if (strcmp(parts->service, "scp"))
		return 0;
	int i = find_variable_by_host("scp-client-style", parts->host);
	return (i ? get_num_var(i) : 0) & 7;	// Just take user's bits.
}

// Our winpty helper comes in both 32-bit and 64-bit versions.
// Determine which type of Cygwin install is on the path, if any, and
// what %w should expand to.
get_winpty(char *winpty, int turn_off_echo)
{
	char exename[FNAMELEN];

	*winpty = 0;
	if (opsys != OS_WINDOWS)
		return;
	if (!cygwin_type) {
		char *buf = temp_buf();
		shell("", "uname -m", buf, SHELL_HIDE);		// Get text with Cygwin install type.
		save_var bufname = buf;
		on_exit delete_buffer(buf);
		point = 0;
		cygwin_type = search(1, "_64") ? 2 : re_search(-1, "i.86") ? 1 : 3;
		show_text(-2, 0, "Cygwin type detected: %d", cygwin_type);
	}
	if (cygwin_type == 3)
		return;
	get_executable_directory(exename);
	convert_to_8_3_filename(exename);
	sprintf(winpty, "\"%swinpty%s\" -a%s",
			exename,
			(cygwin_type == 1) ? "32" : "64",
			turn_off_echo ? " -n" : "");
}

// Put value of sftp or ssh command line variable for this OS into
// dest.  Def is a default value.  Pat is a pattern with %s where the
// OS name goes.

char *ssh_command_variable(char *dest, char *def, char *pat, struct url_parts *parts, int turn_off_echo)
{
	char s[FNAMELEN], winpty[FNAMELEN], *vals[1];
	int i;
	get_winpty(winpty, turn_off_echo);
	vals[0] = winpty;
	strcpy(dest, def);
	sprintf(s, pat, opsys == OS_UNIX ? "unix" : "windows");
	i = find_variable_by_host(s, parts->host);
	if (i)
		expand_string_template(dest, get_str_var(i), "w", vals);
}

// Transfer a file using scp or sftp, using the command specified by
// template, populated by values from parts.  Parts specifies the remote
// file name, and p indicates the local one; the direction of transfer
// is implicit in the command template.

// If use_tmp, create a temporary file first, writing the current buffer
// to it (if mode is -1) or leaving it empty (mode 0), and recording the
// temp name in p.

do_scp_xfer(struct url_parts *parts, struct scp_job *p, char *template,
			int mode, int use_tmp)
{
	char tmpfile[FNAMELEN], cmd[FNAMELEN], *args[5], *s, cmdname[FNAMELEN];

	if (use_tmp) {
		new_file_write("", translation_type, NULL, FILE_IO_TEMPFILE, mode);
		strcpy(p->file, get_tempfile_name());
		p->flags |= SCP_TMP_FILE;
	}
	if (!(p->flags & (SCP_SCP_XFER | SCP_CLIENT_MODE))
		  && !strchr(p->file, '%')) {
		scp_run_helper(parts, p);
		p->flags |= SCP_IS_XFER;
		sprintf(cmd, (p->op == FTP_SEND) ? "put \"%s\" \"%%s\""
										 : "get \"%%s\" \"%s\"", p->file);
		if (is_mswindows)
			for (s = cmd; s = strchr(s, '\\'); )
				*s = '/';	// Windows sftp gets confused by \'s sometimes.
		scp_send_some_command(parts, p, cmd);
		return;
	}
	strcpy(tmpfile, get_tail(p->file, 1));
	args[0] = parts->usr;
	args[1] = parts->host;
	args[2] = parts->fname;
	args[3] = tmpfile;
	ssh_command_variable(cmdname, "sftp", "scp-%s-sftp-command", parts, 1);
	args[4] = cmdname;
	if (expand_string_template(cmd, template, "uhflc", args))
		error("Invalid scp template.");
	scp_start_process(p, cmd);
}

scp_send_some_command(struct url_parts *parts, struct scp_job *p, char *fmt)
{
	int len = strlen(parts->fname) * 2 + strlen(fmt) + 40;
	if (!(p->flags & SCP_CLIENT_MODE) && p->homedir)
		len += strlen(p->homedir);
	p->cmd = p->cmdlist = malloc(len);
	if (!(p->flags & SCP_CLIENT_MODE)) {
		char dirpart[FNAMELEN];
		if (p->busy == 2)
			sprintf(p->cmd, "pwd\n");
		else if (p->homedir)
			sprintf(p->cmd, "cd \"%s\"\n", p->homedir);
		else
			error("Can't determine home directory.");
		strcpy(dirpart, parts->fname);
		*get_tail(dirpart, 1) = 0;
		if (*dirpart)
			sprintf(p->cmd + strlen(p->cmd), "cd \"%s\"\n", dirpart);
		if (*fmt)
			sprintf(p->cmd + strlen(p->cmd), fmt, get_tail(parts->fname, 1));
	} else
		sprintf(p->cmdlist, "list \"%s\"\n", parts->fname);
	if (p->busy != 2)
		scp_next_cmd(p);
}

scp_run_helper(struct url_parts *parts, struct scp_job *p)
{
	char cmd[FNAMELEN], *args[3], cmdname[FNAMELEN];

	args[0] = parts->usr;
	args[1] = parts->host;
	ssh_command_variable(cmdname, "sftp", "scp-%s-sftp-command", parts, 1);
	args[2] = cmdname;
	if (expand_string_template(cmd, scp_variable(parts, "run-helper"),
							   "uhc", args))
		error("Invalid run helper template.");
	if (!is_process_buffer(p->buf)) {
		if (is_mswindows) {
			char *oldval = getenv("CYGWIN");
			if (!oldval)
				putenv("CYGWIN=nodosfilewarning");
			else if (!strstr(oldval, "nodosfilewarning")) {
				char *newval = malloc(strlen(oldval) + 50);
				sprintf(newval, "CYGWIN=%s nodosfilewarning", oldval);
				putenv(strkeep(newval));
				free(newval);
			}
		}
		scp_start_process(p, cmd);
		if (p->homedir)
			free(p->homedir);
		p->homedir = 0;
		p->busy = 2;
	}
}

struct scp_job *do_scp(struct url_parts *parts, int op)
{
	struct scp_job *p;

	p = prep_scp_buffer(parts, op);
	if (op == FTP_RECV) {
		do_scp_xfer(parts, p, scp_variable(parts, "read-file"), 0, 1);
	} else if (op == FTP_SEND) {
		do_scp_xfer(parts, p, scp_variable(parts, "write-file"), -1, 1);
	} else if (op == FTP_LIST || op == FTP_MISC) {
		scp_run_helper(parts, p);
		*p->file = 0;
		if (op == FTP_LIST) {
			char cmd[FNAMELEN], *val;
			val = get_str_var(find_variable_by_host("scp-list-flags",
				parts->host));
			sprintf(cmd, "ls %s %%s", val ? val : "-l");
			scp_send_some_command(parts, p, (p->flags & SCP_OLD_SFTP)
								  ? "ls %s" : cmd);
		}
	}
	return p;
}

// Process the directory listing from sftp to get a list of matches for
// the pattern (maybe a prefix match), discarding non-matching lines.
// Add rest of url to each match.
scp_filter_sftp_matches(char *filter, int prefix, char *url)
{
	char fname[FNAMELEN];
	delete_final_newlines();
	point = 0;
	for (;;) {	// Result is in ls -l format.
		int start = point, isdir = curchar() == 'd';
		int len = parse_string(1, "^.*" UNIX_DATE " *");
		if (!len)
			break;
		delete(start, start + len);	// Want just the name.
		to_end_line();
		grab(start, point, fname);
		if (!fpatmatch(fname, filter, prefix, 0))
			delete(start, point + 1);	// Discard this line.
		else {
			if (isdir)
				stuff("/");
			if (!nl_forward())
				break;
		}
	}
	delete_final_newlines();
	point = 0;
	strcpy(fname, url);	// Now insert rest of URL before each match.
	*get_tail(get_url_file_part(fname, 0), 1) = 0;
	if (*fname && !is_path_separator(fname[strlen(fname) - 1]))
		strcat(fname, "/");
	if (size())
		simple_re_replace(1, "^", fname);
}

net_during_wait(int cond)
{
	delay(10, cond);
	if (ungot_key == WIN_HELP_REQUEST && menu_command
		  && !strcmp(menu_command, "password-prompt")) {
		if (_doing_input && !is_gui)
			error("Can't prompt for a password within another prompt.");
		ungot_key = -1;
		password_prompt();
	}
	if (ungot_key == -1 && char_avail()) {	// Needed for scp password prompts on non-GUI.
		wait_for_key();
		if (!IS_WIN_PASSIVE_KEY(key) && !IS_MOUSE_KEY(key))
			ungot_key = key;
	}
}	

// Send a command to helper to get info on this url and its parts, and
// wait for a response, returned in buffer b.  What info to get:
// 0=completions of this file name, 1=absolute form of relative dir
// name, 2=whether file name is for a directory or file (only for
// non-client mode), 3=wildcard matches of this pattern (with no *
// automatically added as for completions).

// For completions and wildcarding, we can't just ask for matches of
// pat*.  If pat* happens to match a single directory, say patdir,
// current sftp servers will list the matches in patdir, not the name
// patdir.  Ls's -d flag would prevent this, but current sftp servers
// don't recognize it.  So we ask for all matches in the parent dir,
// then filter the matches ourselves.

scp_get_response(int code, char *url, struct url_parts *parts, int b)
{
	char cmd[FNAMELEN], filter[FNAMELEN], *s, *val;
	struct scp_job *p;

	save_var bufnum = b;
	buf_zap(bufnum);
	p = do_scp(parts, FTP_MISC);
	strcpy(p->file, url);
	*get_url_file_part(p->file, 0) = 0;
	if (!(p->flags & SCP_CLIENT_MODE)) {
		switch (code) {
			case 1:
				s = "";
				break;
			case 2:
				p->flags |= SCP_CHECK_IS_DIR;
				s = "cd \"%s\"\n\n";
				break;
			default:		// Cases 0 and 3: list matches.
				p->flags |= SCP_CHECK_MATCHES;
				val = get_str_var(find_variable_by_host("scp-list-flags",
					parts->host));
				sprintf(s = cmd, "ls %s", (p->flags & SCP_OLD_SFTP) ? ""
					: val ? val : "-l");
				if (s[strlen(s) - 1] != ' ')
					strcat(s, " ");
				strcat(s, "%s");
				strcpy(filter, get_tail(parts->fname, 1));
				*get_tail(parts->fname, 1) = 0;	// Must list all files.
				break;
		}
		scp_send_some_command(parts, p, s);
	} else {
		sprintf(cmd, (code == 1 || code == 2) ? "prdir %s\n" : "comp %s\n",
				parts->fname);
		p->cmd = p->cmdlist = strsave(cmd);
		if (p->busy != 2)
			scp_next_cmd(p);
	}
	delayed_say(0, 100, 30, "Working...");
	while (p->busy && !user_abort)
		net_during_wait(COND_KEY);
	if ((p->flags & SCP_RESTART) && code == 2 && !(p->flags & SCP_CLIENT_MODE)) {
		p->flags &= ~SCP_RESTART;
		scp_restart_process(p);
		while (!user_abort && p->cmd && *p->cmd)
			net_during_wait(COND_KEY);
	}
	set_buf_mode_message(bufnum, NULL);
	user_abort = 0;
	delayed_say(0, 0, 0, "");
	say("");
	point = 0;
	if (!(scp_flags(parts) & SCP_CLIENT_MODE)) {
		switch (code) {
			case 0:
			case 3:		// Discard the extra info and filter nonmatches.
				scp_filter_sftp_matches(filter, code != 3, url);
				break;
			case 1:
				buf_printf(b, "%s%s", p->file, p->lastdir);
				set_buf_point(b, 0);
				break;
		}
	}
}

// Return control buffer associated with scp data buffer b, or 0 if none.
int scp_find_control(int b)
{
	struct scp_job *p;

	for (p = scp_job; p; p = p->next)
		if (b == p->destbuf && buf_exist(p->buf))
			return p->buf;
	return 0;
}

ftp_file_match(char *pat, int flags)
{
	char mypat[FNAMELEN * 2], name[FNAMELEN * 2], *p;

	save_var new_buffer_translation_type = FILETYPE_AUTO;
	save_var bufnum = zap("-ftp-file-match-listing");
	strcpy(mypat, pat);
	if (!(flags & EXACTONLY))
		strcat(mypat, "*");
	strcpy(name, mypat);
	p = get_tail(name, 1);	// If name contains wildcards, strip them.
	if (index(p, '*') || index(p, '?')) {
		*p = 0;								// Make FTP listing
		save_var ftp_compatible_dirs |= 1;	// show just the basename.
	}
	url_operation(name, FTP_LIST | FTP_WAIT);
	point = 0;
	bprintf("     Directory of %s\n", name);
	while (re_search(1, "^.")) {
		point--;
		stuff("  ");
		point++;
	}
	modified = 0;
	dired_orig_position = point = 0;
	dired_ftp_activity(NET_DONE, bufnum, 0, 0);
	color_dired_range(0, size());
	int b = zap(_MATCH_BUF);
	point = 0;
	for (point = 0; point < size() && nl_forward(); )
		if (dired_getname(name, 2) && (dired_isdir() ? !(flags & FM_NO_DIRS)
			: !(flags & FM_ONLY_DIRS)))
			if (fpatmatch(name, mypat, 0, 0))
				buf_printf(b, "%s%s\n", name, dired_isdir() == 1 ? "/" : "");
	set_buf_point(b, 0);
}

// Write to buffer b a list of URLs based on url, that match the
// partial host name.
remote_host_match(int b, char *url, char *parthost, int flags)
{
	char host[FNAMELEN], newurl[FNAMELEN], pat[FNAMELEN];
	int hostpos = strlen(url) - strlen(parthost);
	buf_zap(b);
	strcpy(newurl, url);
	if (exist(OLDRESP)) {
		save_var bufname = OLDRESP;
		save_var point = 0;
		sprintf(pat, "(%s):[/\\][/\\](%q[^/\\\n]+)", url_services, parthost);
		while (re_search(1, pat)) {
			grab(find_group(2, 1), find_group(2, 0), host);
			strcpy(newurl + hostpos, host);
			if ((flags & EXACTONLY)
				  ? !strcmp(parthost, host)
				  : !strnfcmp(parthost, host, strlen(parthost)))
				buf_printf(b, "%s/\n", newurl);
		}
		buf_sort_and_uniq(b);
	}
	set_buf_point(b, 0);
}

char *remote_file_match(char *s, int flags)
{
	struct url_parts parts;
	int b = create(_MATCH_BUF);

	if (!parse_url(s, &parts))
		return NULL;
	if (flags & STARTMATCH) {
		if (!*parts.fname && (!*parts.host || !is_path_separator(s[strlen(s) - 1])))
			remote_host_match(b, s, parts.host, flags);	// Complete on partial hostname.
		else if (!strcmp(parts.service, "scp"))
			scp_get_response((flags & EXACTONLY) ? 3 : 0, s, &parts, b);
		else if (!strcmp(parts.service, "ftp"))
			ftp_file_match(s, flags);
		else
			previous_response_matches(b, s, flags);
		if (flags & FM_NO_DIRS) {
			save_var bufnum = b;
			simple_re_replace(0, ".*[/\\]\n", "");
			restore_vars();
		}
	}
	if (!grab_line(b, _this_match))
		return NULL;
	return _this_match;
}

// Take some ftp-syntax commands in s and convert them to sftp syntax.
scp_rewrite_commands(struct scp_job *p, char *s)
{
	int tmp = tmp_buf();
	save_var bufnum = tmp, case_fold = 1;
	stuff(s);
	point = 0;
	if (p->busy == 2)
		bprintf("pwd\n");
	else if (p->homedir)
		bprintf("cd \"%s\"\n", p->homedir);
	else {
		say("Can't determine home directory.\n");
		restore_vars();
		buf_delete(tmp);
		return;
	}
	string_replace("^rmd (.*)", "rmdir \"#1\"", REGEX);
	string_replace("^mkd (.*)", "mkdir \"#1\"", REGEX);
	string_replace("^dele (.*)", "rm \"#1\"", REGEX);
	string_replace("^rnfr (.*)\nrnto (.*)", "rename \"#1\" \"#2\"", REGEX);
	while (re_search(1, "^cpfr (.*)\ncpto (.*)")) {
// 		char from[FNAMELEN], to[FNAMELEN];
// 		grab(find_group(1, 1), find_group(1, 0), from);
// 		grab(find_group(2, 1), find_group(2, 0), to);
		say("Remote=>remote copy not supported.");
		delete(matchstart, matchend);
	}
	s = malloc(size() + 1);
	grab(0, size(), s);
	restore_vars();
	buf_delete(tmp);
	p->cmd = p->cmdlist = s;
}

scp_misc_op(char *s, char *cmds)
{
	struct url_parts parts;
	struct scp_job *p;

	if (!parse_url(s, &parts) || strcmp(parts.service, "scp") || !*parts.host)
		return 0;
	p  = do_scp(&parts, FTP_MISC);
	if (p->flags & SCP_CLIENT_MODE)
		p->cmd = p->cmdlist = strsave(cmds);
	else
		scp_rewrite_commands(p, cmds);
	if (p->busy != 2)
		scp_next_cmd(p);
	return 1;
}

// Take a URL with a dirname that refers to the home dir or another
// user's home dir, and replace it with its absolute form.

scp_remote_dirname_absolute(char *s, struct url_parts *parts)
{
	char resp[FNAMELEN];
	int b = create(_MATCH_BUF);
	scp_get_response(1, s, parts, b);
	if (grab_line(b, resp))
		strcpy(s, resp);
}

// Return nonzero if this scp URL of a symbolic link points to a
// directory.  Always returns 0 in client mode, since client dired
// listings of symbolic links already distinguish directories.
int scp_is_url_a_dir(char *s, struct url_parts *parts)
{
	if (scp_flags(parts) & (SCP_CLIENT_MODE | SCP_NO_DIR_CHECK))
		return 0;
	if (!*parts->fname)
		return 1;
	int tmp = tmp_buf();
	scp_get_response(2, s, parts, tmp);
	int isdir = !buf_size(tmp);
	buf_delete(tmp);
	return isdir;
}

// Is buffer b currently used for scp?  Return op type, or -1 if none.
is_busy_doing_scp(int b)
{
	struct scp_job *p;

	save_var bufnum = b;
	for (p = scp_job; p; p = p->next)
		if (p->busy && (p->destbuf == b || p->buf == b))
			return p->op;
	return -1;
}

// If buffer b is currently actively retrieving a remote file, don't
// return until it finishes.

wait_for_remote_file(int b)
{
	save_var bufnum = b;
	for (;;) {
		if (!ftp_job && is_busy_doing_scp(b) == -1
			  && when_net_activity != http_activity
			  && when_activity != https_activity)
			return;
		net_during_wait(0);
		check_abort();
	}
}

is_retrieving_remote_file(int b)
{
	save_var bufnum = b;
	return ftp_job && (ftp_job->operation & FTP_OP_MASK) == FTP_RECV
			|| (is_busy_doing_scp(b) & FTP_OP_MASK) == FTP_RECV
			|| when_net_activity == http_activity
			|| when_activity == https_activity;
}

// Download remote file and save locally.
net_receive_file(char *from, char *to)
{
	struct url_parts parts;
	struct scp_job *p;

	wait_for_remote_file(bufnum);
	prepare_url_operation(from, FTP_RECV, &parts);
	p = prep_scp_buffer(&parts, FTP_RECV);
	strcpy(p->file, to);
	do_scp_xfer(&parts, p, scp_variable(&parts, "read-file"), 0, 0);
}

// Upload local file.
net_send_file(char *from, char *to)
{
	struct url_parts parts;
	struct scp_job *p;

	wait_for_remote_file(bufnum);
	prepare_url_operation(to, FTP_SEND, &parts);
	if (!*parts.fname)
		strcpy(parts.fname, ".");
	p = prep_scp_buffer(&parts, FTP_SEND);
	strcpy(p->file, from);
	do_scp_xfer(&parts, p, scp_variable(&parts, "write-file"), 0, 0);
}

// Parse the URL and begin remote activity with this buffer.
// If the URL specifies ftp://, the op may be FTP_RECV, FTP_LIST, or
// FTP_SEND, but only FTP_RECV is allowed for other types of URL.
// Takes a URL, possibly prompts for a password.  Can modify the URL
// in place.

int url_operation(char *file, int op)
{
	struct url_parts parts;
	int port, i;

	if (prepare_url_operation(file, op, &parts))
		return 1;
	port = numtoi(parts.port);
	if (*parts.port && got_bad_number)
		say("Ignoring invalid port number %s", parts.port);
	if (i = find_variable_by_host("fallback_remote_translation_type", parts.host))
		fallback_translation_type = get_num_var(i);
	if (!strfcmp(parts.service, "http")) {
		if (got_bad_number)
			port = 80;
		if ((op & FTP_OP_MASK) != FTP_RECV)
			error("Can't do that with HTTP files.");
		do_http(&parts, port);
	} else if (!strfcmp(parts.service, "https")) {
		if (got_bad_number)
			port = 443;
		if ((op & FTP_OP_MASK) != FTP_RECV)
			error("Can't do that with HTTPS files.");
		do_https(&parts, port);
	} else if (!strfcmp(parts.service, "telnet")) {
		if (got_bad_number)
			port = 23;
		if ((op & FTP_OP_MASK) != FTP_RECV)
			error("Can't do that with telnet buffers.");
		do_telnet(parts.host, port, bufname);
	} else if (!strfcmp(parts.service, "ssh")) {
		if ((op & FTP_OP_MASK) != FTP_RECV)
			error("Can't do that with ssh buffers.");
		do_ssh(parts.host, bufname, parts.usr);
	} else if (!strfcmp(parts.service, "scp")) {
		do_scp(&parts, op);
	} else
		do_ftp_op(bufname, parts.host, parts.port, parts.usr, parts.pwd,
				  parts.fname, op);
	return 0;		// No error possible yet (asynchronous).
}


/*************************/
/*     Telnet stuff      */
/*************************/

telnet_activity(int activity, int buf, int from, int to)
{
	if (!buf_exist(buf))
		return;
	save_var iter, has_arg;
	save_var bufnum = buf;
	if (!(telnet_interpret_output & 16))
		save_var point;
	switch (activity) {
		case NET_DONE:
			if (!telnet_id)		// A ssh buffer.
				break;
			point = type_point;
			stuff("Connection closed.\n");
			refresh();
			telnet_id = 0;
			break;
		case NET_RECV:
			set_character_color(from, to, color_class process_output);
			if (telnet_interpret_output & 0x100)
				perform_unicode_conversion(bufnum, from, size(), CONV_TO_16, "UTF-8-No-BOM");
			if (telnet_interpret_output & 0x20) {
				point = from;
				simple_re_replace(0, "<#0-#1>", "");
			}
			if (telnet_interpret_output & 0x40)
				remove_xterm_sequences(from);
			if (telnet_interpret_output & 4)
				color_ansi_output(from);
			if (telnet_interpret_output & 8) {
				point = from - 2;
				while (re_search(1, ".\b."))
					if (character(point - 3) == character(point - 1)) {
						delete(point - 3, point - 1);
						set_character_color(point - 1, point, color_class man_bold);
					}
			}
			if (telnet_interpret_output & 1) {
				save_var point = from - 2;
				while (re_search(1, "_\b.|o\b%+")) {
					delete(point - 3, point - 1);
					set_character_color(point - 1, point, color_class process_prompt);
				}
			}
			if (telnet_interpret_output & 0x80)
				process_backspace_sequences(from - 100);
			point = to;
			save_var case_fold = 1;
			if (ungot_key == -1 && (recognize_password_prompt & 1)
				  && parse_string(-1, recognize_password_pattern)
				  && point == size()) {
				ungot_key = WIN_HELP_REQUEST;	// Make Epsilon run this cmd:
				menu_command = "password-prompt";
			}
			if ((net_prompt_get_directory & 1) && point == size())
				net_prompt_get_dir();
			if (telnet_interpret_output & 2) {
				point = from;
				to_begin_line();
				while (re_search(1, "^[^\r\n]*\r+![^\n]"))
					delete(matchstart, matchend);
			}
	}
}

// Set this buffer's current directory by examining the prompt from the
// remote system.
void net_prompt_get_dir()
{
	struct url_parts parts;
	if (!buffer_url || !parse_url(buffer_url, &parts))
		return;				// Get the host name.
	int i = find_variable_by_host("net_prompt_get_directory", parts.host);
	int flags = get_num_var(i);
	char *pat, *s = strchr(buffer_url, ':');	// Copy :host@user portion.
	i = find_variable_by_host("net_prompt_pattern", parts.host);
	if (!(pat = get_str_var(i)) || !s || !(flags & 1))
		return;
	int mode = 0;
	if (parse_string(-1, "\x1FEpsilon:cwd:(.*)\x1F"))
		mode = 1;			// Our shell script modified the prompt.
	else if (!(flags & 2) || !parse_string(-1, pat))
		if (!(flags & 4)) {	// Fall back, no cur dir info.
			mode = 2;
			parse_string(-1, "()");	// This will make dir be empty.
		} else
			return;
	int from = find_group(1, 1), to = find_group(1, 0);
	char dir[FNAMELEN], url[FNAMELEN];
	if (to - from < FNAMELEN - strlen(buffer_url) - 5)
		grab(from, to, dir);
	if (mode == 1)
		delete(matchend, matchstart);
	sprintf(url, "%s%s/", (flags & 8) ? "scp" : "ftp", s);
	if (mode != 2) {
		i = find_variable_by_host("net_prompt_get_dir_helper", parts.host);
		(*(int (*)()) i)(mode, url, dir, &parts);
	}
	process_current_directory = url;
	if (flags & 16)
		set_character_color(give_begin_line(), point, color_class process_prompt);
	if (flags & 32) {
		save_var use_process_current_directory = 0;	// Already got it.
		do_when_process_input(bufnum, 0, point);	// Helps to color user input later.
	}
}

// Modify url by adding the path corresponding to dir. Mode: 0=got dir
// from parsing the prompt, 1=got dir from magic Epsilon sequence. This
// version is for Unix-style hosts. A per-host replacement like
// net_prompt_get_dir_helper_example_com() is allowed to completely
// replace the contents of url if needed. It may use the components of
// the buffer's URL found in *parts.

void net_prompt_get_dir_helper(int mode, char *url, char *dir, struct url_parts *parts)
{
	UNREFERENCED_PARAMETER(mode);		// A user-written replacement might use these.
	UNREFERENCED_PARAMETER(parts);
	if (strcmp(dir, "~")) {		// If dir is not home dir, add it.
		if (dir[0] == '~' && dir[1] == '/')
			strcat(url, dir + 2);	// Path is relative to home dir, so 1 /.
		else if (strcmp(dir, "/"))		// Use 2 /s to make abs path, unless root dir.
			strcat(url, dir);
		strcat(url, "/");
	}
}

command send_invisible()
{
	char res[FNAMELEN];

	iter = 0;
	if (!can_send_invisible())
		error("This is not a process or telnet buffer.");
	comp_read(res, "Text to send without inserting into buffer: ", (int (*)()) 0,
			  POP_UP_PROMPT | PASSWORD_PROMPT, "");
	do_send_invisible(res, !has_arg);
}

// Can we send text to this telnet or process buffer without it echoing?
int can_send_invisible()
{
	return telnet_id || (is_process_buffer(bufnum)
						 && (has_feature & FEAT_PROC_SEND_TEXT));
}

// Send the line to the process or via telnet, maybe adding a
// terminator.  Return 0 if error.
int do_send_invisible(char *s, int add_term)
{
	if (telnet_id) {
		telnet_send(telnet_id, s);
		if (add_term)
			telnet_send(telnet_id, "\r");
	} else if (is_process_buffer(bufnum)) {
		process_send_text(bufnum, s, strlen(s));
		if (add_term)
			process_send_text(bufnum, "\n", 1);
	} else
		return 0;
	return 1;
}

int password_prompt()
{
	char msg[FNAMELEN], res[FNAMELEN];
	int scp_buf = scp_find_control(bufnum);

	if (scp_buf)
		save_var bufnum = scp_buf, point = size();
	else if (can_send_invisible())
		save_var point = type_point;
	else
		return;
	grab_full_line(bufnum, msg);
	restore_vars();
	comp_read(res, *msg ? msg : "Password: ", (int (*)()) 0,
			  POP_UP_PROMPT | PASSWORD_PROMPT, "");
	if (scp_buf) {
		process_send_text(scp_buf, res, strlen(res));
		process_send_text(scp_buf, "\n", 1);
	} else
		do_send_invisible(res, 1);
}

void do_telnet(char *host, int port, char *bname)
{
	to_buffer(bname);
	type_point = 0;
	telnet_mode();
	retain_forced_mode = 1;
	discardable_buffer = 1;
	buffer_not_saveable = 1;
	set_read_only(0);
	when_net_activity = telnet_activity;
	telnet_id = telnet_host(host, port, bname);
	char url[FNAMELEN];
	sprintf(url, "telnet://%s", host);
	buffer_url = strkeep(url);
}

command telnet()
{
	char hostname[FNAMELEN], buf[FNAMELEN], *p;
	int port = 23, i = 1;

	inet_only();
	tiled_only();
	if (has_arg){
		port = iter;
		iter = 0;
	}
	*hostname = 0;
	get_str_auto_def(hostname, "Host");
	if (!*hostname)
		return;
	if (!has_arg && (p = strchr(hostname, ':')) && numtoi(p + 1)
		  && !got_bad_number) {
		port = numtoi(p + 1);
		*p = 0;
	}
	sprintf(buf, "%s", hostname);
	if (port != 23)
		sprintf(buf + strlen(buf), ":%d", port);
	p = buf + strlen(buf);
	for (;;) {		// Find an unused buffer.
		if (!exist(buf))
			break;
		bufname = buf;
		if (!telnet_id)
			break;
		sprintf(p, " %d", i++);
	}
	do_telnet(hostname, port, buf);
}

telnet_send_line() on telnet_tab['\n'], telnet_tab['\r']
{
	char telnet_cbuf[200], hist_buf[FNAMELEN];

	if (point < type_point || !telnet_id) {
		enter_key();
		return;
	}
	sprintf(hist_buf, "-process-history-%s", bufname);
	create(hist_buf);
	process_cmds_to_end(hist_buf);
	while (point > type_point) {
		int i = MIN(point - type_point,
					sizeof(telnet_cbuf) / sizeof(*telnet_cbuf) - 1);
		grab(type_point, type_point + i, telnet_cbuf);
		if (telnet_server_echoes(telnet_id))
			delete(type_point, type_point + i);
		else
			type_point += i;
		telnet_send(telnet_id, telnet_cbuf);
		if (i > 0)
			buffer_printf(hist_buf, "%s", telnet_cbuf);
	}
	if (!has_arg) {
		if (!telnet_server_echoes(telnet_id))
			insert('\n'), type_point = point++;
		telnet_send(telnet_id, "\r\n");
	}
	buffer_printf(hist_buf, "\n");
}

telnet_send_interrupt() on telnet_ctrlc_tab[CTRL('c')]
{
	do_send_invisible("\03", (mode_keys == ssh_tab));
}

maybe_restart_remote_process()
{
	if (buffer_url && !telnet_id && !is_process_buffer(bufnum)
		  && is_remote_command_line(buffer_url))
		url_operation(buffer_url, FTP_RECV);
}

int is_remote_command_line(char *cmd)
{
	return (!strncmp(cmd, "telnet://", 9) || !strncmp(cmd, "ssh://", 6));
}

// Find a buffer connected to the given host using the service and
// return its number, or 0.
int find_remote_buffer(char *url)
{
	save_var bufnum;
	int i = buf_list(0, 2);
	do {
		bufnum = i;
		if (buffer_url && !strcmp(url, buffer_url)
			  && (telnet_id || is_process_buffer(bufnum)))
			return i;
	} while (i = buf_list(-1, 1));
	return 0;
}

is_remote_buffer(int buf)
{
	save_var bufnum = buf;
	return (mode_keys == ssh_tab) || telnet_id != 0
			|| *filename && is_remote_file(filename);
}

telnet_send_literal() on telnet_tab[CTRL('o')]
{
	char p[10];
	sprintf(p, "%c", key);
	telnet_send(telnet_id, p);
}

telnet_mode()
{
	mode_default_settings();
	mode_keys = telnet_tab;
	fix_key_table(reg_tab, find_index("cc-tab"),
				  telnet_tab, find_index("telnet-ctrlc-tab"));
	telnet_tab[ALT('p')] = find_index("process-previous-cmd");
	telnet_tab[ALT('n')] = find_index("process-next-cmd");
	major_mode = _telnet_mode_name;
	tab_size = 8;
	try_calling("telnet-mode-hook");
	make_mode();
}

// For sending chars to ssh job, act like a process, but for getting
// chars, act like telnet.
ssh_activity(int activity, int buf, int param1, int param2)
{
	save_var telnet_interpret_output = ssh_interpret_output, abort_searching = 0;
	if (activity == NET_SEND) {
		save_var bufnum = buf;
		color_process_range(0, size());
		do_when_process_input(buf, param1, param2);
	} else
		telnet_activity(activity, buf, param1, param2);
}

// A variation on process mode for ssh buffers.
ssh_mode()
{
	process_mode();
	mode_keys = ssh_tab;
	major_mode = _ssh_mode_name;
	try_calling("ssh-mode-hook");
	make_mode();
}

// Ssh mode doesn't provide completion, and uses telnet's version of
// Ctrl-C Ctrl-C.
when_loading()
{
	copy_key_table(process_tab, ssh_tab);
	ssh_tab['\t'] = 0;
	ssh_tab[NUMSHIFT(GREYTAB)] = 0;
	fix_key_table(reg_tab, find_index("cc-tab"),
				  ssh_tab, find_index("telnet-ctrlc-tab"));
}

// Fill in the hostname for active ssh buffers and return nonzero.
ssh_get_hostname(char *host)
{
	char *p;
	struct url_parts parts;

	if (mode_keys != ssh_tab)
		return 0;
	if (p = strchr(bufname, '@'))
		p++;
	else if (!strncmp(bufname, "ssh-", 4))
		p = bufname + 4;
	else if (*filename && is_remote_file(filename)
			 && parse_url(filename, &parts) && !strcmp(parts.service, "ssh")) {
		p = parts.host;
	} else
		return 0;
	strcpy(host, p);	// Make a copy of host name.
	if (p = strchr(host, ' '))
		*p = 0;
	return 1;
}

// Create a ssh job in buffer bname connected to host.  Host can be
// user@host format too.

void do_ssh(char *host, char *bname, char *usr)
{
	char cmd[FNAMELEN], *args[3], justhost[FNAMELEN], username[FNAMELEN], *s;
	char cmdname[FNAMELEN];
	struct url_parts parts;

	if (clear_process_buffer & 0x20)
		zap(bname);
	to_buffer(bname);
	type_point = size();
	ssh_mode();
	retain_forced_mode = 1;
	discardable_buffer = 1;
	buffer_not_saveable = 1;
	set_read_only(0);
	when_activity = ssh_activity;
	process_echo = is_unix ? 6 : 4;
	memset(&parts, 0, sizeof(parts));
	strcpy(justhost, host);
	strcpy(username, usr);
	parts.have_usr = *username != 0;
	if (s = strchr(host, '@')) {
		strcpy(justhost, s + 1);
		strncpy(username, host, s - host);
		parts.have_usr = 1;
	}
	strcpy(parts.service, "ssh");
	strcpy(parts.host, justhost);
	strcpy(parts.usr, username);
	args[0] = username;
	args[1] = justhost;
	ssh_command_variable(cmdname, "ssh", "ssh-command-%s", &parts, 0);
	args[2] = cmdname;
	if (expand_string_template(cmd, scp_variable(&parts, ""), "uhc", args))
		error("Invalid ssh template.");
	if (concur_shell("", cmd, NULL, bufname, SHELL_NO_CD_MAGIC))
		error("Couldn't start ssh process");
	char url[FNAMELEN];
	sprintf(url, "ssh://%s", justhost);
	if (parts.have_usr)
		sprintf(url, "ssh://%s@%s", username, justhost);
	buffer_url = strkeep(url);
}

command ssh()
{
	char hostname[FNAMELEN], dir[FNAMELEN];
	struct url_parts parts;

	inet_only();
	if (!(has_feature & FEAT_MULT_CONCUR))
		error("SCP functions require Unix or Windows NT/2K/XP");
	tiled_only();
	*hostname = 0;
	if (get_buffer_directory(dir) && is_remote_file(dir)
		  && parse_url(dir, &parts) && strstr(" ssh scp ", parts.service)) {
		if (parts.have_usr)
			sprintf(dir, "%s@%s", parts.usr, parts.host);
		else
			sprintf(dir, "%s", parts.host);
		get_strdef(hostname, "Ssh to (hostname or user@hostname)", dir);
	} else
		get_str_auto_def(hostname, "Ssh to (hostname or user@hostname)");
	if (!*hostname)
		return;
	do_ssh_pick_buffer(hostname, "");
}

do_ssh_pick_buffer(char *hostname, char *usr)
{
	char buf[FNAMELEN], *p;
	int i = 1;

	sprintf(buf, "ssh-%s", hostname);
	p = buf + strlen(buf);
	for (;;) {		// Find an unused buffer.
		if (!exist(buf))
			break;
		bufname = buf;
		if (!is_process_buffer(bufnum))
			break;
		sprintf(p, " %d", i++);
	}
	do_ssh(hostname, buf, usr);
}

// To use the next-error facility in a telnet buffer, you must first
// tell Epsilon how to access a file on the remote host for that telnet
// session, by writing a conversion function.

// The conversion function takes a a telnet host name like xxx.yyy.com
// and a file name like myfile.c and builds the correct url to access
// that file (for example, ftp://joe@xxx.yyy.com/project/myfile.c).

// This is the user-supplied conversion function.  Its job is to take a
// host name, a file name, and possibly a directory name from a telnet
// buffer connected to that host, and build an appropriate url.

// If it doesn't recognize the host name or can't construct the URL for
// any reason, it should return 0.  Otherwise, it should return 1 after
// writing the new URL to its url parameter.  (But if file is NULL, it
// should simply return 1; this tests whether the host name is known to
// this function).

int telnet_error_converter(char *file, char *dir, char *url)
{
	char host[FNAMELEN], funcname[FNAMELEN], mode[FNAMELEN];
	int i;

	UNREFERENCED_PARAMETER(file);
	UNREFERENCED_PARAMETER(dir);
	UNREFERENCED_PARAMETER(url);
	clean_mode(mode);
	sprintf(funcname, "%s-get-hostname", mode);
	if (i = find_index(funcname))
		if (!(*(int (*)())(i))(host))
			return 0;

#if 0		/* Copy this example for your own rules. */
	if (!strcmp(host, "xxx.yyy.com")) {
		if (file) {
			strcpy(url, "ftp://joe@xxx.yyy.com/");
			if (dir)
				strcat(url, dir);
			else		// Assume this default directory, if we don't see another.
				strcat(url, "project/");
			if (url[strlen(url) - 1] != '/')
				strcat(url, "/");
			strcat(url, file);
		}
		return 1;
	}
#endif
	return 0;
}

// Fill in the hostname for active telnet buffers and return nonzero.

telnet_get_hostname(char *host)
{
	char *p;

	if (mode_keys != telnet_tab || !telnet_id)
		return 0;
	strcpy(host, bufname);	// Make a copy of host name.
	for (p = host + strlen(host); p > host && isdigit(p[-1]); )
		p--;		// Remove any trailing <Space>+port number from name.
	if (p > host && p < host + strlen(host) && p[-1] == ' ')
		*p = 0;
	return 1;
}

// If the current buffer is a telnet buffer and we have a conversion
// rule for it, scan this buffer for the errors.

telnet_error_set_to_telnet()
{
	if (!telnet_error_converter(NULL, NULL, NULL))
		return;
	// Create the error spot if none.  Or if it's in some other buffer,
	// set it to the start of this one.
	if (spot_to_buffer(error_spot) != bufnum) {
		set_error_spot();
		*error_spot = 0;
	}
}


/*************************/
/** End of telnet stuff **/
/*************************/


/* Scan part1 looking for the first character that's a member of the
character set cs.  If found, set part2 to the section of part1
following that character, remove the character and everything after
from part1, and return the character.  If no character in cs occurs
in part1, set part2 to "", don't modify part1, and return 0. */

int split_string(char *part1, char *cs, char *part2)
{
	int i, c;

	for (i = 0; part1[i]; i++)
		if (index(cs, part1[i])) {
			c = part1[i];
			part1[i] = 0;
			strcpy(part2, part1 + i + 1);
			return c;
		}
	*part2 = 0;
	return 0;
}

/* Scan part1 looking for the last character that's a member of the
character set cs.  If found, set part2 to the section of part1
following that character, remove the character and everything after
from part1, and return the character.  If no character in cs occurs
in part1, set part2 to "", don't modify part1, and return 0. */

int reverse_split_string(char *part1, char *cs, char *part2)
{
	int i, c;

	for (i = strlen(part1) - 1; i >= 0; i--)
		if (index(cs, part1[i])) {
			c = part1[i];
			part1[i] = 0;
			strcpy(part2, part1 + i + 1);
			return c;
		}
	*part2 = 0;
	return 0;
}

// The URL component s may have % escape sequences.  Translate them in
// place.

url_unescape(char *s)
{
	char *p, *hex = "0123456789abcdef", *d1, *d2;

	if (!s)
		return;
	while ((p = strchr(s, '%')) && p[1] && p[2]
		   && (d1 = strchr(hex, p[1])) && (d2 = strchr(hex, p[2]))) {
		*p = (d1 - hex) * 16 + (d2 - hex);
		strcpy(p + 1, p + 3);
	}
}

/* Parse a "full URL": one that starts with "service://".  Return
nonzero if it's valid.

To apply a default service to a URL, see if is_remote_file() returns
1 (indicating it's a full URL already).  If not, prepend service: or
service:// to the URL. */

int parse_url(char *url, struct url_parts *p)
{
	memset(p, 0, sizeof(*p));
	strcpy(p->service, url);			// Parse service:// part.
	if (!split_string(p->service, ":", p->usr) || !*p->service
		|| is_path_separator(p->usr[0]) != 1
		|| is_path_separator(p->usr[1]) != 1)	// It's not valid.
		return 0;
	strcpy(p->usr, p->usr + 2);		// Split off the service part.
	split_string(p->usr, "/\\", p->fname);		// Remove the file name.

	if (!reverse_split_string(p->usr, "@", p->host)) {
										// Do we have a user part?
		strcpy(p->host, p->usr);		// No, so this must be the
		*p->usr = 0;					// host part.
	} else
		p->have_usr = 1;

	if (split_string(p->usr, ":", p->pwd))	// User part could have a pwd.
		p->have_password = 1;		// Pwd could be specified but empty.
	url_unescape(p->usr);
	url_unescape(p->pwd);
	if (split_string(p->host, ":", p->port))
		if (!*p->port)	// Host part could have a port (must be nonempty).
			return 0;
	return 1; // *p->host != 0;		// (Host name must be nonempty.) No, let caller check that.
}

// Divide a URL into service, host section, and pathname.  For instance,
// scp://bob:pwd@example.com:1022/path/to/file produces a service of
// scp, a host part of bob:pwd@example.com:1022, and a pathname in the
// fname member of path/to/file.  No elements are unescaped.  Return 0
// if we can't.

int divide_url(char *url, struct url_parts *p)
{
	memset(p, 0, sizeof(*p));
	strcpy(p->service, url);			// Parse service:// part.
	if (!split_string(p->service, ":", p->host) || !*p->service
		  || is_path_separator(p->host[0]) != 1
		  || is_path_separator(p->host[1]) != 1)	// It's not valid.
		return 0;
	strcpy(p->host, p->host + 2);		// Split off the service part.
	split_string(p->host, "/\\", p->fname);		// Remove the file name.
	return 1;
}

// Given a full FTP URL, try to remove any . or .. components from the
// file name part.  Also convert \ to /.

standardize_remote_pathname(char *url)
{
	char after[FNAMELEN], *s = url, *t;

	inet_only();
	while (s = index(url, '\\'))
		*s = '/';			// Convert all \'s to /'s.
	s = get_url_file_part(url, 1);
	if (!s[0])
		strcpy(s, "/");
	else if (s[0] == '/' && s[1] == '.' && s[2] == '.') {
		strcpy(after, s);	// If relative path starts with ..,
		s[1] = 0;			// insert the expanded home dir before it.
		remote_dirname_absolute(url);
		strcat(url, after);
	} else if (s[0] == '/' && s[1] == '~' && (t = index(s + 1, '/'))
			   && t[1] == '.' && t[2] == '.') {
		strcpy(after, t);	// If relative path starts with ~<any>/..,
		t[1] = 0;			// replace with the expanded home dir.
		remote_dirname_absolute(url);
		strcat(url, after);
	}
	if (*s)
		drop_dots(s);
}

// Change a relative remote directory name to an absolute one in place.

remote_dirname_absolute(char *s)
{
	char *t1, *t2, cmd[FNAMELEN], *name;
	struct url_parts parts;
	int port;

	if (prepare_url_operation(s, FTP_MISC | FTP_WAIT, &parts))
		return;
	if (!strcmp(parts.service, "scp")) {
		scp_remote_dirname_absolute(s, &parts);
		return;
	}
	noteput("Expanding home directory, please wait...");
	t1 = temp_buf();
	t2 = temp_buf();
	name = get_url_file_part(s, 0);
	if (*parts.fname)
		sprintf(cmd, "CWD %s\nPWD\n", parts.fname);
	else
		sprintf(cmd, "CWD\nPWD\n");
	port = numtoi(parts.port);
	if (got_bad_number)
		port = 21;
	ftp_op(t1, t2, parts.host, port, parts.usr, parts.pwd, cmd,
		   FTP_MISC | FTP_WAIT);
	save_var bufname = t2;
	point = 0;
	if (re_search(1, "^257 \"(.*)\"")) {
		grab(find_group(1, 1), find_group(1, 0), name);
		restore_vars();
		delete_buffer(t1);
		delete_buffer(t2);
		note("");
	} else {
		say("Can't expand pathname %s, log %s is %d",
			s, t2, size());
		buffer_printf("Epsnet-Debug", "Can't expand pathname %s,"
					  " log %s is %d\n", s, t2, size());
	}
}

// Return nonzero if the given ftp or scp URL refers to a directory.

int is_ftp_url_a_dir(char *s)
{
	char cmd[FNAMELEN];
	struct url_parts parts;
	int port, res;

	if (prepare_url_operation(s, FTP_MISC | FTP_WAIT, &parts))
		return 0;
	if (!strcmp(parts.service, "scp"))
		return scp_is_url_a_dir(s, &parts);
	if (ftp_compatible_dirs & 4)
		return 0;
	if (*parts.fname)
		sprintf(cmd, "CWD %s", parts.fname);
	else
		return 1;
	port = numtoi(parts.port);
	if (got_bad_number)
		port = 21;
	char *t1 = temp_buf(), *t2 = temp_buf();
	set_mode_message("[FTP: Resolving link] ");
	ftp_op(t1, t2, parts.host, port, parts.usr, parts.pwd, cmd,
		   FTP_MISC | FTP_WAIT);
	set_mode_message(NULL);
	save_var bufname = t2;
	point = 0;
	res = !re_search(1, "^(550|530) ");
	restore_vars();
	delete_buffer(t1);
	delete_buffer(t2);
	return res;
}


// Return a pointer in this URL to the start of the filename part
// (which follows the / after the host name or port number).
// If separator is nonzero, point to the / itself.
// If not a URL, return a pointer to the start of the file name.

char *get_url_file_part(char *url, int separator)
{
	if (!is_remote_file(url))
		return url;
	while (*url && is_path_separator(*url) != 1)
		url++;			// Move past the service.
	while (*url && is_path_separator(*url) == 1)
		url++;			// Move past the :// part.
	while (*url && is_path_separator(*url) != 1)
		url++;			// The next \ or / should start the file name.
	if (!separator && is_path_separator(*url) == 1)
		url++;
	return url;
}

ftp_misc_operation(char *url, char *cmd)
{
	char *t1;
	struct url_parts parts;

	if (scp_misc_op(url, cmd))
		return;
	t1 = temp_buf();
	on_exit delete_buffer(t1);
	if (!prepare_url_operation(url, FTP_MISC, &parts))
		do_ftp_op(t1, parts.host, parts.port, parts.usr, parts.pwd, cmd,
				  FTP_MISC);
}

// Interpret . and .. in a pathname and remove them, modifying s.
// A .. sequence with nothing before it is simply removed.

drop_dots(char *s)
{
	int i, j, slashes[30], numslashes = 0, t;
	char buf[FNAMELEN], *p;

	while (is_path_separator(*s) == 1)	// Never change /'s at the start.
		s++;
	if (!*s)
		return;
	slashes[numslashes++] = 0;
	for (i = 0; s[i]; i++)		// Locate all / or \ chars.
		if (is_path_separator(s[i]) == 1)	// Each starts a "component".
			slashes[numslashes++] = i;
	slashes[numslashes] = i;
	for (i = j = 0; j < numslashes; i++, j++) {
		slashes[i] = slashes[j];
		t = slashes[j];
		if (is_path_separator(s[t]) == 1)
			t++;
		if (s[t] == '.')	// If it starts with a "." ...
			if (slashes[j + 1] == t + 1)
				i--;						// Found "."
			else if (slashes[j + 1] == t + 2 && s[t + 1] == '.')
				i -= (i > 0) ? 2 : 1;		// Found ".."
	}
	numslashes = i;			// Collect desired path components.
	if (is_path_separator(s[slashes[0]]) == 1)
		slashes[0]++;
	for (p = buf, i = 0; i < numslashes; i++) {
		j = slashes[i];		// Component starts here.
		do {
			*p++ = s[j++];	// Copy this component.
		} while (s[j] && is_path_separator(s[j]) != 1);
	}
	*p = 0;
	strcpy(s, buf);
}

// Return info on the remote file: 1 if it's not a remote file at all, 2
// if we can't parse it, 3 if it uses a service other than ftp or scp, 4
// if it has no file name, 5 if it uses wildcards, 6 if it ends in a
// path separator, 7 if it names a user's home directory using ~, 8 if
// we tested and found it was a directory, 9 if we skipped testing
// because we're just starting up (testing could fail if site asks us to
// prompt for a password), otherwise 0 (i.e. for an ordinary file). 4-8
// mean dired should handle it.

remote_file_type(char *file)
{
	struct url_parts parts;
	char *s;

	save_var case_fold = 1;
	if (!is_remote_file(file))
		return 1;
	if (!parse_url(file, &parts) || !*parts.host)
		return 2;
	if (!strstr(" ftp scp ", parts.service))
		return 3;
	if (!*parts.fname)
		return 4;
	if (is_pattern(parts.fname) & 2)
		return 5;		// Use dired if it has * or ? chars.
	if (is_path_separator(parts.fname[strlen(parts.fname) - 1]))
		return 6;
	s = index(parts.fname, '~');	// a user's home dir
	if (s && !index(s, '/') && !index(s, '\\'))
		return 7;
	if (!last_index)		// During init, reject synchronous requests
		return 9;			// since we're not ready to prompt yet.
	if (is_ftp_url_a_dir(file))
		return 8;
	return 0;
}

// Like is_directory() but works with remote files. Can be slow if we
// have to ask the remote system.
int url_is_a_directory(char *s)
{
	switch (remote_file_type(s)) {
		case 1: return is_directory(s);
		case 4:
		case 6:
		case 7:
		case 8:	return 1;
	}
	return 0;
}

// Take a time in seconds, display in 3:12:34 format.

ftp_format_time(char *res, int seconds)
{
	int hours, minutes;

	minutes = (seconds / 60) % 60;
	hours = minutes / 3600;
	seconds %= 60;
	if (hours)
		sprintf(res, "%d:%02d:%02d", hours, minutes, seconds);
	else
		sprintf(res, "%d:%02d", minutes, seconds);
}

ftp_connection_list(char *msg)
{
	char elapsed_msg[40], remaining_msg[40], *s;
	int elapsed = (time_ms() - ftp_job->time_started) / 1000;

	ftp_format_time(elapsed_msg, elapsed);
	switch (ftp_job->operation) {
		case FTP_RECV:		s = "got"; break;
		case FTP_SEND:		s = "sent"; break;
		default:			sprintf(msg, "FTP job, %s elapsed",
									elapsed_msg);
							return;
	}
	if (ftp_job->file_size) {	// time remaining = bytes to go
		if (ftp_job->so_far)	//    * time so far / bytes so far
			ftp_format_time(remaining_msg,
							muldiv(ftp_job->file_size - ftp_job->so_far,
								   elapsed, ftp_job->so_far));
		else
			strcpy(remaining_msg, "?:?");
		sprintf(msg, "%s %d%%, %s elapsed, %s remains",
				s, muldiv(ftp_job->so_far, 100, ftp_job->file_size),
				elapsed_msg, remaining_msg);
	} else
		sprintf(msg, "%s %d bytes, %s elapsed",
				s, ftp_job->so_far, elapsed_msg);
}

// Build a buffer containing a list of connections.

make_connect_list()
{
	int i;
	char msg[500], con[200], *type;

	save_var bufnum;
	zap(CONNECT_LIST_BUF);
	i = active_callbacks();
	if (i == 1)
		sprintf(con, "There is 1 active connection.");
	else
		sprintf(con, "There are %d active connections.", i);
	buffer_printf(CONNECT_LIST_BUF, "%s\n", con);

	i = buf_list(0, 2);		// First list active FTP/Telnet buffers
	do {
		bufnum = i;
		if (telnet_id)
			type = "Telnet session";
		else if (ftp_job && ftp_job->buf == bufnum) {
			ftp_connection_list(msg);
			type = msg;
		} else
			continue;
		buffer_printf(CONNECT_LIST_BUF, "%s: %s\n", bufname, type);
	} while (i = buf_list(-1, 1));

	i = buf_list(0, 2);		// Now list all log files, active or not.
	do {
		bufnum = i;
		if (strncmp(bufname, "FTP Log", 7))
			continue;
		if (!ftp_job)				// The job finished already.
			sprintf(msg, "Inactive");
		else if (ftp_job->operation == FTP_MISC)	// No assoc. buffer.
			sprintf(msg, "Active");
		else				// It's got an associated buffer.
			sprintf(msg, "for %s", buf_exist(ftp_job->buf)
					? bufnum_to_name(ftp_job->buf) : "deleted buffer");
		buffer_printf(CONNECT_LIST_BUF, "%s: %s\n", bufname, msg);
	} while (i = buf_list(-1, 1));
	bufname = CONNECT_LIST_BUF;
	delete_final_newlines();
	point = 0;
}

command show_connections() on reg_tab[ALT(CTRL('C'))]
{
	char msg[500], *s;

	inet_only();
	tiled_only();
	make_connect_list();
	if (num_of_popups() > 0 && !strcmp(bufname, CONNECT_LIST_BUF))
		return;
	save_var menu_width = 60;
	*msg = 0;
	if (get_choice(name_to_bufnum(CONNECT_LIST_BUF), msg,
			   "Remote Connections", NULL, "Go to", "Close", "")) {
		s = strstr(msg, ": ");
		if (s) {
			*s = 0;
			if (exist(msg))
				to_buffer(msg);
		}
	}
	user_abort = 0;
}
