/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1985, 2020 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// This file compares file dates and issues warnings.

#include "eel.h"
#include "diff.h"

get_file_date()		/* record this file's date for later comparison */
{
	if (check_file(filename, &file_info) != CHECK_FILE)
		file_info.year = 0;
}

// The file has been modified on disk; maybe re-read it without prompting.
maybe_auto_read(struct file_info *now)
{
	int pos;

	// Only auto-read if size on disk is > 90% of current size, or
	// old file is very small, and new file's not huge.
	if (!modified && (now->fsize / 9 >= size() / 10 || size() < 300)
		  && now->fsize < 10000000) {
		pos = point;
		if (!read_file(filename, reading_translation_type()))
			say("%f: Modified file read from disk.", filename);
		else
			auto_read_changed_file = _force_auto_read_changed_file = 0;
		point = pos;
		to_begin_line();
		maybe_add_readonly_once(now);
		return 1;
	}
	return 0;
}

// If the file's become read-only, set buffer to match.

maybe_add_readonly(struct file_info *now)
{
	if (!(now->attr & ATTR_READONLY) || (file_info.attr & ATTR_READONLY))
		return;
	if (modified) {
		char msg[FNAMELEN];
		save_var window_bufnum = bufnum;
		refresh();
		sprintf(msg, "%s is now read-only.  Set buffer read-only too? ",
				get_tail(filename, 0));
		if (!ask_yn("File is read-only", msg, "Yes", "No", 1))
			return;
	}
	do_readonly_warning();
}

// Like above, but also remembers new attribute.

maybe_add_readonly_once(struct file_info *now)
{
	if (readonly_warning & (ROWARN_GREP | ROWARN_REFRESH))
		return;
	maybe_add_readonly(now);
	file_info.attr = now->attr;
}

// Should we skip checking this file because it's on a network or
// removable drive, and the user said to skip such files?
int check_dates_skip_drive()
{
	int i = filename_rules(filename), j = 0;
	if (i & FSYS_LOCAL)
		j = 1;
	else if (i & FSYS_NETWORK)
		j = 2;
	else if (i & FSYS_REMOVABLE)
		j = 4;
	else if (i & FSYS_CDROM)
		j = 8;
	return (file_date_skip_drives & j);
}

 /* Compare this file's time info with version on disk. */
/* Return 0 to indicate ok, 1 to indicate user said to cancel saving
 * the file, or 2 to tell grep that it skipped asking to read a modified
 * file from disk. */
check_dates(save)
{
	struct file_info now, *then = &file_info;
	int ret = 1, temp, orig = bufnum, pos;

	save_var reject_client_connections |= 2;
	if (num_of_popups() > 0)	// Don't pop up over bufed or similar.
		return 0;
	if (file_date_skip_drives && check_dates_skip_drive())
		return 0;
	ignoring_file_change = 0;	// Resume checking on getting focus.
	if (!want_warn || !filename || then->check_type != CHECK_FILE
			|| check_file(filename, &now) != CHECK_FILE)
		return 0;
	update_readonly_warning(&now);
	if ((readonly_warning & ROWARN_REFRESH) && !save
		  && (file_info.attr ^ now.attr) & ATTR_READONLY)
		return 2;

	if (!compare_dates(&now, then)) {
		maybe_add_readonly_once(&now);
		return 0;
	}
	switch_to_buffer(bufnum);
	maybe_add_readonly_once(&now);

	if ((auto_read_changed_file || _force_auto_read_changed_file)
		&& !save && maybe_auto_read(&now))
		return 1;
	if ((readonly_warning & (ROWARN_GREP | ROWARN_REFRESH)) && !save)
		return 2;
	tiled_only();
	switch_to_buffer(bufnum);
	maybe_ding(bell_on_date_warning);
	temp = tmp_buf();
	save_var _view_right = 0, _view_bottom = screen_lines - 30;
	save_var return_raw_buttons = 1;
	while (!user_abort) {
		show_date_warning(temp, orig, then, &now, save);
		bufnum = orig;
		pos = point;
		if (is_gui && want_gui_prompts) {
			int win, owin = window_handle;

			remove_final_view();
			save_var _doing_input = DI_VIEW;
			win = add_popup(0, 0, 75, 15, _view_border, temp);
			go_to_link = 0;
			display_dialog_box("FileDateBox2", "File has changed on disk",
							   win, 0, 0,
							   "Compare", "Read", save ? "Save" : "Ignore");
			text_color = (is_gui && want_gui_prompts) ? color_class text
						 : color_class viewed_text;
			point = 0;
			view_loop(win);
			switch (key_is_button) {
				case 1:	key = 'C'; break;
				case 2:	key = 'R'; break;
				case 3:	key = save ? 'S' : 'I'; break;
			}
			window_handle = owin;
			remove_window(win);
			if (user_abort)
				break;
		} else {
			view_buf(temp, 0);
			if (key == abort_key)
				user_abort = 1;
			if (user_abort)
				break;
			getkey();
		}
		char junk[FNAMELEN];
		if (key == WIN_DRAG_DROP)	// Ignore dropped files here.
			while(drag_drop_result(junk))
				;
		key = toupper(key);
		if (key == 'C') {
			ignoring_file_change = 2;	// Don't check for a while.
			refresh();
			split_window_vertically();
			window_number--;	/* to 1st window */
			find_in_other_buf(filename, get_default_translation_type(filename));
			point = pos;
			to_begin_line();
			set_buffer_filename("");
			error("This buffer has disk version.");
			break;		/* abort out of calling func */
		} else if (key == 'R') {
			refresh();
			if (!read_file(filename, reading_translation_type())) {
				say("%s: Read from disk.", filename);
				decompress_after_reading();
			}
			point = pos;
			to_begin_line();	// Don't check immediately after dialog ends.
			ignoring_file_change = is_gui ? 2 : 0;
			break;
		} else if (key == 'D') {
			save_var compare_to_prior_version_style = 1;
			switch (do_compare_to_prior_version(0)) {
				case 1:
					say("The version on disk is identical.");
					delay(200, 0);
					break;
				case 0:
					say("Displaying changes from disk version to edited version.");
					window_one();
					buf_delete(temp);
					quick_abort();		/* abort out of calling func */
					break;
			}
			break;
		} else if (key == (save ? 'S' : 'I')) {
			if (key == 'I')
				ignoring_file_change = 100;	// Don't check on getting focus.
			else
				ignoring_file_change = 2;	// Don't check immediately after dialog ends.
			refresh();
			ret = 0;
			break;
		} else if (key == '!')
			want_warn = !want_warn;
		else if (key == 'A')
			ask_read_changed_file();
		say("");
	}
	buf_delete(temp);
	return user_abort ? 1 : ret;
}

show_date_warning(temp, orig, then, now, save)
struct file_info *now, *then;
{
	char msg[FNAMELEN], *file;
	int mod, warn, auto;

	save_var bufnum = orig;
	file = filename;
	mod = modified;
	warn = want_warn;
	auto = auto_read_changed_file;
	bufnum = temp;
	buf_zap(temp);
	bprintf("%s\nhas changed since it was read from disk", file);
	bprintf(mod ? "!\n" : ".\nFortunately, you haven't edited it.\n");

	stuff_file_date("\nThe version being edited was dated %s on disk.\n",
		then);
	stuff_file_date("         The copy on disk is dated %s.\n\n", now);
	current_time_and_date(msg);
	bprintf("                         It is now %s.\n\n", msg);

	bprintf("Options:\n");
	bprintf("D) Display a diff listing comparing the two versions.\n");
	bprintf("C) Compare the two versions by reading the disk version ");
	bprintf("in its own buffer.\n");
	bprintf("   You may then use compare-windows or diff to ");
	bprintf("see the changes.\n");
	bprintf("R) Read in the new version now, replacing the ");
	bprintf("version already read in.\n");
	if (save)
		bprintf("S) Save your changes, replacing ");
	else
		bprintf("I) Ignore ");
	bprintf("the version on disk.\n\n");
	bprintf("!) %s this warning for this buffer.  ",
		warn ? "Disable" : "Re-enable");
	bprintf("(You may then choose\n");
	bprintf("   another option.)\n");

	bprintf("A) %s auto-read.  Read changed files without prompting for this buffer,\n"
			"   if it's safe. ", auto ? "Disable" : "Enable");
	bprintf("(You may then choose another option.)\n");

	bprintf("\nYour choice? ");
	point = 0;
}

ask_read_changed_file()
{
	char ques[200];
	int res;

	sprintf(ques, "%s for this buffer. %s auto-reading for new buffers too? ",
			auto_read_changed_file ? "Disabling" : "Enabling",
			auto_read_changed_file ? "Disable" : "Enable");
	if (!(is_gui && want_gui_prompts))
		res = ask_yn("Set default", ques, "Yes", "No", 2) ? 1 : 2;
	else
		res = button_dialog("Set default", ques, "Yes", "No", "Cancel", 2);
	if (!user_abort) {
		if (res == 1 || res == 2)
			auto_read_changed_file = !auto_read_changed_file;
		if (res == 1)
			auto_read_changed_file.default = auto_read_changed_file;
	}
	user_abort = 0;
}

// Do these two file_info structures indicate the file has been modified?
// Return nonzero if modified, zero if unchanged.
compare_dates(a, b)
struct file_info *a, *b;
{
	int diff;

	// Avoid Windows network bug that returns size of zero just after
	// writing.  Check sizes only if both are nonzero.
	if (b->fsize && a->fsize)
		if (a->fsize != b->fsize)
			return 1;
	// We test raw times to avoid an NT problem with daylight saving
	// time.  (Files on some file systems change their timestamps when
	// daylight saving time starts or ends, while files on other file
	// systems do not.)  Also we can easily ignore small time differences
	// which can occur over networks or in various other situations.
	diff = b->raw_file_date - a->raw_file_date;
	if (diff < 0)
		diff = -diff;
	if (a->raw_file_date && diff < file_date_tolerance)
		return 0;	// Raw times are available, and close, so they're equal.

	return a->year != b->year || a->month != b->month
			|| a->day != b->day || a->hour != b->hour || a->minute != b->minute
			|| a->second != b->second;
}

stuff_file_date(fmt, p)	/* bprintf the date in p using fmt's %s */
char *fmt;
struct file_info *p;
{
	char msg[FNAMELEN];

	format_date(msg, p->year, p->month, p->day,
		p->hour, p->minute, p->second);
	bprintf(fmt, msg);
}

// Returns 0=Monday ... 6=Sunday given day 1..31, month 1..12, year.
int day_of_week(int day, int month, int year)
{
	if (month < 3)
		month += 12, year--;
	return ((month * 13 + 3) / 5 + day + year
			+ year / 4 - year / 100 + year / 400) % 7;
}


format_date(msg, year, month, day, hour, minute, second)
char *msg;
{
	char ampm = 'a', args[24][30], *vals[24], *s, *t;
	char *months = "January February March April May June July August "
				   "September October November December ";
	char *weekdays = "Monday Tuesday Wednesday Thursday Friday Saturday Sunday ";
	int i, dow;

	sprintf(args[0], "%d", hour);
	sprintf(args[1], "%02d", hour);
	if (hour >= 12)
		ampm = 'p', hour -= 12;
	sprintf(args[2], "%d", hour ? hour : 12);
	sprintf(args[3], "%02d", hour ? hour : 12);
	sprintf(args[4], "%d", minute);
	sprintf(args[5], "%02d", minute);
	sprintf(args[6], "%d", second);
	sprintf(args[7], "%02d", second);
	sprintf(args[8], "%c", ampm);
	sprintf(args[9], "%c", toupper(ampm));
	sprintf(args[10], "%d", day);
	sprintf(args[11], "%02d", day);
	sprintf(args[12], "%d", month);
	sprintf(args[13], "%02d", month);
	sprintf(args[14], "%d", year);
	sprintf(args[15], "%02d", year % 100);
	for (i = month - 1, s = months; i > 0 && s; i--) {
		s = strchr(s, ' ');
		if (s)
			s++;
	}
	if (month >= 1 && month <= 12 && s && (t = strchr(s, ' ')))
		sprintf(args[16], "%.*s", t - s, s);
	else
		sprintf(args[16], "???");
	strncpy(args[18], args[16], 3);
	for (i = 0; i < strlen(args[16]) + 1; i++)
		args[17][i] = toupper(args[16][i]);
	for (i = 0; i < strlen(args[18]) + 1; i++)
		args[19][i] = toupper(args[18][i]);
	dow = day_of_week(day, month, year);
	for (i = dow, s = weekdays; i > 0 && s; i--) {
		s = strchr(s, ' ');
		if (s)
			s++;
	}
	if (dow >= 0 && dow <= 6 && s && (t = strchr(s, ' ')))
		sprintf(args[20], "%.*s", t - s, s);
	else
		sprintf(args[20], "???");
	strncpy(args[22], args[20], 3);
	for (i = 0; i < strlen(args[20]) + 1; i++)
		args[21][i] = toupper(args[20][i]);
	for (i = 0; i < strlen(args[22]) + 1; i++)
		args[23][i] = toupper(args[22][i]);
	for (i = 0; i < 24; i++)
		vals[i] = args[i];
	if (expand_string_template(msg, date_format, "iIhHnNeEaAdDmMyYfFtTwWsS", vals))
		error("Invalid date-format variable.");
}

/* Write the date in p into s in readable format. */
format_file_date(char *s, struct file_info *p)
{
	format_date(s, p->year, p->month, p->day, p->hour, p->minute, p->second);
}

// Fill in s with current time and date in text form.
current_time_and_date(char *s)
{
	struct time_info time;

	time_and_day(&time);
	format_date(s, time.year, time.month, time.day,
				time.hour, time.minute, time.second);
}

command insert_date() on cc_tab[ALT('d')]
{
	char time[FNAMELEN];

	current_time_and_date(time);
	stuff(time);
}
