/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1985, 2020 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// This file has commands for manipulating files.

#include "eel.h"
#include "proc.h"
#include "colcode.h"
#include "epsnet.h"
#include "filter.h"
#include "info.h"
#include "grep.h"

// Load the filter from filter.txt into the dlg-filter-str variable, if
// we haven't already.
prepare_dlg_filter()
{
	char pats[FNAMELEN], *s;

	save_var bufnum = create("-dlg-filter");
	if (!size()) {
		if (s = lookpath("filter.txt"))
			file_read(s, FILETYPE_AUTO);
		else
			stuff("All Files *.*\n");
		point = 0;
		simple_re_replace(1, "^#.*", "");
		while (search(1, ";")) {
			replace(point - 1, ' ');
			grab(point, give_end_line(), pats);
			to_end_line();
			delete(point, point + 1);
			insert(0);
			stuff(pats);
			insert(0);
		}
		insert(0);
		point = 0;
		simple_re_replace(1, "\n", "");
		int i = new_variable("dlg-filter-str", NT_VAR, TYPE_CARRAY, size() + 1);
		s = varptr(i);
		grab(0, size(), varptr(i));
	}
}

use_common_open_file_dlg(fname, title, flags, save)
char *fname, *title;
int *flags;
{
	char dir[FNAMELEN];
	int ret;

	if (common_open_use_directory & 1) {
		getcd(dir);
		common_open_curdir = dir;
	} else if (common_open_use_directory & 12) {
		if (get_buffer_directory(dir) || (common_open_use_directory & 8))
			common_open_curdir = dir;
	}
	prepare_dlg_filter();
	ret = common_file_dlg(fname, title, flags, save,
						  varptr(find_index("dlg-filter-str")),
						  custom_filter, &filter_index);
	if (ret) {
		absolute(fname);
		if (common_open_use_directory & 2)
			chdir(common_open_curdir);
	}
	return ret;
}

// Should we use the GUI dialog?
use_common_dialog(int var)
{
	if (!(has_feature & FEAT_FILE_DIALOG))
		return 0;
	switch (var) {
		case 0:	return 0;	// Never use.
		case 2:	return 1;	// Always use.
		case 3:	return !run_by_mouse;	// Unless via menu or toolbar.
	}
	return run_by_mouse;	// Use if command was invoked by menu or toolbar.
}

// Should we use the GUI file dialog?
use_common_file_dialog()
{
	return use_common_dialog(want_common_file_dialog);
}

command force_common_file_dialog()
{
	save_var want_common_file_dialog;
	switch (want_common_file_dialog) {
		case 0:		want_common_file_dialog = 2; break;
		case 2:		want_common_file_dialog = 0; break;
		default:	want_common_file_dialog = 3; break;
	}
	do {
		mention("Common dialog: %s ", want_common_file_dialog ? "On" : "Off");
		getkey();
	} while (key == MOUSE_MOVE);
	say("");
	do_topkey();
}

/* Ask for a file name and save buffer to that file if one is given. */
command write_file() on cx_tab[CTRL('W')]
{
	char tmp[FNAMELEN], tmp2[FNAMELEN];
	int flags = OFN_NOREADONLYRETURN | OFN_HIDEREADONLY;

	iter = 0;
	if (use_common_file_dialog()) {
		strcpy(tmp, force_save_as == FSA_NEWFILE ? "" : filename);
		if (!use_common_open_file_dlg(tmp, "Save file as", &flags, 1))
			return 2;
	} else
		get_file_dir(tmp, "Write file: ");
	if (*tmp && url_is_a_directory(tmp) && *filename) {
		make_copy_name(filename, tmp, tmp2);
		strcpy(tmp, tmp2);
	}
	if (*tmp) {
		if (!is_remote_file(tmp) && get_file_read_only(tmp) == 1)
			error("File %s is read-only.", get_tail(tmp, 0));
		switch (warn_existing_file(tmp)) {
			case 1: return 0;
			case -1: save_var file_write_newfile = 1;
		}
		if (has_arg) {
			translation_type = give_line_translate(tmp);
			make_mode();
		}
		set_buffer_filename(tmp);
		already_made_backup = 0;
		modified = 1;
		return do_save_file(1, 0, 1);
	}
	say("Buffer %s not saved, no file name.", bufname);
	return 2;
}

command copy_to_file() on reg_tab[FCTRL(7)]
{
	int flags = OFN_NOREADONLYRETURN | OFN_HIDEREADONLY;
	char tmp[FNAMELEN], tmp2[FNAMELEN], orig[FNAMELEN];

	iter = 0;
	strcpy(orig, filename);
	if (use_common_file_dialog()) {
		strcpy(tmp, force_save_as == FSA_NEWFILE ? "" : filename);
		if (!use_common_open_file_dlg(tmp, "Copy to file", &flags, 1))
			return;
	} else
		get_file(tmp, "Copy to file",
				 force_save_as == FSA_NEWFILE ? "" : filename);
	if (*tmp && url_is_a_directory(tmp) && *filename) {
		make_copy_name(filename, tmp, tmp2);
		strcpy(tmp, tmp2);
	}
	switch (warn_existing_file(tmp)) {
		case 1: return;
		case -1: save_var file_write_newfile = 1;
	}
	filename = tmp;
	save_var already_made_backup = 0, force_save_as = 0;
	save_var synchronous_file_io = 1;	// Don't revert filename until we've reported status.
	save_var modified;
	on_exit filename = orig;
	do_save_file(1, 0, 0);
}

// Maybe warn if file s already exists.  Return 1 to abort, 0 if user
// said to go ahead, or turned off checking, or -1 if file didn't
// exist when we checked.  (If possible, cancel writing if it appears
// later.)
int warn_existing_file(char *s)
{
	char msg[FNAMELEN];

	if (is_remote_file(s))
		inet_only();
	if (!warn_before_overwrite || warn_before_overwrite == 2 && is_remote_file(s))
		return 0;
	if (is_remote_file(s)
		  ? !remote_file_match(s, STARTMATCH | EXACTONLY)
		  : (check_file(s, NULL) != CHECK_FILE))
		return -1;
	sprintf(msg, "Overwrite %r? ", s);
	if (!ask_yn("Confirm overwrite", msg, "Overwrite", "Cancel", 1))
		return 1;
	return 0;
}

command print_setup()
{
	page_setup_dialog();
}

// Printing helper functions for Windows.

#define FORMFEED	"\x0C"		/* a string with a Ctrl-L */

// Return start of printed page following the one that starts at pt.
next_printed_page(pt)
{			// Either go one screenful down, or till next Ctrl-L.
	int full = move_by_screen_lines(pt, text_height());
	save_var point = pt, narrow_end = size() - full;
	return search(1, FORMFEED) ? point : full;
}

next_formfeed()		// How much of this page should we print?
{					// Stop before next Ctrl-L, if any.
	save_var point;
	if (search(1, FORMFEED))	// Go to next Ctrl-L
		return point - 1;	// and return new position.
	return point;		// Or return end of buffer.
}

// Send current page to the printer.
show_printed_page()
{
	save_var narrow_end = size() - next_formfeed();
	redisplay();	// Fill this window.
	print_window(window_handle); // Send this window to the printer.
}

// How many printed pages will this buffer generate?

count_printed_pages()
{
	int i = 0;

	window_start = 0;
	while (window_start < size() - narrow_end && !user_abort) {
		window_start = next_printed_page(window_start);
		i++;
	}
	return i;
}

print_this_page(page)	// Did the user ask to print this page number?
{
	if (print_options.flags & PD_PAGENUMS)
		if (page < print_options.frompage || page > print_options.topage)
			return 0;
	return 1;
}

// Maybe show headings at the top of the page.
print_headings(fname, page, total_pages, date)
char *fname, *date;
{
	char buf[FNAMELEN];
	int scheme = find_index(print_color_scheme);

	if (print_heading & PRHEAD_FNAME) {
		sprintf(buf, "%s", fname);
		print_line(buf, scheme);
	}
	if (print_heading & (PRHEAD_PAGECNT | PRHEAD_DATE)) {
		*buf = 0;
		if (print_heading & PRHEAD_PAGECNT)
			sprintf(buf, "Page %d of %d     ", page, total_pages);
		if (print_heading & PRHEAD_DATE)
			strcat(buf, date);
		print_line(buf, scheme);
	}
	if (print_heading)
		print_line("", scheme);
}

// Cycle the pages of this buffer through the printing window.
// Return # of pages we printed.

send_printed_pages(fname)
char *fname;
{
	int total_pages, page = 1, printed = 0;
	char date[FNAMELEN];
	struct time_info t;

	// Collect info we need for the page header.
	time_and_day(&t);
	format_date(date, t.year, t.month, t.day, t.hour, t.minute, t.second);
	save_var display_column = print_long_lines_wrap ? -1 : 0;
	total_pages = count_printed_pages();

	window_start = point = 0;
	while (point < size() - narrow_end) {
		if (print_this_page(page)) {
			note("Printing page %d of %d...", page, total_pages);
			print_headings(fname, page, total_pages, date);
			show_printed_page();
			if (!print_eject())
				break;
			printed++;
		}
		point = window_start = next_printed_page(window_start);
		page++;
		if (user_abort)
			break;
	}
	return printed;
}

gui_print(jobname, fname)
char *jobname, *fname;
{
	int height = print_options.height;	// Leave room for our heading.
	int win, cnt, scheme;

	if (print_heading)		// any heading takes 2 lines (one blank)
		height -= 2;
	if ((print_heading & (PRHEAD_PAGECNT | PRHEAD_DATE))
		&& (print_heading & PRHEAD_FNAME))
		height--;			// a two-line heading
	if (!start_print_job(jobname))
		error("Printing failed.");
	save_var window_handle;
	win = create_invisible_window(print_options.width, height, bufnum);
	if (print_color_scheme && (scheme = find_index(print_color_scheme))
		&& name_type(scheme) == NT_COLSCHEME)
		window_color_scheme = scheme;
	cnt = send_printed_pages(fname);
	remove_window(win);
	end_print_job();
	say("%d page%s printed.", cnt, cnt == 1 ? "" : "s");
}

command print_buffer() on reg_tab[FALT(9)]
{
	do_print_region(window_bufnum, !is_highlight_on(), has_arg);
}

command print_region() on reg_tab[FSHIFT(9)]
{
	do_print_region(window_bufnum, 0, has_arg);
}

command print_buffer_no_prompt()
{
	do_print_region(window_bufnum, !is_highlight_on(), 1);
}

// Make other changes to the buffer before printing it.
do_print_adjustments()
{
	if (print_doublespaced)
		simple_re_replace(1, "\n", "\n\n");
}

// Print the region, or the whole buffer if "all" is nonzero.
// If silent, try not to display any dialog, just do it.

do_print_region(b, all, silent)
{
	int buf, err, skipped_lines, cnt, cols, width = 80;
	char to[FNAMELEN], title[FNAMELEN], *dest;

	iter = 0;
	dest = is_unix ? print_destination_unix : print_destination;
	if ((has_feature & FEAT_OS_PRINTING) && want_gui_printing) {
		print_options.flags = all ? 0 : PD_SELECTION;
		if (silent)
			print_options.flags |= PD_RETURNDEFAULT;
		if (!select_printer(&print_options))
			return -1;		// User canceled.
		all = !(print_options.flags & PD_SELECTION);
		width = print_options.width;
	} else if (!silent) {
		get_strdef(to, "Print to device (or !command)", dest);
		strcpy(dest, to);
	}
	if (!*filename)
		strcpy(title, bufname);
	else if (print_heading & PRHEAD_PATH) {
		strcpy(title, filename);
		abbreviate_file_name(title, width);
	} else
		strcpy(title, get_tail(filename, 1));
	save_var bufnum = b;
	buf = tmp_buf();
	save_spot point, mark;
	fix_region();		// handle other types of regions
	recolor_buffer_range(0, size());
	skipped_lines = all ? 0 : lines_between(0, point, 0);
	if (all)
		buf_xfer_colors(buf, narrow_position(0), narrow_position(size()));
	else if (region_type() == REGRECT)
		extract_rectangle_colors(buf, 0, 1);
	else
		buf_xfer_colors(buf, point, mark);
	bufnum = buf;
	safe_copy_buffer_variables(buf, b);

	point = 0;
	if (!print_tabs || print_line_numbers)
		tab_convert(0, size(), 0);
	if (print_line_numbers) {
		cnt = lines_between(0, size(), 0) + 1 + skipped_lines;
		for (cols = 0; cnt > 0; cols++)
			cnt /= 10;		// Compute maximum width of line numbers.
		cnt = 1;
		do {
			if (point >= size())
				break;
			bprintf("%*d  ", cols, skipped_lines + cnt++);
			set_character_color(give_begin_line(), point, -1);
		} while (nl_forward());
		point = 0;
	}
	do_print_adjustments();
	if ((has_feature & FEAT_OS_PRINTING) && want_gui_printing) {
		gui_print(title, title);
		err = 0;
	} else
		err = send_to_file_or_cmd(buf, dest, title);
	bufnum = b;
	buf_delete(buf);
	if (!err && !all)
		highlight_off();
	if (!err && !((has_feature & FEAT_OS_PRINTING) && want_gui_printing))
		say("%s printed.", all ? "Buffer" : "Region");
	return err;
}

// Copy buffer b to file, or send to cmd.  Include title if possible.
send_to_file_or_cmd(int b, char *file, char *title)
{
	int err;
	char *thisfile = file, cmdline[FNAMELEN], tempfile[FNAMELEN];
	char tmp[FNAMELEN], *s;

	if (!size()) {
		say("Nothing to print.");
		return -1;
	}
	save_var bufnum = b;
	if (err = new_file_write(thisfile, translation_type, NULL,
							 *file == '!' ? FILE_IO_TEMPFILE : -1, -1)) {
		file_error(err, (*file == '!') ? get_tempfile_name() : thisfile,
				   "write error");
		maybe_ding(bell_on_write_error);
	}
	if (!err && *file == '!') {
		strcpy(tempfile, get_tempfile_name());
		convert_to_8_3_filename(tempfile);
		build_filename(cmdline, file + 1, tempfile);
		if (s = strstr(cmdline, "%t")) {
			strncpy(tmp, cmdline, s - cmdline);
			strcat(tmp, title);
			strcat(tmp, s + 2);
			strcpy(cmdline, tmp);
		}
		err = do_push(cmdline, 0, -1);
		delete_file(tempfile);
	}
	return err;
}

ask_save_buffer()
{
	if (!(is_gui && want_gui_prompts)) {
		int res = ask_yn("Unsaved changes", "Save buffer? ",
					  "Yes", "No", 1);
		check_abort();
		return res;
	}
	switch (button_dialog("Unsaved changes", "This buffer contains "
						  "unsaved changes.  Save first? ",
						 "Save", "Discard", "Cancel", 1)) {
		case 1:	return 1;
		case 2: return 0;
	}
	quick_abort();
}

int reading_translation_type()
{
	if (translation_type & FORCED_TRANS)
		return translation_type;
	return FILETYPE_AUTO;	// Let both encoding and translation be detected.
}

command visit_file() on cx_tab[CTRL('V')]
{
	char tmp[FNAMELEN];

	iter = 0;
	save_var show_mouse_choices = 1;
	get_file(tmp, "Visit file", force_save_as == FSA_NEWFILE ? "" : filename);
	if (!*tmp)
		return;
	if (is_process_buffer(bufnum)
		  && !ask_yn("Confirm replacing buffer contents", "Are you sure "
					 "you want to begin executing the lines of this file? ",
					 "Yes", "No", 2))
		return;
	if (!has_arg && modified) {	/* buffer need to be saved? */
		if (ask_save_buffer())
			do_save_file(1, 1, 1);
	}
	read_file(tmp, reading_translation_type());
	do_file_read_kibitz();
}

/* read the specified file into the current buffer */
read_file(tmp, strip)
char *tmp;
{
	int err;

	err = do_file_read(tmp, strip);
	if (err == 2) {
		zap(bufname);
		modified = 0;
		say("%f: New file.", tmp);
	} else if (err && err != EREADABORT) {
		file_error(err, tmp, "read error");
		maybe_ding(bell_on_read_error);
	} else if (file_info.check_type == CHECK_FILE
				&& (file_info.attr & ATTR_READONLY))
		do_readonly_warning();
	else
		update_readonly_warning(&file_info);
	if (window_bufnum == bufnum)
		switching_to_buffer();
	if (all_read_only)
		set_read_only(4);
	make_mode();
	return err;
}

do_readonly_warning()	// alert user: just read a read-only file
{
	if (readonly_warning & ROWARN_MSG)
		say("%f: File is read-only.", filename);
	if (readonly_warning & ROWARN_BUF_RO) {
		_buf_readonly = 2;	// set from file
		call_on_modify = 1;
	}
	maybe_ding(readonly_warning & ROWARN_BELL);
	make_mode();
}

update_readonly_warning(p)  // if file's no longer readonly, and we made
struct file_info *p;	// buffer readonly because it was, make writeable
{
	if (p->check_type == CHECK_FILE && !(p->attr & ATTR_READONLY)
	    && (_buf_readonly == 2)) {
		_buf_readonly = 0;	// if set from file, reset
		make_mode();
	}
}

command change_file_read_only()
{
	int old;
	char fname[FNAMELEN];

	iter = 0;
	save_var show_mouse_choices = 1;
	get_file(fname, "Change read-only status of file",
			force_save_as == FSA_NEWFILE ? "" : filename);
	old = get_file_read_only(fname);
	if (old == -1 || !set_file_read_only(fname, !old))
			error("Can't change read-only status of %r", fname);
	say("%r is now %s.", fname, !old ? "read-only" : "writable");
	check_dates(0);
}

do_revert_file(int trans)
{
	char pr[FNAMELEN], rel[FNAMELEN];

	save_var point;
	if (!*filename || force_save_as)
		error("No file for this buffer.");
	iter = 0;
	if (modified) {		/* buffer need to be saved? */
		relative(filename, rel);
		sprintf(pr, "Discard changes and revert to %s? ", rel);
		if (!ask_yn("Discard changes", pr, "Discard", "Cancel", 1))
			return 0;
	}
	read_file(filename, trans);
	build_first = 1;
	fix_window_start();
	return 1;
}

command revert_file()
{
	do_revert_file(reading_translation_type());
}

command refresh_files()
{
	TIMER talk_now;
	int i, cnt = 0, res;

	time_begin(&talk_now, 100);		// Wait a second before chatter.
	save_var readonly_warning;
	save_var bufnum, window_bufnum;
	i = buf_list(0, 0);
	do {
		bufnum = i;
		readonly_warning |= ROWARN_REFRESH;
		save_var buf_accessed;	// Don't update due to checking.
		res = check_dates(0);
		if (res)
			cnt++;
		if (res == 2) {
			readonly_warning &= ~ROWARN_REFRESH;
			to_buffer_num(i);
			check_dates(0);
			time_begin(&talk_now, 100);
		}
		check_abort();
		if (*filename && time_done(&talk_now)) {
			note("%r: refreshing.", filename);
			time_begin(&talk_now, 100);
		}
	} while (i = buf_list(1, 1));
	if (!cnt)
		say("No files have been modified on disk since they were read.");
	else if (time_done(&talk_now))
		say("");
}

do_file_read(s, strip)	/* read file, updating file_info */
char *s;
{
	int i;

	highlight_off();
	filename = s;
	unique_file_identifier[0] = 0;
	if (is_remote_file(s))
		return find_remote_file(s, strip);
	i = new_file_read(s, strip, &file_info, -1, -1);
	unique_filename_identifier(filename, unique_file_identifier);
	undo_set_flag(UNDO_READ_FILE);
	_read_aborted = (i == EREADABORT) ? 1 : (i && i != 2) ? 2 : 0;
	file_convert_read(0);
	return i;
}

command insert_file() on cx_tab['i']
{
	char file[FNAMELEN];
	int flags = OFN_HIDEREADONLY;

	save_var show_mouse_choices = 1, file_pattern_rules |= FPAT_PERMIT_NO_URLS;
	if (use_common_file_dialog()) {
		strcpy(file, _region_file);
		if (!use_common_open_file_dlg(file, "Insert file", &flags, 0))
			return;
		save_var file_pattern_wildcards = 0; // []'s etc in file name are literal.
	} else {
		if (!insert_file_remembers_file || !*_region_file)
			get_buffer_directory(_region_file);
		get_file(file, "Insert file", _region_file);
	}
	if (!*file)
		return;
	strcpy(_region_file, file);
	mark = point;			/* put mark before it */
	set_region_type(REGNORM);
	build_first = 1;
	do_insert_file(file, give_line_translate(file));
	iter = 0;
}

do_insert_file(file, strip)	/* insert file into current buffer at pt */
char *file;
{
	int orig = bufnum, buf, err;

	bufnum = buf = tmp_buf();
	err = err_file_read(file, strip);
	if (!err)
		buf_xfer(orig, 0, size());
	bufnum = orig;
	buf_delete(buf);
	return err;
}

quiet_file_read(char *file, int strip)
{
	if (is_remote_file(file)) {
		save_var synchronous_file_io = 1;
		return find_remote_file(file, strip);
	}
	return file_read(file, strip);
}

err_file_read(file, strip)	/* like file_read, but show error */
char *file;
{
	int err;
	if (is_remote_file(file)) {
		save_var synchronous_file_io = 1;
		err = find_remote_file(file, strip);
	} else
		err = file_read(file, strip);
	if (window_bufnum == bufnum)
		switching_to_buffer();
	_read_aborted = (err == EREADABORT);
	if (err && !_read_aborted) {
		file_error(err, file, "read error");
		maybe_ding(bell_on_read_error);
	}
	return err;
}

command write_region() on cx_tab['w']
{
	char file[FNAMELEN];
	int b = -1;
	int flags = OFN_NOREADONLYRETURN | OFN_HIDEREADONLY;

	if (use_common_file_dialog()) {
		strcpy(file, _region_file);
		if (!use_common_open_file_dlg(file, "Write region to file", &flags, 1))
			return;
	} else {
		if (!insert_file_remembers_file || !*_region_file)
			get_buffer_directory(_region_file);
		get_file(file, "Write region to file", _region_file);
	}
	if (!*file)
		return;
	if (warn_existing_file(file) > 0)
		return;
	if (has_arg)
		save_var translation_type = give_line_translate(file);
	strcpy(_region_file, file);
	save_spot point, mark;
	fix_region();		// handle other types of regions
	highlight_off();
	if (region_type() == REGRECT) {
		b = tmp_buf();
		extract_rectangle(b, 0);
		save_var bufnum = b;
		mark = 0;
	}
	if (!write_part(file, translation_type, point, mark))
		note_written(file);
	if (b != -1)
		buf_delete(b);
}

note_written(char *fname)
{
	if (is_remote_file(fname))
		show_text(0, 1, "%f: write request queued.", fname);
	else
		show_text(0, 1, "%f written.", fname);
}

write_part(file, strip, start, end)
char *file;
{
	int orig = bufnum, buf, err;

	buf = tmp_buf();
	buf_xfer(buf, start, end);
	bufnum = buf;
	if (is_remote_file(file)) {
		filename = file;
		err = save_remote_file(file, strip, NULL);
	} else
		err = file_write(file, strip);
	bufnum = orig;
	if (!is_remote_file(file))
		buf_delete(buf);
	if (err) {
		file_error(err, file, "write error");
		maybe_ding(bell_on_write_error);
	}
	return err;
}

command save_file() on cx_tab[CTRL('S')]
{
	tiled_only();
	if (has_arg) {
		translation_type = give_line_translate(filename);
		make_mode();
	}
	return do_save_file(1, 1, 1);
}

do_write_backup()
{
	char othername[FNAMELEN];

	build_filename(othername, backup_name, filename);
	if ((want_backups && !already_made_backup || want_backups == 2)
		  && fnamecmp(filename, othername)) {
		make_backup(filename, othername);
		already_made_backup = 1;
	}
}

prepare_new_file()
{
}

prepare_save_file()
{
	if (GET_LINE_TRANSLATE(translation_type) == FILETYPE_BINARY)
		return;
	int i = get_mode_variable("%s-delete-trailing-spaces");
	if (got_bad_number)
		i = default_delete_trailing_spaces;
	if (i)
		simple_re_replace(1, "[ \t]+$", "");

	i = get_mode_variable("%s-add-final-newline");
	if (got_bad_number)
		i = default_add_final_newline;
	if (i && size() > 0 && character(size() - 1) != '\n') {
		save_var point = size();
		insert('\n');
	}
}

// Ask user to confirm saving current file, using this prompt.  Return
// zero if user said no.

int ask_confirm_save(char *fmt)
{
	char othername[FNAMELEN], msg[FNAMELEN];

	relative(filename, othername);
	if (!(is_gui && want_gui_prompts))
		abbreviate_file_name(othername, screen_cols - strlen(fmt) - 2);
	sprintf(msg, fmt, othername);
	return ask_yn("Confirm save", msg, "Save", "Cancel", 0);
}

/* Returns 0=ok, 1=err, 2=user canceled. Prints status in echo area (except with
 * backup==2 and save-without-prompt). */
int do_save_file(int backup, int checkdate, int getdate)
{
	struct file_info *finfo;
	int err, old_abort = user_abort;

	iter = 0;
	if (!*filename || force_save_as)
		return write_file();
	if (checkdate && check_dates(1)) {
		user_abort = old_abort;	// Ignore aborting from modified-on-disk prompt.
		return 2;
	}
	if ((buffer_flags(0) & BF_UNICODE) && !GET_ENCODING(translation_type)
		  && (has_feature & FEAT_UNICODE)) {
		if (!buffer_has_pattern("<#256-#65535>"))
			save_var translation_type = SET_ENCODING(translation_type, 1);
		else		// Force to raw temporarily.
			pick_write_encoding();	// Have 16-bit chars, get an encoding for them.
		if (!GET_ENCODING(translation_type))
			error("Invalid encoding.");
	}
	if (_read_aborted)
		if (!ask_confirm_save((_read_aborted == 1)
							  ? "Only part of %s was read from disk.  "
							  "Save anyway? "
							  : "Error when reading %s.  Save anyway? "))
			return 2;
	if (is_retrieving_remote_file(bufnum)
		  && !ask_confirm_save("%s is still being downloaded.  "
							   "Save partial file now? "))
			return 2;
	if (backup)
		do_write_backup();
	prepare_save_file();
	finfo = getdate ? &file_info : 0;
	if (err = file_convert_write(filename, translation_type, finfo)) {
		auto_read_changed_file = 0;
		file_info.check_type = 0;	// Prevent checking date.
		file_error(err, filename, "write error");
		maybe_ding(bell_on_write_error);
	} else {
		if (!save_without_prompt || backup != 2)
			note_written(filename);
		undo_set_flag(UNDO_WRITE_FILE);
		if (finfo)
			update_readonly_warning(finfo);
		maybe_delete_autosave();
	}
	return (err != 0);
}

autosave_ok(int buf)
{
	save_var bufnum = buf;
	return want_auto_save && *filename
			&& !(buffer_not_saveable & (NOSAVE_ALL | NOSAVE_AUTOSAVE))
			&& !discardable_buffer && !is_remote_file(filename)
			&& check_file(filename) == CHECK_FILE;
}

delete_all_autosaves()
{
	save_var bufnum;
	if (!want_auto_save || (want_auto_save & 8))
		return;		// No auto-save, or no auto-cleanup, do nothing.
	int i = buf_list(0, 0);
	do {
		bufnum = i;
		maybe_delete_autosave();
	} while (i = buf_list(1, 1));
}

maybe_delete_autosave()
{
	auto_save_failed = 0;
	if (!(want_auto_save & 4) && strcmp(auto_save_name, "%f")
		  && autosave_ok(bufnum)) {
		char name[FNAMELEN];
		build_filename(name, auto_save_name, filename);
		delete_file(name);
		if (check_file(filename) != CHECK_FILE)
			error("Deleted auto-saved file and now original file is gone: %f",
				  filename);
	}
}

do_when_idle_autosave(int cnt)
{
	if (auto_save_idle_seconds < 1 || !want_auto_save)
		return -1;
	if (!cnt)
		time_begin(&auto_save_idle_timer, auto_save_idle_seconds * 100);
	int left = time_remaining(&auto_save_idle_timer);
	if (left > 0)
		return left;
	_auto_save_counter = 0;
	auto_save_buffers();
	time_begin(&auto_save_idle_timer, auto_save_idle_seconds * 100);
	return -1;
}

auto_save_buffers()	/* maybe save buffers */
{
	int i;

	save_var bufnum;
	i = buf_list(0, 0);	/* start with oldest buf */
	do {
		bufnum = i;
		if (modified && autosave_ok(bufnum) && size() <= auto_save_biggest_file)
			auto_save_one();
	} while (i = buf_list(1, 1));
}

auto_save_one()
{
	char name[FNAMELEN], altname[FNAMELEN];
	int err, origfile = !strcmp(auto_save_name, "%f"), from, to;

	if (auto_save_failed)	// It failed last time, so no saving for now.
		return;
	if (!modified_buffer_region(&from, &to, "auto-save"))
		return;			// Buffer hasn't been modified since last auto-save.
	build_filename(name, auto_save_name, filename);
	if (_read_aborted || !autosave_ok(bufnum))		// If we didn't read
		return;						// whole file in, don't save it.
	else if (!origfile) {			// auto-save to a different file
		sprintf(altname, "%s.epstmp", name);
		delete_file(altname);
		rename_file(name, altname);
		err = file_convert_write(name, translation_type, NULL);
		save_var case_fold = (get_foldfile() != 0);
		if (strfcmp(altname, name))
			delete_file(altname);
	} else if (check_dates(1))	// make sure other prog hasn't changed
		return;
	else				// get new time info
		err = file_convert_write(filename, translation_type, &file_info);
	if (err) {
		save_var no_popup_errors = 1; // Don't pop up dialog when autosave fails.
		file_error(err, name, "couldn't autosave");
		maybe_ding(bell_on_autosave_error);
		auto_save_failed = 1;
	} else {
		if (want_auto_save & 2)
			show_text(0, 1, "%f autosaved.", name);
		reset_modified_buffer_region("auto-save");
	}
	if (!origfile)
		modified = 1;
}

// Is this buffer unsaved?
int is_unsaved_buffer()
{
	if (discardable_buffer || (buffer_not_saveable & (NOSAVE_ALL | NOSAVE_WARN)))
		return 0;
	if (modified && *filename && !is_dired_buf())
		if (size() || check_file(filename) == CHECK_FILE)
			return 1;
	if (ftp_job && ftp_job->buf == bufnum && ftp_job->operation == FTP_SEND)
		return 1;
	return 0;
}

unsaved_buffers()	/* return 1 if any unsaved buffers exist */
{
	int i = buf_list(0, 0), ret = 0;

	save_var bufnum;
	do {
		bufnum = i;
		if (save_without_prompt && *filename && !force_save_as)
			do_save_file(2, 1, 1);
		if (is_unsaved_buffer()) {
			ret = 1;
			break;
		}
	} while (i = buf_list(1, 1));
	return ret;
}

maybe_save_all()	/* possibly save all buffers, depending on variable */
{			/* returns 0 normally; 1 means don't do cmd */
	switch (save_when_making) {
		case 0:		return 0;	/* never do it */
		case 1:		break;		/* always do it */
		default:	if (!unsaved_buffers())
						return 0;
					switch (bufed_ask_save()) {
						case 0:	return 1;
						case 1: return 0;
						case 2:	break;
					}
					break;
	}
	save_all_buffers();
	return 0;
}


limit_buffer_size(int max)
{
	if (max > 0) {
		save_var narrow_start = 0, narrow_end = 0;
		save_var point = size() - max;
		to_begin_line();
		delete(0, point);
	}
}

// Move to the end of this buffer; also chop lines from the start so it
// doesn't exceed limit chars (-1=no limit).

buffer_to_end(char *buf, int limit)
{
	if (exist(buf)) {
		save_var bufname = buf;
		limit_buffer_size(limit);
		point = size();
	}
}

// Try to save all buffers (except/only) those that need user to pick a
// name. Update counts of what we did. Add file name of normal saves to
// end of msg if it fits.
do_save_all_buffers(int newfile_bufs, int *count, int *queued, int *failed, char *msg)
{
	char rel[FNAMELEN];
	int i;

	save_var bufnum, window_bufnum;
	i = buf_list(0, 0);
	do {
		bufnum = i;
		if (!*filename || !modified ||
			  (buffer_not_saveable & (NOSAVE_ALL | NOSAVE_SAVEALL))
			  || is_retrieving_remote_file(i))
			continue;
		if (newfile_bufs ? (force_save_as != FSA_NEWFILE)
						 : (force_save_as == FSA_NEWFILE))
			continue;	// Only do one group.
		if (newfile_bufs) {
			if (!size())		// Nothing to save, so don't bother to prompt.
				continue;
			if (save_all_without_asking || _doing_input) {
				buffer_printf(MSG_BUFFER, "save-all-buffers: Could not"
							  " save buffer %s, no file name\n", bufname);
				(*failed)++;
				continue;
			}
		}
		if (newfile_bufs || force_save_as == FSA_READONLY)
			to_buffer_num(i);	// Show the buffer we'll be prompting for.
		switch (do_save_file(1, 1, 1)) {
			case 2:
				buffer_printf(MSG_BUFFER, "save-all-buffers: Cancelled saving buffer %s\n",
							  bufname);
				(*failed)++;
				break;
			case 1:
				buffer_printf(MSG_BUFFER, "save-all-buffers: Could not save %s\n",
							  filename);
				(*failed)++;
				break;
			case 0:
				aside("save-all-buffers: %s %s",
					  is_remote_file(filename) ? "Queued" : "Saved", filename);
				relative(filename, rel);
				if (ptrlen(msg) - strlen(msg) > strlen(rel) + 10) {
					if (*msg)
						strcat(msg, ", ");
					strcat(msg, rel);
				}
				if (!save_without_prompt)
					if (is_remote_file(filename))
						(*queued)++;
					else
						(*count)++;
				break;
		}
	} while (i = buf_list(1, 1));
}

command save_all_buffers() on cx_tab['s']
{
	int count = 0, failed = 0, queued = 0;
	char msg[FNAMELEN];

	*msg = 0;
	save_var no_popup_errors = 1;
	buffer_to_end(MSG_BUFFER, message_history_size);
	do_save_all_buffers(0, &count, &queued, &failed, msg);
	do_save_all_buffers(1, &count, &queued, &failed, msg);
	if (failed) {
		if (in_bufed())
			do_bufed(0);
		count += queued;
		error("%d file%s saved/queued, %d file%s could not be saved, see %s buffer.",
			  count, (count != 1) ? "s" : "",
			  failed, (failed != 1) ? "s" : "", MSG_BUFFER);
	}
	if (count) {
		if (queued)
			say("%d file%s saved, %d file%s queued: %s",
				  count, (count != 1) ? "s" : "",
				  queued, (queued != 1) ? "s" : "", msg);
		else
			say("%d file%s saved: %s", count, (count != 1) ? "s" : "", msg);
		if (in_bufed())
			do_bufed(0);
	} else if (queued)
		say("%d file%s queued for saving: %s", queued, (queued != 1) ? "s" : "", msg);
	else
		say("No files needed saving.");
	return count;
}

set_buffer_filename(f)
char *f;
{
	char abs[FNAMELEN];
	char bname[FNAMELEN+10];

	if (*f) {
		strcpy(abs, f);
		absolute(abs);
		filename = abs;
		do_make_bname(abs, bname, bufname);
		if (!exist(bname))
			change_buffer_name(bname);
		force_save_as = 0;
	} else
		filename = f;
	unique_filename_identifier(filename, unique_file_identifier);
	if (window_bufnum == bufnum)
		switching_to_buffer();
}

command set_file_name()
{
	char tmp[FNAMELEN];

	get_file_dir(tmp, "Enter new output file name: ");
	if (!*tmp || warn_existing_file(tmp) > 0)
		return;
	set_buffer_filename(tmp);
	modified = 1;
}

command set_want_backup_file()
{
	want_backups = has_arg? (iter != 0) : !want_backups;
	say("Epsilon will%s create a backup file for this buffer.",
	    want_backups ? "" : " not");
}

command write_files_and_exit()
{
	if (unsaved_buffers()) {
		save_all_buffers();
		refresh();
	}
	exit();
	iter = 1;
}

// If file is in cygwin format, change it to native format in place.
// If 'from_prompt' is non-zero that means the file name was extracted from a
// command line prompt, in an attempt the track the current directory in
// Windows NT.

convert_cygwin_filename(char *file, int from_prompt)
{
	char buf[FNAMELEN], *home;

	if (opsys != OS_WINDOWS || !cygwin_filenames)
		return;
	if ((cygwin_filenames & 1) && file[0] == '/' && file[1] == '/'
		  && isalpha(file[2]) && file[3] == '/') {
		strcpy(buf, file + 3);		// Rewrite //f/file into f:/file.
		sprintf(file, "%c:%s", file[2], *buf ? buf : "/");
	} else if ((cygwin_filenames & 2) && !strncmp(file, "/cygdrive/", 10)) {
		sprintf(buf, "%c:%s", file[10], file[11] ? file + 11 : "/");
		strcpy(file, buf);
	} else if ((cygwin_filenames & 4) && file[0] == '~'
			   && (!file[1] || file[1] == '/')
			   && (home = getenv("HOME"))) {
		strcpy(buf, file + 1);
		sprintf(file, "%s%s", home, buf);
	} else if ((cygwin_filenames & (from_prompt ? 8 : 16)) && *cygwin_root
			   && file[0] == '/') {
		sprintf(buf, "%s%s", cygwin_root, file);
		strcpy(file, buf);
	} else if ((cygwin_filenames & 32) && (has_feature & FEAT_MULT_CONCUR)
			  && !strchr(file, '\\'))
		replace_file_with_output(file, "cygpath -w %s");
}

// Build a command line by combining cmdpat (must contain %s) with file
// and run that program.  If it returns one line of output, replace
// file with it.

replace_file_with_output(char *file, char *cmdpat)
{
	char cmdline[FNAMELEN];
	int outbuf = tmp_buf();
	on_exit buf_delete(outbuf);
	save_var bufnum = outbuf;
	sprintf(cmdline, cmdpat, file);
	if (pipe_text(bufname, bufname, cmdline, NULL,
				  PIPE_SYNCH | PIPE_CLEAR_BUF | PIPE_NOREFRESH, 0)
	   && lines_between(0, size(), 0) == 1 && character(size() - 1) == '\n')
		grab(0, size() - 1, file);
}

// Maybe display why the file was read as binary.
do_file_read_kibitz()
{
	if (file_read_kibitz && *get_file_read_kibitz())
		say("%f: %s.", filename, get_file_read_kibitz());
}

// Read a file with no character set conversion (even if that's the
// default).
command find_unconverted_file()
{
	save_var default_character_set = 1;
	find_file();
}

command find_file() on cx_tab[CTRL('F')]
{
	char fname[FNAMELEN];
	int want_readonly = 0, flags = 0;

	tiled_only();
	was_quoted = 0;
	save_var show_mouse_choices = 1;
	if (use_common_file_dialog()) {
		*fname = 0;
		if (!use_common_open_file_dlg(fname, "Open file", &flags, 0))
			return;
		if (flags & OFN_READONLY)
			want_readonly = 1;
		save_var file_pattern_wildcards = 0; // []'s etc in file name are literal.
	} else
		get_file_dir(fname, "Find file: ");
	if (!*fname)
		dired_one(fname);
	else if (was_quoted && !is_directory(fname) && !is_remote_file(fname))
		find_it(fname, give_line_translate(fname));
	else
		do_find(fname, give_line_translate(fname));
	do_file_read_kibitz();
	if (want_readonly)
		force_save_as = FSA_READONLY;
	iter = 0;
}

command new_file()
{
	iter = 0;
	tiled_only();
	make_newfile_buf();
}

make_newfile_buf()
{
	char bname[FNAMELEN], fname[FNAMELEN];
	int unique = 2;

	strcpy(bname, "New File");
	for (;;) {
		sprintf(fname, "New%d%s", unique - 1, new_file_ext);
		absolute(fname);
		if (!exist(bname) && !do_file_match(fname, STARTMATCH | EXACTONLY)
			|| unique >= 99)
			break;
		sprintf(bname, "New File %d", unique++);
	}
	to_buffer(bname);
	force_save_as = FSA_NEWFILE;
	filename = fname;
	fundamental_mode();
	try_calling(new_file_mode);
	check_filevars();
}

command find_read_only_file()
{
	find_file();
	if (!is_dired_buf()) {
		_buf_readonly = 1;	// not set from file
		call_on_modify = 1;
		force_save_as = FSA_READONLY;
		make_mode();
	}
}

// Search on line for link, set file to its name.
lst_find_linked_file(char *file)
{
	to_begin_line();
	point += parse_string(1, "[ \t]*");
	if (!parse_string(1, ".*[^ \t\n]![ \t\n]*$", file))
		error("Can't find any file here.");
}

#define URL_PATTERN	"|((ftp|http|scp|telnet|ssh|https):[/\\][/\\][^\n\t >]+)"

default_find_linked_file(char *file)
{
	char *pat;

	save_var case_fold = 1;
	if (opsys == OS_UNIX)
		pat = "(%$({<ident>+}|<ident>+)/)?[+-9A-z~@]+[/.][+-9A-z@]*[--9+A-z]" URL_PATTERN;
	else
		pat = "(%%<ident>+%%[\\/])?([a-zA-Z]:)?[+-9A-z~]+<.>[+-9A-z]*[--9+A-z]" URL_PATTERN;
	point -= parse_string(-1, "[^ \t\n!-*]*");
	if (!re_search(1, pat)) {
		to_begin_line();
		if (!re_search(1, pat))
			error("Can't find any file name on this line.");
	}
	if (strchr("\"'`<", character(matchstart)))
		matchstart++;
	if (strchr(")", character(matchend - 1))
		  && strchr("\"'`.,>", character(matchend - 2)))
		matchend--;
	if (strchr("\"'`.,>", character(matchend - 1)))
		matchend--;
	grab(matchstart, matchend, file);
}

// Look for a file name after point on the current line.

find_linked_file_here(char *file)
{
	char subr[FNAMELEN], curdir[FNAMELEN], mode[FNAMELEN], *s;
	int i = 0;

	save_var narrow_end = size() - give_end_line();
	if (*filename) {
		s = get_extension(filename);
		sprintf(subr, "%s_find_linked_file", *s ? (s + 1) : "none");
		i = find_index(subr);		// Any special method for this file extension?
	}
	if (!i) {
		clean_mode(mode);
		sprintf(subr, "%s_find_linked_file", mode);
		i = find_index(subr);		// Any special method for this buffer?
	}
	switch (name_type(i)) {		// i has name table index of subr/var
		case NT_SUBR: case NT_COMMAND:
		case NT_AUTOSUBR: case NT_AUTOLOAD:
			(*(int (*)())i)(file);		// Cast i to a func ptr, then call it.
			break;
		default:		// Look in file for something like a file name.
			default_find_linked_file(file);
			break;
	}
	get_buffer_directory(curdir);
	if (!is_remote_file(file) && file[0] != ((opsys == OS_UNIX) ? '$' : '%'))
		absolute(file, curdir);
}

// Look for a file name on this line, then view it.

command find_linked_file() on cx_tab[CTRL('L')]
{
	char file[FNAMELEN], curdir[FNAMELEN], subst[FNAMELEN];

	*file = 0;
	if (is_highlight_on()) {
		save_var point, mark;
		fix_region();
		if (mark - point < FNAMELEN - 1 && !lines_between(point, mark, 0))
			grab(point, mark, file);
	}
	if (!*file)
		find_linked_file_here(file);
	get_buffer_directory(curdir);
	if (filename_subst_envvars(file, subst))
		strcpy(file, subst);
	if (!is_remote_file(file))
		absolute(file, curdir);
	quiet_set_bookmark();
	tiled_only();
	locate_window("", file);
	// If it's the name of an existing local file but contains extended
	// file pattern chars, treat them as normal characters for now so
	// we don't get a dired.
	if (!is_remote_file(file) && is_pattern(file) == 1 && check_file(file))
		save_var file_pattern_wildcards = 0;
	do_find(file, give_line_translate(file));
}

// Search for a URL at or after point, return nonzero.
int find_url_here()
{
	int start = point;
	to_begin_line();
	while (re_search(1, "(www%.|(https?|ftp):[/\\][/\\])[^\n\t <>\"]+"
					 "[^\n\t <>()\"',.]|mailto:<word|@|.>+"))
		if (matchend >= start)
			return 1;
	return 0;
}

command view_web_site() on reg_tab[FSHIFT(8)]
{
	char url[2048], def[2048];

	*def = 0;
	html_help_check_ok();
	save_var point, mark;
	if (is_highlight_on()) {
		fix_region();
		if (mark - point < sizeof(def) / sizeof(char) - 1
			  && !lines_between(point, mark, 0))
			grab(point, mark, def);
	}
	if (!*def)
		if (find_url_here())
			if (character(matchstart) == 'w') {	// No http://
				strcpy(def, "http://");
				grab(matchstart, matchend, def + strlen(def));
			} else
				grab(matchstart, matchend, def);
	restore_vars();
	get_strdef(url, "View web site", def);
	if (!*url)
		return;
	if (is_mswindows)
		do_shell_execute(url, "Open", "view web page");
	else
		show_url(url);
}

// Insert the environment and any other subst variables and their
// values into the current buffer in the form VAR=VALUE, one line each.
get_subst_var_list()
{
	char dir[FNAMELEN];
	int buf = name_to_bufnum("-subst-vars"), orig = bufnum;
	if (buf) {
		grab_buffer(buf);
		return;
	}
	save_var bufnum = zap("-subst-vars");
	discardable_buffer = 1;				// Now get all env vars and values, plus shell folders.
	get_substitution_variables((file_interpret_env_vars & 16) ? 3 : 1);
	save_var point = 0;
	while (re_search(1, "^<word|@|(|)>+<^word|@|(|)|=|newline>")) {
		delete(point - 1, point);		// Remove invalid chars from the var names.
		to_begin_line();
	}
	point = 0;
	get_customization_directory(dir);
	bprintf("@ecustom=%s\n", dir);
	save_var show_status = 0;
	if (file_interpret_env_vars & 64)		// Strip vars that don't name a directory.
		do_drop_matching_lines(REGEX, is_mswindows
							   ? "^<^=>+=(\\\\|([a-z]+:)?[\\/])" : "^<^=>+=/", 0);
	if (file_interpret_env_vars & 32) {
		point = size();
		char *s = lookpath("subst_vars");
		if (s)
			do_insert_file(s, FILETYPE_AUTO);
	}
	buf_xfer(orig, 0, size());
}

// Get the value of this subst var from the list and return it, or 0 if
// none.
char *get_other_subst_var(char *var)
{
	char pat[FNAMELEN], value[FNAMELEN];
	int b = tmp_buf();
	on_exit buf_delete(b);
	save_var bufnum = b;
	get_subst_var_list();
	save_var show_status = 0, case_fold = 1, point = 0;
	sprintf(pat, "^%q=", var);
	if (!re_search(1, pat))
		return 0;
	grab(point, give_end_line(), value);
	return strkeep(value);
}

char *filename_subst_var_work_area;

// Return the value of a subst var, but if the result looks like a list
// of directories, make into an extended file pattern and return that.
char *filename_subst_var(char *var)
{
	char *s, *p;

	s = get_other_subst_var(var);
	char sep = (opsys == OS_WINDOWS) ? ';' : ':';
	if (s && (file_interpret_env_vars & 8) && strchr(s, sep)) {
		if (!filename_subst_var_work_area && ptrlen(filename_subst_var_work_area) < strlen(s) + 10)
			filename_subst_var_work_area = realloc(filename_subst_var_work_area, strlen(s) + 10);
		sprintf(filename_subst_var_work_area, "{%s", s);
		s = filename_subst_var_work_area;
		for (p = s; p = strchr(p, sep); p++) {
			if (p > s && is_path_separator(p[-1]) == 1) {
				strcpy(p - 1, p);		// Strip trailing / or \.
				p--;
			}
			*p = ',';			// Change each sep to comma.
		}
		strcat(filename_subst_var_work_area, "}");
		return filename_subst_var_work_area;
	}
	return s;
}

// If the filename src has an environment variable name, substitute
// its value while copying it to dest and return nonzero. Otherwise
// just return zero (using dest as a scratch area).
int filename_subst_envvars(char *src, char *dest)
{
	if (file_interpret_env_vars & 1)
		if (opsys == OS_WINDOWS)
			return filename_subst_envvars_windows(src, dest);
		else if (opsys == OS_UNIX)
			return filename_subst_envvars_unix(src, dest);
	return 0;
}

int filename_subst_envvars_windows(char *src, char *dest)
{
	char *s, *t, var[FNAMELEN], *val;
	int res = 0;
	while (s = strchr(src, '%')) {
		strncpy(dest, src, s - src);
		if (t = strchr(s + 1, '%')) {
			strncpy(var, s + 1, t - s - 1);
			if (val = filename_subst_var(var)) {
				strcat(dest, val);
				src = t + 1;
				res = 1;
			} else {
				strncpy(dest, s, t - s);
				src = t + 1;
			}
		} else {
			strcpy(dest, s);
			src += strlen(src);
		}
		dest = dest + strlen(dest);
	}
	strcpy(dest, src);
	return res;
}

int filename_subst_envvars_unix(char *src, char *dest)
{
	char *s, *t, var[FNAMELEN], *val;
	int res = 0;
	while (s = strchr(src, '$')) {
		strncpy(dest, src, s - src);
		if (s[1] == '{') {
			if (t = strchr(s + 1, '}')) {
				strncpy(var, s + 2, t - s - 2);
				if (val = filename_subst_var(var)) {
					strcat(dest, val);
					src = t + 1;
					res = 1;
				} else {
					strncpy(dest, s, t - s);
					src = t + 1;
				}
			} else {
				strcpy(dest, s);
				src += strlen(src);
			}
		} else {
			for (t = s + 1; *t && (isident(*t) || strchr("@", *t)); )
				t++;
			strncpy(var, s + 1, t - s - 1);
			if (val = filename_subst_var(var)) {
				strcat(dest, val);
				src = t;
				res = 1;
			} else {
				strncpy(dest, s, t - s);
				src = t;
			}
		}
		dest = dest + strlen(dest);
	}
	strcpy(dest, src);
	return res;
}


// Call dired if file has wildcards or is a directory name.
// For remote files, assume it's an FTP directory name if the name
// ends in /, if there's no name at all, or if the ftp spec includes
// wildcards; otherwise test to see if it's a dir, and if so, mark it
// with a / at the end.

do_find(char *file, int strip)
{
	char f[FNAMELEN];

	if (is_remote_file(file))
		switch (remote_file_type(file)) {
			case 4: case 5: case 6: case 7:
				return dired_one(file);
			case 8:
				strcpy(f, file);
				if (!is_path_separator(f[strlen(f) - 1]))
					strcat(f, "/");
				return dired_one(f);
			default:
				return find_it(file, strip);
		}
	if (is_pattern(file) || is_directory(file))
		return dired_one(file);
	else
		return find_it(file, strip);
}

find_it(fname, strip)	/* find named file (must be in absolute form) */
char *fname;
{
	if (!*fname)
		return 2;
	if (look_file(fname)) {
		save_var _force_auto_read_changed_file = 1;
		to_buffer(bufname);
		if (!size() && !file_info.check_type)
			find_in_this_buf(fname, strip);
		else
			maybe_restart_remote_process();
		return 0;
	}
	return find_in_other_buf(fname, strip);
}

find_in_other_buf(fname, transl)
char *fname;
{
	char bname[FNAMELEN+10];

	make_bname(fname, bname);
	to_buffer(bname);
	return find_in_this_buf(fname, transl);
}

find_in_this_buf(char *fname, int transl)
{
	int err;

	if (default_character_set == 2)
		file_io_converter = oem_file_converter;
	else
		file_io_converter = 0;
	err = read_file(fname, transl);
	unique_filename_identifier(fname, unique_file_identifier);
	if (err != EREADABORT) {
		call_mode(filename);
		make_mode();
		check_filevars();
		if (err == 2)		// New file
			prepare_new_file();
	}
	return err;
}

// Get the default translation.  If fname is non-null, maybe use
// alternative variable for rule.
int get_default_translation_type(char *fname)
{
	struct url_parts parts;
	int i;

	if (fname && is_remote_file(fname) && parse_url(fname, &parts)
		  && (i = find_variable_by_host("force_remote_translation_type",
										parts.host)))
		return get_num_var(i);
	return default_translation_type;
}

// Find a file using default translation.
std_find_it(char *fname)
{
	int ret = find_it(fname, get_default_translation_type(fname));
	if (!ret)
		do_file_read_kibitz();
	return ret;
}

// Find a file.  If has_arg, ask for its translation.
int ask_find_it(char *fname)
{
	int ret = find_it(fname, give_line_translate(fname));
	do_file_read_kibitz();
	return ret;
}

// Find a file variable that matches this pattern; return nonzero if we
// did.  This can be on an -*- var: value -*- line at the start, or in a
// Local Variables section at the end.  If oldpat is non-NULL, it says
// what old-style pattern to look for, where just a value is present
// between -*- markers.
int look_for_filevar_pattern(char *var, char *value, char *oldpat)
{
	char fullpat[2000];

	save_var point = 0, case_fold = 1;
	save_var narrow_end = size() - 6000;	// Might have very long lines.
	if (parse_string(1, "#!"))	// Look in first line or two only.
		nl_forward();
	narrow_end = size() - give_end_line();
	point = 0;
	if (search(1, "-*-")) {	// Quick test first.
		point = 0;
		if (oldpat) {
			sprintf(fullpat, ".*-%%*-[ \t]*%s[ \t]*-%%*-", oldpat);
			if (re_search(1, fullpat))
				return 1;
		}
		point = 0;
		sprintf(fullpat, ".*-%%*-.*</word>%s:[ \t]*%s[ \t]*(;.*)?[ \t]*-%%*-",
				var, value);
		if (re_search(1, fullpat))
			return 1;
	}
	restore_vars();
	save_var point = size() - file_variables_search_limit, case_fold = 1;
	sprintf(fullpat, "</word>%s:[ \t]*%s", var, value);
	if (search(1, "Local Variables:") && re_search(1, fullpat))
		return 1;
	return 0;
}

// Try to parse a single element, either a number or a suitable
// variable name (perhaps with - before it), whose value we retrieve
// from buffer in_buf. On success, set *val to
// its value and return 1, else return 0.
int filevar_parse_value(int *val, int in_buf)
{
	char buf[FNAMELEN];
	if (parse_string(1, "-?(0x[0-9a-f]+|0o[0-7]+|ob[01]+|[0-9]+|nil</word>])", buf)) {
		*val = numtoi(buf);
		return 1;
	}
	if (!parse_string(1, "-?[a-z][-a-z0-9]*", buf))
		return 0;
	int i = find_index(buf + (buf[0] == '-'));
	if (!i || vartype_class(i) != 1)
		return 0;
	save_var bufnum = in_buf;
	*val = (buf[0] == '-') ? -get_num_var(i) : get_num_var(i);
	return 1;
}


// Check for file vars in the current buffer, then apply to target
// buffer.
check_buffer_filevars(int target)
{
	char buf[FNAMELEN], var[FNAMELEN], when_setting[FNAMELEN], *s;
	int i, cur = bufnum, val, val2, operator;

	save_var bufnum;
	if (look_for_filevar_pattern("tab-(size|width)", "([0-9]+)", NULL)) {
		grab(find_group(2, 1), find_group(2, 0), buf);
		i = numtoi(buf);
		bufnum = target;
		if (i > 0)
			tab_size = i;
		bufnum = cur;
	}
	if (look_for_filevar_pattern("(margin-right|fill-column)",
								 "([0-9]+)", NULL)) {
		grab(find_group(2, 1), find_group(2, 0), buf);
		i = numtoi(buf);
		bufnum = target;
		if (i > 0)
			margin_right = i;
		bufnum = cur;
	}
	if (look_for_filevar_pattern("(auto-)?fill-mode",
								 "([0-9]+|nil)", NULL)) {
		grab(find_group(2, 1), find_group(2, 0), buf);
		i = numtoi(buf);
		bufnum = target;
		call_maybe_break_line(i);
		bufnum = cur;
	}
	if (look_for_filevar_pattern("(indent-with-tabs|indent-tabs-mode)",
								 "([0-9]+|nil)", NULL)) {
		grab(find_group(2, 1), find_group(2, 0), buf);
		bufnum = target;
		indent_with_tabs = numtoi(buf);
		bufnum = cur;
	}
	if (look_for_filevar_pattern("(write-line-translate)",
								 "([0-9]|dos|windows|binary|unix|mac|auto)", NULL)) {
		grab(find_group(2, 1), find_group(2, 0), buf);
		bufnum = target;
		set_line_translate_from_text(buf);
		bufnum = cur;
	}
	save_var point = size() - file_variables_search_limit, case_fold = 1;
	if (!search(1, "Local Variables:"))
		return;
	while (re_search(1, "</word>([a-z]<hyphen|ident>*):[ \t]*!<hyphen|word>")) {
		grab(find_group(1, 1), find_group(1, 0), var);
		i = find_index(var);
		if (!i || name_type(i) != NT_BUFVAR || !name_user(i) || vartype_class(i) != 1) {
			aside("Ignoring file variable %s.", var);
			continue;	// Must be a buffer-specific numeric user var
		}
		s = name_name(i);
		if (word_in_list(s, "|fill-mode|indent-with-tabs|", 1))
			continue;		// These already received special handling above.
		if (!filevar_parse_value(&val, target)) {
			aside("Ignoring unrecognized value for %s", var);
			continue;
		}
		if (parse_string(1, "<hspace>*([-+*/])<hspace>*")) {
			operator = character(find_group(1, 1));
			point = matchend;
			if (filevar_parse_value(&val2, target))
				switch (operator) {
					case '+':		val += val2; break;
					case '-':		val -= val2; break;
					case '*':		val *= val2; break;
					case '/':		val /= val2; break;
				}
		}
		bufnum = target;
 		set_num_var(i, val);
 		sprintf(when_setting, "when-setting-%s", var);
 		try_calling(when_setting);
		bufnum = cur;
	}
}

// If pat matches here, set *res to the integer at the match in the
// pattern with the group'th set of parens.  (If group is <= 0, set *res
// to -group.)  Return 1 if it matched.
check_one_vi_filevar(char *pat, int group, int *res)
{
	char val[FNAMELEN];
	if (!parse_string(1, pat))
		return 0;
	if (group > 0) {	// Get the number from the match.
		grab(find_group(group, 1), find_group(group, 0), val);
		*res = numtoi(val);
	} else				// Return a fixed number (for boolean settings).
		*res = -group;
	return 1;
}

// Apply any vi-format settings line in the region we've narrowed to.
// We only recognize the common ones that correspond well to Epsilon
// variables.
check_some_vi_filevars()
{
	int res, had_set;
	save_var point = 0;
	while (re_search(1, "([ \t\n](vi|vim)|[ \t]ex):[ \t]*(set? )?")) {
		had_set = parse_string(-1, "set? ");
		do {		// Found a settings line, apply all settings.
			if (check_one_vi_filevar("(textwidth|tw)=([0-9]+)", 2, &res))
				margin_right = res;
			if (check_one_vi_filevar("(tabstop|ts)=([0-9]+)", 2, &res))
				tab_size = res;
			if (check_one_vi_filevar("(shiftwidth|sw|softttabstop|sts)"
									 "=([0-9]+)", 2, &res))
				soft_tab_size = res;
			if (check_one_vi_filevar("no(expandtab|et)", -1, &res)
				  || check_one_vi_filevar("(expandtab|et)", 0, &res))
				indent_with_tabs = res;	// Set to 1 or 0.
			point += res = parse_string(1, "[^ \n:]*[ :]");
			if (had_set && character(point - 1) == ':')
				break;		// A set command ends with next :.
		} while (res > 0);	// To next setting on this line.
	}
}

// Scan first five lines, then last five, for a line with settings in
// the format used by vi/vim/ex.  Vi/vim calls these "modelines".
check_vi_filevars()
{
	save_var point = 0, case_fold = 1;
	save_var narrow_end = size() - 6000;	// Might have very long lines.
	move_by_lines(5);
	narrow_end = size() - point;	// Only look at these.
	check_some_vi_filevars();
	narrow_end = 0;		// Next, check lines at end of buffer.
	point = size();		// Limit scan if lines are very long.
	save_var narrow_start = point - 6000;
	move_by_lines(-6);
	narrow_start = point;
	check_some_vi_filevars();
}

// Look for an .epsilon_vars file with settings for many files and apply
// its settings to the current buffer.  Do both global and per-directory
// files.
apply_vars_files()
{
	char varsfile[FNAMELEN], global[FNAMELEN], dir[FNAMELEN];

	*varsfile = 0;
	if (!stricmp(get_tail(filename, 1), ".epsilon_vars"))
		return;
	if (get_buffer_directory(dir))
		sprintf(varsfile, "%s.epsilon_vars", dir);
	get_customization_directory(global);
	strcat(global, ".epsilon_vars");
	apply_vars_file(global);
	if (*varsfile && !is_remote_file(varsfile) && strcmp(varsfile, global)) {
		if ((use_file_variables & 16) && check_file(varsfile) != CHECK_FILE)
			if (!look_up_tree(varsfile, ".epsilon_vars", dir, NULL))
				return;
		apply_vars_file(varsfile);
	}
}

// Apply the settings from the given .epsilon_vars file (local or
// global).
apply_vars_file(char *varsfile)
{
	char bname[FNAMELEN], pat[FNAMELEN];
	char *file = filename, *mode = major_mode, *s;
	int orig = bufnum;

	if (check_file(varsfile) != CHECK_FILE)
		return;
	save_var bufnum;
	if (!look_file(varsfile)) {	// Load the .epsilon_vars file we found.
		make_bname(varsfile, bname);
		bufnum = create(bname);
		find_in_this_buf(varsfile, get_default_translation_type(varsfile));
	}
	int tmp = tmp_buf();	// Make a buffer with just the matching settings.
	buf_printf(tmp, "Local Variables:\n");
	on_exit buf_delete(tmp);
	save_var bufnum;
	save_var point = 0, case_fold = 0, abort_searching = ABORT_ERROR;
	save_var file_variables_search_limit = 1000000;	// Remove usual limit for an .epsilon_vars file.
	while (re_search(1, "^(Extensions|Modes|Filenames|Basenames): *([^ \n]*)") == 1) {
		grab(find_group(2, 1), find_group(2, 0), pat);
		switch (toupper(character(matchstart))) {
			case 'E':	s = get_extension(file); break;
			case 'M':	s = mode; break;
			case 'F':	s = file; break;
			default:	s = get_tail(file, 1); break;
		}
		if (fpatmatch(s, pat, 0, 0)) {
			int start = give_end_line();
			if (re_search(1, "^(Extensions|Modes|Filenames|Basenames): *"))
				to_begin_line();
			buf_xfer(tmp, start, point);
			buf_printf(tmp, "\n");
		}
	}
	bufnum = tmp;	// Now apply settings from tmp to orig buffer.
	point = 0;
	simple_re_replace(1, "^#.*", "");
	check_buffer_filevars(orig);
	if (!retain_forced_mode)
		check_filevar_mode(orig);
}

// Look for file variables and set our variables appropriately.
check_filevars()
{
	if (use_file_variables & 2)		// Look for general file first.
		apply_vars_files();
	if (use_file_variables & 4)		// Apply vi/vim rules next, before
		check_vi_filevars();		// native ones.
	if (use_file_variables & 1)		// Read here, apply to this buffer.
		check_buffer_filevars(bufnum);
}

// Set buffer b to mode specified by filevar and return 1, or return 0
// if none.
check_filevar_mode(int b)
{
	char buf[FNAMELEN], mode[FNAMELEN];

	if (!look_for_filevar_pattern("mode", "([^:; \t\n]+)", "([^:; \t\n]+)"))
		return 0;
	grab(find_group(1, 1), find_group(1, 0), buf);
	if (!strfcmp(buf, "C++"))
		strcpy(mode, "cpp-mode");
	else
		sprintf(mode, "%s-mode", buf);
	save_var bufnum = b;
	return try_calling(mode);
}

command reset_mode()
{
	retain_forced_mode = 0;
	call_mode(filename);
}

/* Go into appropriate mode based on extension of filename f. */
call_mode(char *f)
{
	if (!retain_forced_mode && !check_filevar_mode(bufnum)) {
		if (!strfcmp(get_extension(f), ".in"))	// Remove .in extension.
			*get_extension(f = strkeep(f)) = 0;
		call_by_suffix(f, "suffix-%s");
	}
}

// Run try_calling() on a subr name derived from a file name, first
// converting any characters invalid in a function name to underscores.
int safe_try_calling(char *func)
{
	char safe[FNAMELEN], *s, *t;
	for (s = func, t = safe; *s; s++, t++)
		*t = (*s < 128 && isident(*s)) ? *s : '_';
	*t = 0;
	return try_calling(safe);
}

call_by_suffix(file, pattern)	// construct a subr name from extension
char *file, *pattern;		// of file name, try to call it
{				// return 1 if we found something to call
	char *s, subr[FNAMELEN], def[FNAMELEN];

	s = get_extension(file);
	sprintf(subr, pattern, *s ? (s + 1) : "none");
	sprintf(def, pattern, "default");
	return safe_try_calling(subr) || safe_try_calling(def);
}

// The current file has no extension, or one we don't recognize. Examine
// it and guess its mode, figuring out a pretend extension that yields
// the right mode. If we can, use pat as a sprintf pattern, put the
// result in res, and return 1. Otherwise return 0.

guess_mode_without_extension(char *res, char *pat)
{
	if (!strfcmp(get_tail(filename), "makefile")
		  || !strnfcmp(get_tail(filename), "makefile.", 9)) {
		sprintf(res, pat, "mak");	// Recognize special file names
		return 1;					// w/o .mak extension.
	}
	save_var point = 0;
	if (parse_string(1, "#%!.*/(env )?perl")) {
		sprintf(res, pat, "perl");
		return 1;
	}
	if (parse_string(1, "#%!.*/(env )?python")) {
		sprintf(res, pat, "python");
		return 1;
	}
	if (parse_string(1, "[ \t\n]*%<%??(%!DOCTYPE[ \t]+html|html|head|%!--)")) {
		sprintf(res, pat, "html");
		return 1;
	}
	if (parse_string(1, "[ \t\n]*%<%??(%!DOCTYPE|xml|xst)")) {
		sprintf(res, pat, "xml");
		return 1;
	}
	if (parse_string(1, "#%!.*/php|(.*\n)?[ \t\n]*%<%??(php)")) {
		sprintf(res, pat, "php");
		return 1;
	}
	if (parse_string(1, "(:.*\n)?#%!.*/(env )?"
					 "[a-z]?(sh|csh|[bd]ash|ksh|tcsh|zsh)</word>|:\n")
		  || !strnfcmp(get_tail(filename), ".profile", 8)) {
		sprintf(res, pat, "sh");
		return 1;
	}
	if (parse_string(1, "#%!.*/(env )?(node|nodejs)")) {
		sprintf(res, pat, "js");
		return 1;
	}
	if (parse_string(1, "%$title ")) {
		sprintf(res, pat, "gms");
		return 1;
	}
	if (parse_string(1, "%%%!PS-Adobe-")) {
		sprintf(res, pat, "ps");
		return 1;
	}
	if (size() > 1000)		// Look in 1st 1000 chars only.
		save_var narrow_end = size() - 1000;
	if (parse_string(1, "[ \t\n]*(//|/%*|#)")
		&& re_search(1, "^[ \t]*(#define|#include)[ \t]+")) {
		// It starts with a comment or preprocessor line, and it's
		// got a #define or #include, so assume it's a C++ header
		// file.
		sprintf(res, pat, "hpp");
		return 1;
	}
	point = 0;
	if (GET_LINE_TRANSLATE(translation_type) == FILETYPE_UNIX
		  && parse_string(1, "([ \t]*\n)*#")
		  || !strfcmp(get_tail(filename), ".epsilon_vars")) {
		sprintf(res, pat, "conf");
		return 1;
	}
	return 0;
}

suffix_none()		// Try to detect the file type from its contents.
{
	char mode[FNAMELEN];

	if (guess_mode_without_extension(mode, "suffix-%s"))
		if (try_calling(mode))
			return;
	suffix_default();
}

suffix_default()
{
	char mode[FNAMELEN];

	if (guess_mode_without_extension(mode, "suffix-%s"))
		if (try_calling(mode))
			return;
	fundamental_mode();
}

build_filename(result, template, file)	/* interpret filename templates */
char *result, *template, *file;	/* copy res to pat, replacing % */
{				/* codes with parts of filename */
	struct url_parts parts;
	char *start;
	int len;

	for (; *result = *template; template++) {
		len = -1;
		if (template[0] == '%' && template[1]) {
			switch (*result = *++template) {
				case 'p':
					start = file;
					if (is_remote_file(file) && divide_url(file, &parts))
						start = parts.fname;
					len = get_tail(start, 0) - start;
					break;
				case 'b':
					start = get_tail(file, 0);
					len = get_extension(start) - start;
					break;
				case 'e':
					start = get_extension(file);
					len = strlen(start);
					break;
				case 'f':
					start = file;
					len = strlen(start);
					break;
				case 'r':
					relative(file, result);
					result += strlen(result) - 1;
					break;
				case 'h':
					*result = 0;
					if (is_remote_file(file) && divide_url(file, &parts))
						strcpy(result, parts.host);
					result += strlen(result) - 1;
					break;
				case 'x':
					get_executable_directory(result);
					result += strlen(result) - 1;
					break;
				case 'X':
					get_executable_directory(result);
					convert_to_8_3_filename(result);
					result += strlen(result) - 1;
					break;
				case 'c':
					get_customization_directory(result);
					result += strlen(result) - 1;
					break;
			}
		}
		if (len >= 0) {
			strncpy(result, start, len);
			result += strlen(result);
		} else
			result++;
	}
}

put_directory(s)		// put directory of current file in s
char *s;
{
	strcpy(s, filename);
	*get_tail(s, 0) = 0;
}

// Look in dir & its parents for a file.
// If found, put the file's name in res & return 1.
// If not found, return 0.
// If dir is null, use the current directory.
// If stop is non-null and non-empty, Epsilon will only look for
// matches within it.  (It must be in absolute form.)

look_up_tree(res, file, dir, stop)
char *res, *file, *dir, *stop;
{
	char subdir[FNAMELEN], stopdir[FNAMELEN];
	int len, last = 1000;

	strcpy(stopdir, stop ? stop : "");
	if (*stopdir == '~')		// Expand home directory for Unix.
		absolute(stopdir);
	if (dir)
		strcpy(subdir, dir);
	else
		getcd(subdir);
	for (;;) {
		if (stopdir && *stopdir && fnamecmp(subdir, stopdir)
			&& !is_in_tree(subdir, stopdir))
			return 0;
		strcpy(res, file);
		absolute(res, subdir);
		if (check_file(res, NULL) == CHECK_FILE)
			return 1;
		if (strchr(file, '\\') || strchr(file, '/') || strchr(file, ':'))
			return 0;
		*get_tail(subdir, 0) = 0;
		len = strlen(subdir);
		if (is_path_separator(subdir[0]) && is_path_separator(subdir[1])
			  && !strpbrk_cnt(subdir, "/\\", 3))
			return 0;				// A UNC path; stop when we hit share name.
		if (is_remote_file(subdir) && !strpbrk_cnt(subdir, "/\\", 2))
			return 0;
		if (len > 3 && is_path_separator(subdir[len - 1]))
			subdir[--len] = 0;
		if (len >= last)
			break;
		last = len;
	}
	return 0;
}

is_in_tree(file, tree)		// is file inside this directory tree?
char *file, *tree;		// assumes both are in absolute form
{
	char rel[FNAMELEN];

	relative(file, rel, tree);
	return fnamecmp(file, rel) != 0;
}

wild_pattern(s)		/* does this filename use the {}[],; wildcards? */
char *s;
{
	return is_pattern(s) & 1;
}

char *low_file_match(char *s, int flags)
{
	char *remote_file_match();

	if (is_remote_file(s))
		return (file_pattern_rules & FPAT_PERMIT_NO_URLS) ? 0
				: remote_file_match(s, flags);
	return file_match(s, flags);
}

char *do_file_match(s, flags)	// hook this function to change the rules
char *s;			// for matching file names
{
	char pat[FNAMELEN * 2];

	if (strlen(s) > FNAMELEN) {
		char *t = malloc(strlen(s) + 1), *r;
		strcpy(t, s);
		drop_quotes(t);
		r = low_file_match(t, flags);
		free(t);
		return r;
	}
	strcpy(pat, s);
	drop_quotes(pat);
	return low_file_match(pat, flags);
}

command list_files()	// list files that match a pattern
{			// with num arg, include directories
	char filepat[FNAMELEN], *s;
	int count = 0;

	iter = 0;
	tiled_only();
	get_file_dir(filepat, "List files matching pattern: ");
	if (!*filepat)
		return;
	zap("file-list");
	to_buffer("file-list");
	noteput("Listing files...");
	s = do_file_match(filepat, STARTMATCH | EXACTONLY
			| (has_arg ? 0 : FM_NO_DIRS));
	for (; s; s = do_file_match(filepat, 0)) {
		if (++count % 10 == 0)
			noteput("Listing file %d... %s", count, s);
		bprintf("%s\n", s);
		check_abort();
	}
	do_sort_region(0, size(), 0, 0);
	point = modified = 0;
	say("%d files found.", count);
}

command locate_file()
{
	char file[FNAMELEN], pat[FNAMELEN], tmp[FNAMELEN];
	int d;

	get_string(file, "Locate files matching the pattern: ");
	if (is_unix) {
		sprintf(pat, "%s%s", locate_path_unix, file);
	} else {
		strcpy(pat, "{");
		for (d = 'C'; d < 'Z'; d++) {
			sprintf(tmp, "%c:\\", d);
			if (filename_rules(tmp) & FSYS_LOCAL)
				sprintf(pat + strlen(pat), "%c,", d);
		}
		if (strlen(pat) < 2)
			error("Can't determine local drive list.");
		pat[strlen(pat) - 1] = '}';
		sprintf(pat + strlen(pat) - 1, "}:\\**\\%s", file);
	}
	sayput("Listing files matching %s...", pat);
	dired_one(pat);
}

int ask_just_line_translate()
{
	int res = -1;

	while (res == -1) {
		sayput("File translation? D=DOS, B=Binary, U=Unix, M=Mac, A=Auto [A] ");
		getkey();
		check_abort();
		if (IS_ENTER_KEY(key))
			res = FILETYPE_AUTO;
		switch (toupper(key)) {
			case 'B':	res = FILETYPE_BINARY; break;
			case 'D':	res = FILETYPE_MSDOS; break;
			case 'U':	res = FILETYPE_UNIX; break;
			case 'M':	res = FILETYPE_MAC; break;
			case 'A':	res = FILETYPE_AUTO; break;
		}
	}
	if (res != FILETYPE_AUTO)
		res |= FORCED_TRANS;
	return res;
}

int ask_line_translate()
{
	return give_line_translate(NULL);
}

// If has_arg, ask for translation and encoding rules and return their
// code.  Otherwise return default ones (based on specified filename, if
// non-null).
int give_line_translate(char *fname)
{
	int res = -1;

	if (!has_arg)
		return get_default_translation_type(fname);
	iter = 1;
	res = ask_just_line_translate();
	if (has_feature & FEAT_UNICODE) {
		int conv = ask_encoding("File uses encoding", 0);
		if (conv >= 0)
			res = MAKE_TRANSLATE(conv, res, (res != FILETYPE_AUTO || conv > 0));
	}
	say("");
	return res;
}

command set_line_translate()
{
	int trans, enc = GET_ENCODING(translation_type);

	iter = 1;
	trans = ask_just_line_translate();
	translation_type = MAKE_TRANSLATE(enc, trans,
									  (trans != FILETYPE_AUTO || enc > 0));
	switch (GET_LINE_TRANSLATE(translation_type)) {
		case FILETYPE_MSDOS:   say("Read/write as Windows/DOS file: ^M^J separates lines");
							   break;
		case FILETYPE_BINARY:  say("Binary file: No changes when reading or writing");
							   break;
		case FILETYPE_UNIX:    say("Read/write as Unix file: ^J separates lines");
							   break;
		case FILETYPE_MAC:     say("Read/write as Macintosh file: ^M separates lines");
							   break;
		case FILETYPE_AUTO:	   say("Automatically detect file type when reading");
							   break;
		default: say("Unknown translation type"); break;
	}
	make_mode();
	iter = 0;
}

set_line_translate_from_text(char *s)
{
	int trans, enc = GET_ENCODING(translation_type);

	if (isdigit(*s))
		trans = numtoi(s);
	else switch (tolower(*s)) {
		case 'd':
		case 'w':	trans = FILETYPE_MSDOS; break;
		case 'b':	trans = FILETYPE_BINARY; break;
		case 'u':	trans = FILETYPE_UNIX; break;
		case 'm':	trans = FILETYPE_MAC; break;
		case 'a':	trans = FILETYPE_AUTO; break;
	}
	translation_type = MAKE_TRANSLATE(enc, trans, 1);
}

// Prompt for name of encoding.  Pass 0 to include auto-detect, 1 to
// skip it.
int ask_encoding(char *pr, int i)
{
	char conv[FNAMELEN], *s;

	zap(_MATCH_BUF);
	save_var recall_id = "encoding";
	for (; s = encoding_to_name(i); i++)
		buffer_printf(_MATCH_BUF, "%s\n", s);
	s = encoding_to_name(GET_ENCODING(translation_type));
	prompt_comp_read(conv, pr, general_matcher, COMP_FOLD, s ? s : "");
	if ((i = encoding_from_name(conv)) >= 0)
		return i;
	say("Unknown encoding %s", conv);
	return -1;
}

command set_encoding()
{
	int e = ask_encoding("Set buffer encoding to", 0);
	int t = GET_LINE_TRANSLATE(translation_type);

	iter = 1;
	if (e < 0)
		return;
	translation_type = MAKE_TRANSLATE(e, t, (e || t != FILETYPE_AUTO));
	make_mode();
}

// Append info to buf describing some characters that require encoding.
list_chars_needing_encoding(int buf)
{
	set_buf_point(buf, buf_size(buf));
	save_var point = 0;
	int cnt = 0;
	while (re_search(1, "<#256-#65535>+")) {
		char charname[FNAMELEN], text[FNAMELEN];
		int i, len, ch = character(matchstart);
		len = matchend - matchstart;
		unicode_character_name(charname, ch);
		buf_printf(buf, "Line %d: U+%04x %s%s\n in \"",
				   lines_between(0, point) + 1, ch, charname,
				   len > 1 ? " and others" : "");
		i = matches_at_length(point - len, -1, "<ascii&!newline>+");
		i = MIN(i, 15);
		grab(point - len - i, point - len, text);
		buf_printf(buf, "%s", text);
		len = MIN(len, 20);
		grab(point - len, point, text);
		buf_printf(buf, "%p%s", color_class grep_match, text);
		i = matches_at_length(point, 1, "<ascii&!newline>+");
		i = MIN(i, 15);
		grab(point, point + i, text);
		buf_printf(buf, "%p%s\"\n", -1, text);
		if (cnt++ > 100) {
			buf_printf(buf, "Remaining Unicode characters omitted.\n");
			return;
		}
	}
}

// Copy lines from this buffer that match the regex into end of buf,
// highlighting matches.  Limit to 1st max characters.
grab_regex_match_lines(int buf, char *pat, char *prefix, char *suffix, int max)
{
	int opoint = buf_size(buf);
	set_buf_point(buf, opoint);
	save_var point = 0;
	if (max > size())
		save_var narrow_end = size() - max;
	while (re_search(1, pat)) {
		buf_printf(buf, "%s", prefix);
		buf_xfer_colors(buf, give_begin_line(), give_end_line());
		buf_printf(buf, "%s\n", suffix);
		if (!nl_forward())
			break;
	}
	save_var bufnum = buf;
	save_var point = opoint;
	while (re_search(1, pat)) {
		set_character_color(matchstart, matchend, color_class grep_match);
	}
}

// Get user to select an encoding for writing this file.
pick_write_encoding()
{
	int tmp = tmp_buf();
	int t = GET_LINE_TRANSLATE(translation_type);
	int e = GET_ENCODING(translation_type);
	int def = encoding_from_name(default_write_encoding);

	if (is_window(window_handle) == ISPOPUP)
		error("Choose an encoding for this buffer first.");
	save_var window_bufnum;
	to_buffer_num(bufnum);
	save_var _view_title = "Choose an encoding";
	buf_printf(tmp, "%s\ncontains characters that cannot be represented\n",
			   bufname);
	buf_printf(tmp, "in a disk file without using an encoding.  Please choose an\n");
	buf_printf(tmp, "encoding for this file.  The default encoding shown or any\n");
	buf_printf(tmp, "of the UTF- encodings will preserve all the characters.\n");
	buf_printf(tmp, "See \"Using National Characters\" in the online manual.\n\n");
	buf_printf(tmp, "Here are some of the characters that can't be represented\n");
	buf_printf(tmp, "without an encoding:\n\n");
	list_chars_needing_encoding(tmp);
	if (def < 0)
		def = 2;		// Default to UTF-8-No-BOM
	save_var translation_type = MAKE_TRANSLATE(def, t, 1);
	char *enc = encoding_to_name(GET_ENCODING(translation_type));
	say("Set buffer encoding to: %s", enc ? enc : "");
	view_buf(tmp, 0);
	e = ask_encoding("Set buffer encoding to", 1);
	iter = 1;
	buf_delete(tmp);
	restore_vars();
	if (e < 1)
		e = def;
	translation_type = MAKE_TRANSLATE(e, t, (e || t != FILETYPE_AUTO));
	make_mode();
}

get_foldfile()
{
	if (preserve_filename_case & 16)
		return 0;
	else if (preserve_filename_case & 32)
		return COMP_FOLD;
	if (opsys == OS_UNIX && is_unix_flavor != IS_UNIX_MACOS && is_unix_flavor != IS_UNIX_CYGWIN)
		return 0;
	return COMP_FOLD;
}

command cd() on reg_tab[FKEY(7)]	/* change current directory */
{
	char resp[FNAMELEN], dir[FNAMELEN];

	save_var recall_id = "changedir";
	if (get_buffer_directory(dir))
		save_completion(dir, 0);
	getcd(dir);
	if (*dir && is_path_separator(dir[strlen(dir) - 1]) != 1)
		sprintf(dir + strlen(dir), "%c", path_sep);
	save_var show_mouse_choices = 1, insert_default_response = 3;
	save_var completion_lister = list_file_matches;
	save_var abort_file_matching = ABORT_JUMP;
	prompt_comp_read(resp, "Change to directory", f_match,
					 COMP_FILE | FOLDFILE | FM_ONLY_DIRS, dir);
	if (*resp && chdir(resp)) {
		if (errno == 2)
			error("%s: no such directory", resp);
		else
			file_error(errno, resp, "unknown error");
	} else {
		getcd(dir);
		say("Current directory: %f", dir);
	}
}

// Get default directory for this buffer and return nonzero.
// If none, set dir to current directory and return zero.

get_buffer_directory(dir)
char *dir;
{
	int ok = 1;

	*dir = 0;
	if (is_dired_buf())		/* get directory name from dired */
		dired_getdir(dir);
	else if (is_process_buffer(bufnum))	// process buffer?
		strcpy(dir, process_current_directory);
	else if (!strcmp(bufname, INFO_BUF))
		ok = 0;
	else {
		strcpy(dir, filename);
		if (!*filename || (force_save_as & FSA_NEWFILE)) {
			ok = 0;			// No file name, so use Epsilon's cur dir
			*dir = 0;		// but say we couldn't find anything specific.
		}
		absolute(dir);
		if (ok)			/* want only directory part */
			*get_tail(dir, 1) = 0;
	}
	if (!*dir) {
		ok = 0;
		getcd(dir);
	}
	if (*dir) {
		dir += strlen(dir);	/* ptr to end of string */
		if (!is_path_separator(dir[-1])) {
			*dir++ = path_sep; /* make sure it ends with slash */
			*dir = 0;
		}
	}
	return ok;
}

// Called whenever a window starts displaying a different buffer
// or the user switches windows.

switching_to_buffer()
{		// put "timestamp" on this buffer showing last access
	hide_auto_show_delimiter();
	buf_accessed = ++buf_accessed_clock;
}
