/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1996, 2012 Lugaru Software Ltd.  All rights reserved.  *
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

#include "eel.h"
#include "proc.h"
#include "gams.h"
#include "colcode.h"

// GAMS is a language for mathematical programming.  See www.gams.com.

// These keywords are followed by identifiers and stuff we parse specially.
#define GAMS_DATA_KEYWORDS "set|sets|scalar|scalars|parameter|parameters|table" \
						   "|variable|variables|equation|equations|model|models" \
						   "|file|files|acronym|acronyms"

// These keywords (plus the ones above) start units.  Semicolon is
// optional before them.
#define GAMS_UNIT_STARTERS "option|options|loop|while|alias|abort|solve|display"

// Other keywords.
#define GAMS_KEYWORDS "card|ord|inf|na|eps|yes|no|smax|smin|not|and|or|sum|prod" \
					  "|exp|all|assign|diag|else|eq|for|ge|gt|if|le|lt|maximizing" \
					  "|minimizing|ne|putpage|puttl|repeat|sameas|system|then" \
					  "|until|using|xor"

// These can come before a VARIABLE declaration.
#define GAMS_TYPE_KEYWORDS "binary|negative|positive|integer|free|semicont" \
						   "|semiint|sos1|sos2|nonnegative|nonpositive"

// Identifiers that aren't really GAMS keywords but should be
// highlighted anyway.
#define GAMS_SPECIALS	"abs|arctan|ceil|cos|errorf|exp|floor|include|log" \
						"|log10|mapval|max|min|mod|normal|power|put|putclose" \
						"|round|sign|sin|sqr|sqrt|trunc|undf|uniform"

// A GAMS statement ends with a semicolon, or with a line break
// followed by a keyword.
#define GAMS_STATEMENT_END	"(;|[ \t]*\n([ \t\n]|[*$].*\n)*(" \
							GAMS_DATA_KEYWORDS "|" GAMS_UNIT_STARTERS "|" \
							GAMS_TYPE_KEYWORDS ")</word|lparen>)"

// A list of possible suffixes.  For speed we color any word
// using the syntax of a suffix, even if it's not in this list.
#define GAMS_SUFFIXES	"date|ifile|ofile|page|rdate|rfile|rtime|sfile|time" \
						"|title|ts|tl|te|l|lo|m|prior|scale|up|lj|nj|sj|tj" \
						"|lw|nw|sw|tw|nd|nr|nz|cc|hdcc|tlcc|cr|hdcr|tlcr|ll" \
						"|hdll|tlll|lp|ws|pc|ps|pw|tm|bm|case|bratio|domlim" \
						"|domusd|holdfixed|iterlim|iterusd|limcol|limrow" \
						"|modelstat|numequ|numinfes|numnopt|numnz|numunbnd" \
						"|numvar|optca|optcr|optfile|reslim|resusd|scaleopt" \
						"|solprint|solveopt|solvestat|sysout|workspace"

// A quoted string starts at position p.  Color it and move past.
gams_parse_and_color_string(int p)
{
	char pat[50];
	int c = character(p);

	point = p + 1;
	sprintf(pat, "[%c\n]", c);
	re_search(1, pat);
	set_character_color(p, point, color_class gams_string);
}

// Return the comment character in use for the given buffer position.
// Inline may be nonzero for inline comment chars (else one-line
// comments).  Return 0 if not set, 2 if set to {{ }}, else char code.
// See gams_have_comment() for supported delimiters.
int gams_get_comment_char(int p, int inline)
{
	char *regname = inline ? "gams-inlinecom" : "gams-eolcom";
	int pos, i = get_tagged_region(regname, p, &pos);
	if (i >= 0)
		return i;
	save_var point = pos;	// Start at unknown section.
	to_begin_line();		// Assume it's the same as prev section, then scan it.
	int prev = (point > 0) ? get_tagged_region(regname, point - 1) : 0;
	set_tagged_region(regname, point, size(), prev);
	while (re_search(1, inline ? "^[$].*inlinecom +[</{]" : "^[$].*eolcom +[!/#?_]")) {
		if (curchar() == '{' && character(point - 1) == '{')	// Use special code for {{ }}.
			set_tagged_region(regname, point, size(), 2);
		else
			set_tagged_region(regname, point, size(), character(point - 1));
	}
	return get_tagged_region(regname, p);
}

// Color a comment block.  Also maybe color MPSGE code inside it.
gams_color_comment(from, to)
{
	int t, c;

	set_character_color(from, to, color_class gams_comment);
	save_var point = to;
	point -= parse_string(-1, "%$offtext.*");
	save_var narrow_end = size() - point;
	point = from;
	point += parse_string(1, "%$ontext.*\n");
	save_var narrow_start = point;
	if (!parse_string(1, "%$model:"))
		return;
	set_character_color(point, size() - narrow_end, -1);
	while (point < to) {
		if (!re_search(1, "^[*$]|[!#?\"'{]|/![*/]|%<%<|_!%$"))
			return;
		t = matchstart;
		c = character(point - 1);
		if (c == '!') {
			nl_forward();
			set_character_color(t, point, color_class gams_string);
		} else if (c == '$' && parse_string(1, "[a-z]+:")) {
			set_character_color(t, matchend, color_class gams_keyword);
			nl_forward();
		} else
		gams_color_common(c, t);
	}
}

// Besides the standard GAMS comment syntax of starting lines with *,
// you can use directives eolcom and inlinecom in a GAMS file to define
// additional comment characters.  This mode only recognizes particular
// characters.  The eolcom (comment till end of line) characters may be
// #, !, //, ??, or the sequence _$.  The inlinecom (comment till
// closing delimiter) pairs may be /* */ as in C, << >>, { }, or {{ }}.

// Is p at the start of a gams comment?  Return 1 if line comment, 2 if
// block comment, else 0.

int gams_have_comment(int p)
{
	switch (character(p)) {
		case '!':
			if (gams_get_comment_char(point, 0) == '!')
				return 1;
			break;
		case '<':
			if (gams_get_comment_char(point, 1) == '<' && character(p + 1) == '<')
				return 2;
			break;
		case '/':
			if (character(p + 1) == '*' && gams_get_comment_char(point, 1) == '/')
				return 2;
			if (character(p + 1) == '/' && gams_get_comment_char(point, 0) == '/')
				return 1;
			break;
		case '{':
			if (character(p + 1) == '{' && gams_get_comment_char(point, 1) == 2)
				return 2;			// Check for special code 2 for {{ }}.
			if (gams_get_comment_char(point, 1) == '{')
				return 2;
			break;
		case '#':
			if (gams_get_comment_char(point, 0) == '#')
				return 1;
			break;
		case '?':
			if (gams_get_comment_char(point, 0) == '?')
				return 1;
			break;
		case '*':
			if (matches_at(p, 1, "^%*"))
				return 1;
			break;
		case '_':			// Found _$, possible comment.
			if (gams_get_comment_char(point, 0) == '_' && character(p + 1) == '$')
				return 1;
			break;
		case '$':
			if (matches_at(p, 1, "^%$ontext"))
				return 1;
	}
	return 0;
}

// Move to the end of a known comment that starts with position p.
gams_move_past_comment(int p)
{
	switch (character(p)) {
		case '!':
		case '#':
		case '*':
		case '?':			// Found ??.
		case '_':			// Found _$.
			nl_forward();
			break;
		case '<':
			search(1, ">>");
			break;
		case '/':
			if (character(p + 1) == '/')
				nl_forward();
			else
				search(1, "*/");
			break;
		case '{':
			search(1, character(p + 1) == '{' ? "}}" : "}");
			break;
		case '$':
			search(1, "$offtext");
			break;
	}
}

// Move past and color comments and whitespace.
gams_move_past_spaces_and_comments()
{
	int t, orig;

	do {
		orig = point;
		point += parse_string(1, "[ \t\n]*");
		t = point;
		if (gams_have_comment(point)) {
			gams_move_past_comment(point);
			set_character_color(t, point, color_class gams_comment);
		}
	} while (point > orig);
}

// We're at the start of an identifier.  Move past it.  We include any
// suffix, as in var.lo, and any indexes, as in var.lo(23, 4).
move_past_gams_identifier()
{
	int p;

	point += parse_string(1, "[a-z0-9_]+");
	p = point;
	point += parse_string(1, "%.[a-z0-9]+");
	set_character_color(p, point, color_class gams_suffix);
	if (curchar() == '(')
		while (re_search(1, "[)\"']"))
			switch (character(point - 1)) {
				case ')':	return;
				case '"':	search(1, "\""); break;
				case '\'':	search(1, "'"); break;
			}
}

color_gams_control(int from, int to)
{
	set_character_color(from, to, color_class gams_control);
	save_var point = from;
	save_var narrow_start = from, narrow_end = size() - to;
	if (re_search(1, "[\"']"))
		gams_parse_and_color_string(point - 1);
}

// At position p we found something we might want to color, identified
// by the character c (usually its first character).  Color and move
// past.  Ignore it if we don't want to color it.  Return 0 if we don't
// recognize anything we handle.

int gams_color_common(int c, int p)
{
	switch (c) {
		case '"':		// found a string
		case '\'':
			gams_parse_and_color_string(p);
			return 1;
		case '*':	// Found possible comment.
		case '#':
		case '!':
		case '{':
		case '?':		// Found ??.
		case '<':		// Found <<.
		case '/':		// Found /* or //
		case '_':		// Found _$
			if (gams_have_comment(p)) {
				gams_move_past_comment(p);
				set_character_color(p, point, color_class gams_comment);
				return 1;
			}
			break;
		case '$':			// found control statement
			if (parse_string(1, "ontext")) {
				search(1, "\n$offtext");
				gams_color_comment(p, point);
				return 1;
			}
			nl_forward();
			color_gams_control(p, point);
			return 1;
	}
	return 0;
}

// Recolor some data in the middle of a data block.  Return 1 if data
// ended before the block did (must recolor more if so), else 0.
int gams_recolor_data_block(int sdata, int edata, int sblock, int eblock)
{
	save_var narrow_end = size() - eblock;
	save_var point = edata;
	nl_forward();
	narrow_end = size() - point;
	point = sdata;
	save_var narrow_start = sblock;
	nl_reverse();
	to_begin_line();
	narrow_start = 0;
	gams_color_data_list();
	return point < size() - narrow_end;
}

color_gams_range(from, to) // recolor just this section
{			// last colored region may go past to
	int t = -1, talk = 0, c, talk_at = from + 40000, p, q, sdata, edata, data = 0;

	if (from >= to)
		return to;
	if (modified_buffer_region(&p, &q, "gams-regions")) {
		// Buffer's changed, so forget saved comment-char info after
		// this point.
		set_tagged_region("gams-inlinecom", p, size(), -1);
		set_tagged_region("gams-eolcom", p, size(), -1);
		if (get_tagged_region("color-gams-data", from, &sdata, &edata) > 0
			  && sdata < p && q < edata) {	// Modification within data block.
			if (!gams_recolor_data_block(p, q, sdata, edata)) {
				reset_modified_buffer_region("gams-regions");
				return to;
			}
		}
		set_tagged_region("color-gams-data", p, size(), -1);
		set_tagged_region("needs-color", to, size(), -1);
	}
	reset_modified_buffer_region("gams-regions");

	save_var point, matchstart, matchend, case_fold = 1;
	set_character_color(from, to, -1);
	point = from;
	c = get_tagged_region("color-gams-as-unit", give_begin_line() - 1, &p);
	if (c && p < point)	// Jump back over data areas.
		point = p;
	if (data < 0)
		while (re_search(-1, GAMS_STATEMENT_END))	// Move back to start of a statement.
			if (get_character_color(point) != color_class gams_comment
				  && get_tagged_region("color-gams-as-unit", matchstart) == -1)
				break;
	from = point;
	set_character_color(point, to, -1);
	while (point < to) {
		if (!re_search(1, "[0-9A-Za-z_.]+|^[*$]|[!#?\"'{]|/![*/]|%<%<|_!%$")) {
			t = size();
			break;
		}
		t = matchstart;
		c = character(point - 1);
		if (!gams_color_common(c, t))
			if (isalnum(c) || strchr("_.", c))	// found identifier, kywd, or number
				gams_keyword_color(t, point);
		if (point > talk_at) {
			note("Coloring GAMS file: %d%% complete...",
				 muldiv(point - from, 100, to - from));
			talk_at = point + 2000;
			talk = 1;
		}
	}
	if (to < t)
		set_character_color(to, t, -1);
	if (talk)
		note("");
	return point;
}

// Set color of "identifier" in this buffer range (actually anything
// with alpha or digits).
gams_keyword_color(from, to)
{
	char buf[500];

	if (point - from > sizeof(buf) / sizeof(*buf) - 10)
		save_var point = from + sizeof(buf) / sizeof(*buf) - 10;
	buf[0] = '|';		// get identifier, between | chars
	grab(from, point, buf + 1);
	if (index("0123456789-", buf[1]) || buf[1] == '.' && isdigit(buf[2])) {
		set_character_color(from, to, -1);
		return;
	}
	strcpy(buf + point - from + 1, "|");
	if (is_gams_data(buf)) {
		set_character_color(from, to, color_class gams_keyword);
		gams_color_descriptions();
		set_tagged_region("color-gams-as-unit", from, point - 1, 1);
		set_tagged_region("color-gams-as-unit", point - 1, point, -1);
	} else if (is_gams_keyword(buf))
		set_character_color(from, to, color_class gams_keyword);
	else {
		set_character_color(from, to, color_class gams_variable);
		if (parse_string(-1, "%.[a-z]+"))
			set_character_color(matchend, point, color_class gams_suffix);
	}
}

// We've just moved past a / at the start of a data list.  Color until
// the end.
gams_color_data_list()
{
	int t, c;

	while (re_search(1, "^[*$]|[/!#?\"'{]|%<%<|_!%$")) {
		t = matchstart;
		c = character(point - 1);
		if (c == '/' && !strchr("*/", curchar()))
			return;
		gams_color_common(c, t);
	}
}

// We've found a keyword like SET or SCALAR that's followed by one or
// more identifiers.  Each identifier might include a suffix and perhaps
// parenthesized stuff, like var.lo(1,2).  After an identifier, there
// can be descriptive text from there to the end of the line, or the
// next comma, semicolon, or slash.  The descriptive text may be quoted
// to include such delimiter characters.  After a slash (or, for TABLE
// only, after a newline) there will be data.

gams_color_descriptions()
{
	int descr, done, table, c, p, q;

	set_tagged_region("color-gams-data", point, size(), -1);
	if (parse_string(-1, GAMS_TYPE_KEYWORDS)) {
		gams_move_past_spaces_and_comments();
		if (parse_string(1, GAMS_DATA_KEYWORDS)) {
			set_character_color(matchstart, matchend, color_class gams_keyword);
			point = matchend;
		}
	}
	table = parse_string(-1, "table");	// Special rules for this keyword.
	gams_move_past_spaces_and_comments();
	move_past_gams_identifier();
	do {
		if (parse_string(1, GAMS_STATEMENT_END))
			return;
		point += parse_string(1, "[ \t]*");
		descr = point;
		done = 0;
		c = curchar();
		if (strchr("\"'", c)) {
			gams_parse_and_color_string(p = point);
		} else {
			if (re_search(1, "[\n,/]|" GAMS_STATEMENT_END)) {
				c = character(p = point - 1);
				if (isalpha(c)) {		// relaxed syntax rule: \n plus keyword means ;
					c = ';';
					p = matchstart;
					point = p + 1;
				}
			} else {
				done = 1;
				p = point;
				c = ';';
			}
		}
		set_character_color(descr, p, color_class gams_string);
		switch (c) {
			case ';':
				return;
			case '/':
				q = give_end_line();
				gams_color_data_list();
				set_tagged_region("color-gams-data", q, give_begin_line(), 1);
				if (parse_string(1, GAMS_STATEMENT_END)) {
					point = matchstart;
					return;
				}
				point += parse_string(1, "[ \t\n,]*");
				gams_move_past_spaces_and_comments();
				move_past_gams_identifier();
				break;
			case '\n':
			case ',':
			case '"':
			case '\'':
				if (parse_string(1, GAMS_STATEMENT_END)) {
					point = matchstart;
					return;
				}
				point += parse_string(1, "[ \t\n]*");
				if (table) {
					if (re_search(1, GAMS_STATEMENT_END))
						point = matchstart;
					return;
				}
				gams_move_past_spaces_and_comments();
				move_past_gams_identifier();
				break;
		}
	} while (!done);
}

// Is text in p (must be surrounded by |'s) a keyword that starts some data?
// If so, we're going to interpret the rest of this section (until a ;)
// specially.

is_gams_data(p)
char *p;
{
	if (strstr("|" GAMS_DATA_KEYWORDS "|" GAMS_TYPE_KEYWORDS "|", p))
		return 1;
	return 0;
}

is_gams_keyword(p)	// is text in p (must be surrounded by |'s) a keyword?
char *p;
{
	if (strstr("|" GAMS_UNIT_STARTERS "|" GAMS_KEYWORDS "|", p))
		return 1;
	if (strstr("|" GAMS_SPECIALS "|", p))
		return 2;		// Not really keywords for parsing, but highlight anyway.
	return 0;
}

// This runs after we compile a GAMS file using compile-buffer.  It
// finds the matching .lst file, then translates any error messages
// there into a traditional error message format, inserting the error
// lines into compilation-output where next-error and friends will look
// for them.
gams_read_list_file(int res, int buf)
{
	char list_file[FNAMELEN], orig_file[FNAMELEN], num[100], inc_file[FNAMELEN];
	char gams_err[FNAMELEN];
	int out = create("compilation-output"), i;

	if (res == 2)
		say("Interrupted, skipping error message conversion.");
	if (res == 2 || res == 3 || res == 5)	// Do_compile() didn't complete
		return;								// successfully.
	save_var bufnum = buf;
	strcpy(orig_file, filename);
	strcpy(list_file, filename);
	strcpy(get_extension(list_file), ".lst");
	int tmp = tmp_buf();
	save_var bufnum = tmp;
	if (!err_file_read(list_file, FILETYPE_AUTO)) {
		save_var point = 0;
		while (re_search(1, "^%*%*%*%* *%$")) {
			i = give_end_line();
			grab(point, i, gams_err);	// Get the error code.
			nl_reverse();		// Now copy the source line.
			buf_xfer(out, give_begin_line(), i + 1);
			point = i + 1;
			if (parse_string(1, " *([0-9]+)")) {	// Error in main file.
				grab(find_group(1, 1), find_group(1, 0), num);
				buf_printf(out, "%s %d: GAMS error %s\n",
						   orig_file, numtoi(num) - 1, gams_err);
			} else if (parse_string(1, "<*>+ *([0-9]+.+)\n *([0-9]+)")) {
				grab(find_group(1, 1), find_group(1, 0), gams_err);
				grab(find_group(2, 1), find_group(2, 0), num);
				buf_printf(out, "%s %d: GAMS error %s\n",
						   orig_file, numtoi(num) - 1, gams_err);
			} else if (parse_string(1, "%*%*%*%* *LINE +([0-9]+) +"
									"INPUT +([^ \n].*)")) {	// In another file.
				grab(find_group(1, 1), find_group(1, 0), num);
				grab(find_group(2, 1), find_group(2, 0), inc_file);
				buf_printf(out, "%s %d: GAMS error %s\n",
						   inc_file, numtoi(num), gams_err);
			}
		}
	}
	restore_vars();
	buf_delete(tmp);
	next_error();
}

command gams_mode()
{
	mode_default_settings();
	major_mode = _gams_mode_name;
	compile_buffer_cmd = compile_gams_cmd;	// can compile this
	post_compile_hook = gams_read_list_file;
	if (auto_show_gams_delimiters)
		auto_show_matching_characters = gams_auto_show_delim_chars;
	strcpy(comment_start, "^*[ \t]*");
	strcpy(comment_pattern, "^*.*$");
	strcpy(comment_begin, "* ");
	strcpy(comment_end, "");
	recolor_range = color_gams_range;	// set up coloring rules
	recolor_from_here = recolor_from_top;
	coloring_flags |= COLOR_INVALIDATE_RESETS;
	if (want_code_coloring)		// maybe turn on coloring
		when_setting_want_code_coloring();
	try_calling("gams-mode-hook");
	make_mode();
}

suffix_gms()		// .gms file invokes GAMS mode
{
	gams_mode();
}

suffix_set()		// .set file invokes GAMS mode
{
	gams_mode();
}

// These extensions aren't enabled by default, as they're commonly in
// use for non-GAMS purposes.

suffix_inc()
{
	char ext[FNAMELEN];
	if (gams_files)
		gams_mode();
	else if (!guess_mode_without_extension(ext, "suffix-%s") || !try_calling(ext))
		asm_mode();
}

suffix_map()
{
	if (gams_files)
		gams_mode();
	else
		fundamental_mode();
}

suffix_dat()
{
	if (gams_files)
		gams_mode();
	else
		fundamental_mode();
}
