/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1985, 2020 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// This file has the commands for getting help on Epsilon.

#include "eel.h"
#include "menu.h"
#include "info.h"
#include "perl.h"
#include "lowlevel.h"

command help() on reg_tab[CTRL('_')], reg_tab[ALT('?')], reg_tab[FKEY(1)],
				  reg_tab[GREYHELP]
{
	if (_show_help_on)
		help_on_command(_show_help_on);
	else if (_help_on_cmd)
		help_on_command(name_name(_help_on_cmd));
	else if (_doing_input == DI_VIEWLAST || _doing_input == DI_LINEINPUT)
		help_on_command(name_name(last_index));
	else
		general_help();
}

general_help()
{
	if (run_by_mouse)
		switch (get_epsilon_help_format()) {
			case HELP_FORMAT_WINHELP:
				win_help_contents("");		// Show Epsilon's contents topic.
				return;
			case HELP_FORMAT_HTML:
				if (try_html_help_string("") == 1)
					return;
		}
	prompt_for_help();
}

prompt_for_help()
{
	for (;;) {
		mention("Help on (? for choices): ");
		do {
			getkey();
		} while (!IS_TRUE_KEY(key));
		check_abort();
		say("");
		if (binding_of(key, root_keys) == (short) help || key == '?')
			dodesc("help", "help", 2);
		else if (key < 128)
			switch (tolower(key)) {
				case 'k': describe_key(); return;
				case 'c': describe_command(); return;
				case 'a': apropos(); return;
				case 'b': show_bindings(); return;
				case 'q': what_is(); return;
				case 'l': show_last_keys(); return;
				case 'v': about_epsilon(); return;
				case 'r': describe_variable(); return;
				case 'm': help_on_mode(); return;
				case 'i': info(); return;
				case 'f': epsilon_info_look_up(); return;
				case 'w': windows_only_help();
						  win_help_contents(""); return;
				case 'h': epsilon_html_look_up(); return;
				case 't': context_help(); return;
				case CTRL('C'): info_goto_epsilon_command(); return;
				case CTRL('K'): info_goto_epsilon_key(); return;
				case CTRL('V'):
				case CTRL('R'): info_goto_epsilon_variable(); return;
			}
	}
}

int get_epsilon_help_format()
{
	char varname[FNAMELEN];
	sprintf(varname, "epsilon-help-format-%s-%s",
			is_unix ? "unix" : is_mswindows ? "win" : "other",
			(is_gui || is_unix == IS_UNIX_XWIN) ? "gui" : "console");
	int code = get_num_var(find_index(varname));
	if (code == HELP_FORMAT_WINHELP && !(has_feature & FEAT_WINHELP))
		code = 0;		// Ignore setting, it's only available in Windows.
	if (!code)	// Pick a default format.
		if (has_feature & FEAT_WINHELP_NATIVE)	// WinHelp is always installed.
			code = HELP_FORMAT_WINHELP;
		else if (is_unix == IS_UNIX_XWIN || is_gui || is_mswindows)
			code = HELP_FORMAT_HTML;	// Try using a browser.
		else			// Fall back to internal help system.
			code = HELP_FORMAT_POPUP;
	return code;
}

// Run Windows help on file, looking up highlighted text.
// If none, either display the Contents help topic, or display
// the help file's list of search keywords.

windows_help_from(file, show_contents)
char *file;
{
	char buf[FNAMELEN];

	windows_only_help();
	save_spot point, mark;
	fix_region();
	iter = 1;

	if (is_highlight_on() && mark - point < sizeof(buf) / sizeof(*buf)){
		grab(point, mark, buf);
		highlight_off();
		return win_help_string(file, buf);
	} else if (show_contents)
		return win_help_contents(file);
	else
		return win_help_string(file, "");
}

command epsilon_manual()
{
	switch (get_epsilon_help_format()) {
		case HELP_FORMAT_WINHELP:
			win_help_string("", "Epsilon Manual Contents");
			return;
		case HELP_FORMAT_HTML:
			epsilon_manual_html();
			return;
	}
	epsilon_manual_info();
}

html_help_check_ok()
{
	if (is_unix != IS_UNIX_XWIN && !is_gui && !is_mswindows)
		error("Don't know how to run a web browser in this environment, sorry.");
}

// Open this URL in a browser and return 1.  Return 0 if we can't.
int try_show_url(char *url)
{
	char cmd[300], prog[300];

	if (is_mswindows) {
		get_executable_directory(prog);
		strcat(prog, "lhelp.exe");
		sprintf(cmd, "lhelp -u \"%s\"", url);
		winexec(prog, cmd, SW_SHOWMINNOACTIVE, 0);
		return 1;
	} else if (is_unix == IS_UNIX_XWIN) {
		if (run_helper("goto_url", url, ""))
			return 1;
	}
	return 0;
}

// Cause this URL to be displayed by a browser.
show_url(char *url)
{
	if (!try_show_url(url))
		error("Can't start a browser, please access %s manually.", url);
}

// Display topic in a browser and return 1.  Return 0 if running a web
// browser is unsupported, or -1 if we couldn't for some other reason.
int try_html_help_string(char *topic)
{
	char url[FNAMELEN], prog[FNAMELEN], *p, *q, cmdline[FNAMELEN], dir[FNAMELEN];

	if (epsilon_manual_port > 0)
		sprintf(url, "http://localhost:%d/man/", epsilon_manual_port);
	else if (epsilon_manual_port == -1)
		strcpy(url, "https://www.lugaru.com/cgi-bin/lhelp/man/");
	else
		sprintf(url, "https://www.lugaru.com/xman/v%s/", versioned_file_string);
	for (p = topic, q = url + strlen(url); *p; p++, q++)
		if (*p == ' ') {
			strcpy(q, "%20");
			q += 2;
		} else
			*q = *p;
	*q = 0;
	if (!*topic)
		strcat(url, "0");
	p = lookpath("lhelp/epsilon.lhl");
	if (p) {		// Find name of directory containing our lhelp files.
		strcpy(dir, p);
		*get_tail(dir, 0) = 0;
	} else
		get_executable_directory(dir);
	convert_to_8_3_filename(dir);
	sayput("Starting a browser, please wait...");
	expire_message = 1;
	if (opsys == OS_UNIX) {
		get_executable_file(prog, "lhelpd", 1);
		sprintf(cmdline, "%s -p %d -d \"%s\" &", prog, epsilon_manual_port, dir);
		if (epsilon_manual_port > 0)
			shell("", cmdline, "");
	} else {
		get_executable_file(prog, "lhelp.exe", 0);
		sprintf(cmdline, "lhelp -p %d -d \"%s\"", epsilon_manual_port, dir);
		if (epsilon_manual_port > 0)
			winexec(prog, cmdline, SW_SHOWMINNOACTIVE, 0);
	}
	if (!try_show_url(url))
		return -1;
	note("");
	return 1;
}

when_setting_epsilon_manual_port()
{
	char file[FNAMELEN], port[100];
	byte port_bytes[100];
	get_customization_directory(file);
	strcat(file, "epsilon-manual-port");
	sprintf(port, "%d", epsilon_manual_port);
	chars_to_bytes(port_bytes, port);
	int f = lowopen(file, 3);
	lowwrite(f, port_bytes, strlen(port));
	lowclose(f);
}

html_help_string(char *topic)
{
	switch (try_html_help_string(topic)) {
		case 0:
			error("Don't know how to run a web browser in this environment, sorry.");
		case -1:
			error("Error starting a browser.");
	}
}

command epsilon_html_look_up()
{
	char topic[FNAMELEN], def[FNAMELEN], file[FNAMELEN], *p;

	html_help_check_ok();
	*def = 0;
	if (!strcmp(bufname, INFO_BUF)) {
		info_get_file(file);
		if (strstr(file, "epsilon")) {
			p = strchr(filename, ')');
			strcpy(def, p ? (p + 1) : filename);
		}
	}
	get_strnone(topic, "Epsilon help topic to show in browser [title page]: ", def);
	html_help_string(topic);
}

command epsilon_manual_html()
{
	html_help_check_ok();
	html_help_string("");
}

command view_lugaru_web_site()	// point browser to Lugaru's web site
{
	html_help_check_ok();
	show_url("https://www.lugaru.com");
}

windows_only_help()
{
	if (!(has_feature & FEAT_WINHELP))
		error("Sorry, WinHelp isn't available in this environment.");
}

int mshelp2_wait_until_pat(char *pat)
{
	int i;
	for (i = 0; i < 60; i++) {
		point = 0;
		if (re_search(1, pat))
			return 1;
		delay(100, COND_PROC, bufnum);
		if (i > 5)
			noteput("Starting MS-Help2...  Giving up in %d seconds", 60 - i);
	}
	return 0;
}

// If line in buffer matches name, copy line to col and return 1.
mshelp2_check_std_name(char *name, char *col)
{
	point = 0;
	if (search(1, name)) {
		grab_full_line(bufnum, col);
		return 1;
	}
	return 0;
}

// If there's a collection with a date, use it.  Prefer later dates to
// earlier ones.
mshelp2_choose_collection(char *col)
{
	if (*mshelp2_collection) {		// Override default selection.
		strcpy(col, mshelp2_collection);
		return;
	}
	*col = 0;
	if (!mshelp2_wait_until_pat("Using default collection.*\n"))
		return;
	point = matchend;
	bprintf("listcoll\n");
	int before = size();
	if (!mshelp2_wait_until_pat("^--End\n"))
		return;
	int tmp = tmp_buf();
	buf_xfer(tmp, before, matchstart);
	on_exit buf_delete(tmp);
	save_var bufnum = tmp;
	case_fold = 1;
	if (mshelp2_check_std_name("MS.NETFramework.v30", col)
		  || mshelp2_check_std_name("MS.VSCC.v90", col)
		  || mshelp2_check_std_name("MS.NETFramework.v20", col)
		  || mshelp2_check_std_name("MS.VSCC.v80", col)
		  || mshelp2_check_std_name("MS.VSCC.2003", col)
		  || mshelp2_check_std_name("MS.NETFramework", col)
		  || mshelp2_check_std_name("MS.VSCC", col))
		return;
	char *months = "Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec ", pat[100];
	int year, month;
	for (year = 2020; year > 2000; year--) {
		for (month = 12; month > 0; ) {
			sprintf(pat, "qtr.%d%.3s", year, months + --month * 4);
			point = 0;
			if (search(1, pat)) {
				grab_full_line(bufnum, col);
				return;
			}
		}
		sprintf(pat, "%d", year);
		point = 0;
		if (search(1, pat)) {
			grab_full_line(bufnum, col);
			return;
		}
	}
}

show_windows_help2(char *topic)
{
	char prog[FNAMELEN], cmdline[FNAMELEN], *s, col[FNAMELEN];
	int talk = 0, regkey;

	if (!is_mswindows)
		error("This type of help is only available under Windows.");
	save_var bufnum = create("-windows-help2-helper");
	if (!is_process_buffer(bufnum)) {
		if (size())
			say("Possible problem running mshelp2.vbs helper detected.");
		if (RegOpenKey(HKEY_CLASSES_ROOT, "DExplore.AppObj", &regkey))
			return 1;
		RegCloseKey(regkey);
		sayput("Starting MS-Help2...");
		talk = 1;
		s = lookpath("mshelp2.vbs");
		if (!s) {
			get_executable_directory(prog);
			strcat(prog, "mshelp2.vbs");
		} else
			strcpy(prog, s);
		convert_to_8_3_filename(prog);
		sprintf(cmdline, "cscript %s", prog);
		if (concur_shell("", cmdline, NULL, bufname))
			return 2;
		mshelp2_choose_collection(col);
		point = size();
		if (*col)
			bprintf("collection %s\n", col);
	}
	point = size();
	bprintf("keyword %s\n", topic);
	delay(10, 0);
	if (talk)
		say("");
	if (!is_process_buffer(bufnum))
		return 3;
	return 0;
}

// When Epsilon exits, shut down the helper script.  Otherwise the
// external help program will never exit, even if the user closes its
// window.  (Unfortunately this will force the external help program to
// close when Epsilon closes.)
do_when_exiting_mshelp2(int forced)
{
	int cnt = 1;
	UNREFERENCED_PARAMETER(forced);
	if (!exist("-windows-help2-helper"))
		return;
	save_var bufnum = create("-windows-help2-helper");
	if (!is_process_buffer(bufnum))
		return;
	while (cnt <= 5) {
		point = size();
		stuff("exit\r\n");
		delay(500, COND_PROC_EXIT | COND_TRUE_KEY, bufnum);
		if (!is_process_buffer(bufnum))
			return;
		check_abort();
		note("Attempt %d/5: Can't shut down helper in -windows-help2-helper.",
			 ++cnt);
	}
}

// Look for MSDev help in typical places.  Add logic for other C++
// compilers.
context_help_windows_compilers()
{
	char item[FNAMELEN];

	context_help_get_item(item, word_pattern);
	if (!show_windows_help2(item))
		return;
	do_context_help_from_msdn(item);
}

context_help_from_msdn()
{
	char item[FNAMELEN];

	context_help_get_item(item, word_pattern);
	do_context_help_from_msdn(item);
}

do_context_help_from_msdn(char *item)
{
	char dir[FNAMELEN], *s;
	int sz = sizeof(dir) / sizeof(*dir), key;

	strcpy(dir, "C:\\Program Files");
	if (!RegOpenKey(HKEY_LOCAL_MACHINE, "SOFTWARE\\Microsoft\\"
					"Windows\\CurrentVersion", &key)) {
		RegQueryValueEx(key, "ProgramFilesDir", NULL, dir, &sz);
		RegCloseKey(key);
	}
	strcat(dir, "\\*\\**\\msdn*.col");
	if (s = do_file_match(dir, STARTMATCH | EXACTONLY))
		win_help_string(s, item);
	else
		error("Can't find help files.  Set the "
			  "context-help-rule-... variable for this mode.");
}

// Set the text in a popup window used for help.  Fill the text using
// the window's current width, then adjust its height to match the
// text, and position it just above the echo area.
show_popup_help(int win, char *text)
{
	save_var window_handle = win;
	text_color = (is_gui && want_gui_prompts) ? color_class viewed_gui_text
											  : color_class viewed_text;
	set_wattrib(win, PVERTBORDCOLOR, color_class viewed_border);
	set_wattrib(win, PHORIZBORDCOLOR, color_class viewed_border);
	buf_zap(bufnum);
	stuff(text);
	margin_right = window_width - 4;
	fill_region();
	int lines = lines_between(0, size(), 0) + 1;
	window_height = lines + 2;
	window_top = screen_lines - lines - 5;
	display_scroll_bar = 0;
	restore_vars();
	refresh();
}

/* provide help on current command */
help_on_current()
{
	char *name;

	name = name_name(last_index);
	dodesc(name, name, 0);
}

help_on_command(cmd)
char *cmd;
{
	dodesc(cmd, cmd, 0);
}

help_on_mode()		// show help for current buffer's mode
{		// uses convention that foo-mode command goes into foo mode
	char mode_cmd[FNAMELEN];

	sprintf(mode_cmd, "%s-mode", major_mode);
	dodesc(mode_cmd, mode_cmd, 1);
}


/* list all commands with a specifed substring in name or short description */
command apropos()
{
	char str[80], msg[300], name[80];
	int found = 0, start, old = bufnum, helpbuf;

	iter = 0;
	get_string(str, "Apropos: ");
	save_var bufnum;
	if (!*str || !get_doc(0))
		return;
	helpbuf = bufnum;
	noteput("Searching for %s...", str);
	point = 0;
	while (search(1, str)) {
		if (nl_reverse())
			point++;
		if (curchar() != '~')
			nl_forward();
		else {
			found = 1;
			point++;
			parse_string(1, "[^ \t\n]*", name);
			bufnum = old;
			descr_bindings(name, msg);
			buffer_printf(HELPBUF, "%s\n", msg);
			bufnum = helpbuf;
			start = point;
			nl_forward();
			xfer(HELPBUF, start, point);
			add_variable_info(find_index(name));
		}
	}
	if (found) {
		bufname = HELPBUF;
		point = 0;
		string_replace("^[^ \n\t]+", "\1#0\2", REGEX);
	} else
		buffer_printf(HELPBUF, "No matches.\n");
	bufnum = old;
	say("");
	sprintf(msg, " Apropos \"%s\" ", str);
	view_help(msg, 1);
}

add_variable_info(i)		// modify help buf to describe variable i
{
	if (vartype(i)) {
		save_var bufname = HELPBUF;
		point--;
		if (!isspace(character(point - 1)))
			insert(' ');
		bprintf("%s variable.", variable_type_name(vartype(i)));
		to_indentation();
		replace(point, toupper(curchar()));
		point = size();
	}
}


command describe_key()
{
	iter = 0;
	tell_key(1, "Describe key: ");
}

tell_key(full, pr)
char *pr;
{
	char msg[300], tmp[300], *flavor;
	int index;

	index = *ask_key(pr, tmp, 0);
	user_abort = 0;
	say("");
	if (key != full_key)
		sprintf(tmp + strlen(tmp), " (becomes %k)", key);
	if (index <= 0)
		sprintf(msg, "The key %s is not bound to any command.", tmp);
	else {
		switch (name_type(index)) {
			case NT_MACRO:	flavor = "macro"; break;
			case NT_SUBR:	flavor = "subroutine"; break;
			default:	flavor = "command"; break;
		}
		sprintf(msg, "%s runs the %s %s.",
			tmp, flavor, name_name(index));
		if (full == 2)
			info_to_epsilon_node(name_name(index), 1);
		else if (full) {
			switch (get_epsilon_help_format()) {
				case HELP_FORMAT_WINHELP:
					show_help_for(name_name(index));
					break;
				case HELP_FORMAT_HTML:
					if (try_html_help_string(name_name(index)) == 1)
						break;
				default:	// Fall through.
					dodesc(name_name(index), msg, 5);
			}
			return;
		}
	}
	say("%s", msg);
}

show_help_for(name)
char *name;
{
	switch (name_type(find_index(name))) {
		case NT_WINVAR: case NT_BUFVAR: case NT_VAR:
		case NT_BUILTVAR:
			show_var_help(name);
			return;
	}
	show_cmd_help(name);
}

command describe_command()
{
	char cmd[80];

	if (run_by_mouse)
		switch (get_epsilon_help_format()) {
			case HELP_FORMAT_WINHELP:
				win_help_string("", "commands, list of");
				return;
			case HELP_FORMAT_HTML:
				if (try_html_help_string("command reference") == 1)
					return;
		}
	iter = 0;
	save_var show_mouse_choices = 1;
	get_cmd(cmd, "Describe command", "");
	show_cmd_help(cmd);
}

show_cmd_help(char *cmd)
{
	char msg[300];

	switch (get_epsilon_help_format()) {
		case HELP_FORMAT_WINHELP:
			sprintf(msg, "%s command", cmd);
			win_help_string("", msg);
			break;
		case HELP_FORMAT_HTML:
			if (try_html_help_string(cmd) == 1)
				break;
		default:		// Fall through.
			descr_bindings(cmd, msg);
			dodesc(cmd, msg, 5);
			break;
	}
}

command describe_variable()
{
	char var[80];

	if (run_by_mouse)
		switch (get_epsilon_help_format()) {
			case HELP_FORMAT_WINHELP:
				win_help_string("", "all variables, list of");
				return;
			case HELP_FORMAT_HTML:
				if (try_html_help_string("variable reference") == 1)
					return;
		}
	iter = 0;
	save_var show_mouse_choices = 1, show_all_variables = 1;
	get_var(var, "Describe variable", "", MUST_MATCH);
	restore_vars();
	show_var_help(var);
}

show_var_help(var)
char *var;
{
	char msg[300];

	switch (get_epsilon_help_format()) {
		case HELP_FORMAT_WINHELP:
			sprintf(msg, "%s variable", var);
			win_help_string("", msg);
			break;
		case HELP_FORMAT_HTML:
			if (try_html_help_string(var) == 1)
				break;
		default:		// Fall through.
			make_var_help(var, bufnum, window_handle, 0);
			sprintf(msg, " Help on %s ", var);
			view_help(msg, 1);
			break;
	}
}

make_var_help(var, buf, win, strip_links)	// make help on var with
char *var;			// index i using values in buf, win
{					// maybe remove links to other topics from help text
	char *quotes, *val;
	int i, start, is_number;
	char tmp[FNAMELEN * 2];	// temp loc for values

	i = find_index(var);
	is_number = vartype_class(i) == 1;
	save_var window_handle = win;	// if buffer/window-specific,
	save_var bufnum = buf;		// get value from before this
	val = get_var_val(i, tmp, 0); // command
	restore_vars();

	save_var bufnum = zap(HELPBUF);
	if (!i || !vartype(i)) {
		buffer_printf(HELPBUF, "%s is not a variable.", var);
		return;
	} else if (!get_doc(1)) {
		say("");
		buffer_printf(HELPBUF, "\n\nCan't find edoc file.");
	} else if (name_help(i) < 0) {
		buffer_printf(HELPBUF, "\n\nNo information on %s.", var);
	} else {
		point = name_help(i);
		start = point;
		save_var abort_searching = 0;
		do_searching(REGEX, "^!~");
		xfer(HELPBUF, start, point);
	}
	bufname = HELPBUF;
	point = 0;
	if (strip_links)
		remove_links();
	to_end_line();	// add to end of 1st help line
	bprintf("%s%s variable %s.\n", current_column() ? " " : "",
		variable_type_name(vartype(i)), var);
	replace(0, toupper(character(0)));
	quotes = is_number ? "" : "\"";
	bprintf("Current value: %s%s%s",
		quotes, val ? val : "<Not available>", quotes);
	switch (name_type(i)) {	// has a default value too?
		case NT_BUFVAR:
		case NT_WINVAR:	// show default value
			val = get_var_val(i, tmp, 1);
			bprintf("%sefault value: %s%s%s",
				is_number ? ", d" : "\nD", quotes,
				val ? val : "<Not available>", quotes);
	}
	point = 0;
	delete_final_newlines();
}

descr_bindings(cmd, msg)	/* put line with binding info in msg */
char *cmd, *msg;		/* if none, just put cmd */
{
	int ind, n = 0;
	short *mkeys = mode_keys;

	if ((ind = find_index(cmd)) > 0) {
		sprintf(msg, "%s may be invoked by ", cmd);
		n = find_bindings(msg, ind, mkeys, root_keys, "", 0);
		strcat(msg, ".");
	}
	if (!n)
		strcpy(msg, cmd);
}

show_binding(fmt, cmd)	/* print msg, substituting binding of cmd at %s */
char *fmt, *cmd;
{
	char msg[300];

	give_binding(msg, cmd);
	say(fmt, msg);
}

give_binding(s, cmd) // put text telling how to invoke cmd in s
char *s, *cmd;
{
	int ind;

	ind = find_index(cmd);
	*s = 0;
	if (ind <= 0 || !find_bindings(s, ind, mode_keys, root_keys, "", 0))
		sprintf(s, "Alt-x %s", cmd);
}

// Display edoc-style help for str.  Insert "also" text at the start of
// the topic.  Flags: 1 means this is the last action of the command, so
// set up the window and return.  Without it, we don't return until the
// user presses some key, then return that key.  The 2 flag prevents
// displaying the h option.  The 4 flag forces using the internal help
// format.

dodesc(char *str, char *also, int flags)
{
	char *old = bufname, msg[80];
	int start, i, browser_ok = !(flags & 2);

	if (!(flags & 4))
		switch (get_epsilon_help_format()) {
			case HELP_FORMAT_WINHELP:
			case HELP_FORMAT_HTML:
				show_help_for(str);
				return;
		}
	if (!get_doc(0))
		return;
	i = find_index(str);
	if (!i || name_help(i) < 0) {
		gripe("No information on %s", str, flags & 1);
		bufname = old;
		return;
	}
	point = name_help(i);
	while (isspace(curchar()))
		point++;
	start = point;
	re_search(1, "^!~");
	xfer(HELPBUF, start, point);
	bufname = HELPBUF;
	if (also && *also)
		point = 0, stuff(also);
	insert('\n');
	point = size();
	if (is_unix != IS_UNIX_XWIN && !is_gui && !is_mswindows || (flags & 1))
		browser_ok = 0;	// No h key unless we keep control after showing.
	if (browser_ok)
		bprintf("Press H to see this topic in a web browser, with\n"
				"links to the rest of Epsilon's online manual.\n");
	point = 0;
	bufname = old;
	sprintf(msg, " Help on %s ", str);
	view_help(msg, flags & 1);
	if (browser_ok && tolower(key) == 'h') {
		ungot_key = -1;
		html_help_string(str);
		drop_pending_says();
	}
}

get_doc(quiet)	/* make sure doc file is ready to go, clear help buffer */
{		/* if !quiet, complain if can't find edoc file */
	int exists = exist("-edoc"), i;
	char *s, name[80], *oldbuf = bufname;

	create("-edoc");
	bufname = "-edoc";
	if (!exists) {
		sayput("Loading doc file...");
		if (!(s = lookpath(_docname)) || file_read(s, FILETYPE_AUTO)) {
			bufname = oldbuf;
			delete_buffer("-edoc");
			if (quiet)
				say("");
			else
				gripe("Can't find doc file %s", _docname, 0);
			return 0;
		}
		say("");
		for (i = 1; i <= final_index(); i++)
			set_name_help(i, -1);
		point = 0;
		while (re_search(1, "^~")) {
			parse_string(1, "[^ \t\n]*", name);
			if (i = find_index(name))
				set_name_help(i, point + 1);
		}
	}
	point = 0;
	zap(HELPBUF);
	return 1;
}


command what_is() on reg_tab[FKEY(6)]
{
	iter = 0;
	if (has_arg)
		describe_key();
	else
		tell_key(0, "What is key: ");
}

// Retrieve the word at pos in buffer b.
grab_man_word_from_buffer(int b, int pos, char *word, char *pat)
{
	*word = 0;
	save_var bufnum = b, point = pos;
	point--;		// Get to start of word nearest point.
	re_search(1, pat);
	re_search(-1, pat);
	re_search(1, pat);
	if (point - matchstart + 1 < ptrlen(word))
		grab(matchstart, point, word);
}

literal_if_args()
{
	save_var point = 0;
	return !isalpha(curchar()) || search(1, " ");
}

// Matches path lines from man.conf or manpath.conf.
#define MANCONF_PATH	"^(MANDATORY_|OPTIONAL_)?MANPATH" \
						"(_MAP[ \t]+[^ \t\n]+)?[ \t]+([^ \t\n]+)$"

// Compute a list of directories where man pages might be, and put in
// dest.

build_man_path(char *dest)
{
	char *path = getenv("MANPATH"), out[FNAMELEN * 4];
	int tmp;

	if (path && *path) {
		strcpy(dest, path);
		return;
	}
	save_var bufnum = tmp = tmp_buf();
	if (!file_read("/etc/manpath.config", FILETYPE_AUTO)
		  || !file_read("/etc/man.config", FILETYPE_AUTO)) {
		do_drop_matching_lines(REGEX, MANCONF_PATH, 0);
		string_replace(MANCONF_PATH, "#3", REGEX);
	}
	stuff("/usr/man\n/usr/share/man\n");
	*out = 0;
	replace_file_with_output(out, is_mswindows ? "man --path %s" : "manpath %s");
	while (path = strchr(out, ':'))
		*path = '\n';
	stuff(out);
	do_sort_region(0, size(), 0, 0);
	do_uniq(1, 1, 0);
	simple_re_replace(0, "\n", ":");
	simple_re_replace(0, ":<eof>", "");
	grab(0, size(), dest);
	restore_vars();
	buf_delete(tmp);
}

// Return matches for man pages that start with s.
char *man_match(s, flags)
char *s;
{
	char pat[FNAMELEN], match[FNAMELEN], *path, *t, *f, *q;
	char man_path[FNAMELEN * 2];

	if (strchr(s, '/') || strchr(s, '\\'))
		return do_file_match(s, flags);
	save_var cygwin_filenames |= 32;
	if ((flags & STARTMATCH) && *s && (!isalpha(*s) || s[1] == ' ')) {
		zap(_MATCH_BUF);	// Something like "-k foo" or "2 bar".
		buffer_printf(_MATCH_BUF, "%s%s\n", s, (flags & LISTMATCH) ? "" : " ");
		// Just return text itself as the match.
	} else if (flags & STARTMATCH) {
		show_text(-1, 20, "Listing matches...");	// Show for at least .2 secs.
		zap(_MATCH_BUF);
		build_man_path(path = man_path);
		while (!user_abort) {
			t = strchr(path, ':');	// Assume MANPATH uses Unix sep, not DOS.
			if (t)
				strncpy(pat, path, t - path);
			else
				strcpy(pat, path);
			convert_cygwin_filename(pat, 0);
			if (*pat && pat[strlen(pat) - 1] != '/')
				strcat(pat, "/");
			sprintf(pat + strlen(pat), "man*/%s*", s);
			f = do_file_match(pat, STARTMATCH);
			for (; f && !user_abort; f = do_file_match(pat, 0)) {
				strcpy(match, get_tail(f, 0));
				q = get_extension(match);
				if (q && !strcmp(q, ".gz"))
					*q = 0;
				q = get_extension(match);
				if (!q || !*q)
					continue;
				*q = 0;
				buffer_printf(_MATCH_BUF, "%s\n", match);
			}
			if (!t)
				break;
			path = t + 1;
		}
		save_var bufname = _MATCH_BUF;
		point = 0;
		save_var case_fold = 0;
		do_sort_region(0, size(), 0, 0);
		do_uniq(1, 1, 0);
		say("");
		restore_vars();
	}
	return general_matcher(s, flags);
}

// Make find-linked-file use the rules for C mode.
man_find_linked_file(char *file)
{
	c_find_linked_file(file);
}

// Translate _man_program value into program name to run.  Other
// man-like programs can be added here.
char *man_use_program(int type)
{
	if (type == 1)
		return perldoc_command;
	return "man";
}

man_mode()
{
	mode_keys = man_tab;
	major_mode = _man_mode_name;
	mouse_dbl_selects = 1;
	make_mode();
}

man_follow_link() on man_tab['\n'], man_tab['e'], man_tab['\r']
{
	char buf[FNAMELEN];

	save_var show_status = 0, abort_searching = 0;
	grab_man_word_from_buffer(bufnum, point, buf,
							  "[-+_0-9a-z.:]+( *%((..?|[0-9][a-z]*)%))?");
	if (man_rewrite(buf) || _man_program && *buf)
		make_man_page(buf, _man_program);
	else if (key == MOUSE_DBL_LEFT)
		mouse_selecting();
	else
		error("Not a manual page cross-reference; must use open(2) format.");
}

// Rewrite "open(2)" into "2 open"; return nonzero if we did.
man_rewrite(char *buf)
{
	char p1[FNAMELEN], p2[FNAMELEN], *p;

	if (*buf && buf[strlen(buf) - 1] == '.')
		buf[strlen(buf) - 1] = 0;		// Discard final . char.
	if (*buf && (p = strchr(buf, '('))) {
		strcpy(p1, p + 1);
		*p = 0;
		p = strchr(p1, ')');
		if (p)
			*p = 0;
		strcpy(p2, buf);
		for (p = p1; *p; p++)
			*p = tolower(*p);
		sprintf(buf, "%s %s", p1, p2);
		return 1;
	}
	return 0;
}

// Pass type of 0 for normal man page, or a code man_use_program
// recognizes.
show_man_page(int type)
{
	char topic[FNAMELEN], buf[FNAMELEN], pr[FNAMELEN];
	char *s, *prog = man_use_program(type);

	tiled_only();
	if (is_unix_flavor == IS_UNIX_MACOS && (s = getenv("MANPATH"))
		  && !strcmp(s, "/usr/man:/usr/X11R6/man") && !check_file("/usr/man"))
		putenv("MANPATH=");		// Avoid Mac's bogus MANPATH bug.

	save_var show_status = 0, abort_searching = 0, case_fold = 1;
	grab_man_word_from_buffer(bufnum, point, buf,
							  "[-+_0-9a-z.:]+(%(..?%))?");
	if (type == 1 && !strcmp(buf, "use")) {
		save_var point;
		point += parse_string(1, "<word>*<space>*");
		grab_man_word_from_buffer(bufnum, point, buf,
								  "[-+_0-9a-z.:]+(%(..?%))?");
	}
	man_rewrite(buf);
	save_var completion_literal_char = literal_if_args;
	sprintf(pr, "Show %s entry for", type ? prog : "manual");
	prompt_comp_read(topic, pr, prog ? man_match : 0,
					 COMP_FOLD | SPACE_VALID | CAUTIOUS, buf);
	if (!*topic)
		return;
	make_man_page(topic, type);
}

prompt_for_man_page() on man_tab['m']
{
	show_man_page(_man_program);
}

command man()
{
	show_man_page(0);
}

command perldoc()
{
	show_man_page(1);
}

make_man_page(char *topic, int type)
{
	char cmdline[FNAMELEN], buf[FNAMELEN];
	int color;

	save_var case_fold = 1;
	sprintf(cmdline, "%s %s", man_use_program(type), topic);
	sprintf(buf, "man-%s", topic);
	quiet_set_bookmark();
	to_buffer(buf);
	man_mode();
	putenv(strkeep("MAN_KEEP_FORMATTING=Y"));
	if (!size() || type != _man_program) {
		set_read_only(0);
		zap(buf);
		translation_type = FILETYPE_AUTO;
		pipe_text(NULL, buf, cmdline, NULL,
				  PIPE_SYNCH | PIPE_CLEAR_BUF | PIPE_NOREFRESH, 0);
		point = 0;
		if (parse_string(1, "Reformatting.*please wait.*\n"))
			delete(0, matchend);
		simple_re_replace(1, "\n\n\n+", "\n\n");
		set_character_color(0, size(), -1);
		point = 0;
		while (re_search(1, ".\b.")) {
			if (character(point - 3) == '_')
				color = color_class man_underlined;
			else
				color = color_class man_bold;
			delete(point - 3, point - 1);
			set_character_color(point - 1, point, color);
			point--;
		}
		color_ansi_buffer();
		filename = "fake-man-file.man";	// So find-linked-files works.
		force_save_as = FSA_NEWFILE;
		discardable_buffer = buffer_not_saveable = 1;
	}
	modified = point = 0;
	_man_program = type;
	set_read_only(3);
}

#define DEFN(c)    ((mtab[c] == 0) ? rtab[c] : mtab[c])
#define WALLINDENT	24	/* indentation of cmd in wall chart */

command wall_chart()
{
	short *mkeys = mode_keys;

	iter = 0;
	tiled_only();
	noteput("Building wall chart...");
	to_buffer_num(zap("wall"));
	chart_table(mkeys, root_keys, "");
	point = 0;
	modified = 0;
	say("");
}

// Return last key number in range starting at k with identical
// bindings per keytables.
int last_with_same_binding(int k, short *mtab, short *rtab)
{
	int i = get_range(mtab, k, mtab[k]);
	int j = get_range(rtab, k, rtab[k]);

	return MIN(i, j) + k - 1;
}

chart_table(mtab, rtab, prefix)
short *mtab, *rtab;
char *prefix;
{
	int c, last, i, j;
	short *newtab, *newalt;
	char msg[300];

	for (c = 0; c < MAXKEYS; c++) {
		if (DEFN(c) > 0 && name_type(DEFN(c)) != NT_TABLE) {
			strcpy(msg, prefix);
			show_char(msg, c, 1);
			last = last_with_same_binding(c, mtab, rtab);
			i = strlen(msg);
			if (last > c + 1)
				strcat(msg, "\nto"), i = 2;
			j = (WALLINDENT / tab_size) * tab_size;
			if (!j)
				j = tab_size;
			if (j <= i)
				strcat(msg, " ");
			for (; j > i; j -= tab_size)
				strcat(msg, "\t");
			if (name_type(DEFN(c)) == NT_MACRO)
				strcat(msg,"macro ");
			else if (name_type(DEFN(c)) == NT_SUBR)
				strcat(msg,"subroutine ");
			strcat(msg, name_name(DEFN(c)));
			strcat(msg, "\n");
			if (last > c + 1) {
				strcat(msg, prefix);
				show_char(msg, last, 1);
				strcat(msg, "\n");
				c = last;
			}
			stuff(msg);
		}
		i = last_with_same_binding(c, mtab, rtab);
		if (i > c)
			c = i - 1;
	}
	for (c = 0; c < MAXKEYS; c++) {
		if (DEFN(c) > 0 && name_type(DEFN(c)) == NT_TABLE) {
			strcpy(msg, prefix);
			show_char(msg, c, 1);
			strcat(msg, " ");
			newtab = index_table(mtab[c] ? mtab[c] : rtab[c]);
			newalt = index_table(rtab[c] ? rtab[c] : mtab[c]);
			newtab = newtab ? newtab : newalt;
			newalt = newalt ? newalt : newtab;
			chart_table(newtab, newalt, msg);
		}
		i = last_with_same_binding(c, mtab, rtab);
		if (i > c)
			c = i - 1;
	}
}

/* take name table index and make a string showing bindings */
find_bindings(msg, ind, mtab, rtab, prefix, count)
char *msg, *prefix;
short *mtab, *rtab;
{
	char tmp[300];
	short *newtab, *newalt;
	int i, maxwidth = screen_cols - 1;

	i = list_bindings(-1, mtab, rtab, ind);
	while (i != -1) {
		if (DEFN(i) == ind) {
			if (strlen(msg) > maxwidth - 24) {
				if (strlen(msg) < maxwidth - 18)
					strcat(msg, ", ...");
				return count;
			}
			if (count++)
				strcat(msg, ", ");
			strcat(msg, prefix);
			show_char(msg, i, 1);
		} else if (DEFN(i) > 0 && name_type(DEFN(i)) == NT_TABLE) {
			strcpy(tmp, prefix);
			show_char(tmp, i, 1);
			strcat(tmp, " ");
			newtab = index_table(mtab[i] ? mtab[i] : rtab[i]);
			newalt = index_table(rtab[i] ? rtab[i] : mtab[i]);
			newtab = newtab ? newtab : newalt;
			newalt = newalt ? newalt : newtab;
			count += find_bindings(msg, ind, newtab, newalt,
						tmp, count);
		}
		i = list_bindings(i, mtab, rtab, ind);
	}
	return count;
}

command show_bindings() on reg_tab[FKEY(5)]
{
	char cmd[80], msg[160];
	int ind, n;
	short *mkeys = mode_keys;

	iter = 0;
	if (has_arg) {
		describe_command();
		return;
	}
	save_var show_mouse_choices = 1;
	get_cmd(cmd, "Show bindings of command", "");
	if ((ind = find_index(cmd)) <= 0)
		sprintf(msg, "Couldn't find %s", cmd);
	else {
		sprintf(msg, "%s may be invoked by ", cmd);
		n = find_bindings(msg, ind, mkeys, root_keys, "", 0);
		if (!n)
			sprintf(msg, "%s is not bound to any key", cmd);
	}
	say("%s", msg);
}

command configure_epsilon()
{
	char prog[FNAMELEN];
	if (!is_mswindows)
		error("This command is only available in Epsilon for Windows.");
	get_executable_file(prog, "econfig.exe", 0);
	do_shell_execute(prog, "", "run setup");
}

command show_last_keys()
{
	char line[300];
	int i, c, len;

	len = MIN(screen_cols - _view_left - _view_right - 15,
			  sizeof(line) / sizeof(*line));
	if (is_gui && want_gui_prompts)
		len = 48;
	bufnum = zap(HELPBUF);
	bprintf("The last %d keys pressed were:\n", MAXRECENT);
	*line = 0;
	for (i = 0; i < MAXRECENT; i++) {
		c = _recent_keys[(i + _recent_place) % MAXRECENT];
		if (c == ' ')
			strcat(line, "<Sp>");
		else
			show_char(line, c, 1);
		if (strlen(line) >= len)
			bprintf("%s\n", line), *line = 0;
		else
			strcat(line, " ");
	}
	bprintf("%s", line);
	view_help(" Last Keys Pressed ", 1);
}

command about_epsilon()
{
	if (is_gui)
		about_box();
	else {
		save_var bufnum = zap(HELPBUF), window_start;
		about_box();
		save_var indents_separate_paragraphs = 1;
		save_var margin_right = screen_cols - _view_left - _view_right - 10;
		fill_paragraph();
		margin_right = screen_cols - _view_left - _view_right;
		point = 0;
		do {
			center_line();
		} while (nl_forward());
		point = 0;
		restore_vars();
		view_help(" About Epsilon ", 1);
	}
}

command release_notes()
{
	char fname[FNAMELEN], *s;

	tiled_only();
	s = lookpath("readme.txt");
	if (!s)
		error("Can't find readme.txt");
	strcpy(fname, s);
	absolute(fname);
	if (!do_find(fname, FILETYPE_AUTO)) {
		force_save_as = FSA_READONLY;
		set_read_only(1);
	}
}

command tutorial()
{
	char *s;

	tiled_only();
	to_buffer("myteach");
	if (s = lookpath("eteach"))
		do_file_read(s, FILETYPE_AUTO);
	else
		error("Can't find tutorial. Install first.");
	filename = "myteach";
	display_column = -1;
}

gripe(msg, str, last)
	char *msg, *str;
{
	save_var bufnum = zap(HELPBUF);
	bprintf(msg, str);
	insert('\n');
	view_help(" Error ", last);
}

help_linker(link)
char *link;
{
	show_help_for(link);
}

view_help(title, last)	/* show help buffer */
char *title;		/* if last, set up & return immediately */
{			/* else return after user reads it */
	buf_delete_final_newlines(name_to_bufnum(HELPBUF));
	save_var _view_title = title;

	if (last)
		save_var paging_retains_view = 1;
	if (window_bufnum != name_to_bufnum(HELPBUF)
			|| window_handle != top_popup())
		view_linked_buf(name_to_bufnum(HELPBUF), last, help_linker);
	else {
		int height = screen_lines - _view_top - _view_bottom;

		create_links();
		window_title(window_handle, BTOP, TITLECENTER, _view_title);
		window_to_fit(height);
	}
}

top_popup()	/* give window handle of top popup, or 0 if no popups */
{
	save_var window_number = -1;	/* switch to top popup */
	return is_window(window_handle) == ISPOPUP ? window_handle : 0;
}
