/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1996, 2021 Lugaru Software Ltd.  All rights reserved.  *
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

#include "eel.h"
#include "proc.h"
#include "colcode.h"
#include "c.h"
#include "html.h"
#include "indent.h"
#include "vbasic.h"
#include "perl.h"
#include "python.h"
#include "grep.h"
#include "epsnet.h"

// HTML syntax highlighting.

// Recolor just this region of css.  Last colored region may go past to.
color_css_range(from, to)
{
	int t = -1, talk, s, talked = 0;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend, case_fold = 1;
	while (from > narrow_start
		   && get_tagged_region("color-as-unit", from, &s) != -1)
		from = s - 1;
	if (from < to)
		set_character_color(from, to, -1);
	point = from;
	to_begin_line();
	talk = point + 10000;	// show status after coloring this many chars
	while (point < to) {
		if (!re_search(1, "[-0-9A-Za-z_.@%#]+|[\"']|/%*|%<%!--|-->")) {
			t = size();
			break;
		}
		t = matchstart;
		if (character(t) == '<') {
			set_character_color(t, point, color_class html_embedded);
			continue;
		}
		switch (character(point - 1)) {		// check last char
			case '>':	
				set_character_color(t, point, color_class html_embedded);
				break;
			case '*':			// found comment
				search(1, "*/");
				set_character_color(t, point, color_class css_comment);
				break;
			case '"':		// found a string literal
				point = t;
				re_search(1, "\"([^\"\\\n]|\\(.|\n))*[\"\n]");
				set_character_color(t, point, color_class css_string);
				break;
			case '\'':		// found a char const
				point = t;
				re_search(1, "\'([^\'\\\n]|\\(.|\n))*[\'\n]");
				set_character_color(t, point, color_class css_string);
				break;
			default:		// found identifier.
				if (character(matchstart) == '@' && matches_at(t, 1, "(@<word>+([^{\"]*)*)[{\"]")) {
					point = matchend - 1;		// Found a CSS rule like @media or @import
					set_character_color(t, point, color_class css_class);
				} else if (parse_string(1, "((([ \t\n]*[][,:.]?([ \t\n]*"
								 "[-0-9A-Za-z_.:@%#,]+)*)[ \t\n:().<>])*)[ \t\n]*{")) {
					point = find_group(1, 0);
					set_character_color(t, point, color_class css_selector);
				} else if (parse_string(1, "[ \t\n]*:"))
					set_character_color(t, point, color_class css_class);
				else
					set_character_color(t, point, -1);
				break;
		}
		if (point > talk) {
			note("Coloring CSS: %d%% complete...",
				 muldiv(point - from, 100, to - from));
			talk = point + 5000;
			talked = 1;
		}
	}
	if (to < t)
		set_character_color(to, t, -1);
	if (talked)
		note("");
	return point;
}

// Move backward to the nearest line guaranteed to start outside any
// colored region, return pos.  We know safe is ok value to return.
color_css_from_here(safe)
{
	to_begin_line();
	int orig = point;
	// Find a comment delim.  Ignore if it might be in a string.
	while (!re_search(-1, "/<*>|<*>/") || curchar() == '*')
		if (!parse_string(-1, "['\"].*|\\\n.*"))
			return orig;
	return safe;
}

// Does this region have unmatched { with text after it?  Return pos if
// so.
css_unclosed(int start, int end)
{
	save_var narrow_start = start;
	save_var point = end;
	int level = 0;
	while (re_search(-1, "[{}]"))
		if ((level += (curchar() == '{') ? -1 : 1) < 0)
			return parse_string(1, "{[ \t]*(/<*>|\n)") ? -1 : point;
	return -1;
}

css_indenter()
{
	int orig, ind, pos;

	to_begin_line();
	orig = point;
	if (Closeback && parse_string(1, "[ \t]*}")) {
		point = matchend;
		if ((pos = default_move_level(-1, "}", "{")) >= 0)
			ind = get_indentation(pos);
	} else if (!re_search(-1, "[^ \t\n]"))	// Find previous non-blank line.
		ind = 0;
	else if ((pos = css_unclosed(give_begin_line(), orig)) >= 0) {
		// If there's an unmatched ( between start of this line and
		// orig, align to text after (.
		point = pos;
		point += parse_string(1, "{[ \t]*");
		ind = current_column();
	} else  {
		ind = get_indentation(point);
		to_end_line();
		if (parse_string(-1, "{[ \t]*"))
			ind += css_indent;
		else if (parse_string(-1, "}[ \t]*")) {
			if ((pos = default_move_level(-1, "}", "{")) >= 0) {
				// If prev line ended in }, don't use its indent, but
				// indent of word before matching { line.
				point = pos;
				if (parse_string(-1, "<word><wspace>+"))
					point = matchend;
				ind = get_indentation(point);
			} else if (!Closeback)
				ind -= css_indent;
		}
	}
	point = orig;
	to_indentation();
	indent_to_column(ind);
}

do_css_indent()
{
	if (maybe_indent_rigidly(0))
		return;
	if (this_cmd != CMD_INDENT_REG)
		this_cmd = C_INDENT;
	if (current_column() > get_indentation(point) || prev_cmd == C_INDENT) {
		indent_like_tab();
		return;
	}
	css_indenter();
}

css_indent_and_show_matching_delimiter()
{
	normal_character();
	save_spot point;
	css_indenter();
	save_var inside_show_matching_delimiter = 1;
	restore_vars();
	if (Matchdelim && strchr(")]}", key))
		find_delimiter();
}

command css_mode()
{
	html_like_mode(0);
	mode_keys = html_tab;		/* Use these keys. */
	major_mode = _css_mode_name;
	sublanguage_code = SUBLANG_CSS;
	strcpy(comment_start, "(/<*>)[ \t\f]*");
	strcpy(comment_pattern, "/<*>(.|<newline>)*<*>/<FirstEnd>");
	strcpy(comment_begin, "/* ");
	strcpy(comment_end, " */");
	recolor_range = color_css_range;	// set up coloring rules
	recolor_from_here = color_css_from_here;
	idle_coloring_size = 1000;
	try_calling("css-mode-hook");
}

suffix_css()
{
	css_mode();
}

suffix_scss()
{
	css_mode();
}

html_region_type(int pos)
{
	if (sublanguage_code == SUBLANG_CSS)
		return HT_CSS;
	return get_tagged_region("color-as-unit", pos);
}

// Look back for block start/end syntax for CDATA or comments or
// similar, to ensure we don't start parsing in the middle of such a
// block. Position to start of the region we scanned, and return the
// safe location we found, or -1 if none found. (We don't need to rescan
// from point to that location until it changes.)

int html_recolor_from_here_helper()
{
	int start = point;

	// Look for start or end of XML's CDATA block, but ignore the
	// marker if it might be part of a string or comment.
	if (re_search(-1, "%<%!%[CDATA%[|%]%]%>!")
		  && !parse_string(-1, "(['\"#]|//||\\\n).*"))
		return point;
	point = start;
	while (re_search(-1, sublanguage_code == SUBLANG_XML
					 ? "%<%!--|--[ \t\n]*>|%<%?php"
					 : "%<%!--|-[ \t\n]*>|%<[%?]|[%?]%>") && curchar() == '<')
		if (html_region_type(point) == -1)
			return point;
	return -1;
}

// Same as previous function, but keep track of regions we've already
// searched to improve speed on huge buffers.
int html_recolor_from_here(int safe)
{
	int start = point, from, to, p, q;
	if (sublanguage_code == SUBLANG_PHP)	// Can have <?'s in strings.
		return safe;		// Can't parse it backwards, so start from top.

	if (modified_buffer_region(&from, &to, "html-color")) {
		// Buffer's changed, so forget saved info on this range.
		if (get_tagged_region("html-color", from, &p, &q) == 1)
			from = p;
		if (get_tagged_region("html-color", to, &p, &q) == 1)
			to = q;
		set_tagged_region("html-color", from, to, -1);
	}
	reset_modified_buffer_region("html-color");

	if (get_tagged_region("html-color", point, &p, &q) == 1)
		point = p;		// Skip back over section previously scanned.
	p = html_recolor_from_here_helper();
	if (p == -1) {		// Remember we found nothing before starting pos.
		set_tagged_region("html-color", narrow_start, start, 1);
		point = start;
		return start;
	}
	// Remember which region had no matches.
	set_tagged_region("html-color", point, p, 1);
	return p;
}

// Search for the next quote character c, but keep track of nesting of
// <% %> or <? ?> and ignore quotes inside.
html_find_matching_quote(int c)
{
	char pat[FNAMELEN];
	int level = 0, j, orig = point;
	strcpy(pat, "X|%<%%|%%%>|%<%?|%?%>|%<(bean):");
	pat[0] = c;
	while (re_search(1, pat)) {
		switch (character(point - 1)) {
			case '?':
			case '%':			// Found <% or <?.
				level++;
				break;
			case '>':			// Found %> or ?>.
				level--;
				break;
			case ':':
				j = point;
				if (!search(1, ">") || character(point - 2) != '/')
					point = j;
				break;
			default:
				if (level <= 0)
					return;
		}
		if (level > 3 || level < 0) {	// Mismatched, so use simple rules.
			point = orig;
			pat[1] = 0;
			search(1, pat);
			return;
		}
	}
}

// Color the attribute value, then any scripting inside.
html_color_value(int from, int to, int embed)
{
	set_character_color(from, to, color_class html_value);
	save_var narrow_start = point = from, narrow_end = size() - to;
	int cnt = 0;
	while (re_search(1, "[<][%?]")) {
		if (!embed) {
			if (!cnt++)
				set_tagged_region("color-as-unit", from, to, 99);
			// Force recoloring of rest of attribute.
			set_tagged_region("needs-color", point, to, -1);
		}
		color_html_include(-1, embed);
	}
}

// We've just parsed "<tagname ", now color any attributes before the ">".
// Script is 1 if tag name was "script" or similar, 2 if a jsp-style
// script, 3 if a style sheet, 4 if cfscripting, otherwise 0.

color_html_attributes(int script, int orig, int limit, int embed)
{
	int len, start, lang;

	if (script == 2)		// A <jsp:> tag.
		lang = get_html_asp_language(HTML_SPECIFIC_COLORING(0x10, HT_JSCRIPT));
	else if (script == 3)
		lang = HT_CSS;
	else if (script == 4)
		lang = get_html_asp_language(HTML_SPECIFIC_COLORING(0x20, HT_JSCRIPT));
	else if (script)
		lang = get_html_asp_language(html_other_coloring);
	for (;;) {
		if (curchar() == '<' && parse_string(1, "%<%!---")) {
			point = matchend;
			color_html_comment(embed);
			re_search(1, "[ \t\n]*");	// Parse whitespace.
		}
		if (!(len = parse_string(1, "[-_./a-z0-9]*")))	// An attribute?
			break;
		if (script == 1 && parse_string(1, "(language|type)[ \t\n]*=[ \t\n]*"
										"[\"']*(text/)?")) {
			start = matchend;
			if (matches_at(start, 1, "(javascript|jscript|ecmascript)"))
				lang = HTML_SPECIFIC_COLORING(1, HT_JSCRIPT);
			else if (matches_at(start, 1, "(vb)"))
				lang = HTML_SPECIFIC_COLORING(2, HT_VBSCRIPT);
			else if (matches_at(start, 1, "(php)"))
				lang = HTML_SPECIFIC_COLORING(4, HT_PHP);
			else if (matches_at(start, 1, "(python)"))
				lang = HTML_SPECIFIC_COLORING(8, HT_PYTHON);
		}
		if (character(point + len) == ':'
			  && matches_at(point + len, 1, ":[-_./a-z0-9]*")) {
			set_character_color(point, point + len + 1, color_class html_namespace);
			point += len + 1;
			len = matchend - point;
		}
		set_character_color(point, point + len, color_class html_attribute);
		point += len;			// Followed by a value?
		if (len = parse_string(1, "[ \t\n]*=[ \t\n]*")) {
			set_character_color(point, point + len, color_class html_punctuation);
							// Color the = and whitespace.
			point += len;		// Value might be quoted.
			start = point;
			if (curchar() == '"') {
				point++;
				html_find_matching_quote('"');
			} else if (curchar() == '\'') {
				point++;
				html_find_matching_quote('\'');
			} else	// No quotes, assume it's whitespace-delimited.
				re_search(1, "[^ \t\n<>]*");
			html_color_value(start, point, embed);
		}
		re_search(1, "[ \t\n]*");	// Parse whitespace.
		set_character_color(matchstart, matchend, color_class html_punctuation);
	}
	if (curchar() == '>') {
		if (character(point - 1) == '/')
			set_character_color(point - 1, point, color_class html_tag);
		set_character_color(point, point + 1, color_class html_punctuation);
		point++;
	}
	start = point;
	if (script && lang != HT_IGNORE && !parse_string(-1, "/%>")) {
		if (re_search(1, script == 3 ? "%</style%>"
					  : script == 2 ? "%</jsp:.*%>" : "%</(cf)?script(text)?%>")) {
			set_tagged_region("color-as-unit", orig, point, 99);
			point = matchstart;
		} else
			set_tagged_region("color-as-unit", orig, point, 99);
		color_embedded_code(start, point, lang, limit, embed);
	}
}

// Color this region as Java, VBScript, etc., or plain.  If limit > 0,
// we can skip coloring after that point by setting
// html-coloring-postponed.

color_embedded_code(int from, int to, int lang, int limit, int embed)
{
	save_var point;
	save_var narrow_start = from;
	save_var narrow_end = size() - to;
	if (!embed) {
		set_tagged_region("color-as-unit", from, to, lang);
		if (limit > 0 && to > limit + 1000 && lang == HT_PHP
			  && count_windows_with_buf(bufnum, 0) <= 1) {
			set_tagged_region("html-coloring-postponed", limit, to, 1);
			to = limit;		// Note that we only colored part of this.
		}
	}
	if (lang == HT_JSCRIPT) {
		save_var c_extra_keywords |= JS_KEYWORDS;
		save_var recolor_range = color_c_range;
		save_var recolor_from_here = color_c_from_here;
		coloring_flags |= COLOR_INVALIDATE_FORWARD;
		// Invalidating flag fixes issue where deleting a </script>
		// doesn't recolor rest of block as scripting.
	} else if (lang == HT_VBSCRIPT) {
		save_var recolor_range = color_vbasic_range;
		save_var recolor_from_here = recolor_by_lines;
		save_var sublanguage_code = SUBLANG_VBSCRIPT;
	} else if (lang == HT_PHP) {
		save_var recolor_range = color_full_perl_range;
		save_var recolor_from_here = recolor_from_top;
		save_var sublanguage_code = SUBLANG_PHP;
		// Any buffer with embedded PHP must invalidate coloring after
		// edits, as in Perl mode.
		coloring_flags |= COLOR_INVALIDATE_FORWARD | COLOR_INVALIDATE_RESETS;
	} else if (lang == HT_PYTHON) {
		save_var recolor_range = color_python_range;
		save_var recolor_from_here = recolor_from_top;
	} else if (lang == HT_CSS) {
		save_var recolor_range = color_css_range;
		save_var recolor_from_here = recolor_from_top;
	} else {
		set_character_color(from, to, color_class html_attribute);
		return;
	}
	color_for_narrowed_area(from, to);
}

// An HTML comment starts here.  Color it appropriately.
// HTML comments start with <!.  If there's a - immediately after,
// then this is probably a legal HTML comment like <!-- Comment -->
// so search for --> (this allows commenting out HTML code).
// Otherwise, it's not a "standard" comment (might be a legal DOCTYPE
// spec, or a valid <!>, or an illegal comment).  In any case,
// assume it ends at the next >.

color_html_comment(int embed)
{
	int s;

	search(-1, "<");	// Find start of comment.
	s = point;
	if (parse_string(1, "%<%!%[CDATA%[")) {	// xml cdata
		re_search(1, "%]%]%>");
		set_character_color(s, point, color_class html_value);
		return;
	} else if (!sublanguage_code && html_recognize_coldfusion_comments
			   && parse_string(1, "%<%!---") && !parse_string(1, "%<%!----%>")) {
		int level = 0;	// A legal ColdFusion comment.
		while (re_search(1, "%<%!---|---%>"))
			if (character(matchstart) == '<')
				level++;
			else if (--level <= 0)
				break;
		set_character_color(s, point, color_class html_altcomment);
		if (!embed)
			set_tagged_region("color-as-unit", s, point, 99);
		return;
	} else if (parse_string(1, (sublanguage_code == SUBLANG_XML)
							? "%<%!--" : "%<[ \t\n]*%![ \t\n]*--")) {
		point = matchend;		// A legal comment.
		re_search(1, "--[ \t\n]*>");	// Search for end of legal comment.
	} else if (parse_string(1, "%<%!DOCTYPE[^][<>]*%[")) {
		set_character_color(s, s + 2, color_class html_punctuation);
		set_character_color(s + 2, point = matchend, color_class html_tag);
		return;
	} else if (parse_string(1, "%<%!(ENTITY|ELEMENT|ATTLIST)[ \t\n]*")) {
		set_character_color(s, s + 2, color_class html_punctuation);
		set_character_color(s + 2, point = matchend, color_class html_tag);
		s = point;
		search(1, ">");
		set_character_color(s, point, color_class html_value);
		if (character(point - 1) == '>')
			set_character_color(point - 1, point, color_class html_punctuation);
		return;
	} else if (parse_string(1, "%<%!DOCTYPE")) {
		if (re_search(1, "[[>]") && character(matchstart) == '[') {
			search(1, "]");
			search(1, ">");
		}
		set_character_color(s, point, color_class html_value);
		return;
	} else	// Some odd flavor of comment; don't be too picky about its end.
		search(1, ">");
	set_character_color(s, point, color_class html_comment);
}

get_html_asp_language(int def)
{
	save_var point = 0;
	if (re_search(1, "%<(%%@|jsp:directive.page)[^>]*</word>Language=[\"']?")) {
		if (parse_string(1, "php"))
			return HT_PHP;
		if (parse_string(1, "vb"))
			return HT_VBSCRIPT;
		if (parse_string(1, "(java|j|ecma)script"))
			return HT_JSCRIPT;
		if (parse_string(1, "python"))
			return HT_PYTHON;
	}
	return def;
}

// Color text delimited by <% and %> or similar.  Return zero to treat
// this tag like any other; it's not the start of embedding.  If limit
// is > 0, we can skip coloring past that point by setting
// html-coloring-postponed.

color_html_include(int limit, int embed)
{
	int s, t, c, lang, orig = point, stop;
	char *end, *name;

	search(-1, "<");	// Find start of include.
	c = character(point + 1);
	s = point;
	switch (c) {
		case '%':	if (parse_string(1, "%<%%--")) {
						search(1, "--%>");	// Found a JSP "hidden comment".
						set_character_color(s, point, color_class html_comment);
						return 1;
					}
					point += parse_string(1, "%<%%[a-z0-9]*");
					end = "%>";
					lang = get_html_asp_language(sublanguage_code
						== SUBLANG_XML ? xml_asp_coloring : html_asp_coloring);
					break;
		case '?':	if (sublanguage_code == SUBLANG_XML)
						lang = xml_php_coloring;
					else if (parse_string(1, "%<%?php"))
						lang = HTML_SPECIFIC_COLORING(4, HT_PHP);
					else if (parse_string(1, "%<%?python"))
						lang = HTML_SPECIFIC_COLORING(8, HT_PYTHON);
					else
						lang = html_php_coloring;
					point += parse_string(1, "%<%?[a-z0-9]*");
					end = "?>";
					break;
		default:	point++;
					end = ">";
					lang = get_html_asp_language(html_asp_coloring);
					break;
	}
	if (lang == HT_IGNORE) {
		point = orig;
		return 0;
	}
	t = point;
	// First mark whole section embedded.
	set_character_color(s, point, color_class html_embedded);
	if (!embed)
		set_tagged_region("color-as-unit", s, point, 99);
	for (;;) {
		stop = search(1, end) ? matchstart : point;	// Find the end marker.
		color_embedded_code(t, point, lang, limit, embed);	// Color it too for now.
		if (c == '%' && lang == HT_VBSCRIPT)
			break;	// Don't do this for comment in ASP vbscript.
		// If it wound up being colored as a comment or string, that's
		// not really the end, so find the next marker.
		name = name_color_class(get_character_color(stop, &t));
		if (point <= stop || !name || !strstr(name, "comment")
			  && !strstr(name, "string") && !strstr(name, "constant"))
			break;
		// Force recoloring of that entire comment/string.
		set_tagged_region("needs-color", t, s = point, -1);
	}
	// Color the marker ending the embedded code.
	set_character_color(stop, point, color_class html_embedded);
	set_tagged_region("color-as-unit", stop, point, 99);
	return 1;
}

// Say if point is possibly inside some quoted attribute value like
// alt="1<2".  In such places <'s don't start tags.

// Look for the previous quote character of each type.  If there's a =
// before either one, and it looks like part of a tag, return 1 and
// move to the start of that tag.  Otherwise don't move and return 0.

html_in_quoted_attribute()
{
	save_var point;
	if (!re_search(-1, "['\"]"))
		return 0;
	if (!parse_string(-1, "%<" HTML_NAME "([ \t\n]+" HTML_ATTRIBUTE ")*"
					  "[ \t\n]+" HTML_NAME HTML_EQUALS))
		return 0;
	restore_vars();
	point = matchend + 1;
	return 1;
}

// Don't start coloring from inside <>'s or in script or style sections,
// since we won't do it right.  Move point back to a good place.  Look
// back for openers <script, <style, <? or <% or closers </script,
// </style, ?> or %>.  If we found an opener, start from there.
// Otherwise, move back before the nearest good <.  We also return codes
// indicating what we found.

color_html_move_back()
{
	int res = color_html_move_back_helper(), p;
	if (get_character_color(point, &p, NULL) != color_class html_comment)
		return res;
	point = p;		// Anything we found inside a comment doesn't count.
	return 4;		// Go to top of comment instead.
}

// Perform a search but ignore matches inside an embedded block of
// some other language.
int color_html_re_search_ignore_marked(int dir, char *pat)
{
	int res;
	for (;;) {
		res = re_search(dir, pat);
		if (res <= 0 || get_tagged_region("color-perl-as-unit", point) == -1)
			return res;
	}
}

color_html_move_back_helper()
{
	int start = point, p, q, from, to, res;

	save_var narrow_start = point - color_html_look_back;
	int val = get_tagged_region("color-perl-as-unit", point, &p);
	if (val == 2 && p > narrow_start) {	// We're in HTML that's embedded in Perl or PHP code.
		save_var narrow_start = p;		// So don't go back past its start.
	}
	if (modified_buffer_region(&from, &to, "html-move")) {
		if (get_tagged_region("html-move", from, &p, &q) == 1)
			from = p;
		if (get_tagged_region("html-move", to, &p, &q) == 1)
			to = q;
		set_tagged_region("html-move", from, to, -1);
	}
	reset_modified_buffer_region("html-move");
	if (get_tagged_region("html-move", point, &p, &q) == 1)
		point = p;

	res = re_search(-1, "%<%!%[CDATA%[|%]%]%>");
	p = MIN(point + 10, start);
	if (res)
		p = matchstart;
	set_tagged_region("html-move", p, start, 1);
	if (res && sublanguage_code != SUBLANG_XML
		  && !parse_string(-1, "%<script[^<]*%><wspace>*"))
		if (curchar() == '<')
			return 1;
		else {
			point += parse_string(1, "%]%]%>");
			return -1;
		}
	point = start;
	if (sublanguage_code != SUBLANG_XML
		  && color_html_re_search_ignore_marked(-1, "<Langle>[ \t\n]*/?((cf)?script|style)")
		  && parse_string(1, "<Langle>[ \t\n]*((cf)?script|style)"))
		return 2;				// Found a <script>, not a </script>.
	point = start;
	if (sublanguage_code != SUBLANG_XML
		  && color_html_re_search_ignore_marked(-1, "<Langle>[%?]|[%?]<Rangle>") && curchar() == '<'
		  && get_tagged_region("color-as-unit", point) == -1)
		return 3;				// Found a <% or <?, not a %> or ?>.
	point = start;
	do {			// start coloring from the previous tag
		if (!color_html_re_search_ignore_marked(-1, "<langle>"))
			break;	// but ignore <'s before " or ' characters
		if (get_character_color(point, &p, NULL) == color_class html_comment) {
			point = p + 2;	// If we're in a comment,
			color_html_re_search_ignore_marked(-1, "<rangle><!>");	// go back to its start.
		}
	} while (html_in_quoted_attribute());
	return 0;
}

// Color html from point to end.  Move past the part we colored, which
// may extend past end.

color_html_loop(int end, int embed)
{
	int s, t, i, script;

	for (s = point; point < end; ) {
		if (curchar() == '&') {
			set_character_color(s, point, color_class html_text);
			s = point++;	 // Color an entity like &amp; or &#251;
			if (parse_string(1, (sublanguage_code == SUBLANG_XML)
							 ? "(<:|_|p:ID_Start><:|hyphen|.|p:ID_Continue>*|#[0-9]+|#x[0-9a-f]+);?"
							 : "([a-z0-9]+|#[0-9]+|#x[0-9a-f]+);?"))
				point = matchend;
			set_character_color(s, point, color_class html_entity);
			s = point;
			continue;
		}
		if (curchar() != '<') {
			point++;
			continue;
		}
		set_character_color(s, point, color_class html_text);
		i = t = point++;
		while (strchr(" \t\n", curchar()))
			point++;	// Parse < followed by whitespace.
		set_character_color(t, point, color_class html_punctuation);
		if (curchar() == '!')
			color_html_comment(embed);
		else if (strchr("%?", curchar()) && color_html_include(end, embed))
			;
		else if (s = parse_string(1, "/?[-_.a-z0-9]*")) { // Parse the tag.
			set_character_color(point, point + s, color_class html_tag);
			script = 0;
			if (sublanguage_code != SUBLANG_XML) {
				if (s == 6 && parse_string(1, "script"))
					script = 1;
				if (s == 3 && parse_string(1, "jsp:(declaration|expression"
										   "|scriptlet)</word>"))
					script = 2;
				if (s == 5 && parse_string(1, "style"))
					script = 3;
				if (s == 8 && parse_string(1, "cfscript"))
					script = 4;
			}
			point += s;			// Look for javascript tag.
			if (curchar() == ':' && parse_string(1, ":[-_.a-z0-9]*")) {
				set_character_color(point - s, point + 1,
									color_class html_namespace);
				set_character_color(point + 1, matchend,
									color_class html_tag);
				point = matchend;
			}
			for (i = point; strchr(" \t\n", curchar()); )
				point++;	// Parse whitespace.
			set_character_color(i, point, color_class html_punctuation);
			if (parse_string(1, "/[ \t\n]*>")) {
				set_character_color(point, point + 1, color_class html_tag);
				set_character_color(point + 1, matchend,
									color_class html_punctuation);
				point = matchend;
			} else
				color_html_attributes(script, t, end, embed);
		}
		s = point;
	}
	re_search(1, "![<&]");
	if (s < point)
		set_character_color(s, point, color_class html_text);
}

// Recolor just this section. Last colored region may go past to.
int color_html_range(int from, int to)
{
	return do_color_html_range(from, to, 0);
}

// Like above, but embed nonzero means we're coloring an embedded block
// of HTML, so don't use tagged regions to mark it up or do part, just
// color the whole thing.
int do_color_html_range(int from, int to, int embed)
{
	int i, j;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend;
	if (from < to) {
		set_character_color(from, to, color_class html_text);
		if (!embed)
			set_tagged_region("needs-color", from, to, -1);	// Say it needs coloring.
	}
	point = from;
	save_var case_fold = 1;
	if (embed) {
		color_html_loop(to, embed);
		return point;
	}
	color_html_move_back();
	while (point > narrow_start
		   && get_tagged_region("color-as-unit", point - 1, &from) != -1)
		point = from;		// Move before script areas.
	if (to > size() - narrow_end)
		to = size() - narrow_end;
	set_tagged_region("color-as-unit", from, to, -1);
	from = point;
	for (;;) {
		set_tagged_region("html-coloring-postponed", 0, size(), -1); // Clear rgn.
		color_html_loop(to, embed);
		to = 0;
		while (to < size() - narrow_end) { // Colored partial script block...
			if (get_tagged_region("html-coloring-postponed", to, &i, &j) == 1) {
				if (i >= to) {
					point = i;	// ...so invalidate rest of buffer.
					set_tagged_region("needs-color", point, size(), -1);
					set_tagged_region("color-as-unit", point, size(), 98);
					return point;
				}
			}
			if (j <= to)
 				break;
			to = j;
		}

		set_tagged_region("needs-color", from, point, 0);	// Say it's colored.
		i = get_tagged_region("color-as-unit", point, NULL, &to);
		if (i == -1 || i == 99 || point >= size() - narrow_end || to <= point)
			break;
		// The next text was colored as scripting, but it shouldn't be
		// any more, since we never stop coloring inside scripting.
		// Continue coloring it as HTML, and mark the rest of the buffer
		// to be recolored later.
		set_tagged_region("needs-color", point, size(), -1);
	}
	return point;
}

// Search forward on line from here for link, set file to its name.
// Abort if none.
html_find_linked_file(char *file)
{
	char buf[FNAMELEN], *p;
	int orig;

	if (!re_search(1, "((href|src)[ \t\n]*=|@import)[ \t\n]*")) {
		to_begin_line();
		if (!re_search(1, "((href|src)[ \t\n]*=|@import)[ \t\n]*"))
			error("Can't find any link here.");
	}
	orig = point;
	if (curchar() == '"') {
		point++, orig++;
		search(1, "\"");
	} else if (curchar() == '\'') {
		point++, orig++;
		search(1, "'");
	} else	// No quotes, assume it's whitespace-delimited.
		re_search(1, "[^ \t\n<>]*"), point++;
	grab(orig, point - 1, buf);		// Get link from buffer.
	p = strchr(buf, '#');		// Remove #fragment if any.
	if (p)
		*p = 0;
	if (!*buf)
		error("Can't follow fragment links.");
	point = orig;
	if (parse_string(1, "[A-Za-z]+:"))
		strcpy(file, buf);					// An absolute URL.
	else if (parse_string(1, "/")) {		// An absolute URL path.
		strcpy(file, filename);
		strcpy(get_url_file_part(file, 1), buf);
	} else {								// A relative URL path.
		strcpy(file, filename);
		if (string_matches_regex(file, "[a-z]+:[/\\][/\\][^\n /\\]+<eof>", 1))
			sprintf(file + strlen(file), "/%s", buf);
		else
			strcpy(get_tail(file, 1), buf);
	}
}

xml_find_linked_file(char *file)
{
	html_find_linked_file(file);
}

/*
Move in direction dir past a matched pair of characters in HTML buffers.
On success, or if the end of the buffer is encountered before the
first character is found, return the new buffer position.
Otherwise return -1.
*/

html_move_level(int dir, int stop_on_key)
{
	char *findch = (dir > 0) ? "<([" : ">)]";
	char *otherch = (dir < 0) ? "<([" : ">)]";

	recolor_buffer_range(point, point + 1);
	switch (html_region_type(point - (dir < 0))) {
		case HT_JSCRIPT:
			return c_move_level(dir, stop_on_key);
		case HT_VBSCRIPT:
		case HT_PYTHON:
		case HT_CSS:
			findch = (dir > 0) ? "[({" : "])}";
			otherch = (dir < 0) ? "[({" : "])}";
			break;
		case HT_PHP:
			save_var sublanguage_code = SUBLANG_PHP;
			return perl_move_level(dir, stop_on_key);
	}
	return default_move_level(dir, findch, otherch);
}

// Narrowing may not have excluded the <script line itself, so do it
// now.
html_exclude_script_head()
{
	save_spot point = narrow_start;
	if (parse_string(1, "<langle>[%?][A-Za-z0-9]*"))
		point = matchend;
	else if (parse_string(1, "<Langle>[ \t\n]*((cf)?script|style)"))
		search(1, ">");
	if (parse_string(1, "%<%!--"))
		point = matchend;
	if (parse_string(1, "[ \t\n]*%<%!%[CDATA%["))
		point = matchend;
	narrow_start = point;
}

// Return nonzero if we should treat <p> as a paragraph container (used
// with </p>), not a paragraph separator (used alone).
int html_check_paragraph_is_container()
{
	int opens = 0, closes = 0;
	if (html_paragraph_is_container == 2) {
		save_var narrow_start = 0, narrow_end = 0, point = 0;
		narrow_end = size() - narrow_position(300000);
		save_var case_fold = 1;
		while (re_search(1, "%</?p</word>") && opens < 20) {
			if (character(matchstart + 1) == '/')
				closes++;
			else
				opens++;
		}
		html_paragraph_is_container = (closes * 2 > opens || opens < 2);
	}
	return html_paragraph_is_container;
}

// Adjust indentation level for <colgroup> line at point.  Check one tag
// just before it.  Outdent if it's <colgroup> or <col>.

int html_colgroup_adjust()
{
	save_var point, case_fold = 1;
	if (do_color_searching(REGEX | REVERSE, "<c:html-punctuation>[<]"
						   "/?(col|colgroup)</word>|<c:html-text>.") == 1
		  && curchar() == '<' && character(point + 1) != '/')
			return -1;
	return 0;
}

// Search forward (flags & 1: back) from pos for any one of a main list
// of tags.  If the tag found appears in an ok list, return the start of
// the tag.  Return -1 if we found a tag on only the first list, or -2
// if we found nothing.  Each list is |-separated.  The first list
// should not include / (it's assumed); the second must if it wants to
// permit end tags too.
html_find_open_tag(int pos, char *main, char *ok, int flags)
{
	char pat[FNAMELEN], tag[FNAMELEN];
	save_var case_fold = (sublanguage_code != SUBLANG_XML), point = pos;
	sprintf(pat, "<c:html-punctuation>[<](/?(%s))</" HTML_NAME_CHAR ">", main);
	if (do_color_searching(REGEX | ((flags & 1) ? REVERSE : 0), pat) != 1)
		return -2;
	strcpy(tag, "|");
	grab(find_group(1, 1), find_group(1, 0), tag + 1);
	strcat(tag, "|");
	sprintf(pat, "|%s|", ok);
	if (strstr(pat, tag))
		return point;
	return -1;
}

// The following functions help us with elements where the end tag is
// optional.  We want to know if we're inside such an element, so we
// look for start and end tags for that element, as well as elements
// that contain it.

// Look back for an <option>, <optgroup>, or <select> start tag, or
// their end tags.  If it was an <option> start tag, return the
// beginning position of that tag.  Return -1 if it was one of the
// others, or -2 if we found nothing.  So non-negative means we're
// inside an <option> block.
int html_open_option(int p)
{
	return html_find_open_tag((p), "option|optgroup|select", "option", 1);
}

// Look back for an <li> tag (start or end), or one of the elements that
// contain it.  If it was an <li> start tag, return the beginning
// position of that tag.  Return -1 if it was one of the others, or -2
// if we found nothing.  So non-negative means we're inside an <li>
// block.
int html_open_li(int p)
{
	return html_find_open_tag((p), "li|ol|ul|dir|menu", "li", 1);
}

// Look back for a <dd>, <dt>, or <dl> start tag, or their end tags.  If
// it was a <dt> or <dd> start tag, return the beginning position of
// that tag.  Return -1 if it was one of the others, or -2 if we found
// nothing.  So non-negative means we're inside a <dd> or <dt> block.
int html_open_dd_dt(int p)
{
	return html_find_open_tag((p), "dd|dt|dl", "dt|dd", 1);
}

// Look back for a <p>, <pre>, </p>, or </pre> tag.  If it was pre,
// return beginning position of that tag.  Return -1 if it was one of
// the others, or -2 if we found nothing.  So non-negative means we're
// inside a <pre> block.
int html_open_pre(int p)
{
	return html_find_open_tag((p), "p|pre", "pre", 1);
}

// Return nonzero if the tag containing position pos has the form
// <tag/>.  The > must be the specific color class (normally
// html-punctuation, unless the whole tag is some interpolated
// expression inside an attribute value).
html_is_self_delimiting(int pos, char *color)
{
	save_var point = pos;
	char pat[FNAMELEN];
	sprintf(pat, "<c:%s>[>]", color);
	return do_color_searching(REGEX, pat) && parse_string(-1, "/[ \t\n]*[>]");
}

// Assuming pos is inside a tag, is it unclosed?
html_invalid_tag(int pos)
{
	save_var point = pos;
	return do_color_searching(REGEX, "<c:html-punctuation>[<>]")
			&& character(point - 1) == '<';
}

// There's a < at pos in the buffer.  Return 1 if it's a start tag that
// should have a matching end tag.  Return -1 if it's an end tag.
// Return 0 if it's a self-terminating tag like img.  Return -2 if we
// can't tell.

// HTML has some elements (p, option, li, dd, dt) where the end tag is
// optional.  When we can guess that the end tag was intentionally
// omitted, we treat these as self-terminating.  If we can't tell, we
// treat them as start tags.

// Flags: 1 to treat elements in html_no_indent_elements as
// self-terminating.
int html_is_start_tag(int pos, int flags)
{
	char tag[FNAMELEN];
	int i, is_end = 0;

	save_var point = pos, case_fold = (sublanguage_code != SUBLANG_XML);
	save_var narrow_end = 0;	// Use only new limit.
	narrow_end = size() - narrow_position(point + html_tag_match_look_back);
	i = parse_string(1, "%<[ \t\n]*");
	if (!i)
		return -2;
	point += i;
	if (curchar() == '/')			// An end tag.
		is_end = 1, point++;
	point += parse_string(1, "[ \t\n]");	// Get the tag name.
	if (!parse_string(1, "[-_./A-Za-z0-9:]+", tag))
		return -2;
	if (is_end)
		return (flags & 1) && word_in_list(tag, html_no_indent_elements, 1)
				? 0 : -1;
	if (do_color_searching(REGEX, "<c:html-punctuation>%>") != 1)
		return -2;		// Find the closing >.
	if (parse_string(-1, "/%>"))
		return 0;		// It's marked as self-terminating.
	if (sublanguage_code == SUBLANG_XML || (html_style_rules & 1))
		return 1;		// In XML, all self-terminating tags must be marked.
	// In HTML, we must check more.
	if (word_in_list(tag, html_empty_elements, 1))
		return 0;
	if (tolower(tag[0]) == 'c' && tolower(tag[1]) == 'f'
		  && word_in_list(tag + 2, coldfusion_empty_elements, 1))
		return 0;
	if ((flags & 1) && word_in_list(tag, html_no_indent_elements, 1))
		return 0;	// Assume start tags are optional for these.
	if (!strfcmp(tag, "p") && !html_check_paragraph_is_container())
		return 0;
	if (!strfcmp(tag, "option") && html_open_option(pos) >= 0)
		return 0;
	if (!strfcmp(tag, "li")) {
		// These rarely have end tags, so be lenient.
		if (html_open_li(pos) >= 0 || html_find_open_tag(pos + 1,
			"li|ol|ul|dir|menu",
			"li|ol|ul|dir|menu|/ol|/ul|/dir|/menu", 0) >= 0)
			return 0;
	}
	if ((!strfcmp(tag, "dd") || !strfcmp(tag, "dt"))
		  && html_open_dd_dt(pos) >= 0)
		return 0;
	return 1;
}

do_php_indent()
{
	save_var in_perl_buffer = 1;
	save_var sublanguage_code = SUBLANG_PHP;
	save_var perl_indent = php_indent;
	save_var perl_closeback = php_closeback;
	save_var perl_topindent = php_topindent;
	save_var perl_label_indent = php_label_indent;
	save_var perl_contin_offset = php_contin_offset;
	save_var perl_align_contin_lines = php_align_contin_lines;
	save_var perl_brace_offset = php_brace_offset;
	save_var perl_top_braces = php_top_braces;
	save_var perl_top_contin = php_top_contin;
	save_var perl_top_struct = php_top_struct;
	do_c_indent();
}

// Return indent for attrs inside a tag that starts at pos, with current
// indentation cur.
html_indent_in_tag(int pos, int cur)
{
	int attrlen = 0, tagind;
	if (!(html_indenting_rules & 0x100))
		return 0;
	point += parse_string(1, "[ \t]*");
	// Get width of "attname=" on current line.
	if ((html_indenting_rules & 0x8000) && parse_string(1, HTML_NAME "[ \t]*="))
		attrlen = get_column(matchend) - current_column();
	save_var point = pos;
	if (parse_string(-1, "(^|\n)[ \t]*")) {
		// Get indentation of "<tagname".
		tagind = get_indentation(point);
		point += parse_string(1, "[<]" HTML_NAME "[ \t]*");
		if (attrlen && parse_string(1, "[ \t\n]*" HTML_NAME "[ \t]*=")) {
			int i = get_column(matchend) - attrlen;	// Indent to align ='s.
			if (i >= 0)		// Always indent at least to level of tag name.
				return MAX(i, tagind + 1);
		}		// Otherwise, don't align ='s.
		return current_column();
	}
	restore_vars();
	save_var point;
	if (!re_search(-1, "[^ \t\n]"))
		return cur;
	return get_indentation(point);
}

// Search forward counting start and end tags, and return difference in
// counts.
int html_tag_level()
{
	int level = 0;

	if (parse_string(1, "[ \t]*[<]/") && html_is_start_tag(matchend - 2, 1))
		level++;
	while (do_color_searching(REGEX, "<c:html-punctuation>[<]") == 1)
		switch (html_is_start_tag(point - 1, 1)) {
			case -1:	level--; break;
			case 1:		level++; break;
		}
	if (html_indenting_rules & 0x80)
		level = (level > 0) ? 1 : (level < 0) ? -1 : 0;
	return level;
}

int give_html_indent()
{
	if (sublanguage_code == SUBLANG_XML)
		return (xml_indent > 0) ? xml_indent : tab_size;
	return (html_indent > 0) ? html_indent : tab_size;
}

int give_html_indent_by_type(int type)
{
	switch (type) {
		case HT_JSCRIPT:
			return get_c_indent();
		case HT_VBSCRIPT:
			return vbasic_step();
		case HT_PHP:
			return (php_indent > 0) ? php_indent : tab_size;
		case HT_PYTHON:
			return python_indent;
		case HT_CSS:
			return css_indent;
	}
	return give_html_indent();
}

// Given position of </ in an end tag, return the indent of its
// matching start tag, or old if we can't.

html_indent_of_matching_tag(int pos, int old)
{
	char tag[FNAMELEN];
	int res = do_html_find_matching_tag(pos, tag);
	if (res < 0)
		return old;
	save_var point = res;
	point -= parse_string(-1, "%<[ \t]*");
	int ind = get_indentation(point);
	// The start tag might not be the first tag on its line, so adjust
	// the indent using those tags.
	save_var narrow_end = size() - point;
	to_begin_line();
	return ind + html_tag_level() * give_html_indent();
}

// Return indent of this HTML line by examining start and end tags.
int html_indent_level(int manual)
{
	char tag[FNAMELEN];
	save_var case_fold = (sublanguage_code != SUBLANG_XML);
	int ind = get_indentation(point), i, level = 0;
	i = html_in_tag();
	if (i >= 0)
		return html_indent_in_tag(i, ind);
	if (!(html_indenting_rules & 0x200) && !parse_string(1, "[ \t]*[<]"))
		return 0;
	// Indent end tag if due to pressing tab, or to reindent previous
	// line option, but not when auto-fill moves an end tag down.
	if ((manual || indenter_action != IACT_AUTO_FILL)
		  && parse_string(1, "[ \t]*[<]/"))	// Indenting an end tag.
		return html_indent_of_matching_tag(matchend - 2, ind);
	if (sublanguage_code != SUBLANG_XML	// Special rules for some tags.
		  && parse_string(1, "[ \t]*%<(" HTML_NAME ")</" HTML_NAME_CHAR ">")) {
		grab(find_group(1, 1), find_group(1, 0), tag);
		if (!strfcmp(tag, "option") && (i = html_open_option(point)) >= 0)
			return get_indentation(i);
		if (!strfcmp(tag, "li") && (i = html_open_li(point)) >= 0)
			return get_indentation(i);
		if ((!strfcmp(tag, "dd") || !strfcmp(tag, "dt"))
			  && (i = html_open_dd_dt(point)) >= 0)
			return get_indentation(i);
		if (!strfcmp(tag, "colgroup"))
			level += html_colgroup_adjust();
	}
	save_var point;
	if (do_color_searching(REGEX | REVERSE, "<^c:html-comment>[^ \t\n]") != 1)
		return 0;
	save_var narrow_end = size() - give_end_line();
	to_indentation();
	while ((i = html_in_tag()) >= 0 && i < give_begin_line()) {
		point = i;	// Go to start of tag.
		to_indentation();
	}

	ind = current_column();
	level += html_tag_level();
	return ind + level * give_html_indent();
}

int html_indent_script_end(int ch)
{
	char pat[20];
	sprintf(pat, "%%%<%%%c", ch);
	save_var point;
	while (do_color_searching(REGEX | REVERSE, pat) == 1)
		if (get_tagged_region("color-as-unit", point) == 99)
			return get_indentation(point);
	return 0;
}

/* are all chars in this region comments or whitespace? */
int all_comments(int from, int to, int type)
{
	save_var point, narrow_end = size() - to;
	int single_char = -1, slash_slash = 1, slash_star = 1;	// Start with JavaScript comment rules.
	switch (type) {
		case HT_VBSCRIPT:	// Only ' comments
			slash_slash = slash_star = 0;
			single_char = '\'';
			break;
		case HT_PHP:		// C comments plus #
			single_char = '#';
			break;
		case HT_PYTHON:	// Only # comments
			single_char = '#';
			slash_slash = slash_star = 0;
			break;
		case HT_CSS:		// Only /* */ comments.
			slash_slash = 0;
			break;
	}
	for (point = from; point < to; )
		if (isspace(curchar()))
			point++;
		else if (curchar() == single_char || curchar() == '/' && slash_slash && character(point + 1) == '/')
			nl_forward();
		else if (curchar() == '/' && slash_star && character(point + 1) == '*') {
			point += 2;
			if (!search(1, "*/"))
				return 0;
		} else
			return 0;
	return 1;
}

do_html_indent(int manual)
{
	int region, start;

	UNREFERENCED_PARAMETER(manual);
	recolor_buffer_range(point, point + 200);
	if (parse_string(1, "[ \t]*[%?]%>")
		  && get_tagged_region("color-as-unit", matchend - 2) == 99) {
		indent_to_column(html_indent_script_end(character(matchend - 2)));
		return;
	}
	region = get_tagged_region("color-as-unit",
							   point + parse_string(1, "[ \t]*"), &start);
	if (sublanguage_code == SUBLANG_CSS)
		region = HT_CSS;
	int var = (sublanguage_code == SUBLANG_XML) ? xml_auto_indent
			  : html_auto_indent; 
	if (IS_ENTER_KEY(key) && !(var & (region == HT_JSCRIPT ? 2
		: region == HT_VBSCRIPT ? 4 : region == HT_PHP ? 8
		: region == HT_PYTHON ? 16 : region == HT_CSS ? 32 : 1))) {
		if (!(var & 0x4000))
			prev_indenter();
		return;
	}
	save_var narrow_start = start;
	if (region > 0 && region <= HT_CSS && !IS_ENTER_KEY(key)
		  && !(html_indenting_rules & (1 << (region - 1)))) {
		prev_indenter();	// Can move back to prev line's indent.
		return;
	}
	if (parse_string(1, "[ \t]*%<%!%[CDATA"))
		region = 0;
	if (region >= HT_JSCRIPT && region < HT_IGNORE) {
		if (all_comments(narrow_start, point, region)) {	// Start of script block.
			narrow_start = 0;	// 0x400 indent; 0x800 indent even if <% has none.
			int i = get_indentation(start);
			if (!start || start == point)
				i = 0;
			else if ((html_indenting_rules & 0x400)
					 && ((html_indenting_rules & 0x800) || i))
				i += give_html_indent_by_type(region);
			indent_to_column(i);
			return;
		}
		html_exclude_script_head();
	}
	switch (region) {
		case HT_JSCRIPT:
			save_var c_extra_keywords |= JS_KEYWORDS;
			do_c_indent();
			return;
		case HT_VBSCRIPT:
			do_vbasic_indent();
			return;
		case HT_PHP:
			do_php_indent();
			return;
		case HT_PYTHON:
			if (this_cmd != CMD_INDENT_REG)
				do_python_indent();
			return;
		case HT_CSS:
			do_css_indent();
			return;
	}
	narrow_start = 0;
	if (sublanguage_code != SUBLANG_XML && html_open_pre(point) >= 0)
		return;		// Inside open <pre> block.
	indent_to_column(html_indent_level(manual));
}

html_indenter()
{
	do_html_indent(0);
}

html_indent_cmd() on html_tab['\t']
{
	int col = current_column();
	if (maybe_indent_rigidly(0))
		return;
	if (col > get_indentation(point)) { // Not in indentation.
		if (this_cmd != CMD_INDENT_REG)
			to_column(col + tab_size - col % tab_size);
	} else if (prev_cmd == C_INDENT)	/* repeated, make bigger */
		to_column(col + give_html_indent());
	else
		do_html_indent(1);
	if (this_cmd != CMD_INDENT_REG)
		this_cmd = C_INDENT;
}

html_redirect_active_key() on html_tab['{'], html_tab['}'], html_tab[')'],
							  html_tab[']'], html_tab[':'], html_tab['#']
{
	int region, start, end, f = 0, i;
	short *tab = 0;

	recolor_buffer_range(point, point + 1);
	region = get_tagged_region("color-as-unit", point, &start, &end);
	if (sublanguage_code == SUBLANG_CSS)
		region = HT_CSS;
	i = (1 << (region - 1 + 16));
	if (region > 0 && region <= HT_CSS && !(html_indenting_rules & i))
		region = 0;
	switch (region) {
		case HT_JSCRIPT:
			save_var auto_show_matching_characters = c_auto_show_delim_chars;
			save_var mode_move_level = c_move_level;
			tab = c_tab;
			break;
		case HT_VBSCRIPT:
			save_var auto_show_matching_characters = vbasic_auto_show_delim_chars;
			save_var mode_move_level = 0;
			tab = vbasic_tab;
			break;
		case HT_PHP:
			save_var auto_show_matching_characters = perl_auto_show_delim_chars;
			save_var mode_move_level = perl_move_level;
			save_var in_perl_buffer = 1;
			save_var sublanguage_code = SUBLANG_PHP;
			save_var perl_indent = php_indent;
			save_var perl_closeback = php_closeback;
			save_var perl_topindent = php_topindent;
			save_var perl_label_indent = php_label_indent;
			save_var perl_contin_offset = php_contin_offset;
			save_var perl_align_contin_lines = php_align_contin_lines;
			save_var perl_brace_offset = php_brace_offset;
			save_var perl_top_braces = php_top_braces;
			save_var perl_top_contin = php_top_contin;
			save_var perl_top_struct = php_top_struct;
			tab = c_tab;
			break;
		case HT_PYTHON:
			save_var auto_show_matching_characters = python_auto_show_delim_chars;
			save_var mode_move_level = 0;
			tab = python_tab;
			break;
		case HT_CSS:
			save_var auto_show_matching_characters = css_auto_show_delim_chars;
			save_var mode_move_level = 0;
			f = css_indent_and_show_matching_delimiter;
			break;
	}
	if (tab)
		f = binding_of(key, tab);
	if (!f && Matchdelim && strchr(")]}", key)) {
		save_var inside_show_matching_delimiter = 1;
		f = show_matching_delimiter;
	}
	if (f > 0) {
		save_var narrow_start = start, narrow_end = size() - end;
		html_exclude_script_head();
		do_command(f);
		return;
	}
	normal_character();
}

// Are we inside a tag, between its < and >?  If so, return pos of its
// <, else return -1.

int html_in_tag()
{
	save_var point;
	if (do_color_searching(REVERSE | REGEX, "<c:html-punctuation>[<>]") != 1
		  || curchar() != '<')
		return -1;
	return point;
}

// The tag in this range has no match, so highlight it specially.

html_show_mismatch(int start, int end)
{
	if (!(auto_show_html_delimiters & 4))	// Don't highlight mismatches.
		return;
	if (!show_mismatch_region[0]) {
		show_mismatch_region[0] = alloc_spot(1);
		show_mismatch_region[1] = alloc_spot(1);
		if (modify_region(SHOW_MISMATCH_REGION, MRTYPE, -1) != -1)
			remove_region(SHOW_MISMATCH_REGION);
		add_region(show_mismatch_region[0], show_mismatch_region[1],
				   color_class html_mismatch, 0, SHOW_MISMATCH_REGION);
		if (!(auto_show_adjacent_delimiter & 4))
			modify_region(SHOW_MISMATCH_REGION, MRCONTROL,
						  find_index("is_current_window"));
	}
	*show_mismatch_region[0] = start;
	*show_mismatch_region[1] = end;
	modify_region(SHOW_MISMATCH_REGION, MRTYPE, REGNORM);
}

// Highlight matching tags.

// It should return 0 if no highlighting should be done, 1 to make
// Epsilon try to use the auto_show_matching_characters setting for
// simple highlighting, 2 to indicate mismatched delimiters, or 3 to
// indicate matched delimiters.  For 2, we may do our own highlighting
// of the mismatch.

int do_html_auto_show_delimiter()
{
	int res, dir, i;
	char tag[FNAMELEN];

	save_var point, case_fold = (sublanguage_code != SUBLANG_XML);
	save_var matchstart, matchend, abort_searching = 0;
	save_var narrow_start = 0, narrow_end = 0;
	init_auto_show_delimiter();   // Must do this first.
	if (html_in_tag() < 0)
		return (auto_show_html_delimiters & 1) != 0;
	if (!(auto_show_html_delimiters & 2))
		return (auto_show_html_delimiters & 1) != 0;
	point -= parse_string(-1, HTML_NAME);
	*highlight_area_start[0] = point;
	dir = character(point - 1) == '<' ? 1 : -1;
	if ((i = parse_string(-1, "%</?")) && parse_string(1, HTML_NAME)) {
		*highlight_area_end[0] = matchend;
		res = do_html_find_matching_tag(point - i, tag);
		if (res == -1)
			switch (html_is_start_tag(point - i, 1)) {
				case 1:
				case -1:
					html_show_mismatch(point, *highlight_area_end[0]);
					return 2;
				case 0:
					return (auto_show_html_delimiters & 1) != 0;
			}

		if (res < 0)
			return 2;
		point = res;
		if (!parse_string(-dir, HTML_NAME))
			return 2;
	} else
		return (auto_show_html_delimiters & 1) != 0;
	*highlight_area_start[1] = matchstart;   // Mark the far end.
	*highlight_area_end[1] = matchend;
	modify_region(SHOW_MATCHING_REGION, MRTYPE, REGNORM);
			 // Make the highlighting visible.
	return 3;
}

int html_auto_show_delimiter()
{
	int region, start, end;

	modify_region(SHOW_MISMATCH_REGION, MRTYPE, 0);
	save_var point;
	recolor_buffer_range(point, point + 1);
	region = get_tagged_region("color-as-unit", point, &start, &end);
	if (sublanguage_code == SUBLANG_CSS)
		region = HT_CSS;
	save_var narrow_start = start, narrow_end = size() - end;
	if (region < HT_JSCRIPT || region > HT_CSS)
		return do_html_auto_show_delimiter();
	html_exclude_script_head();
	save_var auto_show_matching_characters, mode_move_level = 0;
	switch (region) {
		case HT_JSCRIPT:
			if (!auto_show_c_delimiters)
				return 0;
			auto_show_matching_characters = c_auto_show_delim_chars;
			mode_move_level = c_move_level;
			break;
		case HT_VBSCRIPT:
			if (!auto_show_vbasic_delimiters)
				return 0;
			auto_show_matching_characters = vbasic_auto_show_delim_chars;
			break;
		case HT_PHP:
			if (!auto_show_php_delimiters)
				return 0;
			auto_show_matching_characters = php_auto_show_delim_chars;
			mode_move_level = perl_move_level;
			break;
		case HT_PYTHON:
			if (!auto_show_python_delimiters)
				return 0;
			auto_show_matching_characters = python_auto_show_delim_chars;
			break;
		case HT_CSS:
			if (!auto_show_css_delimiters)
				return 0;
			auto_show_matching_characters = css_auto_show_delim_chars;
			break;
	}
	return default_auto_show_delimiter();
}

command html_list_mismatched_tags() on html_tab[ALT('L')]
{
	char tag[FNAMELEN], fname[FNAMELEN], dir[FNAMELEN];
	int buf = tmp_buf(), cnt = 0, pos = point;
	save_var point = 0;
	save_var case_fold = (sublanguage_code != SUBLANG_XML);
	save_var narrow_start = 0, narrow_end = 0;
	on_exit buf_delete(buf);
	recolor_buffer_range(0, size());
	init_copy_for_grep(fname, dir);
	save_var abort_searching = 0;
	while (do_color_searching(REGEX, "<c:html-punctuation>[<]") == 1
		   && !user_abort) {
		if (do_html_find_matching_tag(point - 1, tag) != -1)
			continue;
		switch (html_is_start_tag(point - 1, 1)) {
			case 1:
			case -1:
				point += parse_string(1, "<wspace|/>*");
				copy_line_for_grep(fname, buf, point, point + strlen(tag),
								   color_class html_mismatch);
				cnt++;
				break;
		}
		if (size() > 20000)
			note("%d mismatched tags, %d%% complete.",
				 cnt, muldiv(point, 100, size()));
	}
	user_abort = 0;
	if (size() > 20000)
		say("");
	point = pos;
	if (buf_size(buf))
		finish_copy_for_grep(buf, "mismatched tags", fname, dir);
	else
		say("No mismatched tags.");
}

// Given a < character at pos that's the beginning of a start tag, find
// its matching end tag, and return the position just after its element
// name.  Given a < character at pos that's the beginning of an end tag,
// find its matching start tag and return the position just before its
// element name.  Return -2 if there's no tag name at pos, or -1 if
// there's no match found for that tag, or -3 if it's a self-terminating
// tag like <tag/>.  Set tag to the tag name found (or "" if none).

int do_html_find_matching_tag(int pos, char *tag)
{
	return try_html_find_matching_tag(pos, tag, 0);
}

// Like above but accept flags: 1=on keypress immediately return -4.
int try_html_find_matching_tag(int pos, char *tag, int flags)
{
	char pat[FNAMELEN], *endcolor = "html-punctuation";
	int level, dir = 0, i;

	*tag = 0;
	save_var point = pos, case_fold = (sublanguage_code != SUBLANG_XML);
	i = parse_string(1, "%<[ \t\n]*");
	if (!i)
		return -2;
	point += i;
	if (curchar() == '/') {
		dir = REVERSE;
		level = 0;
		point++;
	} else
		level = 1;
	point += parse_string(1, "[ \t\n]");
	int color = get_character_syntax_color(point);
	if (color == color_class html_value)
		endcolor = name_color_class(color);
	if (!parse_string(1, "[-_./A-Za-z0-9:]+", tag))
		return -2;
	if (html_is_self_delimiting(point, endcolor))
		return -3;
	sprintf(pat, "%%<(/?<wspace>*<c:%s>%q)</word>",
			(color >= 0) ? name_color_class(color) : "-1", tag);
	while (do_color_searching(REGEX | dir, pat) == 1) {
		if ((flags & 1) && (char_avail() || is_key_repeating()))
			return -4;
		i = (character((dir ? matchend : matchstart) + 1) == '/') ? -1 : 1;
		if (!html_is_self_delimiting(point, endcolor))
			if (!(level += i))
				break;
	}
	if (dir == REVERSE)
		point++;
	if (level)
		return -1;
	return point;
}

command html_find_matching_tag() on html_tab[ALT('=')]
{
	char tag[FNAMELEN];

	save_var point;
	if (curchar() != '<' && !search(-1, "<"))
		error("No tag found.");		// Get to opening <.
	int res = do_html_find_matching_tag(point, tag);
	restore_vars();
	if (res == -2)
		error("No tag found.");
	else if (res == -1 || res == -3) {
		say("No match for %s tag.", tag);
		expire_message = 1;
	} else
		point = res;
}

// Set tag to the name of the tag whose < is at pos, and return 1 if
// it's an opening tag, -1 if it's a closing tag, or 0 if it's bad.

int html_get_tag_name(int pos, char *tag)
{
	save_var point = pos;
	int i = parse_string(1, "%<[ \t\n]*"), ret = 1;
	if (!i)
		return 0;
	point += i;
	if (curchar() == '/') {
		ret = -1;
		point++;
	}
	point += parse_string(1, "[ \t\n]");
	if (!parse_string(1, "[-_./A-Za-z0-9:]+", tag))
		return 0;
	return ret;
}

// Inside a start tag, move after its matching end tag.  Inside some
// other kind of tag, move to its end.  Outside a tag, move by words.

command html_forward_tag() on html_tab[ALT('F')]
{
	if (do_color_searching(REGEX, "<c:html-punctuation>[>]"
						   "|<c:html-text><word>+") != 1
		  || get_character_color(matchstart) == color_class html_text)
		return;
	char tag[FNAMELEN];
	int pos = point, match;
	if (do_color_searching(REGEX | REVERSE, "<c:html-punctuation>[<]") == 1
		  && !parse_string(1, "[<][ \t]*/")
		  && (match = do_html_find_matching_tag(point, tag)) >= 0) {
		point = match;
		do_color_searching(REGEX, "<c:html-punctuation>[>]");
		return;
	}
	point = pos;
}

// Inside an end tag, move before its matching start tag.  Inside some
// other kind of tag, move to its start.  Outside a tag, move by words.

command html_backward_tag() on html_tab[ALT('B')]
{
	if (do_color_searching(REGEX | REVERSE, "<c:html-punctuation>[<]"
						   "|<c:html-text><word>+") != 1
		  || get_character_color(matchstart) == color_class html_text)
		return;
	char tag[FNAMELEN];
	int pos = point, match;
	if (parse_string(1, "[<][ \t]*/")
		  && (match = do_html_find_matching_tag(point, tag)) >= 0) {
		point = match;
		do_color_searching(REGEX | REVERSE, "<c:html-punctuation>[<]");
		return;
	}
	point = pos;
}

// Delete current tag or next one.  It if has a matching start/end tag,
// delete that one too.
command html_delete_tag() on html_tab[ALT('D')]
{
	save_spot point;
	int pos = html_in_tag(), orig = point;
	char tag[FNAMELEN];
	if (pos >= 0)
		point = orig = pos;
	else if (parse_string(1, "<wspace>*[<]"))
		point = matchend - 1;
	else
		error("Not in or before a tag.");
	spot match = alloc_spot();
	on_exit free_spot(match);
	*match = do_html_find_matching_tag(point, tag);
	if (do_color_searching(REGEX, "<c:html-punctuation>[>]") == 1)
		delete(orig, point);
	if (*match >= 0) {
		point = *match;
		if (do_color_searching(REGEX | REVERSE, "<c:html-punctuation>[<]") == 1) {
			orig = point;
			if (do_color_searching(REGEX, "<c:html-punctuation>[>]") == 1)
				delete(orig, point);
		}
	}
}

command html_close_last_tag() on html_tab[ALT('E')]
{
	if (html_in_tag() >= 0)
		error("Can't insert an end tag inside another tag.");
	int tmp = tmp_buf();
	do_list_element_nesting(tmp, 1);
	save_var bufnum = tmp;
	point = size();
	char msg[FNAMELEN];
	if (parse_string(-1, "[<]" HTML_NAME "[>]\n", msg)) {
		msg[strlen(msg) - 2] = 0;
		restore_vars();
		bprintf("</%s>", msg + 1);
		save_spot point;
		search(-1, "<");
		if (current_column() <= get_indentation(point))
			do_html_indent(1);
	} else if (parse_string(-1, ".*\n", msg)) {
		msg[strlen(msg) - 1] = 0;
		error("%s", msg);
	} else
		error("Not in any element.");
}

xml_sort_attrs_for_one_tag(int pos)
{
	save_var point = pos;
	if (do_color_searching(REGEX | REVERSE, "<c:html-punctuation>[<]") != 1)
		return;
	point += parse_string(1, "[<]<wspace>*" HTML_NAME "<wspace>+");
	save_var narrow_start = point;
	if (do_color_searching(REGEX, "<c:html-punctuation>[>]") != 1)
		return;
	point -= parse_string(-1, "<wspace>*/?[>]");
	save_var narrow_end = size() - point, abort_searching = 0;
	simple_re_replace(1, "\n", XML_SORT_NL_MARKER);
	point = 0;
	stuff(XML_SORT_MARKER);
	while (point < size() - narrow_end && !user_abort) {
		point += parse_string(1, HTML_NAME XML_SORT_WSPACE "*");
		int len;
		if (curchar() == '=') {
			len = parse_string(1, "=" XML_SORT_WSPACE "*(<^wspace|\"|'>+"
							   "|\"[^\"]*\"|'[^']*')");
			if (!len)
				break;
			point += len;
		}
		len = parse_string(1, XML_SORT_WSPACE "+");
		if (!len)
			break;
		delete(point, point + len);
		stuff("\n" XML_SORT_MARKER);
	}
	point = size();
	stuff("\n");
	do_sort_region(narrow_start, size() - narrow_end, 0, 0);
	point = 0;
	delete(point, point + parse_string(1, XML_SORT_MARKER));
	simple_re_replace(1, "\n?^" XML_SORT_MARKER, " ");
	point = size();
	delete(point - parse_string(-1, "\n?"), point);
	simple_re_replace(1, XML_SORT_NL_MARKER, "\n");
}

// Fill in name with the name of the n'th attribute (from 0) in the
// current tag and return 1.  If not that many, return 0.
int xml_find_nth_attr(int n, char *name)
{
	*name = 0;
	point = 0;
	if (do_color_searching(REGEX, "<c:html-punctuation>[<]") != 1)
		return 0;
	point += parse_string(1, "<wspace>*" HTML_NAME "<wspace>+");
	save_var narrow_start = point;
	if (do_color_searching(REGEX, "<c:html-punctuation>[>]") != 1)
		return 0;
	point -= parse_string(-1, "<wspace>*/?[>]");
	save_var narrow_end = size() - point, abort_searching = 0;
	point = 0;
	while (point < size() - narrow_end && !user_abort) {
		point += parse_string(1, HTML_NAME, name);
		if (strlen(name) <= 0)
			return 0;
		if (n-- <= 0)
			return 1;
		point += parse_string(1, "<wspace>*");
		if (curchar() == '=')
			point += parse_string(1, "=<wspace>*(<^wspace|\"|'>+"
								  "|\"[^\"]*\"|'[^']*')");
		int len = parse_string(1, "<wspace>+");
		if (!len)
			break;
		point += len;
	}
	return 0;
}

// Find cnt'th attribute of the first tag in the buffer.  Align all
// attributes with that name in the region.  Return 0 if there's no
// n'th attr.
int xml_align_one_attr(int cnt)
{
	char name[FNAMELEN], pat[FNAMELEN];
	if (!xml_find_nth_attr(cnt, name))
		return 0;
	save_spot point;
	sprintf(pat, "( )(<c:html-attribute|html-namespace>%q)", name);
	align_by_pattern(16, narrow_start, size() - narrow_end, pat, 1);
	return 1;
}

// Go to each tag and reorder its attrs, then align them.
command xml_sort_by_attribute_name() on html_tab[ALT('R')]
{
	save_spot point, mark;
	if (is_highlight_on()) {
		fix_region();
		save_var narrow_start = point;
		save_var narrow_end = size() - mark;
		point = 0;
		while (do_color_searching(REGEX, "<c:html-punctuation>[<]")) {
			xml_sort_attrs_for_one_tag(point);
			check_abort();
		}
		int cnt = 0;
		while (xml_align_one_attr(cnt++))
			check_abort();
   } else {
		point++;
		if (do_color_searching(REGEX | REVERSE, "<c:html-punctuation>[<]") == 1)
			xml_sort_attrs_for_one_tag(point + 1);
	}
}

// List element context at point into buffer b.  Flags: 1=multi-line
// format, 2=give up on keypress and return 0.
int do_list_element_nesting(int b, int flags)
{
	char tag[FNAMELEN], depth[FNAMELEN];
	save_var point, case_fold = (sublanguage_code != SUBLANG_XML);
	int open = html_in_tag() != -1;
	while (do_color_searching(REVERSE | REGEX, "<c:html-punctuation>[<]") == 1) {
		if ((flags & 2) && (char_avail() || is_key_repeating()))
			return 0;		// Give up search now?
		set_buf_point(b, 0);
		int i = html_is_start_tag(point, 0);
		if (i == -1) {
			i = try_html_find_matching_tag(point, tag, (flags & 2) ? 1 : 0);
			if (i >= 0) {
				point = i;
				search(-1, "<");	// Get from element name to before its <.
			} else if (i == -4)
				return 0;
			else if (!(flags & 1))
				buf_printf(b, "%s? ", tag);
			else if (i == -1)
				buf_printf(b, "</%s> unmatched at line %d, position %d.\n",
						   tag, lines_between(0, point) + 1, point);
			else
				buf_printf(b, "Can't recognize tag at line %d, position %d.\n",
						   lines_between(0, point) + 1, point);
		} else if ((i == 1 || !i && open) && html_get_tag_name(point, tag) == 1) {
			char *bad = html_invalid_tag(point + 1) ? "?" : "";
			if ((flags & 1) && *bad)
				buf_printf(b, "Unclosed <%s at line %d, position %d.\n",
						   tag, lines_between(0, point) + 1, point);
			else
				buf_printf(b, !(flags & 1) ? (open ? "<%s%s " : "%s%s ")
										   : (open ? "<%s%s\n" : "<%s%s>\n"),
						   tag, bad);
		}
		open = 0;
	}
	if (flags & 1) {
		save_var bufnum = b, point = 0;
		*depth = 0;
		while (re_search(1, "^[<]" HTML_NAME)) {
			set_character_color(matchstart + 1, matchend, color_class html_tag);
			point = matchstart;
			bprintf("%s", depth);
			strcat(depth, "  ");
			to_end_line();
		}
	}
	return 1;
}

command html_list_element_nesting() on html_tab[ALT('i')]
{
	int b = zap("-list-element-nesting"), sz;

	iter = 0;
	do_list_element_nesting(b, 1);
	set_buf_point(b, 0);
	sz = buf_size(b);
	char *name = *filename ? filename : bufname;
	if (!want_gui_prompts || !is_gui)
		buf_set_character_color(b, 0, size(), 0);
	buf_printf(b, "Element nesting in %s%s:\n",
			   strlen(name) > 20 ? "\n" : "", name);
	save_var _view_title = "Element Nesting";
	if (sz)
		view_buf(b, 1);
	else
		say("Not in any element.");
}

// Set display_func_name to the name of the function we're editing, and
// return 1.  If not in a function, set display_func_name to "" and
// return 1.  If user pressed a key and we gave up for now, return 0.

javascript_func_name_finder()
{
	int from, to, orig = point;

	save_var point, case_fold = 1;
	to_end_line();
	while (do_color_searching(REVERSE | REGEX, "^[ \t]*<^c:*comment><^c:*string>"
							  "(function)</word>") == 1) {
		if (!must_find_func_name && (char_avail() || is_key_repeating()))
			return 0;		// Give up search now?
		if (!parse_string(1, "^[ \t]*(function)[ \t]+([a-zA-Z_]<word>*)"))
			continue;
		from = find_group(2, 1);
		to = find_group(2, 0);
		to = MIN(to, from + ptrlen(display_func_name) - 1);
		grab(from, to, display_func_name);
		point = matchend;
		point += parse_string(1, "[ \t]*");
		if (curchar() == '(')
			move_level(1, "(", ")", 0, 0);
		move_level(1, "{", "}", 0, 0);
		if (give_end_line() < orig)
			*display_func_name = 0;
		return 1;
	}
	*display_func_name = 0;
	return 1;
}

// Set display_func_name to the name of the function we're editing, and
// return 1.  If not in a function, set display_func_name to "" and
// return 1.  If user pressed a key and we gave up for now, return 0.
// Only for parts of an .html or .php file that contain php, not html.

inner_php_func_name_finder()
{
	int from, to, orig = point;

	save_var point, case_fold = 1, narrow_end;
	to_end_line();
	while (do_color_searching(REVERSE | REGEX, "^[ \t]*<^c:*comment><^c:*string>"
							  "(function)</word>") == 1) {
		if (!must_find_func_name && (char_avail() || is_key_repeating()))
			return 0;		// Give up search now?
		if (!parse_string(1, "^[ \t]*(function)[ \t]+([a-zA-Z_]<word>*)"))
			continue;
		from = find_group(2, 1);
		to = find_group(2, 0);
		to = MIN(to, from + ptrlen(display_func_name) - 1);
		grab(from, to, display_func_name);
		point = matchend;
		point += parse_string(1, "[ \t]*");
		narrow_end = size() - (orig + 1000);
		if (curchar() == '(')
			move_level(1, "(", ")", 0, 0);
		move_level(1, "{", "}", 0, 0);
		if (give_end_line() < orig)
			*display_func_name = 0;
		narrow_end = 0;
		return 1;
	}
	*display_func_name = 0;
	return 1;
}

// Put contents of buffer b into display_func_name, chopping it off if
// too long.
html_trim_element_nesting(int b)
{
	save_var bufnum = b;
	int space = window_width
				- strlen(get_tail(*filename ? filename : bufname, 0))
				- html_display_nesting_width;
	if (space < 15) {		// If we'd have too little room, skip it.
		*display_func_name = 0;
		return;
	}
	point = size() - space;
	if (point > 0) {
		point += 4;		// Make room for "..."
		if (!search(1, " "))
			if (search(-1, " "))
				point++;
		delete(0, point);
		stuff("... ");
	}
	point = size() - ptrlen(display_func_name) - 10;
	if (point) {
		delete(0, point);
		stuff("... ");
	}
	grab(0, size(), display_func_name);
}

// Set display_func_name to the name of the function we're editing, and
// return 1.  If not in a function, set display_func_name to "" and
// return 1.  If user pressed a key and we gave up for now, return 0.

// For XML/HTML, we usually substitute a list of nested elements for
// the start of the line for the function name.  Since we don't get
// called for interline moves, we don't use the position within the
// current line.

html_func_name_finder()
{
	int start, end, region, res = 1;
	*display_func_name = 0;
	if (sublanguage_code == SUBLANG_CSS)
		return 1;
	recolor_buffer_range(point, point + 1);
	region = get_tagged_region("color-as-unit", point, &start, &end);
	save_var narrow_start = start, narrow_end = size() - end, bufnum;
	if (region <= 0 && html_display_nesting_width >= 5
		  && (html_display_definition
			  & ((sublanguage_code == SUBLANG_XML) ? 4 : 8))) {
		int b = tmp_buf();
		narrow_start = narrow_end = 0;
		save_var point = give_begin_line();
		if (!do_list_element_nesting(b, must_find_func_name ? 0 : 2))
			return 0;
		html_trim_element_nesting(b);
		restore_vars();
		buf_delete(b);
		return 1;
	}
	if (region > 0 && (html_display_definition & 3)) {
		html_exclude_script_head();
		switch (region) {
			case HT_JSCRIPT:
				save_var mode_move_level = c_move_level;
				if (html_display_definition & 1)
					res = javascript_func_name_finder();
				if ((html_display_definition & 2) && res && !*display_func_name)
					strcpy(display_func_name, "JavaScript");
				return res;
			case HT_VBSCRIPT:
				save_var mode_move_level = 0;
				if (html_display_definition & 1)
					res = vbasic_func_name_finder();
				if ((html_display_definition & 2) && res && !*display_func_name)
					strcpy(display_func_name, "VBScript");
				return res;
			case HT_PHP:
				save_var mode_move_level = perl_move_level;
				if (html_display_definition & 1)
					res = inner_php_func_name_finder();
				if ((html_display_definition & 2) && res
					  && !*display_func_name && strcmp(major_mode, "PHP"))
					strcpy(display_func_name, "PHP");
				return res;
			case HT_PYTHON:
				save_var mode_move_level = 0;
				if (html_display_definition & 1)
					res = python_func_name_finder();
				if ((html_display_definition & 2) && res && !*display_func_name)
					strcpy(display_func_name, "Python");
				return res;
			case HT_CSS:
				if (html_display_definition & 2)
					strcpy(display_func_name, "CSS");
				return res;
		}
	}
	return 1;
}

php_func_name_finder()		// Uses same rules as HTML.
{
	return html_func_name_finder();
}

xml_func_name_finder()		// Uses same rules as HTML.
{
	return html_func_name_finder();
}

// Inside embedded code blocks, return the appropriate value, not the
// HTML one.
char *html_comment_start_override()
{
	recolor_buffer_range(point, point + 1);
	switch (html_region_type(point)) {
		case HT_PHP:
			return "(/</|*>|#)[ \t\f]*";
		case HT_JSCRIPT:
			return "(/</|*>)[ \t\f]*";
		case HT_VBSCRIPT:
			return "(rem |')[ \t]*";
		case HT_PYTHON:
			return "#[ \t]*";
		case HT_CSS:
			return "/<*>[ \t\f]*";
	}
	return 0;
}

char *html_comment_pattern_override()
{
	recolor_buffer_range(point, point + 1);
	switch (html_region_type(point)) {
		case HT_PHP:
			return "(//|#).*$|/<*>(.|<newline>)*<*>/<FirstEnd>";
		case HT_JSCRIPT:
			return "//.*$|/<*>(.|<newline>)*<*>/<FirstEnd>";
		case HT_VBSCRIPT:
			return "(rem |').*$";
		case HT_PYTHON:
			return "#.*$";
		case HT_CSS:
			return "/<*>(.|<newline>)*<*>/<FirstEnd>";
	}
	return 0;
}

char *html_comment_begin_override()
{
	save_var narrow_start = 0, narrow_end = 0;
	recolor_buffer_range(point, point + 1);
	switch (html_region_type(point)) {
		case HT_PHP:
			int start;
			save_var point;
			point += parse_string(1, "<hspace>*");
			if (get_character_color(point, &start) == color_class perl_comment)
				if (matches_at(start, 1, "//"))
					return "// ";
				else if (character(start) == '#')
					return "# ";
			switch (php_comment_style) {
				case 1: return "# ";
				case 2: return "// ";
			}
			return "/* ";
		case HT_JSCRIPT:
			return new_c_comments ? "// " : "/* ";
		case HT_VBSCRIPT:
			return "' ";
		case HT_PYTHON:
			return "# ";
		case HT_CSS:
			return "/* ";
	}
	return 0;
}

char *html_comment_end_override()
{
	save_var narrow_start = 0, narrow_end = 0;
	recolor_buffer_range(point, point + 1);
	switch (html_region_type(point)) {
		case HT_PHP:
			switch (php_comment_style) {
				case 1:
				case 2: return "";
			}
			return " */";
		case HT_JSCRIPT:
			return new_c_comments ? "" : " */";
		case HT_VBSCRIPT:
		case HT_PYTHON:
			return "";
		case HT_CSS:
			return " */";
	}
	return 0;
}

// PHP uses the same settings as HTML.
char *php_comment_start_override()
{
	return html_comment_start_override();
}

char *php_comment_pattern_override()
{
	return html_comment_pattern_override();
}

char *php_comment_begin_override()
{
	return html_comment_begin_override();
}

char *php_comment_end_override()
{
	return html_comment_end_override();
}

context_help_override_html(char *mode)
{
	save_var narrow_start = 0, narrow_end = 0;
	recolor_buffer_range(point, point + 1);
	switch (html_region_type(point)) {
		case HT_PHP:
			strcpy(mode, "php");
			return;
		case HT_JSCRIPT:
			strcpy(mode, "jscript");
			return;
		case HT_VBSCRIPT:
			strcpy(mode, "vbscript");
			return;
		case HT_PYTHON:
			strcpy(mode, "python");
			return;
	}
	strcpy(mode, "html");
}

context_help_override_css(char *mode)
{
	strcpy(mode, "html");
}

// Is it OK to break line at pos?  Not if it's inside a quoted
// attribute.  If the character at pos is part of a value and the first
// character of the value is a quote, then it is, so return 0 to
// disallow.  Also enforce the rule on short element names.

int html_restrict_break(int pos)
{
	save_var matchstart, matchend, point;
	if (get_character_syntax_color(pos) == color_class html_value) {
		int start;
		get_character_color(pos, &start);
		return !strchr("\"'", character(start));
	}
	int var = (sublanguage_code == SUBLANG_XML) ? xml_auto_fill_combine
												: html_auto_fill_combine;
	if (var && parse_string(-1, "%<[ \t]*(" HTML_NAME ")[ \t]*")
		  && find_group(1, 0) - find_group(1, 1) <= var)
		return 0;
	return 1;
}

html_narrow_to_paragraph()
{
	char pat[1200];
	char *elems = (sublanguage_code == SUBLANG_XML) ? xml_paragraph_elements
		: html_paragraph_elements;
	save_var point;
	int pos = point, save, inside;
	inside = html_in_tag();
	if (inside >= 0) {		// If we're inside a tag like <p, go past it
		sprintf(pat, "[<](%s)</word>", elems); 	// so we find it in next step.
		if (matches_at(inside, 1, pat))
			point = matchend;
	} else if (parse_string(1, "<hspace>*[<]<alphanum>+"))
		point = matchend;	// Likewise if we're just before one.
	sprintf(pat, "\n<hspace>*\n|[<](%s)</word>", elems);
	if (re_search(-1, pat)) {
		save = point = matchstart;
		if (character(matchend) == '<') {
			point = matchend;
			int end = default_move_level(1, "<", ">");
			point = (end > 0) ? end : save;
			point += parse_string(1, "<hspace>*");
		}
		to_begin_line();
		narrow_start = point;
	}
	point = pos;
	sprintf(pat, "\n<hspace>*\n|[<]/(%s)</word>", elems);
	if (re_search(1, pat)) {
		point = matchstart;
		point -= parse_string(-1, "<wspace>*");
		narrow_end = size() - point;
	}
}

command html_fill_paragraph() on html_tab[ALT('q')]
{
	int region, start, end;

	recolor_buffer_range(point, point + 1);
	region = html_region_type(point);
	switch (region) {
		case HT_JSCRIPT:
			save_var major_mode = _c_mode_name;
			fill_comment();
			return;
		case HT_VBSCRIPT:
		case HT_PHP:
		case HT_PYTHON:
		case HT_CSS:
			fill_comment();
			return;
	}
	if (!is_highlight_on()) {
		save_var narrow_start, narrow_end;
		if (get_character_color(point, &start, &end) == color_class html_comment) {
			narrow_start = start;
			narrow_end = size() - end;
		} else
			html_narrow_to_paragraph();
	}
	save_var mode_restrict_break = html_restrict_break;
	fill_indented_paragraph();
}

// We're inserting the character in key; maybe break this line first
// (or instead).  Return 1 if we still want to insert the character, or
// 0 to ignore it.  But if mode is 0,1,2, then we're only setting (0,1) or
// retrieving (2) the value of the mode-specific variable.

int html_maybe_break_line(int mode)
{
	char *var = (sublanguage_code == SUBLANG_XML) ? &xml_auto_fill_mode
				: &html_auto_fill_mode;
	switch (mode) {
		case 0:
			*var &= ~1;
			return 0;
		case 1:
			*var |= 1;
			return 0;
		case 2:
			return *var & 1;
	}
	if (!(*var & 1))		// We're not in auto-fill mode.
		return 1;
	recolor_buffer_range(point, point + 1);
	int have = 16;	// Assume we're in scripting.
	if (html_region_type(point) == -1)	// We're not. Comment, tag, or text?
		have = (get_character_color(point - 1) == color_class html_comment)
			   ? 8 : (html_in_tag() == -1) ? 2 : 4;
	if (!(have & *var))
		return 1;	// Only do one sort of filling?
	switch (html_region_type(point)) {
		case HT_PHP:			 // Break only if in a comment.
		case HT_VBSCRIPT:
		case HT_PYTHON:
		case HT_CSS:
			return do_maybe_break_line(0);
		case HT_JSCRIPT:
			return do_maybe_break_line(1);
	}
	save_var mode_restrict_break = html_restrict_break;
	return maybe_break_this_line();	// Break always as text.
}

tag_mode_jscript()
{
	char func[TAGLEN];
	int start;

	save_var point = 0, case_fold = 1;
	while (re_search(1, "(^|;)[ \t]*(var|function)[ \t]+(<word>+)")) {
		int var = tolower(character(find_group(2, 1))) == 'v';
		grab(start = find_group(3, 1), find_group(3, 0), func);
		if (!isdigit(*func))
			add_tag(func, start);
		if (var)
			for (;;) {
				point += parse_string(1, "([ \t\n]|//.*)*");
				if (curchar() == '=')
					re_search(1, "![\n,;]");
				if (!parse_string(1, ",[ \t\n]*(<word>+)"))
					break;
				grab(start = find_group(1, 1), find_group(1, 0), func);
				point = matchend;
				if (!isdigit(*func))
					add_tag(func, start);
			}
		else
			c_skip_level('{', '}');
	}
}

// Narrow and tag this region if it's scripting.
html_tag_region(int type, int from, int to)
{
	save_var narrow_start = from;
	save_var narrow_end = size() - to;
	point = 0;
	switch (type) {
		case HT_JSCRIPT:
			tag_mode_jscript();
			break;
		case HT_VBSCRIPT:
			tag_mode_vbasic();
			break;
		case HT_PHP:
			save_var sublanguage_code = SUBLANG_PHP;
			tag_mode_perl();
			break;
		case HT_PYTHON:
			tag_mode_python();
			break;
	}
}

// For each scripting block, tag its contents appropriately.
tag_mode_html()
{
	int from, to, type;
	recolor_buffer_range(0, size());	// Divide into scripting blocks.
	save_var point = 0;
	do {
		type = get_tagged_region("color-as-unit", point, &from, &to);
		html_tag_region(type, from, to);
		check_abort();
		point = to + 1;
	} while (point < size() - narrow_end);
}

tag_mode_xml()
{
	tag_mode_html();
}

tag_mode_php()
{
	tag_mode_html();
}

html_like_mode(int force_xml)
{
	mode_default_settings();
	sublanguage_code = 0;
	mode_keys = html_tab;		/* Use these keys. */
	if (force_xml || matches_at(0, 1, "<wspace>*%<%?xml"))
		sublanguage_code = SUBLANG_XML;
	if (sublanguage_code == SUBLANG_XML) {
		major_mode = strkeep("XML");
		compile_buffer_cmd = compile_xml_cmd;
	} else {
		major_mode = strkeep("Html");
		compile_buffer_cmd = compile_html_cmd;
	}

	strcpy(comment_start, "%<%!-- *");
	strcpy(comment_pattern, "%<%!--(.|<newline>)*--%><FirstEnd>");
	strcpy(comment_begin, "<!-- ");
	strcpy(comment_end, " -->");
	buffer_maybe_break_line = html_maybe_break_line;
	indenter = html_indenter;
	auto_indent = 1;
	mode_move_level = html_move_level;
	if (auto_show_html_delimiters)
		auto_show_matching_characters = html_auto_show_delim_chars;
	mode_auto_show_delimiter = html_auto_show_delimiter;
	recolor_range = color_html_range;	// set up coloring rules
	recolor_from_here = html_recolor_from_here;
	idle_coloring_size = 5000;
	coloring_flags |= COLOR_IGNORE_INDENT;
	if (want_code_coloring)		// maybe turn on coloring
		when_setting_want_code_coloring();
	try_calling("html-mode-hook");
	drop_all_colored_regions();
	make_mode();
}

command html_mode()
{
	html_like_mode(0);
}

command xml_mode()
{
	html_like_mode(1);
}


command php_mode()
{
	html_like_mode(0);
	compile_buffer_cmd = compile_php_cmd;
	major_mode = _php_mode_name;
	sublanguage_code = SUBLANG_PHP;
	auto_show_matching_characters = auto_show_php_delimiters
									? php_auto_show_delim_chars : 0;
	if (php_tab_override > 0)
		tab_size = php_tab_override;
	try_calling("php-mode-hook");
}

// Automatically go into HTML mode on .htm, .html files.

suffix_htm()
{
	html_mode();
}

suffix_html()
{
	html_mode();
}

suffix_shtml()
{
	html_mode();
}

// .cfm, .cfml, .cfc are HTML variants for ColdFusion by Allaire.

suffix_cfml()
{
	html_mode();
}

suffix_cfm()
{
	html_mode();
}

suffix_cfc()
{
	html_mode();
}

// .htx is an HTML variant for Microsoft's Internet Database Connector.

suffix_htx()
{
	html_mode();
}

// .asp, .asa, .aspx, .ascx are HTML variants for Microsoft's web server.

suffix_asp()
{
	html_mode();
}

suffix_asa()
{
	html_mode();
}

suffix_aspx()
{
	html_mode();
}

suffix_ascx()
{
	html_mode();
}

// Recognize XML and related extensions.

suffix_xml()
{
	xml_mode();
}

suffix_cdf()
{
	xml_mode();
}

suffix_osd()
{
	xml_mode();
}

suffix_htt()
{
	html_mode();
}

suffix_wml()
{
	xml_mode();
}

suffix_ui()			// Qt user interface file
{
	xml_mode();
}

suffix_xsl()		// XML style sheet
{
	xml_mode();
}

suffix_xst()		// XML script
{
	xml_mode();
}

suffix_xsd()		// W3 Schema for XML
{
	xml_mode();
}

suffix_jsp()		// Java server page
{
	html_mode();
}

suffix_php()		// PHP server-side page
{
	php_mode();
}

suffix_php3()		// PHP server-side page
{
	php_mode();
}

suffix_php4()		// PHP server-side page
{
	php_mode();
}

suffix_sphp()		// PHP server-side page
{
	php_mode();
}

suffix_sgml()		// Generic SGML file.
{
	xml_mode();
}

suffix_sgm()
{
	xml_mode();
}

suffix_svg()
{
	xml_mode();
}

suffix_xmp()
{
	xml_mode();
}

suffix_rdf()
{
	xml_mode();
}

suffix_rss()
{
	xml_mode();
}

suffix_xsdconfig()
{
	xml_mode();
}

suffix_prx()
{
	html_mode();
}

suffix_asx()
{
	html_mode();
}

suffix_xaml()
{
	xml_mode();
}
