/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 2003, 2020 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// This file has commands that deal with customizing Epsilon.

#include "eel.h"

// Initialize buffer for customizations and return its number.

prep_customizations()
{
	int b;
	char file[FNAMELEN], *p = 0;

	if (!exist("-customizations"))	// Read the file 1st time, in case we
		p = lookpath(einit_file_name);		// didn't at start.
	save_var bufnum = b = create("-customizations");
	save_var use_default = 0;
	if (p) {
		do_file_read(p, default_translation_type);
		ini_mode();
	}
	if (!size()) {
		bprintf("# Epsilon customization file. -*- Mode: Conf -*-\n");
		bprintf("# This file lists your Epsilon customizations, such as\n");
		bprintf("# setting variables, colors, bindings and so forth.\n");
		bprintf("# Epsilon reads it at startup.\n\n");
		ini_mode();
	}
	if (!*filename) {
		get_customization_directory(file);
		strcat(file, einit_file_name);
		filename = file;
	}
	want_backups = 1;
	if (record_customizations == 2)
		save_without_prompt = 1;
	auto_read_changed_file = 1;
	if (!modified)
		check_dates(0);
	point = size();
	return b;
}

// Should we record customizations?

int want_record_customizations()
{
	return !noinit_flag && record_customizations && !now_parsing_cmds;
}

// Save the current setting of name table index i as a customization.

save_customization(int i)
{
	if (want_record_customizations())
		list_one(prep_customizations(), i, 0);
}


// Convert this buffer from a raw list of EEL file names into commands
// to load those files.
fix_extension_list(int buf)
{
	save_var bufnum = buf;
	point = 0;
	do_drop_matching_lines(REGEX, "[bv][0-9]+patch%.e$", 1);
	string_replace("[\\\"]", "\\#0", REGEX);
	string_replace("^.+$", "(load-eel-from-path \"#0\" 2)", REGEX);
}

// Scan the rest of the current buffer for EEL-created things that
// couldn't be defined.  If there are any, remove them from the buffer,
// show them to the user, and return nonzero.  If none, return zero.
list_undefined_customizations()
{
	int list = 0;
	char funcname[FNAMELEN];

	while (re_search(1, "^([^ \n\t:]+): (command|subroutine)\n")) {
		grab(find_group(1, 1), find_group(1, 0), funcname);
		if (!is_runnable(funcname)) {
			if (!list) {
				list = zap("-unknown-customizations");
				save_var bufnum = list;
				bprintf("Epsilon couldn't find definitions for the functions "
						"listed below.\nIf your previous setup included any EEL "
						"extension files, you\nmay want to rerun this command "
						"and include their names.\n\n");
				restore_vars();
			}
			buf_xfer(list, matchstart, matchend);
		}
		delete(matchstart, matchend);
	}
	if (list) {
		set_buf_point(list, 0);
		view_buf(list, 1);
	}
	return list != 0;
}

// Run a program with a list of arguments as a string.  Under Unix, run
// via the shell so it can split up the arguments into an argv array.
// Under Windows, don't use the shell so we don't get a command window
// if the program's GUI-based.  Return the result code from shell().

int shell_with_args(char *cmd, char *args)
{
	return shell_with_args_flags(cmd, args, SHELL_KEEP_ENV);
}

shell_with_args_flags(char *cmd, char *args, int flags)
{
	char full_cmd[1000];
	int err;

	if (is_unix) {	// Run via the shell, so it splits the args at whitespace.
		quote_for_shell(full_cmd, cmd);
		strcat(full_cmd, " ");
		strcat(full_cmd, args);
		err = shell("", full_cmd, "", flags);
	} else		// Run without the shell, so there's no window.
		err = shell(cmd, args, "", flags);
	build_first = 1;
	screen_messed();
	return err;
}

// Comment out old customizations before this position.
comment_old_customizations(int pos)
{
	save_var narrow_end = size() - pos;
	point = 0;		// First delete previous batch of cleanings.
	simple_re_replace(0, "^![-(+&~]", "### ");
}

// Some variables are normally saved via a session file, not an
// einit.ecm or state file.  Immediately apply those here.

custom_apply_immediately()
{
	int tmp = tmp_buf();
	on_exit buf_delete(tmp);
	buf_xfer(tmp, 0, size());
	save_var bufnum = tmp, show_status = 0;
	point = 0;
	do_drop_matching_lines(REGEX, "^%+(push-cmd|_region_file|_grep_files) ", 0);
	save_var _abort_load_on_error = 2;
	parse_cmds();
}

custom_fix_mode_vars()
{
	char start[FNAMELEN], end[FNAMELEN], *s;
	save_var point = 0;
	*start = *end = 0;
	if (re_search(1, "^%+mode-start .*: [0-9]+ \"(.*)\"$"))
		grab(find_group(1, 1), find_group(1, 0), start);
	point = 0;
	if (re_search(1, "^%+mode-end .*: [0-9]+ \"(.*)\"$"))
		grab(find_group(1, 1), find_group(1, 0), end);
	s = strstr(start, "\\z");
	if (s)
		*s = 0;
	s = strstr(end, "\\z");
	if (s)
		*s = 0;
	point = size();
	if (*start || *end)
		bprintf("+mode-format char array variable: %d \"%s%%b%%E%%M%s\"\n",
				ptrlen(mode_format), *start ? start : " ",
				*end ? end : " %d%p %s%f");
}

// Each key binding in src that doesn't match any in current buffer
// indicates the user unbound that key.  Insert a setting for that in
// current buffer.
insert_key_unbindings(int src)
{
	char keyspec[FNAMELEN], pat[FNAMELEN];
	int orig = point, obuf = bufnum;

	save_var point = 0, bufnum = src, point = 0;
	while (re_search(1, "^(~<^wspace>+ \".*\": )")) {
		grab(matchstart, matchend, keyspec);
		bufnum = obuf;
		point = orig;
		sprintf(pat, "^%q", keyspec);
		if (!re_search(1, pat))
			bprintf("%s0\n", keyspec);
		bufnum = src;
	}
}

// Position pos in buf is the start of a binding in .ecm format.  Fill
// start, end, text with its settings and return 1, or 0 if bad format.
int get_binding_range(int buf, int pos, int *start, int *end, char *text)
{
	save_var bufnum = buf, point = pos;
	if (!parse_string(1, "~[^ \t\n:]+ "))
		return 0;
	point = matchend;
	int *kcode = get_keycode();
	if (!kcode || kcode[0] != 2)
		return 0;
	*start = *end = kcode[1];
	if (parse_string(1, " \"")) {	// It uses a range spec.
		point++;
		kcode = get_keycode();
		if (!kcode || kcode[0] != 2)
			return 0;
		*end = kcode[1];
	}
	return parse_string(1, ": .*", text) != 0;
}

// Search buffer buf for an .ecm-format binding line that starts with
// init, skipping over those binding only keys < start.  Locate the one
// with the lowest-numbered range of keys.  Fill in its brange, erange,
// and def, and return 1.  If none, return 0.
find_lowest_binding_range(int buf, char *init, int start,
						  int *brange, int *erange, char *def)
{
	int bpos, epos, found = 0;
	save_var bufnum = buf, point = 0;
	char pat[FNAMELEN], text[FNAMELEN];
	sprintf(pat, "^%q", init);
	while (re_search(1, pat)) {
		if (get_binding_range(bufnum, give_begin_line(), &bpos, &epos, text)) {
			if (epos > start && (!found || bpos < *brange)) {
				found = 1;
				*brange = bpos;
				*erange = epos;
				strcpy(def, text);
			}
		}
	}
	return found;
}

// Insert in buffer buf a new binding line, binding keys from start to
// end, with text before and after.  If s1 > e1, ignore.
insert_new_binding_range(int buf, int start, int end,
						 char *before, char *after)
{
	int b[10];
	if (start > end)
		return;
	save_var bufnum = buf;
	bprintf("%s", before);
	b[0] = 2;
	b[1] = start;
	stuff_macro(b, 1);
	if (end > start) {
		bprintf(" ");
		b[1] = end;
		stuff_macro(b, 1);
	}
	bprintf("%s\n", after);
}

// Look through all multiple-key bindings in .ecm-format buffer b1.  For
// each, see if b2 has any corresponding binding lines that intersect
// b1's range but aren't identical.  If so, split up b1's range into a
// series of ranges, so each range is completely inside or outside b2's
// ranges.  This rewriting lets us compare binding changes line-by-line.
scan_binding_ranges(int b1, int b2)
{
	char init[FNAMELEN], def1[FNAMELEN], def2[FNAMELEN];
	int s1, e1, s2, e2;
	save_var bufnum = b1;
	point = 0;
	while (re_search(1, "~[^ \t\n:]+ !.*\" \".*\": ")) {
		grab(matchstart, matchend, init);
		int *kcode = get_keycode();
		if (!kcode || kcode[0] != 2)
			continue;
		s1 = kcode[1];
		if (!parse_string(1, " \""))
			continue;
		point++;
		kcode = get_keycode();
		if (!kcode || kcode[0] != 2)
			continue;
		e1 = kcode[1];
		if (!parse_string(1, ": .*", def1))
			continue;
		if (!find_lowest_binding_range(b2, init, s1, &s2, &e2, def2))
			continue;
		if (s1 >= s2 && e1 <= e2)	// This range is contained within other.
			continue;
		if ((s1 != s2 || e1 != e2) && s2 <= e1) {
			delete(give_begin_line(), give_end_line() + 1);
			int pos = point;
			insert_new_binding_range(b1, s1, s2 - 1, init, def1);
			insert_new_binding_range(b1, MAX(s1, s2), MIN(e1, e2), init, def1);
			insert_new_binding_range(b1, e2 + 1, e1, init, def1);
			point = pos;
		}
	}
}

// Return a buffer listing the differences in settings between orig and
// cur, two buffers with lists of settings.  Put the original settings
// corresponding to the modified settings into a buffer
// "original-settings".
int make_difference_list(int orig, int cur)
{
	noteput("Comparing...");
	save_var show_status = 0;
	scan_binding_ranges(orig, cur);
	scan_binding_ranges(cur, orig);
	save_var bufnum = orig;
	save_var case_fold = 1;
	do_sort_region(0, size(), 0, 0);
	save_var bufnum = cur;
	save_var case_fold = 1;
	do_sort_region(0, size(), 0, 0);
	int out = tmp_buf();
	do_compare_sorted(orig, cur, "original-settings",
					  bufnum_to_name(out), NULL);
	bufnum = out;
	point = 0;
	save_var case_fold = 1;
	custom_apply_immediately();
	// Drop vars starting with -, but not color scheme ones.
	do_drop_matching_lines(REGEX, "^%+-[^o][^u][^r][^-]", 1);
	drop_boring_change_lines();
	custom_fix_mode_vars();
	insert_key_unbindings(name_to_bufnum("original-settings"));
	do_drop_matching_lines(REGEX, "^%+(push-cmd|echo-line|permanent-menu"
						   "|avoid-(top|bottom)-lines|bufed-width"
						   "|mode-start|mode-end) ", 1);
	do_drop_matching_lines(REGEX, "^&.* color class for after-exiting: .*$", 1);
	do_drop_matching_lines(REGEX, "^&.* color class for .*: "
						   "%[0x4000000 on 0x0%]$", 1);
	do_drop_matching_lines(REGEX, "^~.*: case-indirect$", 1);
	// We might have original.sta from 9.0 with epsilon.sta from 9.02,
	// due to 9.02 updater failing to update original.sta.  Ignore
	// resulting differences.  We ignore the latex setting only if it
	// looks like this is a 9.0/9.02 update.
	if (do_drop_matching_lines(REGEX, "^(fix-buf-accessed-clock|fix-tex-math-coloring"
						   "|http-activity|tex-start-construct): subroutine$", 1))
		do_drop_matching_lines(REGEX, "^%+latex-2e-or-3 char variable: 0$", 1);
	do_drop_matching_lines(REGEX, "^%+ftp-passive-transfers char variable: 1$", 1);

	return out;		// Don't count vars saved in session file, others.
}

// Check each line in the range to see if it names an existing file.
// Return 0 if everything's OK, otherwise ask the user if he wants to
// try again and return 1 if yes, 2 to give up.
test_valid_files(int start, int end)
{
	char file[FNAMELEN], msg[FNAMELEN], *s;
	int ret = 0;

	save_var point = start, narrow_end = size() - end;
	while (!ret && grab_line(bufnum, file)) {
		if (s = strstr(file, " (not found)"))
			*s = 0;
		if (check_file(file) != CHECK_FILE)
			ret = 1;
		else if (is_relative(file))
			ret = 3;
	}
	if (!ret)
		return ret;
	if (ret == 3)
		sprintf(msg, "\"%s\" not an absolute path, try again? ", file);
	else
		sprintf(msg, "File \"%s\" not found, try again? ", file);
	return ask_yn("Try again?", msg, "Yes", "No", 1) ? 1 : 2;
}

// Run Epsilon with its original state file, sending it a list of EEL
// extensions to load, and have it generate a settings file.  Then
// compare to the current settings.

command list_customizations()
{
	char cmd[FNAMELEN], *args, time[FNAMELEN];
	char ext_file[FNAMELEN], settings_file[FNAMELEN], *s;
	int tmp, cur, ext, i, err;
	int cust = prep_customizations(), clist;

	iter = 1;
	if (!has_arg)
		clean_customizations();
	tiled_only();
	get_customization_directory(ext_file);
	strcat(ext_file, "extensions-list");
	get_customization_directory(settings_file);
	strcat(settings_file, "settings-list");
	delete_file(settings_file);
	i = name_match("-loaded-eel-file-", STARTMATCH);
	save_var bufnum = ext = zap("-list-customizations");
	give_binding(cmd, "exit-level");
	bprintf("Epsilon will include the following EEL extension files as part\n"
			"of your customizations.  Edit this list if needed, then\n"
			"press %s to begin.\n\n%p\n", cmd, color_class viewed_text);
	point--;
	for (; i; i = name_match("", 0))	// List new extensions by searching
		if (name_type(i) == NT_VAR) {	// for id variables.
			s = get_str_var(i);
			bprintf("%s%s\n", s, (check_file(s) == CHECK_FILE)
					? "" : " (not found)");
		}
	to_buffer_num(ext);
	point = 0;
	do {
		recursive_edit();
		if (!buf_exist(ext))
			return;
		point = 0;
		point += parse_string(1, "(.|\n.)*\n*");
		i = test_valid_files(point, size());
		if (i == 2) {
			buf_zap(ext);
			return;
		}
	} while (i);
	point = 0;		// Drop the explanatory text.
	delete(0, parse_string(1, "(.|\n.)*\n*"));
	simple_re_replace(1, " *%(not found%)", "");
	if (err = file_write(ext_file, FILETYPE_BINARY)) {
		file_error(err, ext_file, "write error");
		return;
	}
	save_var reject_client_connections = 1;
	strcpy(cmd, argv[0]);
	convert_to_8_3_filename(cmd);
	sayput("Running %s to create original settings list...", cmd);
	args = " -s original -p0 -noinit -r build-settings-file";
	err = shell_with_args(cmd, args);
	if (err)
		error("Error running %s %s", cmd, args);
	sayput("Reading original settings list...");
	bufnum = tmp = tmp_buf();
	if (err_file_read(settings_file, FILETYPE_BINARY))
		return;
	bufnum = cur = tmp_buf();
	list_all_to_buffer(cur, 3);
	clist = make_difference_list(tmp, cur);
	restore_vars();
	buf_delete(tmp);
	buf_delete(cur);
	to_buffer_num(cust);
	current_time_and_date(time);
	save_spot point;
	i = point;
	bprintf("\n# List-customizations created these settings %s.\n", time);
	fix_extension_list(ext);
	grab_buffer(ext);
	grab_buffer(clist);
	buf_delete(clist);
	delete_file(settings_file);
	delete_file(ext_file);
	bprintf("\n");
	point = i;
	if (!has_arg)
		comment_old_customizations(point);
	if (!list_undefined_customizations())
		say("");
}

// Read extensions file, load those extensions, then write a list of
// settings.

build_settings_file()
{
	char ext_file[FNAMELEN], settings_file[FNAMELEN], eel_file[FNAMELEN];
	int tmp = tmp_buf(), err;

	get_customization_directory(ext_file);
	strcat(ext_file, "extensions-list");
	get_customization_directory(settings_file);
	strcat(settings_file, "settings-list");
	save_var bufnum = tmp;
	if (err_file_read(ext_file, FILETYPE_BINARY)) {
		show_text(0, 300, "Can't read extensions list file, aborting.");
		leave(1);
	}
	point = 0;
	while (grab_line(bufnum, eel_file)) {
		sayput("Loading %f...", eel_file);
		if (load_eel_from_path(eel_file, 2))
			return;
	}
	buf_zap(bufnum = tmp);
	sayput("Listing original settings...");
	list_all_to_buffer(tmp, 3);
	if (err = file_write(settings_file, FILETYPE_BINARY)) {
		file_error(err, settings_file, "write error");
		delete_file(settings_file);	// Ensure caller notices failure.
	}
	say("");
	leave(0);
}

// Parse a paren command in a command file buffer and move past it.
skip_paren_cmd()
{
	point++;		// Pass ( char.
	point += parse_string(1, "([^()'\"]|'.'|'\\.'|\"([^\"]|\\<any>)*\")*");
	if (curchar() != ')')
		error("Can't parse this command.");
	point++;	// Pass ) char.
}

// Extract name of a thing from a regex match, then return nonzero if
// there are future definitions of it.  To check, build a pattern using
// the rule, substituting the match for <> in the rule.

int check_dup_customization(int group, char *rule)
{
	char thing[FNAMELEN], pat[FNAMELEN];

	grab(find_group(group, 1), find_group(group, 0), thing);
	strcpy(pat, rule);
	sprintf(strstr(pat, "<>"), "%q", thing);
	strcat(pat, strstr(rule, "<>") + 2);	// Add the part after the <>.
	save_var point;
	to_end_line();
	return re_search(1, pat);
}

command edit_customizations()
{
	iter = 0;
	tiled_only();
	to_buffer_num(prep_customizations());
}

command clean_customizations()
{
	int dup;

	iter = 0;
	tiled_only();
	to_buffer_num(prep_customizations());
	point = 0;		// First delete previous batch of cleanings.
	simple_re_replace(0, "###.*\n", "");
	point = 0;
	while (point < size() - narrow_end) {
		dup = 0;
		if (parse_string(1, "^(%+<word|dash>+ )"))
			dup = check_dup_customization(1, "^<>");
		else if (parse_string(1, "^(%&<word|dash>+ color class "
							  "for <word|dash>+: )"))
			dup = check_dup_customization(1, "^<>");
		else if (parse_string(1, "^%(bind-to-key \"[^\"\n]+\" (\"[^\"\n]+\")%)"))
			dup = check_dup_customization(1, "^%(bind-to-key \"[^\"\n]+\" <>%)");
		else if (parse_string(1, "^(<word|dash|#128-#255>+: macro )"))
			dup = check_dup_customization(1, "^<>");
		else if (parse_string(1, "^(%(load-eel-from-path \"[^\"\n]+\" 2%)$)"))
			dup = check_dup_customization(1, "^<>");
		else if (parse_string(1, "^(~<^wspace>+ \".*\": )"))
			dup = check_dup_customization(1, "^<>");
		if (dup) {
			stuff("### ");
			nl_forward();
			continue;
		}
		switch (curchar()) {
			case '(':	skip_paren_cmd(); break;
			case ';':
			case '#':	nl_forward(); break;
			default:
				nl_forward();
				break;
		}
	}
}

// Compile the eel file src into the directory dest using the compiler
// pathname eel, and write the resulting .b file name into the buffer
// list (if nonzero).  If src contains a series of newline-separated
// names, do this on each.

compile_for_import_customizations(char *src, char *eel, char *dest, int list)
{
	char buf[2000], *s, cmd[FNAMELEN], args[FNAMELEN], rdest[FNAMELEN], newsrc[FNAMELEN];

	for (;;) {
		strcpy(rdest, dest);
		convert_to_8_3_filename(rdest);
		strcpy(buf, src);
		if (!*buf)
			return;
		if (s = strchr(buf, '\n'))
			*s = 0;
		if (s = strstr(buf, " (not found)"))
			*s = 0;
		convert_to_8_3_filename(buf);
		if (check_file(buf) != CHECK_FILE)
			error("Can't find %s", buf);
		strcpy(newsrc, rdest);
		sprintf(newsrc + strlen(newsrc), "%c%s", path_sep, get_tail(buf, 0));
		copyfile(buf, newsrc);
		strcpy(cmd, eel);
		convert_to_8_3_filename(cmd);
		sprintf(rdest + strlen(rdest), "%c%s", path_sep, get_tail(buf, 0));
		strcpy(get_extension(rdest), byte_extension);
		delete_file(rdest);
		sprintf(args, " -o %s %s", rdest, newsrc);
		show_text(-2, 1, "Run %s %s", cmd, args);
		if (shell_with_args(cmd, args) || check_file(rdest) != CHECK_FILE)
			error("Error compiling %s with cmd line %s %s", buf, cmd, args);
		if (list)
			buf_printf(list, "%s\n", rdest);
		if (s = strchr(src, '\n'))
			src = s + 1;
		else
			break;
	}
}

// Delete each color-setting line in newbuf that's also in oldbuf.
import_remove_unchanged_colors(int oldbuf)
{
	char newline[FNAMELEN];
	save_var bufnum;
	
	point = 0;
	set_buf_point(oldbuf, 0);
	while (point < size() && parse_string(1, ".*\n", newline)) {
		if (*newline == '\t' && count_matches_in_buffer(oldbuf, newline)) {
			point--;
			delete(give_begin_line(), give_end_line() + 1);
		}
	}
}

command import_customizations()
{
	int tmp = zap("import-customizations"), i, ver = 0, err, clean = !has_arg;
	int debug = (iter == 99);
	char buf[2000], maindir[FNAMELEN], bindir[FNAMELEN];
	char origsta[FNAMELEN], cursta[FNAMELEN], *s, *exe_ext = "", pat[FNAMELEN];
	char answer[6][2000], ranswer[6][2000], oldname[FNAMELEN];
	char origdir[FNAMELEN], *epsout = lookpath("epsout.e");

	iter = 1;
	tiled_only();
	getcd(origdir);
	on_exit chdir(origdir);
	putenv("EPSRUNS");		// Prevent error msg in old Epsilons.
	save_var bufnum;
	if (!epsout)
		error("The file epsout.e is missing.");
	for (i = EELVERSION / 10 - 1; i >= 8 ; i--) {
		if (is_mswindows) {
			sprintf(buf, "EPSPATH%02d", (i < 10) ? (i * 10) : i);
			if (!verenv(buf))
				continue;
		} else {
			sprintf(buf, "~/.epsilon/epsilon-v%d.sta", i);
			if (!check_file(buf))
				continue;
		}
		buf_printf(tmp, "%d\n", ver = i);
	}
	set_buf_point(tmp, 0);
	switch (count_lines_in_buf(tmp, 0)) {
		case 0:		ver = 0; break;
		case 1:		break;
		default:
			*buf = 0;
			buf_delete_final_newlines(tmp);
			if (!get_choice(tmp, buf, " Update from version ",
							"Update from which version? ",
							"OK", "Cancel", ""))
				return;
			ver = numtoi(buf);
			say("");
	}
	*maindir = *bindir = *origsta = *cursta = 0;
	if (is_mswindows && ver) {
		exe_ext = ".exe";
		sprintf(buf, "EPSPATH%02d", (ver < 10) ? (ver * 10) : ver);
		if (s = verenv(buf)) {
			strcpy(maindir, s);
			while (s = strchr(maindir, ';')) {
				*s = 0;
				sprintf(origsta, "%s%coriginal.sta", maindir, path_sep);
				if (check_file(origsta) == CHECK_FILE)
					break;
				strcpy(maindir, s + 1);
			}
			absolute(maindir);
			sprintf(origsta, "%s%coriginal.sta", maindir, path_sep);
		}
		sprintf(buf, "BINPATH%02d", (ver < 10) ? (ver * 10) : ver);
		if (*maindir)
			sprintf(bindir, "%s\\bin", maindir);
		s = verenv(buf);
		if (s)
			strcpy(bindir, s);
		absolute(bindir);
		if (*maindir)
			sprintf(cursta, "%s\\epsilon.sta", maindir);
		if (ver >= 13) {
			get_customization_directory(cursta);
			sprintf(cursta + strlen(cursta), "\\epsilon-v%d.sta", ver);
		}
	} else if (ver) {		// Set vars for Unix.
		if (is_unix_flavor != IS_UNIX_MACOS)
			sprintf(buf, "/{usr/local,opt}/epsilon%d*/bin/epsilon", ver);
		else if (ver == 13)
			sprintf(buf, "/Applications/Eps*app/Contents/bin/epsilon-dialogs");
		else
			sprintf(buf, "/Applications/Eps*app/Contents/Misc/version%d", ver);
		int max = 0;		// Find newest match of pattern.
		struct file_info fi;
		s = file_match(buf, STARTMATCH);
		*buf = 0;
		while (s && check_file(s, &fi)) {
			if (!max || fi.raw_file_date > max) {
				strcpy(buf, s);
				max = fi.raw_file_date;
			}
			s = file_match("", 0);
		}
		if (*buf) {
			get_tail(buf, 0)[-1] = 0;
			strcpy(bindir, buf);
			strcpy(origsta, buf);
			if (is_unix_flavor == IS_UNIX_MACOS && ver > 13) {
				strcat(bindir, "/../MacOS");
				strcat(origsta, "/X");	// Don't go up a level, overwrite the X
			}
			strcpy(get_tail(origsta, 0), "original.sta");
		}
		sprintf(cursta, "~/.epsilon/epsilon-v%d.sta", ver);
	}
	save_var window_bufnum;
	to_buffer_num(tmp);
	buf_zap(bufnum);
	sprintf(oldname, "Epsilon %d", ver);
	if (!ver)
		strcpy(oldname, "the old version");
	bprintf("Epsilon will try to import any customizations you saved in your\n");
	bprintf("state file in %s.  Please correct any of\n", oldname);
	give_binding(buf, "exit-level");
	bprintf("these settings as necessary, then press %s\n", buf);
	bprintf("to begin the import process.\n\n");

	if (!ver) {
		bprintf("Epsilon couldn't find any %s,\n", is_mswindows
				? "Epsilon registry entries"
				: "customized state file in ~/.epsilon");
		bprintf("so you'll have to fill in this information.\n\n");
	}
	bprintf("1. Your customized state file for %s is:\n%p%s\n%p\n",
			oldname, color_class viewed_text,
			*cursta ? cursta : "full path for your .sta file", -1);
	if (*cursta && check_file(cursta) != CHECK_FILE)
		bprintf("(This file doesn't exist.)\n\n");
	bprintf("2. The original pre-customization state file for %s\n", oldname);
	bprintf("is:\n%p%s\n%p\n",
			color_class viewed_text,
			*origsta ? origsta : "full path for original.sta", -1);
	if (*origsta)
		if (check_file(origsta) != CHECK_FILE)
			bprintf("(This file doesn't exist.)\n\n");
	bprintf("3. If you customized %s by loading EEL files,\n", oldname);
	bprintf("please list each .e file's pathname here, one per line:\n%p\n%p\n",
			color_class viewed_text, -1);
	sprintf(buf, "full path for epsilon%s", exe_ext);
	if (*bindir)
		sprintf(buf, "%s%c%s%s", bindir, path_sep,
				(ver < 10) ? "e32" : "epsilon", exe_ext);
	bprintf("4. The Epsilon%s executable file for %s is:\n%p%s\n%p\n",
			is_mswindows ? " (GUI, not console version)" : "", oldname,
			color_class viewed_text, buf, -1);
	if (*bindir && check_file(buf) != CHECK_FILE)
		bprintf("(This file doesn't exist.)\n\n");
	sprintf(buf, "full path for eel%s", exe_ext);
	if (*bindir)
		sprintf(buf, "%s%ceel%s", bindir, path_sep, exe_ext);
	bprintf("5. The EEL compiler executable for %s is:\n%p%s\n%p\n",
			oldname, color_class viewed_text, buf, -1);
	if (*bindir && check_file(buf) != CHECK_FILE)
		bprintf("(This file doesn't exist.)\n\n");
	get_customization_directory(buf);
	strcat(buf, "upd");
	mkdir(buf);
	bprintf("6. Use this working directory for intermediate files:\n%p%s\n%p\n",
			color_class viewed_text, buf, -1);
	if (buf && check_file(buf) != CHECK_DIR)
		bprintf("(Warning: Couldn't create this directory.)\n\n");
	for (;;) {
		err = point = 0;
		recursive_edit();
		if (!buf_exist(tmp))
			return;
		bufnum = tmp;
		point = 0;		// Now retrieve the answers.
		for (i = 0; i < 6; i++) {
			sprintf(pat, "^%d%%. [^:]+:\n((.|\n.)*)\n\n", i + 1);
			if (!re_search(1, pat)) {
				say("Can't interpret response %d", i + 1);
				err = 1;
				break;
			}
			grab(find_group(1, 1), find_group(1, 0), answer[i]);	// 8.3 version.
			grab(find_group(1, 1), find_group(1, 0), ranswer[i]);	// For display.
			if (i == 2) {
				err = test_valid_files(find_group(1, 1), find_group(1, 0));
				if (err == 2)
					quick_abort();
				err = err ? 3 : 0;	// Don't prompt again.
			} else if (check_file(answer[i]) != ((i == 5)
				? CHECK_DIR : CHECK_FILE)) {
				say("Response %d: Can't read %s \"%f\".",
					  i + 1, (i == 5) ? "directory" : "file", answer[i]);
				err = 1;
				break;
			} else
				convert_to_8_3_filename(answer[i]);
		}
		if (!err)
			break;
		if (err != 3
			  && !ask_yn("Try again", "Would you like to correct your response? ",
						 "Yes", "No", 0))
			quick_abort();
	}
	if (chdir(answer[5]))
		error("Couldn't set directory to %f.", answer[5]);
	buf_zap(tmp);	// Compile epsout & extensions, then save list of them.
	compile_for_import_customizations(epsout, answer[4], answer[5], 0);
	compile_for_import_customizations(answer[2], answer[4], answer[5], tmp);
	if (err = file_write("extensions-list", FILETYPE_BINARY)) {
		file_error(err, "extensions-list", "write error");
		return;
	}

	// Run the old version with orig state file.
	save_var reject_client_connections = 1;
	sprintf(buf, " \"-s%s\" -l%s%cepsout.b -rexport_current_colors -r%sbuild_settings_file",
			answer[1], answer[5], path_sep, debug ? "debug_" : "");
	show_text(-2, 1, "Run %s %s", answer[3], buf);
	shell_with_args(answer[3], buf);
	rename_file("mycolors.e", "mycolors-old.e");
	buf_zap(tmp);
	if (err = file_write("extensions-list", FILETYPE_BINARY)) {
		file_error(err, "extensions-list", "write error");
		return;
	}
	if (err_file_read("settings-list", FILETYPE_BINARY))
		return;
	// Run the old version with new state file.
	sprintf(buf, " \"-s%s\" -l%s%cepsout.b -rexport_current_colors -r%sbuild_settings_file",
			answer[0], answer[5], path_sep, debug ? "debug_" : "");
	show_text(-2, 1, "Run %s %s", answer[3], buf);
	shell_with_args(answer[3], buf);
	int cur = tmp_buf();
	on_exit buf_delete(cur);
	bufnum = cur;
	if (err_file_read("settings-list", FILETYPE_BINARY))
		return;
	int clist = make_difference_list(tmp, cur);
	on_exit buf_delete(clist);	// Now format the list of differences.
	if (clean)
		clean_customizations();
	int cust = prep_customizations();
	bufnum = cust;
	current_time_and_date(buf);
	save_var point;
	i = point;
	bprintf("\n# Import-customizations created these settings %s\n", buf);
	bprintf("# from %s.\n", ranswer[0]);
	int ext = tmp_buf();
	on_exit buf_delete(ext);
	buf_printf(ext, "%s\n", ranswer[2]);
	fix_extension_list(ext);
	grab_buffer(ext);
	grab_buffer(clist);
	bprintf("\n");
	int custom_colors = 0;
	if (check_file("mycolors.e")) {
		get_customization_directory(buf);
		strcat(buf, "mycolors.e");
		get_customization_directory(oldname);
		strcat(oldname, "backup-mycolors.e");
		rename_file(buf, oldname);
		int old_colors = tmp_buf();
		save_var bufnum = old_colors;
		if (err_file_read("mycolors-old.e", FILETYPE_AUTO))
			say("Error reading mycolors-old.e");
		save_var bufnum = tmp_buf();
		if (err_file_read("mycolors.e", FILETYPE_AUTO))
			say("Error reading mycolors.e");
		else {
			import_remove_unchanged_colors(old_colors);
			substitute_color_names();
			if (count_matches_in_buffer(bufnum, "\t")) {
				file_write(buf, FILETYPE_AUTO);
				custom_colors = 1;
			}
		}
	}
	restore_vars();
	to_buffer_num(cust);
	point = size();
	if (custom_colors)
		bprintf("(load-eel-from-path \"mycolors.e\" 2)\n");
	point = i;
	if (clean)
		comment_old_customizations(point);
	list_undefined_customizations();
	say("Done.  Save file, then restart Epsilon, to use these new settings.");
}
