/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1996, 2020 Lugaru Software Ltd.  All rights reserved.  *
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

#include "eel.h"
#include "proc.h"
#include "colcode.h"
#include "misclang.h"
#include "c.h"

// Move past a string in PostScript.

fwd_postscript_string()
{
	int level = 0;

	while (re_search(1, "[\\()]")) {
		switch (character(point - 1)) {
			case '\\':		// Quote next char.
				point++;
				break;
			case '(':		// Ignore balanced parens within.
				level++;
				break;
			case ')':
				if (--level <= 0)
					return 1;
				break;
		}
	}
	return 0;
}

color_postscript_range(from, to)
{
	int t = -1, talk, r, talk_at = 0;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend, case_fold = 1;
	set_character_color(from, to, -1);
	point = from;
	to_begin_line();	// Move back over continued lines.
	point -= parse_string(-1, "(.*\\\n)+");
	while (point > narrow_start
		   && get_tagged_region("color-as-unit", point - 1, &r) != -1)
		point = r;		// Move before binary blocks.
	talk = (to - point > 1000000);	// Show status during long delays.
	while (point < to) {
		if (!re_search(1, "[%(]|/[-_a-z0-9]+|</word>(doclutimage|doNimage)</word>")) {
			t = size();
			break;
		}
		t = matchstart;
		switch (character(t)) {		// check first char
			case '%':
				nl_forward();
				set_character_color(t, point, color_class postscript_comment);
				if (parse_string(-1, "^%%%%Begin(Binary|Data).*\n")) {
					r = point;
					if (re_search(1, "\n%%%%End(Binary|Data)"))
						point = matchstart;
					set_character_color(r, point, color_class postscript_text);
					set_tagged_region("color-as-unit", t, point, 1);
				}
				break;
			case '(':		// A text string.
				point--;
				fwd_postscript_string();
				set_character_color(t, point, color_class postscript_text);
				break;
			case '/':		// Found /literalname.
				set_character_color(t, point, color_class postscript_literal);
				break;
			default:		// Found start of Windows PS driver binary code.
				re_search(1, "~>[ \n]*(Z|endimage)");	// Find end.
				set_character_color(t, point, color_class postscript_text);
				set_tagged_region("color-as-unit", t, point, 1);
				break;
		}
		if (talk && point > talk_at + 20000) {
			note("Coloring PostScript: %d%% complete...",
				 muldiv(point - from, 100, to - from));
			talk_at = point;
		}
	}
	if (to < t)
		set_character_color(to, t, -1);
	if (talk_at)
		note("");
	return point;
}

command postscript_mode()
{
	mode_default_settings();
	major_mode = _postscript_mode_name;
	mode_keys = pscript_tab;		/* Use these keys. */
	if (auto_show_postscript_delimiters)
		auto_show_matching_characters = postscript_auto_show_delim_chars;
	strcpy(comment_start, "%%[ \t]*");
	strcpy(comment_pattern, "%%.*$");
	strcpy(comment_begin, "% ");
	strcpy(comment_end, "");
	comment_column = 0;
	recolor_range = color_postscript_range;	// set up coloring rules
	recolor_from_here = recolor_from_top;
	coloring_flags |= COLOR_INVALIDATE_FORWARD | COLOR_INVALIDATE_RESETS;
	idle_coloring_size = 10000;
	if (want_code_coloring)		// maybe turn on coloring
		when_setting_want_code_coloring();
	buffer_maybe_break_line = generic_maybe_break_line;
	fill_mode = (misc_language_fill_mode & 1) != 0;
	try_calling("postscript-mode-hook");
	make_mode();
}

when_loading()
{
	pscript_tab[ALT('q')] = (short) fill_comment;
	pscript_tab[ALT('Q')] = (short) prefix_fill_paragraph;
}

suffix_ps()
{
	postscript_mode();
}

suffix_eps()
{
	postscript_mode();
}

color_conf_range(from, to)
{
	int t = -1;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend;
	point = to;				// Color entire lines.
	nl_forward();
	to = point;
	save_var narrow_end = 0;
	set_character_color(from, to, -1);
	point = from;

	while (point < to) {
		if (!re_search(1, "([^ \t\n#:;,]+[ \t]*)="
					   "([ \t]*([^ \t\n#:;,]+|'[^\n']*'|\"[^\n\"]*\"))|^;|[#]")) {
			t = size();
			break;
		}
		t = matchstart;
		switch (character(point - 1)) {		// check last char
			case '#':
			case ';':
				nl_forward();
				set_character_color(t, point, color_class conf_comment);
				break;
			default:
				set_character_color(t, find_group(1, 0),
									color_class conf_attribute);
				set_character_color(find_group(2, 1), point,
									color_class conf_value);
				break;
		}
	}
	if (to < t)
		set_character_color(to, t, -1);
	return point;
}

command conf_mode()
{
	mode_default_settings();
	major_mode = _conf_mode_name;
	mode_keys = conf_tab;		/* Use these keys. */
	strcpy(comment_start, "#[ \t]*");
	strcpy(comment_pattern, "#.*$");
	strcpy(comment_begin, "# ");
	strcpy(comment_end, "");
	comment_column = 0;
	recolor_range = color_conf_range;	// set up coloring rules
	recolor_from_here = recolor_by_lines;
	idle_coloring_size = 10000;
	if (want_code_coloring)		// maybe turn on coloring
		when_setting_want_code_coloring();
	buffer_maybe_break_line = generic_maybe_break_line;
	fill_mode = (misc_language_fill_mode & 2) != 0;
	if (auto_show_conf_delimiters)
		auto_show_matching_characters = conf_auto_show_delim_chars;
	try_calling("conf-mode-hook");
	make_mode();
}

when_loading()
{
	conf_tab[ALT('q')] = (short) fill_comment;
	conf_tab[ALT('Q')] = (short) prefix_fill_paragraph;
}

suffix_conf()
{
	conf_mode();
}

// Return code for word from here to point (something with alpha or
// digits).  Return for 1 for command, 2 for comment, 0 if unknown.
is_batch_keyword(int from)
{
	char buf[500], *s;

	save_var case_fold = 1;
	if (point - from > sizeof(buf) / sizeof(char) - 10)
		save_var point = from + sizeof(buf) / sizeof(char) - 10;
	buf[0] = '|';		// get identifier, between | chars
	grab(from, point, buf + 1);
	if ((s = strchr(buf + 1, '.')) && matches_at(from, -1, "\n[ \t]*"))
		*s = 0;
	strcat(buf, "|");
	if (strstr("|rem|", buf) && !matches_at(from, -1, "[^ \n\t@][@ \t]*"))
		return 2;
	if (strstr("|assoc|at|attrib|break|cacls|call|cd|chcp|chdir"
			   "|chkdsk|chkntfs|cls|cmd|cmdextversion|color|comp"
			   "|compact|convert|copy|date|defined|del|dir|diskcomp"
			   "|diskcopy|do|doskey|echo|else|endlocal|equ|erase"
			   "|errorlevel|exist|exit|fc|find|findstr|for|format"
			   "|ftype|geq|goto|graftabl|gtr|help|if|in|label|leq"
			   "|lss|md|mkdir|mode|more|move|neq|not|path|pause"
			   "|popd|print|prompt|pushd|rd|recover|ren|rename"
			   "|replace|rmdir|set|setlocal|shift|sort|start|subst"
			   "|time|title|tree|type|ver|verify|vol|xcopy|cmd"
			   "|command|", buf))
			return 1;
	return 0;
}

color_batch_range(from, to)
{
	int t = -1;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend, case_fold = 1;
	point = to;				// Color entire lines.
	nl_forward();
	to = point;
	save_var narrow_end = 0;
	set_character_color(from, to, -1);
	point = from;

	while (point < to) {
		if (!re_search(1, "[-a-z0-9_.%/\\]+|[<>|]|^::")) {
			t = size();
			break;
		}
		t = matchstart;
		if (strchr("<>|", character(point - 1))) {
			set_character_color(t, point, color_class batch_punctuation);
			continue;
		} else if (character(point - 1) == ':') {
			nl_forward();
			set_character_color(t, point, color_class batch_comment);
			continue;
		}
		switch (is_batch_keyword(t)) {
			case 1:
				set_character_color(t, point, color_class batch_keyword);
				break;
			case 2:
				nl_forward();
				set_character_color(t, point, color_class batch_comment);
				break;
			default:
				set_character_color(t, point, -1);
				break;
		}
	}
	if (to < t)
		set_character_color(to, t, -1);
	return point;
}

command batch_mode()
{
	mode_default_settings();
	major_mode = _batch_mode_name;
	mode_keys = batch_tab;		/* Use these keys. */
	strcpy(comment_start, "rem[ \t]*");
	strcpy(comment_pattern, "rem.*$");
	strcpy(comment_begin, "rem ");
	strcpy(comment_end, "");
	comment_column = 0;
	recolor_range = color_batch_range;	// set up coloring rules
	recolor_from_here = recolor_by_lines;
	idle_coloring_size = 10000;
	if (want_code_coloring)		// maybe turn on coloring
		when_setting_want_code_coloring();
	buffer_maybe_break_line = generic_maybe_break_line;
	fill_mode = (misc_language_fill_mode & 4) != 0;
	if (auto_show_batch_delimiters)
		auto_show_matching_characters = batch_auto_show_delim_chars;
	try_calling("batch-mode-hook");
	make_mode();
}

when_loading()
{
	batch_tab[ALT('q')] = (short) fill_comment;
	batch_tab[ALT('Q')] = (short) prefix_fill_paragraph;
}

suffix_bat()
{
	batch_mode();
}

suffix_cmd()
{
	batch_mode();
}

suffix_btm()
{
	batch_mode();
}

is_nsis_keyword(int p)	// Is text at p a keyword?
{
	char buf[500];

	if (point - p > sizeof(buf) / sizeof(char) - 10)
		save_var point = p + sizeof(buf) / sizeof(char) - 10;
	buf[0] = '|';		// get identifier, between | chars
	grab(p, point, buf + 1);
	strcpy(buf + point - p + 1, "|");
	if (buf[0] == '!'
		  && strstr("|AddIncludeDir|AddPluginDir|appendfile|cd|define"
					"|delfile|echo|else|endif|error|execute|if|ifdef"
					"|ifmacrodef|ifmacrondef|ifndef|include|insertmacro"
					"|macro|macroend|packhdr|searchparse|searchreplace"
					"|system|tempfile|undef|verbose|warning|", buf + 1))
		return 1;
	if (strstr("|Abort|AddBrandingImage|AddSize|AllowRootDirInstall"
			   "|AllowSkipFiles|AutoCloseWindow|BGFont|BGGradient"
			   "|BrandingText|BringToFront|Call|CallInstDLL|Caption|ChangeUI"
			   "|CheckBitmap|ClearErrors|CompletedText|ComponentText"
			   "|CopyFiles|CRCCheck|CreateDirectory|CreateFont|CreateShortCut"
			   "|Delete|DeleteINISec|DeleteINIStr|DeleteRegKey|DeleteRegValue"
			   "|DetailPrint|DetailsButtonText|DirText|DirVar|DirVerify"
			   "|EnableWindow|EnumRegKey|EnumRegValue|Exch|Exec|ExecShell"
			   "|ExecShellWait|ExecWait|ExpandEnvStrings|File|FileBufSize"
			   "|FileClose|FileErrorText|FileOpen|FileRead|FileReadByte"
			   "|FileReadUTF16LE|FileReadWord|FileSeek|FileWrite"
			   "|FileWriteByte|FileWriteUTF16LE|FileWriteWord|FindClose"
			   "|FindFirst|FindNext|FindWindow|FlushINI|Function|FunctionEnd"
			   "|GetCurInstType|GetCurrentAddress|GetDlgItem|GetDLLVersion"
			   "|GetDLLVersionLocal|GetErrorLevel|GetFileTime"
			   "|GetFileTimeLocal|GetFullPathName|GetFunctionAddress"
			   "|GetInstDirError|GetLabelAddress|GetTempFileName|Goto"
			   "|HideWindow|Icon|IfAbort|IfErrors|IfFileExists|IfRebootFlag"
			   "|IfSilent|InitPluginsDir|InstallButtonText|InstallColors"
			   "|InstallDir|InstallDirRegKey|InstProgressFlags|InstType"
			   "|InstTypeGetText|InstTypeSetText|Int64Cmp|Int64CmpU|Int64Fmt"
			   "|IntCmp|IntCmpU|IntFmt|IntOp|IntPtrCmp|IntPtrCmpU|IntPtrOp"
			   "|IsWindow|LangString|LangStringUP|LicenseBkColor|LicenseData"
			   "|LicenseForceSelection|LicenseLangString|LicenseText"
			   "|LoadLanguageFile|LockWindow|LogSet|LogText|ManifestDPIAware"
			   "|ManifestSupportedOS|MessageBox|MiscButtonText|Name|Nop"
			   "|OutFile|Page|PageCallbacks|PageEx|PageExEnd|Pop|Push|Quit"
			   "|ReadEnvStr|ReadINIStr|ReadRegDWORD|ReadRegStr|Reboot|RegDLL"
			   "|Rename|RequestExecutionLevel|ReserveFile|Return|RMDir"
			   "|SearchPath|Section|SectionEnd|SectionGetFlags"
			   "|SectionGetInstTypes|SectionGetSize|SectionGetText"
			   "|SectionGroup|SectionGroupEnd|SectionIn|SectionSetFlags"
			   "|SectionSetInstTypes|SectionSetSize|SectionSetText"
			   "|SendMessage|SetAutoClose|SetBrandingImage|SetCompress"
			   "|SetCompressionLevel|SetCompressor|SetCompressorDictSize"
			   "|SetCtlColors|SetCurInstType|SetDatablockOptimize|SetDateSave"
			   "|SetDetailsPrint|SetDetailsView|SetErrorLevel|SetErrors"
			   "|SetFileAttributes|SetFont|SetOutPath|SetOverwrite"
			   "|SetPluginUnload|SetRebootFlag|SetRegView|SetShellVarContext"
			   "|SetSilent|ShowInstDetails|ShowUninstDetails|ShowWindow"
			   "|SilentInstall|SilentUnInstall|Sleep|SpaceTexts|StrCmp"
			   "|StrCmpS|StrCpy|StrLen|SubCaption|SubSection|SubSectionEnd"
			   "|Unicode|UninstallButtonText|UninstallCaption"
			   "|UninstallExeName|UninstallIcon|UninstallSubCaption"
			   "|UninstallText|UninstPage|UnRegDLL|Var|VIAddVersionKey"
			   "|VIFileVersion|VIProductVersion|WindowIcon|WriteINIStr"
			   "|WriteRegBin|WriteRegDWORD|WriteRegExpandStr|WriteRegMultiStr"
			   "|WriteRegStr|WriteUninstaller|XPStyle|",
			   buf))
			return 1;
	return 0;
}

color_nsis_range(from, to)
{
	int t = -1;
	char pat[FNAMELEN];

	if (from >= to)
		return to;
	save_var point, matchstart, matchend, case_fold = 1;
	point = to;				// Color entire lines.
	nl_forward();
	to = point;
	save_var narrow_end = 0;
	point = from;
	to_begin_line();
	while (parse_string(-1, "\\\n") && matchend < point) {
		point = matchend;
		to_begin_line();
	}
	from = point;
	set_character_color(from, to, -1);

	while (point < to) {	// Search for special char or word or comment.
		if (!re_search(1, "[;#\"'`]|[^ \n\t;#\"'`/][^ \n\t;#]*|/%*")) {
			t = size();
			break;
		}
		t = matchstart;
		if (t >= to)
			return to;
		if (strchr(";#", character(t))) {
			nl_forward();
			set_character_color(t, point, color_class nsis_comment);
		} else if (character(t) == '/') {
			search(1, "*/");
			set_character_color(t, point, color_class nsis_comment);
		} else if (strchr("\"'`", character(t))) {
			sprintf(pat, "([^\n$%c]|<$>[^\\\n]|<$>\\.|\\\n)*(%c|$)",
					character(t), character(t));
			point += parse_string(1, pat);
			set_character_color(t, point, color_class nsis_string);
		} else if (character(t) == '!') {
			set_character_color(t, point, color_class nsis_preprocessor);
		} else if (is_nsis_keyword(t))
			set_character_color(t, point, color_class nsis_keyword);
		else
			set_character_color(t, point, -1);
	}
	if (to < t)
		set_character_color(to, t, -1);
	return point;
}

command nsis_mode()
{
	mode_default_settings();
	major_mode = strkeep("NSIS");
	mode_keys = nsis_tab;		/* Use these keys. */
	strcpy(comment_start, "[;#][ \t]*");
	strcpy(comment_pattern, "[;#].*$");
	strcpy(comment_begin, "; ");
	strcpy(comment_end, "");
	comment_column = 0;
	recolor_range = color_nsis_range;	// set up coloring rules
	recolor_from_here = recolor_by_lines;
	idle_coloring_size = 10000;
	if (want_code_coloring)		// maybe turn on coloring
		when_setting_want_code_coloring();
	buffer_maybe_break_line = generic_maybe_break_line;
	fill_mode = (misc_language_fill_mode & 0x200) != 0;
	auto_show_matching_characters = nsis_auto_show_delim_chars;
	mode_auto_show_delimiter = ifdef_auto_show_delimiter;
	try_calling("nsis-mode-hook");
	make_mode();
}

when_loading()
{
	nsis_tab[ALT('q')] = (short) fill_comment;
	nsis_tab[ALT('Q')] = (short) prefix_fill_paragraph;
}

tag_mode_nsis()	/* tag all functions, macros, and global defs in the file */
{
	char func[TAGLEN];
	int start;

	save_var case_fold = 1, point = 0;
	while (re_search(1, "^[ \t]*(Function|[!]Macro|[!]Define)<hspace>+([a-z0-9_]+)</word>")) {
		grab(start = find_group(2, 1), find_group(2, 0), func);
		add_tag(func, start);
	}
}

suffix_nsi()
{
	nsis_mode();
}

suffix_nsh()
{
	nsis_mode();
}

suffix_nsdinc()
{
	nsis_mode();
}

