/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1985, 2020 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// This file has commands for parsing error messages.

#include "eel.h"
#include "proc.h"
#include "epsnet.h"

command next_error()
{
	do_next_error(iter, NULL);
	iter = 1;
}

command previous_error()
{
	do_next_error(-iter, NULL);
	iter = 1;
}

proc_view_error_lines(int buf, int pos)
{
	if (!process_view_error_lines)
		return;
	if (process_view_error_lines < -90
		  || process_view_error_lines > screen_lines - 2)
		process_view_error_lines = screen_lines / 3;
	if (process_view_error_lines < 0)	// It's a percentage.
		save_var process_view_error_lines
				= muldiv(-process_view_error_lines, screen_lines, 100);
	int owin = window_handle;
	int win = buf_in_window(buf);
	if (win >= 0)
		window_handle = win;
	else {
		window_number = 0;
		window_number--;		// Get to last window on screen.
		while (window_height < process_view_error_lines) {
			int oh = window_height++;
			if (oh == window_height)
				if (give_window_space(BTOP))
					break;
		}
		if (window_height > process_view_error_lines) {
			window_split(HORIZONTAL);
			window_height = process_view_error_lines;
		}
		to_buffer_num(buf);
	}
	point = window_start = pos;
	set_window_wrap(wrap_view_error_lines);
	if (is_window(owin))
		window_handle = owin;
	else
		window_number--;
}

proc_view_start_compile(int buf)
{
	if (!process_view_error_lines)
		return;
	if (buf_in_window(buf) >= 0)
		proc_view_error_lines(buf, buf_size(buf));
}

// A reference to this file was found in compiler output in the
// specified buffer.  Determine the directory it's in and make it an
// absolute pathname.  We recognize the "Running" message we insert,
// and the Entering and Leaving messages created by Gnu Make that
// indicate it's changing directories.

process_find_directory(char *file, int buf, int pos)
{
	char dir[FNAMELEN], url[FNAMELEN];
	int level = 0, got = 0, c;

	save_var bufnum = buf, point = pos;
	*dir = 0;
	if (is_process_buffer(bufnum) && *process_current_directory
		  && !is_remote_buffer(bufnum))
		strcpy(dir, process_current_directory);
	while (!got && re_search(-1, "(Entering|Leaving|Running `.*' in)"
							 " directory `(.*)'$")) {
		switch (c = character(find_group(1, 1))) {
			case 'L':	level++; break;
			case 'E':	if (--level >= 0)
							break;
			case 'R':	got = c;
						grab(find_group(2, 1), find_group(2, 0), dir);
						break;
		}
	}
	if (telnet_error_converter(file, *dir ? dir : NULL, url)) {
		strcpy(file, url);
		return;
	}
	if (got != 'R')	// "Running" message is from us, so don't translate it.
		convert_cygwin_filename(dir, 0);
	if (opsys == OS_DOS && strchr(file, '/'))
		convert_cygwin_filename(file, 0);
	if (*dir)
		absolute(file, dir);
	else
		absolute(file);
}

// We found message at pos in procbuf referencing the specified java
// file, but it's not where we expected based on cur dir.  Reparse
// message and see if we can find it, modifying file name if found.

proc_find_java_source(char *file, int procbuf, int pos)
{
	char basename[FNAMELEN], path[FNAMELEN], *s, *classpath;

	classpath = getenv("CLASSPATH");
	save_var bufnum = procbuf;
	strcpy(basename, get_tail(file, 1));
	if (classpath && matches_at(pos, 1, ".*( |^)(<word|$>+(%.<word|$>+)+)")) {
		grab(find_group(2, 1), find_group(2, 0), path);
		if (s = strrchr(path, '.'))		// drop function name
			*s = 0;
		if (s = strrchr(path, '.'))		// drop file name
			*s = 0;
		while (s = strrchr(path, '.'))		// make rest into path
			*s = path_sep;
		sprintf(path + strlen(path), "%c", path_sep);
		strcat(path, basename);
		if (s = look_on_path(path, 0, classpath))
			strcpy(file, s);
	}
}

// If the file doesn't exist, substitute a file with the same basename.
proc_fix_filename(char *fname)
{
	if (!(process_next_error_options & 1) || is_remote_file(fname)
		  || check_file(fname))
		return;
	int orig = bufnum;
	int i = buf_list(0, 2);	/* start with most recently created buf */
	do {
		bufnum = i;
		if (*filename
			  && !fnamecmp(get_tail(filename, 0), get_tail(fname, 0))
			  && force_save_as != FSA_NEWFILE
			  && check_file(filename) == CHECK_FILE) {
			show_text((process_next_error_options & 2) ? 0 : -2, see_delay,
					  "No file %f found, picking another %s",
					  fname, get_tail(fname, 0));
			strcpy(fname, filename);
			return;
		}
	} while (i = buf_list(-1, 1));
	bufnum = orig;
}

// First, define some regex pieces that could be helpful in multiple
// definitions.

// These file name matchers say, optionally a drive spec c:, then a
// basename, a dot, and an extension.  The OK_SPACE version permits
// spaces in the basename, starting with the 2nd character.  Both permit
// tags like <3> at the end of the extension.
#define FNAME_NO_SPACE	"(([a-z]:)?[+-9A-z~]+<.>[+-9A-z]+[0-9<>]*)"
#define FNAME_OK_SPACE	"(([a-z]:)?[+-9A-z~][+-9A-z~ ()]*<.>[+-9A-z]+[0-9<>]*)"

// Some things we want to ignore in an error message.

// 3> or E " or W " or ***, for example.
#define START_STUFF1	"[1-9]>|e \"|w \"|<*>*"
// One of these three keywords, or [someother].
#define START_STUFF2	"error|warning|fatal| *%[[a-z]+%]"
// Optionally, a code like E123 with a space before it, then perhaps
// some punc chars * : after.  Or just punc chars " or (.
#define START_STUFF3	"( [a-z][0-9]+)?[*:]*|[\"(]?"

// Between a file name and line number, we must have space, open paren,
// comma or colon, then perhaps L or the word "line" or a [, then
// perhaps space.  Also skip over any " at the start, which may have
// surrounded file name.
#define BEFORE_NUM		"\"?([ (:,]( *line|L|%[)?) *"
// The line number.
#define LINE_NUM		"([0-9]+)"
// Stuff to require after line number.  Either ) or colon, or else some
// stuff with colon or semicolon after it, or else a column number and ].
#define END_STUFF		"(<rparen|:>|[^0-9\\\n].*[:;]|,[0-9]+%][ \n]*)"

/* If you change ERROR_PATTERN, you may have to change the next 3 macros
 * as well. */
#define ERROR_PATTERN	"^(" START_STUFF1 \
						"(" START_STUFF2 "|[a-z0-9]+ )" \
						START_STUFF3 "|[ \t]*from) *" \
						FNAME_NO_SPACE \
						BEFORE_NUM LINE_NUM END_STUFF ".*"
#define FILE_PAREN	4	/* filename is in 4th () pair above */
#define LINE_PAREN	8	/* line# is in 8th pair of parens above */

#define SIMP_PATTERN	"[:;)]| line "	/* search with simple pattern first */

// This version permits spaces in file names, but with more
// restrictions on ancillary text before it.
#define ERROR_PATTERN2	"^(" START_STUFF1 \
						"(" START_STUFF2 ")" \
						"( [a-z][0-9]+)?[*:]*|[\"(]?) *" \
						FNAME_OK_SPACE \
						BEFORE_NUM LINE_NUM END_STUFF ".*"
#define FILE_PAREN2	4	/* filename is in 4th () pair above */
#define LINE_PAREN2	8	/* line# is in 8th pair of parens above */

#define ERROR_PATTERN3	".* at (.+) line ([0-9]+)[.,].*"
#define FILE_PAREN3	1	/* filename in () pair above */
#define LINE_PAREN3	2	/* line# in pair of parens above */

// An error pattern that permits file names with no . character, but
// with more restrictions on the format.
#define ERROR_PATTERN4	"^([+-9A-z~:]+[0-9<>]*)" BEFORE_NUM \
						LINE_NUM "[ :)]*(error|warning).*[:;].*"
#define FILE_PAREN4	1	/* filename is in 1st () pair above */
#define LINE_PAREN4	2	/* line# is in 2nd pair of parens above */

// An error pattern for Java runtime errors.

#define ERROR_PATTERN5	"^.*%(([+-9A-z~:]+):([0-9]+)%)"
#define FILE_PAREN5	1	/* filename is in 1st () pair above */
#define LINE_PAREN5	2	/* line# is in 2nd pair of parens above */

// An error pattern that permits file names with no . character,
// allowing spaces, but with even more restrictions on the rest of the
// message.
#define ERROR_PATTERN6	"^([a-z]:\\[ +-9A-z~]+)[(]" \
						LINE_NUM "[ :)]*(error|warning).*[:;].*"
#define FILE_PAREN6	1	/* filename is in 1st () pair above */
#define LINE_PAREN6	2	/* line# is in 2nd pair of parens above */

// You can define ERROR_PATTERN7 with another alternate pattern.
// Epsilon will look for either one, on lines that contain SIMP_PATTERN.
// Define FILE_PAREN7 and LINE_PAREN7 to match.  (Also ERROR_PATTERN8
// etc.)  You can also define an error pattern where the PAREN codes
// refer to an empty set of parens, for error messages that don't
// include a file name; Epsilon will jump to the process buffer line
// reporting the error for these.

do_next_error(int count, char *cmd)	// return 0 if no errors, else 1
{
	char file[FNAMELEN], number[30], line[FNAMELEN], *p;
	int lineno, origpt, dir, from, to, is_bufname, file_paren, line_paren;
	int orig = bufnum, procbuf, start, have_file = 1;

	tiled_only();
	_last_was_grep = 0;
	if (spot_to_buffer(error_spot) <= 0 && !exist(PROCBUF))
		error("No process buffer to read errors from.");
	tiled_only();
	if (spot_to_buffer(error_spot) > 0)
		bufnum = spot_to_buffer(error_spot);
	else
		bufname = PROCBUF;
	telnet_error_set_to_telnet();
	procbuf = bufnum;
	origpt = point;
	case_fold = 1;
	if (spot_to_buffer(error_spot) > 0)
		point = *error_spot;
	else {
		error_spot = alloc_spot();
		point = 0;
	}
	if (!count && procbuf == orig)
		point = origpt;

	noteput("Searching...");
	if (!count)
		find_error(-1, &file_paren, &line_paren), count = 1;
	dir = (count > 0) ? 1 : -1;
	for (; count; count -= dir) {
		if (!find_error(dir, &file_paren, &line_paren)) {
			point = origpt;
			bufnum = orig;
			if (_had_errors == 2)
				return 0;
			else if (_had_errors)
				say("No more errors.");
			else if (cmd)
				say("No errors running \"%s\".", cmd);
			else
				say("No errors.");
			return 0;
		}
	}
	_had_errors = 1;
	grab(find_group(file_paren, 1), find_group(file_paren, 0), file);
	grab(find_group(line_paren, 1), find_group(line_paren, 0), number);
	start = find_group(0, 1);
	from = find_group(line_paren, 0);
	to = find_group(0, 0);
	if (to - from < 20) {	// error message at end is too short
		from = start;	// so try including whole line
		if (to - from > screen_cols - 5) // Only rightmost if it's too big.
			from = to - screen_cols + 5;
	}
	if (to > from + sizeof(line) / sizeof(*line) - 1)
		to = from + sizeof(line) / sizeof(*line) - 1;
	grab(from, to, line);
	is_bufname = (parse_string(1, "Buffer ") != 0);
	if (dir > 0)
		to_end_line();
	*error_spot = point;
	point = origpt;
	lineno = strtoi(number, 10);
	bufnum = orig;
	if (!*file || !*number) {
		locate_window(bufnum_to_name(procbuf), "");
		point = *error_spot;
		to_begin_line();
		say("=> %s", line);
		return 1;
	}
	if (is_bufname)
		locate_window(file, "");
	else {
		process_find_directory(file, procbuf, from);
		if (!check_file(file) && !strcmp(get_extension(file), ".java"))
			proc_find_java_source(file, procbuf, start);
		proc_fix_filename(file);
		locate_window("", file);
		if (process_view_error_lines
			  && window_number == number_of_windows() - 1)
			window_number = 0;
		if ((process_next_error_options & 4) && !is_remote_file(file)
			  && check_file(file) != CHECK_FILE)
			have_file = 0;
		else
			std_find_it(file);
	}
	if (have_file) {
		go_line(lineno);
		if (is_remote_file(file))
			ftp_go_to_line = lineno;
	} else
		say("%f not found", file);
	for (p = line; *p && index("()[] :;,\t\n0123456789", *p); p++)
		;			// skip over these chars at start of msg
	say("=> %s", p);
	restore_vars();
	proc_view_error_lines(procbuf, start);
	return 1;
}

/* Find next error in direction, return 1 and set vars to find_group() paren count */
find_error(int dir, int *file_paren, int *line_paren)
{
	int old = point;

	if (point > 0 && dir > 0)
		to_end_line();
	else
		to_begin_line();
	while (re_search(dir, SIMP_PATTERN)) {
		to_begin_line();
		if ((!*ignore_error || !parse_string(1, ignore_error, NULL))
			  && !parse_string(1, "In file included from ", NULL)) {
			if (parse_string(1, ERROR_PATTERN, (char *)0)) {
				*file_paren = FILE_PAREN;
				*line_paren = LINE_PAREN;
				return 1;
			}
#ifdef ERROR_PATTERN2
			if (parse_string(1, ERROR_PATTERN2, (char *)0)) {
				*file_paren = FILE_PAREN2;
				*line_paren = LINE_PAREN2;
				return 1;
			}
#endif
#ifdef ERROR_PATTERN3
			if (parse_string(1, ERROR_PATTERN3, (char *)0)) {
				*file_paren = FILE_PAREN3;
				*line_paren = LINE_PAREN3;
				return 1;
			}
#endif
#ifdef ERROR_PATTERN4
			if (parse_string(1, ERROR_PATTERN4, (char *)0)) {
				*file_paren = FILE_PAREN4;
				*line_paren = LINE_PAREN4;
				return 1;
			}
#endif
#ifdef ERROR_PATTERN5
			if (parse_string(1, ERROR_PATTERN5, (char *)0)) {
				*file_paren = FILE_PAREN5;
				*line_paren = LINE_PAREN5;
				return 1;
			}
#endif
#ifdef ERROR_PATTERN6
			if (parse_string(1, ERROR_PATTERN6, (char *)0)) {
				*file_paren = FILE_PAREN6;
				*line_paren = LINE_PAREN6;
				return 1;
			}
#endif
#ifdef ERROR_PATTERN7
			if (parse_string(1, ERROR_PATTERN7, (char *)0)) {
				*file_paren = FILE_PAREN7;
				*line_paren = LINE_PAREN7;
				return 1;
			}
#endif
#ifdef ERROR_PATTERN8
			if (parse_string(1, ERROR_PATTERN8, (char *)0)) {
				*file_paren = FILE_PAREN8;
				*line_paren = LINE_PAREN8;
				return 1;
			}
#endif
#ifdef ERROR_PATTERN9
			if (parse_string(1, ERROR_PATTERN9, (char *)0)) {
				*file_paren = FILE_PAREN9;
				*line_paren = LINE_PAREN9;
				return 1;
			}
#endif
		}
		if (dir > 0)
			nl_forward();
	}
	*error_spot = old;
	return 0;
}

