/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1985, 2020 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// This file has commands for running other programs.

#include "eel.h"
#include "proc.h"
#include "colcode.h"
#include "kill.h"
#include "epsnet.h"

/* Process commands. */

no_running(int talk)			/* make sure no process is running */
{
	if (!is_process_buffer(name_to_bufnum(PROCBUF)))
		return 0;
	if (talk)
		say("A process is already running in buffer %s.", PROCBUF);
	if (strcmp(bufname, PROCBUF)) {
		tiled_only();
		locate_window(PROCBUF, "");
		maybe_refresh();
	}
	point = size();
	return 1;
}

char *get_cmdline(msg, buf)	/* prompt for & return cmd */
char *msg, *buf;		/* if !has_arg, prepare cmd in buf */
{				/* but don't return it */
	if (has_arg) {
		get_strdef(buf, msg, push_cmd);
		strcpy(push_cmd, buf);
		iter = 1;
	}
	build_filename(buf, push_cmd, filename);
	return has_arg ? buf : "";
}

command push() on cx_tab[CTRL('E')]
{
	char cmd[FNAMELEN * 2];

	_last_was_grep = 0;
	save_var clear_process_buffer = (clear_process_buffer & 3) != 0;
	do_push(get_cmdline("Push With Command", cmd), capture_output, 1);
}

// Should we switch back to the original directory after starting the
// process?  Yes, if start-process-in-buffer-directory is 2.  However,
// if we're sharing the current directory between Epsilon and the
// process (under Win95/98/ME only), remain in this directory.

should_restore_process_directory()
{
	if (start_process_in_buffer_directory != 2)
		return 0;
	if (is_mswindows && use_process_current_directory
		  && can_get_process_directory == 1)
		return 0;
	return 1;
}

do_push(cmdline, cap, show)
char *cmdline;	// cap nonzero means capture output, show means show it to user
{		// but show -1 means don't wait for keystroke
	char dir[FNAMELEN], old[FNAMELEN];
	int result, tbuf;

	if (!(has_feature & FEAT_MULT_CONCUR) && !another_process())
		save_var restart_concurrent = 0;
	if (is_unix == IS_UNIX_XWIN && !*cmdline) {
		cmdline = push_cmd_unix_interactive;
		if (is_unix_flavor == IS_UNIX_MACOS)
			cmdline = push_cmd_unix_macos_interactive;
	}
	iter = 1;
	before_push();
	if (start_process_in_buffer_directory) {
		getcd(old);
		get_buffer_directory(dir);
		chdir(dir);
	}
	result = shell("", cmdline, cap ? bufnum_to_name(tbuf = tmp_buf())
					: "", SHELL_COPY_OUTPUT);
	if (should_restore_process_directory())
		chdir(old);
	if (cap) {
		process_captured(tbuf, cmdline);
		buf_delete(tbuf);
	}
	after_push();
	if (cmdline[0] && cap && show) {
		locate_window(PROCBUF, "");
		point = size();
	}
	if (result == -1 && errno) {
		file_error(errno, "During exec", "Couldn't exec");
		quick_abort();
	}
	maybe_restart_concurrent();
	if (cmdline[0] && result)
		say("Process returned %d", result);
	else
		check_dates(0);
	return result;
}

before_push()
{
	if (is_gui || opsys == OS_UNIX)
		return;
	save_screen(&push_info);
	sayput("");
	wait_for_unseen_msgs();
	term_position(0, screen_lines - 1);
	term_mode(0);
}

after_push()
{
	if (is_gui || opsys == OS_UNIX)
		return;
	build_first = 1;
	screen_messed();
	if (term_mode(1))
		restore_screen(&push_info);
	say("");
}

process_captured(b, cmdline)	/* move captured output to process buffer */
char *cmdline;
{
	save_var bufnum;
	create(PROCBUF);
	bufname = PROCBUF;
	process_mode();
	if (clear_process_buffer)
		zap(PROCBUF);
	point = is_process_buffer(bufnum) ? type_point : size();
	set_error_spot();
	_had_errors = 0;
	if (point > 0 && character(point - 1) != '\n')
		insert('\n');
	if (cmdline[0])
		bprintf("%s\n", cmdline);
	grab_buffer(b);
	if (point > type_point)
		type_point = point;
}

// Put name of alternative process buffer in buf, which starts with
// name of primary process buffer.  Prefer current buffer, if its name
// has the right pattern.

pick_other_proc_buffer(char *buf)
{
	int i = 1;
	if (is_process_buffer(name_to_bufnum(buf)))
		if (!is_process_buffer(bufnum) && !strncmp(bufname, "process-", 8)
			  && strtoi(bufname + 8, 10) && !got_bad_number)
			strcpy(buf, bufname);
		else
			do {
				sprintf(buf, "process-%d", i++);
			} while (is_process_buffer(name_to_bufnum(buf)));
}

command start_process() on cx_tab[CTRL('M')]
{
	char cmd[FNAMELEN * 2], *cmdline, *sh, fixup[FNAMELEN], *p;
	char cust1[FNAMELEN], cust2[FNAMELEN], buf[FNAMELEN];

	save_var clear_process_buffer = (clear_process_buffer & 5) != 0;
	_last_was_grep = 0;
	tiled_only();
	error_if_input(1);
	strcpy(buf, PROCBUF);
	if (has_arg && (has_feature & FEAT_MULT_CONCUR))
		pick_other_proc_buffer(buf);
	else if (no_running(1))
		return;
	iter = has_arg = 0;
	cmdline = get_cmdline("Start Process With Command", cmd);
	create_concur(cmdline, buf);
	locate_window(buf, "");
	sh = getenv("EPSSHELL");
	if (!sh)
		sh = getenv("SHELL");
	if (opsys == OS_UNIX && sh) {
		*cust1 = *cust2 = *fixup = 0;
		if (p = lookpath("epsilon.sh", 0))
			strcpy(cust1, p);
		if (p = lookpath("epsilon.csh", 0))
			strcpy(cust2, p);
		if (!*cust1 || !*cust2)
			sprintf(fixup, "echo \"Customization file not found, "
					"no customizations performed.\">/dev/null");
		else
			switch (get_shell_type_from_name(sh)) {
				case 1:
					sprintf(fixup, ". \"%s\"; set +o emacs", cust1);
					break;
				case 2:
					sprintf(fixup, "source \"%s\"", cust2);
					break;
				default:
					sprintf(fixup, "echo \"Unrecognized shell, no customizations "
							"performed.\">/dev/null");
					break;
			}
		delay(40, 0, bufnum);	// Let's first get a prompt if we can.
		point = type_point;
		bprintf("%s\n", fixup);
	}
	point = size();
	process_mode();
}

// Return 1 for bash-like, 2 for csh-like, 0 otherwise. Arg is full
// path to the shell.
int get_shell_type_from_name(char *sh)
{
	if (string_matches_regex(sh, ".*/(z|ba|.ba|k)?sh", 1))
		return 1;
	if (string_matches_regex(sh, ".*/.?csh", 1))
		return 2;
	return 0;
}

create_concur(char *cmdline, char *buf)
{
	char dir[FNAMELEN], old[FNAMELEN];
	int err = 0;

	create(buf);
	if (clear_process_buffer)
		zap(buf);
	if (start_process_in_buffer_directory) {
		getcd(old);	// Don't switch to proc buffer before
		get_buffer_directory(dir);	// doing this.
		chdir(dir);
	}
	bufname = buf;
	point = size();
	if (!is_process_buffer(bufnum))
		err = do_concur_shell("", cmdline, buf);
	if (should_restore_process_directory())
		chdir(old);
	if (err) {
		file_error(errno, "During exec", "Couldn't exec");
		quick_abort();
	}
}

command stop_process() on cc_tab[CTRL('C')]
{
	halt_process(has_arg ? iter : 0, is_process_buffer(bufnum) ? bufnum : 0);
	iter = 1;
}

int exit_this_process(char *buf)
{
	save_var bufname = buf;
	if (is_process_buffer(name_to_bufnum(buf)) == ISPROC_PIPE)
		return 0;
	type_point = point = size();
	stuff("exit\n");
	delay(500, COND_PROC_EXIT, bufnum);
	return !is_process_buffer(name_to_bufnum(buf));
}

command exit_process()
{
	tiled_only();
	if (!exist(PROCBUF) || !is_process_buffer(name_to_bufnum(PROCBUF)))
		error("No process to exit.");
	if (!exit_this_process(PROCBUF) && !char_avail()) {
		locate_window(PROCBUF, "");
		error("Couldn't exit from concurrent process.");
	}
}

command kill_process()
{
	if (!is_process_buffer(bufnum))
		kill_ftp_job(bufnum);
	else if (!process_kill(bufnum))
		say("Unable to kill the process.");
}

command compile_buffer() on reg_tab[FALT(3)]
{		// compile the current buffer (mode-specific)
	int res, buf = bufnum;
	char cmd[FNAMELEN], oscmd[FNAMELEN], full[FNAMELEN], *ext;

	save_var clear_process_buffer = (clear_process_buffer & 9) != 0;
	tiled_only();
	if (use_compile_command_file_variable && look_for_filevar_pattern(
		"compile-command", "([^-\" \n\t;]([^-;\n]|-[^*])*|\"[^\n\"]+\")", NULL)) {
		grab(find_group(1, 1), find_group(1, 0), oscmd);
		if (*oscmd == '"')
			drop_quotes(oscmd);
		if (use_compile_command_file_variable != 1)
			get_strdef(cmd, "Command to compile this buffer", oscmd);
		else
			strcpy(cmd, oscmd);
		if (*cmd)
			save_var compile_buffer_cmd = cmd;
	}
	if (!compile_buffer_cmd) {
		ext = get_extension(filename);
		sprintf(cmd, "compile-%s-cmd", *ext ? (ext + 1) : "none");
		sprintf(oscmd, "%s-%s", cmd, is_unix ? "unix" :
				is_mswindows ? "win32" : "other");
		if ((ext = get_str_var(find_index(oscmd)))
			  || (ext = get_str_var(find_index(cmd))))
			save_var compile_buffer_cmd = ext;
	}
	if (!compile_buffer_cmd || !*compile_buffer_cmd) {
		get_string(cmd, "Command to compile this buffer: ");
		if (!*cmd)
			error("No command specified.");
		compile_buffer_cmd = strsave(cmd);
	}
	save_var concurrent_make;
	if (pre_compile_hook)	// hook can change above vars
		if ((*pre_compile_hook)())
			return;		// hook says don't do compile
	if (concurrent_compile < 3)
		concurrent_make = concurrent_compile;
	if (concurrent_compile == 0 && (is_gui || is_mswindows))	// Run in the
		concurrent_make = 4;		// concur proc if any, but keep control.
	if (start_process_in_buffer_directory && !is_remote_file(filename)) {
		char dir[FNAMELEN], old[FNAMELEN];

		getcd(old);
		get_buffer_directory(dir);
		chdir(dir);
		build_filename(full, compile_buffer_cmd, filename);
		chdir(old);
		if (start_process_in_buffer_directory == 2)	// For error msgs.
			save_var start_process_in_buffer_directory = 1;
	} else
		build_filename(full, compile_buffer_cmd, filename);
	res = do_compile(full);
	save_var bufnum = buf;
	if (post_compile_hook)
		(*post_compile_hook)(res, buf);
}

run_on_file_template(char *template)
{
	char fname[FNAMELEN];
	save_var start_process_in_buffer_directory = 1;
	build_filename(fname, template, filename);
	do_compile(fname);
}

load_this_bytecode_file(res, buf)	// load .b file made from current file
{
	char file[FNAMELEN];

	if (res && res != 4)
		return;
	save_var bufnum = buf;
	strcpy(file, filename);
	strcpy(get_extension(file), byte_extension);
	if (check_file(file, NULL) == CHECK_FILE) {
		load_commands(filename);
		say("%f compiled and loaded.%s", *filename ? filename : "Buffer",
		   res == 4 ? "  (There were warnings.)" : "");
	}
}

compile_via_dll()	// Try to compile .e file using a DLL, return 1 if ok.
{
	if (!(has_feature & FEAT_EEL_COMPILE))
		return 0;			// Not supported under this platform.
	_had_errors = 0;
	save_var point;
	switch (eel_compile(bufname, 0, compile_eel_dll_flags, "eel-errors")) {
		case 2:		// Couldn't access DLL.
			say("Unable to compile internally, running process.");
			return 0;
		case 1:
			break;
		case 0:
			say("%f compiled and loaded.%s",
				filename ? filename : "Buffer",
				buffer_size("eel-errors") ? "  (There were warnings.)" : "");
			break;
		default:
			say("Unable to compile internally, running process.");
			return 0;
	}
	restore_vars();
	save_var bufname = "eel-errors";
	if (size()) {
		point = 0;			// Starting at the beginning,
		set_error_spot();	//    scan in this buffer.
		do_next_error(1, NULL);	// Scan for error messages.
	} else {
		restore_vars();
		delete_buffer("eel-errors");
	}
	return 1;
}

command make() on cx_tab['m']
{
	char cmd[FNAMELEN];

	save_var clear_process_buffer = (clear_process_buffer & 17) != 0;
	tiled_only();
	get_cmdline("Make using command", cmd);
	save_var start_process_in_buffer_directory
			= start_make_in_buffer_directory;
	do_compile(cmd);
}

// Count the number of compilation buffers running and update the mode
// line message.  Use offset to correct for a compilation buffer we know
// is about to go away or start up.

update_mode_more_extra(int offset)
{
	char *s;
	int cnt = 0;

	for (s = buf_match("compilation-output", STARTMATCH); s
			 ; s = buf_match("compilation-output", 0)) {
		if (is_process_buffer(name_to_bufnum(s)))
			  cnt++;
	}
	cnt += offset;
	if (cnt < 0)
		cnt = 0;
	switch (cnt) {
		case 0:		*mode_more_extra = 0; break;
		case 1:		strcpy(mode_more_extra, "[Compiling] "); break;
		default:	sprintf(mode_more_extra, "[Compiling %d] ", cnt); break;
	}
	all_must_build_mode = 1;
	maybe_refresh();
}

compile_handler(int activity, int buf, int from, int to)
{
	UNREFERENCED_PARAMETER(to);
	if (activity == NET_DONE) {
		save_var bufnum = buf, abort_searching = 0;
		point = size();
		if (size() && character(point - 1) != '\n')
			bprintf("\n");
		bprintf("Process exited with status %d\n", from);
		update_mode_more_extra(-1);
		if (post_compile_hook_buffer && buf_exist(post_compile_hook_buffer)) {
			save_var bufnum = post_compile_hook_buffer;
			if (post_compile_hook)		// Track this compilation back to orig
				(*post_compile_hook)(6, bufnum); // buffer, then run its hook.
		}
	}
}

// Run the given cmd, then search for errors.
// Return 0 if the compile was ok,
// 1 if the process returned an error code,
// 2 if we ran it concurrently and the user started typing,
// 3 if the user didn't ok saving,
// 4 if the process returned 0, but next-error found some errors.
// 5 if we can't tell when the process has finished.
// 6 is reserved; we don't return it, but a post_compile_hook function
// can receive when the compilation finishes it if the user interrupted
// us and we returned 2.
do_compile(cmd)
char *cmd;
{
	int err = 0, nerr, sep_buf_ok = 0;
	int obuf = bufnum;
	char *procbuf = PROCBUF, curdir[FNAMELEN];

	_last_was_grep = 0;
	if (maybe_save_all())
		return 3;
	if (start_process_in_buffer_directory)
		get_buffer_directory(curdir);
	else
		getcd(curdir);
	drop_final_slash(curdir);
	if (!(has_feature & FEAT_MULT_CONCUR) && !another_process())
		save_var restart_concurrent = 0;

	// If we're allowed to run non-concur, then use separate buffer.
	if (concurrent_make < 2 && (has_feature & FEAT_MULT_CONCUR)
		  && compile_in_separate_buffer)
		sep_buf_ok = 1;
	else if (concurrent_make == 2)
		create_concur("", PROCBUF);
	else if (!sep_buf_ok && concurrent_make == 0)
		try_exit_concurrent();
	if (!sep_buf_ok && clear_process_buffer)
		zap(PROCBUF);
	if (sep_buf_ok) {
		int i;
		char newname[FNAMELEN];
		bufnum = create(procbuf = "compilation-output");
		for (i = 2; is_process_buffer(bufnum); i++) {
			sprintf(procbuf = newname, "compilation-output-%d", i);
			bufnum = create(procbuf);
		}
		if (clear_process_buffer)
			zap(procbuf);
        process_mode();
		point = size();
		set_error_spot();
		_had_errors = 0;
		bprintf("Running `%s' in directory `%s'\n", cmd, curdir);
		type_point = size();
		if (compile_in_separate_buffer == 2)
			locate_window(procbuf, "");
		if (!concur_shell("", cmd, curdir, procbuf)) {
			update_mode_more_extra(0);
			when_activity = compile_handler;
			post_compile_hook_buffer = obuf;
			proc_view_start_compile(bufnum);
			while (process_exit_status == PROC_STATUS_RUNNING) {
				if (char_avail() && getkey() != MOUSE_MOVE
					  && !IS_WIN_PASSIVE_KEY(key)) {
					ungot_key = full_key;	// Wait until user presses a key
					return 2;			// or process finishes.
				}
				delay(10, COND_KEY, bufnum);	// Delay 0.1 secs
			}
			err = process_exit_status;
			bufnum = obuf;
			check_dates(0);
		} else
			err = -1;
	} else if (concurrent_make == 4 && is_process_buffer(name_to_bufnum(PROCBUF))) {
		save_var bufname = PROCBUF;	// Run the program in the concur
		set_error_spot();			// proc buffer, but don't process
		_had_errors = 0;			// the user's keys until it finishes.
		if (*cmd) {
			point = size();
			bprintf("%s\n", cmd);
		}
		proc_view_start_compile(bufnum);
		maybe_refresh();
		iter = 1;
		delay(-1, COND_PROC, bufnum);
		if (!is_process_buffer(name_to_bufnum(PROCBUF)) || !process_input())
			return 2;	/* else assume it's compiled */
	} else if (no_running(0)) {	/* goes to end of proc buffer */
		set_error_spot();
		_had_errors = 0;
		to_begin_line();
		bprintf("%pRunning `%s' in directory `%s'\n",
				color_class process_output, cmd, *process_current_directory
				? process_current_directory : curdir);
		type_point = point = size();
		if (*cmd)
			bprintf("%s\n", cmd);
		maybe_refresh();
		iter = 1;
		if (opsys != OS_UNIX)
			delay(-1, COND_PROC | COND_KEY, bufnum);
		if (!is_process_buffer(name_to_bufnum(PROCBUF)) || !process_input())
			return 2;	/* else assume it's compiled */
		if (!(has_feature & FEAT_DETECT_CONCUR_WAIT))
			return 5;
	} else {
		bufnum = create(PROCBUF);
		point = type_point = size();
		buffer_printf("process", "%p\nRunning `%s' in directory `%s'\n",
				color_class process_output, cmd, curdir);
		bufnum = obuf;
		err = do_push(cmd, 1, 0);
	}
	maybe_restart_concurrent();
	nerr = do_next_error(1, cmd);	// scan for error messages
	if (!nerr && err) {	// no message recognized, but error code
		say("Process returned %d", err);
		locate_window(procbuf, "");
		point = size();		/* show end of process buffer */
		return err;
	} else
		bufnum = obuf;
	if (err)
		return 1;
	else if (nerr)
		return 4;
	return 0;
}

// The following functions save & restore a concurrent process.  To use,
// you must first execute:
// 	if (!another_process()) save_var restart_concurrent = 0;
// Then call try_exit_concurrent(), run the process, and then call
// maybe_restart_concurrent().
// These functions are disabled in versions that support multiple
// concurrent processes, as they're not useful there.

try_exit_concurrent()
{
	if (!(has_feature & FEAT_MULT_CONCUR) && another_process()) {
		exit_process();
		if (another_process())
			error("Couldn't exit from concurrent process.");
	}
}

maybe_restart_concurrent()
{
	if (!(has_feature & FEAT_MULT_CONCUR) && restart_concurrent
		  && !another_process())
		do_concur_shell("", "", PROCBUF);
}

// Epsilon calls this function whenever something happens with the
// process.

concur_handler(int activity, int buf, int from, int to)
{
	switch (activity) {
		case NET_SEND:	do_when_process_input(buf, from, to); break;
		case NET_RECV:	do_when_process_output(buf, from, to); break;
	}
}

do_concur_shell(char *prog, char *cmd, char *buf)
{
	save_var bufnum = create(buf);
	when_activity = concur_handler;
	restore_vars();
	return concur_shell(prog, cmd, "", buf);
}

process_find_linked_file(char *file)
{
	int buf = bufnum, pos = point;

	save_var _had_errors = 2;
	if (do_next_error(0, NULL)) {		// Found an error message here.
		save_var bufnum = buf, point = pos;
		quiet_set_bookmark();
		quick_abort();
	}
	say("");
	default_find_linked_file(file);
}

command view_process() on reg_tab[FSHIFT(3)]	// put up a list of errors,
{					//  let user select one, and go there
	int w = window_handle, win, buf, pos, proc;

	tiled_only();
	if (buf_exist(spot_to_buffer(error_spot)))
		proc = spot_to_buffer(error_spot);
	else if (exist(PROCBUF))
		proc = name_to_bufnum(PROCBUF);
	else
		error("No process buffer to read errors from.");
	save_var menu_width = screen_cols - _view_left - _view_right;
	save_var _doing_input = DI_LINEINPUT;

	save_var bufnum = proc;
	if (spot_to_buffer(error_spot) != bufnum) {
		pos = 0;
		set_error_spot();
	} else
		pos = *error_spot;
	win = new_menu(buf = tmp_buf());
	grab_buffer(proc);
	point = pos;
	to_begin_line();
	window_title(win, BTOP, TITLECENTER, " Compiler Errors ");
	window_to_fit(screen_lines - 7);
	save_var _window_flags |= HIGHLIGHT_LINE;
	save_var search_in_menu = 0;
	save_var display_column = -1;	// ensure all of msg is visible
	if (is_gui && want_gui_prompts) {
		one_window_to_dialog("Select an error message", win,
							 "To Error", "Cancel", "");
	} else
		say("Select an error message and press <Enter>.");
	set_read_only(1);
	select_menu_item(0, win, w, 0);
	bufnum = buf;
	pos = point;
	bufnum = proc;
	buf_delete(buf);
	check_abort();
	point = pos;
	set_error_spot();
	ungot_key = -1;
	restore_vars();
	do_next_error(0, NULL);
}

color_process_range(from, to)
{
	if (from >= to || spot_to_buffer(process_input_starts_at) <= 0
		  || !(process_coloring_rules & 1))
		return to;
	save_var point, matchstart, matchend;
	point = *process_input_starts_at;
	if (get_character_color(point - 1) != color_class process_prompt)
		return to;
	to_end_line();
	set_character_color(*process_input_starts_at, point,
						color_class process_input);
	if (curchar() == '\n')
		set_character_color(point, point + 1, -1);
	return to;
}

process_cmds_to_end(char *buf)
{
	save_var bufname = buf;
	point = 0;
	save_var case_fold = 0;
	do_uniq(1, 1, 0);		// discard adjacent duplicate cmds silently
	point = size();
}

// This process buffer region might be a prompt containing a directory
// name.  If this is NT, where we can't get concurrent process directory
// names in any other way, grab it and set Epsilon to that dir.
// We assume here that the prompt is the usual "$P$G".

try_grab_directory_name(int from, int to)
{
	char dir[FNAMELEN];

	if (!is_mswindows || !use_process_current_directory
		  || to - from > 200 || can_get_process_directory == 1)
		return;					// Too long for a likely dir name.
	if (*process_prompt_pattern) {
		save_var narrow_end = size() - to;
		if (!matches_at(from, 1, process_prompt_pattern)
			  && !matches_at(from, -1, process_prompt_pattern))
			return;
		grab(find_group(1, 1), find_group(1, 0), dir);
	} else {
		if (character(--to) != '>')
			return;
		grab(from, to, dir);
	}
	convert_cygwin_filename(dir, 1);
	if (is_directory(dir)) {
		process_current_directory = dir;
		if (use_process_current_directory == 2)
			chdir(dir);
		can_get_process_directory = 2;
	}
}

// Called in versions that only permit one process buffer.
when_process_input(int type, int start)
{
	do_when_process_input(name_to_bufnum(PROCBUF), type, start);
}

when_process_output(int start, int end)
{
	do_when_process_output(name_to_bufnum(PROCBUF), start, end);
}

notice_password_prompt(int pos)
{
	UNREFERENCED_PARAMETER(pos);
	save_var case_fold = 1, point = size();
	if ((recognize_password_prompt & 2) && (has_feature & FEAT_PROC_SEND_TEXT)
		  && parse_string(-1, recognize_password_pattern)
		  && ungot_key == -1 && !_doing_input && !number_of_popups()) {
		ungot_key = WIN_HELP_REQUEST;	// Make Epsilon run this cmd:
		menu_command = "password-prompt";
	}
}

// The process is now ready for or has just been sent some input.
do_when_process_input(int buf, int type, int start)
{
	char hist_buf[FNAMELEN];

	save_var abort_searching = 0;
	save_var bufnum = buf;
	save_var point = start;
	if (type == PROCESS_INPUT_LINE) {
		if (opsys == OS_UNIX) {	// Our only opportunity to color.
			if (spot_to_buffer(process_input_starts_at) <= 0)
				process_input_starts_at = alloc_spot();
			*process_input_starts_at = start;
			if (process_coloring_rules & 1)
				set_character_color(give_begin_line(), start,
									color_class process_prompt);
		}
		sprintf(hist_buf, "-process-history-%s", bufname);
		create(hist_buf);
		process_cmds_to_end(hist_buf);
		nl_forward();
		if (point > start + 1)		// don't store empty (just \n) cmds
			xfer(hist_buf, start, point);
		process_cmds_to_end(hist_buf);
	} else if (type == PROCESS_INPUT_CHAR) {
		if (process_coloring_rules & 1)
			set_character_color(start, start + 1, color_class process_input);
	} else if (!type) {		// Got a request for line input (might now wait).
		if (spot_to_buffer(process_input_starts_at) <= 0)
			process_input_starts_at = alloc_spot();
		*process_input_starts_at = start;
		point = start;
		to_begin_line();
		if (process_coloring_rules & 1)
			set_character_color(point, start, color_class process_prompt);
		try_grab_directory_name(point, start);
	}
}

// Return RGB for ANSI codes 0..7.  Optionally adjust foreground colors
// to show up better on default white background.
ansi_color_number_to_rgb(int code, int foreground)
{
	int bgcolor = (alter_color(color_class process_output, -1) >> 4) & 7;
	int light_background = foreground && (process_coloring_rules & 4)
						   && (bgcolor == 7);

	switch (code) {
		case 1:		return red;
		case 2:		return green;
		case 3:		return light_background ? light_olive : yellow;
		case 4:		return blue;
		case 5:		return magenta;
		case 6:		return cyan;
		case 7:		return light_background ? light_grey : white;
		default:	return black;
	}
}

// Return current_output_color modified to reflect this ANSI m code.

// Per ECMA-48 standard, codes are: 0=default, 1=bold, 2=faint,
// 3=italic, 4=underlined, 5=slow blink, 6=fast blink, 7=reversed,
// 8=hidden, 9=crossed-out, 10=main font, 11-19=alt fonts, 20=gothic
// font, 21=double underline, 22=non-bold/faint, 23=not ital/gothic,
// 24=not underlined, 25=not blinking, 27-29=undo 7-9, 30-37
// foreground color black, red, green, yellow, blue, magenta, cyan,
// white, 39 default fg color, 40-49 background colors similar.

update_ansi_color_class(int val)
{
	int fg, bg, style, class;

	class = current_output_color ? current_output_color
								 : color_class process_output;
	fg = get_foreground_color(class);
	bg = get_background_color(class);
	style = fg & EFONT_STYLE;
	fg &= ~EFONT_STYLE;
	if (val == 0) {
		return 0;
	} else if (val >= 30 && val <= 37)
		fg = ansi_color_number_to_rgb(val - 30, 1);
	else if (val >= 40 && val <= 47)
		bg = ansi_color_number_to_rgb(val - 40, 0);
	else if (val == 1)
		style |= EFONT_BOLD;
	else if (val == 3)
		style |= EFONT_ITALIC;
	else if (val == 4)
		style |= EFONT_UNDERLINED;
	else if (val == 22)
		style &= ~EFONT_BOLD;
	else if (val == 23)
		style &= ~EFONT_ITALIC;
	else if (val == 24)
		style &= ~EFONT_UNDERLINED;
	else if (val == 39)
		fg = black;
	else if (val == 49)
		bg = white;
	else
		return current_output_color;
	return define_color_class(NULL, fg | style, bg);
}

// Delete ANSI color setting commands, typically produced by the "ls"
// command, and replace them with buffer coloring.

color_ansi_output(int start)
{
	char num[100];
	int val, orig = start, lastcolor = current_output_color;

	save_var point = start - 1;	// For Unix, color user input after system echos it.
	if ((get_character_color(point) == color_class process_prompt
		 || get_character_color(point) == color_class process_input)
		  && curchar() != '\n')
		set_character_color(start, give_end_line(),
							color_class process_input);
	point = start - 25;
	save_var case_fold = 0;
	while (re_search(1, "\x1B%[[0-9;]*m")) {
		int begin = matchstart, end = matchend;
		point = matchstart + 2;
		while (parse_string(1, "[0-9]*[;m]", num)) {
			val = numtoi(num);
			current_output_color = update_ansi_color_class(val);
			if (character(point - 1) == 'm')
				break;
		}
		point = end;
		delete(begin, end);
		if (point > orig)
			set_character_color(orig, point, lastcolor ? lastcolor
				: color_class process_output);
		lastcolor = current_output_color;
		orig = point;
	}
	set_character_color(orig, point, lastcolor ? lastcolor
											   : color_class process_output);
}

color_ansi_buffer()
{
	save_var current_output_color;
	color_ansi_output(0);
}

remove_xterm_sequences(int start)
{
	save_var point = start - 25, narrow_start = point;
	save_var case_fold = 0;
	simple_re_replace(1, "\x1B%][0-9];.*\7", "");
	simple_re_replace(1, "\x1B%[K", "");
}

// If the current line would be off the bottom of the window, adjust
// the window so it's on the last line instead.  (If we didn't,
// redisplay would center the window around this line.)

line_up_to_bottom()
{
	int new_window_start = prev_screen_line(text_height() - 1);

	if (new_window_start > window_start)
		window_start = new_window_start;
}

// Interpreting xxxx\b\b\b to yield x would mean we'd hide some styles
// of progress messages.  So we look only for things like xxxx\b\b\byyy,
// converting to xyyy.

process_backspace_sequences(int start)
{
	save_var point = start;
	while (re_search(1, "\b![^\n\b]")) {
		int limit = parse_string(1, "[^\n\b]+");
		int before = parse_string(-1, "\b+");
		if (before > limit)
			before = limit;
		point -= before;
		limit = parse_string(-1, "[^\n\b]+");
		if (limit < before)
			before = limit;
		if (!before)
			break;
		delete(point - before, point + before);
	}
}

process_remove_controls(int start)
{
	save_var point = start, narrow_start = start;
	simple_re_replace(1, "\x1B\x5B[&=<>?]?[-0-9;]*[a-zA-Z@]", "");
	simple_re_replace(1, "[\x01-\x08\x0B-\x1A\x1C-\x1F]", "");
}

process_unicode_conversion(int start)
{
	save_var point;
	perform_unicode_conversion(bufnum, start, size(), CONV_TO_16, "UTF-8-No-BOM");
}

// The process has just produced some output.

do_when_process_output(int buf, int start, int end)
{
	save_var bufnum = buf;
	save_var abort_searching = 0;
	set_character_color(start, end, color_class process_output);
	if (process_coloring_rules & 64)
		process_unicode_conversion(start);
	if (process_coloring_rules & 32)
		remove_xterm_sequences(start);
	if (process_coloring_rules & 2)
		color_ansi_output(start);
	if (process_coloring_rules & 8)
		process_backspace_sequences(start - 100);
	if (process_coloring_rules & 16)
		process_remove_controls(start);
	notice_password_prompt(start);
	limit_buffer_size(process_max_size);
	save_var position_window_on_screen_line = 100;
	if (buf_in_window(buf) >= 0)
		refresh();
}

command process_mode()
{
	mode_keys = process_tab;
	major_mode = _process_mode_name;
	tab_size = process_tab_size;
	discardable_buffer = 1;
	recolor_range = color_process_range;	// set up coloring rules
	recolor_from_here = recolor_from_top;
	when_setting_want_code_coloring();	// maybe turn on coloring
	fill_mode = 0;
	try_calling("process-mode-hook");
	make_mode();
}

command process_enter() on process_tab['\r']
{
	int set_err = 1;

	if ((process_enter_whole_line & 3) == 2 && point < type_point) {
		char line[FNAMELEN];

		to_begin_line();
		point += parse_string(1, "[^>$#%]*[>$#%] *");
		parse_string(1, ".*", line);
		point = size();	// grab current line excluding prompt
		stuff(line);	// and copy to the end
		return;
	}
	if ((process_enter_whole_line & 3) && point >= type_point)
		to_end_line();
	if (parse_string(-1, "Terminate batch job.*"))
		set_err = 0;
	enter_key();

// Pressing Enter should usually set the error spot.  When shouldn't it?
// 1. When you've pre-typed the next compilation command in the middle
// of a previous compile?
// 2. When you've just hit enter without typing anything?
//	if (point > type_point + 1 && (spot_to_buffer(error_spot) != bufnum
//								   || point < *error_spot))
// 3. When you're confirming aborting a batch file in Windows.

	if (set_err)
		set_error_spot(), (*error_spot)--;
	_last_was_grep = 0;
	if (point >= window_end && process_output_to_window_bottom)
		line_up_to_bottom();
}

process_get_previous_cmd(int down)
{
	int buf = bufnum;
	char cmd[16384], hist_buf[FNAMELEN];

	sprintf(hist_buf, "-process-history-%s", bufname);
	point = size();
	if (!exist(hist_buf))
		return;
	*cmd = 0;
	if (has_arg) {
		iter = 0;
		save_var bufname = hist_buf;
		if (point == size())
			point--;
		to_begin_line();
		restore_vars();
		if (!get_choice(name_to_bufnum(hist_buf), cmd,
						"Previous commands", "Select a previous command",
						"OK", "Cancel", ""))
			return;
	}
	if (spot_to_buffer(process_input_starts_at) <= 0)
		process_input_starts_at = alloc_spot();
	if (lines_between(*process_input_starts_at, point))
		*process_input_starts_at = type_point;
	if (!lines_between(*process_input_starts_at, point))
		delete(*process_input_starts_at, point);
	else
		*process_input_starts_at = point;
	if (*cmd)
		stuff(cmd);
	else {
		save_var bufname = hist_buf;
		down_or_up_lines(down);
		buf_xfer(buf, give_begin_line(), give_end_line());
	}
	restore_vars();
	if (process_enter_whole_line & 4)
		point = *process_input_starts_at;
}

command process_previous_cmd() on process_tab[ALT('p')]
{
	process_get_previous_cmd(0);
}

command process_next_cmd() on process_tab[ALT('n')]
{
	process_get_previous_cmd(1);
}

// Like normal kill-word, but don't include prompt.
command process_backward_kill_word() on process_tab[ALT(CTRL('H'))],
										process_tab[NUMALT(GREYBACK)]
{
	int start;

	check_modify(bufnum);
	start = point;
	backward_word();
	if (spot_to_buffer(process_input_starts_at) > 0
		  && *process_input_starts_at < start
		  && point < *process_input_starts_at)
		point = *process_input_starts_at;
	do_save_kill(start, point);
}

command process_yank()
{
	int buf, lines;
	char msg[FNAMELEN];

	get_clipboard();
	buf = name_to_bufnum(_cur_kill_buf);
	lines = buf_exist(buf) ? count_lines_in_buf(buf, 0) : 0;
	sprintf(msg, "Are you sure you want to yank %d line%s "
			"for immediate execution? ", lines, (lines == 1) ? "" : "s");
	if (!in_macro() && _len_def_mac <= 0 && point >= type_point
		  && (process_yank_confirm > 0 && lines >= process_yank_confirm)
		  && is_process_buffer(bufnum))
		if (!ask_yn("Confirm yank", msg, "Yes", "No", 0))
			return;
	delete_if_highlighted();
	retrieve(0);
}

when_loading()
{
	fix_key_table(reg_tab, (short) yank, process_tab, (short) process_yank);
}

// Search for an Epsilon for Unix helper program and run it on the named
// file.  First look in ~/.epsilon, then on path.  If not found, run the
// version in the executable's directory.  Optionally pass additional
// arguments for the program.  Return nonzero if OK.
int run_helper(char *prog, char *file, char *moreargs)
{
	char cmd[FNAMELEN], quoted[FNAMELEN];

	quote_for_shell(quoted, file);
	sprintf(cmd, "~/.epsilon/%s %s %s >/dev/null 2>&1",
			prog, quoted, moreargs);
	if (!shell("", cmd, ""))	// Run user's version; fail silently.
		return 1;		// If we failed, look on path.
	sprintf(cmd, "%s %s %s >/dev/null 2>&1", prog, quoted, moreargs);
	if (!shell("", cmd, ""))	// Search along the path; fail silently.
		return 1;		// If we failed, try again with std copy we provide.
	get_executable_file(cmd, prog, 1);
	sprintf(cmd + strlen(cmd), " %s %s", quoted, moreargs);
	if (!shell("", cmd, ""))
		return 1;
	return 0;
}

int do_run_viewer(char *fname, char *op, char *dirname)
{
	char orig_dir[FNAMELEN];
	int ret;

	if (opsys != OS_UNIX || is_unix_flavor == IS_UNIX_CYGWIN)
		return run_viewer(fname, op, dirname);
	getcd(orig_dir);
	if (dirname && *dirname)
		chdir(dirname);
	if (!stricmp(op, "open"))
		op = "";	// Use the default operation for this.
	ret = run_helper("epsilon-viewer", fname, op);
	chdir(orig_dir);
	return ret;
}

// Write to dest the full pathname of the file prog in the same
// directory as Epsilon's executable, optionally inside quotes.
get_executable_file(char *dest, char *prog, int quoted)
{
	char path[FNAMELEN];

	get_executable_directory(path);
	convert_to_8_3_filename(path);
	strcat(path, prog);
	if (quoted)
		quote_for_shell(dest, path);
	else
		strcpy(dest, path);
}	

// Transfer the list of matches into the process buffer, adding
// formatting.

proc_complete_show_matches()
{
	int buf = name_to_bufnum(_MATCH_BUF), orig = bufnum;

	columnize_buffer_text(buf, text_width(), 3);
	save_spot point;
	if (point >= type_point)
		point = type_point - 1;
	to_begin_line();
	buf_xfer_colors(buf, point, size());
	save_var bufnum = buf;
	insert('\n');
	buf_xfer_colors(orig, 0, size());
}

// Prepare a list of matches of pat in the match buffer.  If initial is
// nonzero, use special rules for getting command name matches from the
// path.  If skip_dots, ignore . and .. dirs.

int proc_complete_list_matches(char *pat, int initial,
							   int skip_dots, int flags)
{
	int proc_tab_size = tab_size;
	char *pat2 = 0, file[FNAMELEN], *s, *t, cmdnames = 0;

	if (initial && (s = getenv("PATH")) && !strchr(pat, '/') &&
		  (opsys == OS_UNIX || !strchr(pat, '\\') && !strchr(pat, ':'))) {
		pat2 = malloc(strlen(s) + strlen(pat) + 100);
		if (opsys == OS_UNIX)
			sprintf(pat2, "{%s}/%s*", s, pat);
		else
			sprintf(pat2, "{%s,.}/%s*.{%s}", s, pat,
					process_completion_windows_programs);
		while (s = strchr(pat2, path_list_char))
			*s = ',';
		flags |= EXACTONLY;
		pat = pat2;
		cmdnames = 1;
	}
	zap(_MATCH_BUF);
	char last[FNAMELEN], *orig;
	*last = 0;
	if (s = do_file_match(pat, STARTMATCH | flags))
		do {
			if (opsys == OS_UNIX && initial && lowaccess(s, LOWACC_X))
				continue;
			orig = s;
			if (cmdnames) {
				strcpy(file, s);
				s = get_tail(file, 0);
				if (opsys != OS_UNIX)
					*get_extension(s) = 0;
			}
			if (skip_dots && *get_tail(s, 0) == '.'
				  && string_matches_pattern(get_tail(s, 0), "%.%.?[/\\]?<eof>"))
				continue;
			if (opsys == OS_WINDOWS && (cygwin_filenames & 64))
				while (t = strchr(s, '\\'))
					*t = '/';
			buffer_printf(_MATCH_BUF, "%s\n", s);
			strcpy(last, orig);
		} while ((s = do_file_match(pat, 0)) && !user_abort);
	free(pat2);	// Free if we allocated it.
	save_var bufname = _MATCH_BUF;
	save_var abort_searching = 0;
	do_sort_region(0, size(), 0, 0);
	do_uniq(1, 1, 0);
	tab_size = proc_tab_size;
	int cnt = lines_between(0, size(), 0);
	if (cnt == 1 && (process_completion_style & 8) && *last) {
		absolute(last);
		say("Completed: %s", last);
	}
	return cnt;
}

// Prepare a list of matches of pat in the match buffer.  If initial is
// nonzero, use special rules for getting command name matches from the
// path.  Also ensure we use process's directory.

int proc_complete_do_list_matches(char *pat, int initial,
								  int skip_dots, int flags)
{
	char orig_dir[FNAMELEN], dir[FNAMELEN];
	int cnt;

	if (!can_get_process_directory)
		error("Can't determine process's current directory.");
	getcd(orig_dir);
	if (*process_current_directory) {
		strcpy(dir, process_current_directory);
		chdir(dir);
	} else
		getcd(dir);
	delayed_say(1, 200, 100, "Listing matches for %s...", pat);
	cnt = proc_complete_list_matches(pat, initial, skip_dots, flags);
	delayed_say(1, 0, 0, "");
	chdir(orig_dir);
	return cnt;
}

// Is the current file name on this line inside an open "?  If so,
// return nonzero and with matchend set to just before it. Also check
// for ' under Unix.

process_line_inside_quoting()
{
	if (opsys == OS_UNIX
		  && parse_string(-1 , "^([^'\n]|'[^'\n]*'|)*!'[^'\n]*"))
		return 1;
	return parse_string(-1, "^([^\"\n]|\"[^\"\n]*\"|)*!\"[^\"\n]*");
}

// Prepare a list of env var name matches of pat in the match buffer.
int proc_complete_list_envvars(char *var)
{
	char pat[FNAMELEN];
	save_var bufnum = zap(_MATCH_BUF);
	get_substitution_variables(1);
	save_var abort_searching = 0, show_status = 0;
	point = 0;
	simple_re_replace(1, "=.*", "");	// Discard values.
	sprintf(pat, "^%q", var);
	if (strcmp(var, "*"))
		do_drop_matching_lines(REGEX, pat, 0); // Discard non-matching lines.
	do_sort_region(0, size(), 0, 0);
	return lines_between(0, size(), 0);
}

command process_complete() on process_tab['\t'], process_tab[NUMSHIFT(GREYTAB)]
{
	char pat[FNAMELEN], pref[FNAMELEN];
	int cnt = 0, start, end, buf = create(_MATCH_BUF), envvars = 0;
	int offset = IS_SHIFT_KEY(full_key) ? -1 : 1;

	start = end = point;
	if (start < type_point) {
		if (!maybe_indent_rigidly(0))
			indent_like_tab();
		return;
	}
	*pat = 0;
	if (!(process_completion_style & 16)
		  && parse_string(-1, (opsys == OS_UNIX) ? "%$%{?(<ident>*)" : "%%(<ident|(|)>*)")) {
		grab(start = find_group(1, 1), find_group(1, 0), pat);
		envvars = start - matchend;	// 2 if ${, 1 if $ or %
		point = end;
	} else if (process_line_inside_quoting() || parse_string(-1, "[^\n\t <>=\"]+", pat)) {
		start = point;
		point = end;
		if (process_line_inside_quoting())
			grab(start = matchend + 1, end, pat);
		if (start < type_point && type_point < end) {
			start = type_point;		// exclude prompt
			grab(start, end, pat);
		}
	}
	if (!*pat)
		strcpy(pat, "*");
	if (!(process_completion_style & 1) || prev_cmd != CMD_PROC_COMPL) {
		int dircmdlen = matches_at_length(start, -1, process_completion_dircmds);
		int dirsonly = dircmdlen && ((dircmdlen == start - type_point)
					 || matches_at(start - dircmdlen, -1, "[][#&;()] *"));
		int initial = all_blanks(type_point, start)
					  || matches_at(start, -1, "([&;|]|</word>(sudo)) +");
		if (envvars)
			cnt = proc_complete_list_envvars(pat);
		else
			cnt = proc_complete_do_list_matches(pat, initial,
												process_completion_style & 2,
												dirsonly ? FM_ONLY_DIRS : 0);
		check_abort();
		offset = 0;
		if (!cnt)
			error("No matches for %s.", pat);
		set_buf_point(buf, 0);
	}
	if (!(process_completion_style & 1))
		find_buffer_prefix_folding(name_to_bufnum(_MATCH_BUF), pref, !is_unix);
	else {
		buf_delete_final_newlines(buf);
		grab_line_offset(buf, pref, offset, 1);
	}
	int needs_quoting = strpbrk(pref, opsys == OS_UNIX ? " \\" : " ") && !envvars;
	int quotechar = opsys == OS_UNIX ? '\'' : '"';
	if ((process_completion_style & 1) || strlen(pref) > end - start) {
		delete(start, end);
		if (needs_quoting && !process_line_inside_quoting())
			insert(quotechar);
		stuff(pref);
		if ((process_completion_style & 1)) {
			if (is_path_separator(character(point - 1)))
				delete(point - 1, point);
			this_cmd = CMD_PROC_COMPL;
			return;
		}
		if (cnt == 1 && !is_path_separator(character(point - 1))) {
			if (needs_quoting && process_line_inside_quoting())
				insert(quotechar);
			stuff(!envvars ? " " : (opsys != OS_UNIX) ? "%" : envvars == 2 ? "}" : "");
		}
	} else if (prev_cmd == CMD_PROC_COMPL) {
		proc_complete_show_matches();
		this_cmd = CMD_PROC_COMPL;
	} else if (cnt == 1 && !is_path_separator(character(point - 1))) {
		if (needs_quoting && process_line_inside_quoting())
			insert(quotechar);
		stuff(" ");
	} else if (!(process_completion_style & 4)
			   && buffer_size(_MATCH_BUF) < screen_cols - 20 - strlen(pat)
			   && buffer_size(_MATCH_BUF) < ptrlen(pref) - 2) {
		save_var bufname = _MATCH_BUF;
		simple_re_replace(1, "\n", " ");
		grab(0, size(), pref);
		say("%d matches for %s: %s", cnt, pat, pref);
		expire_message = 1;
	} else {
		say("%d matches for %s: press again to see matches.", cnt, pat);
		expire_message = 1;
		this_cmd = CMD_PROC_COMPL;
	}
}

set_error_spot()	/* maybe allocate, then set to point */
{
	if (spot_to_buffer(error_spot) == bufnum)
		*error_spot = point;
	else {
		if (spot_to_buffer(error_spot) > 0)
			free_spot(error_spot);
		error_spot = alloc_spot();
	}
}

// Run a process (cmdline), pass it a buffer of input, retrieve
// its output.  Either the input or output buffer names may be NULL.
// Start the process in curdir (or Epsilon's current directory, if
// curdir == NULL).  Returns 0 and sets errno if the function could
// not be started, returns 1 if the function started successfully.

// The PIPE_SYNCH flag means don't return from the function until the
// process has finished.  The PIPE_CLEAR_BUF flag means empty the
// output buffer before inserting the process's text (but do nothing
// if the process can't be started); it's convenient when the input
// and output buffers are the same, to filter a buffer in place.
// PIPE_NOREFRESH keeps Epsilon from automatically displaying the
// output buffer as it receives text.

// If handler is nonzero, it's the index of a function to call each
// time text is received from the process, and when the process
// terminates.  There must be an output buffer for a handler to be
// called.

pipe_text(char *input, char *output, char *cmdline, char *curdir,
			 int flags, int handler)
{
	int inbuf = 0, outbuf = 0;

	if (input && *input)
		inbuf = create(input);
	if (output && *output) {
		outbuf = create(output);
		save_var bufnum = outbuf;
		if (handler)
			when_activity = handler;
		restore_vars();
	}
	return buf_pipe_text(inbuf, outbuf, cmdline, curdir ? curdir : "",
						 flags);
}

command filter_region() on reg_tab[ALT('|')]
{
	int in_place = (iter > 1), b, width, orig = bufnum;
	char cmd[FNAMELEN], bname[FNAMELEN];

	get_str_auto_def(cmd, in_place ? "Send region to command" :
			   "Filter region through command");
	strcpy(bname, "filter-output");
	if (iter < 0)	// Iter > 1: in place, iter < 1: ask where; iter 1: default buffer
		get_buf(bname, "Into buffer", bname);
	b = zap(bname);
	iter = 0;
	if (region_type() == REGRECT) {
		width = extract_rectangle(b, 0);
	} else {
		save_spot point, mark;
		fix_region();		// handle other types of regions
		buf_xfer(b, point, mark);
	}
	if (!buf_pipe_text(b, b, cmd, "", PIPE_CLEAR_BUF
					   | ((in_place | (process_echo & 8)) ? PIPE_SYNCH : 0))) {
		if (bufnum != b)
			buf_delete(b);
		file_error(errno, "During exec", "Couldn't exec");
		quick_abort();
	}
	if (in_place) {
		if (region_type() == REGRECT) {
			save_var bufnum = b;
			xfer_rectangle(orig, width, 1);
		} else {
			delete(point, mark);
			restore_vars();
			grab_buffer(b);
			fix_region();
		}
	} else {
		tiled_only();
		to_buffer_num(b);
		point = 0;
	}
	highlight_off();
}

command shell_command() on reg_tab[ALT('!')]
{
	int in_place = (iter > 1), b;
	char cmd[FNAMELEN], bname[FNAMELEN];

	get_str_auto_def(cmd, in_place ? "Insert output of shell command" :
					 "Run shell command");
	strcpy(bname, "filter-output");
	if (iter < 0)	// Iter > 1: in place, iter < 1: ask where; iter 1: default buffer
		get_buf(bname, "Into buffer", bname);
	b = zap(bname);
	iter = 0;
	if (!buf_pipe_text(0, b, cmd, "", PIPE_CLEAR_BUF | (in_place ? PIPE_SYNCH : 0))) {
		if (bufnum != b)
			buf_delete(b);
		file_error(errno, "During exec", "Couldn't exec");
		quick_abort();
	}
	if (in_place) {
		grab_buffer(b);
		fix_region();
	} else {
		to_buffer_num(b);
		point = 0;
	}
	highlight_off();
}
