/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2021 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 14.01 to 14.02.

#include "eel.h"
#include "c.h"
#include "colcode.h"
#include "diff.h"
#include "html.h"
#include "perl.h"
#include "python.h"

c_indenter()		/* replace surrounding indentation with new */
{
	if (c_extra_keywords & IDL_KEYWORDS)
		return;
	if (in_shell_buffer)
		prev_indenter();
	else if (!(in_c_comment(c_look_back) & IN_COMMENT))
		indent_to_column(c_compute_indent());
	else if ((sublanguage_code == SUBLANG_PHP ? reindent_php_comments :
			  in_perl_buffer ? reindent_perl_comments : reindent_c_comments)
			 & (indenter_action == IACT_REINDENT_PREV ? 4 :
				(this_cmd == CMD_INDENT_REG) ? 16 : 1))
		indent_c_comment();
}

is_label(p)	/* does line at p contain only case: or label? */
{
	save_var point = p;
	while (parse_label())	/* go past any labels */
		point = matchend;
	return parse_string(1, "[ \t\f]*(\n|/<*>|//)");
}

indent_case(orig)	// indent case-like statement at orig
{
	char pat[FNAMELEN];

	sprintf(pat, "</word>case</word>|(</word>(default|public|"
			"private|protected%s)[ \t\n]*:)|[{}]",
			(c_extra_keywords & CSHARP_KEYWORDS) ? "|internal" : "");

	point = orig;
	for (;;) {		// look for case-like statement, or {}'s
		cind_search(pat);
		if (in_c_comment(c_look_back))
			continue;
		else if (curchar() != '}')
			break;
		point++;
		if (!c_rev_skip_level('{', '}')) // move before matching {
			longjmp(&c_indent_giveup, 1);
	}
	if (curchar() != '{')	// copy indentation from prior case
		return get_indentation(point);

	cind_space();	// skip over switch-like construct before {
	if (character(point - 1) == ')') { // back past () of switch
		if (!c_rev_skip_level('(', ')'))
			longjmp(&c_indent_giveup, 1);
	} else
		to_statement_start();
	to_indentation();
	past_labels();	// might be "case 3: switch ()", get to switch
	c_skip_fwd_space(0);
	sprintf(pat, "(public|private|protected%s)</word>",
			(c_extra_keywords & CSHARP_KEYWORDS) ? "|internal" : "");
	return current_column() + get_c_indent() +
			(matches_at(orig, 1, pat)
			 ? c_access_spec_offset : c_case_offset);
}

// Move forward past spaces, tabs, optionally newlines, and C/C++
// comments.
c_skip_fwd_space(int skip_newlines)
{
	for (;;)
		switch (curchar()) {
			case '\n':
				if (!skip_newlines)
					return;
			case ' ': case '\t':
				point++;
				break;
			case '/':
				switch (character(point + 1)) {
					case '/':
						to_end_line();
						break;
					case '*':
						point + 2;
						search(1, "*/");
						break;
					default:
						return;
				}
				break;
			default:
				return;
		}
}

// Return 1 if these files match according to compare-files rules., 2 if
// they match except for whitespace, or 0 if they don't match at all.
int files_match(char *f1, struct file_info *i1,
				char *f2, struct file_info *i2)
{
	int t1, t2, transl = FILETYPE_AUTO, orig = bufnum;

	if (!compare_files_method)
		return !compare_dates(i1, i2);
	if (compare_files_method == 1) {
		transl = FILETYPE_BINARY;
		if (i1->fsize != i2->fsize)
			return 0;
	}
	save_var abort_file_io = ABORT_JUMP;
	t1 = tmp_buf();
	t2 = tmp_buf();
	on_exit buf_delete(t1);
	on_exit buf_delete(t2);
	save_var bufnum = t1;
	quiet_file_read(f1, transl);
	save_var narrow_start = narrow_end = 0;
	bufnum = t2;
	quiet_file_read(f2, transl);
	save_var narrow_start = narrow_end = 0;
	bufnum = orig;
	if (compare_files_ignore_spaces) {
		char *pat = (compare_files_ignore_spaces == 1) ? "[ \t]+" : "[ \t\n]+";
		if (buffers_match_ignoring(t1, t2, pat))
			return buffers_identical(t1, t2) ? 1 : 2;
		else
			return 0;
	}
	return buffers_identical(t1, t2);
}

// Generate a diff in out.  Return 0 if a file wasn't found.
compare_files_diff(char *file1, char *file2, char *out, int style)
{
	struct screen_info wins;
	int err1 = 0, err2 = 0, temp1 = 0, temp2 = 0, had_narrow1 = 0, had_narrow2 = 0;

	save_var synchronous_file_io = 1;
	save_var readonly_warning |= ROWARN_GREP, no_popup_errors = 1;
	readonly_warning &= ~(ROWARN_MSG | ROWARN_BELL);
	wins.wins = 0;
	save_screen(&wins);
	on_exit if (temp1 && buf_exist(temp1)) buf_delete(temp1);
	on_exit if (temp2 && buf_exist(temp2)) buf_delete(temp2);
	on_exit restore_screen(&wins);
	one_window();
	if (!look_file(file2)) {
		err2 = std_find_it(file2);
		if (!err2)
			temp2 = bufnum;
	} else
		window_bufnum = bufnum;
	had_narrow2 = narrow_start > 0 || narrow_end > 0;
	save_var narrow_start = narrow_end = 0;
	window_split(HORIZONTAL);
	if (!look_file(file1)) {
		err1 = std_find_it(file1);
		if (!err1)
			temp1 = bufnum;
	} else
		window_bufnum = bufnum;
	had_narrow1 = narrow_start > 0 || narrow_end > 0;
	save_var narrow_start = narrow_end = 0;
	if (!err1 && !err2)
		do_diff(out, style);
	if (err1)
		file_error(err1, file1, "read error");
	else if (err2)
		file_error(err2, file2, "read error");
	else {
		save_var bufname = out;
		save_var point = 0;
		point += parse_string(1, "Comparing.*\nwith.*\n");
		if (had_narrow1)
			bprintf("%pIgnored narrowing for %s\n", -1, file1);
		if (had_narrow2)
			bprintf("%pIgnored narrowing for %s\n", -1, file2);
		diff_mode();
		return 1;
	}
	return 0;
}

diff_view_both() on diff_tab['c']
{
	diff_view_both_helper(HORIZONTAL);
}

diff_view_side_by_side() on diff_tab['s']
{
	diff_view_both_helper(VERTICAL);
}

do_set_show_graphic(mode)	// set graphic rep by mode code
{				// 0: multi-char, 1: graphic, 2: hexadecimal, 3: int'l
	int i, std, stdctrl, replacement, dispwidth;

	if (show_spaces && _display_class == _show_disp_class)
		save_var _display_class = _std_disp_class;
	std = (mode == 0) ? BM : (mode == 2) ? BHEX : BNORMAL;
	stdctrl = (mode == 3) ? BC : std;
	for (i = 0; i < 128; i++)
		if (i < 32 || i == 127) {
			if (i != '\n' && i != '\t')
				_display_class[i] = mode ? stdctrl : BC;
			replacement = set_character_property(i + 128, CPROP_DEF_REPL, -1);
			dispwidth = set_character_property(replacement, CPROP_DISP_WIDTH, -1);
			if (mode == 0 || dispwidth == 0 && mode != 2)
				_display_class[i + 128] = BMC;
			else if (mode != 2 && replacement != i + 128
					 && set_character_property(i + 128, CPROP_DISP_WIDTH, -1) == 0)
				// If orig char was zero-width but repl isn't, use repl.
				_display_class[i + 128] = replacement;
			else
				_display_class[i + 128] = std;
		} else {
			_display_class[i] = BNORMAL;
			_display_class[i + 128] = std;
		}
	for (i = 128; i < 160; i++) {
		replacement = set_character_property(i, CPROP_DEF_REPL, -1);
		dispwidth = set_character_property(replacement, CPROP_DISP_WIDTH, -1);
		_echo_disp_class[i] = dispwidth > 0 ? replacement : BMC;
	}
	for (i = 256; i  < 65536; i++)
		_display_class[i] = (mode == 2) ? BHEX : BNORMAL;
	fix_display_class();
	make_show_spaces();
	full_redraw = 1;
}

guess_mode_without_extension(char *res, char *pat)
{
	if (!strfcmp(get_tail(filename), "makefile")
		  || !strnfcmp(get_tail(filename), "makefile.", 9)) {
		sprintf(res, pat, "mak");	// Recognize special file names
		return 1;					// w/o .mak extension.
	}
	save_var point = 0;
	if (parse_string(1, "#%!.*/(env )?perl")) {
		sprintf(res, pat, "perl");
		return 1;
	}
	if (parse_string(1, "#%!.*/(env )?python")) {
		sprintf(res, pat, "python");
		return 1;
	}
	if (parse_string(1, "[ \t\n]*%<%??(%!DOCTYPE[ \t]+html|html|head|%!--)")) {
		sprintf(res, pat, "html");
		return 1;
	}
	if (parse_string(1, "[ \t\n]*%<%??(%!DOCTYPE|xml|xst)")) {
		sprintf(res, pat, "xml");
		return 1;
	}
	if (parse_string(1, "#%!.*/php|(.*\n)?[ \t\n]*%<%??(php)")) {
		sprintf(res, pat, "php");
		return 1;
	}
	if (parse_string(1, "(:.*\n)?#%!.*/(env )?"
					 "[a-z]?(sh|csh|[bd]ash|ksh|tcsh|zsh)</word>|:\n")
		  || !strnfcmp(get_tail(filename), ".profile", 8)) {
		sprintf(res, pat, "sh");
		return 1;
	}
	if (parse_string(1, "#%!.*/(env )?(node|nodejs)")) {
		sprintf(res, pat, "js");
		return 1;
	}
	if (parse_string(1, "%$title ")) {
		sprintf(res, pat, "gms");
		return 1;
	}
	if (parse_string(1, "%%%!PS-Adobe-")) {
		sprintf(res, pat, "ps");
		return 1;
	}
	if (size() > 1000)		// Look in 1st 1000 chars only.
		save_var narrow_end = size() - 1000;
	if (parse_string(1, "[ \t\n]*(//|/%*|#)")
		&& re_search(1, "^[ \t]*(#define|#include)[ \t]+")) {
		// It starts with a comment or preprocessor line, and it's
		// got a #define or #include, so assume it's a C++ header
		// file.
		sprintf(res, pat, "hpp");
		return 1;
	}
	point = 0;
	if (GET_LINE_TRANSLATE(translation_type) == FILETYPE_UNIX
		  && parse_string(1, "([ \t]*\n)*#")
		  || !strfcmp(get_tail(filename), ".epsilon_vars")) {
		sprintf(res, pat, "conf");
		return 1;
	}
	return 0;
}

when_setting_epsilon_manual_port()
{
	char file[FNAMELEN], port[100];
	byte port_bytes[100];
	get_customization_directory(file);
	strcat(file, "epsilon-manual-port");
	sprintf(port, "%d", epsilon_manual_port);
	chars_to_bytes(port_bytes, port);
	int f = lowopen(file, 3);
	lowwrite(f, port_bytes, strlen(port));
	lowclose(f);
}

// Recolor just this region of css.  Last colored region may go past to.
color_css_range(from, to)
{
	int t = -1, talk, s, talked = 0;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend, case_fold = 1;
	while (from > narrow_start
		   && get_tagged_region("color-as-unit", from, &s) != -1)
		from = s - 1;
	if (from < to)
		set_character_color(from, to, -1);
	point = from;
	to_begin_line();
	talk = point + 10000;	// show status after coloring this many chars
	while (point < to) {
		if (!re_search(1, "[-0-9A-Za-z_.@%#]+|[\"']|/%*|%<%!--|-->")) {
			t = size();
			break;
		}
		t = matchstart;
		if (character(t) == '<') {
			set_character_color(t, point, color_class html_embedded);
			continue;
		}
		switch (character(point - 1)) {		// check last char
			case '>':	
				set_character_color(t, point, color_class html_embedded);
				break;
			case '*':			// found comment
				search(1, "*/");
				set_character_color(t, point, color_class css_comment);
				break;
			case '"':		// found a string literal
				point = t;
				re_search(1, "\"([^\"\\\n]|\\(.|\n))*[\"\n]");
				set_character_color(t, point, color_class css_string);
				break;
			case '\'':		// found a char const
				point = t;
				re_search(1, "\'([^\'\\\n]|\\(.|\n))*[\'\n]");
				set_character_color(t, point, color_class css_string);
				break;
			default:		// found identifier.
				if (character(matchstart) == '@' && matches_at(t, 1, "(@<word>+([^{\"]*)*)[{\"]")) {
					point = matchend - 1;		// Found a CSS rule like @media or @import
					set_character_color(t, point, color_class css_class);
				} else if (parse_string(1, "((([ \t\n]*[][,:.]?([ \t\n]*"
								 "[-0-9A-Za-z_.:@%#,]+)*)[ \t\n:().<>])*)[ \t\n]*{")) {
					point = find_group(1, 0);
					set_character_color(t, point, color_class css_selector);
				} else if (parse_string(1, "[ \t\n]*:"))
					set_character_color(t, point, color_class css_class);
				else
					set_character_color(t, point, -1);
				break;
		}
		if (point > talk) {
			note("Coloring CSS: %d%% complete...",
				 muldiv(point - from, 100, to - from));
			talk = point + 5000;
			talked = 1;
		}
	}
	if (to < t)
		set_character_color(to, t, -1);
	if (talked)
		note("");
	return point;
}

int give_html_indent_by_type(int type)
{
	switch (type) {
		case HT_JSCRIPT:
			return get_c_indent();
		case HT_VBSCRIPT:
			return vbasic_step();
		case HT_PHP:
			return (php_indent > 0) ? php_indent : tab_size;
		case HT_PYTHON:
			return python_indent;
		case HT_CSS:
			return css_indent;
	}
	return give_html_indent();
}

/* are all chars in this region comments or whitespace? */
int all_comments(int from, int to, int type)
{
	save_var point, narrow_end = size() - to;
	int single_char = -1, slash_slash = 1, slash_star = 1;	// Start with JavaScript comment rules.
	switch (type) {
		case HT_VBSCRIPT:	// Only ' comments
			slash_slash = slash_star = 0;
			single_char = '\'';
			break;
		case HT_PHP:		// C comments plus #
			single_char = '#';
			break;
		case HT_PYTHON:	// Only # comments
			single_char = '#';
			slash_slash = slash_star = 0;
			break;
		case HT_CSS:		// Only /* */ comments.
			slash_slash = 0;
			break;
	}
	for (point = from; point < to; )
		if (isspace(curchar()))
			point++;
		else if (curchar() == single_char || curchar() == '/' && slash_slash && character(point + 1) == '/')
			nl_forward();
		else if (curchar() == '/' && slash_star && character(point + 1) == '*') {
			point += 2;
			if (!search(1, "*/"))
				return 0;
		} else
			return 0;
	return 1;
}

do_html_indent(int manual)
{
	int region, start;

	UNREFERENCED_PARAMETER(manual);
	recolor_buffer_range(point, point + 200);
	if (parse_string(1, "[ \t]*[%?]%>")
		  && get_tagged_region("color-as-unit", matchend - 2) == 99) {
		indent_to_column(html_indent_script_end(character(matchend - 2)));
		return;
	}
	region = get_tagged_region("color-as-unit",
							   point + parse_string(1, "[ \t]*"), &start);
	if (sublanguage_code == SUBLANG_CSS)
		region = HT_CSS;
	int var = (sublanguage_code == SUBLANG_XML) ? xml_auto_indent
			  : html_auto_indent; 
	if (IS_ENTER_KEY(key) && !(var & (region == HT_JSCRIPT ? 2
		: region == HT_VBSCRIPT ? 4 : region == HT_PHP ? 8
		: region == HT_PYTHON ? 16 : region == HT_CSS ? 32 : 1))) {
		if (!(var & 0x4000))
			prev_indenter();
		return;
	}
	save_var narrow_start = start;
	if (region > 0 && region <= HT_CSS && !IS_ENTER_KEY(key)
		  && !(html_indenting_rules & (1 << (region - 1)))) {
		prev_indenter();	// Can move back to prev line's indent.
		return;
	}
	if (parse_string(1, "[ \t]*%<%!%[CDATA"))
		region = 0;
	if (region >= HT_JSCRIPT && region < HT_IGNORE) {
		if (all_comments(narrow_start, point, region)) {	// Start of script block.
			narrow_start = 0;	// 0x400 indent; 0x800 indent even if <% has none.
			int i = get_indentation(start);
			if ((html_indenting_rules & 0x400)
				  && ((html_indenting_rules & 0x800) || i))
				indent_to_column(i + give_html_indent_by_type(region));
			return;
		}
		html_exclude_script_head();
	}
	switch (region) {
		case HT_JSCRIPT:
			save_var c_extra_keywords |= JS_KEYWORDS;
			do_c_indent();
			return;
		case HT_VBSCRIPT:
			do_vbasic_indent();
			return;
		case HT_PHP:
			do_php_indent();
			return;
		case HT_PYTHON:
			if (this_cmd != CMD_INDENT_REG)
				do_python_indent();
			return;
		case HT_CSS:
			do_css_indent();
			return;
	}
	narrow_start = 0;
	if (sublanguage_code != SUBLANG_XML && html_open_pre(point) >= 0)
		return;		// Inside open <pre> block.
	indent_to_column(html_indent_level(manual));
}

// Color this range without changing narrowing.  Color more if
// necessary.
color_full_perl_range(from, to)
{
	int end;

	do {
		from--;		// Get to uncolored section.
	} while (from > narrow_start
			 && (get_character_color(from, &from) != -1
				 || get_tagged_region("color-perl-as-unit", from) != -1));
	set_character_color(from, to, -1);
	end = color_this_perl_range(from, to);
	// It looks like the region ends in the middle of some syntax
	// element, due to the number of "'s changing or similar.  Color
	// more.
	if (get_character_color(end) != -1) {
		// Recolor the part we need to show right now.
		end = color_this_perl_range(end, must_color_through);
		// Mark the rest of the buffer so it gets recolored later.
		set_tagged_region("needs-color", end, size() - narrow_end, -1);
	}
	return end;
}

// Is this << very likely to be the start of a here doc, not a left
// shift operator? If its right-hand value contains EOF, assume it is.
perl_likely_here_doc(int pos)
{
	save_var case_fold = 0;
	return matches_at(pos, 1, "%<%<[ '\"]*<ident>*EOF<ident>*[ '\"]*;");
}

color_this_perl_range(from, to)
{
	int t = -1, talk, c, talk_at = 0, after_here_doc = 0;
	char *langname;
	TIMER talk_now;

	save_var point, matchstart, matchend, case_fold = 0;
	if (talk = (to - from > 30000))	// Show status during long delays.
		time_begin(&talk_now, 50);		// But wait a bit before chatter.
	point = from;
	langname = sublanguage_code == SUBLANG_PHP ? "PHP"
			   : in_shell_buffer ? "Shell" : "Perl";

	while (point < to) {
		if (!re_search(1, "[0-9A-Za-z_]+|[\"'`/#$&<=\n]")) {
			t = size();
			break;
		}
		t = matchstart;
		switch (c = character(point - 1)) {		// check last char
			case '#':			// found comment
				if (in_shell_buffer && !parse_string(-1, "([ \n\t|&;()<>]|^)#"))
					break;
				to_end_line();
				set_character_color(t, point, color_class perl_comment);
				break;
			case '&':
				if (in_shell_buffer)
					break;
				if (parse_string(1, "((<word>+('|::))*|'|::)<word>+"))
					set_character_color(t, point = matchend,
										color_class perl_function);
				break;
			case '$':
				if (in_shell_buffer && strchr("\"'", curchar())) {
					set_character_color(t, point, color_class perl_string);
					break;			// Translated string.
				}
				if (in_shell_buffer) {
					if (curchar() == '(') {
						shell_parse_substitution(point - 1);
						break;
					}
					if (parse_string(1, "<word>+|%{([^{}\n]|%{[^{}\n]*%})+%}"
									 "|[^ \t\n`(]"))
						set_character_color(t, point = matchend,
											color_class perl_variable);
					break;
				}
				if (sublanguage_code == SUBLANG_PHP) {
					if (parse_string(1, "<word>+((::|-%>)<word>+)*"
									 "|%{([^{}\n]|%{[^{}\n]*%})+%}|[^\n`]"))
						set_character_color(t, point = matchend,
											color_class perl_variable);
					break;
				}
				if (!isident(curchar()) && !strchr("#{}", curchar())) {
					if (character(point++) == '^')
						point++;	// $', $?, $^Z or similar special var
					set_character_color(t, point, color_class perl_variable);
				} else if (parse_string(1, "[#$@]*(((<word>+('|::))*|'|::)<word>+|[-+])")) {
					set_character_color(t, point = matchend,
										color_class perl_variable);
				} else if (parse_string(1, "#?{"))
					set_character_color(t, point = matchend - 1,
										color_class perl_variable);
				break;
			case '\'':
				if (in_shell_buffer) {
					if (!(matches_at_length(t, -1, "\\+") & 1)) {
						search(1, "'");	// No \-quoting in '' for shell mode.
						set_character_color(t, point, color_class perl_string);
					}
				} else if (!strchr("&$@%*", character(t - 1)))
					color_perl_fwd_quote(c, t, 0);	// old syntax pkg'func
				break;
			case '"':		// found a string literal
			case '`':
				if (in_shell_buffer && (matches_at_length(t, -1, "\\+") & 1))
					break;
				color_perl_fwd_quote(c, t, 1);
				break;
			case '/':
				if (sublanguage_code == SUBLANG_PHP) {
					if (curchar() == '/') {
						nl_forward();
						set_character_color(t, point, color_class perl_comment);
					} else if (curchar() == '*') {
						search(1, "*/");
						set_character_color(t, point, color_class perl_comment);
					}
					break;
				}

				if (curchar() == '/') {  // Found defined-or operator //
					point++;
					break;
				}
				// if previous thing looks like an expression, skip it
				if (!in_shell_buffer && !perl_after_expression(t, 1)) {
					point--;
					color_perl_delimiters(0, 1, 2);
				}
				break;
			case '<':
				if (sublanguage_code == SUBLANG_PHP) {	// Look for <<<word
					if (parse_string(1, "%<%<(<alpha|_><word>*)"))
						color_perl_here_doc(t, 1);
					break;
				}
				if ((perl_after_expression(t, 0) || character(t - 1) == '<')
					  && !perl_likely_here_doc(point - 1))
					break; // It's a shift or lessthan op, not "here" doc or input op.
				if (curchar() == '<') {		// found <<, possible "here" document.
					point++;	// Next must be an identifier or text in quotes.
					if (parse_string(1, in_shell_buffer
									 ? "[ \t]*[-\\]?(<alpha|_><word>*)"
									 : "~?(<alpha|_><word>*)"))
						color_perl_here_doc(t, 1);
					else if (parse_string(1, "~?[ \t]*\"([^\"\n]+)\""))
						color_perl_here_doc(t, 1);
					else if (parse_string(1, "~?[ \t]*'([^'\n]+)'"))
						color_perl_here_doc(t, 0);
					else if (parse_string(1, "~?[ \t]*\\(<alpha|_><word>*)"))
						color_perl_here_doc(t, 0);
					else if (parse_string(1, "~?[ \t]*`([^`\n]+)`"))
						color_perl_here_doc(t, 1);
					else
						break;
					after_here_doc = 1;
					break;
				}
				if (in_shell_buffer || sublanguage_code == SUBLANG_PHP)
					break;
				point--;	// File glob operator.
				color_perl_delimiters(0, 1, 0);
				break;
			case '\n':
				if (after_here_doc)
					if (get_tagged_region("color-perl-as-unit", point) != -1)
						to_end_line();	// Already colored this line.
					else
						after_here_doc = 0;
				break;
			case '=':			// Found POD (documentation).
				if (in_shell_buffer || sublanguage_code == SUBLANG_PHP)
					break;
				if (parse_string(-1, "^=")) {
					search(1, "\n=cut");
					set_character_color(t, point, color_class perl_comment);
				}
				break;
			default:		// found identifier, kywd, or number
				color_perl_word(t);
				break;
		}
		if (talk && point > talk_at + 20000 && time_done(&talk_now)) {
			note("Coloring %s program: %d%% complete...",
				 langname, muldiv(point - from, 100, to - from));
			talk_at = point;
		}
	}
	if (talk_at)
		note("");
	return point;
}

load_binding()		/* new-style binding declaration */
{
	char name[80], junk[80];
	int *onekey, start;
	short *tab;
	int i, now_type, cnt = 1;

	point++;			/* skip initial ~ */
	parse_string(1, "[^ \t\n:]+", name);
	i = find_index(name);
	now_type = (i > 0) ? name_type(i) : -1;
	if (now_type != NT_TABLE)
		init_error("%s is not a table.", name);
	tab = index_table(i);
	point++;			/* skip space before key */
	onekey = get_keycode();
	if (!onekey || onekey[0] != 2)
		init_error("Bad key name.%s", "");
	start = onekey[1];
	if (parse_string(1, " \"")) {
		point++;
		onekey = get_keycode();
		if (!onekey || onekey[0] != 2)
			init_error("Bad final key name.%s", "");
		cnt = onekey[1] - start + 1;
	}
	if (!parse_string(1, ": ", junk) || !parse_string(1, ".*", name))
		init_error("Bad key binding.%s", "");
	i = find_index(name);
	if (i <= 0) {		/* allow name table index too */
		i = numtoi(name);
		if (!i && *name != '0') {	/* not a name, not a number */
			init_warning("No such command %s.", name);
			return;
		}
	}
	set_range(tab, start, i, cnt);
	rebuild_menu();
}

when_loading()
{
	update_display_class();
	epsilon_manual_port = 7567;
}
