/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2020 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 14 B6 to 14 B7.

#include "eel.h"
#include "c.h"
#include "chelp.h"
#include "colcode.h"
#include "dired.h"
#include "grep.h"
#include "html.h"
#include "hex.h"
#include "info.h"
#include "makemode.h"
#include "misclang.h"
#include "rmode.h"
#include "perl.h"
#include "proc.h"
#include "session.h"
#include "tags.h"
#include "tex.h"

void hscroll_color_for_area(int start, int end)
{
	int i, from, to, next, max = text_height();
	int last_column = text_width() + display_column;
	save_var point;
	for (i = 0; i < max; i++) {		// For each screen line of this window
		start = window_line_to_position(i);
		if (start >= end)	// If its start is past the region to color, done.
			break;
		// To avoid an expensive move_to_column(), we first check if
		// this line is already entirely colored.
		if (!get_tagged_region("needs-color", start, &from, &to)) {
			// The start of this line is already colored and it's part
			// of the already-colored region from...to.
			next = window_line_to_position(i + 1);	// Start of next screen line.
			if (next == -1 && i == max - 1)	// Last screen line, so we can't get
				next = end;		// start of the next one -- substitute.
			if (next != -1 && next <= to)
				continue;		// Entire line is already colored.
		}
		point = start;		// Find buffer pos of the last column we need to color.
		move_to_column(last_column);
		if (curchar() == '\n')
			point++;
		color_for_area(start, point);
	}
}

// Like above, but don't disturb narrowing.
color_for_narrowed_area(from, to)
{
	int i, j = 0, give_up = 0, last_j;

	save_var point, abort_searching = 0;
	init_color_code();
	to = MIN(to, size() - narrow_end);
	for (;;) {
		last_j = j;
		if (get_tagged_region("needs-color", from, &i, &j)) {
			if (i < from) {
				point = from;		// From isn't colored, so start before it.
				color_go_back_lines();	// Color a few extra lines before
										// to speed up scrolling.
				if (point > i) // Next, find starting place known to be on 
					i = recolor_from_here(i); // the edge of a colored region.
			}				// (Or use known safe place if it's closer.)
		} else {
			if (j >= to) {		// From is colored, see if good region extends
				return;			// far enough.  If not, get following bad reg.
			} else if (!get_tagged_region("needs-color", j, &i, &j)) {
				int a, b, res;

				say("Coloring internal error 1");
				buffer_printf("debug-color", "Coloring internal error 1:\n");
				buffer_printf("debug-color", "bad color 0 from %d to %d in %s, size %d\n",
							  i, j, bufname, size());
				res = get_tagged_region("needs-color", from, &a, &b);
				buffer_printf("debug-color", "color %d at %d (from %d to %d) in %s, size %d\n",
							  res, from, a, b, bufname, size());
				res = get_tagged_region("needs-color", j - 1, &a, &b);
				buffer_printf("debug-color", "color %d at %d (from %d to %d) in %s, size %d\n",
							  res, j - 1, a, b, bufname, size());
				return;
			}
		}
		j = MIN(j, to);		// Color only some of it.
		if (i >= j)
			return;
		j = call_recolor_range(i, j);
		set_tagged_region("needs-color", i, j, 0);
		if (j <= last_j && give_up++ > 100) {
			say("Coloring internal error 2");
			when_displaying = 0;
			return;
		}
	}
}

int call_recolor_range(int from, int to)
{
	if (want_debug_coloring)		// Clean slate to see just what it colors.
		set_tagged_region("debug-color", 0, size(), -1);
	int ret = recolor_range(from, to);
	if (want_debug_coloring)
		check_debug_coloring(from, to, ret);
	return ret;
}

#undef set_character_color
// Record that we're set some color for this region.
debug_set_character_color(int from, int to, int color)
{
	set_character_color(from, to, color);
	if (want_debug_coloring) {
		set_tagged_region("debug-color", from, to, 1);
		if (want_debug_coloring & 16) {
			char bname[FNAMELEN];
			sprintf(bname, "debug-color-%s", bufname);
			create(bname);
			buffer_to_end(bname, -1);
			buffer_printf(bname, "Painted %d...%d in ", from, to);
			char *name = name_color_class(color);
			if (name)
				buffer_printf(bname, "%s\n", name);
			else
				buffer_printf(bname, "color class %d\n", color);
		}
	}
}

// A coloring function was asked to color from...to, and reported that
// it colored from...ret. Check if it really did. If not, report the
// bug.
check_debug_coloring(int from, int to, int got)
{
	char bname[FNAMELEN];
	sprintf(bname, "debug-color-%s", bufname);
	int start, end, buf = -1, ok = 0, p, i;
	if (get_tagged_region("debug-color", from, &start, &end) == 1
		  && end >= got)
		ok = 1;
	if (!ok || (want_debug_coloring & 8)) {
		buf = create(bname);
		buffer_to_end(bname, -1);
	}
	if (ok) {		// No errors detected.
		if (want_debug_coloring & 8)
			buf_printf(buf, "Colored %d...%d, went to %d, OK\n", from, to, got);
		return;
	}
	save_var narrow_start = 0, narrow_end = 0;
	say("Possible coloring bug detected, see report in buffer %s", bname);
	buf_printf(buf, "Colored %d...%d, went to %d, some regions not colored:\n", from, to, got);
	p = from;
	do {
		i = get_tagged_region("debug-color", p, &start, &end);
		if (i == -1) {	// Missed
			buf_printf(buf, "  %d...%d was not colored\n", start, end);
			if (want_debug_coloring & 2)
				set_character_color(start, end, color_class debug_text);
		} else if (want_debug_coloring & 4)
			buf_printf(buf, "  %d...%d was colored\n", start, end);
		p = end;
	} while (p < to);
}

// The EEL debugger's r command calls this. Call recursive-edit, but 
// preserve key, true_key, bufnum, window_bufnum, window_handle, and be
// careful restoring since the user might have deleted buffers or
// windows.
recursive_edit_from_debug()
{
	int obuf = bufnum, trouble = 0;
	save_var key, full_key;
	if (!recursive_edit_preserve())
		trouble = 1;
	if (buf_exist(obuf))
		bufnum = obuf;
	else 
		trouble = 1;
	if (trouble)
		say("Warning: previous buffer or window was deleted during recursive edit!");
}

command find_linked_file()
{
	char file[FNAMELEN], curdir[FNAMELEN], subst[FNAMELEN];

	*file = 0;
	if (is_highlight_on()) {
		save_var point, mark;
		fix_region();
		if (mark - point < FNAMELEN - 1 && !lines_between(point, mark, 0))
			grab(point, mark, file);
	}
	if (!*file)
		find_linked_file_here(file);
	get_buffer_directory(curdir);
	if (filename_subst_envvars(file, subst))
		strcpy(file, subst);
	if (!is_remote_file(file))
		absolute(file, curdir);
	quiet_set_bookmark();
	tiled_only();
	locate_window("", file);
	// If it's the name of an existing local file but contains extended
	// file pattern chars, treat them as normal characters for now so
	// we don't get a dired.
	if (!is_remote_file(file) && is_pattern(file) == 1 && check_file(file))
		save_var file_pattern_wildcards = 0;
	do_find(file, give_line_translate(file));
}

guess_mode_without_extension(char *res, char *pat)
{
	if (!strfcmp(get_tail(filename), "makefile")
		  || !strnfcmp(get_tail(filename), "makefile.", 9)) {
		sprintf(res, pat, "mak");	// Recognize special file names
		return 1;					// w/o .mak extension.
	}
	save_var point = 0;
	if (parse_string(1, "#%!.*/(env )?perl")) {
		sprintf(res, pat, "perl");
		return 1;
	}
	if (parse_string(1, "#%!.*/(env )?python")) {
		sprintf(res, pat, "python");
		return 1;
	}
	if (parse_string(1, "[ \t\n]*%<%??(%!DOCTYPE[ \t]+html|html|head|%!--)")) {
		sprintf(res, pat, "html");
		return 1;
	}
	if (parse_string(1, "[ \t\n]*%<%??(%!DOCTYPE|xml|xst)")) {
		sprintf(res, pat, "xml");
		return 1;
	}
	if (parse_string(1, "[ \t\n]*%<%??(php)")) {
		sprintf(res, pat, "php");
		return 1;
	}
	if (parse_string(1, "(:.*\n)?#%!.*/(env )?"
					 "(sh|csh|[bd]ash|ksh|tcsh|zsh)</word>|:\n")
		  || !strnfcmp(get_tail(filename), ".profile", 8)) {
		sprintf(res, pat, "sh");
		return 1;
	}
	if (parse_string(1, "#%!.*/(env )?(node|nodejs)")) {
		sprintf(res, pat, "js");
		return 1;
	}
	if (parse_string(1, "%$title ")) {
		sprintf(res, pat, "gms");
		return 1;
	}
	if (parse_string(1, "%%%!PS-Adobe-")) {
		sprintf(res, pat, "ps");
		return 1;
	}
	if (size() > 1000)		// Look in 1st 1000 chars only.
		save_var narrow_end = size() - 1000;
	if (parse_string(1, "[ \t\n]*(//|/%*|#)")
		&& re_search(1, "^[ \t]*(#define|#include)[ \t]+")) {
		// It starts with a comment or preprocessor line, and it's
		// got a #define or #include, so assume it's a C++ header
		// file.
		sprintf(res, pat, "hpp");
		return 1;
	}
	point = 0;
	if (GET_LINE_TRANSLATE(translation_type) == FILETYPE_UNIX
		  && parse_string(1, "([ \t]*\n)*#")
		  || !strfcmp(get_tail(filename), ".epsilon_vars")) {
		sprintf(res, pat, "conf");
		return 1;
	}
	return 0;
}

color_conf_range(from, to)
{
	int t = -1;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend;
	point = to;				// Color entire lines.
	nl_forward();
	to = point;
	save_var narrow_end = 0;
	set_character_color(from, to, -1);
	point = from;

	while (point < to) {
		if (!re_search(1, "([^ \t\n#:;,]+[ \t]*)="
					   "([ \t]*([^ \t\n#:;,]+|'[^\n']*'|\"[^\n\"]*\"))|^;|[#]")) {
			t = size();
			break;
		}
		t = matchstart;
		switch (character(point - 1)) {		// check last char
			case '#':
			case ';':
				nl_forward();
				set_character_color(t, point, color_class conf_comment);
				break;
			default:
				set_character_color(t, find_group(1, 0),
									color_class conf_attribute);
				set_character_color(find_group(2, 1), point,
									color_class conf_value);
				break;
		}
	}
	if (to < t)
		set_character_color(to, t, -1);
	return point;
}

color_batch_range(from, to)
{
	int t = -1;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend, case_fold = 1;
	point = to;				// Color entire lines.
	nl_forward();
	to = point;
	save_var narrow_end = 0;
	set_character_color(from, to, -1);
	point = from;

	while (point < to) {
		if (!re_search(1, "[-a-z0-9_.%/\\]+|[<>|]|^::")) {
			t = size();
			break;
		}
		t = matchstart;
		if (strchr("<>|", character(point - 1))) {
			set_character_color(t, point, color_class batch_punctuation);
			continue;
		} else if (character(point - 1) == ':') {
			nl_forward();
			set_character_color(t, point, color_class batch_comment);
			continue;
		}
		switch (is_batch_keyword(t)) {
			case 1:
				set_character_color(t, point, color_class batch_keyword);
				break;
			case 2:
				nl_forward();
				set_character_color(t, point, color_class batch_comment);
				break;
			default:
				set_character_color(t, point, -1);
				break;
		}
	}
	if (to < t)
		set_character_color(to, t, -1);
	return point;
}

color_nsis_range(from, to)
{
	int t = -1;
	char pat[FNAMELEN];

	if (from >= to)
		return to;
	save_var point, matchstart, matchend, case_fold = 1;
	point = to;				// Color entire lines.
	nl_forward();
	to = point;
	save_var narrow_end = 0;
	point = from;
	to_begin_line();
	while (parse_string(-1, "\\\n") && matchend < point) {
		point = matchend;
		to_begin_line();
	}
	from = point;
	set_character_color(from, to, -1);

	while (point < to) {	// Search for special char or word or comment.
		if (!re_search(1, "[;#\"'`]|[^ \n\t;#\"'`/][^ \n\t;#]*|/%*")) {
			t = size();
			break;
		}
		t = matchstart;
		if (t >= to)
			return to;
		if (strchr(";#", character(t))) {
			nl_forward();
			set_character_color(t, point, color_class nsis_comment);
		} else if (character(t) == '/') {
			search(1, "*/");
			set_character_color(t, point, color_class nsis_comment);
		} else if (strchr("\"'`", character(t))) {
			sprintf(pat, "([^\n$%c]|<$>[^\\\n]|<$>\\.|\\\n)*(%c|$)",
					character(t), character(t));
			point += parse_string(1, pat);
			set_character_color(t, point, color_class nsis_string);
		} else if (character(t) == '!') {
			set_character_color(t, point, color_class nsis_preprocessor);
		} else if (is_nsis_keyword(t))
			set_character_color(t, point, color_class nsis_keyword);
		else
			set_character_color(t, point, -1);
	}
	if (to < t)
		set_character_color(to, t, -1);
	return point;
}

// If this .tex file has a mainfile declaration (a line "% mainfile:
// filename.tex" or "% !TeX root=filename.tex" within 3 lines of its
// start or end), fill in mainfile and return 1, else set to "" and
// return 0.
int jump_to_dvi_get_mainfile(char *mainfile)
{
	char fname[FNAMELEN];
	char *pat = "^<hspace>*%%<hspace>*(mainfile:|%!TeX root<hspace>*=)<hspace>*(<^wspace>.*tex)";
	*mainfile = 0;
	save_var narrow_start = 0, narrow_end = 0;
	save_var point = 0;
	move_by_lines(3);
	if (!re_search(-1, pat)) {
		point = size();
		move_by_lines(-4);
		if (!re_search(1, pat))
			return 0;
	}
	grab(find_group(2, 1), find_group(2, 0), fname);
	strcpy(mainfile, filename);
	strcpy(get_tail(mainfile, 1), fname);
	return 1;
}

// The user's specified a command line, so run the command, passing it
// the TeX source file name, line number, and the name of the .dvi or
// .pdf file.

jump_to_dvi_by_command()
{
	char dvi[FNAMELEN], *args[5], lnum[20], cmd[FNAMELEN], buf[FNAMELEN];
	char basename[FNAMELEN], path[FNAMELEN], only_ext[FNAMELEN], msg[FNAMELEN];
	int i = 1;

	*dvi = 0;
	if (!has_arg) {
		if (*jump_to_dvi_main_file)
			strcpy(dvi, jump_to_dvi_main_file);
		if ((!*dvi || check_file(dvi) != CHECK_FILE) && *filename) {
			if (!jump_to_dvi_get_mainfile(dvi))
				strcpy(dvi, filename);
			strcpy(get_extension(dvi), jump_to_dvi_extension);
		}
		if (*dvi && check_file(dvi) != CHECK_FILE)
			*dvi = 0;
	}
	sprintf(only_ext, "|%s|", jump_to_dvi_extension);
	save_var only_file_extensions = only_ext;
	if (!*dvi) {
		sprintf(msg, "%s file to view: ", jump_to_dvi_extension);
		get_file_dir(dvi, msg);
		strcpy(jump_to_dvi_main_file, dvi);
	}
	if (!*dvi) {
		strcpy(dvi, filename);
		strcpy(get_extension(dvi), jump_to_dvi_extension);
	}
	if (check_file(dvi) != CHECK_FILE)
		error("Can't locate file %s.", dvi);
	sprintf(lnum, "%d", lines_between(0, point, 1) + 1);
	strcpy(basename, get_tail(filename, 1));
	*get_extension(basename) = 0;
	strcpy(path, filename);
	*get_tail(path, 1) = 0;
	args[0] = lnum;		// Prep the values to substitute in the template.
	args[1] = filename;
	args[2] = dvi;
	args[3] = basename;
	args[4] = path;
	if (expand_string_template(cmd, jump_to_dvi_command, "lfdbp", args))
		error("Invalid jump-to-dvi-command setting.");
	do {
		sprintf(buf, "-jump-to-dvi-output-%d", i++);
	} while (exist(buf) && is_process_buffer(name_to_bufnum(buf)));
	pipe_text(NULL, buf, cmd, NULL, 0, 0);
	iter = 0;
}

// Go to the place where this text appears in the DVI previewer or PDF
// viewer. The PDF support works with PDF viewers that support SyncTeX.
// The DVI support works with previewers that accept appropriate command
// line parameters, and also via DDE, with Y&Y's old DVIWindo previewer,
// version 2.1.4 and later.

command jump_to_dvi()
{
	int conv;
	char msg[FNAMELEN], dvifile[FNAMELEN];

	if (*jump_to_dvi_command) {	// The usual case.
		jump_to_dvi_by_command();
		return;
	}
	if (!is_mswindows)
		error("You must set jump-to-dvi-command first.");
	iter = 0;							// Using the old DDE method.
	strcpy(dvifile, filename);
	strcpy(get_extension(dvifile), ".dvi");
	conv = dde_open("DVIWindo", "SRCSpecial");
	if (!conv)
		error("Couldn't open DDE channel to DVI viewer; have you set jump-to-dvi-command?");
	if (check_file(dvifile)) {		// The file exists, so open it.
		sprintf(msg, "[Open(\"%s\")]", dvifile);
		if (!dde_execute(conv, msg, 10000))
			say("Error sending open message to DVI viewer.");
	}	// Otherwise assume the user's already opened the file.
	sprintf(msg, "[Source(\"%s\");Line(%d)]",
			get_tail(filename, 1), lines_between(0, point, 0));
	if (!dde_execute(conv, msg, 10000))
		say("Error sending message to DVI viewer.");
	if (!dde_close(conv))
		say("Error closing channel to DVI viewer.");
}

command tex_mouse_select() on tex_tab[MOUSE_LEFT_DN], tex_tab[MOUSE_DBL_LEFT]
{
	if (mouse_shift & KB_CTRL_DN)
		jump_to_dvi();
	else
		mouse_select();
}

when_loading()
{
	if (!strcmp(compile_tex_cmd, "tex --interaction scrollmode \"%r\""))
		strcpy(compile_tex_cmd, "tex --interaction nonstopmode -file-line-error \"%r\"");
	if (!strcmp(compile_latex_cmd, "latex --interaction scrollmode \"%r\""))
		strcpy(compile_latex_cmd, "latex --interaction nonstopmode -file-line-error \"%r\"");
}
