/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2020 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 14 B7 to 14 B8.

#include "eel.h"
#include "c.h"
#include "chelp.h"
#include "colcode.h"
#include "dired.h"
#include "epsnet.h"
#include "grep.h"
#include "html.h"
#include "hex.h"
#include "info.h"
#include "makemode.h"
#include "misclang.h"
#include "rmode.h"
#include "perl.h"
#include "proc.h"
#include "session.h"
#include "tags.h"
#include "tex.h"

command quoted_insert();

short *ask_key(char *pr, char *keyname, int flags)
{
	int index, k, fk, ask_if_ambiguous = flags & 1, warn_if_std = flags & 2;
	short *tab, *alttab;
	char msg[FNAMELEN], *p;

	tab = mode_keys;
	alttab = root_keys;
	*keyname = 0;			// put readable form of key here
	key_binding[0] = 1;		// also save sequence in this global
	for (;;) {
		sprintf(msg, "%s%s", pr, keyname);
		full_getkey(msg, 0);
		if (root_keys[key] == (short) quoted_insert) {
			mention("%s<Quote> ", msg);
			getkey();
			user_abort = 0;
		}
		check_abort();
		p = keyname + strlen(keyname);
		k = full_key;
		index = (tab[k] == 0) ? alttab[k] : tab[k];
		if (index <= 0) {
			k = key;
			index = (tab[k] == 0) ? alttab[k] : tab[k];
		}
		show_char(keyname, ask_if_ambiguous ? k : full_key, 1);
		key_binding[key_binding[0]] = k; // add to constructed key
		key_binding[0]++;
		if (index <= 0 || name_type(index) != NT_TABLE)
			break;
		alttab = index_table((tab[k] && !alttab[k]) ?
							 tab[k] : alttab[k]);
		tab = index_table(index);
		if (!alttab)
			alttab = tab;
		strcat(keyname, " ");
		warn_if_std = 0;
	}
	say("");
	if (warn_if_std && (key >= ' ' && key <= 65535 || key == '\r' || key == '\n')) {
		sprintf(msg, "Are you sure you want to rebind the standard key %k? ", key);
		if (!ask_yn("Confirm rebinding a standard key", msg, "Yes", "No", 0))
			quick_abort();
	}
	if (ask_if_ambiguous && full_key != key) {
		k = key;
		fk = full_key;
		sprintf(msg, "Bind only specific key %k (not general key %k)? ",
				fk, k);
		if (ask_yn("Specific key or general?", msg, "Yes", "No", 2)) {
			key = k = fk;
			*p = 0;
			show_char(keyname, k, 1);
		}
	} else if (full_key != key && k == full_key)
		key = k;
	return ((tab[k] == 0) ? alttab : tab) + k;
}

/* dquote "(" snormal OR bslash anything ")" STAR dquote */
#define C_STRING "\"([^\"\\\n]|\\(.|\n))*\""

/* squote "(" cnormal OR bslash anything ")" STAR squote */
#define C_CCONST "'([^'\\\n]|\\(.|\n))*'"

c_skip_item(item, flags)
{
	char *s; int m;

	if ((c_extra_keywords & CSHARP_KEYWORDS) && item == '"') {
		if (flags & RE_REVERSE) {
			if (parse_string(-1, "@\"([^\"]|\"\")*\"[^\"]*")) {
				point = matchend;	// If this a C# verbatim literal,
				return 1;			// go to its start.
			}
		} else {
			if (!search(1, "\""))
				return 0;	// If this is a C# verbatim literal,
			point--;		// follow its syntax rules.
			return re_search(flags, (character(point - 1) == '@')
							 ? "\"([^\"]|\"\")*\"![^\"]" : C_STRING);
		}
	} else if ((c_extra_keywords & CPP_KEYWORDS) && item == '"'
			   && move_past_cpp_raw_literal(0, (flags & RE_REVERSE)))
		return 1;
	switch (item) {
		case '/':	m = (flags & RE_REVERSE) ? -1 : 1;
					return search(m, (m < 0) ? "*/" : "/*")
							&& search(m, (m < 0) ? "/*" : "*/");
		case '"':	s = C_STRING; break;
		case '\'':	s = C_CCONST; break;
	}
	return re_search(flags, s);
}

// We're displaying the buffer region start...end, so syntax-highlight
// it if needed. But since we're horizontally scrolling, and lines might
// extend far to the right, we'll avoid wasting time coloring those
// parts, by coloring line by line.
void hscroll_color_for_area(int start, int end)
{
	int i, from, to, next, max = text_height();
	int last_column = text_width() + display_column;
	save_var point;
	for (i = 0; i < max; i++) {		// For each screen line of this window
		start = window_line_to_position(i);
		if (start >= end)	// If its start is past the region to color, done.
			break;
		// To avoid an expensive move_to_column(), we first check if
		// this line is already entirely colored.
		if (!get_tagged_region("needs-color", start, &from, &to)) {
			// The start of this line is already colored and it's part
			// of the already-colored region from...to.
			next = window_line_to_position(i + 1);	// Start of next screen line.
			if (next == -1 && i == max - 1)	// Last screen line, so we can't get
				next = end;		// start of the next one -- substitute.
			if (next != -1 && next <= to)
				continue;		// Entire line is already colored.
		}
		point = start;		// Find buffer pos of the last column we need to color.
		move_to_column(last_column);
		if (curchar() == '\n') {
			point++;
			hscroll_color_extra(start);
		}
		color_for_area(start, point);
	}
}

// We're at the start of a line. Move forward over additional lines as
// long as we stay within a certain number of characters of start.
int hscroll_color_extra(int start)
{
	save_var narrow_end = size() - (start + 2000);
	int p = point;
	while (nl_forward())
		p = point;
	point = p;		// Start of last full line in region.
}

int do_color_c_range(int from, int to)
{
	int t = -1, talk, s, talk_at = 0;
	char pat[200];
	TIMER talk_now;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend;
	c_init_color(from, to);
	point = from;
	if (talk = (to - from > 30000))	// Show status during long delays.
		time_begin(&talk_now, 50);		// Wait a bit before chatter.
	save_var case_fold = 0;
	strcpy(pat, "/<*>|//|^[ \t]*#|[\"']");
	if (!minimal_coloring)
		strcat(pat, "|[A-Za-z_][A-Za-z0-9_]*"
			   "|-?%.?[0-9]([A-Za-z0-9._']|[Ee]-)*");
	if (c_extra_keywords & JS_KEYWORDS)
		strcat(pat, "|`");	// JavaScript: `${template} ${literals}` syntax
	while (point < to) {
		if (!re_search(1, pat)) {
			t = size();
			break;
		}
		t = matchstart;
		switch (character(point - 1)) {		// check last char
			case '/':			// found // one-line comment
				while (nl_forward() && character(point - 2) == '\\')
					;		// Continued // comment.
				set_character_color(t, point, color_class c_comment);
				break;
			case '*':			// found /* starting comment
				search(1, "*/");
				set_character_color(t, point, color_class c_comment);
				if (get_character_color(point, (int *) 0, &s)
					  == color_class c_comment && s > to) {
					c_init_color(point, s);
					to = s;	// Comment extended further before; rescan it.
				}
				break;
			case '#':			// found preproc line
				if (matches_at(t, 1, c_color_preproc_if_pattern)) {
					if (locate_by_ifdef(1))
						nl_forward();
					else
						point = size();
					set_character_color(t, point, color_class c_comment);
				} else {
					c_preproc_color();	// move past correct part
					set_character_color(t, point, color_class c_preprocessor);
				}
				break;
			case '"':		// found a string literal
				point = t;
				if ((c_extra_keywords & OBJC_KEYWORDS) && character(t - 1) == '@')
					t--;
				if ((c_extra_keywords & CPP_KEYWORDS) && character(t - 1) == 'R'
					  && move_past_cpp_raw_literal(0, 0))
					t--;	// Color R too.
				else if ((c_extra_keywords & CSHARP_KEYWORDS)
						 && character(t - 1) == '@')
					t--, re_search(1, "\"([^\"]|\"\")*\"![^\"]");
				else
					re_search(1, "\"([^\"\\\n]|\\(.|\n))*[\"\n]");
				set_character_color(t, point, color_class c_string);
				if (get_character_color(point, (int *) 0, &s) == 
					  color_class c_string && s > to)  // fix up after
					c_init_color(point, to = s); // quoted "'s
				break;
			case '\'':		// found a char const
				point = t;
				re_search(1, "\'([^\'\\\n]|\\(.|\n))*[\'\n]");
				set_character_color(t, point, color_class c_charconst);
				break;
			case '`':		// found JavaScript template literal
				color_js_template_literal(t);
				break;
			default:		// found identifier, kywd, or number
				if ((c_extra_keywords & (OBJC_KEYWORDS | JAVA_KEYWORDS))
					&& character(t - 1) == '@')
						t--;
				set_character_color(t, point, c_keyword_color(t));
				if ((c_extra_keywords & IDL_KEYWORDS)	// Special rule for
					  && (point - t == 4)			// uuid's in IDL files.
					  && parse_string(-1, "uuid")
					  && parse_string(1, "%([0-9A-Fa-f]"))
					maybe_color_uuid();
				break;
		}
		if (talk && point > talk_at + 30000 && time_done(&talk_now)) {
			note("Coloring C program: %d%% complete...",
				 muldiv(point - from, 100, to - from));
			talk_at = point;
		}
	}
	c_init_color(to, t);
	if (talk_at)
		note("");
	return point;
}

// Parse a C++ raw string literal. Point+offset is just before the
// opening ". Move to just after the final " and return 1. If invalid
// start, just return 0. If unterminated, move to EOF and return 2.

// If reverse is nonzero, we try instead to parse it backwards, and
// return 0 if it wasn't a raw string literal after all. We expect
// point+offset to be just after the final ", and move to just before
// the starting R on success.
int move_past_cpp_raw_literal(int offset, int reverse)
{
	char delim[FNAMELEN], pat[FNAMELEN];

	if (reverse) {
		int orig = point;
		if (!matches_at(point + offset, -1, "%)([^\n()\\]*)\""))
			return 0;
		grab(find_group(1, 1), find_group(1, 0), delim);
		sprintf(pat, "R\"%q%%(", delim);
		if (re_search(-1, pat))
			return 1;
		point = orig;
		return 0;
	}		
	if (!matches_at(point + offset, 1, "\"([^\n()\\]*)%("))
		return 0;
	grab(find_group(1, 1), find_group(1, 0), delim);
	sprintf(pat, "%%)%q\"", delim);
	return re_search(1, pat) ? 1 : 2;
}

is_c_keyword(p)		// is text in p (must be surrounded by |'s) a keyword?
char *p;
{
	if (strstr("|if|else|for|do|while|return|break|continue|switch|"
		   "case|default|goto|char|short|int|static|volatile|const|"
		   "float|double|long|void|enum|", p))
		return 1;
	if (p[1] == '_' &&
	    strstr("|_asm|_based|_cdecl|_emit|_except|_export|_far|_fastcall|"
		   "_fortran|_huge|_interrupt|_loadds|_near|_pascal|_saveregs|"
		   "_segment|_segname|_stdcall|_syscall|_try|__int64|", p))
		return 1;
	if (!(c_extra_keywords & (EEL_KEYWORDS | CPP_KEYWORDS | JAVA_KEYWORDS |
							  JS_KEYWORDS | IDL_KEYWORDS | CSHARP_KEYWORDS | OBJC_KEYWORDS))
		  && strstr("|restrict|_Alignas|_Alignof|_Atomic|_Bool|_Complex|"
					"_Generic|_Imaginary|_Noreturn|_Static_assert|_Thread_local|"
					"_Decimal128|_Decimal32|_Decimal64|", p))
		return 1;		// C language only
	if ((c_extra_keywords & EEL_KEYWORDS) &&
	    strstr("|keytable|buffer|window|command|on|color_class|"
		   "spot|save_spot|save_var|on_exit|user|zeroed|byte|int32|", p))
		return 1;
	if ((c_extra_keywords & CPP_KEYWORDS) &&
	    strstr("|asm|catch|class|delete|friend|inline|new|operator|"
			   "private|protected|public|template|try|this|virtual|"
			   "throw|_self|and|and_eq|bitand|bitor|bool|compl|const_cast|"
			   "dynamic_cast|explicit|false|mutable|namespace|not|not_eq|"
			   "or|or_eq|reinterpret_cast|static_cast|true|typeid|typename|"
			   "using|wchar_t|xor|xor_eq|", p))
		return 1;
	if ((c_extra_keywords & CPP_KEYWORDS) &&
		  strstr("|alignas|alignof|char8_t|char16_t|char32_t|concept|consteval|"
				 "constexpr|constinit|co_await|co_return|co_yield|decltype|"
				 "export|noexcept|nullptr|requires|static_assert|thread_local|", p))
		return 1;		// Later C++ keywords
	if ((c_extra_keywords & CPP_KEYWORDS) &&
		  strstr("|override|final|import|module|", p))
		return 1;		// C++ "identifiers with special meaning"
	if ((c_extra_keywords & (JAVA_KEYWORDS | JS_KEYWORDS)) &&
	    strstr("|abstract|assert|boolean|byte|catch|class|extends|false|final|"
			   "finally|function|implements|import|in|instanceof|"
			   "interface|native|new|null|package|private|protected|"
			   "public|strictfp|super|synchronized|this|throw|throws|transient|"
			   "true|try|var|with|", p))
		return 1;
	if (!(c_extra_keywords & (JAVA_KEYWORDS | JS_KEYWORDS)) &&
	    strstr("|auto|extern|register|unsigned|signed|sizeof|"
			   "struct|union|typedef|", p))
		return 1;		// C keywords not in Java.
	if ((c_extra_keywords & IDL_KEYWORDS) &&
		strstr("|appobject|auto_handle|bindable|boolean|broadcast|byte|"
			   "callback|coclass|code|comm_status|context_handle|"
			   "control|decode|defaultbind|dispinterface|displaybind|"
			   "dual|enable_allocate|encode|error_status_t|"
			   "explicit_handle|fault_status|handle|handle_t|hidden|"
			   "hyper|idempotent|ignore|immediatebind|import|in|"
			   "include|interface|library|licensed|local|maybe|module|", p))
		return 1;
	if ((c_extra_keywords & IDL_KEYWORDS) &&
		strstr("|ms_union|nocode|nonextensible|notify|object|odl|"
			   "oleautomation|optional|out|pipe|propget|propput|"
			   "propputref|public|ptr|readonly|ref|requestedit|"
			   "restricted|retval|small|source|string|unique|"
			   "usesgetlasterror|v1_enum|vararg|", p))
		return 1;
	if ((c_extra_keywords & CSHARP_KEYWORDS) &&
		  strstr("|abstract|as|base|bool|byte|catch|checked|decimal|delegate|"
				 "event|explicit|false|finally|fixed|foreach|implicit|in|"
				 "interface|internal|is|lock|namespace|new|null|object|"
				 "operator|out|override|params|private|protected|public|"
				 "readonly|ref|sbyte|sealed|sizeof|stackalloc|string|this|"
				 "throw|true|try|typeof|uint|ulong|unchecked|unsafe|ushort|"
				 "using|virtual|", p))
		return 1;
	if ((c_extra_keywords & OBJC_KEYWORDS) &&
		  strstr("|self|super|BOOL|id|Class|nil|YES|NO|NS_DURING|NS_HANDLER|"
				 "NS_ENDHANDLER|in|out|inout|bycopy|byref|oneway|@interface|"
				 "@implementation|@end|@class|@private|@protected|@public|"
				 "@encode|@protocol|@selector|@try|@catch|@finally|@throw|"
				 "@synchronized|Nil|IBAction|IBOutlet|", p))
			return 1;
	return 0;
}

net_during_wait(int cond)
{
	delay(10, cond);
	if (ungot_key == WIN_HELP_REQUEST && menu_command
		  && !strcmp(menu_command, "password-prompt")) {
		if (_doing_input && !is_gui)
			error("Can't prompt for a password within another prompt.");
		ungot_key = -1;
		password_prompt();
	}
	if (ungot_key == -1 && char_avail()) {	// Needed for scp password prompts on non-GUI.
		wait_for_key();
		if (!IS_WIN_PASSIVE_KEY(key) && !IS_MOUSE_KEY(key))
			ungot_key = key;
	}
}	

remote_file_type(char *file)
{
	struct url_parts parts;
	char *s;

	save_var case_fold = 1;
	if (!is_remote_file(file))
		return 1;
	if (!parse_url(file, &parts) || !*parts.host)
		return 2;
	if (!strstr(" ftp scp ", parts.service))
		return 3;
	if (!*parts.fname)
		return 4;
	if (is_pattern(parts.fname) & 2)
		return 5;		// Use dired if it has * or ? chars.
	if (is_path_separator(parts.fname[strlen(parts.fname) - 1]))
		return 6;
	s = index(parts.fname, '~');	// a user's home dir
	if (s && !index(s, '/') && !index(s, '\\'))
		return 7;
	if (!last_index)		// During init, reject synchronous requests
		return 9;			// since we're not ready to prompt yet.
	if (is_ftp_url_a_dir(file))
		return 8;
	return 0;
}

guess_mode_without_extension(char *res, char *pat)
{
	if (!strfcmp(get_tail(filename), "makefile")
		  || !strnfcmp(get_tail(filename), "makefile.", 9)) {
		sprintf(res, pat, "mak");	// Recognize special file names
		return 1;					// w/o .mak extension.
	}
	save_var point = 0;
	if (parse_string(1, "#%!.*/(env )?perl")) {
		sprintf(res, pat, "perl");
		return 1;
	}
	if (parse_string(1, "#%!.*/(env )?python")) {
		sprintf(res, pat, "python");
		return 1;
	}
	if (parse_string(1, "[ \t\n]*%<%??(%!DOCTYPE[ \t]+html|html|head|%!--)")) {
		sprintf(res, pat, "html");
		return 1;
	}
	if (parse_string(1, "[ \t\n]*%<%??(%!DOCTYPE|xml|xst)")) {
		sprintf(res, pat, "xml");
		return 1;
	}
	if (parse_string(1, "[ \t\n]*%<%??(php)")) {
		sprintf(res, pat, "php");
		return 1;
	}
	if (parse_string(1, "(:.*\n)?#%!.*/(env )?"
					 "[a-z]?(sh|csh|[bd]ash|ksh|tcsh|zsh)</word>|:\n")
		  || !strnfcmp(get_tail(filename), ".profile", 8)) {
		sprintf(res, pat, "sh");
		return 1;
	}
	if (parse_string(1, "#%!.*/(env )?(node|nodejs)")) {
		sprintf(res, pat, "js");
		return 1;
	}
	if (parse_string(1, "%$title ")) {
		sprintf(res, pat, "gms");
		return 1;
	}
	if (parse_string(1, "%%%!PS-Adobe-")) {
		sprintf(res, pat, "ps");
		return 1;
	}
	if (size() > 1000)		// Look in 1st 1000 chars only.
		save_var narrow_end = size() - 1000;
	if (parse_string(1, "[ \t\n]*(//|/%*|#)")
		&& re_search(1, "^[ \t]*(#define|#include)[ \t]+")) {
		// It starts with a comment or preprocessor line, and it's
		// got a #define or #include, so assume it's a C++ header
		// file.
		sprintf(res, pat, "hpp");
		return 1;
	}
	point = 0;
	if (GET_LINE_TRANSLATE(translation_type) == FILETYPE_UNIX
		  && parse_string(1, "([ \t]*\n)*#")
		  || !strfcmp(get_tail(filename), ".epsilon_vars")) {
		sprintf(res, pat, "conf");
		return 1;
	}
	return 0;
}

// Get user to select an encoding for writing this file.
pick_write_encoding()
{
	int tmp = tmp_buf();
	int t = GET_LINE_TRANSLATE(translation_type);
	int e = GET_ENCODING(translation_type);
	int def = encoding_from_name(default_write_encoding);

	if (is_window(window_handle) == ISPOPUP)
		error("Choose an encoding for this buffer first.");
	save_var window_bufnum;
	to_buffer_num(bufnum);
	save_var _view_title = "Choose an encoding";
	buf_printf(tmp, "%s\ncontains characters that cannot be represented\n",
			   bufname);
	buf_printf(tmp, "in a disk file without using an encoding.  Please choose an\n");
	buf_printf(tmp, "encoding for this file.  The default encoding shown or any\n");
	buf_printf(tmp, "of the UTF- encodings will preserve all the characters.\n");
	buf_printf(tmp, "See \"Using National Characters\" in the online manual.\n\n");
	buf_printf(tmp, "Here are some of the characters that can't be represented\n");
	buf_printf(tmp, "without an encoding:\n\n");
	list_chars_needing_encoding(tmp);
	if (def < 0)
		def = 2;		// Default to UTF-8-No-BOM
	save_var translation_type = MAKE_TRANSLATE(def, t, 1);
	char *enc = encoding_to_name(GET_ENCODING(translation_type));
	say("Set buffer encoding to: %s", enc ? enc : "");
	view_buf(tmp, 0);
	e = ask_encoding("Set buffer encoding to", 1);
	iter = 1;
	buf_delete(tmp);
	restore_vars();
	if (e < 1)
		e = def;
	translation_type = MAKE_TRANSLATE(e, t, (e || t != FILETYPE_AUTO));
	make_mode();
}

// Flavor bits: 1: in buffers, not files; 2: show grep buffer in next
// window, not this one; 4: search this buffer only and don't ask.
grep_prompt(int flags, int flavor)
{
	char filepat[FNAMELEN], old_cd[FNAMELEN], buf_dir[FNAMELEN];
	jmp_buf *old_level = top_level, this_level;
	int inbufs = (flavor & 1), nextwin = (flavor & 2);
	int thisbuf = (flavor & 4);

	_last_was_grep = 1;
	tiled_only();
	iter = 0;
	error_if_input(1);	/* error if we're in the middle of input */
	flags = get_search_string("List lines with: ",
				  flags | (inbufs ? BUFGREP : FILEGREP));
	flags &= ~(INCR | FAILING | BAD | REVERSE);
	all_must_build_mode = 1;
	if (thisbuf)
		strcpy(filepat, bufname);
	else if (!get_grep_file_showing_path(inbufs, filepat))
		return;
	all_must_build_mode = 1;
	top_level = &this_level;
	getcd(old_cd);
	if (grep_default_directory == 3 ||
		  (grep_prompt_with_buffer_directory == PBD_TEMPCD && wild_pattern(filepat)))
		if (get_buffer_directory(buf_dir)) {
			chdir(buf_dir);
			on_exit chdir(old_cd);
		}
	if (inbufs && (search_in_region & 2) && !strcmp(filepat, bufname)
		  && is_highlight_on()) {
		save_var point, mark;
		fix_region();
		save_var narrow_start = point, narrow_end = size() - mark;
	}
	if (nextwin)
		next_window();
	if (!setjmp(top_level))
		do_grep("grep", default_search_string(flags), flags, filepat, inbufs);
	top_level = old_level;
}

prompt_file_replace(flags, inbufs)
{		// prompt for replacing in files/buffers, then do it
	char *str, with[10000], filepat[FNAMELEN];
	char old_cd[FNAMELEN], buf_dir[FNAMELEN];
	jmp_buf *old_level = top_level, this_level;

	flags = get_replace_text(flags | (inbufs ? BUFGREP : FILEGREP),
				 with);
	if (flags == -1 || !get_grep_file(inbufs, filepat))
		return;
	str = default_search_string(flags);
	top_level = &this_level;
	getcd(old_cd);
	if (grep_default_directory == 3 ||
		  (grep_prompt_with_buffer_directory == PBD_TEMPCD && wild_pattern(filepat)))
		if (get_buffer_directory(buf_dir)) {
			chdir(buf_dir);
			on_exit chdir(old_cd);
		}
	if (!setjmp(top_level))
		do_file_replace(str, with, flags, filepat, inbufs);
	top_level = old_level;
}

// Returns 0 if user clicked on selected text but didn't drag it
// anywhere, 1 for plain click, 2 for selection, 3 for extending
// selection with Shift, or 4 if user dragged and dropped selected text,
// 5 if user clicked on menu or empty echo area or other special, 6 if
// on window border, 7 if a double-click was used to select an item like
// Enter would (but at prompts, we'd usually jump out of the recursive
// edit instead of returning to the caller in that case).
command mouse_select() on reg_tab[MOUSE_LEFT_DN], reg_tab[MOUSE_DBL_LEFT]
{
	int k = key, ret;

	if (mouse_other_handler())
		return 5;
	else if (!in_text_area(mouse_x, mouse_y)) { // clicked in border
		mouse_move_border(0);
		ret = 6;
	} else if (IS_MOUSE_DOUBLE(key) && mouse_dbl_selects) {
		run_enter_key();
		ret = 7;
	} else {
		ret = mouse_selecting();
		mouse_call_hook(k);	// call correct hook for this key
	}
	return ret;
}

// Using left button to select text. Returns 0 if user clicked on
// selected text but didn't drag it anywhere, 1 for plain click, 2 for
// selection, 3 for extending selection with Shift, or 4 if user dragged
// and dropped selected text.
int mouse_selecting()
{
	int orig, by_words = IS_MOUSE_DOUBLE(key), extend, ret = 1;

	save_var mouse_auto_on = 0;	/* movements won't show mouse */
	to_mouse_window(mouse_x, mouse_y, mouse_screen);
	orig = point;
	refresh();
	to_mouse_position(mouse_x, mouse_y);
	extend = (mouse_shift & KB_SHIFT_DN);
	if (!extend) {	/* not extending */
		if (is_highlight_on() && point_in_region(orig)) {
			ret = drag_and_drop(orig) ? 4 : 0;
			shift_selecting = 2;
			return ret;
		}
		mark = point;		/* put one end here */
		if (virtual_space >= 2)
			mark_to_column = cursor_to_column;
		highlight_off();
	} else
		ret = 3;
	refresh();
	select_loop(by_words, 1);
	if (point != mark && ret == 1)
		ret = 2;
	if (interpret_mouse_variable(mouse_selection_copies) && point != mark) {
		if (extend)
			pop_kill();
		if (opsys == OS_UNIX)
			save_var clipboard_format = 1;
		do_copy_region();
		highlight_on();
	}
	if (!extend)
		shift_selecting = 2;
	return ret;
}

/* let user pick a spot to drag text to. Called whenever user clicks on
highlighted text. Return 1 if we dragged the text, or 0 if user didn't
move out of region before releasing mouse button. */
int drag_and_drop(orig)
{
	spot temp_point;
	int region, type = region_type();

	save_var mouse_cursor = drag_and_drop_pointer;
	highlight_off();	// turn off for this window
	temp_point = alloc_spot(); // and display a region that's independent
	*temp_point = orig;	// of point
	region = add_region(temp_point, mark_spot, color_class highlight,
			    type);
	modify_region(region, MRCONTROL, find_index("_dragging_control"));
	save_var _dragging_control = 1;
	select_loop(0, 0);			// wait until user releases key
	set_region_type(type);	// select_loop() changes type
	remove_region(region);	// now free temporary region
	free_spot(temp_point);
	if (point_in_region(orig)) {	// unselect region
		highlight_off();
		mark = point;
		return 0;
	} else {
		drag_drop_xfer(orig);		// moved it somewhere
		highlight_on();
		return 1;
	}
}

command start_process() on cx_tab[CTRL('M')]
{
	char cmd[FNAMELEN * 2], *cmdline, *sh, fixup[FNAMELEN], *p;
	char cust1[FNAMELEN], cust2[FNAMELEN], buf[FNAMELEN];

	save_var clear_process_buffer = (clear_process_buffer & 5) != 0;
	_last_was_grep = 0;
	tiled_only();
	error_if_input(1);
	strcpy(buf, PROCBUF);
	if (has_arg && (has_feature & FEAT_MULT_CONCUR))
		pick_other_proc_buffer(buf);
	else if (no_running(1))
		return;
	iter = has_arg = 0;
	cmdline = get_cmdline("Start Process With Command", cmd);
	create_concur(cmdline, buf);
	locate_window(buf, "");
	sh = getenv("EPSSHELL");
	if (!sh)
		sh = getenv("SHELL");
	if (opsys == OS_UNIX && sh) {
		*cust1 = *cust2 = *fixup = 0;
		if (p = lookpath("epsilon.sh", 0))
			strcpy(cust1, p);
		if (p = lookpath("epsilon.csh", 0))
			strcpy(cust2, p);
		if (!*cust1 || !*cust2)
			sprintf(fixup, "echo \"Customization file not found, "
					"no customizations performed.\">/dev/null");
		else
			switch (get_shell_type_from_name(sh)) {
				case 1:
					sprintf(fixup, ". \"%s\"; set +o emacs", cust1);
					break;
				case 2:
					sprintf(fixup, "source \"%s\"", cust2);
					break;
				default:
					sprintf(fixup, "echo \"Unrecognized shell, no customizations "
							"performed.\">/dev/null");
					break;
			}
		delay(40, 0, bufnum);	// Let's first get a prompt if we can.
		point = type_point;
		bprintf("%s\n", fixup);
	}
	point = size();
	process_mode();
}

// Return 1 for bash-like, 2 for csh-like, 0 otherwise. Arg is full
// path to the shell.
int get_shell_type_from_name(char *sh)
{
	if (string_matches_regex(sh, ".*/(z|ba|.ba|k)?sh", 1))
		return 1;
	if (string_matches_regex(sh, ".*/.?csh", 1))
		return 2;
	return 0;
}

// Prepare a list of env var name matches of pat in the match buffer.
int proc_complete_list_envvars(char *var)
{
	char pat[FNAMELEN];
	save_var bufnum = zap(_MATCH_BUF);
	get_substitution_variables(1);
	save_var abort_searching = 0, show_status = 0;
	point = 0;
	simple_re_replace(1, "=.*", "");	// Discard values.
	sprintf(pat, "^%q", var);
	if (strcmp(var, "*"))
		do_drop_matching_lines(REGEX, pat, 0); // Discard non-matching lines.
	do_sort_region(0, size(), 0, 0);
	return lines_between(0, size(), 0);
}

command process_complete() on process_tab['\t'], process_tab[NUMSHIFT(GREYTAB)]
{
	char pat[FNAMELEN], pref[FNAMELEN];
	int cnt = 0, start, end, buf = create(_MATCH_BUF), envvars = 0;
	int offset = IS_SHIFT_KEY(full_key) ? -1 : 1;

	start = end = point;
	if (start < type_point) {
		if (!maybe_indent_rigidly(0))
			indent_like_tab();
		return;
	}
	*pat = 0;
	if (!(process_completion_style & 16)
		  && parse_string(-1, (opsys == OS_UNIX) ? "%$%{?(<ident>*)" : "%%(<ident|(|)>*)")) {
		grab(start = find_group(1, 1), find_group(1, 0), pat);
		envvars = start - matchend;	// 2 if ${, 1 if $ or %
		point = end;
	} else if (process_line_inside_quoting() || parse_string(-1, "[^\n\t <>=\"]+", pat)) {
		start = point;
		point = end;
		if (process_line_inside_quoting())
			grab(start = matchend + 1, end, pat);
		if (start < type_point && type_point < end) {
			start = type_point;		// exclude prompt
			grab(start, end, pat);
		}
	}
	if (!*pat)
		strcpy(pat, "*");
	if (!(process_completion_style & 1) || prev_cmd != CMD_PROC_COMPL) {
		int dircmdlen = matches_at_length(start, -1, process_completion_dircmds);
		int dirsonly = dircmdlen && ((dircmdlen == start - type_point)
									 || matches_at(start - dircmdlen, -1, "[][#&;()] *"));
		int initial = all_blanks(type_point, start)
					  || matches_at(start, -1, "([&;|]|</word>(sudo)) +");
		if (envvars)
			cnt = proc_complete_list_envvars(pat);
		else
			cnt = proc_complete_do_list_matches(pat, initial,
												process_completion_style & 2,
												dirsonly ? FM_ONLY_DIRS : 0);
		check_abort();
		offset = 0;
		if (!cnt)
			error("No matches for %s.", pat);
		set_buf_point(buf, 0);
	}
	if (!(process_completion_style & 1))
		find_buffer_prefix_folding(name_to_bufnum(_MATCH_BUF), pref, !is_unix);
	else {
		buf_delete_final_newlines(buf);
		grab_line_offset(buf, pref, offset, 1);
	}
	int needs_quoting = strpbrk(pref, opsys == OS_UNIX ? " \\" : " ") && !envvars;
	int quotechar = opsys == OS_UNIX ? '\'' : '"';
	if ((process_completion_style & 1) || strlen(pref) > end - start) {
		delete(start, end);
		if (needs_quoting && !process_line_inside_quoting())
			insert(quotechar);
		stuff(pref);
		if ((process_completion_style & 1)) {
			if (is_path_separator(character(point - 1)))
				delete(point - 1, point);
			this_cmd = CMD_PROC_COMPL;
			return;
		}
		if (cnt == 1 && !is_path_separator(character(point - 1))) {
			if (needs_quoting && process_line_inside_quoting())
				insert(quotechar);
			stuff(!envvars ? " " : (opsys != OS_UNIX) ? "%" : envvars == 2 ? "}" : "");
		}
	} else if (prev_cmd == CMD_PROC_COMPL) {
		proc_complete_show_matches();
		this_cmd = CMD_PROC_COMPL;
	} else if (cnt == 1 && !is_path_separator(character(point - 1))) {
		if (needs_quoting && process_line_inside_quoting())
			insert(quotechar);
		stuff(" ");
	} else if (!(process_completion_style & 4)
			   && buffer_size(_MATCH_BUF) < screen_cols - 20 - strlen(pat)
			   && buffer_size(_MATCH_BUF) < ptrlen(pref) - 2) {
		save_var bufname = _MATCH_BUF;
		simple_re_replace(1, "\n", " ");
		grab(0, size(), pref);
		say("%d matches for %s: %s", cnt, pat, pref);
		expire_message = 1;
	} else {
		say("%d matches for %s: press again to see matches.", cnt, pat);
		expire_message = 1;
		this_cmd = CMD_PROC_COMPL;
	}
}

skip_c_delims(int start, int end, int restore)
{
	int level = 0, c, orig = point, iflevel = 0, skipping = 0;
	char buf[4], pat[80];

	strcpy(buf, "X|\\");
	sprintf(pat, "[%c%c\"']|/<*|/>|(" PREPROC_LINE ")", end, start);
	while (re_search(1, pat)) {
		buf[0] = c = character(point - 1);
		if (c == '\"' || c == '\'') {
			if ((c_extra_keywords & CPP_KEYWORDS) && c == '\''
				  && isalnum(curchar()) && isalnum(character(point - 2))) {
				// Possible C++14 use of ' as digit separator. Check
				// both decimal/binary/octal and hex cases.
				if (parse_string(-1, "</word>(0b)?[0-9']+'") && parse_string(1, "[0-9']+")
					  || parse_string(-1, "</word>0x[0-9a-f']+'") && parse_string(1, "[0-9a-f']+")) {
					point = matchend;
					continue;
				}
			}
			if ((c_extra_keywords & CPP_KEYWORDS) && parse_string(-1, "R\"")
				  && move_past_cpp_raw_literal(-1, 0))
				;	// We've just moved past C++ raw string literal.
			else if ((c_extra_keywords & CSHARP_KEYWORDS)
					 && character(point - 2) == '@')
				re_search(1, "([^\"]|\"\")*\"![^\"]");
			else
				while (re_search(1, buf)
					   && character(point - 1) == '\\')
					point++;
		} else if (c == '*')				/* Found block comment */
			search(1, "*/");
		else if (c == '/')		/* Found c++ one line comment */
			nl_forward();
		else if (c == start) {			/* Found char like (. */
			if (!skipping)
				level++;
		} else if (c == end) {			/* Found char like ). */
			if (!skipping) {
				if (!--level)			/* Found proper end. */
					return 0;
				else if (level < 0) {	/* Found extra closing delim, give up. */
					if (restore)
						point = orig;
					return 2;
				}
			}
		} else if (parse_string(-1, PREPROC_ELSE)) {
			skipping |= (1 << iflevel);	// Found #else or #elif line, skip through #endif.
		} else if (parse_string(-1, PREPROC_ENDIF)) {
			skipping &= ~(1 << iflevel);	// found #endif line, maybe stop skipping.
			if (iflevel > 0)			/* Ignore extra #endif's. */
				iflevel--;
		} else if (parse_string(-1, PREPROC_IF)) {
			iflevel++;				// found #if/#ifdef line
		}
	}
	if (restore)						/* Couldn't find match, give up. */
		point = orig;
	return 1;
}

skip_initializer()		// move past c initializer
{
	int level = 0;

	c_fwd_space();
	if (curchar() == '{')	// don't bother skipping over {}'s,
		return;		// caller does that anyway
	while (re_search(1, "[,;(){}'\"]"))
		switch (character(point - 1)) {
			case '"':
			case '\'':	if ((c_extra_keywords & CPP_KEYWORDS)
							  && matches_at(point - 2, 1, "[0-9a-f]'[0-9a-f]"))
							continue;		// A number like 123'456'789.
						point--;
						c_skip_item(curchar(), RE_FORWARD);
						break;
			case ',':
			case ';':	if (level < 0)
							break;
						point--;
						return;
			case '(':	level--; break;
			case ')':	level++; break;
			case '{':
			case '}':	point--;
						return;
		}
}

command tex_mouse_select() on tex_tab[MOUSE_LEFT_DN], tex_tab[MOUSE_DBL_LEFT]
{
	int had_ctrl = (mouse_shift & KB_CTRL_DN);
	if (mouse_select() <= 1 && had_ctrl)
		jump_to_dvi();
}

when_loading()
{
	strcpy(tex_auto_show_delim_chars, "{[()]}");
}
