/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1996, 2020 Lugaru Software Ltd.  All rights reserved.  *
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

#include "eel.h"
#include "rmode.h"
#include "proc.h"
#include "colcode.h"

// We just found a " or a ' character.  Find the other end and color.

rmode_string_color(int c)
{
	char pat[100];
	int start = point - 1;

	sprintf(pat, "\\.|%c|\n", c);
	while (re_search(1, pat))
		if (character(matchstart) != '\\')
			break;
	set_character_color(start, point, color_class rmode_string);
}

color_rmode_range(from, to) // recolor just this section
{			// last colored region may go past to
	int t = -1, talk, s, talk_at = 0;
	TIMER talk_now;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend;
	if (from < to)
		set_character_color(from, to, -1);
	point = from;
	if (talk = (to - from > 30000))	// Show status during long delays.
		time_begin(&talk_now, 50);		// Wait a bit before chatter.
	while (point < to) {
		if (!re_search(1, "[A-Za-z_][A-Za-z0-9._]*"
					   "|-?%.?[0-9]([A-Za-z0-9._]|[Ee]-)*j?|[\"'#%]")) {
			t = size();
			break;
		}
		t = matchstart;
		switch (character(point - 1)) {		// check last char
			case '#':			// found comment
				nl_forward();
				set_character_color(t, point, color_class rmode_comment);
				break;
			case '"':		// found a string literal
				rmode_string_color('"');
				if (get_character_color(point, (int *) 0, &s) == 
					  color_class rmode_string && s > to)  // fix up after
					if (point < (to = s))	 // quoted "'s
						set_character_color(point, to, -1);
				break;
			case '%':	// Special operator delimited by %'s.
				re_search(1, "[^%]*%%?");
				set_character_color(t, point, color_class rmode_operator);
				break;
			case '\'':
				rmode_string_color('\'');
				if (get_character_color(point, (int *) 0, &s) == 
					  color_class rmode_string && s > to)  // fix up after
					if (point < (to = s))	 // quoted "'s
						set_character_color(point, to, -1);
				break;
			default:		// found identifier, kywd, or number
				set_character_color(t, point, rmode_keyword_color(t));
				break;
		}
		if (talk && point > talk_at + 2000 && time_done(&talk_now)) {
			note("Coloring Rmode program: %d%% complete...",
				 muldiv(point - from, 100, to - from));
			talk_at = point;
		}
	}
	if (to < t)
		set_character_color(to, t, -1);
	if (talk_at)
		note("");
	return point;
}

rmode_paren_follows() // is there whitespace followed by a "(" at point in buffer?
{		// if so, we assume it's a function name
	point += parse_string(1, "[ \t\n]*");
	for (;; point++) {
		switch (curchar()) {
			case ' ': case '\t': case '\n':
				continue;
			case '(':		return 1;
			case '#':		nl_forward();
							break;
			default:		return 0;
		}
	}
}

rmode_keyword_color(from)	// return color for "identifier" from here to point
{			// (something with alpha or digits)
	char buf[500];

	if (point - from > sizeof(buf) / sizeof(*buf) - 10)
		save_var point = from + sizeof(buf) / sizeof(*buf) - 10;
	buf[0] = '|';		// get identifier, between | chars
	grab(from, point, buf + 1);
	if (index("0123456789-", buf[1]) || buf[1] == '.' && isdigit(buf[2]))
		return color_class rmode_number;
	strcpy(buf + point - from + 1, "|");
	if (is_rmode_keyword(buf))
		return color_class rmode_keyword;
	if (color_class rmode_function != color_class rmode_identifier
		  && rmode_paren_follows())
		return color_class rmode_function;
	return color_class rmode_identifier;
}

is_rmode_keyword(p)	// is text in p (must be surrounded by |'s) a keyword?
char *p;
{
	save_var case_fold = 0;
	if (strstr("|if|else|repeat|while|function|for|in|next|break|"
			   "|TRUE|FALSE|NULL|Inf|NaN"
			   "|NA|NA_integer_|NA_real_|NA_complex_|NA_character_|", p))
		return 1;
	return 0;
}


// Does this region have unmatched (, [, or {?
int rmode_open_delimiter(int start, int end)
{
	save_var point = start;
	save_var narrow_end = size() - end;
	while (re_search(1, "![[({]") && get_character_syntax_color(point) == -1)
		if (to_default_move_level(1, "[({", "])}", 1) == -1)
			return 1;
	return 0;
}

// Does this region have unmatched ), ] or }?
int rmode_close_delimiter(int start, int end)
{
	save_var narrow_start = start;
	save_var point = end;
	while (re_search(-1, "[])}]!") && get_character_syntax_color(point - 1) == -1)
		if (to_default_move_level(-1, "])}", "[({", 1) == -1)
			return 1;
	return 0;
}

// If this is a continuation line, move back one line and return 1.
rmode_back_over_continuation_line(int orig)
{
	if (parse_string(-1, "\\\n[ \t]*"))
		point = matchend;
	else if (!rmode_close_delimiter(give_begin_line(), orig) || !nl_reverse())
		return 0;
	return 1;
}

// Indent one more if there's an open delimiter.

do_rmode_indent()
{
	int ind, orig = give_begin_line(), close;

	close = matches_at(orig, 1, "[ \t]*[])}]");
	save_var point;
	do {		// Find prev non-comment or non-blank line.
		if (!nl_reverse())		// First line in the file, no indent.
			return 0;
		to_begin_line();
	} while (parse_string(1, rmode_indent_to_comment ? "[ \t]*\n"
						  : "[ \t]*[#\n]"));
	while (rmode_back_over_continuation_line(orig))
		;
	to_indentation();
	ind = current_column();
	recolor_buffer_range(point, orig);
	// Is there an open delim on prev line?  If so, indent one level
	// more.
	if (rmode_open_delimiter(point, orig))
		ind += rmode_indent;
	else if (close)
		ind -= rmode_indent;
	restore_vars();
	indent_to_column(ind);
}

rmode_indenter() on rmode_tab['\t']
{
	standard_tab_cmd(do_rmode_indent, rmode_indent, 1);
}

command rmode_close() on rmode_tab['}'], rmode_tab[')'], rmode_tab[']']
{		/* only if typed inside indentation, & it might need fixup */
	normal_character();
	if (current_column() - 1 <= get_indentation(point)) {
		save_spot point;
		to_indentation();
		do_rmode_indent();
	}
}


command r_mode()
{
	mode_default_settings();
	mode_keys = rmode_tab;		/* Use these keys. */
	major_mode = _rmode_mode_name;
	compile_buffer_cmd = compile_rmode_cmd;	// can compile this
	auto_indent = 1;
	if (rmode_tab_override > 0)
		tab_size = rmode_tab_override;
	indenter = do_rmode_indent;
	soft_tab_size = rmode_indent;
	strcpy(comment_start, "#[ \t]*");
	strcpy(comment_pattern, "#.*$");
	strcpy(comment_begin, "# ");
	strcpy(comment_end, "");
	recolor_range = color_rmode_range;	// set up coloring rules
	recolor_from_here = recolor_from_top;
	if (want_code_coloring)		// maybe turn on coloring
		when_setting_want_code_coloring();
	if (auto_show_rmode_delimiters)
		auto_show_matching_characters = rmode_auto_show_delim_chars;
	buffer_maybe_break_line = generic_maybe_break_line;
	fill_mode = (misc_language_fill_mode & 0x400) != 0;
	try_calling("rmode-mode-hook");
	drop_all_colored_regions();
	make_mode();
}

when_loading()
{
	rmode_tab[ALT('q')] = (short) fill_comment;
}

// Do R users use the .r extension?
suffix_r()
{
	r_mode();
}

tag_mode_r()
{
	char func[TAGLEN];
	int start, color;

	r_mode();
	recolor_buffer_range(0, size());	// Just to skip over comments
	save_var point = 0, case_fold = 0;				// & strings.
	while (re_search(1, "([a-zA-Z._]<word>*)[ \t]*%<-[ \t]*(function)%(")) {
		color = get_character_color(matchstart);
		if (color == color_class rmode_comment
			  || color == color_class rmode_string)
			continue;
		grab(start = find_group(1, 1), find_group(1, 0), func);
		add_tag(func, start);
	}
}
