/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1989, 2020 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// This file handles remembering the window and buffer configuration
// you used the last time you ran Epsilon.

#include "eel.h"
#include "proc.h"
#include "bufed.h"
#include "dired.h"
#include "session.h"
#include "filter.h"
#include "browse.h"

/* functions for remembering window configurations, which files
   we were editing, etc., from session to session */

#define DEFSESSNAME	"epsilon.ses"

give_session_file(file)
char *file;
{
	char *s, *p, dir[FNAMELEN];

	if (explicit_session_file) {
		strcpy(file, explicit_session_file);
		return;
	}
	s = cmd_line_session_file;
	if (!s)
		s = verenv("ESESSION");
	if (!s && *session_file_name)
		s = session_file_name;
	if (!s)
		s = DEFSESSNAME;
	if (look_up_tree(file, s, NULL, session_tree_root))
		return;
	strcpy(file, s);
	if (*session_default_directory)
		strcpy(dir, session_default_directory);
	else if (p = verenv("EPSPATH")) {
		strcpy(dir, p);
		p = index(dir, path_list_char);
		if (p)
			*p = 0;
		if (!is_unix && !strcmp(dir, "~"))
			get_customization_directory(dir);
	} else
		get_customization_directory(dir);
	absolute(file, dir);
}

ask_session_file(file, pr)	/* ask user for a session file */
char *file, *pr;
{
	char def[FNAMELEN];

	save_var only_file_extensions = "|.ses|";
	give_session_file(def);
	get_file(file, pr, def);
}

/* Write out the session automatically when exiting.  Report any error
 * and return nonzero error code. */
save_session()
{
	char file[FNAMELEN];

	if (preserve_session_once)
		save_var preserve_session = preserve_session_once - 1;
	if (!(preserve_session & (SAVE_RESTORE_SESSION | SAVE_SESSION)))
		return 0;
	if (!preserve_session_flags && preserve_session != SAVE_SESSION)
		return 0;	// Never read a session, don't write (unless flags say to).
	give_session_file(file);
	int err = do_save_session(file);
	if (err && session_warn_when_saving) {
		save_var no_popup_errors = 1;
		file_error(err, file, "session file write error");
		if (ask_yn("Can't save session, exit anyway?",
				   "Error writing session file, exit anyway? ",
				   "Yes", "No", 1) != 1)
			quick_abort();
	}
	return err;
}

command write_session()
{
	char file[FNAMELEN];
	int err;

	ask_session_file(file, "Write session to");
	err = do_save_session(file);
	if (err) {
		file_error(err, file, "write error");
		maybe_ding(bell_on_write_error);
	} else {
		say("Session saved on %s", file);
		if (explicit_session_file)
			free(explicit_session_file);
		explicit_session_file = strsave(file);
	}
}

do_save_session(file)	/* write out the session putting fname in file */
char *file;		/* and return error code */
{
	int b, err;

	make_session(b = tmp_buf(), file);
	save_var bufnum = b;
	err = file_write(file, 0);
	restore_vars();
	buf_delete(b);
	return err;
}

char *give_buffer_name(buf)	/* give name or "" if bad buffer # */
{
	return buf_exist(buf) ? bufnum_to_name(buf) : "";
}

// Write info on window 
session_save_window(int b, int i)
{
	struct window_info w;

	get_window_info(i, &w);
	buf_printf(b, "%d %d %d %d\n",
			   w.left, w.top, w.right, w.bottom);
	buf_printf(b, "%d %d %d %d\n",
			   w.textcolor, w.hbordcolor, w.vbordcolor, w.titlecolor);
	buf_printf(b, "%d %d\n", w.borders, w.other);
	buf_printf(b, "%s\n", give_buffer_name(w.bufnum));
	buf_printf(b, "%s\n", give_buffer_name(w.prevbuf));
	buf_printf(b, "%d %d %d\n", w.point, w.dcolumn, w.dpoint);
}

session_save_misc(int b)
{
	buf_printf(b, "%d %d %d %d\n",
			   _default_flags, _default_regex_flags, _last_was_regex,
			   dired_sorts_files);
	buf_printf(b, "%s" SES_DELIM "%s" SES_DELIM,
			   _default_search ? _default_search : "",
			   _default_regex_search ? _default_regex_search : "");
	buf_printf(b, "%s" SES_DELIM "%s" SES_DELIM,
			   _default_replace ? _default_replace : "",
			   _default_regex_replace ? _default_regex_replace : "");
	buf_printf(b, "%s" SES_DELIM "%s" SES_DELIM,
			   push_cmd, _region_file);
	buf_printf(b, "%s" SES_DELIM "\n", _grep_files);
}

make_session(b, sesfile)
char *sesfile;
{
	int i, wins, owin = window_number;
	char sdir[FNAMELEN];	// dir of session file
	char file[FNAMELEN];

	strcpy(sdir, sesfile);
	absolute(sdir);
	*get_tail(sdir, 0) = 0;
	buf_printf(b, SES_HDR "\n");
	save_buffer_list(sdir, b);
	buf_printf(b, "\n%d\n", _prev_num);
	for (i = 0; i < _prev_num; i++)
		buf_printf(b, "%s\n", give_buffer_name(_prev_old_buffers[i]));
	save_bookmark_session(b);

	wins = number_of_windows() - number_of_popups();
	buf_printf(b, "%d %d %d %d %d %d\n",
			   wins, screen_cols, screen_lines, is_gui, opsys,
			   buf_accessed_clock);
	for (i = 0; i < wins; i++)
		session_save_window(b, i);
	buf_printf(b, "%d %d %d %d %d 0 0\n", owin, echo_line,
			   permanent_menu, avoid_top_lines, avoid_bottom_lines);
	session_save_misc(b);
	make_session_hook(b);
	save_recall_session(b);
	getcd(file);
	buf_printf(b, "CWD: %s\n", file);
	buf_printf(b, "OPENCWD: %s\n", common_open_curdir);
	buf_printf(b, "FILTER: %d\n", filter_index);
	buf_printf(b, "BROWSERFILE: %s\n", browser_file);
}

session_save_buffer(int b, char *sdir)
{
	char file[FNAMELEN];
	int transl = GET_LINE_TRANSLATE(translation_type);
	int uni_compat = 0;

	if (has_feature & FEAT_UNICODE) {
		char *s = encoding_to_name(GET_ENCODING(translation_type));
		if (s)
			uni_compat = strstr(s, "UTF-16") != 0;
		if (uni_compat)
			transl = FILETYPE_BINARY;
	}
	buf_printf(b, "%s\n", bufname);
	if (is_dired_buf()) {
		insert_char_in_buf(b, 0);	// Hide file pattern from old versions.
		buf_printf(b, "%s\n", get_dired_buffer_pattern());
	} else {
		relative(filename, file, sdir);
		buf_printf(b, "%s\n", file);
	}
	buf_printf(b, "%d %d %d %d %d %d %d",
			   point, mark, transl,
			   is_dired_buf(), dired_buffer_pattern_wildcards,
			   dired_sorts_files, force_save_as);
	if ((has_feature & FEAT_UNICODE) && (translation_type & FORCED_TRANS))
		buf_printf(b, " %d @%s", GET_LINE_TRANSLATE(translation_type),
				   encoding_to_name(GET_ENCODING(translation_type)));
	buf_printf(b, "\n");
	buf_printf(b, "%d %d %d %d\n", over_mode, fill_mode,
			   indent_with_tabs, margin_right);
	transl = FILETYPE_AUTO;
	if (uni_compat || (translation_type & FORCED_TRANS))
		transl = GET_LINE_TRANSLATE(translation_type);
	// This flavor of the translation type is used by non-Unicode/
	// pre-v12 versions only.
	buf_printf(b, "%d %d %d %d %d %d %d %d %d %d\n", tab_size,
			   case_fold, 0, want_warn,
			   (_buf_readonly == 4 ? 0 : _buf_readonly), buf_accessed,
			   (file_io_converter == oem_file_converter),
			   uni_compat,
			   transl, soft_tab_size);
}

save_buffer_list(char *sdir, int b)
{
	int i, k = 0, d = 0, tmp = tmp_buf(), nums[3];

	save_var bufnum;
	minimize_buf_accessed();
	// All buffers should now have a buf_accessed setting from 0 to buf_accessed_clock.
	for (i = buf_list(0, 0); i; i = buf_list(1, 1)) {
		bufnum = i;
		if (buf_accessed < 0 || buf_accessed > buf_accessed_clock)
			error("Internal error: buffer %s clock %d, not 0..%d",
				  bufname, buf_accessed, buf_accessed_clock);
		if (i == b || !remember_this_buffer())
			continue;
		buf_printf(tmp, "%09d %d %d\n", buf_accessed, is_dired_buf(), bufnum);
	}
	bufnum = tmp;
	do_sort_region(0, size(), 0, 1);
	point = 0;
	while (grab_numbers(tmp, nums) == 3) {
		if (nums[1] ? (session_restore_max_directories >= 0
					   && ++k > session_restore_max_directories)
					: (session_restore_max_files >= 0
					   && ++d > session_restore_max_files))
			continue;
		bufnum = nums[2];
		session_save_buffer(b, sdir);
	}
	restore_vars();
	buf_delete(tmp);
}

remember_this_buffer()	/* should we record this buffer in session file? */
{
	char *fname = filename;
	int limit;

	if (*bufname == '-')
		return 0;
	if (is_process_buffer(bufnum) || (buffer_not_saveable & (NOSAVE_ALL | NOSAVE_SESSION))
		  || _read_aborted || (force_save_as & FSA_NEWFILE))
		return 0;
	if (file_info.year && file_info.check_type != CHECK_FILE)
		return 0;
	if (is_dired_buf()) {
		fname = get_dired_buffer_pattern();
		if (!(session_restore_directory_buffers
			  & (is_remote_file(fname) ? 2 : strstr(fname, "**") ? 4 : 1)))
			return 0;
	} else {
		if (!*filename)
			return 0;
		limit = is_remote_file(filename) ? session_restore_biggest_remote_file
										 : session_restore_biggest_file;
		if (limit >= 0 && size() > limit)
			return 0;		// too big to restore
		if (!size() && check_file(filename) != CHECK_FILE)
			return 0;
	}
	if (session_restore_exclude_file_types) {
		int flags = filename_rules(fname);
		if ((flags & FSYS_NETWORK) && (session_restore_exclude_file_types & 1))
			return 0;
		if ((flags & FSYS_REMOVABLE) && (session_restore_exclude_file_types & 2))
			return 0;
		if ((flags & FSYS_CDROM) && (session_restore_exclude_file_types & 4))
			return 0;
	}
	return 1;
}

save_bookmark_session(b)
{
	int i, buf;
	spot m;

	buf_printf(b, "%d %d\n", _bookmark_slot, _last_bookmark);
	for (i = 0; i < NUM_BOOKMARKS; i++) {
		m = _bookmark_spots[i];
		if (m && (buf = spot_to_buffer(m)) > 0)
			buf_printf(b, "%s\n%d\n", give_buffer_name(buf), *m);
		else
			buf_printf(b, "\n0\n");
	}
}

save_recall_session(b)	// store saved prompt responses in session file
{
	save_var bufnum = create(OLDRESP);
	point = size() - recall_maximum_session;
	to_begin_line();	// only keep last N chars worth
	buf_xfer(b, point, size());
	buf_printf(b, SES_DELIM "\n");
}

command read_session() on reg_tab[FSHIFT(5)]
{
	char file[FNAMELEN];
	int res;

	if (maybe_save_all())
		return 0;
	ask_session_file(file, "Read session from");
	preserve_session_flags |= 2;
	res = load_this_session(file);
	return res;
}

load_session()		/* maybe search for a session file & load it */
{			/* return 1 if session was loaded */
	char file[FNAMELEN];

	if (preserve_session_once)
		save_var preserve_session = preserve_session_once - 1;
	if (!(preserve_session & (SAVE_RESTORE_SESSION | RESTORE_SESSION)))
		return 0;
	give_session_file(file);
	preserve_session_flags |= 1;
	show_text(-2, 0, "Try to load session file %s", file);
	return load_this_session(file);
}

load_this_session(file)		/* return 1 if session was loaded */
char *file;
{
	int b, err, old = bufnum;

	bufnum = b = create(SES_BNAME);
	err = file_read(file, 0);
	if (!err) {
		if (explicit_session_file)
			free(explicit_session_file);
		explicit_session_file = strsave(file);
		maybe_refresh();
		err = restore_session(b);
		switching_to_buffer();
		if (buf_in_window(b) == -1)
			buf_delete(b);
		if (!err)
			show_text(0, -1, "");
		else if (err != 100)
			show_text(0, -1, "Invalid format, code %d, in session %s",
				err, file);
		return 1;
	} else if (err != 2) {
		file_error(err, file, "couldn't read session");
		maybe_ding(bell_on_read_error);
	}
	bufnum = old;
	buf_delete(b);
	return 0;
}

session_restore_one_file_1(int *nums, int i)
{
	UNREFERENCED_PARAMETER(i);
	over_mode = nums[0];
	fill_mode = nums[1];
	indent_with_tabs = nums[2];
	margin_right = nums[3];
}

session_restore_one_file_2(int *nums, int i)
{
	tab_size = nums[0];
	case_fold = nums[1]; // nums[2] var was removed
	want_warn = nums[3];
	if (nums[4] == 1)
		set_read_only(1);
	buf_accessed = i > 5 ? nums[5] : 0;
	if (i >= 10)
		soft_tab_size = nums[9];
	undo_mainloop();
}

// Read one file, return -1 to continue, 0 if user aborted, or error code.

int session_restore_one_file(int b, char *file, char *sdir, int *readerrs)
{
	int i, trans = FILETYPE_AUTO;
	int num1[10], num1cnt, num2[10], num2cnt, num3[10], num3cnt;
	char firstline[150];
	jmp_buf this_level;

	grab_line(b, firstline);
	if ((num1cnt = break_into_numbers(firstline, num1)) < 3)
		return 2;
	if ((num2cnt = grab_numbers(b, num2)) < 4)
		return 3;
	if ((num3cnt = grab_numbers(b, num3)) < 5)
		return 4;
	if (session_restore_files)
		filename = file;
	save_var top_level = &this_level;
	if (setjmp(top_level))
		return -1;
	if (num1cnt >= 4 && num1[3] && session_restore_files) {	// restore dired buf
		filename = "";
		if (num1cnt >= 5)
			save_var file_pattern_wildcards = num1[4];
		save_var show_status = 0;
		if (num1cnt >= 6)
			dired_sorts_files = num1[5];
		create_dired_listing(file, 0);
		if (is_remote_file(get_dired_buffer_pattern()))
			dired_orig_position = -1;
	} else if (*file && session_restore_files) {
		if (!is_remote_file(file))
			absolute(file, sdir);
		if (has_feature & FEAT_UNICODE) {
			if (num1cnt >= 7) {		 // If explicit trans, force.
				char *p;
				if (p = strchr(firstline, '@')) {
					i = encoding_from_name(p + 1);
					if (i)
						trans = MAKE_TRANSLATE(i, num1[7], 1);
				}
			}
		} else {
			if (num3cnt > 8 && num3[7])
				trans = num1[2];	// If explicit trans, or common unicode, force.
		}
		i = do_file_read(file, trans);
		force_save_as = 0;
		if (num1cnt >= 7)
			force_save_as = num1[6];
		if (user_abort || i == EREADABORT)
			return 0;
		if (!i || i == 2) {
			if (!i)
				set_readonly_from_file();
			window_bufnum = bufnum;		// make suffix- kbd macros work
			call_mode(filename);
			make_mode();
			check_filevars();
		} else if (!*readerrs++) {
			file_error(i, file, "read error");
			maybe_ding(bell_on_read_error);
		} else
			show_text(0, -1, "%d files could not be read", *readerrs);
	}
	if (session_restore_files) {
		if (bufnum != b)	/* same tmp name as session buffer */
			point = num1[0];
		mark = num1[1];
		if (is_dired_buf())
			move_to_column(goal_column);
	}
	if (session_restore_files) {
		session_restore_one_file_1(num2, num2cnt);
		session_restore_one_file_2(num3, num3cnt);
	}
	return -1;
}

int session_restore_prev_buffers(int b)
{
	int nums[10], i, j;
	char buf[FNAMELEN];

	if (grab_numbers(b, nums) != 1)
		return 1;
	j = nums[0];
	if (session_restore_files)
		_prev_num = j;
	for (i = 0; i < j; i++) {
		grab_line(b, buf);
		if (session_restore_files)
			_prev_old_buffers[i] = name_to_bufnum(buf);
	}
	return 0;
}

int session_restore_windows(int b, int def_buf)
{
	int nums[10], i, wins, err;

	i = grab_numbers(b, nums);
	if (i < 3)
		return -1;
	wins = nums[0];
	if (((has_feature & FEAT_EEL_RESIZE_SCREEN))
		  && !want_lines && !want_cols)
		try_set_screen_size(nums[2], nums[1], "");
	if (i > 5)
		buf_accessed_clock = nums[5];
	fix_buf_accessed_clock();
	err = rest_sess_wins(b, wins, !session_restore_files, def_buf,
						 nums[2], nums[1]);
	if (err) {
		def_buf = give_prev_buf();
		if (def_buf)
			to_buffer_num(def_buf);
	}
	return err;
}

int session_restore_misc(int b)
{
	int nums[10], i;
	char file[FNAMELEN];

	i = grab_numbers(b, nums);
	if (i < 3)
		return 1;
	_default_flags = nums[0];
	_default_regex_flags = nums[1];
	_last_was_regex = nums[2];
	if (i >= 4)
		dired_sorts_files = nums[3];
	grab_string_expanding(b, &_default_search, SES_DELIM, FNAMELEN);
	grab_string_expanding(b, &_default_regex_search, SES_DELIM, FNAMELEN);
	grab_string_expanding(b, &_default_replace, SES_DELIM, FNAMELEN);
	grab_string_expanding(b, &_default_regex_replace, SES_DELIM, FNAMELEN);
	grab_string(b, push_cmd, SES_DELIM);
	grab_string(b, _region_file, SES_DELIM);
	grab_string(b, _grep_files, SES_DELIM);
	restore_session_hook(b);
	restore_recall(b);
	grab_line(b, file);		// skip \n after recall buffer
	grab_line(b, file);
	if (!strncmp(file, "CWD: ", 5)) {
		switch (session_restore_directory) {
			default:	if (directory_flags & DIR_CWD_RESTORE)
			case 1:			chdir(file + 5);
			case 0:		break;
		}
		grab_line(b, file);
		if (!strncmp(file, "OPENCWD: ", 9))
			common_open_curdir = file + 9;
		grab_line(b, file);
		if (!strncmp(file, "FILTER: ", 8))
			filter_index = numtoi(file + 8);
		grab_line(b, file);
		if (!strncmp(file, "BROWSERFILE: ", 13))
			strcpy(browser_file, file + 13);
	}
	return 0;
}

// Read the next line, a file name, from buffer b and put in file.  This
// is complicated because we want to ignore a null character at the
// start of the file name.  We put one there before file names previous
// versions won't understand (specifically dired patterns) so they'll be
// ignored.
get_file_name_from_session_file(int b, char *file)
{
	save_var bufnum = b;
	if (!curchar())
		point++;
	grab_line(b, file);
}

restore_session(b)	/* if error in session file, return error code # */
{			/* for debugging purposes, else 0 */
	int j, err, readerrs = 0, def_buf = 0;
	char buf[FNAMELEN], file[FNAMELEN];
	char sdir[FNAMELEN];	// dir of session file

	give_session_file(sdir);
	absolute(sdir);
	*get_tail(sdir, 0) = 0;
	bufnum = b;
	point = 0;
	try_calling("restore-session-conversion-hook");
	if (!grab_line(b, buf) || strcmp(buf, SES_HDR))
		return 1;
	while (grab_line(b, buf)) {
		get_file_name_from_session_file(b, file);
		if (session_restore_files)
			bufnum = create(buf);
		if (!def_buf) {
			if (is_gui)
				show_text(0, 1, "Restoring session...");
			else
				show_text(-1, 1, "Restoring session...");
			def_buf = bufnum;
		}
		j = session_restore_one_file(b, file, sdir, &readerrs);
		if (j != -1)
			return j;
	}
	if (session_restore_prev_buffers(b))
		return 12;
	if (restore_bookmark_session(b))
		return 13;
	err = session_restore_windows(b, def_buf);
	if (err < 0)
		return 5;
	if (session_restore_misc(b))
		err = 11;
	return err ? err : readerrs ? 100 : 0;
}

set_readonly_from_file()
{
	if (file_info.check_type == CHECK_FILE &&
			(file_info.attr & ATTR_READONLY) &&
			(readonly_warning & ROWARN_BUF_RO)) {
		_buf_readonly = 2;	// set from file
		call_on_modify = 1;
	}
}

restore_bookmark_session(b)
{
	int nums[10], i;
	char buf[FNAMELEN];

	if (grab_numbers(b, nums) < 2)
		return 1;
	_bookmark_slot = nums[0];
	_last_bookmark = nums[1];
	for (i = 0; i < NUM_BOOKMARKS; i++) {
		grab_line(b, buf);
		if (grab_numbers(b, nums) < 1)
			return 1;
		if (*buf && exist(buf)) {
			bufname = buf;
			_bookmark_spots[i] = alloc_spot(0);
			*_bookmark_spots[i] = nums[0];
			if (!size() && is_remote_file(filename))
				_bookmark_saved_pos[i] = nums[0];
		}
	}
	return 0;
}

/* restore windows from b */
/* if skip, just parse them, don't restore */

rest_sess_wins(b, wins, skip, def_buf, lines, cols)
{
	int nums[10], i, err = 0, bnum, must_default;
	char buf[FNAMELEN];
	struct window_info w;

	for (i = 0; i < wins; i++) {
		if (grab_numbers(b, nums) < 4)
			err = 6;
		w.left = nums[0];
		w.top = nums[1];
		w.right = nums[2];
		w.bottom = nums[3];
		if (grab_numbers(b, nums) < 4 || w.left < 0 || w.top < 0 || w.right < 0 || w.bottom < 0)
			err = 7;
		w.textcolor = nums[0];
		w.hbordcolor = nums[1];
		w.vbordcolor = nums[2];
		w.titlecolor = nums[3];
		if (grab_numbers(b, nums) < 2)
			err = 8;
		w.borders = nums[0];
		w.other = nums[1];
		grab_line(b, buf);
		w.point = -1;
		w.dpoint = -1;
		bnum = name_to_bufnum(buf);
		must_default = !bnum;
		grab_line(b, buf);
		w.prevbuf = name_to_bufnum(buf);
		w.bufnum = bnum ? bnum : w.prevbuf ? w.prevbuf :
			(bnum = give_prev_buf()) ? bnum : def_buf;
		if (grab_numbers(b, nums) < 3)
			err = 9;
		w.dcolumn = nums[1];
		if (!must_default) {
			w.point = nums[0];
			w.dpoint = nums[2];
		}
		if (err)
			break;
		if (!skip) {
			window_create(!i, &w);
			fix_window_start();
		}
	}
	nums[1] = screen_lines - 1;	/* echo line */
	nums[2] = 0;			/* want perm menu? */
	nums[3] = 0;			/* top avoid lines */
	nums[4] = 1;			/* bottom avoid lines */
	if (!err && !grab_numbers(b, nums))
		err = 10;
	if (err)
		one_window();
	else if (!skip) {
		echo_line = nums[1];
		select_low_window(nums[0], nums[3], nums[4], lines, cols);
		i = permanent_menu;
		restore_menu(nums[2]);
		set_menu_bar(i);	// undo setting in session file
	}
	return err;
}

restore_recall(b)	// get saved prompt responses from session
{
	int orig;

	save_var bufnum = b;
	orig = point;
	if (search(1, SES_DELIM))
		buf_xfer(zap(OLDRESP), orig, matchstart);
}

// Set buf_accessed_clock so it's higher than any buffer's buf_accessed.

fix_buf_accessed_clock()
{
	int i = buf_list(0, 2);
	save_var bufnum;
	do {
		bufnum = i;
		if (buf_accessed_clock <= buf_accessed)
			buf_accessed_clock = buf_accessed + 1;
	} while (i = buf_list(-1, 1));
}

// If the user manually changes preserve-session, it overrides -p.
when_setting_preserve_session()
{
	preserve_session_once = 0;
}

make_session_hook(b)
{
	b = b;				/* silence eel's warning */
}

restore_session_hook(b)
{
	b = b;				/* silence eel's warning */
}
