/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1986, 2020 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// This file has commands for setting and showing variables.

#include "eel.h"
#include "lowlevel.h"

char *get_var_val(i, tmp, def)	/* return value of variable with index i */
char *tmp;			/* (maybe in tmp), or 0 if bad type */
{
	char *p = tmp;

	save_var use_default = def;
	switch (vartype(i)) {
		case TYPE_BYTE:
		case TYPE_CHAR:
		case TYPE_SHORT:
		case TYPE_INT32:
		case TYPE_INT:
			sprintf(tmp, "%d", get_num_var(i));
			break;
		case TYPE_CPTR:
			p = get_str_var(i);
			return p ? p : "<Null Pointer>";
		case TYPE_CARRAY:
			strncpy(tmp, varptr(i), ptrlen(tmp) - 1);
			break;
		default:
			return 0;
	}
	return p;
}

do_setting(int i, char *tmp, int def)
{
	save_var use_default = def;
	switch (vartype(i)) {
		case TYPE_BYTE:
		case TYPE_CHAR:
		case TYPE_SHORT:
		case TYPE_INT32:
		case TYPE_INT:
			if (*tmp)
				set_num_var(i, exptoi(tmp));
			break;
		case TYPE_CARRAY:
			strncpy(varptr(i), tmp, ptrlen(varptr(i)) - 1);
			int j = ptrlen(varptr(i)) - strlen(tmp);
			if (j > 0)
				set_chars(varptr(i) + strlen(tmp), 0, j);
			else
				say("%s truncated: %d characters, room for %d.",
					  name_name(i), strlen(tmp), ptrlen(varptr(i)) - 1);
			break;
		case TYPE_CPTR:
			set_str_var(i, strsave(tmp)); break;
	}
}

// Set variable at index i to tmp.  For buffer/window-specific
// variables: If def == 0, set for this buffer/window.  Def == 1: set
// default.  Def == 2: both.  Def == 3: set default and all non-system
// buffers/windows.  Def == 4: set default & all buffers/windows.

set_var_val(i, tmp, def)
char *tmp;
{
	char when_setting[FNAMELEN];
	int old_record_cust = record_customizations;

	if (strstr(" selected-color-scheme -our-mono-scheme -our-color-scheme "
			   "-our-gui-scheme -our-unixconsole-scheme ", name_name(i))
			   && isalpha(*tmp)) {
		if (name_type(find_index(tmp)) != NT_COLSCHEME)
			error("%s is not a color scheme", tmp);
		sprintf(tmp, "%d", find_index(tmp));
	}
	do_setting(i, tmp, def);
	if (def == 2)
		do_setting(i, tmp, 0);
	sprintf(when_setting, "when-setting-%s", name_name(i));
	try_calling(when_setting);
	if (def) {
		save_customization(i);
		// Also record turning it off.
		if (old_record_cust && !record_customizations && !noinit_flag
			  && !now_parsing_cmds && i == find_index("record-customizations"))
			list_one(prep_customizations(), i, 0);
	}
	if (def >= 3)
		switch (name_type(i)) {
			case NT_WINVAR:
				save_var window_handle;
				int orig = window_number;
				do {
					if (!system_window || (def == 4))
						do_setting(i, tmp, 0);
				} while (++window_number != orig);
				break;
			case NT_BUFVAR:
				save_var bufnum;
				int buf = buf_list(0, 2);
				do {
					bufnum = buf;
					if (*bufname != '-' || (def == 4))
						do_setting(i, tmp, 0);
				} while (buf = buf_list(-1, 1));
				break;
		}
}

safe_set_var_val(i, tmp, def)	/* set var at index i to tmp */
char *tmp;
{
	jmp_buf this_level;
	save_var top_level = &this_level;
	if (!setjmp(top_level))
		set_var_val(i, tmp, def);
}

command set_variable() on reg_tab[FKEY(8)]
{
	char res[80];
	int i;

	get_var(res, "Set variable", "", MUST_MATCH);
	i = find_index(res);
	if (name_type(i) != NT_VAR && name_type(i) != NT_BUFVAR
	    && name_type(i) != NT_WINVAR && name_type(i) != NT_BUILTVAR)
		error("%s: not a variable", res);
	save_var _show_help_on = res;
	do_set_variable(i);
}

command set_any_variable()
{
	save_var show_all_variables = 1;
	set_variable();
}

char *variable_type_name(i)	// translate type code into text
{
	switch (i) {
		case TYPE_BYTE:		return "byte";
		case TYPE_CHAR:		return "character";
		case TYPE_SHORT:	return "short integer";
		case TYPE_INT32:	return "int32";
		case TYPE_INT:		return "integer";
		case TYPE_CARRAY:
		case TYPE_CPTR:		return "string";
	}
	return "other";
}

do_set_variable(i)
{
	char msg[FNAMELEN], res[2048], *tname, *p, tmp[2048], prompt[FNAMELEN];
	int def;

	tname = variable_type_name(vartype(i));
	if (!(p = get_var_val(i, tmp, 0)))
		error("Can't interpret type of %s", name_name(i));
	sprintf(msg, "Set %s variable %s to", tname, name_name(i));
	if (run_by_mouse && (has_feature & FEAT_GUI_DIALOGS)) {
		get_strpopup(res, msg, p, name_name(i));
		def = 2;			// set both
	} else {
		build_prompt(prompt, msg, p, insert_default_response, 0);
		comp_read(res, prompt, (char *(*)()) 0, 0, p);
		def = ask_default("Set", name_name(i), 1);
	}
	set_var_val(i, res, def);		/* set default */
	do_show_variable(name_name(i), def);
	full_redraw = 1;	/* might have set a display-type variable */
}

command show_variable() on reg_tab[FCTRL(8)]
{
	char res[80];

	save_var show_all_variables = 1;
	get_var(res, "Show variable", "", MUST_MATCH);
	do_show_variable(res, (var_settable(res) == 1) ? 1
				: ask_default("Show", res, 0));
}

do_show_variable(name, def)
char *name;
{
	char *vl = "Value", tmp[200], tmp2[20];
	int i, type, hasdef = 1;

	i = find_index(name);
	if (name_type(i) == NT_BUFVAR)
		vl = def ? "Default value" : "This buffer's value";
	else if (name_type(i) == NT_WINVAR)
		vl = def ? "Default value" : "This window's value";
	else if (name_type(i) == NT_VAR || name_type(i) == NT_BUILTVAR)
		hasdef = 0;
	else
		error("%s: not a variable", name);
	type = var_settable(name);
	if (!type)
		error("Can't interpret type of %s", name_name(i));
	else if (type == 1 && hasdef)
		say("Default value of %s is %s, and current value is %s",
			name_name(i), get_var_val(i, tmp, 1),
				get_var_val(i, tmp2, 0));
	else
		say("%s of %s is: %s", vl, name_name(i),
				get_var_val(i, tmp, def));
}

var_settable(char *name)	/* is variable settable? 1=number, 2=string, else 0 */
{
	return vartype_class(find_index(name));
}

vartype_class(int i)
{
	switch (vartype(i)) {
		case TYPE_BYTE: case TYPE_CHAR: case TYPE_SHORT: case TYPE_INT: case TYPE_INT32:
			return 1;
		case TYPE_CARRAY: case TYPE_CPTR:
			return 2;
	}
	return 0;
}

/* Ask if we want to get/set default.  Return 0=current only, 1=default
 * only, 2=both, 3=plus all buffers/windows, 4=even system ones.  If
 * !bothok, only first two options are available. */
ask_default(ask, name, bothok)
char *ask, *name;
{
	int res = iter, type = name_type(find_index(name));

	iter = 0;
	if (type != NT_BUFVAR && type != NT_WINVAR)
		return 1;
	if (has_arg && res >= 0 && res <= 4)
		return res;
	sayput(bothok
		? "%s D=default value of %s, C=current value, B=both, or A=all? [B] "
		: "%s D=default value of %s or C=current value? [C] ",
		ask, name);
	do {
		getkey();
	} while (!IS_TRUE_KEY(key));
	check_abort();	/* N means C for compatibility */
	switch (toupper(key)) {
		case 'A':		if (bothok)
							return 3;
		default:		if (bothok)
							return 2;
		case 'N': case 'C':	return 0;
		case 'D':		return 1;
	}
}

command create_variable()
{
	char res[80];
	int type = -1, ind, len = 1;

	iter = 0;
	save_var show_all_variables = 1;
	get_var(res, "Create variable", "", 0);
	if (find_index(res)
		  && !ask_yn(" Replace ", "Replace existing item with that name? ",
					 "Yes", "No", 1))
		error("No changes made.");
	do {
		sayput("Character, Short, Int32, Integer, or Array of characters"
		       " [C, S, 3, I, A] ");
		getkey();
		check_abort();
		switch (toupper(key)) {
			case 'C': type = TYPE_CHAR; break;
			case 'S': type = TYPE_SHORT; break;
			case '3': type = TYPE_INT32; break;
			case 'I': type = TYPE_INT; break;
			case 'A': type = TYPE_CARRAY;
				  len = get_number("Length of new variable: ");
				  break;
			default: if (IS_TRUE_KEY(key))
					 maybe_ding(bell_on_bad_key);
				 break;
		}
	} while (type < 0);
	ind = new_variable(res, NT_VAR, type, len);
	set_name_user(ind, 1);		/* make it a user variable */
	do_set_variable(ind);
}

command delete_name()
{
	char name[FNAMELEN];

	save_var show_mouse_choices = 1;
	get_any(name, "Delete (function, variable name...)", "");
	switch (drop_name(name)) {
		case 1: error("%s: not found", name); break;
		case 2: error("%s: in use", name); break;
	}
}

command change_name()
{
	char old[FNAMELEN], new[FNAMELEN], msg[FNAMELEN];

	save_var show_mouse_choices = 1;
	get_any(old, "Rename (function, variable name...)", "");
	if (!find_index(old))
		error("%s: not found", old);
	sprintf(msg, "New name for %s", old);
	get_any(new, msg, "");
	switch (replace_name(old, new)) {
		case 1: error("%s: not found!", old); break;
		case 2: error("%s exists, delete first", new); break;
	}
}

show_abort_key(keyname, akey)
char *keyname;
{
	*keyname = 0;
	if (akey == ESC)
		strcpy(keyname, "<Esc>");
	else if (akey >= 0)
		show_char(keyname, akey, 1);
	else
		strcpy(keyname, "disabled");
}

command set_abort_key()
{
	char keyname[FNAMELEN], msg[FNAMELEN];
	int newkey;

	iter = 0;
	show_abort_key(keyname, abort_key);
	mention("Set abort key (now %s) to (D to disable): ", keyname);
	do {
		newkey = getkey();
	} while (!IS_TRUE_KEY(newkey));
	user_abort = 0;
	show_abort_key(keyname, newkey);
	sprintf(msg, "Set abort key to %s? ", keyname);
	if (ask_yn("Set abort key", msg, "Yes", "No", 2)) {
		abort_key = (toupper(newkey) == 'D') ? -1 : newkey;
		say("The abort key is now %s", keyname);
		if (abort_key != -1) {
			reg_tab[abort_key] = (short) abort;
			save_binding(reg_tab + abort_key);
		}
		copy_bindings();	/* fix up changed bindings */
		if (want_record_customizations()) {
			save_var bufnum = prep_customizations();
			bprintf("+abort-key int variable: %d\n", abort_key);
		}
	} else
		say("No change.");
}

// Modify the provided state file name to point it to a home dir, or
// initialize it if it's "".

set_default_state_file_name(char *state_file)
{
	char fname[FNAMELEN], newname[FNAMELEN];

	build_filename(fname, "%x%f", default_state_file_name);
	strcpy(newname, default_state_file_name);
	absolute(newname);
	// Force a different name if the original is read-only or the
	// suggested name differs from the usual one in the executable's
	// directory.
	if (*state_file && (strcmp(newname, fname)
						|| get_file_read_only(state_file) == 1)) {
		get_customization_directory(state_file);
		absolute(state_file);		// Strip off any trailing /.
		sprintf(state_file + strlen(state_file), "%c%s",
				path_sep, default_state_file_name);
	} else if (!*state_file)
		strcpy(state_file, default_state_file_name);
}

command write_state() on reg_tab[FCTRL(3)]
{
	char fname[FNAMELEN];

	set_default_state_file_name(state_file);
	save_var only_file_extensions = "|.sta|";
	if (quiet_write_state)
		strcpy(fname, state_file);
	else
		get_file(fname, "Write current state on file", state_file);
	if (!*fname)
		return;

	do_save_state(fname);
	strcpy(state_file, fname);
	iter = 0;
}

do_save_state(fname)		/* save state in file, aborting on error */
char *fname;
{
	int err;
	char file[FNAMELEN], backupname[FNAMELEN];

	strcpy(file, fname);
	copy_bindings();	/* fix up changed bindings */
	sayput("Writing...");

	strcpy(get_extension(file), state_extension);
	if (want_state_file_backups) {
		build_filename(backupname, state_file_backup_name, file);
		make_backup(file, backupname);
	}
	if (err = save_state(file)) {
		file_error(err, file, "write error");
		maybe_ding(bell_on_write_error);
		quick_abort();
	} else
		say("State saved on %f.", file);
}

command create_file_associations()
{
	configure_epsilon();
}

ev_header(char *buf)
{
	save_var bufname = buf;
	if (lookpath("include/eel.h", 0))
		bprintf("#include <eel.h>\n");
	else
		bprintf("#include \"eel.h\"\n");
}

ev_template()		// Replaced by functions below.
{
	error("Error in eval or execute-eel.");
}

void eval_template(char *expr, char *src, int in_buf)
{
	char quoted[300], *p, *q, *format = 0, copy[300];

	strcpy(copy, expr);
	expr = copy;
	for (p = expr; p = strchr(p, ';'); p++) {
		if (strchr(p, '"'))
			continue;
		format = p + 1;
		*p = 0;
		break;
	}
	save_var bufnum = zap(src);
	for (p = expr, q = quoted; *q = *p; p++, q++)
		if (*p == '"') {
			*q++ = '\\';
			*q = '"';
		}
	ev_header(src);
	bprintf("ev_template(){\n");
	bprintf("char *strval;\n");
	bprintf("  if (sizeof(%s) <= sizeof(int))\n", expr);
	char prefix[20];
	*prefix = 0;
	if (format)
		if (p = strpbrk(format, "xo"))
			sprintf(prefix, "0%c", *p);
	if (in_buf)
		bprintf("    bprintf(\"%s%%%s\", (%s));\n",
				prefix, format ? format : "d", expr);
	else
		bprintf("    say(\"%%s == %s%%%s\", \"%s\", (%s));\n",
				prefix, format ? format : "'d", quoted, expr);
	bprintf("  else {\n");
	bprintf("    strval = (char *)(%s);\n", expr);
	bprintf("    if (!strval) strval = \"<null>\";\n");
	if (in_buf)
		bprintf("    bprintf(\"%%%s\", strval);\n", format ? format : "s");
	else
		bprintf("    say(\"%%s == %%%s\", \"%s\", strval);\n",
				format ? format : "s", quoted);
	bprintf("}\n");
	bprintf("}\n");
}

command eval()
{
	char expr[FNAMELEN], *src = temp_buf();

	get_string(expr, "Expression: ");
	if (!*expr)
		return;
	eval_template(expr, src, has_arg);

	// Compiles the buffer src, the status messages go into the
	// buffer named "eval-comp".
	if (eel_compile(src, 0, "-s -b", "eval-comp") != 0){
		to_buffer("eval-comp");
		discardable_buffer = 1;
		delete_buffer(src);
		error("Cannot evaluate expression.");
	}
	delete_buffer(src);
	ev_template();
	iter = 0;
}

int ev_numexpr()		// Replaced by functions below.
{
	return 0;
}

// Take an expression like 1+2 and evaluate it.  If it's invalid,
// return 0.  Set got_bad_number to indicate if it's valid.
int evaluate_numeric_expression(char *expr)
{
	char *src = temp_buf();
	on_exit delete_buffer(src);
	save_var bufnum = zap(src);
	got_bad_number = 0;
	ev_header(src);
	bprintf("ev_numexpr(){ return %s; }\n", expr);
	if (!eel_compile(src, 0, "-s -b", "eval-comp"))
		return ev_numexpr();
	got_bad_number = 1;
	return 0;
}

void execute_eel_template(char *expr, char *src)
{
	save_var bufnum = zap(src);
	ev_header(src);
	bprintf("ev_template(){ %s; }\n", expr);
}

do_execute_eel(char *expr)
{
	char *src = temp_buf();
	on_exit delete_buffer(src);
	execute_eel_template(expr, src);

	// Compiles the buffer src, the status messages go into the
	// buffer named "eval-comp".
	if (eel_compile(src, 0, "-s -b", "eval-comp") != 0){
		to_buffer("eval-comp");
		discardable_buffer = 1;
		error("Could not compile EEL statement.");
	}
	ev_template();
}

inline_eel(char *code)
{
	char *src = temp_buf();
	on_exit delete_buffer(src);
	ev_header(src);
	buffer_printf(src, "\n%s\n", code);

	// Compiles the buffer src, the status messages go into the
	// buffer named "eval-comp".
	if (eel_compile(src, 0, "-s -b", "eval-comp") != 0){
		to_buffer("eval-comp");
		error("Could not compile EEL code.");
	}
}

command execute_eel()
{
	char expr[FNAMELEN];

	get_string(expr, "Execute EEL statement: ");
	if (*expr)
		do_execute_eel(expr);
}
