/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1985, 2021 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// This file has commands that deal with command files.

#include "eel.h"

jmp_buf *jbuf;
int *keyseq;		/* sequence of key codes, in format used by macros */

command load_file()
{
	char fname[FNAMELEN];

	save_var show_mouse_choices = 1;
	get_file(fname, "Load definitions from file", "");
	if (*fname)
		do_load_file(fname);
}

command load_buffer()
{
	char bname[FNAMELEN], *old = bufname;
	int err;

	get_buf(bname, "Load definitions from buffer", bufname);
	if (exist(bname))
		bufname = bname;
	else
		error("No such buffer");
	err = parse_cmds();
	bufname = old;
	if (err) to_buffer(bname);		/* show offending command */
}

do_load_file(fname)	/* load command file */
char *fname;
{
	int err;
	char *old = bufname, *temp_load;

	temp_load = temp_buf();
	bufname = temp_load;
	filename = fname;
	if (err = do_file_read(fname, FILETYPE_AUTO)) {
		if (_abort_load_on_error != 2) {
			file_error(err, fname, "read error");
			maybe_ding(bell_on_read_error);
		}
	} else if (parse_cmds()) {
		bufname = old;
		if (_abort_load_on_error == 2) {
			save_var see_delay = 300;
			say("Error found loading %s, see %s buffer.", fname, MSG_BUFFER);
		} else {
			to_buffer(temp_load);	/* show bad commands in window */
			quick_abort();
		}
	}
	bufname = old;
	delete_buffer(temp_load);
}

load_cmd_file_from_path(fname)
char *fname;
{
	char *s;

	if (!(s = lookpath(fname, 1)))
		error("Can't find %s", fname);
	do_load_file(s);
}

// Create an EEL color settings file from the changes file
// written by an older version of Epsilon (v7 or before).
// V8.0's changes file doesn't include 32-bit color information--
// load color customizations in v8.0 from the appropriate EEL file.

command import_colors()
{
	char name[FNAMELEN];
	int tmp, i;

	if (!ask_yn("Confirm", "This command is only for updating from Epsilon 7 or older. Continue? ",
				"Yes", "No", 1))
		return;
	get_file(name, "Load changes from file", "changes");
	tmp = tmp_buf();
	to_buffer_num(tmp);
	if (read_file(name, FILETYPE_AUTO))
		quick_abort();

	strcpy(name, "mycolors.e");
	absolute(name);
	if (warn_existing_file(name) > 0)
		return;
	do_find(name, FILETYPE_AUTO);
	buf_zap(bufnum);
	grab_buffer(tmp);
	point = 0;
	do_drop_matching_lines(REGEX, "^&.* color class for ", 0);
	do_drop_matching_lines(0, "[0x4000000 on 0x0]", 1);	// Unknown color.
	while (re_search(1, ": *([0-9a-fx]+)$")) {
		grab(find_group(1, 1), find_group(1, 0), name);
		point = find_group(1, 1);		// Change old-style color numbers
		delete(point, find_group(1, 0));	// to names.
		i = numtoi(name);
		give_color_combo(name, i);
		stuff(name);
	}
	point = 0;
	for (i = 0; i < 10; i++) {	// Assume class names have no more than 10 -'s.
		noteput("Working... %d%%", 20 + i * 5);
		string_replace("(color class for .*)-", "#1_", REGEX);	// Conv to _'s.
		if (!replace_num_changed)
			break;
	}
	for (i = 0; i < 10; i++) {
		string_replace("(=<word>*)-", "#1_", REGEX);	// Conv to _'s.
		if (!replace_num_changed)
			break;
	}

	noteput("Working... 70%%");
	// Make into valid EEL definitions.
	string_replace("^&(.*) color class for (.*): (.*)",
				   "color_scheme \"#1\" color_class #2 #3;", REGEX);
	noteput("Working... 80%%");
	// Convert pre-v8 color names to EEL names.
	string_replace(" (blue|green|cyan|red|magenta)", " dark_#1", REGEX);
	string_replace(" white", " grey", REGEX);
	string_replace(" brightblack", " dark_grey", REGEX);
	noteput("Working... 90%%");
	string_replace(" bright(blue|green|cyan|red|magenta|white)", " #1", REGEX);

	string_replace(" *= *", " = color_class ", REGEX);
	string_replace("%[(.* on .*)%]", "#1", REGEX);
	string_replace(" on 0x1000000;", ";", REGEX);

	stuff("// Automatically generated by import-colors.\n");
	stuff("// Save this file, then press Alt-F3 to import colors.\n\n");
	stuff("#include \"eel.h\"\n\n");
	point = 0;
	buf_delete(tmp);
	add_color_class_declarations();
	say("Done.");
}

export_colors_one_scheme(int i)
{
	char *nm;
	int j, k, fg, bg;

	nm = name_name(i);
	noteput("Working...%s", nm);
	bprintf("color_scheme \"%s\" {\n", nm);
	save_var selected_color_scheme = i;
	for (j = 0; j < number_of_color_classes(); j++) {
		k = get_foreground_color(j, 1);
		if (k & ECOLOR_UNKNOWN)
			continue;
		bprintf("\tcolor_class %s ", name_color_class(j));
		if (k & ECOLOR_COPY) {
			bprintf("= color_class %s;\n", name_color_class(COLOR_STRIP_ATTR(k)));
			continue;
		}
		fg = get_foreground_color(j, 0);
		bg = get_background_color(j, 0);
		bprintf("MAKE_RGB(%d, %d, %d)",
				GETRED(fg), GETGREEN(fg), GETBLUE(fg));
		if (fg & EFONT_BOLD)
			bprintf(" | EFONT_BOLD");
		if (fg & EFONT_ITALIC)
			bprintf(" | EFONT_ITALIC");
		if (fg & EFONT_UNDERLINED)
			bprintf(" | EFONT_UNDERLINED");
		if (!j || !(bg & ETRANSPARENT)
			  && bg != get_background_color(0))
			bprintf(" on MAKE_RGB(%d, %d, %d)",
					GETRED(bg), GETGREEN(bg), GETBLUE(bg));
		stuff(";\n");
	}
	stuff("};\n\n");
}

// Create an EEL file with the current color settings.

command export_colors()
{
	char name[FNAMELEN];
	int i, max = final_index();

	iter = 0;
	strcpy(name, "mycolors.e");
	absolute(name);
	if (warn_existing_file(name) > 0)
		return;
	do_find(name, FILETYPE_AUTO);
	buf_zap(bufnum);
	stuff("// Automatically generated by export-colors.\n\n");
	stuff("#include \"eel.h\"\n\n");
	if (has_arg) {
		for (i = 1; i <= max; i++) {
			check_abort();
			if (name_type(i) == NT_COLSCHEME)
				export_colors_one_scheme(i);
		}
	} else
		export_colors_one_scheme(selected_color_scheme);
	restore_vars();
	point = 0;
	for (i = 0; i < 10; i++) {	// Assume class names have no more than 10 -'s.
		noteput("Working...%d", i);
		string_replace("(color_class .*)-", "#1_", REGEX);
		if (!replace_num_changed)	// Convert -'s in class names to _'s.
			break;
	}
	add_color_class_declarations();
	substitute_color_names();
	say("Done.");
}

// For each "= color_class foo;", add a declaration at the top.
add_color_class_declarations()
{
	int tmp;

	tmp = tmp_buf();
	buf_xfer(tmp, 0, size());
	save_var bufnum = tmp, show_status = 0;
	point = 0;
	do_drop_matching_lines(0, "= color_class", 0);
	point = 0;
	string_replace(".* = ", "", REGEX);
	point = 0;
	do_sort_region(0, size(), 0, 0);
	point = 0;
	do_uniq(1, 1, 0);
	restore_vars();
	point = 0;
	re_search(1, "(#include .*\n)+\n*");
	to_begin_line();
	grab_buffer(tmp);
	stuff("\n");
	buf_delete(tmp);
}


// Replace MAKE_RGB() references in the current buffer with
// the matching color names by interpreting the eel.h file.

substitute_color_names()
{
	char *codes, name[100], defn[100];
	int tmp, orig = bufnum;

	codes = lookpath("eel.h");
	if (!codes)
		codes = lookpath("include/eel.h");
	if (!codes)
		error("Can't locate eel.h to substitute color names");
	bufnum = tmp = tmp_buf();
	if (err_file_read(codes, FILETYPE_AUTO))
		quick_abort();
	filename = "";	// Don't accidentally overwrite the real eel.h.
	save_var case_fold = 1;
	if (!re_search(1, "#define[ \t]+black"))
		error("Can't locate start of color names");
	delete(0, give_begin_line());
	if (!re_search(1, "#define[ \t]+END_COLOR_MARKER"))
		error("Can't locate end of color names");
	delete(give_begin_line(), size());
	point = 0;
	string_replace(",[ \t]*", ", ", REGEX);	// standardize MAKE_RGB() calls
	while (re_search(1, "#define[ \t]+")) {
		if (!parse_string(1, "[a-z0-9_]+", name))
			error("Can't interpret color name");
		point += parse_string(1, "[ \t]*");
		if (!parse_string(1, "MAKE_RGB%([a-z0-9, \t]+%)", defn))
			error("Can't interpret color definition");
		bufnum = orig;
		string_replace(defn, name, 0);
		bufnum = tmp;
	}
	bufnum = orig;
	buf_delete(tmp);
}

// In a list-all buffer, delete some variable references that aren't
// preservable customizations.
drop_boring_change_lines()
{
	point = 0;
	string_replace("%+(maxmem|minmem|m?colors|normal-cursor"
				   "|-display-characters|buf-accessed-clock|c-look-back"
				   "|-bookmark-.*|-num-kill|-cur-kill-buf|noinit-flag"
				   "|-errors-in-cmd-file|-prev-kbd-extended|eel-version"
				   ") .*\n", "", REGEX);
	string_replace("when-loading: .*\n", "", REGEX);
}

command load_changes()
{
	char name[FNAMELEN], *tmp;
	int i;

	get_file(name, "Load changes from file", "changed");
	tmp = temp_buf();
	to_buffer(tmp);
	if (read_file(name, FILETYPE_MSDOS))
		quick_abort();
	drop_boring_change_lines();
	if (re_search(1, "^%&.*\n"))
		bprintf("# COLOR SETTINGS IGNORED: use import-colors\n");
	point = 0;
	string_replace("^%&.*\n", "", REGEX);	// discard color info
	save_var _abort_load_on_error = 0;
	parse_cmds();
	point = 0;
	i = *get_color_scheme_variable();
	if (name_type(i) == NT_COLSCHEME)
		selected_color_scheme = i;
}

parse_cmds()	/* load cmds in current buf, return 1 if error occurred */
{
	point = 0;
	save_var _errors_in_cmd_file = 0;
	save_var now_parsing_cmds = 1;
    jmp_buf my_jbuf;
    save_var jbuf = &my_jbuf;
	if (setjmp(jbuf))			/* errors come here */
		if (_abort_load_on_error)
			return 1;
	for (;;) {
		re_search(1, "[ \n\t]*");	/* skip whitespace */
		if (point >= size() - narrow_end)
			break;
		if (_abort_load_on_error != 2)
			noteput("Loading: %d%%...", muldiv(point, 100, size()));
		switch (curchar()) {
			case ';':		/* found comment */
			case '#': nl_forward(); break;
			case '(': paren_cmds(); break;
			case '+': load_var(); break;
			case '~': load_binding(); break;
			case '&': load_color(); break;
			default:  load_other(); break;
		}
	}
	copy_bindings();
	int i = *get_color_scheme_variable();
	if (name_type(i) == NT_COLSCHEME)
		selected_color_scheme = i;
	if (!_abort_load_on_error
		  || !_errors_in_cmd_file && _abort_load_on_error != 2)
		say("%d error%s detected.", _errors_in_cmd_file,
			  _errors_in_cmd_file == 1 ? "" : "s");
	else if (_abort_load_on_error != 2)
			say("%d error%s detected, see buffer %s.", _errors_in_cmd_file,
				  _errors_in_cmd_file == 1 ? "" : "s", MSG_BUFFER);
	return _errors_in_cmd_file != 0;
}

load_var()	/* variable definition */
{
	int i, new_type, vtype, hasnum;
	char name[80], junk[80], num[20], *p, *tmp;

	point++;			/* skip initial + */
	parse_string(1, "[^ \t\n:]+", name);
	i = find_index(name);
	new_type = parse_string(1, " buffer", junk)
			? NT_BUFVAR : parse_string(1, " window", junk)
			? NT_WINVAR : NT_VAR;
	if (parse_string(1, " (char ptr )?variable: %<complex>", junk)) {
		if (i <= 0 || name_type(i) != new_type)
			init_warning("Variable %s too complex to define.", name);
		nl_forward();
		return;
	} else if (parse_string(1, " char array ", junk))
		vtype = TYPE_CARRAY;
	else if (parse_string(1, " int ", junk))
		vtype = TYPE_INT;
	else if (parse_string(1, " int32 ", junk))
		vtype = TYPE_INT32;
	else if (parse_string(1, " short ", junk))
		vtype = TYPE_SHORT;
	else if (parse_string(1, " char ", junk))
		vtype = TYPE_CHAR;
	else
bad:		init_error("Bad definition for variable %s.", name);
	if (!parse_string(1, "variable: ", junk))
		goto bad;
	hasnum = parse_string(1, "[-0-9][a-zA-Z0-9]*", num);
	if (vtype == TYPE_CARRAY) {
		if (i <= 0 && hasnum) {
			i = new_variable(name, new_type, TYPE_CARRAY, numtoi(num));
			set_name_user(i, 1);		// Make it a user variable.
		}
		if (!parse_string(1, " *<dquote>", junk)
				|| name_type(i) != new_type
				|| vartype(i) != TYPE_CARRAY)
			goto bad;
		use_default = 1;
		if (hasnum && new_type == NT_VAR && numtoi(num) > ptrlen(varptr(i))) {
			drop_name(name);		// Make it bigger by recreating it.
			i = new_variable(name, new_type, TYPE_CARRAY, numtoi(num));
			set_name_user(i, 1);		// Make it a user variable.
			tmp = p = varptr(i);
		}
		tmp = p = varptr(i);
		use_default = 0;
		for (; curchar() != '"'; point++) {
			if (ptrlen(p) < 1)
				init_error("Definition too long for variable %s.", name);
			if (curchar() != '\\') {
				*p++ = curchar();
				continue;
			}
			point++;
			switch (curchar()) {
				case 't': *p++ = '\t'; break;
				case 'r': *p++ = '\r'; break;
				case 'n': *p++ = '\n'; break;
				case 'z': *p++ = 0; break;
				default: *p++ = curchar();
			}
		}
		while (p - tmp < ptrlen(tmp))
			*p++ = 0;
		if (!parse_string(1, "<dquote>$", junk))
			goto bad;
		char when_setting[FNAMELEN];
		sprintf(when_setting, "when-setting-%s", name_name(i));
		try_calling(when_setting);
	} else if (vtype) {
		if (i <= 0) {
			i = new_variable(name, new_type, vtype, 1);
			set_name_user(i, 1);		// Make it a user variable.
		}
		if (!hasnum && parse_string(1, ".*", junk))
			sprintf(num, "%d", find_index(junk));
		set_var_val(i, num, 2);
	}
}

load_color()
{
	int i, j, fore, back;
	char scheme[FNAMELEN], class[FNAMELEN], num[FNAMELEN], num2[FNAMELEN];

	point++;			/* skip initial & */
	parse_string(1, "[^ \t\n:]+", scheme);
	i = find_index(scheme);
	if (!parse_string(1, " color class for ", class)
			|| !parse_string(1, "[^ \t\n:]+", class)
			|| !parse_string(1, ": ", num)
			|| !parse_string(1, "[[=]?[-a-zA-Z0-9]+", num)) {
		init_warning("Bad color definition for %s.", scheme);
		nl_forward();
		return;
	}
	if (i <= 0)
		i = new_variable(scheme, NT_COLSCHEME, 0);
	for (j = 0; strcmp(class, name_color_class(j)); )
		if (++j >= number_of_color_classes()) {
			init_warning("Undefined color class %s.", class);
			nl_forward();
			return;
		}
	if (*num == '=') {		// copy definition
		for (fore = 0; strcmp(num + 1, name_color_class(fore)); )
			if (++fore >= number_of_color_classes()) {
				init_warning("Undefined color class %s.", num + 1);
				nl_forward();
				return;
			}
		fore |= ECOLOR_COPY;
		back = 0;
	} else if (*num == '[') {		// rgb-style definition
		fore = numtoi(num + 1);
		if (!parse_string(1, " on ", num)
			  || !parse_string(1, "[0-9a-fx]+]", num)) {
			init_warning("Bad rgb color definition for %s.", scheme);
			nl_forward();
			return;
		}
		back = numtoi(num);
		fore = SWAP_OLDCOLOR(fore);		// If that's all, then it uses pre-v14 coding.
		back = SWAP_OLDCOLOR(back);
		if (parse_string(1, "#RGB %[", num)) {	// Found new-style coding.
			if (!parse_string(1, "[0-9a-fx]+", num)
				  || !parse_string(1, " on ", num2)
				  || !parse_string(1, "[0-9a-fx]+]", num2)) {
				init_warning("Bad new-rgb color definition for %s.", scheme);
				nl_forward();
				return;
			}
			fore = numtoi(num);
			back = numtoi(num2);
		}
	} else {
		fore = find_color_code(num);
		if (fore < 0) {
			init_warning("Bad color name %s.", num);
			nl_forward();
			return;
		}
		if (parse_string(1, " on ", num)
			&& parse_string(1, "[-a-zA-Z0-9]+", num)) {
			back = find_color_code(num);
			if (back < 0) {
				init_warning("Bad color name %s.", num);
				nl_forward();
				return;
			}
		}
		fore = attr_to_rgb(fore);
		back = attr_to_rgb(back);
	}
	set_color_pair(j, fore, back, i);
}

char color_names[] = "|black|blue|green|cyan|red|magenta|brown|white|"
		     "brightblack|brightblue|brightgreen|brightcyan|"
		     "brightred|brightmagenta|yellow|brightwhite|";

give_color_name(s, i)	// copy name of color i (0..15) into char array
char *s;
{
	char *p;
	int len;

	strcpy(s, "unknown");
	if (i < 0 || i > 15)
		return;
	for (p = color_names + 1; i > 0; i--, p++)
		if (!(p = index(p, '|')))
			return;
	len = index(p, '|') - p;
	strncpy(s, p, len);
	s[len] = 0;
}

give_color_combo(s, i)	// copy name of color combo i (0..255) into char array
char *s;
{
	if (i < 256) {
		give_color_name(s, i & 15);
		s += strlen(s);
		s += sprintf(s, " on ");
		give_color_name(s, (i >> 4) & 15);
	} else
		sprintf(s, "0x%x", i);
}

find_color_code(s)	// return 0..15 for given color code, or -1 if bad
char *s;
{
	char pat[50], *p, *q;
	int i;

	if (isdigit(*s))
		return numtoi(s);
	sprintf(pat, "|%s|", s);
	p = strstr(color_names, pat);
	if (!p)
		return -1;
	for (i = 0, q = color_names + 1; q < p; q++, i++)
		if (!(q = index(q, '|')))
			return -1;
	return i;
}

load_binding()		/* new-style binding declaration */
{
	char name[80], junk[80];
	int *onekey, start;
	short *tab;
	int i, now_type, cnt = 1;

	point++;			/* skip initial ~ */
	parse_string(1, "[^ \t\n:]+", name);
	i = find_index(name);
	now_type = (i > 0) ? name_type(i) : -1;
	if (now_type != NT_TABLE)
		init_error("%s is not a table.", name);
	tab = index_table(i);
	point++;			/* skip space before key */
	onekey = get_keycode();
	if (!onekey || onekey[0] != 2)
		init_error("Bad key name.%s", "");
	start = onekey[1];
	if (parse_string(1, " \"")) {
		point++;
		onekey = get_keycode();
		if (!onekey || onekey[0] != 2)
			init_error("Bad final key name.%s", "");
		cnt = onekey[1] - start + 1;
	}
	if (!parse_string(1, ": ", junk) || !parse_string(1, ".*", name))
		init_error("Bad key binding.%s", "");
	i = find_index(name);
	if (i <= 0) {		/* allow name table index too */
		i = numtoi(name);
		if (!i && *name != '0') {	/* not a name, not a number */
			init_warning("No such command %s.", name);
			return;
		}
	}
	set_range(tab, start, i, cnt);
	rebuild_menu();
}

load_other()
{
	char name[80], junk[20];
	int i;

	parse_string(1, "[^ :\n]+", name);
	i = find_index(name);
	if (parse_string(1, ": macro ", junk)) {
		get_keyseq();
		if (i > 0)
			if (name_type(i) == NT_MACRO || !strcmp(name, "last-kbd-macro"))
				drop_name(name);
			else {
				init_warning("%s is already defined.  Use Alt-x delete-name "
							 "to overwrite.", name);
				return;
			}
		name_macro(name, keyseq);
	} else if (parse_string(1, ": table", junk)) {
		if (i > 0) {
			if (name_type(i) == NT_TABLE)
				return;
			drop_name(name);
		}
		new_table(name);
	} else if (parse_string(1, ": command", junk)) {
		if (i <= 0 || name_type(i) != NT_COMMAND
			  && name_type(i) != NT_AUTOLOAD)
			init_warning("%s is not a currently defined command.", name);
	} else if (parse_string(1, ": subroutine", junk)) {
		if (i <= 0 || name_type(i) != NT_SUBR
			  && name_type(i) != NT_AUTOSUBR)
			init_warning("%s is not a currently defined subroutine.", name);
	} else
		init_error("%s: unrecognized syntax.", name);
}

paren_cmds()		/* handle parenthesized commands */
{
	char funcname[80], numarg[80];
	int i;

	must_char('(');
	re_search(1, "[ \t\n]*");
	*numarg = 0;
	parse_string(1, "-?[0-9][0-9a-z]*[ \t\n]+", numarg);
	parse_string(1, "[^ \t\n()\"']*", funcname);
	if (!strcmp(funcname,"define-macro"))
		cf_macro();
	else if (!strcmp(funcname, "bind-to-key"))
		cf_bind();
	else if (!strcmp(funcname, "create-prefix-command"))
		cf_prefix();
	else if (!strcmp(funcname, "set-variable"))
		cf_set_variable();
	else if (!strcmp(funcname, "change-name"))
		cf_change_name();
	else if (i = is_runnable(funcname))
		cf_any(i, numarg);
	else
		init_error("Bad command %s in command file. ", funcname);
	re_search(1, "[ \t\n]*");
	must_char(')');
}

cf_macro()		/* handle define-macro */
{
	char name[80];

	parse_var_name(name);
	get_keyseq();
	name_macro(name, keyseq);
}

cf_bind()		/* handle bind-to-key */
{
	int index;
	char name[80];

	parse_var_name(name);
	index = find_index(name);
	if (index)
		put_on_key(index);
	else
		init_error("No such command %s.", name);
	rebuild_menu();
}

cf_prefix()		/* handle create-prefix-command */
{
	int kt = make_anon_keytable();

	put_on_key(kt);
	set_case_indirect(index_table(kt));
	set_mouse_keys(index_table(kt));
}

put_on_key(index)		/* parse key seq and bind it to index */
{
	short *curtab;
	int c, i = 1;

	get_keyseq();
	curtab = root_keys;
	for (;;) {
		c = keyseq[i];
		if (++i >= keyseq[0]) break;
		if (name_type(curtab[c]) != NT_TABLE)
			init_error("Expected \" in sequence%s.", "");
		curtab = index_table(curtab[c]);
	}
	curtab[c] = index;
}

get_keyseq()			/* parse key names, put keys in keyseq */
{
	re_search(1, "[ \t\n]*");
	keyseq = get_keycode();
	if (!keyseq)
		init_error("Bad character.%s", "");
}

get_str(s)		/* get a quoted string, and put it in s */
char *s;
{
	re_search(1, "[ \t\n]*");
	must_char('"');
	parse_string(1, "(<^dquote|\\>|\\<any>)*", s);
	must_char('"');
}

dequote_str(char *s)
{
	for (; *s; s++)
		if (*s == '\\')
			strcpy(s, s + 1);
}

// Parse a variable name, optionally quoted, and put in s.
parse_var_name(char *s)
{
	re_search(1, "[ \t\n]*");
	if (!parse_string(1, "[-_a-zA-Z][-_a-zA-Z0-9]*", s))
		get_str(s);
}

must_char(c)	/* move past current character, making sure that it's c */
{
	char buf[2];

	if (curchar() != c) {
		sprintf(buf,"%c", c);
		init_error("Expected %s.", buf);
	} else
		point++;
}

have_char(c)	/* if this character is c, move past and return 1, else 0 */
{
	int res = (curchar() == c);

	if (res) point++;
	return res;
}

// Get variable name, value, possible default option, and set.
cf_set_variable()
{
	char varname[FNAMELEN], val[FNAMELEN], def[FNAMELEN];
	int i;

	parse_var_name(varname);
	point += parse_string(1, "[ \t\n]*");
	if (!parse_string(1, "-?[0-9][0-9a-zA-Z]*", val)) {
		get_str(val);
		dequote_str(val);
	}
	point += parse_string(1, "[ \t\n]*");
	*def = 0;
	parse_string(1, "-?[0-9][0-9a-zA-Z]*", def);
	i = find_index(varname);
	if (!vartype(i))
		init_error("%s: not a variable", varname);
	if (!var_settable(varname))
		init_error("%s: can't set variable of this type", varname);
	safe_set_var_val(i, val, *def ? numtoi(def) : 3);
}

cf_change_name()
{
	char old[FNAMELEN], new[FNAMELEN];

	parse_var_name(old);
	point += parse_string(1, "[ \t\n]*");
	parse_var_name(new);
	switch (replace_name(old, new)) {
		case 1: init_error("%s: not found!", old); break;
		case 2: init_error("%s exists, delete first", new); break;
	}
}

// Parse args and run command i, using numeric prefix arg in numarg if
// nonempty.
cf_any(int i, char *prefix_arg)
{
	int numargs[10], j, numcnt = 0, strcnt = 0;
	char strbuf[3000], *s, *p, after[300], argtypes[30], *strarray[10];
	jmp_buf *old_level = top_level, this_level;

	save_var bufnum;
	*argtypes = 0;
	s = strbuf;
	for (;;) {
		point += parse_string(1, "[ \t\n]*");
		if (curchar() == ')')
			break;
		if (parse_string(1, "-?([0-9][0-9xoa-f]*|'.'|'\\.')", s)) {
			numargs[numcnt++] = numtoi(s);
			if (got_bad_number)
				init_error("Bad number %s in command file.", s);
			strcat(argtypes, "i");
		} else if (curchar() == '"') {
			point++;
			j = parse_string(1, "([^\"]|\\<any>)*\"");
			if (j > ptrlen(s) - 1)
				init_error("String too long in command file.%s", "");
			grab(point, point + j - 1, s);
			point += j;
			for (p = s; *p; p++)
				if (p[0] == '\\' && p[1]) {
					strcpy(after, p + 1);
					strcpy(p, after);
				}
			strarray[strcnt] = s;
			strcnt++;
			s += strlen(s) + 1;
			strcat(argtypes, "s");
		} else
			init_error("Error in command file.%s", "");
	}
	top_level = &this_level;
	save_var iter = 1, has_arg = 0;
	if (*prefix_arg)
		iter = numtoi(prefix_arg), has_arg = 1;
	if (!setjmp(top_level))
		call_with_arg_list(i, argtypes, numargs, strarray);
	top_level = old_level;
}

init_error(msg, s)		/* display error and abort loading */
char *msg, *s;
{
	init_warning(msg, s);
	longjmp(jbuf, 1);
}

init_warning(msg, s)		/* display error and return */
char *msg, *s;
{
	char full_msg[FNAMELEN];
	_errors_in_cmd_file++;
	if (_abort_load_on_error == 1) {
		sprintf(full_msg, msg, s);
		say("%s (line %d)", full_msg, lines_between(0, point, 0) + 1);
	} else if (_abort_load_on_error == 2) {
		buffer_printf(MSG_BUFFER, "%s %d: ", einit_file_name,
					  lines_between(0, point, 0) + 1);
		buffer_printf(MSG_BUFFER, msg, s);
		buffer_printf(MSG_BUFFER, "\n");
	} else {
		if (!nl_forward())
			insert('\n');
		stuff("# COMMAND FILE ERROR: ");
		bprintf(msg, s);
		stuff("\n");
	}
}

command insert_macro()		/* insert command-file definition of macro */
{
	char name[80];
	int i;

	save_var show_mouse_choices = 1;
	get_macname(name, "Insert macro", "last-kbd-macro");
	i = find_index(name);
	if (i && name_type(i) == NT_MACRO) {
		bprintf("(define-macro \"%s\" ", name);
		stuff_macro(get_macro(i), 0);
		stuff(")\n");
	} else
		error("Not a macro");
}

command insert_binding()
{
	char name[300];
	int index;

	index = *ask_key("Save binding of key: ", name, 0);
	if (index <= 0)
		error("Key not bound");
	bprintf("(bind-to-key \"%s\" ", name_name(index));
	stuff_macro(key_binding, 1);
	bprintf(")\n");
}

stuff_macro(m, oneline)	/* insert cmd-file format macro text in buffer */
int *m;		// if oneline, use \n instead of actual newline
{
	char tmp[300];
	int i, prev;

	stuff("\"");
	for (i = 1; i < m[0]; i++)
		switch (m[i]) {
			case '-':	prev = character(point - 1);
					if (index("ACSFMN", prev))
						insert('\\');
				/* "F-1" in text resembles special key */
			default:			/* fall through */
					if (!oneline && m[i] == '\n') {
						insert('\n');
						break;
					}
					tmp[0] = 0;
					show_char(tmp, m[i]);
					stuff(tmp);
					break;
			case CTRL('\\'):
					stuff("C-\\\\"); break;
			case ALT(CTRL('\\')):
					stuff("C-A-\\\\"); break;
			case ALT('<'):
			case ALT('"'):
			case ALT('\\'):	stuff("A-");	/* fall through */
			case '<':		/* quote these */
			case '"':
			case '\\':	insert('\\');
					insert(m[i] & 0x7f);
					break;
		}
	stuff("\"");
}

// Output color info on this class in current scheme to buf.
list_one_color_setting(int buf, int class, int raw)
{
	char colorpair[FNAMELEN];
	int fg, bg;

	fg = get_foreground_color(class, raw);
	if ((fg & ETRANSPARENT) && !raw)
		fg = get_foreground_color(0);
	bg = get_background_color(class, raw);
	if ((bg & ETRANSPARENT) && !raw)
		bg = get_background_color(0);
	if (fg & ECOLOR_COPY)
		sprintf(colorpair, "=%s", name_color_class(COLOR_STRIP_ATTR(fg)));
	else if (!(fg & ~0xffffff) && fg == attr_to_rgb(rgb_to_attr(fg))
			 && !(bg & ~0xffffff) && bg == attr_to_rgb(rgb_to_attr(bg)))
		give_color_combo(colorpair, alter_color(class, -1));
	else
		sprintf(colorpair, "[0x%x on 0x%x]#RGB [0x%x on 0x%x]",
				SWAP_OLDCOLOR(fg), SWAP_OLDCOLOR(bg), fg, bg);
	buf_printf(buf, "&%s color class for %s: %s\n",
			   name_name(selected_color_scheme),
			   name_color_class(class), colorpair);
}

list_color_scheme(int i, int raw)
{
	int j;

	save_var selected_color_scheme = i;
	for (j = 0; j < number_of_color_classes(); j++)
		list_one_color_setting(bufnum, j, raw);
}

command list_colors()
{
	int i, max = final_index();

	tiled_only();
	zap("list-colors");
	to_buffer("list-colors");
	for (i = 1; i <= max; i++) {
		check_abort();
		if (name_type(i) == NT_COLSCHEME)
			list_color_scheme(i, 0);
	}
	point = 0;
}

list_table(int i)
{
	char tabname[FNAMELEN];
	int j, onekey[2], cnt;
	short *k;

	strcpy(tabname, name_name(i));
	bprintf("%s: table\n", tabname);
	k = index_table(i);
	onekey[0] = 2;
	for (j = 0; j < MAXKEYS; j += cnt) {
		cnt = get_range(k, j, k[j]);
		if (k[j] > 0) {
			bprintf("~%s ", tabname);
			onekey[1] = j;
			stuff_macro(onekey, 1);
			if (cnt > 1) {
				bprintf(" ");
				onekey[1] = j + cnt - 1;
				stuff_macro(onekey, 1);
			}
			if (name_name(k[j]))
				bprintf(": %s\n", name_name(k[j]));
			else
				bprintf(": %d\n", k[j]);
		}
	}
}

// Append to buffer b a description of name table entry i.  Flags:
// 1=ignore non-user variables, 2=include color schemes, 4=use final
// colors in schemes, ignoring flags to build colors from others.
list_one(int b, int i, int flags)
{
	int j, last, val, type;
	char *p, *nm, *tname;

	save_var bufnum = b;
	save_var point = size();
	save_var use_default;
	nm = name_name(i);
	switch (name_type(i)) {
		case NT_COMMAND:
		case NT_AUTOLOAD:
			bprintf("%s: command\n", nm); break;
		case NT_SUBR:
		case NT_AUTOSUBR:
			bprintf("%s: subroutine\n", nm); break;
		case NT_MACRO:
			bprintf("%s: macro ", nm);
			stuff_macro(get_macro(i), 1);
			insert('\n');
			break;
		case NT_TABLE:
			list_table(i);
			break;
		case NT_WINVAR:
		case NT_BUFVAR: case NT_VAR:
			use_default = 1;
			p = varptr(i);
			type = 0;
			if ((flags & 1) && !name_user(i) && (strncmp(nm, "-our-", 5)
				|| !strstr(nm, "-scheme")))
				break;
			switch (vartype(i)) {
				case TYPE_BYTE:	val = *(byte *) p;
								tname = "byte ";
								break;
				case TYPE_CHAR:	val = *(char *) p;
								tname = "char ";
								break;
				case TYPE_SHORT: val = *(short *) p;
								 tname = "short ";
								 break;
				case TYPE_INT32: val = *(int32 *) p;
								tname = "int32 ";
								break;
				case TYPE_INT:	val = *(int *) p;
								tname = "int ";
								break;
				case TYPE_CPTR:	tname = "char ptr ";
								type = 2; break;
				case TYPE_CARRAY: tname = "char array ";
								  type = 1; break;
				default:	tname = ""; type = 2; break;
			}
			bprintf("+%s %s%svariable", nm,
					(name_type(i) == NT_WINVAR)
					? "window " :
					(name_type(i) == NT_BUFVAR)
					? "buffer " : "", tname);
			if (type == 0) {
				if (val && !strncmp(nm, "-our-", 5)
					  && parse_string(-1, "^%+-our-(color|mono|gui"
									  "|unixconsole)-scheme .*"))
					bprintf(": %s\n", name_name(val));
				else
					bprintf(": %d\n", val);
			} else if (type == 2)
				stuff(": <complex>\n");
			else {
				last = ptrlen(p);
				bprintf(": %d \"", last);
				while (last > 0 && !p[last - 1])
					last--;
				for (j = 0; j < last; j++)
					switch (p[j]) {
						case '\t':	stuff("\\t"); break;
						case '\r':	stuff("\\r"); break;
						case '\n':	stuff("\\n"); break;
						case 0:		stuff("\\z"); break;
						case '\\':
						case '"':	insert('\\');
						default:	insert(p[j]); break;
					}
				stuff("\"\n");
			}
			use_default = 0;
			break;
		case NT_COLSCHEME:
			if (flags & 2)
				list_color_scheme(i, !(flags & 4));
			break;
	}
}

// List everything in name table to buffer buf, plus a few special
// things, according to flags: 1=ignore non-user variables, 2=include
// color schemes, 4=use final colors in schemes, ignoring flags to build
// colors from others, 32=check for aborting, 64=show progress.

list_all_to_buffer(int buf, int flags)
{
	save_var bufnum = buf, use_default;
	int i, max = final_index(), percent, opercent = -1;
	for (i = 1; i <= max; i++) {
		if (flags & 32)
			check_abort();
		if ((flags & 64) && (percent = muldiv(i, 100, max)) != opercent) {
			noteput("Describing state: %d%%...", percent);
			opercent = percent;
		}
		list_one(bufnum, i, flags);
	}
	point = size();
	buf_printf(bufnum, "+path-sep char variable: %d\n", path_sep);
	buf_printf(bufnum, "+abort-key int variable: %d\n", abort_key);
}

command list_all()
{
	int flags = 32 | 64;
	tiled_only();
	zap("list-all");
	to_buffer("list-all");
	save_var use_default;
	if (has_arg)
		flags |= iter & 31;
	iter = 1;
	list_all_to_buffer(bufnum, flags);
	point = 0;
	say("");
}
