/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1988, 2021 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// This file contains the commands relating to tags.

#include "eel.h"
#include "tags.h"
#include "c.h"
#include "colcode.h"
#include "bufed.h"
#include "browse.h"

init_tags()		/* read tags if necessary */
{
	char buf[FNAMELEN];

	if (tags_from_bsc)
		return;
	if (!exist("-tags")) {
		if (!look_up_tree(buf, initial_tag_file, NULL, NULL))
			strcpy(buf, initial_tag_file);
		absolute(buf);
		sayput("Loading tags from %s...", buf);
		load_tags(buf);
		say("");
	}
	_tags_buffer = name_to_bufnum("-tags");
}

tags_available()
{
	char buf[FNAMELEN];

	if (tags_from_bsc)
		return 0;
	return (exist("-tags") || look_up_tree(buf, initial_tag_file, NULL, NULL));
}

tags_header()	/* return length of tags file header */
{		/* header line says if tags file is sorted */
	point = 0;		/* also goes to top */
	return parse_string(1, "<tab>TAGS<tab>(UN)?SORTED\n", NULL);
}

are_tags_sorted()	/* -1: no header, 0=not sorted, 1=sorted */
{			/* also goes to top */
	point = 0;
	if (tags_format)
		return -1;
	return tags_header() ? (character(6) == 'S') : -1;
}

mark_tags(sorted)	/* update header line (adding if needed) */
{
	save_var bufnum = _tags_buffer;
	if (!tags_format && are_tags_sorted() != sorted) {
		delete(0, tags_header());
		bprintf("\tTAGS\t%sSORTED\n", sorted ? "" : "UN");
	}
}

load_tags(file)
char *file;
{
	int i;

	save_var synchronous_file_io = 1;
	tags_from_bsc = 0;
	tags_format = TAG_FORMAT_NATIVE;
	*_last_tag = *_last_tag_file = 0;
	if (!strfcmp(get_extension(file), ".bsc")) {
		tags_from_bsc = 1;
		strcpy(bsc_file, file);
		absolute(bsc_file);
		return;
	}
	save_var bufnum = _tags_buffer = zap("-tags");
	i = read_file(file, FILETYPE_AUTO);
	set_read_only(0);
	if (i == 2)			/* new file */
		mark_tags(0);		/* make new-fmt tags file */
	else if (i) {			/* error reading file */
		restore_vars();
		buf_delete(_tags_buffer);
		quick_abort();
	} else if (!tags_header()) {
		if (!ask_yn("Convert", "Convert old-format tags file? ",
					"Yes", "No", 1))
			quick_abort();
		sayput("Converting old-format tags file...");
		if (!string_replace(";", "\t", 0)) {
			mark_tags(0);
			do_save_tags();
		}
	}
}

do_save_tags()		/* save tags to their file if unsaved */
{
	int oldbuf = bufnum;

	if (_tags_buffer && buf_exist(_tags_buffer) && auto_save_tags && !tags_format) {
		bufnum = _tags_buffer;
		if (modified) {
			if (size() > 2000000)
				sayput("Saving tags...");
			do_save_file(1, 1, 1);
		}
		bufnum = oldbuf;
	}
}

command sort_tags()	/* sort & save if needed */
{
	init_tags();
	if (tags_from_bsc || tags_format)
		error("This type of tag file may not be sorted.");
	bufnum = _tags_buffer;
	if (are_tags_sorted() == 1)
		return;
	sayput("Sorting tags...");
	save_var narrow_start = tags_header();	/* don't sort header */
	save_var show_status = 1;
	save_var abort_searching = ABORT_ERROR;
	if (sort_another("-tags", 0, 0))
		return;		/* user aborted */
	narrow_start = 0;
	delete(0, tags_header());	/* remove header */
	mark_tags(1);		/* add header saying it's sorted */
	do_save_tags();
	say("");
}

char *tag_match(s, flags)
char *s;
{
	char tmp[TAGLEN];

	if (tags_from_bsc)
		return browse_matcher(s, flags);
	save_var bufnum;
	save_var abort_searching = 0;
	bufnum = _tags_buffer;
	save_var case_fold = !tag_case_sensitive;
	if (flags & STARTMATCH)
		point = 0;
	if (*s) {
		sprintf(tmp, "^([^\t\n]+(::|'))?%s", s);
		do {
			if (!search(1, s))
				return 0;
		} while (!parse_string(-1, tmp)
				 || (flags & EXACTONLY) && curchar() != '\t');
		to_begin_line();
	} else if (!nl_forward() || point == size())
		return 0;
	parse_string(1, "[^\t\n]+", _this_match);
	to_end_line();
	return _this_match;
}

	/* make list of matches for s in mbuf */
// If s is empty, show list of matches for last tag.
tag_lister(char *s, char *(*finder)(), int flags, int mbuf)
{
	char pat[TAGLEN], look_for[TAGLEN];
	int sorted;

	finder = finder;	/* silence eel's warning */
	flags = flags;
	save_var bufnum;
	save_var abort_searching = 0;
	if (!*s) {
		tags_look_for(look_for);
		if (*look_for) {
			s = look_for;
			save_var tag_list_exact_only = 1;
		}
	}
	if (want_sorted_tags && !*s && !tags_format)
		sort_tags();
	bufnum = _tags_buffer;
	save_var case_fold = !tag_case_sensitive;
	sorted = are_tags_sorted();	/* also goes to top */
	if (*s) {
		sprintf(pat, "^([^\t\n]+(::|'))?%q", s);
		while (search(1, s)) {
			if (parse_string(-1, pat)
				  && (!tag_list_exact_only || curchar() == '\t')) {
				to_end_line();
				buf_xfer(mbuf, matchend, point + 1);
			}
		}
	} else {
		nl_forward();
		buf_xfer(mbuf, point, size());
	}
	bufnum = mbuf;
	if (!tags_format && want_sorted_tags && !sorted && !user_abort) {
		abort_searching = ABORT_ERROR;
		sort_another(bufnum_to_name(mbuf), 0, 0);
	}
	point = 0;
	if (tag_display_width)
		tab_size = tag_display_width;
	else
		tab_size = (screen_cols - _view_left - _view_right) / 2 - 1;
	goal_column = display_column = 0;
}

get_tag(res, pr)	/* do completion on tags */
char *res, *pr;		/* returns either tag, or tag <Tab> file */
{			/* or tag <Tab> file <Tab> pos */
	save_var menu_width = screen_cols - _view_left - _view_right;
	if (tag_display_width)
		menu_width = MIN(menu_width, tag_display_width * 2);
	if (!tags_from_bsc)
		save_var completion_lister = tag_lister;
	comp_read(res, pr, tag_match, (tag_case_sensitive ? 0 : COMP_FOLD)
			  | MUST_MATCH | NONE_OK | CAUTIOUS, "");
}

tag_searcher_c(char *tag)
{
	char classname[FNAMELEN], fulltag[FNAMELEN], pr[FNAMELEN], *s;
	int err, cnt = 0, selected = 0, try = 1;

	if (parse_string(-1, "-%>|%.[ \t\n]"))	// Ignore func's class.
		*classname = 0;
	else
		find_c_func_info(NULL, classname, NULL, 0);
again:
	strcpy(fulltag, tag);
	if (!strchr(tag, ':')) {		// Class not known yet.
withclass:
		if (*classname && strcmp(classname, tag)) {		// See if current class
			sprintf(fulltag, "%s::%s", classname, tag);	// has such a member.
			if (!(err = go_tag(fulltag, 1)))
				return err;
			bufnum = window_bufnum;
		}			// Is there more than one function by that name in any class?
		*fulltag = 0;
		bufnum = window_bufnum;
		if (s = tag_match(tag, STARTMATCH | EXACTONLY)) {
            strcpy(fulltag, s);
			do {
				cnt++;
			} while (tag_match(tag, EXACTONLY));
			if (cnt != 1) {
				sprintf(pr, "%d matches for %s: ", cnt, tag);
				save_var menu_width = screen_cols - _view_left - _view_right;
				if (tag_display_width)
					menu_width = MIN(menu_width, tag_display_width * 2);
				if (!tags_from_bsc)
					save_var completion_lister = tag_lister;
				save_var insert_default_response = 2, tag_list_exact_only = 1;
				ungot_key = '?';
				comp_read(fulltag, pr, tag_match, (tag_case_sensitive ? 0 : COMP_FOLD)
						  | MUST_MATCH | NONE_OK | CAUTIOUS, tag);
				selected = 1;
			}
		} else
			strcpy(fulltag, tag);
		err = go_tag(fulltag, 1);
	} else {
		err = go_tag(fulltag, 1);
		if (err && *classname)
			goto withclass;
	}
	strcpy(_last_tag, tag);
	bufnum = window_bufnum;
	if (!tags_format && err && !selected && find_tagger(0)) {
		// Try retagging current file before giving up.
		sprintf(pr, "Tag %s has moved, retag file %s? ",
				fulltag, filename);
		if (!tag_ask_before_retagging || try > 1
			  || ask_yn("Retag file", pr, "Yes", "No", 1)) {
			if (try < 2)
				do_retag_file(filename);
			err = go_tag(fulltag, 1);
			if (try < 2)
				do_save_tags();
			if (err && try++ < 2)
				goto again;
			if (!err)
				say("Found %s in this file.", *tag ? tag : "tag");
		}
	}
	if (!err && !selected && strcmp(tag, fulltag))
		say("Searched for %s, found %s", tag, fulltag);
	report_tag_error(err);
}

// Adjust the name of a tag to indicate what we'll really search for.
tag_prompter_c(char *tag)
{
	char *p = _last_tag, *q;

	for (;;) {
		if ((q = strstr(p, "::")) && q[2])
			p = q + 2;		// Remove classname:: prefix.
		else if ((q = strstr(p, "'")) && q[1])
			p = q + 1;		// Remove classname' prefix for Perl.
		else
			break;
	}
	strcpy(tag, p);
}

// Search for another instance of the last tag (1==next instance,
// -1==previous, 0==same as before.  Return nonzero if we couldn't find
// any such tag, zero if could (and move to it).

tag_to_next_c(char *pat, int n_fwd)
{
	char raw_tag[FNAMELEN];

	tag_prompter_c(raw_tag);		// strip class names
	sprintf(pat, "^([^\t\n]+(::|'))?%q\t", raw_tag);
	return !tag_search_by_offset(pat, n_fwd);
}

// For modes that don't specify different rules, use rules for C.
tag_searcher_default(char *s)
{
	tag_searcher_c(s);
}

tag_prompter_default(char *s)
{
	tag_prompter_c(s);
}

int tag_to_next_default(char *pat, int n_fwd)
{
	return tag_to_next_c(pat, n_fwd);
}

modal_goto_tag(char *tag)
{
	char fulltag[FNAMELEN], pr[FNAMELEN];
	int i, err;

	quiet_set_bookmark();
	i = get_mode_based_index("tag_searcher_%s");
	if (*tag && !strchr(tag, '\t'))		// If not yet fully specified,
		switch (name_type(i)) {		// Call mode-specific version.
			case NT_SUBR: case NT_COMMAND:
			case NT_AUTOSUBR: case NT_AUTOLOAD:
				(*(int (*)())i)(tag);
				return;
		}
	strcpy(fulltag, tag);
	err = go_tag(fulltag, 1);
	bufnum = window_bufnum;
	if (!tags_format && err && find_tagger(0)) {
		sprintf(pr, "Tag %s has moved, retag file %s? ", fulltag, filename);
		if (!tag_ask_before_retagging
			  || ask_yn("Retag file", pr, "Yes", "No", 1)) {
			do_retag_file(filename);
			err = go_tag(fulltag, 1);
			do_save_tags();
			if (!err)
				say("Found %s in this file.", *fulltag ? fulltag : "tag");
		}
	}
	report_tag_error(err);	// Default behavior.
}

// Put tag name at point in tag.
grab_tag_name(char *tag)
{
	save_var point;
	char *tag_pat = get_mode_string_variable("tag-pattern-%s");
	if (got_bad_number || !tag_pat)
		tag_pat = tag_pattern_default;
	point--;
	re_search(1, tag_pat);
	re_search(-1, tag_pat);
	re_search(1, tag_pat);	// Needed in case we start inside "operator".
	re_search(-1, tag_pat);
	grab(point, matchstart, tag);
}

int get_tag_window_direction(char *tag)
{
	while (get_direction() == -1 && (!IS_ASCII_KEY(key)
									 || !strchr("125znpo.", key))) {
		noteput("Show %s in window (press an arrow key, "
				"or one of .np125z) ", tag && *tag ? tag : "tag");
		getkey();
		check_abort();
	}
	say("");
	return key;
}

command pluck_tag() on cx_tab[',']
{		 /* read a function name at point & go there via tags */
	char tag[TAGLEN];

	tiled_only();
	init_tags();
	iter = 0;
	grab_tag_name(tag);
	if (has_arg)
		save_var tag_to_window_key = get_tag_window_direction(tag);
	modal_goto_tag(tag);
}

// Figure out what name to include in the prompt.
tags_look_for(char *tag)
{
	int i;

	strcpy(tag, _last_tag);
	if (buf_exist(prompt_orig_buf))
		save_var bufnum = prompt_orig_buf;
	i = get_mode_based_index("tag_prompter_%s");
	switch (name_type(i)) {		// Call mode-specific version.
		case NT_SUBR: case NT_COMMAND:
		case NT_AUTOSUBR: case NT_AUTOLOAD:
			(*(int (*)())i)(tag);
	}
}

command goto_tag() on cx_tab['.'] /* asks for a tag, then goes there */
{
	char tag[TAGLEN], pr[FNAMELEN], look_for[FNAMELEN];
	int to_win = has_arg && !iter;

	tiled_only();
	init_tags();
	iter = 0;
	*tag = 0;
	tags_look_for(look_for);
	if (!has_arg || to_win) {
		if (!*_last_tag)
			sprintf(pr, "Find tag: ");
		else if (tags_from_bsc)
			sprintf(pr, "Find tag [next %s]: ", look_for);
		else
			sprintf(pr, "Find tag [next %s of %d]: ",
					look_for, count_tags(look_for));
		if (*_last_tag)
			save_completion(_last_tag, tag_match);
		get_tag(tag, pr);
	}
	if (to_win)
		save_var tag_to_window_key = get_tag_window_direction(tag);
	modal_goto_tag(tag);
}

command next_tag() on reg_tab[NUMCTRL(GREYPLUS)]
{
	quiet_set_bookmark();
	next_prev_tag(iter);
	iter = 0;
}

command previous_tag() on reg_tab[NUMCTRL(GREYMINUS)]
{
	quiet_set_bookmark();
	next_prev_tag(-iter);
	iter = 0;
}


next_prev_tag(dir)
{
	init_tags();
	if (tags_from_bsc) {
		char fname[FNAMELEN];
		int linenum;

		if (!browser_find_match(fname, &linenum, dir))
			say("No %smatches.", dir ? "more " : "");
		else
			get_to_tag_match_line(fname, linenum);
		if (dir)
			say("Going to %s instance of %s",
				(dir > 0) ? "next" : "previous", _last_tag);
	} else
		must_go_tag("", dir);
}

position_to_tag_line()
{
	to_begin_line();
	if (point < window_start || point > window_end)
		window_start = prev_screen_line(show_tag_line);
}

// Maybe switch windows, then go to file with tag, maybe reading it,
// returning read error code or 0 if OK.
int tag_find_file()
{
	save_var synchronous_file_io = 1;
	switch (tag_to_window_key) {
		case '1':
		case '2':
		case '5':
		case 'z':
			has_arg = iter = 0;
			do_command(cx_tab[key]);
			break;
		case 'n':
		case 'p':
		case 'o':
			if (number_of_windows() < 2)
				split_window();
			window_number += tag_to_window_key == 'p' ? -1 : 1;
			break;
		case '.':		// Stay in this window.
			break;
		default:
			int dir = get_direction_from_key(tag_to_window_key);
			if (dir != -1)
				window_handle = window_from_direction(dir, window_handle);
			else
				locate_window("", _last_tag_file);
			break;
	}
	return std_find_it(_last_tag_file);
}

// Get to a tag match by line number (counting from 1).

get_to_tag_match_line(char *fname, int linenum)
{
	char tdir[FNAMELEN];

	strcpy(_last_tag_file, fname);
	strcpy(tdir, bsc_file);
	*get_tail(tdir, 0) = 0;
	absolute(_last_tag_file, tdir);
	if (tag_find_file())
		return;	// Error reading the file, already reported.
	go_line(linenum);
	window_start = prev_screen_line(show_tag_line);
}

// Return nonzero if we can't find the tag.
go_bsc_tag(char *s)
{
	char fname[FNAMELEN];
	int linenum;

	if (!*s) {		// Old way to go to next tag.
		next_prev_tag(1);
		return 0;
	}
	if (browser_query(bsc_file, s, 0) == -1)	// error
		return 1;
	strcpy(_last_tag, s);
	if (!browser_find_match(fname, &linenum, 0))
		return 1;
	get_to_tag_match_line(fname, linenum);
	return 0;
}

// Take result code from go_tag() and report error if any.
report_tag_error(int err)
{
	if (err == 1)
		error("%s is not in the tag list", _last_tag);
	else if (err == 2)
		error("%s is not in file %s", _last_tag, _last_tag_file);
}

/* Find this tag, complain if not found. */
/* If s is "", look for a duplicate tag.  In that case, n_fwd determines
   which one: 1==next, -1==previous, 0==same as before. */

must_go_tag(s, n_fwd)
char *s;
{
	report_tag_error(go_tag(s, n_fwd));
}

// Go to the appropriate file and line for the specified tag s.  Return
// 0 if ok, 1 if there's no tag by that name, 2 if there was a tag entry
// but we couldn't find any mention at that file and line.

// If s is "", look for a duplicate tag.  In that case, n_fwd determines
// which one: 1==next, -1==previous, 0==same as before.

// If using the default tag file doesn't succeed, try each library tag
// file (nyi) until one does.  

go_tag(char *s, int n_fwd)
{
	int res = go_tag_helper(s, n_fwd);
	if (!res)
		return 0;
	return res;
}


/* Return 0 if ok, 1 if not a tag, 2 if not in that file. */
/* If s is "", look for a duplicate tag.  In that case, n_fwd determines
   which one: 1==next, -1==previous, 0==same as before. */
/* This routine hands off to the BSC stuff if necessary, and also
   does retagging and searching again when appropriate. */

go_tag_helper(s, n_fwd)
char *s;
{
	char msg[FNAMELEN];
	short err;

	if (!*s && !*_last_tag)
		error("No current tag.");
	if (tags_from_bsc)
		return go_bsc_tag(s);
	err = do_go_tag(s, n_fwd);
	if (err != 2)
		return err;
	if (narrow_start || narrow_end) {
		say("Tag not found in narrowed buffer, removing narrowing.");
		narrow_start = narrow_end = 0;
		make_mode();
		err = do_go_tag(_last_tag, n_fwd);
		if (err != 2)
			return err;
	}
	if (!tags_format && find_tagger(0)) {
		sprintf(msg, "Tag %s has moved, retag file %s? ",
				_last_tag, _last_tag_file);
		if (tag_ask_before_retagging)
			if (!ask_yn("Retag file", msg, "Yes", "No", 1))
				return 0;
		do_retag_file(_last_tag_file);		// Tag has moved, so retag.
		do_save_tags();
	}
	return do_go_tag(_last_tag, n_fwd);
}

// Search for nth occurrence of the tag pat, wrapping if necessary.
// N may be negative (or 0 to repeat the last one).
// Return 1 if we found one, 0 if there are none.

tag_search_by_offset(char *pat, int n_fwd)
{
	int dir = RE_FORWARD;

	if (n_fwd < 0){
		dir = RE_REVERSE;
		n_fwd = -n_fwd;
		to_begin_line();
	} else if (n_fwd > 0) {
		to_end_line();
	} else {			// n_fwd == 0: return current match again
		to_begin_line();
		n_fwd = 1;
	}
	while (n_fwd-- > 0)
		if (!re_search(dir, pat)) {
			point = (dir == RE_FORWARD) ? 0 : size();
			if (!re_search(dir, pat))	// Wrap and repeat.
				return 0;
		}
	return 1;
}

// How many tags have this name?
count_tags(char *s)
{
	char pat[FNAMELEN], *p, tag[FNAMELEN], fast_pat[FNAMELEN];
	int cnt = 0;

	strcpy(tag, s);
	if (p = index(tag, '\t'))
		*p = 0;		/* remove file name from tag spec */
	sprintf(pat, "^([^\t\n]+(::|'))?%q\t", tag);
	sprintf(fast_pat, "%s\t", tag);
	save_var bufnum = _tags_buffer;
	save_var case_fold = !tag_case_sensitive;
	save_var point = 0;
	while (search(1, fast_pat))
		if (parse_string(-1, pat))
			cnt++;
	return cnt;
}

// Write a pattern to pat that will match if we're probably at the
// specifed tag's definition.  When tag-with-text is zero, we can't
// depend on the line-based search when the tag includes a class but the
// match doesn't.
tag_make_loose_pattern(char *tag, char *pat)
{
	*pat++ = '(';
	for (; *tag; tag++) {
		sprintf(pat, "%q", tag);
		if (*pat == '%')
			pat++;
		pat++;
		if (tag[1] &&  !(isalnum(tag[0]) || tag[0] == '_')
			  && (isalnum(tag[1]) || tag[1] == '_')) {
			strcpy(pat, ")?(");
			pat += 3;
		}
	}
	strcpy(pat, ")[^a-zA-Z0-9_]");
}

/* Go to tag s, return 0=ok, 1=no such tag, 2=tag moved. */
/* If s is "", look for a duplicate tag.  In that case, n_fwd determines
   which one: 1==next, -1==previous, 0==same as before. */

do_go_tag(s, n_fwd)
char *s;
{
	char pat[2048], tdir[FNAMELEN], *file, *p, *q;
	int pos, oldbuf = bufnum, i;

	i = get_mode_based_index("tag_to_next_%s");
	bufnum = _tags_buffer;
	save_var case_fold = !tag_case_sensitive;
    save_var narrow_start, narrow_end;
	*pat = 0;
	if (*s) {
		if (p = index(s, '\t')) {
			strcpy(_last_tag_file, p + 1);
			if (strchr(p + 1, '\t'))	// Have full info already.
				sprintf(pat, "^%q", s);
			*p = 0;		/* remove file name from s */
			if (p = index(_last_tag_file, '\t'))
				*p = 0;
		} else
			*_last_tag_file = 0;
		strcpy(_last_tag, s);
		if (!*_last_tag)	/* no last tag */
			return 0;
		if (!*pat)
			if (*_last_tag_file && p && !strchr(p + 1, '\t') && !tags_format)
				sprintf(pat, "^%q\t%q\t%q", _last_tag, _last_tag_file, p + 1);
			else if (*_last_tag_file)
				sprintf(pat, "^%q\t%q\t", _last_tag, _last_tag_file);
			else
				sprintf(pat, "^%q\t", _last_tag);
		point = 0;
		if (!re_search(1, pat)) {		/* look for 1st */
            if (!tags_format)
				return 1;
		}
	} else {		/* find next matching tag */
		pos = point;
		sprintf(pat, "^%q\t", _last_tag);
		switch (name_type(i)) {		// Call mode-specific version.
			case NT_SUBR: case NT_COMMAND:
			case NT_AUTOSUBR: case NT_AUTOLOAD:
				if ((*(int (*)())i)(pat, n_fwd))
					return 1;
				break;
			default:
				if (!tag_search_by_offset(pat, n_fwd))
					return 1;
		}
	}
	grab(give_begin_line(), give_end_line(), pat);	/* tag, file, pos */
	file = index(pat, '\t');
	*file++ = 0;

	p = index(file, '\t');		/* find start of pos */
	pos = strtoi(p + 1, 10);
	*p = 0;
	q = index(p + 1, '\t');		/* q has line's text, if recorded */
	strcpy(_last_tag_file, file);
	put_directory(tdir);
	bufnum = oldbuf;
	absolute(_last_tag_file, tdir);
	tag_find_file();
	if (q)
		sprintf(tdir, "%q[^a-zA-Z0-9_]", _last_tag);
	else			// Might not include class info.
		tag_make_loose_pattern(_last_tag, tdir);
	point = pos;
	if (!parse_string(1, tdir, NULL))
		if (!q || !prox_line_search(q + 1))	/* search by text */
			return 2;
	to_begin_line();
	window_start = prev_screen_line(show_tag_line);
	return 0;
}

prox_line_search(p)	/* search for line with p on it closest to point */
char *p;		/* and go there, return 1, else 0 */
{
	int orig = point, fwd, bwd, fwdok, bwdok;

	fwdok = line_search(1, p);
	fwd = point;
	point = orig;
	bwdok = line_search(-1, p);
	bwd = point;
	point = orig;
	if (!fwdok && !bwdok)	/* neither one found it, fail */
		return 0;
	if (fwdok && bwdok)	/* if both found it, use closest */
		fwdok = (fwd - point) <= (point - bwd);
	point = fwdok ? fwd : bwd;
	return 1;
}

command select_tag_file() on cx_tab[ALT(',')]
{				/* switch to a particular tags file */
	char file[FNAMELEN], *def = initial_tag_file, defbuf[FNAMELEN];
	int old = bufnum;

	do_save_tags();
	if (tags_from_bsc)
		def = bsc_file;
	else if (tag_options & 2) {	// Scan tree for tag file suggestion.
		char dirbuf[FNAMELEN], *dirptr = 0;
		if ((tag_options & 4) && get_buffer_directory(dirbuf))
			dirptr = dirbuf;	// Scan from current file.
		if (!look_up_tree(defbuf, initial_tag_file, dirptr, NULL))
			strcpy(defbuf, initial_tag_file);
		def = defbuf;
	} else if (exist("-tags")) {
		bufname = "-tags";
		def = filename;
		bufnum = old;
	}
	save_var only_file_extensions = "|.tag|.bsc|";
	get_file(file, "Tag or browser file", def);
	load_tags(file);
	say("Tags loaded from %s", file);
}

command clear_tags()	/* erase all tags */
{
	init_tags();
	if (tags_from_bsc)
		error("Delete the .BSC file %s to remove all tags.", bsc_file);
	buf_zap(_tags_buffer);
	mark_tags(1);
	do_save_tags();
}

// List the names of all files we've tagged in buffer b.  The file
// names will be relative to the tag file's directory.
get_tagged_file_names(int b)
{
	int p, above = 0;
	char this[FNAMELEN], last[FNAMELEN];

	buf_zap(b);
	save_var bufnum = _tags_buffer;
	point = 0;
	*last = 0;
	while (nl_forward()) {
		if (!search(1, "\t"))
			break;
		p = point;
		if (!search(1, "\t"))
			break;
		grab(p, point - 1, this);
		if (!strcmp(this, last))
			continue;
		strcpy(last, this);
		buf_printf(b, "%s\n", this);
		if (point > above) {
			noteput("Extracting file names: %2d%%...",
					muldiv(point, 100, size()));
			above += 50000;
			check_abort();
		}
	}
	bufnum = b;
	do_sort_region(0, size(), 0, 0);
	do_uniq(1, 1, 0);
	point = 0;
	say("");
}

command untag_files()
{
	int t, p, count;
	char this[FNAMELEN], resp[FNAMELEN], msg[FNAMELEN];

	init_tags();
	if (tags_from_bsc)
		error("Can't modify a .BSC file.");
	t = zap(TAGGED_FILES_BUF);
	get_tagged_file_names(t);

	save_var bufnum, window_bufnum;
	to_buffer_num(t);
	p = tmp_buf();
	buf_xfer(p, 0, size());
	point = 0;
	stuff("# Press Ctrl-X Ctrl-Z to untag all files not named below.\n");
	stuff("# Only files remaining in this list will keep their tags.\n\n");
	recursive_edit();
	if (!buf_exist(t) || !buf_exist(p)) {
		say("No files untagged.");
		return;
	}
	to_buffer_num(t);
	simple_re_replace(1, "^(# .*)?\n", "");
	case_fold = sort_case_fold = 1;
	do_sort_region(0, size(), 0, 0);
	bufnum = p;
	case_fold = sort_case_fold = 1;
	do_sort_region(0, size(), 0, 0);
	restore_vars();
	do_compare_sorted(t, p, NULL, "-untag-these-files", NULL);
	buf_delete(p);

	save_var bufname = "-untag-these-files";
	count = lines_between(0, size(), 0);
	restore_vars();
	sprintf(msg, (count == 1) ? " Untag this %d file? "
							  : " Untag these %d files? ", count);
	if (!(is_gui && want_gui_prompts))
		say("%s", msg + 1);
	else
		say("");
	if (count) {
		save_var _view_title = msg;
		save_var paging_retains_view = 1;
		view_buffer("-untag-these-files", 0);
		if (is_gui && want_gui_prompts && key == abort_key)
			count = 0;		// User clicked Cancel.
		else if (is_gui && want_gui_prompts) {	// Already prompted so don't
			if (toupper(key) != 'Y' && !IS_ENTER_KEY(key) // ask again.
				&& toupper(key) != 'O')
				count = 0;
			ungot_key = -1;
		} else {
			get_string(resp, msg + 1);
			if (toupper(*resp) != 'Y')
				count = 0;
		}
		user_abort = 0;
	}
	if (count) {
		save_var bufname = "-untag-these-files";
		for (point = 0; point < size(); nl_forward()) {
			grab(give_begin_line(), give_end_line(), this);
			if (size() > 1000)
				noteput("Deleting tags: %2d%%...", muldiv(point, 100, size()));
			do_delete_tags(this);
		}
		do_save_tags();
		say("%d file%s untagged.", count, count == 1 ? "" : "s");
	} else
		say("No files untagged.");

	restore_vars();
	delete_buffer("-untag-these-files");
	buf_delete(t);
}

command retag_files()
{
	char file[FNAMELEN], tdir[FNAMELEN];
	int t, aborted = 0;

	init_tags();
	save_var synchronous_file_io = 1;
	if (tags_from_bsc)
		error("Can't modify a .BSC file.");
	say("");
	if (!ask_yn("Retag files", "Retag all files? ", "Yes", "No", 1))
		return;
	t = zap(TAGGED_FILES_BUF);
	get_tagged_file_names(t);
	to_another_buffer("-tag-errors");
	delete_buffer("-tag-errors");
	save_var tag_batch_mode = 1;
	put_tag_directory(tdir);
	while (grab_line(t, file) > 0 && !user_abort) {
		absolute(file, tdir);
		if (do_retag_file(file))
			user_abort = aborted = 1;
	}
	if (!aborted)
		do_save_tags();
	if (exist("-tag-errors")) {
		to_buffer("-tag-errors");
		bprintf("\nOld tags for the above files were removed.\n");
		point = 0;
		discardable_buffer = 1;
	}
	if (!aborted)
		say("All files retagged.");
}

command tag_files() on cx_tab[ALT('.')]
{
	char pat[FNAMELEN], *s, file[FNAMELEN], def[FNAMELEN];
	int aborted = 0;

	init_tags();
	save_var synchronous_file_io = 1;
	if (tags_from_bsc)
		error("Must tag files using the BSCMAKE program.");
	if (has_arg)
		save_var tag_declarations = 3;
	save_var recall_id = "f-match tag-files";
	get_last_recall((char *(*)()) f_match, def);
	get_file(pat, "Add/update tags for files matching", def);
	iter = 0;
	if (!(s = do_file_match(pat, STARTMATCH | EXACTONLY | FM_NO_DIRS)))
		error("No matches");
	to_another_buffer("-tag-errors");
	delete_buffer("-tag-errors");
	save_var tag_batch_mode = 1;
	if (tag_options & 1)
		clear_tags();
	for (; s; s = do_file_match(pat, EXACTONLY | FM_NO_DIRS)) {
		strcpy(file, s);
		absolute(file);
		if (aborted = do_retag_file(file))
			break;
	}
	if (!aborted) {
		do_save_tags();
		say("%s tagged.", pat);
	}
	if (exist("-tag-errors")) {
		to_buffer("-tag-errors");
		discardable_buffer = 1;
	}
}

// Locate the appropriate tagging function for this buffer, if any.
// Return 1 if we found one, 0 otherwise.  If run, also run the
// function.

int find_tagger(int run)
{
	char subr[FNAMELEN], mode[FNAMELEN], *s;

	if (tags_from_bsc)
		return 0;
	*subr = 0;
	s = get_extension(filename);
	if (*s == '.')
		sprintf(subr, "tag-suffix-%s", s + 1);
	clean_mode(mode);
	if (!*subr || !is_runnable(subr))
		if (*mode)
			sprintf(subr, "tag-mode-%s", mode);
	if (!*subr || !is_runnable(subr))
		guess_mode_without_extension(subr, "tag-mode-%s");
	if (!*subr || !is_runnable(subr))
		strcpy(subr, "tag-suffix-default");
	time_begin(&tag_show_msg, 200);
	return run ? try_calling(subr) : is_runnable(subr);
}

tag_a_file(char *s)
{
	jmp_buf this_level;
	int ok = 1, err = 0, orig = bufnum, temp = 0, oldmod, aborted = 0;

	bufnum = _tags_buffer;	/* go to right place in tags buffer */
	point = 0;
	nl_forward();
	oldmod = modified;	/* want to see if we add new tags */
	modified = 0;
	bufnum = orig;
	save_var top_level = &this_level;
	if (!setjmp(top_level)) {
		if (!look_file(s)) {
			if (!temp)
				temp = tmp_buf();
			bufnum = temp;
			save_var readonly_warning = 0;
			err = read_file(s, get_default_translation_type(s));
			filename = s;
			call_mode(filename);
			make_mode();
			check_filevars();
		}
		if (!err) {
			save_var point = 0;
			ok = find_tagger(1);
		}
	} else
		aborted = 1;
	bufnum = _tags_buffer;
	if (modified)		/* we added new tags, mark unsorted */
		mark_tags(0);
	else
		modified = oldmod;
	bufnum = orig;
	if (temp)
		buf_delete(temp);
	restore_vars();
	if (err == EREADABORT)
		aborted = 1, err = 0;
	if (err && tag_batch_mode)
		buffer_printf("-tag-errors", "Couldn't read the file %s.\n", s);
	else if (err)		/* couldn't read file */
		quick_abort();	/* already showed error msg */
	else if (!ok && tag_batch_mode)
		buffer_printf("-tag-errors", "Don't know how to tag the file %s.\n", s);
	else if (!ok)
		error("Don't know how to tag the file %s", s);
	return aborted || user_abort;
}

tag_mode_asm()	/* tag all labels or procs in the file */
{
	char func[TAGLEN];
	int start;

	save_var case_fold = 1, point = 0;
	while (re_search(1, "^[ \t]*([a-z0-9@$_]+)[ \t]*:")) {
		grab(start = find_group(1, 1), find_group(1, 0), func);
		add_tag(func, start);
	}
	point = 0;
	while (re_search(1, "^[ \t]*([a-z0-9@$_]+)[ \t]+proc[ \t\n;]+")) {
		grab(start = find_group(1, 1), find_group(1, 0), func);
		add_tag(func, start);
	}
}

/* Tag all C functions in this file. Must also handle C++, JavaScript,
 * C#, Java, etc. based on c_extra_keywords. */
tag_mode_c()
{
	int end;

	save_var case_fold = 0;	// don't confuse STRUCT with struct
	tag_modifier_kwd = 0;
	*c_tagging_class = 0;	// initially, not inside a class block
	if (old_tagging_class) {
		free(old_tagging_class);
		old_tagging_class = 0;
	}
	while (re_search(1, "([{};<]|/<*|/>|[a-zA-Z0-9_]+)")) {
		end = point;		/* find braces, ;, <, comment open, or ident */
		point = matchstart;
		switch (curchar()) {
			case '{':	skip_c_braces();
						break;
			case '<':	skip_cpp_template();
						break;
			case '}':	tag_remove_class();
						point++;
						break;
			case ';':	tag_modifier_kwd = 0;	// assume next var
						point++;		// is being defined
						break;
			case '/':	if (character(point + 1) == '*') {
							point += 2;
							search(1, "*/");
						} else
							nl_forward();	/* c++ comment */
						break;
			default:	if (!good_c_tag())
							point = end;
						break;
		}
	}
}

// We're just before an instance of the start character.  Go to the
// matching end character and return 0.  If we can't find it, return
// 1 if too many start characters or 2 if too many end characters,
// and restore point to start if restore!=0.  Attempt to handle
// preprocessor conditionals by skipping over #else clauses when
// appropriate.

skip_c_delims(int start, int end, int restore)
{
	int level = 0, c, orig = point, iflevel = 0, skipping = 0;
	char buf[4], pat[80];

	strcpy(buf, "X|\\");
	sprintf(pat, "[%c%c\"']|/<*|/>|(" PREPROC_LINE ")", end, start);
	while (re_search(1, pat)) {
		buf[0] = c = character(point - 1);
		if (c == '\"' || c == '\'') {
			if ((c_extra_keywords & CPP_KEYWORDS) && c == '\''
				  && isalnum(curchar()) && isalnum(character(point - 2))) {
				// Possible C++14 use of ' as digit separator. Check
				// both decimal/binary/octal and hex cases.
				if (parse_string(-1, "</word>(0b)?[0-9']+'") && parse_string(1, "[0-9']+")
				   || parse_string(-1, "</word>0x[0-9a-f']+'") && parse_string(1, "[0-9a-f']+")) {
					point = matchend;
					continue;
				}
			}
			if ((c_extra_keywords & CPP_KEYWORDS) && parse_string(-1, "R\"")
				  && move_past_cpp_raw_literal(-1, 0))
				;	// We've just moved past C++ raw string literal.
			else if ((c_extra_keywords & CSHARP_KEYWORDS)
					 && character(point - 2) == '@')
				re_search(1, "([^\"]|\"\")*\"![^\"]");
			else
				while (re_search(1, buf)
					   && character(point - 1) == '\\')
					point++;
		} else if (c == '*')				/* Found block comment */
			search(1, "*/");
		else if (c == '/')		/* Found c++ one line comment */
			nl_forward();
		else if (c == start) {			/* Found char like (. */
			if (!skipping)
				level++;
		} else if (c == end) {			/* Found char like ). */
			if (!skipping) {
				if (!--level)			/* Found proper end. */
					return 0;
				else if (level < 0) {	/* Found extra closing delim, give up. */
					if (restore)
						point = orig;
					return 2;
				}
			}
		} else if (parse_string(-1, PREPROC_ELSE)) {
			skipping |= (1 << iflevel);	// Found #else or #elif line, skip through #endif.
		} else if (parse_string(-1, PREPROC_ENDIF)) {
			skipping &= ~(1 << iflevel);	// found #endif line, maybe stop skipping.
			if (iflevel > 0)			/* Ignore extra #endif's. */
				iflevel--;
		} else if (parse_string(-1, PREPROC_IF)) {
			iflevel++;				// found #if/#ifdef line
		}
	}
	if (restore)						/* Couldn't find match, give up. */
		point = orig;
	return 1;
}

skip_cpp_template()		/* skip over matched <> pairs */
{
	int start = point;		/* Don't look too far for them. */

	save_var narrow_end = size() - (point + 200);
	if (skip_c_delims('<', '>', 1))		// unmatched character
		point = start + 1;	// Skip over that <, we can't match it.
}

skip_c_braces()		/* skip over c function definition */
{
	if (parse_string(-1, "<dquote>[A-Za-z0-9_]*<dquote>[ \t\n]*")
		  && !at_c_comment(point, 0) && !parse_string(-1, "color_scheme.*")) {
		point++;	// if C++'s extern "C" { or similar, don't skip
		return;
	}
	skip_c_delims('{', '}', 0);
}

good_c_tag()	// tag name at point if a valid c tag
{		// return 0 to go back to orig point before searching again
	char func[TAGLEN], tmp[TAGLEN];
	int start = point, i = 1, skip = 0;

	if (!parse_string(1, "[a-zA-Z0-9_]+", func))
		return 0;
	else if (!strcmp(func, "struct") || !strcmp(func, "union"))
		return tag_c_struct(0);		// maybe struct tag
	else if (!strcmp(func, "class") || !strcmp(func, "__interface"))
		return tag_c_struct(1);		// or similar class or managed c++ def
	else if (!strcmp(func, "interface") || !strcmp(func, "package"))
		return tag_c_struct(1);		// for java or managed c++
	else if ((c_extra_keywords & OBJC_KEYWORDS) && character(start - 1) == '@'
			 && word_in_list(func, "|interface|implementation|protocol|", 0))
		return tag_objc_block();
	else if (!strcmp(func, "enum"))
		return tag_c_struct(2);		// or similar enum def
	else if (!strcmp(func, "namespace"))
		return tag_c_struct(3);		// or namespace def
	else if (!strcmp(func, "operator") && parse_string(1, CPP_OPLIST)) {
		grab(matchstart, matchend, func + strlen(func));
		point = matchend;
	} else if (!strcmp(func, "extern")) {
		tag_modifier_kwd |= 1;	// next vars are just declarations
		return 1;
	} else if (!strcmp(func, "typedef")) {
		tag_modifier_kwd |= 2;	// next vars are typedefs
		return 1;
	} else if ((c_extra_keywords & CPP_KEYWORDS) && !strcmp(func, "using")) {
		tag_modifier_kwd |= 4;	// just a declaration 
		return 1;
	} else if (is_java_buffer() && (!strcmp(func, "import")
									|| !strcmp(func, "package"))) {
		tag_modifier_kwd |= 4;
		return 1;
	} else if (!strcmp(func, "__declspec")) {
		return 1;
	} else if (!strcmp(func, "__attribute__") && parse_string(1, "<wspace>*%(%(")
			   && !skip_c_delims('(', ')', 1)) {
			return 1;
	} else if (parse_string(-1, "^[ \t]*#[^#\n]*"))
		return tag_c_define(func);	// do preproc line
	c_fwd_space();
	sprintf(tmp, "(%s)</word>", c_prototype_macro_names);
	if (parse_string(1, tmp)) {
		point = matchend;	// skip over arg-hiding macro
		c_fwd_space();
	}
	if (curchar() == '<')
		tag_check_template(func);
	if (curchar() == '(') {
		if (matches_at(start, -1, "~[ \t]*")) {
			point = matchend;
			parse_string(1, "~[ \t]*[a-zA-Z0-9_]+", func);
		}
		i = tag_c_func(1);
		if (!(tag_which_items & 1)
			  || (c_extra_keywords & JS_KEYWORDS) && !strcmp(func, "function"))
			skip = 1;
	} else if (curchar() == ':' && parse_string(1, "::[ \t\n]*"
			"(_?(near|far)[ \t\n]+)?([a-zA-Z0-9_~]+|" CPP_OPSPEC ") *%(")) {
		strcat(func, "::");	// make raw class::func w/o spaces
		grab(find_group(3, 1), find_group(3, 0), func + strlen(func));
		i = tag_c_func(1);
		if (!(tag_which_items & 1))
			skip = 1;
	} else if (c_extra_keywords & JS_KEYWORDS) {
		i = tag_jscript_ident(func, start);
	} else if (tag_c_variable() && ((tag_declarations & 2) || !(tag_modifier_kwd & 1)
									|| (tag_modifier_kwd & 2))) {
		i = 2;
		if (!(tag_which_items & 2))
			skip = 1;
	}
	if ((i == 2 || i == 3) && !isdigit(*func) && !(tag_modifier_kwd & 4)) {
		if (*c_tagging_class && !index(func, ':')) {
			strcpy(tmp, func);	// put class:: before func name
			sprintf(func, "%s::%s", c_tagging_class, tmp);
		}
		if (!skip)
			add_tag(func, start);
	}
	return i;
}

// If we have template syntax, add it.
tag_check_template(char *func)
{
	int pos = point;
	if (!find_template_end() || !parse_string(1, "[ \t\n]*(::<ident>+)?%("))
		point = pos;
	else
		grab(pos, matchstart, func + strlen(func));
}

// Skip over a thing that might appear before the function body, and
// return 1 if we did.

tag_c_skip_before_body(int show)
{
	if (parse_string(1, "#.*(\\\n.*)*")) {
		point = matchend;
		return 1;
	}
	if (parse_string(1, "(try|const|volatile|NOTHROW|override|"
					 "final|import|module)</ident>")) {
		point = matchend;
		return 1;
	}
	if (parse_string(1, "(AVAILABLE|DEPRECATED_IN)_MAC_OS_X_VERSION_<word>+")) {
		point = matchend;
		return 1;
	}
	if (parse_string(1, "throws[ \t]")) {	// for java
		point = matchend;
		skip_identifier_list();
		return 1;
	}
	if (parse_string(1, "(throw|noexcept|__attribute__|__DARWIN_ALIAS)[ \t]*!%(")) {
		point = matchend;	// throw for C++, attribute for gcc
		if (skip_c_delims('(', ')', 1)) {
			if (show)
				buffer_printf("-tag-errors", "Couldn't parse %s at line %d\n",
							  filename, lines_between(0, point, 0) + 1);
			return 0;
		}
		return 1;
	}
	if (parse_string(1, "(noexcept)</word>")) {
		point = matchend;
		return 1;
	}
	if ((c_extra_keywords & JS_KEYWORDS) && parse_string(1, ":<wspace>*<alnum>+")) {
		point = matchend;
		return 1;
	}
	return 0;
}

// We're positioned between an identifier and a left paren (and not
// inside a function body).  Return 2 (yes) if it looks like a function
// definition (or optionally a declaration), and we should therefore tag
// it.  Move just before the function body or semicolon if so. Return 3
// if it's a C++ definition using =0 instead of a body, or a JavaScript
// lambda function using =>.

// Otherwise, return 0 if we should go back to our starting point and
// look for more identifiers from there, or 1 if we should stay where
// this function moved to, and continue searching from there.

// We also use this function to help recognize function definitions in
// other contexts than tagging.

tag_c_func(show)
{
	int ret = 2;

	if (skip_c_delims('(', ')', 1)) {
		if (show && !in_c_comment(0))
			buffer_printf("-tag-errors", "Warning: couldn't parse %s at line %d\n",
						  filename, lines_between(0, point, 0) + 1);
		return 0;	/* after identifier, skip over arg list */
	}
	c_fwd_space();
	while (tag_c_skip_before_body(show))
		c_fwd_space();
	if (parse_string(1, "on[ \t]+<word>+[ \t]*<[>")) {	// eel command
		point = matchstart - 1;
		c_skip_level('[', ']');
		search(1, "\n{");
		point--;
		ret = 2;
	} else if (parse_string(1, "[A-Za-z{]")) // func definition
		ret = skip_param_decls();
	else if (curchar() == ':')		// C++ initializer
		ret = skip_cpp_init();
	else if (curchar() == '=')		// C++ function using =0.
		return 3;
	else if (!parse_string(1, "[,;]"))	// not even a declaration
		ret = (point >= size());	// Skip parens if it looks like a macro.
	else if (!(tag_declarations & 1)	// a declaration, don't tag it?
			|| (c_extra_keywords & JS_KEYWORDS))
		ret = 1;
	if (ret)
		tag_modifier_kwd = 0;
	return ret;
}

skip_cpp_init()		// found base class initializer, skip past it
{
	if (!search(1, "{"))
		return 0;
	point--;
	return 2;
}

// We're just past the ) of an apparent func definition.
// There should be a { now, with perhaps some old-style param declarations
// before it.  Skip past the { and return 2.  If this doesn't look like
// a func definition after all, go back to where started and return 1.

skip_param_decls()
{
	int orig = point, last;

	save_var narrow_end = size() - (point + 5000);
	c_fwd_space();
	if (curchar() == '{')		// No params.
		return 2;
	do {
		last = point;
		if (curchar() == '{' && semi_before(point))
			return 2;	// had ; just before {, good decls
		point += parse_string(1, "[][a-zA-Z0-9_@;,()*]*");
		c_fwd_space();
	} while (last != point);
	point = orig;		// these weren't param decls, go back
	return 1;
}

semi_before(p)	// is there a semicolon just before p (ignoring comments)?
{
	save_var point = p;
	return !cind_try_space() && character(point - 1) == ';';
}

// Remember we're now in this class (or namespace or similar).
tag_add_class(char *class, char *sep)
{
	if (!old_tagging_class)
		old_tagging_class = (char **) malloc(sizeof(*old_tagging_class) * 20);
	char **p = old_tagging_class;
	while (*p)	// Add ptr to end of class list to end of this list.
		p++;
	*p = c_tagging_class + strlen(c_tagging_class);
	if (*c_tagging_class)
		strcat(c_tagging_class, sep);
	strcat(c_tagging_class, class);
}

// Remember we're now in this class (or namespace or similar).
tag_c_add_class(char *class)
{
	tag_add_class(class, "::");
}

// Leave a class's scope.
tag_remove_class()
{
	char **p;
	if (old_tagging_class && *old_tagging_class) {
		for (p = old_tagging_class; *p; p++)
			;
		**--p = 0;		// Chop off last class.
		*p = 0;			// Forget chop-off point.
	} else
		*c_tagging_class = 0;
}

add_class_tag(char *class, char *sep, char *func, int pos)
{
	char fulltag[FNAMELEN * 2];
	if (class && *class && *sep)
		sprintf(fulltag, "%s%s%s", class, sep, func);
	else
		strcpy(fulltag, func);
	add_tag(fulltag, pos);
}

// Move past a series of comma-separated identifiers with
// whitespace/comments between.  Return how many.
int skip_identifier_list()
{
	int cnt = 0;
	for (;; point++) {
		c_fwd_space();
		if (!parse_string(1, "[a-zA-Z0-9_.]+"))
			break;
		point = matchend;
		if (curchar() == '<')
			c_skip_level('<', '>');
		cnt++;
		c_fwd_space();
		if (curchar() != ',')
			break;
	}
	return cnt;
}

// Return a regex pattern that matches an identifier, including :: or .
// as apppropriate.
char *tag_c_ident_pattern()
{
	return (c_extra_keywords & (CSHARP_KEYWORDS | JS_KEYWORDS | JAVA_KEYWORDS)) ?
			"<word|.>+" : "<word>+(::<word>+)?";
}

// Tag the following structure/class/enum if we're defining a tag for it.
// Flavor: 0=struct/union, 1=class/interface, 2=enum, 3=namespace.
tag_c_struct(int flavor)
{
	char func[TAGLEN], fullname[TAGLEN], *s;
	int start, i;

	if (flavor == 1 && (c_extra_keywords & OBJC_KEYWORDS)
		  && parse_string(-1, "@class"))
		return tag_objc_class_list();
	c_fwd_space();
	start = i = point;
	if (parse_string(1, "__declspec</word>")) {
		point = matchend;
		c_fwd_space();
		if (curchar() == '(')
			c_skip_level('(', ')');
		c_fwd_space();
	}

	if (!parse_string(1, tag_c_ident_pattern(), func)) {
		if (flavor != 1)
			*func = 0;	// all but class don't require a name
		else
			return 1;
	}
	if (flavor == 0 && (c_extra_keywords & CPP_KEYWORDS))
		flavor = 1;	// In C++, treat structs just like classes.
	c_fwd_space();	// skip compiler-specific junk betw "class" & name
	if (parse_string(1, "(final)</word>")) {
		point = matchend;
		c_fwd_space();
	}
	if (flavor == 1 && !is_java_buffer() && !(c_extra_keywords & JS_KEYWORDS)
		  && parse_string(1, tag_c_ident_pattern())) {
		i = point;
		parse_string(1, tag_c_ident_pattern(), func);
		c_fwd_space();
	}
	while (strchr(func, '.')) {
		strcpy(fullname, func);
		s = strchr(fullname, '.');
		sprintf(s, "::%s", func + 1 + (s - fullname));
		strcpy(func, fullname);
	}
	if (strchr(";[,", curchar()) && i > start) {
		point = i;	// A plain definition.
		return 1;
	}
	if (curchar() != '{' && (flavor != 1 || !parse_other_class()))
		return 1;	// classes can have other stuff before {
	if (flavor == 1 || flavor == 3 || flavor == 0) {
		tag_c_add_class(func);
		i = ++point;	// skip over { so we do stuff within
		c_fwd_space();
		if (parse_string(1, "#(else|elif)"))
			if (locate_by_ifdef(1) != 2)
				point = i;
	} else if (flavor == 2) {
		tag_c_enum();
		strcpy(fullname, func);
		if (*c_tagging_class)
			sprintf(fullname, "%s::%s", c_tagging_class, func);
		if (*func && (tag_which_items & 32))
			add_tag(fullname, start);
		return 1;
	}
	if (*func && (tag_which_items & 16))
		add_tag(c_tagging_class, start);
	return 1;
}

/* We've parsed "class foo" but there's no { character right after.
 * See if there's some inheritance information we can skip past to find
 * the { character.
 * Return nonzero if we found ": foo {" or similar (C++).
 * Return nonzero if we found Java's "extends foo implements bar {" or similar.
 * In either case, move before the final { character.  Otherwise return
 * zero.
 */
parse_other_class()
{
	c_fwd_space();
	if (is_java_buffer() && curchar() == '<')
		c_skip_level('<', '>');
	c_fwd_space();
	if (curchar() == ':') {
		point++;
		for (;;) {
			c_fwd_space();
			if (parse_string(1, tag_c_ident_pattern())
				  || parse_string(1, "[,:]"))
				point = matchend;
			else if (!find_template_end())
				break;
		}
		c_fwd_space();
	} else if (c_extra_keywords & (JAVA_KEYWORDS | JS_KEYWORDS)) {
		while (strchr("ei", curchar())
			   && parse_string(1, "(extends|implements)</word>")) {
			point = matchend;
			skip_identifier_list();
		}
	}
	return curchar() == '{';
}

tag_c_enum()	// just before { of enum list; tag them all
{
	char name[TAGLEN];

	do {
		point++;
		c_fwd_space();
		if (!parse_string(1, "[a-zA-Z0-9_]+", name))
			return;
		if (tag_which_items & 8)
			add_tag(name, matchstart);
		c_fwd_space();
		if (curchar() == '=')
			skip_initializer();
	}  while (curchar() == ',');
	point += re_search(1, "<wspace>*}");
}

tag_c_define(keyw)	// found a preproc line, maybe enter #define
char *keyw;		// either way, skip past line
{
	char func[TAGLEN], pat[FNAMELEN];
	int start = point;

	sprintf(pat, "[ \t]+%s</word>[ \t]*($|/[/*])", tag_c_preprocessor_skip_pat);
	if (!strcmp(keyw, "if") && parse_string(1, pat))
		if (!locate_by_ifdef(1))
			point = start;
	if (!strcmp(keyw, "asm"))	// Found #asm, look for matching #endasm.
		return re_search(1, "^[ \t]*#[^#\n]*endasm");
	if (!strcmp(keyw, "define")) {
		point += parse_string(1, "[ \t]*");
		start = point;
		if (parse_string(1, "[a-zA-Z0-9_]+", func) && !isdigit(*func)
			  && (tag_which_items & 4))
			add_tag(func, start);
	}
	if (parse_string(1, ".*/<*>"))
		search(1, "*/");
	do {
		if (!nl_forward())
			break;
	} while (character(point - 2) == '\\');
	return 1;
}

tag_c_variable()	// is this a var definition?
{
	char pat[FNAMELEN * 2];
	sprintf(pat,  "[ \t\n]*<)>+[ \t\n]*!(%s)?<(>", c_prototype_macro_names);
	if (parse_string(1, "([) \t\n]|<[>[^]\n]*<]>|//.*)*[,;=]")) {
		point = matchend - 1;	// ordinary variable definition
		if (character(point) == '=')
			point++, skip_initializer();
	} else if (parse_string(1, pat))
		re_search(1, "![;{]");	// func ptr definition
	else
		return 0;
	return 1;
}

skip_initializer()		// move past c initializer
{
	int level = 0;

	c_fwd_space();
	if (curchar() == '{')	// don't bother skipping over {}'s,
		return;		// caller does that anyway
	while (re_search(1, "[,;(){}'\"]"))
		switch (character(point - 1)) {
			case '"':
			case '\'':	if ((c_extra_keywords & CPP_KEYWORDS)
							  && matches_at(point - 2, 1, "[0-9a-f]'[0-9a-f]"))
							continue;		// A number like 123'456'789.
						point--;
						c_skip_item(curchar(), RE_FORWARD);
						break;
			case ',':
			case ';':	if (level < 0)
							break;
						point--;
						return;
			case '(':	level--; break;
			case ')':	level++; break;
			case '{':
			case '}':	point--;
						return;
		}
}

int tag_jscript_ident(char *func, int start)
{
	char tmp[FNAMELEN];
	int i = 1;
	c_fwd_space();
	while (curchar() == '.') {
		point++;
		if (parse_string(1, "[a-zA-Z0-9_]+", tmp))
			sprintf(func + strlen(func), "::%s", tmp);
		c_fwd_space();
	}
	if (curchar() == '=') {
		point++;
		c_fwd_space();
		if (parse_string(1, "async</word>[ \t\n]+")) {
			point = matchend;
			c_fwd_space();
		}				
		if (parse_string(1, "function</word>")) {
			point = matchend;
			c_fwd_space();
			if (curchar() == '*') {
				point++;
				c_fwd_space();
			}
			if (curchar() == '(')
				i = tag_c_func(1);
			else {
				if (parse_string(1, "[a-zA-Z0-9_.]+", tmp)) {
					add_tag(tmp, start);
					i = 2;
				}
			}
		} else if (curchar() == '(') {
			if (parse_string(1, "%([ \t\n]*function</word>"))
				i = 2;
			else
				i = tag_c_func(1);
		}
	} else if (!strcmp(func, "var")) {
		c_fwd_space();
		if (parse_string(1, "[a-zA-Z0-9_.]+", func)) {
			if (tag_c_variable() && (tag_declarations & 2)) {
				i = 2;
				if (!(tag_which_items & 2))
					i = 1;
			}
		}
	}
	return i;
}

c_fwd_space()	/* move fwd past whitespace or comments */
{
	while (point < size() - narrow_end)
		switch (curchar()) {
		case ' ': case '\t': case '\n':
			point++; break;
		case '/':
			switch (character(point + 1)) {
			case '*':	point += 2; search(1, "*/"); break;
			case '/':	nl_forward(); break;
			default:	return;
			}
			break;
		default:
			return;
		}
}

is_java_buffer()
{
	return (c_extra_keywords & JAVA_KEYWORDS);
}

// Found start of an Objective-C block; parse it.
int tag_objc_block()
{
	char class[FNAMELEN], func[FNAMELEN];
	int i, instance_vars = 0, end;

	point += parse_string(1, "[ \t]*");
	if (!parse_string(1, "<word>+", class))
		return 1;
	if (tag_which_items & 16)
		add_tag(class, matchstart);
	tag_c_add_class(class);
	c_fwd_space();
	if (curchar() == '(')	// Skip category.
		c_skip_level('(', ')');
	c_fwd_space();
	point += parse_string(1, ":[ \t]*<word>+");		// Skip superclass.
	c_fwd_space();
	if (curchar() == '<')	// Skip protocol list.
		c_skip_level('<', '>');
	c_fwd_space();
	if (curchar() == '{') {	// Include instance variables.
		instance_vars++;	// Any vars before next } are instance vars.
		point++;
	}
	tag_modifier_kwd = 0;
	while (re_search(1, "(^@end|^[-+][ \t]*|[{};<]|/<*|/>|[a-zA-Z0-9_]+)")
		   && character(matchstart) != '@') {
		end = point;
		point = matchstart;
		switch (curchar()) {
			case '-':
			case '+':
				if (i = find_objc_method(func))
					if ((i == 1 || (tag_declarations & 1) || instance_vars)
						  && (tag_which_items & 1))
						add_class_tag(c_tagging_class, "::", func,
									  start_of_function);
				re_search(1, "![@;{]");
				break;
			case '{':
				skip_c_braces();
				break;
			case '<':
				skip_cpp_template();
				break;
			case '}':
				if (instance_vars)
					instance_vars = 0;
				else
					tag_remove_class();
				point++;
				break;
			case ';':
				tag_modifier_kwd = 0;	// assume next var
				point++;		// is being defined
				break;
			case '/':
				if (character(point + 1) == '*') {
					point += 2;
					search(1, "*/");
				} else
					nl_forward();	/* c++ comment */
				break;
			default:
				if (!good_c_tag())
					point = end;
				break;
		}
	}
	tag_remove_class();
	return 1;
}

// Found @class, parse class names that follow.
tag_objc_class_list()
{
	char class[FNAMELEN];

	for (;;) {
		c_fwd_space();
		if (!parse_string(1, "<word>+", class))
			break;
		if (tag_which_items & 16)
			add_tag(class, give_begin_line());
		c_fwd_space();
		point += parse_string(1, ",");
	}
	return 1;
}

add_tag(func, pos)
char *func;
{
	char rel[FNAMELEN], *file = filename, tdir[FNAMELEN];

	if (tag_relative) {		/* use relative version instead */
		put_tag_directory(tdir);
		relative(file, rel, tdir);
		file = rel;
	}
	if (tag_show_percent == 2)
		noteput("%r: adding %s...", file, func);
	else if (tag_show_percent == 1 && time_done(&tag_show_msg)) {
		noteput("Working %2d%%...", muldiv(point, 100, size()));
		time_begin(&tag_show_msg, 50);
	}
	if (tag_by_text) {
		int start;

		save_var point;
		point = pos;
		to_begin_line();
		start = point;
		to_end_line();
		buf_printf(_tags_buffer, "%s\t%s\t%d\t",
			func, file, pos);
		buf_xfer(_tags_buffer, start, point);
		buf_printf(_tags_buffer, "\n");
	} else
		buf_printf(_tags_buffer, "%s\t%s\t%d\n", func, file, pos);
}

put_tag_directory(dir)	// put directory of current tag file in dir
char *dir;
{
	save_var bufnum = _tags_buffer;
	put_directory(dir);
}

do_retag_file(file)	/* takes abs pathname, retags it */
char *file;
{
	char rel[FNAMELEN];

	relative(file, rel);
	delete_tags(file);
	if (tag_batch_mode)
		noteput("%s: searching for tags...", rel);
	return tag_a_file(file);
}

delete_tags(file)	/* delete all tags pointing to file (abs name) */
char *file;
{
	char rel[FNAMELEN], tdir[FNAMELEN];

	save_var abort_searching = 0;
	put_tag_directory(tdir);
	relative(file, rel, tdir);
//	noteput("%s: deleting old tags...", rel);
	do_delete_tags(tag_relative ? rel : file);
}

do_delete_tags(file)
char *file;
{
	int start, del;
	char pat[FNAMELEN], str[FNAMELEN];

	save_var bufnum;
	bufnum = _tags_buffer;
	point = 0;
	sprintf(pat, "^[^\t\n]+\t%q\t", file);
	sprintf(str, "\t%s\t", file);	// Quick search first.
	while (search(1, str)) {
		to_begin_line();
		del = parse_string(1, pat);		// Confirm this match.
		start = point;
		nl_forward();
		if (del)
			delete(start, point);
	}
}

// What's the orig buffer position associated with this line of tag buffer?
tag_line_to_position(int line, char *tagname)
{
	char buf[FNAMELEN];

	save_var bufname = "-list-definitions";
	go_line(line);
	if (parse_string(1, "^([^\t\n]+)\t[^\t\n]+\t([^\t\n]+)\t")) {
		grab(find_group(2, 1), find_group(2, 0), buf);
		if (tagname)
			grab(find_group(1, 1), find_group(1, 0), tagname);
		return strtoi(buf, 10);
	} else if (tagname)
		*tagname = 0;
	return 0;	// Shouldn't happen; we couldn't find tag info on the line.
}

list_definitions_when_displaying(from, to)
{
	char tagname[FNAMELEN], msg[FNAMELEN];
	int dlg = window_handle;

	from = from;		// not used
	if (1) {
		set_wattrib(window_handle, PVERTBORDCOLOR, color_class text);
		set_wattrib(window_handle, PHORIZBORDCOLOR, color_class text);
//		set_wattrib(window_handle, PTITLECOLOR, color_class text);
		text_color = color_class text;
	}
	to = tag_line_to_position(lines_between(0, point, 0) + 1, tagname);
	save_var window_handle = _bufed_orig_window;
	if (!(list_definitions_live_update & 1))
		save_var point;
	point = to;
	if (*tagname) {
		strcpy(msg, tag_declarations ? " Select a declaration or definition - "
									 : " Select a definition - ");
		if (list_definitions_live_update & 2)
			sprintf(msg + strlen(msg), "Line %d ",
					buf_position_to_line_number(bufnum, to));
		sprintf(msg + strlen(msg), "%s", tagname);
		window_title(dlg, BTOP, TITLELEFT(2), msg);
		set_window_caption(dlg, msg);
	}
	if (list_definitions_live_update & 1)
		position_to_tag_line();
}

// Make and display list of definitions, letting user pick one.  Flags:
// 1=instead, go to next definition and return. 2=previous definition.
// 4=force including declarations. 8=fall back to including
// declarations if no definitions.
int make_list_definitions(int flags, int repeat)
{
	char tag[FNAMELEN], pat[FNAMELEN], *title;
	int tags, tmp, pos, before_first = 0;
	int orig = bufnum, oldpoint = point, starthere = -1, old_ws = window_start;
	int orig_tab_size;

	if (when_displaying == list_definitions_when_displaying)
		return;		// Ignore double-presses.
	error_if_input(1);
	tags = zap("-list-definitions");
	tmp = tmp_buf();
	delete_buffer("-tag-errors");
	save_var point = 0, _tags_buffer = tags, tag_show_percent = 1;
	save_var tag_by_text = 1, tag_which_items = list_which_definitions;
	save_var tags_from_bsc = 0, tags_format = 0;
	if (flags & 4)
		save_var tag_declarations = 3;
	if (!*filename || !find_tagger(1))
		error("Don't know how to list definitions in this buffer.");
	orig_tab_size = tab_size;
	if (!buffer_size("-list-definitions") && (flags & 8)) {
		save_var tag_declarations = 3;
		point = 0;
		find_tagger(1);
	}
	if (!buffer_size("-list-definitions"))
		error("No definitions or declarations found in this buffer.");
	say("");
	restore_vars();
	if (want_code_coloring && recolor_range)
		recolor_buffer_range(0, size());	// Update colors for whole buffer.
	save_var point, bufnum = tags;
	point = 0;		// Copy colored lines of original file into tmp, based on tags.
	while (point < size() && parse_string(1, "^[^\t\n]+\t[^\t\n]+\t([^\t\n]+)\t")) {
		grab(find_group(1, 1), find_group(1, 0), pat);
		pos = strtoi(pat, 10);		// Position in original buffer.
		bufnum = orig;
		point = pos;
		if (give_begin_line() > oldpoint && starthere < 0) {	// Look for
			starthere = buffer_size(bufnum_to_name(tmp)) - 1;	// starting pos.
			if (starthere < 0)		// Initial position was before 1st one.
				before_first = 1;
		}
		buf_xfer_colors(tmp, give_begin_line(), give_end_line() + 1);
		bufnum = tags;
		nl_forward();
	}
	bufnum = tmp;
	delete_final_newlines();
	point = (starthere == -1) ? size() : starthere;
	to_begin_line();	// Begin at corresponding point to original location.
	goal_column = 0;
	tab_size = orig_tab_size;
	when_displaying = list_definitions_when_displaying;
	restore_vars();
	*tag = 0;
	save_var menu_width = screen_cols - _view_left - _view_right;
	save_var _bufed_orig_window = window_handle;	// Save it here.
	title = tag_declarations ? "Select a declaration or definition"
							 : "Select a definition";
	char *msg = "";
	if (flags & 3) {
		save_var bufnum = tmp;
		do {
			if (before_first && (flags & 1))
				continue;
			if (!((flags & 1) ? nl_forward() : nl_reverse()))
				msg = "No more declarations.";
		} while (--repeat > 0);
		pos = tag_line_to_position(lines_between(0, point, 0) + 1, NULL);
		restore_vars();
		point = pos;
		position_to_tag_line();
	} else if (get_choice(tmp, tag, title, title, "OK", "Cancel", "")) {
		save_var bufnum = tmp;
		pos = tag_line_to_position(lines_between(0, point, 0) + 1, NULL);
		restore_vars();
		point = oldpoint;
		quiet_set_bookmark();
		point = pos;
		position_to_tag_line();
	} else
		point = oldpoint, window_start = old_ws;
	say(msg);
	buf_delete(tmp);
	buf_delete(tags);
}

command goto_next_declaration() on cc_tab[CTRL('N')]
{
	make_list_definitions(1 | 4, iter);
	iter = 0;
}

command goto_previous_declaration() on cc_tab[CTRL('P')]
{
	make_list_definitions(2 | 4, iter);
	iter = 0;
}

command goto_next_definition()
{
	make_list_definitions(1, iter);
	iter = 0;
}

command goto_previous_definition()
{
	make_list_definitions(2, iter);
	iter = 0;
}

command list_definitions() on reg_tab[ALT('\'')]
{
	iter = 0;
	make_list_definitions((has_arg ? 4 : 0) | 8, 1);
}
