/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1988, 2017 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// A mode for Tcl files.

#include "eel.h"
#include "tcl.h"
#include "proc.h"
#include "colcode.h"
#include "perl.h"
#include "c.h"

tcl_parse_string(int t)
{
	int level = 0;

	while (re_search(1, "[][\"\\$]")) {
		set_character_color(t, point - 1, (level <= 0)
							? color_class tcl_string : -1);
		t = point - 1;
		switch (character(point - 1)) {
			case '[':
				point += parse_string(1, "[ \t]*");
				set_character_color(t, point, -1);
				t = point;
				point += parse_string(1, TCL_IDENT);
				set_character_color(t, point, color_class tcl_command);
				t = point;
				level++;
				break;
			case ']':
				set_character_color(t, point, -1);
				t = point;
				level--;
				break;
			case '$':
				set_character_color(t, point - 1, (level <= 0)
									? color_class tcl_string : -1);
				t = point - 1;
				if (parse_string(1, TCL_IDENT))
					set_character_color(t, point = matchend,
										color_class tcl_variable);
				t = point;
				break;
			case '"':
				set_character_color(t, point, color_class tcl_string);
				return;
			case '\\':
				point++;
				break;
		}
	}
	set_character_color(t, point, (level <= 0)
						? color_class tcl_string : -1);
}

tcl_parse_command(int t)
{
	point += parse_string(1, "[ \t]*");
	set_character_color(t, point, -1);
	t = point;
	if (parse_string(1, "#")) {
		to_end_line();
		set_character_color(t, point, color_class tcl_comment);
	} else if (parse_string(1, TCL_IDENT))
		set_character_color(t, point = matchend,
							color_class tcl_command);
}

color_tcl_range(from, to)
{
	int t = -1;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend;
	point = to;				// Color entire lines.
	nl_forward();
	to = point;
	save_var narrow_end = size() - to;
	set_character_color(from, to, -1);
	point = from;
	to_begin_line();
	tcl_parse_command(point);

	while (point < to) {
		if (!re_search(1, "[[\\$;\n\"]")) {
			t = size();
			break;
		}
		t = matchstart;
		switch (character(point - 1)) {		// check last char
			case ';':
			case '[':
			case '\n':
				tcl_parse_command(t);
				break;
			case '\\':
				point++;
				break;
			case '$':
				if (parse_string(1, TCL_IDENT))
					set_character_color(t, point = matchend,
										color_class tcl_variable);
				break;
			case '"':
				tcl_parse_string(t);
				break;
		}
	}
	if (to < t)
		set_character_color(to, t, -1);
	return point;
}

// Does this region have unmatched { with text after it?  Return pos if
// so.
tcl_unclosed(int start, int end)
{
	save_var narrow_start = start;
	save_var point = end;
	int level = 0;
	while (re_search(-1, "[{}]"))
		if ((level += (curchar() == '{') ? -1 : 1) < 0)
			return parse_string(1, "{[ \t]*(#|$)") ? -1 : point;
	return -1;
}

tcl_indenter()
{
	int orig = point, ind, pos;

	to_begin_line();
	if (parse_string(1, "[ \t]*}")) {
		point = matchend;
		if ((pos = default_move_level(-1, "}", "{")) >= 0)
			ind = get_indentation(pos);
	} else if (!re_search(-1, "[^ \t\n]"))	// Find previous non-blank line.
		ind = 0;
	else if ((pos = tcl_unclosed(give_begin_line(), orig)) >= 0) {
		// If there's an unmatched ( between start of this line and
		// orig, align to text after (.
		point = pos;
		point += parse_string(1, "{[ \t]*");
		ind = current_column();
	} else  {
		ind = get_indentation(point);
		to_end_line();
		if (parse_string(-1, "{[ \t]*"))
			ind += tcl_indent;
		else if (parse_string(-1, "}[ \t]*"))
			if ((pos = default_move_level(-1, "}", "{")) >= 0) {
				// If prev line ended in }, don't use its indent, but
				// indent of word before matching { line.
				point = pos;
				if (re_search(-1, "[^ \t\n]"))
					ind = get_indentation(point);
			}
	}
	point = orig;
	to_indentation();
	indent_to_column(ind);
}

tcl_indent_cmd() on tcl_tab['\t']
{
	standard_tab_cmd(tcl_indenter, tcl_indent, 0);
}

tcl_indent_and_show_matching_delimiter() on tcl_tab['{'], tcl_tab['}']
{
	normal_character();
	save_spot point;
	tcl_indenter();
	save_var inside_show_matching_delimiter = 1;
	restore_vars();
	if (Matchdelim)
		find_delimiter();
}

command tcl_mode()
{
	mode_default_settings();
	mode_keys = tcl_tab;		/* Use these keys. */
	major_mode = _tcl_mode_name;
	compile_buffer_cmd = compile_tcl_cmd;	// can compile this
	strcpy(comment_start, "#[ \t]*");
	strcpy(comment_pattern, "#.*$");
	strcpy(comment_begin, "# ");
	strcpy(comment_end, "");
	buffer_maybe_break_line = generic_maybe_break_line;
	fill_mode = (misc_language_fill_mode & 128) != 0;
	recolor_range = color_tcl_range;	// set up coloring rules
	recolor_from_here = recolor_from_top;
	indenter = tcl_indenter;
	auto_indent = 1;
	if (want_code_coloring)		// maybe turn on coloring
		when_setting_want_code_coloring();
	if (auto_show_tcl_delimiters)
		auto_show_matching_characters = tcl_auto_show_delim_chars;
	try_calling("tcl-mode-hook");
	drop_all_colored_regions();
	make_mode();
}


when_loading()
{
	tcl_tab[ALT('q')] = (short) fill_comment;
	tcl_tab[ALT('Q')] = (short) prefix_fill_paragraph;
}

// Set display_func_name to the name of the function we're editing, and
// return 1.  If not in a function, set display_func_name to "" and
// return 1.  If user pressed a key and we gave up for now, return 0.

tcl_func_name_finder()
{
	int from, to;

	save_var point, case_fold = 0;
	to_end_line();
	while (re_search(-1, "^[ \t]*(proc)[ \t]+(" TCL_IDENT ")|^}")) {
		if (!must_find_func_name && (char_avail() || is_key_repeating()))
			return 0;		// Give up search now?
		if (character(matchstart) == '}')
			break;
		from = find_group(2, 1);
		to = find_group(2, 0);
		to = MIN(to, from + ptrlen(display_func_name) - 1);
		grab(from, to, display_func_name);
		return 1;
	}
	*display_func_name = 0;
	return 1;
}

tag_mode_tcl()
{
	char func[TAGLEN];
	int start;

	save_var case_fold = 1;
	while (re_search(1, "^[ \t]*(proc)[ \t]*(" TCL_IDENT ")")) {
		grab(start = find_group(2, 1), find_group(2, 0), func);
		add_tag(func, start);
	}
}

suffix_tcl()
{
	tcl_mode();
}

suffix_ttml()
{
	tcl_mode();
}
