/************************************************** -*- tab-width:4 -*- *
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*	Copyright (C) 2006, 2020 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// A minor mode to continuously highlight misspelled words.

#include "eel.h"
#include "vspell.h"
#include "colcode.h"

// If the file for this dictionary size, region and group exists,
// insert it into this buffer at the start.
grab_spell_dict(char *dictsz, char *region, char *group)
{
	char fname[FNAMELEN];
	get_customization_directory(fname);
	sprintf(fname + strlen(fname), "spell/final/%s-%s.%s",
			region, group, dictsz);
	point = 0;
	if (check_file(fname) == CHECK_FILE)
		do_insert_file(fname, FILETYPE_AUTO);
}

// Translate some more options into dictionary names.
grab_spell_dict_region(char *dictsz, char *region, char *other)
{
	if (strchr(other, 'Aa'))
		grab_spell_dict(dictsz, region, "abbreviations");
	if (strchr(other, 'Cc'))
		grab_spell_dict(dictsz, region, "contractions");
	if (strchr(other, 'Uu'))
		grab_spell_dict(dictsz, region, "upper");
	if (strchr(other, 'Pp'))
		grab_spell_dict(dictsz, region, "proper-names");
	grab_spell_dict(dictsz, region, "words");
}

// Translate region code and variant spec into dictionary names and
// load them along with the main dictionary for this size.
grab_spell_dict_size(char *dictsz, int region, char *other)
{
	if (strpbrk(other, "Ll"))
		grab_spell_dict_region(dictsz, "variant_2", other);
	if (strpbrk(other, "MLml"))
		grab_spell_dict_region(dictsz, "variant_1", other);
	if (strpbrk(other, "SMLsml"))
		grab_spell_dict_region(dictsz, "variant_0", other);
	switch (toupper(region)) {
		case 'A': grab_spell_dict_region(dictsz, "american", other); break;
		case 'B': grab_spell_dict_region(dictsz, "british", other); break;
		case 'Z': grab_spell_dict_region(dictsz, "british_z", other); break;
		case 'C': grab_spell_dict_region(dictsz, "canadian", other); break;
	}
	grab_spell_dict_region(dictsz, "english", other);
}

spell_configure_helper(int w)
{
	char file[FNAMELEN], ok[FNAMELEN], pr[FNAMELEN];

	sprintf(ok, "NAI%sU", (is_unix_flavor == IS_UNIX_MACOS || is_mswindows) ? "M" : "");
	sprintf(pr, "External spell helper? (None, Aspell, Ispell%s, Unchanged) [U] ",
			is_mswindows ? ", MicroSpell" :
			(is_unix_flavor == IS_UNIX_MACOS) ? ", macOS" : "");
	show_popup_help(w, "The speller can optionally use an external helper "
					"program to provide better, faster suggestions when "
					"correcting misspellings.  If you have one of the supported "
					"programs installed, select it now.\n\nThe aspell and "
					"earlier ispell programs are commonly installed on Unix "
					"systems, and may also be used on Windows systems via the "
					"Cygwin system.\n\nMicroSpell is a commercial spelling "
					"checker that provides hotkey spell checking under "
					"Windows.  See www.microspell.com.  You must have version "
					"9.1f or later.\n\nThe macOS speller is available under "
					"macOS systems (but not over ssh).\n\nThis option sets "
					"the spell-helper-program variable.  If your aspell/ispell "
					"program is not on your path, you'll have to modify this "
					"variable to include its full path.");
	switch (get_key_response(pr, ok, 'U', "spell-configure")) {
		case 'A':	strcpy(spell_helper_program, "aspell -a");
					break;
		case 'I':	strcpy(spell_helper_program, "ispell -a");
					break;
		case 'M':	if (is_unix_flavor == IS_UNIX_MACOS) {
						strcpy(spell_helper_program, "\"%xmacspell\"");
						break;
					}
					get_executable_file(file, "mspellcmd.exe", 0);
					if (!check_file(MICROSPELL_CMDLINE))
						say("Please install %s to %s as admin", file, MICROSPELL_CMDLINE);
					sprintf(spell_helper_program, "\"%s\"", MICROSPELL_CMDLINE);
					break;
		case 'N':	*spell_helper_program = 0;
					break;
	}
}

int spell_download_one(char *dir, char *arch, int sz, int ask)
{
	char archpath[FNAMELEN], url[FNAMELEN], cmdline[FNAMELEN], *s;
	struct file_info fi;
	sprintf(archpath, "%s/%s", dir, arch);
	sprintf(url, "http://www.lugaru.com/%s", arch);
	s = lookpath(arch);
	if (check_file(archpath, &fi) != CHECK_FILE || fi.fsize * 10 < sz * 9) {
		if (s && check_file(s, &fi) == CHECK_FILE && fi.fsize * 10 >= sz * 9)
			strcpy(archpath, s);
		else if (!ask)
			return 0;
		else if (!ask_yn("Download and install dictionary file?", 
						 "Download and install a dictionary file "
						 "(about 700 KB)? ", "Yes", "No", 1))
			return 1;
		else {
			say("Please wait...");
			if (download_file_to_disk(url, archpath, sz))
				error("Couldn't download dictionary file from %s", url);
		}
	}
	if (check_file(archpath, &fi) != CHECK_FILE || fi.fsize * 10 < sz * 9)
		error("Bad dictionary archive %f, please retrieve manually.",
			  archpath);
	convert_to_8_3_filename(archpath);
	sprintf(cmdline, is_mswindows ? "\"%s\"" : "\"tar\" xzf \"%s\"", archpath);
	pipe_text(NULL, "-spell-config", cmdline,
			  dir, PIPE_SYNCH | PIPE_NOREFRESH, 0);
	return 2;
}

spell_download(char *dir)
{
	if (!spell_download_one(dir, is_mswindows ? "spell.exe" : "spell.tgz",
							is_mswindows ? 2_398_133 : 2_258_676, 0))
		spell_download_one(dir, is_mswindows ? "spell2.exe" : "spell2.tgz",
						   is_mswindows ? 771_401 : 692_231, 1);
}

// Construct a words.lst file by downloading a dictionary archive and
// combining files per user specifications.

command spell_configure()
{
	char dir[FNAMELEN], file[FNAMELEN], other[FNAMELEN];
	if (exist(WORD_LIST_BUFFER))
		reload_spell_word_list();
	save_var want_gui_prompts = 0;
	get_customization_directory(dir);
	absolute(dir);
	strcat(dir, "/spell");
	mkdir(dir);
	if (check_file(dir) != CHECK_DIR)
		error("Couldn't create working directory %f", dir);
	sprintf(file, "%s/final/english-words.10", dir);
	if (check_file(file) != CHECK_FILE)		// Download and extract.
		spell_download(dir);
	if (check_file(file) != CHECK_FILE)
		error("Couldn't extract files from dictionary archive.");
	
	int region = get_key_response("Use American, British, British with -ize, "
								  "or Canadian spellings [ABZC]? ",
								  "ABZC", 0, "spell-configure");

	int help = tmp_buf(), w = add_popup(10, 10, 60, 10, _view_border, help);
	on_exit buf_delete(help), remove_window(w);
	window_title(w, BTOP, TITLECENTER,  " Spell Configuration Help ");

	show_popup_help(w, "Select a dictionary size.  Bigger dictionaries "
					" recognize rarer words, but won't catch misspellings "
					"that result in rare words.");
	int dictsz = get_key_response("Dictionary size (small, medium, large) "
								  "[SML, default L]? ", "SML", 'L',
								  "spell-configure");

	save_var insert_default_response = 1;
	show_popup_help(w, "Should the dictionary include abbreviations, "
					"contractions, uppercase words, or proper names?  Enter "
					"A, C, U, and P, in any combination, to select these.\n\nThe "
					"Abbreviations category includes abbreviations and acronyms "
					"like \"TV\", \"VIP\", and \"rpm\".\n\nThe Contractions "
					"category includes \"doesn't\" and \"I'm\".\n\nThe Uppercase "
					"category includes words that must start with an uppercase "
					"letter, like \"Doctor\", \"December\", or \"Europe\".\n\n"
					"The Proper Names category includes less common names like "
					"\"Brian\" or \"Altman\" that are not included in the "
					"Uppercase category.\n\nSome words appear in more than one "
					"category.");
	get_strdef(other, "Include Abbrevs, Contractions, "
			   "Uppercase, Proper names in dictionary", "ACUP");

	show_popup_help(w, "Should the speller permit variant spellings, instead "
					"of permitting only a single preselected spelling of each "
					"word?\n\nThis permits words like \"fiascos\" in addition to "
					"\"fiascoes\" or \"Hanukah\" in addition to \"Hanukkah\".");
	sprintf(other + strlen(other), "%c",
			get_key_response("Permit variant spellings (None, Some, Many, Lots) "
							 "[Some]? ", "NSML", 'S', "spell-configure"));
	spell_configure_helper(w);
	save_var bufnum = zap("-spell-config");
	switch (dictsz) {
		case 'L':
			grab_spell_dict_size("70", region, other);
			grab_spell_dict_size("60", region, other);
			grab_spell_dict_size("55", region, other);
		case 'M':
			grab_spell_dict_size("50", region, other);
			grab_spell_dict_size("40", region, other);
		case 'S':
			grab_spell_dict_size("35", region, other);
			grab_spell_dict_size("20", region, other);
			grab_spell_dict_size("10", region, other);
	}
	sprintf(file, "%s/../%s", dir, WORD_ORDER_FILE);
	int err = file_write(file, FILETYPE_AUTO);
	if (err)
		file_error(err, file, "write error");
	case_fold = 0;
	do_sort_region(0, size(), 0, 0);
	sprintf(file, "%s/../%s", dir, WORD_FILE_NAME);
	err = file_write(file, FILETYPE_AUTO);
	if (err)
		file_error(err, file, "write error");
	restore_vars();
	reload_spell_word_list();
	say("Dictionary configured, run spell-mode or spell-buffer.");
}

char *get_spell_language_prefix()
{
	return spell_language_prefix ? spell_language_prefix : "";
}

// Return the regex pattern for this mode that finds words we want to
// check.  It can use color class assertions.
char *get_spell_word_pattern()
{
	char *s = get_mode_string_variable("%s-spell-word-pattern");
	if (s)
		return s;
	if (when_displaying == recolor_partial_code)
		return default_color_spell_word_pattern;
	else
		return default_spell_word_pattern;
}

// Load word list file.  Flavor is SPELLB_ORDERED for ordered dictionary
// (common words first), SPELLB_SORTED for sorted version of that,
// SPELLB_GLOBAL for global ignore list, SPELLB_PERDIR for per-directory
// ignore list, SPELLB_PERFILE for per-file ignore list, SPELLB_PERMODE
// for per-mode ignore list, or SPELLB_PEREXTEN for per-extension ignore
// list.  Return buffer number, or 0 if no file, or it's empty.  If
// action is 1, force reloading file.  If action is 2, force creating a
// buffer for user file.

int spell_load_file(int flavor, int action)
{
	char fname[FNAMELEN], bname[FNAMELEN], mode[FNAMELEN], *s = 0, *t;
	char lookfor[FNAMELEN];
	int b, bit = (1 << flavor);
	switch (flavor) {
		case SPELLB_ORDERED:
			strcpy(bname, WORD_ORDER_BUFFER);
			break;
		case SPELLB_SORTED:
			strcpy(bname, WORD_LIST_BUFFER);
			break;
		case SPELLB_GLOBAL:
			strcpy(bname, "-spell-ignore-global");
			break;
		case SPELLB_PERDIR:
			if (!*filename)
				return 0;
			sprintf(bname, "-spell-ignore-%s", filename);
			*get_tail(bname, 0) = 0;
			break;
		case SPELLB_PERFILE:
			if (!*filename)
				return 0;
			sprintf(bname, "-spell-ignore-%s", filename);
			break;
		case SPELLB_PERMODE:
			clean_mode(mode);
			sprintf(bname, "-spell-ignore-mode-%s", mode);
			break;
		case SPELLB_PEREXTEN:
			if (!*filename)
				return 0;
			sprintf(bname, "-spell-ignore-extension-%s",
					get_extension(filename));
			break;
	}
	strcat(bname, get_spell_language_prefix());
	if ((b = name_to_bufnum(bname)) && !action)
		return buf_size(b) || (action == 2) ? b : 0;
	switch (flavor) {
		case SPELLB_ORDERED:
			s = WORD_ORDER_FILE;
			break;
		case SPELLB_SORTED:
			s = WORD_FILE_NAME;
			break;
		case SPELLB_GLOBAL:
			s = "ignore.lst";
			break;
		case SPELLB_PERDIR:
			strcpy(fname, filename);
			strcpy(get_tail(fname, 0), ".directory.espell");
			strcat(fname, get_spell_language_prefix());
			break;
		case SPELLB_PERFILE:
			strcpy(fname, filename);
			sprintf(get_tail(fname, 0), ".%s.espell", get_tail(filename, 0));
			strcat(fname, get_spell_language_prefix());
			break;
		case SPELLB_PERMODE:
			sprintf(fname, "ignore.%s.mode.lst", mode);
			strcpy(s = mode, fname);
			break;
		case SPELLB_PEREXTEN:
			sprintf(fname, "ignore.%s.extension.lst", get_extension(filename));
			strcpy(s = mode, fname);
			break;
	}
	if (s) {	// Search for these.
		create(bname);
		sprintf(lookfor, "%s%s", s, get_spell_language_prefix());
		if (t = lookpath(lookfor))
			strcpy(fname, t);
		else if (action == 2 && (flavor == SPELLB_GLOBAL
								 || flavor == SPELLB_PERMODE
								 || flavor == SPELLB_PEREXTEN)) {
			get_customization_directory(fname);
			strcat(fname, lookfor);
		} else
			return 0;
	} else {	// Look just once from each buffer.
		if (action)		// Clear the missing bit.
			vspell_data.missing_file &= ~bit;
		if (!(vspell_data.missing_file & bit)
			  && (check_file(fname) != CHECK_FILE))
			vspell_data.missing_file |= bit;	// Remember it's missing.
		if ((vspell_data.missing_file & bit) && action != 2)
			return 0;
	}
	save_var bufnum = create(bname);
	filename = fname;
	case_fold = 0;
	if (modified && action != 2)
		return 0;		// Don't mess with buffer while user modifies it.
	if (action == 2)
		file_read(fname, FILETYPE_AUTO);
	else if (err_file_read(fname, FILETYPE_AUTO) && action != 2)
		return 0;
	switch (flavor) {	// Sort only user's ignore buffers.
		case SPELLB_GLOBAL:
		case SPELLB_PERDIR:
		case SPELLB_PERFILE:
		case SPELLB_PERMODE:
		case SPELLB_PEREXTEN:
			save_var abort_searching = 0, show_status = 0, modified;
			buf_sort_and_uniq(bufnum);
			break;
	}
	return bufnum;
}

// Load the two dictionary files.  Sort the user one.  Return 1 if ok.
int load_spell_word_list(int force_reload)
{
	if (!spell_load_file(SPELLB_ORDERED, force_reload)) {
		if (!vspell_data.told_err) {
			if (*get_spell_language_prefix())
				say("Can't find speller word list %f%s.",
					  WORD_ORDER_FILE, get_spell_language_prefix());
			else
				say("Can't find speller word list %f, run spell-configure.",
					  WORD_ORDER_FILE);
			vspell_data.told_err = 1;
		}
		return 0;
	}
	spell_load_file(SPELLB_SORTED, force_reload);
	spell_load_file(SPELLB_GLOBAL, force_reload);
	spell_load_file(SPELLB_PERDIR, force_reload);
	spell_load_file(SPELLB_PERFILE, force_reload);
	spell_load_file(SPELLB_PERMODE, force_reload);
	spell_load_file(SPELLB_PEREXTEN, force_reload);
	return 1;
}

// Discard speller word list and reload it, then ensure next refresh
// uses new list.
reload_spell_word_list()
{
	load_spell_word_list(1);
	int owin = window_number;
	do {
		vspell_data.window_start = -1;	// Force update in every window.
		check_abort();
	} while (++window_number != owin);
}

// Return 1 if this line appears in the buffer buf, which we assume is
// already sorted.
int word_in_sorted_buffer(int buf, char *line)
{
	char have_line[2048];
	save_var bufnum = buf;
	int sz = 1, here, res;
	while (sz < size())		// Round the size up to a power of two.
		sz *= 2;
	here = sz /= 2;			// Start in the middle.
	while (sz > 0) {
		sz /= 2;			// Next time move by a smaller amount.
		if (here >= size()) {
			here -= sz;
			continue;
		}
		point = here;
		to_begin_line();
		grab(point, give_end_line(), have_line);
		if (res = strcmp(line, have_line))	// If not found.
			here += (res > 0) ? sz : -sz; // Move by an ever-shrinking distance.
		else				// We found the line.
			return 1;
	}
	return 0;
}

// Is this word in the dictionary buffer buf?  If it contains any
// upper case letters, also check the lower cased version.
int word_in_dictionary(int buf, char *word)
{
	char alt[FNAMELEN];

	if (!buf)
		return 0;
	int i, have_upper = 0;
	if (word_in_sorted_buffer(buf, word))
		return 1;
	strcpy(alt, word);		// Build lower case version.
	for (i = 0; alt[i]; i++) {
		if (isupper(alt[i]))
			have_upper = 1;
		alt[i] = tolower(alt[i]);
	}
	if (have_upper && word_in_sorted_buffer(buf, alt))
		return 1; // Check lowercase version if there were any uppercase chars.
	if (strlen(word) > 3 && word[strlen(word) - 2] == '\''
		  && tolower(word[strlen(word) - 1]) == 's'
		  && !(get_num_var(vspell_get_mode_variable()) & 0x200)) {
		strcpy(alt, word);
		alt[strlen(alt) - 2] = 0;		// Remove possessive 's and retry.
		if (word_in_sorted_buffer(buf, alt))
			return 1;
	}
	return 0;
}

// Return 1 if this word is spelled correctly.  Dictionary buffers
// should already be loaded; create empty ones if not.
int good_word_spelling(char *word)
{
	char listbuf[FNAMELEN];
	sprintf(listbuf, "%s%s", WORD_LIST_BUFFER, get_spell_language_prefix());
	return word_in_dictionary(create(listbuf), word)
			|| word_in_dictionary(spell_load_file(SPELLB_GLOBAL, 0), word)
			|| word_in_dictionary(spell_load_file(SPELLB_PERDIR, 0), word)
			|| word_in_dictionary(spell_load_file(SPELLB_PERFILE, 0), word)
			|| word_in_dictionary(spell_load_file(SPELLB_PERMODE, 0), word)
			|| word_in_dictionary(spell_load_file(SPELLB_PEREXTEN, 0), word);
}

// Get word that matched pattern we just searched for.  Return 0 if it's
// misspelled, nonzero otherwise.
int spell_get_and_check_word(int opt, char *word)
{
	int start = matchstart, end = matchend;

	save_var matchstart, matchend;
	if (matchend - matchstart > FNAMELEN - 1)
		return 1;
	grab(matchstart, matchend, word);
	if (spell_ignore_word(opt, matchstart, matchend, word))
		return 2;
	if (good_word_spelling(word))
		return 3;
	char *s = get_mode_string_variable("%s-spell-ignore-pattern-prefix");
	if (s && *s && matches_at(start, -1, s))
		return 4;
	s = get_mode_string_variable("%s-spell-ignore-pattern-suffix");
	if (s && *s && matches_at(end, 1, s))
		return 4;
	return 0;
}

spell_add_to_helper_ignore_list(int flavor, char *word)
{
	if (!*spell_helper_program || flavor != SPELLB_GLOBAL
		  || !(get_num_var(vspell_get_mode_variable()) & 0x4000))
		return;
	save_var bufnum = run_spell_helper_program();
	point = size();
	bprintf("*%s\n#\n", word);
}

// Add this word to the specified ignore list and refresh.
spell_ignore_this_word(int flavor, char *word, int doc)
{
	save_var bufnum = doc;
	char fname[FNAMELEN];
	int ignore = spell_load_file(flavor, 2);
	if (!ignore)
		error("No file associated with this buffer.");
	bufnum = ignore;
	point = size();
	bprintf("%s%s\n", size() && !parse_string(-1, "\n") ? "\n" : "", word);
	if (global_spell_options & 1)
		do_save_file(0, 0, 0);
	bufnum = doc;
	switch (flavor) {
		case SPELLB_PERDIR:
			strcpy(fname, filename);
			*get_tail(fname, 0) = 0;
			absolute(fname);
			say("\"%s\" added to ignore list for %f.", word, fname);
			break;
		case SPELLB_PERFILE:
			say("\"%s\" added to ignore list for %f.", word, filename);
			break;
		case SPELLB_PERMODE:
			say("\"%s\" added to ignore list for %s mode.", word, major_mode);
			break;
		case SPELLB_PEREXTEN:
			say("\"%s\" added to ignore list for %s extension.",
				  word, get_extension(filename));
			break;
		default:
			say("\"%s\" added to global ignore list.", word);
			break;
	}
	spell_add_to_helper_ignore_list(flavor, word);
	reload_spell_word_list();
}

maybe_delete_spell_ignore_list(char *file)
{
	char fname[FNAMELEN];

	if (!(global_spell_options & 2) || !*file)
		return;
	strcpy(fname, file);
	sprintf(get_tail(fname, 0), ".%s.espell", get_tail(file, 0));
	delete_file(fname);
}

// Wait until some new output from the spell helper program matches the
// pattern.
spell_wait_for_response(char *pat)
{
	save_var bufnum = create("-spell-helper"), point;
	int pos = point;
	delayed_say(0, 200, 50, "Waiting for spell helper program...");
	do {
		point = pos;
		delay(1, COND_KEY);
		check_abort();
		if (!is_process_buffer(bufnum))
			error("Error talking to %s", spell_helper_program);
	} while (!re_search(1, pat));
	delayed_say(0, 0, 0, "");
}

// Run the program specified in spell_helper_program if it isn't
// already running.  It will output a greeting line starting with @.
// Return the buffer where it's running.
int run_spell_helper_program()
{
	save_var bufnum = create("-spell-helper");
	point = size();
	if (is_process_buffer(bufnum))
		return bufnum;
	char prog[FNAMELEN];
	expand_string_template(prog, spell_helper_program, "", NULL);
	convert_to_8_3_filename(prog);
	concur_shell("", prog, NULL, bufname);
	spell_wait_for_response("^@");
	return bufnum;
}

// Replace the current buffer's contents with a list of suggestions for
// the word generated by the helper program.
spell_get_helper_suggestions(char *word)
{
	int orig = bufnum;
	buf_zap(bufnum);
	save_var bufnum = run_spell_helper_program();
	point = size();
	int pos = point;
	bprintf("^%s\n", word);
	spell_wait_for_response("\n\n");
	buf_xfer(orig, pos, size());
	restore_vars();
	if (parse_string(-1, "<*>\n*")) {
		buf_zap(bufnum);
		bprintf("%s\n", word);
	} else {
		simple_re_replace(1, "^[^&].*\n", "");
		simple_re_replace(1, "<bof>.*: *", "");
		simple_re_replace(1, " *, *", "\n");
		simple_re_replace(1, "\n\n", "\n");
		simple_re_replace(1, "<bof>\n+<eof>", "");
	}
}

// This subroutine constructs and returns a malloc'd regex pattern.  The
// pattern matches a line containing exactly the given word, plus
// variations on that word generated by transposing a pair of adjacent
// characters, inserting or deleting a single character, or replacing
// one character with another.  If exact, instead return a pattern that
// matches exactly.

char *get_misspell_pattern(char *word, int exact)
{
	int len = strlen(word), i, c;
	char pat[FNAMELEN], *ret;
	ret = malloc(4 * (len + 1) * (len + 1) + 5);
	if (len < 2)
		exact = 1;
	if (exact) {
		sprintf(ret, "^%s$", word);
		return ret;
	}
	strcpy(ret, "^(");
	for (i = 0; i < len - 1; i++) {		// Find with transposed letter.
		sprintf(pat, "%s|", word);
		c = pat[i];
		pat[i] = pat[i + 1];
		pat[i + 1] = c;
		strcat(ret, pat);
	}
	for (i = 0; i < len; i++) {		 // Find with wrong letter.
		sprintf(pat, "%.*s.%s|", i, word, word + i + 1);
		strcat(ret, pat);
	}
	for (i = 0; i <= len; i++) {	// Find with extra letter.
		sprintf(pat, "%.*s.%s|", i, word, word + i);
		strcat(ret, pat);
	}
	for (i = 0; i < len; i++) {		// Find with missing letter.
		sprintf(pat, "%.*s%s|", i, word, word + i + 1);
		strcat(ret, pat);
	}
	strcpy(ret + strlen(ret) - 1, ")$");
	return ret;
}

// Search this buffer for lines matching goal.  For each, copy the
// corresponding line in the buffer buf to the buffer dest.  Return
// number of matches.  If fuzzy, find approximate matches too.
int spell_copy_matching(char *goal, int fuzzy, char *buf, int dest)
{
	char word[FNAMELEN], *pat = get_misspell_pattern(goal, !fuzzy);
	int words = name_to_bufnum(buf), cnt = 0;
	point = 0;
	while (re_search(1, pat)) {
		get_line_from_buffer(words, lines_between(0, point) + 1, word);
		buf_printf(dest, "%s\n", word);
		cnt++;
	}
	free(pat);
	return cnt;
}

// Split word into two words; if they're both in the dictionary, insert
// into the word list in the current buffer, which has more common
// words first.  Insert the new line by the least common of the two
// words.
spell_insert_runtogethers(char *word)
{
	char w1[FNAMELEN], w2[FNAMELEN], pat[FNAMELEN];
	int len = strlen(word), i, p1, p2;

	sprintf(pat, "%s%s", WORD_LIST_BUFFER, get_spell_language_prefix());
	int buf = create(pat);
	for (i = 2; i < len - 1; i++) {
		strncpy(w1, word, i);
		strcpy(w2, word + i);
		if (!word_in_dictionary(buf, w1) || !word_in_dictionary(buf, w2))
			continue;		// Fast binary search rejects most combinations.
		sprintf(pat, "^%s$", w1);
		point = 0;
		if (!re_search(1, pat))
			continue;
		p1 = give_begin_line();
		sprintf(pat, "^%s$", w2);
		point = 0;
		if (!re_search(1, pat))
			continue;
		p2 = give_begin_line();
		point = MAX(p1, p2);
		bprintf("%s %s\n", w1, w2);
	}
}

// Return a buffer with a list of correctly spelled words similar to
// the misspelled goal word.

int spell_find_suggestions(char *goal, int max, int case_repl)
{
	int poss = tmp_buf(), sugg = zap("-spell-suggestions");
	char sound[FNAMELEN], *wild, ord[FNAMELEN], pat[FNAMELEN];

	sprintf(ord, "%s%s", WORD_ORDER_BUFFER, get_spell_language_prefix());
	on_exit buf_delete(poss);
	if (*spell_helper_program) {
		save_var bufnum = sugg;
		spell_get_helper_suggestions(goal);
		return sugg;
	}
	int fast = get_num_var(vspell_get_mode_variable()) & 0x1000;
	if (fast || !exist(ord) || !buffer_size(ord))
		sprintf(ord, "%s%s", WORD_LIST_BUFFER, get_spell_language_prefix());
	save_var bufnum = poss;
	if (fast) {
		bprintf("%s\n", goal);
		phoneticize_lines(sugg, 5);	// Find the sound code for the goal word.
		get_line_from_buffer(sugg, 0, sound);	// Record it here.
		buf_zap(sugg);
		buf_zap(poss);
		if (!exist(SOUNDS_BUFFER)) {	// Make sound codes for
			load_spell_word_list(0);	// all dictionary words.
			bufname = ord;	
			phoneticize_lines(zap(SOUNDS_BUFFER), 5);
		}
		bufname = SOUNDS_BUFFER;
		// Copy words with the same sound to poss.  If there aren't many
		// and the sound is more than one character long, try copying
		// variations too.
		if (spell_copy_matching(sound, 0, ord, poss) < max
			  && strlen(sound) > 1)
			spell_copy_matching(sound, 1, ord, poss);
	} else {		// Make a copy of buffer into poss, insert runtogethers.
		grab_buffer(name_to_bufnum(ord));
		spell_insert_runtogethers(goal);
	}
	bufname = bufnum_to_name(poss);
	wild = get_misspell_pattern(goal, 0);
	point = 0;
	while (re_search(1, wild))
		buf_xfer(sugg, give_begin_line(), give_end_line() + 1);
	free(wild);
	bufnum = sugg;
	if (case_repl)
		spell_fix_suggestion_case(case_repl);
	save_var case_fold = 0;
	point = 0;
	sprintf(pat, "^%s$", goal);
	if (re_search(1, pat))
		bprintf(" (OK)");
	return sugg;
}

// The buffer contains a list of suggested words.  For each entirely
// lowercase one, uppercase it if rule is 1, or capitalize it if rule
// is 2.
spell_fix_suggestion_case(int rule)
{
	save_var point = 0, case_fold = 0;
	while (re_search(1, "^[a-z]+$"))
		switch (rule) {
			case 1: do_uppercase_region(matchstart, point); break;
			case 2: do_capitalize_region(matchstart, point); break;
		}
}

// Copy suggestion starting with digit i from buf into word, then clean
// up.  Return 1 if OK.
int spell_get_suggestion_line(int buf, int i, char *word)
{
	char pat[FNAMELEN], *s;
	save_var bufnum = buf, point = 0;
	sprintf(pat, "^%c ", i);
	if (!re_search(1, pat))
		return 0;
	grab(point, give_end_line(), word);
	if (s = strstr(word, " (OK)"))
		*s = 0;
	return 1;
}

int spell_correct_move_highlight(int hline, int win, int dir, int setkey)
{
	save_var window_handle = win;
	if (hline == -1)
		make_line_highlight();
	hline += dir;
	int lines = lines_between(0, size(), 0);
	hline = (hline + lines) % lines;
	go_line(hline + 1);
	if (setkey && character(point + 1) == ' ')
		key = curchar();
	return hline;
}

spell_ignore_word_this_run(char *word)
{
	say("Will resume checking \"%s\" when this command ends.", word);
	save_var bufnum = create(SPELL_THIS_RUN);
	case_fold = 0;
	point = size();
	bprintf("%s\n", word);
	buf_sort_and_uniq(bufnum);
}

// Ask user what to do about word at point.  Assumes word list is
// already loaded.  If mult, doing a series of corrections, and can
// return 1 to indicate recursive edit.  Return 0 normally, 2 if no
// word to correct.
do_spell_correct(int mult)
{
	char word[FNAMELEN];
	save_var point--;
	re_search(RE_FORWARD | RE_IGNORE_COLOR, get_spell_word_pattern());
	re_search(RE_REVERSE | RE_IGNORE_COLOR, get_spell_word_pattern());
	int wstart, wend, i = 1, hline = -1, have_file = *filename;
	grab(wstart = matchend, wend = matchstart, word);
	if (!*word)
		return 2;
	int case_repl = (get_num_var(vspell_get_mode_variable()) & 0x2000)
					? 0 : check_for_case_replace(wstart, wend);
	build_window();
	int wcol = window_left + column_in_window;
	int wline = window_top + line_in_window, obuf = bufnum;
	int sugg = spell_find_suggestions(word, 9, case_repl);
	save_var bufnum = sugg;
	remove_line_highlight();
	go_line(10);
	delete(point, size());
	int wcnt = lines_between(0, size());
	point = 0;
	if (point < size()) {
		do {
			bprintf("%d ", i++);
		} while (nl_forward() && point < size());
		stuff("\n");
	} else
		stuff("No suggestions\n\n");
	stuff("O Ignore word once\n");
	if (mult)
		stuff("R Ignore on this run only\n");
	if (have_file) {
		stuff("F Ignore in this file\n");
		stuff("D Ignore in this directory\n");
	}
	stuff("M Ignore for this mode only\n");
	int by_exten = have_file && (global_spell_options & 0x10);
	if (by_exten)
		stuff("E Ignore for this extension only\n");
	stuff("G Ignore globally\n");
	if (mult)
		stuff("Ctrl-R for recursive edit\n");
	point = 0;
	int cnt = lines_between(0, size());
	restore_vars();
	highlight_on();
	save_spot mark = wstart;
	save_spot point = wend;
	on_exit highlight_off();
	save_var want_gui_prompts = 0;
	save_var window_handle;
	int row = 5;
	if (wline + cnt + 4 < screen_lines)
		row = wline + 2;			// Position below text.
	else if (wline > cnt + 2)
		row = wline - cnt - 3;		// Or above.
	else if (row + cnt >= screen_lines - 1)
		if ((row = screen_lines - cnt - 3) + cnt <= wline)
			row = 0;
	int col = wcol - 3;				// Align with text.
	if (col + 35 > screen_cols)		// Too far to right, so move back.
		col = screen_cols - 35;
	if (col < 0)					// Too far to left, move right.
		col = 0;							// If not visible above or below, 
	if (row <= wline && wline < row + cnt)	// move left or right
		if (wcol + 35 > screen_cols)
			col = wcol - 35;
		else
			col = wcol + strlen(word) + 3;
	int win = add_popup(col, row, 30, cnt + 2, _view_border, sugg);
	on_exit remove_window(win);
	text_color = (is_gui && want_gui_prompts) ? color_class viewed_gui_text
											  : color_class viewed_text;
	set_wattrib(win, PVERTBORDCOLOR, color_class viewed_border);
	set_wattrib(win, PHORIZBORDCOLOR, color_class viewed_border);
	window_title(win, BTOP, TITLECENTER,  " Suggestions ");

	display_scroll_bar = 0;
	for (;;) {
		char keys[20];
		sprintf(keys, "O%s%sM%sG", mult ? "R" : "", have_file ? "FD" : "",
				by_exten ? "E" : "");
		if (wcnt > 1)
			noteput("1-%d to replace, ignore \"%s\" Once, or add "
					"to an ignore list [1-%d%s] ", wcnt, word, wcnt, keys);
		else if (wcnt == 1)
			noteput("1 to replace, ignore \"%s\" Once, or add "
					"to an ignore list [1%s] ", word, keys);
		else
			noteput("No suggestions.  Ignore \"%s\" Once, or add "
					"to an ignore list [%s] ", word, keys);
		refresh();
		getkey();
		say("");
		check_abort();
		if (binding_of(key, reg_tab) == down_line)
			hline = spell_correct_move_highlight(hline, win, 1, 0);
		else if (binding_of(key, reg_tab) == up_line)
			hline = spell_correct_move_highlight(hline, win, -1, 0);
		else if (IS_ENTER_KEY(key) && hline != -1)
			hline = spell_correct_move_highlight(hline, win, 0, 1);
		if (IS_MOUSE_LEFT(key)
			  && window_at_coords(mouse_y, mouse_x, mouse_screen) == win
			  && (i = mouse_y - window_text_top(win)) >= 0 && i < cnt)
			hline = spell_correct_move_highlight(-1, win, i + 1,
				IS_MOUSE_DOUBLE(key));
		if (tolower(key) == 'g')
			spell_ignore_this_word(SPELLB_GLOBAL, word, obuf);
		else if (tolower(key) == 'r' && mult) {
			spell_ignore_word_this_run(word);
			break;
		} else if (tolower(key) == 'd' && have_file) {
			spell_ignore_this_word(SPELLB_PERDIR, word, obuf);
			break;
		} else if (tolower(key) == 'f' && have_file) {
			spell_ignore_this_word(SPELLB_PERFILE, word, obuf);
			break;
		} else if (tolower(key) == 'm')
			spell_ignore_this_word(SPELLB_PERMODE, word, obuf);
		else if (tolower(key) == 'e' && by_exten)
			spell_ignore_this_word(SPELLB_PEREXTEN, word, obuf);
		if (key == CTRL('R') && mult)
			return 1;
		if (IS_ASCII_KEY(key) && strchr("sogm", tolower(key))
			  || key == abort_key || key == ESC)
			break;
		if (isdigit(key) && key != '0' && key - '0' <= wcnt) {
			restore_vars();
			spell_get_suggestion_line(sugg, key, word);
			point = wstart;
			delete(wstart, wend);
			stuff(word);
			return 0;
		} else if (IS_ALT_KEY(key) && isdigit(i = KEY_PLAIN(key))
				   && i != '0' && i - '0' <= wcnt) {
			spell_get_suggestion_line(sugg, i, word);
			copy_line_to_clipboard(word, 0);
		}
	}
	this_cmd = SPELL_CORRECT;
	return 0;
}

command spell_correct() on cc_tab[CTRL('O')]
{
	if (load_spell_word_list(0))
		do_spell_correct(0);
}	

// Maybe ignore the word found in this range due to special rules.
int spell_ignore_word(int opt, int from, int to, char *word)
{
	UNREFERENCED_PARAMETER(word);
	if (opt & 0x2)		// Skip words with underscores.
		if (character(from - 1) == '_' || character(to) == '_')
			return 1;
	if (opt & (0x4 | 0x8))	// Skip words by case.
		switch (check_alt_case_replace(from, to)) {
			case 3: case 4:
				if (opt & 0x4)	// Skip MixedCaseWords.
					return 1;
				break;
			case 1:
				if (opt & 0x8)	// Skip all uppercase words.
					return 1;
				break;
		}
	if (opt & 0x10)			// Follows a digit.
		if (isdigit(character(from - 1)))
			  return 1;
	if (opt & 0x20)			// Before a digit.
		if (isdigit(character(to)))
			return 1;
	return 0;
}

// Called during buffer display.  Mark up misspelled words in the
// current window using a region.
vspell_buffer_when_displaying()
{
	int from, to, cnt = 0;
	if (!modified_buffer_region(&from, &to, "vspell")
		  && vspell_data.window_start == window_start
		  && vspell_data.window_end == window_end
		  && vspell_data.display_column == display_column
		  && !vspell_data.skipped_word)
		return;
	vspell_data.window_start = window_start;
	vspell_data.window_end = window_end;
	reset_modified_buffer_region("vspell");
	if (modify_region(SPELL_REGION_HANDLE, MRTYPE, -1) == REGNORM)
		remove_region(SPELL_REGION_HANDLE);
	if (!vspell_data.want)
		return;
	int opos = point;		// Ignore the word just before point when typing.
	char *pat = get_spell_word_pattern();
	save_var narrow_start = window_start
							- matches_at_length(window_start, -1, pat);
	save_var narrow_end = size() - window_end
						  - matches_at_length(window_end, 1, pat);
	save_var point = 0, abort_searching = ABORT_ERROR;
	int opt = get_num_var(vspell_get_mode_variable());
	vspell_data.skipped_word = 0;
	while (re_search(1, pat) == 1) {
		char word[FNAMELEN];
		if (spell_get_and_check_word(opt, word))
			continue;
		int skey = key;
		skey = (skey == GREYBACK) ? '\b' : IS_ASCII_KEY(skey) ? skey : '/';
		if (is_current_window && prev_cmd != SPELL_CORRECT && matchstart <= opos
			  && opos <= matchend + (strchr("'", skey) != 0)
			  && (isalnum(skey) || (strchr("_'", skey)) || skey == '\b')) {
			vspell_data.skipped_word = 1;	// Catch it next time.
			continue;
		}
		if (cnt++ > 2000)	// More than this many misspellings in the window?
			break;			// Must be something wrong; bail out.
		spot start = alloc_spot(0), end = alloc_spot(1);
		*start = point - strlen(word);
		*end = point;
		add_region(start, end, color_class spell_misspelled, REGNORM,
				   SPELL_REGION_HANDLE);
		if (display_column >= 0)	// Don't check words we can't see.
			if (current_column() - strlen(word) > display_column + window_width)
				nl_forward();
			else if (current_column() < display_column - 50)
				move_to_column(display_column - 50);
	}
	modify_region(SPELL_REGION_HANDLE, MRAUTODEL, 1);
}

// Return index of buffer variable for this mode.
int vspell_get_mode_variable()
{
	return get_mode_based_index("%s-spell-options");
}

show_minor_mode_spell(char *m)
{
	if (vspell_data.want)
		strcpy(m, "Sp");
}


// Is spell mode permitted in this buffer?  Disable for system ones.
// Must at least be off for our word list files.
int spell_mode_ok(int buf)
{
	save_var bufnum = buf;
	if (!strcmp(bufname, "#diff#"))
		return 1;
	return !strchr("-#", *bufname);
}

// Called whenever the mode line changes.  Maybe enable spell by
// applying this mode's rules to buffer.
do_when_make_mode_vspell()
{
	if (use_default)
		return;
	vspell_data.want = get_num_var(vspell_get_mode_variable()) & 1;
	if (vspell_data.buf_want)
		vspell_data.want = (vspell_data.buf_want == 1);
	if (vspell_data.last_want == vspell_data.want)
		return;
	if (!spell_mode_ok(bufnum))
		vspell_data.want = 0;
	vspell_data.last_want = vspell_data.want;
	if (modify_region(SPELL_REGION_HANDLE, MRTYPE, -1) == REGNORM)
		remove_region(SPELL_REGION_HANDLE);	// Remove if it exists.
	if (vspell_data.want) {
		if (load_spell_word_list(0)) {
			add_buffer_when_displaying(bufnum, vspell_buffer_when_displaying);
			vspell_data.window_start = -1;
		} else
			vspell_data.want = vspell_data.last_want = vspell_data.buf_want = 0;
	} else
		delete_buffer_when_displaying(bufnum, vspell_buffer_when_displaying);
}

set_spell_mode(int per_buf)
{
	int old = get_num_var(vspell_get_mode_variable()), want;

	if (vspell_data.buf_want)
		old = (vspell_data.buf_want == 1);
	want = (old & ~1) | (has_arg ? (iter != 0) : (~old & 1));
	if ((want & 1) && !spell_load_file(SPELLB_ORDERED, 0)) {
		if (*get_spell_language_prefix()) {
			say("No dictionary found for \"%s\" language, install one.",
				  get_spell_language_prefix());
			return;
		}
		say("No dictionary found, running spell-configure.");
		spell_configure();
		return;
	}
	if ((want & 1) && !spell_mode_ok(bufnum))
		error("Spell mode not available for this system buffer.");
	if (per_buf)
		vspell_data.buf_want = want ? 1 : 2;
	else {
		set_num_var(vspell_get_mode_variable(), want);
		vspell_data.buf_want = 0;
	}
	do_when_make_mode_vspell();
	save_customization(vspell_get_mode_variable());
	if ((want & 1) != vspell_data.want)
		say("Spell mode not available for this type of buffer.");
	else if (per_buf)
		say("Turning Spell minor mode %s for this buffer.",
			  vspell_data.want ? "on" : "off");
	else
		say("Setting %s to turn Spell minor mode %s.",
			  name_name(vspell_get_mode_variable()),
			  vspell_data.want ? "on" : "off");
	make_mode();
}

command spell_mode()
{
	set_spell_mode(0);
}

command buffer_spell_mode()
{
	set_spell_mode(1);
}

command spell_buffer_or_region()
{
	if (!load_spell_word_list(0))
		return;
	save_spot point, mark;
	spot end = alloc_spot(1);
	on_exit if (spot_to_buffer(end) != -1) free_spot(end);
	on_exit highlight_off();
	if (is_highlight_on()) {
		fix_region();
		*end = mark;
	} else {
		point = 0;
		*end = size();
	}
	on_exit delete_buffer(SPELL_THIS_RUN);
	int opt = get_num_var(vspell_get_mode_variable());
	char *pat = get_spell_word_pattern();
	while (point < *end && re_search(1, pat) == 1) {
		char word[FNAMELEN];
		mark = matchstart;
		if (spell_get_and_check_word(opt, word))
			continue;
		int this_run = name_to_bufnum(SPELL_THIS_RUN);
		if (this_run && word_in_dictionary(this_run, word))
			continue;
		highlight_on();
		if (do_spell_correct(1) == 1) {
			highlight_on();
			show_binding("Continue spell checking by typing %s", "exit-level");
			if (!recursive_edit_preserve())
				break;
		}
	}
	say("Spell checking completed.");
}

command spell_grep()
{
	if (!load_spell_word_list(0))
		return;
	char fname[FNAMELEN], dir[FNAMELEN];
	save_var point, mark;
	int pos = point;
	if (is_highlight_on()) {
		fix_region();
		save_var narrow_start = point, narrow_end = size() - mark;
	}
	int buf = tmp_buf(), cnt = 0, eline = 0, lstart = 0;
	on_exit buf_delete(buf);
	recolor_buffer_range(0, size());
	init_copy_for_grep(fname, dir);
	save_var abort_searching = 0;
	point = 0;
	int opt = get_num_var(vspell_get_mode_variable());
	char *pat = get_spell_word_pattern();
	while (point < size() && re_search(1, pat) == 1
		   && !user_abort) {
		char word[FNAMELEN];
		mark = matchstart;
		if (spell_get_and_check_word(opt, word))
			continue;
		if (point < eline) {
			buf_set_character_color(buf, lstart + mark - give_begin_line(),
									lstart + point - give_begin_line(),
									color_class spell_misspelled);
		} else
			lstart = copy_line_for_grep(fname, buf, mark, point,
										color_class spell_misspelled);
		eline = give_end_line();
		cnt++;
		if (size() > 20000)
			note("%d misspelled words, %d%% complete.",
				 cnt, muldiv(point, 100, size()));
	}
	user_abort = 0;
	if (size() > 20000)
		say("");
	point = pos;
	if (buf_size(buf))
		finish_copy_for_grep(buf, "misspelled words", fname, dir);
	else
		say("No misspelled words in %s.",
			  is_highlight_on() ? "region" : "buffer");
	highlight_off();
}
